#! /usr/bin/perl -wT
################################################################################
#
# MAM Assignment object
#
# File   :  Assignment.pm
#
################################################################################
#
#    This file includes parts which are copyrighted by Adaptive Computing
#    Enterprises, Inc., and parts which are copyrighted by Battelle Memorial
#    Institute.
#
#    The terms and conditions for the use and redistribution of these parts
#    are governed by the Moab Accounting Manager License and the BSD License
#    respectively. Refer to the LICENSE file for details.
#
#    Copyright (C) 2006 - 2016 by Adaptive Computing Enterprises, Inc.
#    All Rights Reserved.
#
#    The Moab Accounting Manager License specifies the terms and conditions
#    for use and redistribution.
#
#    The Moab Accounting Manager License applies to the Moab Accounting
#    Manager software offered by Adaptive Computing Enterprises, Inc.
#    By installing or using this software, Licensee accepts a non-exclusive
#    license from Adaptive Computing Enterprises, Inc. and is bound to accept
#    acknowledgement of and abide by the notices and conditions of the
#    Moab Accounting Manager License.
#
#    Copyright (C) 2003 - 2005 Pacific Northwest National Laboratory,
#    Battelle Memorial Institute. All rights reserved.
#
#    The BSD License specifies the terms and conditions for use and
#    redistribution.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions
#    are met:
#
#    - Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
#    - Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
#    - Neither the name of Battelle nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
#    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
#    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
#    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#    POSSIBILITY OF SUCH DAMAGE.
#
################################################################################

use strict;

=head1 NAME

MAM::Assignment - represents an object property to be assigned

=head1 DESCRIPTION

The B<MAM::Assignment> module defines functions to build and parse Assignments

=head1 CONSTRUCTOR

my $assignment = new MAM::Assignment(name => $name[, value => $value][, op => $op]);

=head1 ACCESSORS

=over 4

=item $name = $assignment->getName();

=item $part = $assignment->getPart();

=item $value = $assignment->getValue();

=item $op = $assignment->getOperator();

=back 

=head1 MUTATORS

=over 4

=item $assignment = $assignment->setName($name);

=item $assignment = $assignment->setPart($part);

=item $assignment = $assignment->setValue($value);

=item $assignment = $assignment->setOperator($op);

=back

=head1 OTHER METHODS

=over 4

=item $string = $assignment->toString();

=back

=head1 EXAMPLES

use MAM::Assignment;

my $assignment1 = new MAM::Assignment(name => 'Active', value => 'True');
my $assignment2 = new MAM::Assignment(name => 'Licenses', part => 'Matlab', value => '2', op => 'Inc');

=head1 REQUIRES

Perl 5.8

=head1 AUTHOR

Scott Jackson, scottmo@adaptivecomputing.com

=cut

##############################################################################

package MAM::Assignment;

# ----------------------------------------------------------------------------
# Constructor
# ----------------------------------------------------------------------------

sub new
{
    my ($class, %arg) = @_;
    bless {
        _name => $arg{name} || '',    # SCALAR
        _part => $arg{part} || '',    # SCALAR
        _value => defined $arg{value} ? $arg{value} : '',    # SCALAR
        _op => $arg{op} || '',                               # SCALAR
    }, $class;
}

# ----------------------------------------------------------------------------
# Accessors
# ----------------------------------------------------------------------------

# Get the assignment name
sub getName
{
    my ($self) = @_;
    return $self->{_name};
}

# Get the assignment part
sub getPart
{
    my ($self) = @_;
    return $self->{_part};
}

# Get the assignment value
sub getValue
{
    my ($self) = @_;
    return $self->{_value};
}

# Get the assignment operator
sub getOperator
{
    my ($self) = @_;
    return $self->{_op};
}

# ----------------------------------------------------------------------------
# Mutators
# ----------------------------------------------------------------------------

# Set the assignment name
sub setName
{
    my ($self, $name) = @_;
    $self->{_name} = $name if defined $name;
    return $self;
}

# Set the assignment part
sub setPart
{
    my ($self, $part) = @_;
    $self->{_part} = $part if defined $part;
    return $self;
}

# Set the assignment value
sub setValue
{
    my ($self, $value) = @_;
    $self->{_value} = $value if defined $value;
    return $self;
}

# Set the assignment operator
sub setOperator
{
    my ($self, $op) = @_;
    $self->{_op} = $op if defined $op;
    return $self;
}

# ----------------------------------------------------------------------------
# Other Methods
# ----------------------------------------------------------------------------

# Convert assignment to printable string
sub toString
{
    my ($self) = @_;
    my $string = '<Set';
    $string .= " name=\"$self->{_name}\"" if $self->{_name} ne '';
    $string .= " part=\"$self->{_part}\"" if $self->{_part} ne '';
    $string .= " op=\"$self->{_op}\"" if $self->{_op} ne '';
    $string .= '>';
    $string .= $self->{_value} if $self->{_value} ne '';
    $string .= '</Set>';
    return $string;
}

1;
