#! /usr/bin/perl -wT
################################################################################
#
# MAM Base object
#
# File   :  Base.pm
#
################################################################################
#
#    This file includes parts which are copyrighted by Adaptive Computing
#    Enterprises, Inc., and parts which are copyrighted by Battelle Memorial
#    Institute.
#
#    The terms and conditions for the use and redistribution of these parts
#    are governed by the Moab Accounting Manager License and the BSD License
#    respectively. Refer to the LICENSE file for details.
#
#    Copyright (C) 2006 - 2016 by Adaptive Computing Enterprises, Inc.
#    All Rights Reserved.
#
#    The Moab Accounting Manager License specifies the terms and conditions
#    for use and redistribution.
#
#    The Moab Accounting Manager License applies to the Moab Accounting
#    Manager software offered by Adaptive Computing Enterprises, Inc.
#    By installing or using this software, Licensee accepts a non-exclusive
#    license from Adaptive Computing Enterprises, Inc. and is bound to accept
#    acknowledgement of and abide by the notices and conditions of the
#    Moab Accounting Manager License.
#
#    Copyright (C) 2003 - 2005 Pacific Northwest National Laboratory,
#    Battelle Memorial Institute. All rights reserved.
#
#    The BSD License specifies the terms and conditions for use and
#    redistribution.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions
#    are met:
#
#    - Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
#    - Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
#    - Neither the name of Battelle nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
#    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
#    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
#    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#    POSSIBILITY OF SUCH DAMAGE.
#
################################################################################

use strict;

=head1 NAME

MAM::Base - implements base actions

=head1 DESCRIPTION

The B<MAM::Base> module handles base object-oriented actions through class methods such as create, query, modify, delete, undelete, undo, redo, usage, etc.

=head1 METHODS

=over 4

=item $response = MAM::Base->create($request, $requestId[, $noLog]);

=item $response = MAM::Base->delete($request, $requestId[, $noLog]);

=item $response = MAM::Base->execute($proxy);

=item $response = MAM::Base->fire($request, $requestId);

=item $response = MAM::Base->modify($request, $requestId[, $noLog]);

=item $response = MAM::Base->prune($request, $requestId);

=item $response = MAM::Base->query($request, $requestId[, $chunkNum]);

=item $response = MAM::Base->redo($request, $requestId);

=item $response = MAM::Base->refresh($request[, $requestId]);

=item $response = MAM::Base->send($request, $requestId);

=item $response = MAM::Base->undelete($request, $requestId);

=item $response = MAM::Base->undo($request, $requestId);

=item $response = MAM::Base->usage($request);

=back

=head1 EXAMPLES

use MAM::Base;

my $response = MAM::Base->execute($proxy);

my $response = MAM::Base->query($request, $requestId);

my $response = MAM::Base->modify($request, $requestId);

=head1 REQUIRES

Perl 5.8

=head1 AUTHOR

Scott Jackson, scottmo@adaptivecomputing.com

=cut

##############################################################################

package MAM::Base;

use vars qw($log);
use Crypt::CBC;
use Date::Manip;
use DBI qw( :sql_types );
use Error qw(:try);
use MIME::Base64;
use Time::Local;
use MAM::Cache;
use MAM::Global;
use MAM::Request;
use MAM::Response;

# ----------------------------------------------------------------------------
# $response = execute($proxy);
# ----------------------------------------------------------------------------

# Performs a base action
sub execute
{
    my ($class, $proxy) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $request      = $proxy->getRequest();
    my $action       = $request->getAction();
    my $requestId    = $proxy->getRequestId();
    my $lastResponse = $proxy->getResponse();
    my $response;

    # Provide usage if help flag specified

    my $showUsage = $request->getOptionValue('ShowUsage');
    if (defined $showUsage && $showUsage =~ /^true$/i)
    {
        # Call the usage for the base action
        $response = MAM::Base->usage($request, $requestId);
    }

    # Otherwise try to perform the base action

    # Create
    elsif ($action eq 'Create')
    {
        $response = MAM::Base->create($request, $requestId);
    }

    # Query
    elsif ($action eq 'Query')
    {
     # If this is a continuation of a chunked response pass in next chunk number
        if ($lastResponse && $lastResponse->getChunkMax() == -1)
        {
            $response =
              MAM::Base->query($request, $requestId,
                $lastResponse->getChunkNum() + 1);
        }
        else
        {
            $response = MAM::Base->query($request, $requestId);
        }
    }

    # Modify
    elsif ($action eq 'Modify')
    {
        $response = MAM::Base->modify($request, $requestId);
    }

    # Delete
    elsif ($action eq 'Delete')
    {
        $response = MAM::Base->delete($request, $requestId);
    }

    # Undelete
    elsif ($action eq 'Undelete')
    {
        $response = MAM::Base->undelete($request, $requestId);
    }

    # Refresh
    elsif ($action eq 'Refresh')
    {
        $response = MAM::Base->refresh($request, $requestId);
    }

    # Fire
    elsif ($action eq 'Fire')
    {
        $response = MAM::Base->fire($request, $requestId);
    }

    # Send
    elsif ($action eq 'Send')
    {
        $response = MAM::Base->send($request, $requestId);
    }

    # Undo
    elsif ($action eq 'Undo')
    {
        $response = MAM::Base->undo($request, $requestId);
    }

    # Redo
    elsif ($action eq 'Redo')
    {
        $response = MAM::Base->redo($request, $requestId);
    }

    # Prune
    elsif ($action eq 'Prune')
    {
        $response = MAM::Base->prune($request, $requestId);
    }

    else
    {
        $log->error("Unsupported action: ($action)");
        $response = new MAM::Response()
          ->failure('313', "Unsupported base action: ($action)");
    }

    return $response;
}

# ----------------------------------------------------------------------------
# $response = create($request, $requestId, $noLog);
# ----------------------------------------------------------------------------

# Create (ANY)
sub create
{
    my ($class, $request, $requestId, $noLog) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object      = $request->getObject();
    my $actor       = $request->getActor();
    my %assignments = map { ${_}->getName(), $_ } $request->getAssignments();
    my $database    = $request->getDatabase();
    my @options     = $request->getOptions();
    my $now         = $requestTime;
    my $warnings    = 0;
    my $dbh         = $database->getHandle();
    my $txnId       = $database->nextId('Transaction');
    my $sql;
    my $response;

    # Set Event ArmTime if FireTime specified
    if (   $object eq 'Event'
        && defined $assignments{FireTime}
        && ! defined $assignments{ArmTime})
    {
        $assignments{ArmTime} =
          new MAM::Assignment(name => 'ArmTime', value => $now);
    }

    # First check if object or association already exists (even if deleted)
    my @primaryKeys         = ();
    my @undeleteConditions  = ();
    my @undeleteAssignments = ();
# Select Name,DataType,PrimaryKey,Fixed FROM Attribute WHERE Object=$object AND Hidden=False AND DataType!=AutoGen
    my $results = $database->select(
        object     => 'Attribute',
        selections => [
            new MAM::Selection(name => 'Name'),
            new MAM::Selection(name => 'DataType'),
            new MAM::Selection(name => 'PrimaryKey'),
            new MAM::Selection(name => 'Fixed')
        ],
        conditions => [
            new MAM::Condition(name => 'Object', value => $object),
            new MAM::Condition(name => 'Hidden', value => 'False'),
            new MAM::Condition(
                name  => 'DataType',
                value => 'AutoGen',
                op    => 'NE'
            )
        ],
        chunkNum => 0
    );
    foreach my $attributeRow (@{$results->{data}})
    {
        my $name       = $attributeRow->[0];
        my $dataType   = $attributeRow->[1];
        my $primaryKey = $attributeRow->[2];
        my $fixed      = $attributeRow->[3];
        if ($primaryKey eq 'True')
        {
            push @primaryKeys, $name;
            push @undeleteConditions,
              new MAM::Condition(
                name  => $name,
                value => $request->getAssignmentValue($name)
              );
        }
        elsif ($fixed ne 'True')
        {
            my $value = $request->getAssignmentValue($name);
            if (defined $value)
            {
                push @undeleteAssignments,
                  new MAM::Assignment(name => $name, value => $value);
            }
            else
            {
                push @undeleteAssignments,
                  new MAM::Assignment(name => $name, value => 'NULL');
            }
        }
    }
    my @conditions = @undeleteConditions;
    push @conditions,
      new MAM::Condition(name => 'Deleted', value => 'True', group => '+1');
    push @conditions,
      new MAM::Condition(
        name  => 'Deleted',
        value => 'False',
        conj  => 'Or',
        group => '-1'
      );
    if (@primaryKeys)
    {
# SELECT Deleted from $object WHERE (Deleted=True OR Deleted=False) AND (build conditions from all assignments which are primary keys)
        my $results = $database->select(
            object     => $object,
            selections => [new MAM::Selection(name => 'Deleted')],
            conditions => \@conditions,
            chunkNum   => 0
        );
        if (@{$results->{data}})
        {
            # $object exists and is not deleted
            if (${$results->{data}}[0]->[0] eq 'False')
            {
                return new MAM::Response()
                  ->failure('740', "$object already exists");
            }

          # $object is deleted. We need to undelete it and update its attributes
            else
            {
                # Perform the undelete
                my $count = $database->undelete(
                    object     => $object,
                    actor      => $actor,
                    conditions => \@undeleteConditions,
                    requestId  => $requestId,
                    txnId      => $txnId
                );
                $count = $count eq '0E0' ? 0 : $count;

                # Modify the undeleted object to have the new values
                if (@undeleteAssignments)
                {
                    my $subModifyRequest = new MAM::Request(
                        database    => $database,
                        object      => $object,
                        action      => 'Modify',
                        assignments => \@undeleteAssignments,
                        conditions  => \@undeleteConditions,
                        options     => \@options
                    );
                    my $proxy = new MAM::Proxy(
                        database  => $database,
                        request   => $subModifyRequest,
                        requestId => $requestId
                    );
                    my $subModifyResponse = $proxy->execute();
                    my $updateCount       = $subModifyResponse->getCount();
                }

                # Query undeleted objects
                my $request = new MAM::Request(
                    database   => $database,
                    object     => $object,
                    action     => 'Query',
                    conditions => [
                        new MAM::Condition(
                            name  => 'TransactionId',
                            value => $txnId
                        )
                    ],
                    options => \@options
                );
                MAM::Proxy->prepareSelections($request);
                my $response = MAM::Base->query($request, $requestId, 0);
                my @data     = $response->getData();
                my %metadata = $response->getMetadata();
                my $message =
                    "Successfully created $count "
                  . lower($object)
                  . ($count == 1 ? '' : 's');

                return new MAM::Response()
                  ->success($count, \@data, \%metadata, $message);
            }
        }
    }

    # If creating a new object type, we need to create a new table
    if ($object eq 'Object')
    {
        my $name = $request->getAssignmentValue('Name');
        $database->createTable(
            object    => $name,
            requestId => $requestId,
            txnId     => $txnId
        );
    }

    # If creating a new attribute, we need to add a column to a table
    elsif ($object eq 'Attribute')
    {
        my $obj  = $request->getAssignmentValue('Object');
        my $name = $request->getAssignmentValue('Name');
        # primarykey might be used later to set NOT NULL

        # If primarykey then should be required (or defaultvalue) and fixed
        my $primaryKey   = $request->getAssignmentValue('PrimaryKey');
        my $required     = $request->getAssignmentValue('Required');
        my $fixed        = $request->getAssignmentValue('Fixed');
        my $defaultValue = $request->getAssignmentValue('DefaultValue');
        if (defined $primaryKey && $primaryKey =~ /True/i)
        {
            if (defined $fixed && $fixed !~ /True/i)
            {
                delete $assignments{'Fixed'};
                $assignments{'Fixed'} =
                  new MAM::Assignment(name => 'Fixed', value => 'True');
            }
            if (   defined $required
                && $required !~ /True/i
                && ! defined $defaultValue)
            {
                delete $assignments{'Required'};
                $assignments{'Required'} =
                  new MAM::Assignment(name => 'Required', value => 'True');
            }
        }

        # Calculate new attribute sequence if not specified
        if (! defined $request->getAssignmentValue('Sequence'))
        {
            my $maxSequence = 0;
            my $results     = $database->select(
                object     => 'Attribute',
                selections => [new MAM::Selection(name => 'Sequence')],
                conditions =>
                  [new MAM::Condition(name => 'Object', value => $obj)]
            );
            foreach my $attributeRow (@{$results->{data}})
            {
                my $sequence = $attributeRow->[0];
                if ($sequence > $maxSequence && $sequence < 900)
                {
                    $maxSequence = $sequence;
                }
            }
            my $newSequence = (($maxSequence / 10) * 10) + 10;
            $assignments{'Sequence'} =
              new MAM::Assignment(name => 'Sequence', value => $newSequence);
        }

        # Add a new key generator row if this is autogenerated field
        # Initialize key generator for this object
        # Only one autogen field is allowed per object
        my $dataType = $request->getAssignmentValue('DataType');
        if (defined $dataType && $dataType eq 'AutoGen')
        {
            $sql =
              "INSERT INTO g_key_generator (g_name,g_next_id) VALUES ('$obj',1)";
            if ($log->is_trace())
            {
                $log->trace("SQL Update: $sql");
            }
            $dbh->do($sql);
        }

        # Add columns to table
        $database->addColumn(
            request   => $request,
            object    => $obj,
            attribute => $name,
            requestId => $requestId,
            txnId     => $txnId
        );
        $database->addColumn(
            request   => $request,
            object    => $obj . 'Log',
            attribute => $name,
            requestId => $requestId,
            txnId     => $txnId
        );
    }

    # If creating a new password, need to encrypt it with symmetric key
    elsif ($object eq 'Password')
    {
        # Decrypt password with symmetric key
        my $password        = $assignments{'Password'}->getValue();
        my $promotionMethod = $clientConfig->{'security.promotion'}
          || $SECURITY_PROMOTION;

        ## Promotion method is mamauth
        #if ($promotionMethod eq 'mamauth')
        #{
        #}
        #
        ## Promotion method is suidperl
        #elsif ($promotionMethod eq 'suidperl')

        # Promotion method is suidperl or mamauth
        if ($promotionMethod eq 'suidperl' || $promotionMethod eq 'mamauth')
        {
            my $symmetricKey = $localSiteConfig->{'token.value'};
            my $key = pack('a24', $symmetricKey);
            if (length $password < 8)
            {
                return new MAM::Response()
                  ->failure('740',
                    'The password must be at least 8 characters');
            }
            # DES input must be 8 bytes long
            utf8::encode($password);
            my $cipher = new Crypt::CBC(
                {   key    => $key,
                    cipher => 'Crypt::DES_EDE3',
                    header => 'randomiv',         # No longer default as of 2.17
                    regenerate_key => 0,
                    padding        => 'standard',
                    prepend_iv     => 1,
                }
            );
            my $cipherPayload =
              substr($cipher->encrypt($password), 8);    # Remove 'RandomIV'
            my $encryptedPassword = encode_base64($cipherPayload, '');
            $assignments{'Password'} = new MAM::Assignment(
                name  => 'Password',
                value => $encryptedPassword
            );
        }
    }

    # Mark metadata as stale if necessary
    if (   $object eq 'Object'
        || $object eq 'Attribute'
        || $object eq 'Action'
        || $object eq 'Role'
        || $object eq 'RoleAction'
        || $object eq 'RoleUser'
        || $object eq 'Password')
    {
        MAM::Cache->setStale(1);
    }

    # Perform the insert
    my @assignments = values %assignments;
    my %args        = (
        object      => $object,
        actor       => $actor,
        assignments => \@assignments,
        options     => \@options,
        requestId   => $requestId,
        txnId       => $txnId,
    );
    $args{noLog} = $noLog if defined $noLog;
    my $count = $database->insert(%args);

    my $message =
        "Successfully created $count "
      . lower($object)
      . ($count == 1 ? '' : 's');

    # Query created object
    my $subQueryRequest = new MAM::Request(
        database => $database,
        object   => $object,
        action   => 'Query',
        conditions =>
          [new MAM::Condition(name => 'TransactionId', value => $txnId)],
        options => \@options
    );
    MAM::Proxy->prepareSelections($subQueryRequest);
    my $subQueryResponse = MAM::Base->query($subQueryRequest, $requestId, 0);
    my @data             = $subQueryResponse->getData();
    my %metadata         = $subQueryResponse->getMetadata();

    # If creating a new object type, we need to add the default actions
    # and the default attributes, and create the indexes
    if ($object eq 'Object')
    {
        my $name = $request->getAssignmentValue('Name');

        # Add the default actions (Create, Delete, Modify, Query, Undelete)
        my $actionCount = 0;
        foreach my $action ('Create', 'Delete', 'Modify', 'Query', 'Undelete')
        {
            my @actionAssignments = (
                new MAM::Assignment(name => 'Object',      value => $name),
                new MAM::Assignment(name => 'Name',        value => $action),
                new MAM::Assignment(name => 'Description', value => $action),
            );
            push @actionAssignments,
              new MAM::Assignment(name => 'Display', value => 'False')
              if $action eq 'Undelete';
            my $subCreateRequest = new MAM::Request(
                database    => $database,
                object      => 'Action',
                action      => 'Create',
                assignments => \@actionAssignments,
            );
            my $proxy = new MAM::Proxy(
                database  => $database,
                request   => $subCreateRequest,
                requestId => $requestId
            );
            my $subCreateResponse = $proxy->execute();
            $actionCount += $subCreateResponse->getCount();
        }
        $message .= ", 5 attributes and $actionCount Actions"
          if $actionCount > 0;
    }

    # Set default parent if specified
    my $defaultParent = $request->getOptionValue('DefaultParent');
    if (   defined $defaultParent
        && $defaultParent =~ /^True$/i
        && MAM::Cache->getObjectProperty($object, 'Association') eq 'True')
    {
        my $parentObject = MAM::Cache->getObjectProperty($object, 'Parent');
        my $childObject  = MAM::Cache->getObjectProperty($object, 'Child');

# Select Name FROM Attribute WHERE Object=$childObject AND Values="@?=$parentObject"
        my $results = $database->select(
            object     => 'Attribute',
            selections => [new MAM::Selection(name => 'Name')],
            conditions => [
                new MAM::Condition(name => 'Object', value => $childObject),
                new MAM::Condition(
                    name  => 'Values',
                    value => "@?=$parentObject"
                ),
            ]
        );
        if (@{$results->{data}})
        {
            my $defaultName = ${$results->{data}}[0]->[0];
            my $parentValue = $assignments{$parentObject}->getValue();
            my $childValue  = $assignments{'Name'}->getValue();

            # Set the parent default
            my $subRequest = new MAM::Request(
                database    => $database,
                object      => $childObject,
                action      => 'Modify',
                actor       => $SUPER_USER,
                assignments => [
                    new MAM::Assignment(
                        name  => $defaultName,
                        value => $parentValue
                    )
                ],
                conditions => [
                    new MAM::Condition(
                        name  => 'Name',
                        value => $childValue
                    )
                ]
            );
            my $subResponse = MAM::Base->modify($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Failure')
            {
                $warnings++;
                $message .=
                  "\nWarning -- Unable to set default $parentObject ($parentValue) for $childObject ($childValue): "
                  . $subResponse->getMessage();
            }
        }
        else
        {
            $warnings++;
            $message .=
              "\nWarning -- No default $parentObject attribute found for $childObject";
        }
    }

    if ($warnings)
    {
        $response =
          new MAM::Response()->warning($count, \@data, \%metadata, $message);
    }
    else
    {
        $response =
          new MAM::Response()->success($count, \@data, \%metadata, $message);
    }
    return $response;
}

# ----------------------------------------------------------------------------
# $response = query($request, $requestId[, $chunkNum]);
# ----------------------------------------------------------------------------

# Query (ANY)
sub query
{
    my ($class, $request, $requestId, $chunkNum) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my @objects    = $request->getObjects();
    my $object     = join '_', map ${_}->getName(), @objects;
    my $action     = $request->getAction();
    my $chunkSize  = $request->getChunkSize();
    my @selections = $request->getSelections();
    my @conditions = $request->getConditions();
    my @options    = $request->getOptions();
    my $database   = $request->getDatabase();
    my $count      = 0;
    my @data       = ();
    my @messages   = ();
    my $warnings   = 0;
    $chunkNum = 1 unless defined $chunkNum;
    my $response;

    # Perform the query
    my $results = $database->select(
        objects    => \@objects,
        selections => \@selections,
        conditions => \@conditions,
        options    => \@options,
        chunkNum   => $chunkNum,
        chunkSize  => $chunkSize
    );
    my $data     = $results->{data};
    my $cols     = $results->{cols};
    my $names    = $results->{names};
    my $rows     = $results->{rows};
    my %metadata = %{$results->{metadata}};

    # Populate Data
    foreach my $row (@{$data})
    {
        $count++;
        my $datum = new MAM::Datum($object);
        for (my $i = 0; $i < $cols; $i++)
        {
            my $name  = $names->[$i];
            my $value = $row->[$i];     # Null comes in as undef
            $datum->addProperty($name, $value);
        }
        push @data, $datum;
    }

    # Delete notifications if Delete option is True
    if ($object eq 'Notification')
    {
        my $deleteOption = $request->getOptionValue('Delete');
        my $recursive    = $request->getOptionValue('Recursive');

        if (   defined $deleteOption
            && $deleteOption eq 'True'
            && ((! defined $recursive) || $recursive ne 'True'))
        {
            # DELETE FROM Notification WHERE @conditions
            my $subRequest = new MAM::Request(
                database   => $database,
                object     => 'Notification',
                action     => 'Delete',
                actor      => $SUPER_USER,
                conditions => \@conditions,
                options =>
                  [new MAM::Option(name => 'Recursive', value => 'True'),],
            );
            my $subResponse = MAM::Base->delete($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Failure')
            {
                push @messages,
                  'Warning -- Unable to delete queried notifications: '
                  . $subResponse->getMessage();
                $warnings++;
            }
            if ($subResponse->getCount() > 0)
            {
                push @messages, $subResponse->getMessage();
            }
        }
    }

    if ($warnings)
    {
        $response = new MAM::Response()->warning($count, \@data, \%metadata);
    }
    else
    {
        $response = new MAM::Response()->success($count, \@data, \%metadata);
    }
    $response->setMessage(join "\n", @messages) if @messages;

    # Set chunkNum and chunkMax appropriately
    $response->setChunkNum($chunkNum);
    if ($rows >= $request->getChunkSize())
    {
        $response->setChunkMax(-1);
    }

    return $response;
}

# ----------------------------------------------------------------------------
# $response = modify($request, $requestId, $noLog);
# ----------------------------------------------------------------------------

# Modify (ANY)
sub modify
{
    my ($class, $request, $requestId, $noLog) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object      = $request->getObject();
    my $actor       = $request->getActor();
    my %assignments = map { ${_}->getName(), $_ } $request->getAssignments();
    my @conditions  = $request->getConditions();
    my @options     = $request->getOptions();
    my $database    = $request->getDatabase();
    my $now         = $requestTime;
    my $txnId       = $database->nextId('Transaction');

    # Set Event ArmTime if FireTime specified
    if (   $object eq 'Event'
        && defined $assignments{FireTime}
        && ! defined $assignments{ArmTime})
    {
        $assignments{ArmTime} =
          new MAM::Assignment(name => 'ArmTime', value => $now);
    }

    # If modifying a password, need to encrypt it with symmetric key
    if ($object eq 'Password')
    {
        # Decrypt password with symmetric key
        my $password        = $assignments{'Password'}->getValue();
        my $promotionMethod = $clientConfig->{'security.promotion'}
          || $SECURITY_PROMOTION;

        ## Promotion method is mamauth
        #if ($promotionMethod eq 'mamauth')
        #{
        #}
        #
        ## Promotion method is suidperl
        #elsif ($promotionMethod eq 'suidperl')

        # Promotion method is suidperl or mamauth
        if ($promotionMethod eq 'suidperl' || $promotionMethod eq 'mamauth')
        {
            my $symmetricKey = $localSiteConfig->{'token.value'};
            my $key = pack('a24', $symmetricKey);
            if (length $password < 8)
            {
                return new MAM::Response()
                  ->failure('740',
                    'The password must be at least 8 characters');
            }
            # DES input must be 8 bytes long
            utf8::encode($password);
            my $cipher = new Crypt::CBC(
                {   key    => $key,
                    cipher => 'Crypt::DES_EDE3',
                    header => 'randomiv',         # No longer default as of 2.17
                    regenerate_key => 0,
                    padding        => 'standard',
                    prepend_iv     => 1,            # Prepends RandomIV.{8}
                }
            );
            my $cipherPayload =
              substr($cipher->encrypt($password), 8);    # Remove 'RandomIV'
            my $encryptedPassword = encode_base64($cipherPayload, '');
            $assignments{'Password'} = new MAM::Assignment(
                name  => 'Password',
                value => $encryptedPassword
            );
        }
    }

    # Mark metadata as stale if necessary
    if (   $object eq 'Object'
        || $object eq 'Attribute'
        || $object eq 'Action'
        || $object eq 'Role'
        || $object eq 'RoleAction'
        || $object eq 'RoleUser'
        || $object eq 'Password')
    {
        MAM::Cache->setStale(1);
    }

    # Perform the update
    my @assignments = values %assignments;
    my %args        = (
        object      => $object,
        actor       => $actor,
        assignments => \@assignments,
        conditions  => \@conditions,
        options     => \@options,
        requestId   => $requestId,
        txnId       => $txnId,
    );
    $args{noLog} = $noLog if defined $noLog;
    my $count = $database->update(%args);

    $count = $count eq '0E0' ? 0 : $count;
    my $message =
        "Successfully modified $count "
      . lower($object)
      . ($count == 1 ? '' : 's');

    # Query modified objects
    my $subQueryRequest = new MAM::Request(
        database => $database,
        object   => $object,
        action   => 'Query',
        conditions =>
          [new MAM::Condition(name => 'TransactionId', value => $txnId)],
        options => \@options
    );
    MAM::Proxy->prepareSelections($subQueryRequest);
    my $subQueryResponse = MAM::Base->query($subQueryRequest, $requestId, 0);
    my @data             = $subQueryResponse->getData();
    my %metadata         = $subQueryResponse->getMetadata();

    return new MAM::Response()->success($count, \@data, \%metadata, $message);
}

# ----------------------------------------------------------------------------
# $response = delete($request, $requestId, $noLog);
# ----------------------------------------------------------------------------

# Delete (ANY)
sub delete
{
    my ($class, $request, $requestId, $noLog) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object     = $request->getObject();
    my $actor      = $request->getActor();
    my @conditions = $request->getConditions();
    my @options    = $request->getOptions();
    my $database   = $request->getDatabase();
    my $txnId      = $database->nextId('Transaction');

    # Mark metadata as stale if necessary
    if (   $object eq 'Object'
        || $object eq 'Attribute'
        || $object eq 'Action'
        || $object eq 'Role'
        || $object eq 'RoleAction'
        || $object eq 'RoleUser'
        || $object eq 'Password')
    {
        MAM::Cache->setStale(1);
    }

    # Query objects to be deleted
    my $preQueryRequest = new MAM::Request(
        database   => $database,
        object     => $object,
        action     => 'Query',
        conditions => \@conditions,
        options    => \@options
    );
    MAM::Proxy->prepareSelections($preQueryRequest);
    my $preQueryResponse = MAM::Base->query($preQueryRequest, $requestId, 0);
    my @preQueryData = $preQueryResponse->getData();

    # build a list of primary keys for the object
    my @primaryKeys = ();
    my $results     = $database->select(
        object     => 'Attribute',
        selections => [new MAM::Selection(name => 'Name')],
        conditions => [
            new MAM::Condition(name => 'Object',     value => $object),
            new MAM::Condition(name => 'PrimaryKey', value => 'True')
        ]
    );
    foreach my $attributeRow (@{$results->{data}})
    {
        my $name = $attributeRow->[0];
        push @primaryKeys, $name;
    }

    # Delete dependent associations
    my $associationCount = 0;
    $results = $database->select(
        object     => 'Object',
        selections => [
            new MAM::Selection(name => 'Name'),
            new MAM::Selection(name => 'Parent'),
            new MAM::Selection(name => 'Child')
        ],
        conditions => [
            new MAM::Condition(name => 'Association', value => 'True'),
            new MAM::Condition(
                name  => 'Parent',
                value => $object,
                conj  => 'And',
                group => '+1'
            ),
            new MAM::Condition(
                name  => 'Child',
                value => $object,
                conj  => 'Or',
                group => '-1'
            )
        ]
    );
    foreach my $objectRow (@{$results->{data}})
    {
        my $obj    = $objectRow->[0];
        my $parent = $objectRow->[1];
        my $child  = $objectRow->[2];

        # Iterate over each deleted object
        foreach my $datum (@preQueryData)
        {
            if ($parent eq $object)
            {
                my @conditions = ();
                foreach my $key (@primaryKeys)
                {
                    push @conditions,
                      new MAM::Condition(
                        name  => $parent,
                        value => $datum->getPropertyValue($key)
                      );
                }

                # Recursively delete associations in which object was the parent
                my $subDeleteRequest = new MAM::Request(
                    database   => $database,
                    object     => $obj,
                    action     => 'Delete',
                    actor      => $SUPER_USER,
                    conditions => \@conditions,
                    options    => \@options
                );
                my $proxy = new MAM::Proxy(
                    database  => $database,
                    request   => $subDeleteRequest,
                    requestId => $requestId
                );
                my $subDeleteResponse = $proxy->execute();
                $associationCount += $subDeleteResponse->getCount();
            }
            elsif ($child eq $object)
            {
                my @conditions = ();
                foreach my $key (@primaryKeys)
                {
                    push @conditions,
                      new MAM::Condition(
                        name  => $key,
                        value => $datum->getPropertyValue($key)
                      );
                }

                # Recursively delete associations in which object was the child
                my $subDeleteRequest = new MAM::Request(
                    database   => $database,
                    object     => $obj,
                    action     => 'Delete',
                    actor      => $SUPER_USER,
                    conditions => \@conditions,
                    options    => \@options
                );
                my $proxy = new MAM::Proxy(
                    database  => $database,
                    request   => $subDeleteRequest,
                    requestId => $requestId
                );
                my $subDeleteResponse = $proxy->execute();
                $associationCount += $subDeleteResponse->getCount();
            }
        }
    }

    # Delete associated Password if User is being deleted
    if ($object eq 'User')
    {
        # Iterate over each deleted object
        foreach my $datum (@preQueryData)
        {
            # Delete Password
            my $subDeleteRequest = new MAM::Request(
                database   => $database,
                object     => 'Password',
                action     => 'Delete',
                conditions => [
                    new MAM::Condition(
                        name  => 'User',
                        value => $datum->getPropertyValue('Name')
                    )
                ],
                options => \@options
            );
            my $proxy = new MAM::Proxy(
                database  => $database,
                request   => $subDeleteRequest,
                requestId => $requestId
            );
            my $subDeleteResponse = $proxy->execute();
            $associationCount += $subDeleteResponse->getCount();
        }
    }

    # Perform the base delete
    my %args = (
        object     => $object,
        actor      => $actor,
        conditions => \@conditions,
        options    => \@options,
        requestId  => $requestId,
        txnId      => $txnId,
    );
    $args{noLog} = $noLog if defined $noLog;
    my $count = $database->delete(%args);

    # Query deleted objects
    my $postQueryRequest = new MAM::Request(
        database   => $database,
        object     => $object,
        action     => 'Query',
        conditions => [
            new MAM::Condition(name => 'TransactionId', value => $txnId),
            new MAM::Condition(name => 'Deleted',       value => 'True')
        ],
        options => \@options
    );
    MAM::Proxy->prepareSelections($postQueryRequest);
    my $postQueryResponse = MAM::Base->query($postQueryRequest, $requestId, 0);
    my @postData          = $postQueryResponse->getData();
    my %postMetadata      = $postQueryResponse->getMetadata();

    $count = $count eq '0E0' ? 0 : $count;
    my $message =
        "Successfully deleted $count "
      . lower($object)
      . ($count == 1 ? '' : 's');
    if ($associationCount)
    {
        $message .= " and $associationCount association"
          . ($associationCount == 1 ? '' : 's');
    }

    return new MAM::Response()
      ->success($count, \@postData, \%postMetadata, $message);
}

# ----------------------------------------------------------------------------
# $response = undelete($request, $requestId);
# ----------------------------------------------------------------------------

# Undelete (ANY)
sub undelete
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object     = $request->getObject();
    my $actor      = $request->getActor();
    my @conditions = $request->getConditions();
    my @options    = $request->getOptions();
    my $database   = $request->getDatabase();
    my $txnId      = $database->nextId('Transaction');

    # Mark metadata as stale if necessary
    if (   $object eq 'Object'
        || $object eq 'Attribute'
        || $object eq 'Action'
        || $object eq 'Role'
        || $object eq 'RoleAction'
        || $object eq 'RoleUser'
        || $object eq 'Password')
    {
        MAM::Cache->setStale(1);
    }

    # Query objects to be undeleted
    my @preConditions = @conditions;
    push @preConditions, new MAM::Condition(name => 'Deleted', value => 'True');
    my $preQueryRequest = new MAM::Request(
        database   => $database,
        object     => $object,
        action     => 'Query',
        conditions => \@preConditions
    );
    my $preQueryResponse = MAM::Base->query($preQueryRequest, $requestId, 0);
    my @preQueryData = $preQueryResponse->getData();

    # build a list of primary keys for the object
    my @primaryKeys = ();
    my $results     = $database->select(
        object     => 'Attribute',
        selections => [new MAM::Selection(name => 'Name')],
        conditions => [
            new MAM::Condition(name => 'Object',     value => $object),
            new MAM::Condition(name => 'PrimaryKey', value => 'True')
        ]
    );
    foreach my $attributeRow (@{$results->{data}})
    {
        my $name = $attributeRow->[0];
        push @primaryKeys, $name;
    }

    # Undelete dependent associations
    my $associationCount = 0;
    $results = $database->select(
        object     => 'Object',
        selections => [
            new MAM::Selection(name => 'Name'),
            new MAM::Selection(name => 'Parent'),
            new MAM::Selection(name => 'Child')
        ],
        conditions => [
            new MAM::Condition(name => 'Association', value => 'True'),
            new MAM::Condition(
                name  => 'Parent',
                value => $object,
                conj  => 'And',
                group => '+1'
            ),
            new MAM::Condition(
                name  => 'Child',
                value => $object,
                conj  => 'Or',
                group => '-1'
            )
        ]
    );
    foreach my $objectRow (@{$results->{data}})
    {
        my $obj    = $objectRow->[0];
        my $parent = $objectRow->[1];
        my $child  = $objectRow->[2];

        # Iterate over each undeleted object
        foreach my $datum (@preQueryData)
        {
            if ($parent eq $object)
            {
                my @conditions = ();
                foreach my $key (@primaryKeys)
                {
                    push @conditions,
                      new MAM::Condition(
                        name  => $parent,
                        value => $datum->getPropertyValue($key)
                      );
                }

# Recursively undelete associations in which object was the parent and same transactionId
                my $subUndeleteRequest = new MAM::Request(
                    database   => $database,
                    object     => $obj,
                    action     => 'Undelete',
                    conditions => \@conditions,
                    options    => \@options
                );
                my $proxy = new MAM::Proxy(
                    database  => $database,
                    request   => $subUndeleteRequest,
                    requestId => $requestId
                );
                my $subUndeleteResponse = $proxy->execute();
                $associationCount += $subUndeleteResponse->getCount();
            }
            if ($child eq $object)
            {
                my @conditions = ();
                foreach my $key (@primaryKeys)
                {
                    push @conditions,
                      new MAM::Condition(
                        name  => $key,
                        value => $datum->getPropertyValue($key)
                      );
                }

# Recursively undelete associations in which object was the parent and same requestId
                my $subUndeleteRequest = new MAM::Request(
                    database   => $database,
                    object     => $obj,
                    action     => 'Undelete',
                    conditions => \@conditions,
                    options    => \@options
                );
                my $proxy = new MAM::Proxy(
                    database  => $database,
                    request   => $subUndeleteRequest,
                    requestId => $requestId
                );
                my $subUndeleteResponse = $proxy->execute();
                $associationCount += $subUndeleteResponse->getCount();
            }
        }
    }

    # Perform the base undelete
    my $count = $database->undelete(
        object     => $object,
        actor      => $actor,
        conditions => \@conditions,
        options    => \@options,
        requestId  => $requestId,
        txnId      => $txnId
    );

    $count = $count eq '0E0' ? 0 : $count;
    my $message =
        "Successfully undeleted $count "
      . lower($object)
      . ($count == 1 ? '' : 's');

    # Query undeleted objects
    my $postQueryRequest = new MAM::Request(
        database => $database,
        object   => $object,
        action   => 'Query',
        conditions =>
          [new MAM::Condition(name => 'TransactionId', value => $txnId)],
        options => \@options
    );
    MAM::Proxy->prepareSelections($postQueryRequest);
    my $postQueryResponse = MAM::Base->query($postQueryRequest, $requestId, 0);
    my @postData          = $postQueryResponse->getData();
    my %postMetadata      = $postQueryResponse->getMetadata();

    if ($associationCount)
    {
        $message .= " and $associationCount association"
          . ($associationCount == 1 ? '' : 's');
    }

    return new MAM::Response()
      ->success($count, \@postData, \%postMetadata, $message);
}

# ----------------------------------------------------------------------------
# $response = refresh($request[, $refreshId]);
# ----------------------------------------------------------------------------

# Refresh (Event, Notification, System)
sub refresh
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object   = $request->getObject();
    my $database = $request->getDatabase();

    # Event Refresh
    if ($object eq 'Event')
    {
        my $now = $requestTime;

# Query to see if there are any events with an end time in the past
# where the fire time is after the end time or it is NULL
# SELECT Id FROM Event WHERE EndTime<'now' AND (FireTime>EndTime OR FireTime IS NULL)
        my $results = $database->select(
            object     => 'Event',
            selections => [new MAM::Selection(name => 'Id')],
            conditions => [
                new MAM::Condition(
                    name  => 'EndTime',
                    value => $now,
                    op    => 'LE'
                ),
                new MAM::Condition(
                    name    => 'FireTime',
                    subject => 'Event',
                    value   => 'EndTime',
                    op      => 'GT',
                    conj    => 'And',
                    group   => '+1'
                ),
                new MAM::Condition(
                    name  => 'FireTime',
                    value => 'NULL',
                    conj  => 'Or',
                    group => '-1'
                ),
            ],
        );

        # If so, delete them
        if (@{$results->{data}})
        {
# DELETE from Event WHERE EndTime<'now' AND (FireTime>EndTime OR FireTime IS NULL)
            my $subRequest = new MAM::Request(
                database   => $database,
                object     => 'Event',
                action     => 'Delete',
                actor      => $SUPER_USER,
                conditions => [
                    new MAM::Condition(
                        name  => 'EndTime',
                        value => $now,
                        op    => 'LE'
                    ),
                    new MAM::Condition(
                        name    => 'FireTime',
                        subject => 'Event',
                        value   => 'EndTime',
                        op      => 'GT',
                        conj    => 'And',
                        group   => '+1'
                    ),
                    new MAM::Condition(
                        name  => 'FireTime',
                        value => 'NULL',
                        conj  => 'Or',
                        group => '-1'
                    ),
                ],
            );
            my $subResponse = MAM::Base->delete($subRequest, $requestId);
            return $subResponse;
        }
        else
        {
            return new MAM::Response()
              ->success(1, 'No stale events were located for deletion');
        }
    }

    # Notification Refresh
    elsif ($object eq 'Notification')
    {
        my $now = $requestTime;

      # Query to see if there are any notifications with an end time in the past
      # SELECT Id FROM Notification WHERE EndTime<'now'
        my $results = $database->select(
            object     => 'Notification',
            selections => [new MAM::Selection(name => 'Id')],
            conditions => [
                new MAM::Condition(
                    name  => 'EndTime',
                    value => $now,
                    op    => 'LE'
                ),
            ],
        );

        # If so, delete them
        if (@{$results->{data}})
        {
# DELETE from Notification WHERE EndTime<'now' AND (FireTime>EndTime OR FireTime IS NULL)
# Delete notifications with an end time in the past
# DELETE from Nodification WHERE EndTime<'now'
            my $subRequest = new MAM::Request(
                database   => $database,
                object     => 'Notification',
                action     => 'Delete',
                actor      => $SUPER_USER,
                conditions => [
                    new MAM::Condition(
                        name  => 'EndTime',
                        value => $now,
                        op    => 'LE'
                    ),
                ],
                options =>
                  [new MAM::Option(name => 'Recursive', value => 'True'),],
            );
            my $subResponse = MAM::Base->delete($subRequest, $requestId);
            return $subResponse;
        }
        else
        {
            return new MAM::Response()
              ->success(1, 'No stale events were located for deletion');
        }
    }

    # System Refresh
    elsif ($object eq 'System')
    {
        MAM::Cache->setStale(1);
        return new MAM::Response()
          ->success('Successfully refreshed the metadata cache');
    }

    else
    {
        return new MAM::Response()
          ->failure('315', "$object does not implement the Refresh action");
    }
}

# ----------------------------------------------------------------------------
# $response = fire($request, $requestId);
# ----------------------------------------------------------------------------

# Fire (Event)
sub fire
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object   = $request->getObject();
    my $actor    = $request->getActor();
    my $id       = $request->getOptionValue('Id');
    my @options  = $request->getOptions();
    my $database = $request->getDatabase();
    my $txnId    = $database->nextId('Transaction');
    my $now      = $requestTime;

    if ($object ne 'Event')
    {
        return new MAM::Response()
          ->failure('315', "$object does not implement the Fire action");
    }

    unless (defined $id)
    {
        return new MAM::Response()
          ->failure('314', 'The Id option must be specified');
    }

# Obtain event details
# SELECT Id,FireCommand,FireTime,RetentionPeriodRearmPeriod,EndTime,Notify,RearmOnFailure,FailureCommand,CatchUp FROM Event WHERE Id=$id
    my $results = $database->select(
        object     => 'Event',
        selections => [
            new MAM::Selection(name => 'Id'),
            new MAM::Selection(name => 'FireCommand'),
            new MAM::Selection(name => 'FireTime'),
            new MAM::Selection(name => 'RearmPeriod'),
            new MAM::Selection(name => 'EndTime'),
            new MAM::Selection(name => 'Notify'),
            new MAM::Selection(name => 'RearmOnFailure'),
            new MAM::Selection(name => 'FailureCommand'),
            new MAM::Selection(name => 'CatchUp'),
        ],
        conditions => [
            new MAM::Condition(
                name  => 'Id',
                value => $id,
            ),
        ],
    );
    if (@{$results->{data}})
    {
        my $row            = ${$results->{data}}[0];
        my $id             = $row->[0];
        my $fireCommand    = $row->[1];
        my $fireTime       = $row->[2];
        my $rearmPeriod    = $row->[3];
        my $endTime        = $row->[4];
        my $notify         = $row->[5];
        my $rearmOnFailure = $row->[6];
        my $failureCommand = $row->[7];
        my $catchUp        = $row->[8];

        # Clean up stale events without firing them
        if (   defined $endTime
            && $endTime <= $now
            && (! defined $fireTime || $endTime < $fireTime))
        {
            if ($log->is_debug())
            {
                $log->debug(
                    'This event is stale and will be deleted without being executed'
                );
            }

            # DELETE from Event WHERE Id=$id
            my $subRequest = new MAM::Request(
                database   => $database,
                object     => 'Event',
                action     => 'Delete',
                actor      => $SUPER_USER,
                conditions => [new MAM::Condition(name => 'Id', value => $id)],
            );
            my $subResponse = MAM::Base->delete($subRequest, $requestId);
            return new MAM::Response()->failure('740',
                "The specified event ($id) was ineligible for execution (past its end time) and has been deleted.\n"
                  . $subResponse->getMessage());
        }

        # Deserialize the command
        my $fireCommandRequest = new MAM::Request();
        my $fireCommandResponse;
        try
        {
            $fireCommandRequest->deserialize($fireCommand);
        }
        catch MAM::Exception with
        {
            my $E = shift;
            my $message =
              'Error detected while deserializing fire command: '
              . $E->{'-text'};
            my $code = $E->{'-value'};
            &logAndWarn($message);
            $fireCommandResponse =
              new MAM::Response()->failure($code, $message);
        };

        # Prepare the sub request
        my $proxy;
        unless (defined $fireCommandResponse)
        {
            if ($log->is_debug())
            {
                $log->debug("Preparing fire command for event ($id)");
            }

            # Prepare the command
            try
            {
                $proxy = new MAM::Proxy(
                    database  => $database,
                    request   => $fireCommandRequest,
                    requestId => $requestId,
                );
            }
            catch MAM::Exception with
            {
                my $E = shift;
                my $message =
                  'Error detected while preparing fire command: '
                  . $E->{'-text'};
                my $code = $E->{'-value'};
                &logAndWarn($message);
                $fireCommandResponse =
                  new MAM::Response()->failure($code, $message);
            };
        }

        # Execute the sub request
        unless (defined $fireCommandResponse)
        {
            if ($log->is_info())
            {
                $log->info("Executing fire command for event ($id)");
            }

            # Execute the command
            try
            {
                $fireCommandResponse = $proxy->execute();
                $fireCommandResponse->setChunkMax(0);
            }
            catch MAM::Exception with
            {
                my $E = shift;
                my $message =
                  'Error detected while executing fire command: '
                  . $E->{'-text'};
                my $code = $E->{'-value'};
                &logAndWarn($message);
                $fireCommandResponse =
                  new MAM::Response()->failure($code, $message);
            };
        }

        # Stick response details in the options
        my $status            = $fireCommandResponse->getStatus();
        my $code              = $fireCommandResponse->getCode();
        my $message           = $fireCommandResponse->getMessage();
        my $count             = $status eq 'Success' ? 1 : 0;
        my $markedForDeletion = 0;
        my @messages          = ();
        push @messages, $message if defined $message;

        # Isolate the transaction from the remaining operations
        if ($status eq 'Success' || $status eq 'Warning')
        {
            $database->getHandle()->commit();
        }
        else
        {
            $database->getHandle()->rollback();
        }

        # Rearm the event (on success or rearmOnFailure)
        my $rearmTime;
        my $startTime = $now;
        if ((   $status ne 'Failure' || (defined $rearmOnFailure
                    && $rearmOnFailure eq 'True')
            )
            && defined $fireTime
            && defined $rearmPeriod
            && ((! defined $endTime)
                || (   ($catchUp eq 'True' && $endTime > $fireTime)
                    || ($catchUp eq 'False' && $endTime > $now))
            )
          )
        {
            if ($log->is_info())
            {
                $log->info("Rearming the event ($id)");
            }

            $startTime = $fireTime if $catchUp eq 'True';
            my $baseTime = $startTime;
            my $delta    = 0;

            # Adjust the base time and the delta as necessary
            if ($log->is_trace())
            {
                $log->trace(
                    "Calling toDMF($rearmPeriod, \\\$baseTime[=$baseTime], \\\$delta]"
                );
            }
            my $recur = toDMF($rearmPeriod, \$baseTime, \$delta);
            if ($log->is_debug())
            {
                $log->debug("Calculated frequency description string ($recur)");
            }

            if ($recur =~ /^[\d\*:]+$/)
            {
                if ($log->is_trace())
                {
                    $log->trace(
                        "Calling ParseRecur($recur, \"epoch $baseTime\", \"epoch "
                          . ($startTime + 1)
                          . "\", \"epoch "
                          . ($startTime + $delta + 1)
                          . "\")");
                }
                my @rearmDates = ParseRecur(
                    $recur,
                    "epoch $baseTime",
                    'epoch ' . ($startTime + 1),
                    'epoch ' . ($startTime + $delta + 1)
                );
                if (@rearmDates)
                {
                    my $rearmDate = shift @rearmDates;
                    if ($rearmDate =~
                        /^(\d{4})(\d{2})(\d{2})(\d{2}):(\d{2}):(\d{2})$/)
                    {
                        my ($year, $mon, $day, $hour, $min, $sec) =
                          ($1, $2, $3, $4, $5, $6);
                        $rearmTime =
                          timelocal($sec, $min, $hour, $day, $mon - 1,
                            $year - 1900);
                        if ($log->is_trace())
                        {
                            $log->trace(
                                "Calculated new fire time ($rearmTime)");
                        }
                    }
                    else
                    {
                        $status = 'Warning';
                        $code   = '100';
                        push @messages,
                          "Warning -- The calculated rearm date is not in the expected format ($rearmDate).";
                    }
                }
                else
                {
                    $status = 'Warning';
                    $code   = '100';
                    push @messages,
                      'Warning -- No rearm time was obtained for the event.';
                }
            }
            else
            {
                $status = 'Warning';
                $code   = '100';
                push @messages,
                  "Warning -- Failed to calculate a rearm time for the event: $recur";
            }
        }

        # Build up event assignments
        my $newArmTime =
          defined $fireTime ? ($now < $fireTime ? $now : $fireTime) : $now;
        my @eventAssignments = (
            new MAM::Assignment(
                name  => 'ArmTime',
                value => $newArmTime
            )
        );
        if (   defined $rearmTime
            && $rearmTime > $startTime
            && (   (! defined $endTime)
                || ($rearmTime <= $endTime))
          )
        {
            push @eventAssignments,
              new MAM::Assignment(name => 'FireTime', value => $rearmTime);
        }
        # The rearm period puts the new fire time past the end time
        elsif (defined $endTime
            && defined $rearmTime
            && $rearmTime > $endTime)
        {
            $markedForDeletion = 1;
        }
        # If the event has an fire time, and it was not deleted or rearmed,
        # clear the fire time so it does not fire continuously
        elsif (defined $fireTime)
        {
            if ($log->is_info())
            {
                $log->info('Clearing the event fire time');
                push @messages, 'Clearing the event fire time.';
            }

            # Clear the event fire time
            # UPDATE Event SET FireTime=NULL WHERE Id=$id
            push @eventAssignments,
              new MAM::Assignment(name => 'FireTime', value => 'NULL');
        }

        # Delete obsolete events
        if ($markedForDeletion
            || (   defined $fireTime
                && $fireTime <= $now
                && ! defined $rearmPeriod)
            || (   defined $endTime
                && $endTime <= $now
                && ((! defined $fireTime) || ($endTime <= $fireTime)))
          )
        {
            if ($log->is_info())
            {
                $log->info("Deleting obsolete event ($id)");
            }

            # DELETE from Event WHERE Id=$id
            my $subRequest = new MAM::Request(
                database   => $database,
                object     => 'Event',
                action     => 'Delete',
                actor      => $SUPER_USER,
                conditions => [new MAM::Condition(name => 'Id', value => $id)],
            );
            my $subResponse = MAM::Base->delete($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Failure')
            {
                $status = 'Warning';
                $code   = '100';
                push @messages,
                  'Warning -- Failed deleting obsolete event: '
                  . $subResponse->getMessage();
            }
            else
            {
                push @messages, 'The controlling event has been deleted.';
            }
        }
        # Modify event with new arm time (and new fire time if necessary)
        else
        {
      # Update the event
      # UPDATE Event SET ArmTime=$fireTime|$now,FireTime=$rearmTime WHERE Id=$id
            my $subRequest = new MAM::Request(
                database    => $database,
                object      => 'Event',
                action      => 'Modify',
                actor       => $SUPER_USER,
                assignments => \@eventAssignments,
                conditions  => [
                    new MAM::Condition(
                        name  => 'Id',
                        value => $id
                    )
                ],
            );
            my $subResponse = MAM::Base->modify($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Failure')
            {
                $status = 'Warning';
                $code   = '100';
                push @messages, 'Warning -- Unable to update event: '
                  . $subResponse->getMessage();
            }
        }

        # Send Fire Notification
        if (defined $notify)
        {
            my @notifyOptions = (
                new MAM::Option(name => 'Event',  value => $id),
                new MAM::Option(name => 'Type',   value => 'Fire'),
                new MAM::Option(name => 'Status', value => $status),
                new MAM::Option(name => 'Code',   value => $code),
            );
            push @notifyOptions,
              new MAM::Option(
                name  => 'Message',
                value => join("\n", @messages)
              ) if @messages;
            # Add keys
            my @keys              = ();
            my $fireCommandObject = $fireCommandRequest->getObject();
            foreach my $attribute (
                MAM::Cache->listAttributesWithPropertyValue(
                    $fireCommandObject, 'PrimaryKey', 'True'
                )
              )
            {
                if (defined $fireCommandRequest->getConditionValue($attribute))
                {
                    push @keys,
                      $fireCommandRequest->getConditionValue($attribute);
                }
                elsif (defined $fireCommandRequest->getOptionValue($attribute))
                {
                    push @keys, $fireCommandRequest->getOptionValue($attribute);
                }
            }
            if (@keys)
            {
                push @notifyOptions,
                  new MAM::Option(name => 'Key', value => join ':', @keys),
                  ;
            }

            # Iterate over each notify expression
            foreach my $notification (split /,/, $notify)
            {
                if ($notify =~ /^([+-=])?(?:(\w+):)?([^:\s]+)?$/)
                {
                    my @notificationOptions = ();
                    my ($term, $deliveryMethod, $recipient) = ($1, $2, $3);

                    # Do not send notification if term is not met
                    if (defined $term)
                    {
                        next
                          if ( ($term eq '-' && $status ne 'Failure')
                            || ($term eq '+' && $status eq 'Failure'));
                    }

                    push @notificationOptions,
                      new MAM::Option(
                        name  => 'Recipient',
                        value => $recipient
                      ) if defined $recipient;
                    push @notificationOptions,
                      new MAM::Option(
                        name  => 'DeliveryMethod',
                        value => $deliveryMethod
                      ) if defined $deliveryMethod;

                    my $subRequest = new MAM::Request(
                        database => $database,
                        object   => 'Notification',
                        action   => 'Send',
                        actor    => $SUPER_USER,
                        options  => [@notifyOptions, @notificationOptions],
                    );
                    my $subResponse = MAM::Base->send($subRequest, $requestId);
                    if ($subResponse->getStatus() eq 'Failure')
                    {
                        $status = 'Warning';
                        $code   = '100';
                        push @messages,
                          'Warning -- Unable to send fire notification: '
                          . $subResponse->getMessage();
                    }
                }
                else
                {
                    push @messages,
                      "Warning -- Unable to send fire notification: Invalid notify option format ($notify)";
                }
            }
        }

        # Execute Failure Command on Failure
        if ($status eq 'Failure' && defined $failureCommand)
        {
            # Deserialize the command
            my $failureCommandRequest = new MAM::Request();
            my $failureCommandResponse;
            try
            {
                $failureCommandRequest->deserialize($failureCommand);
            }
            catch MAM::Exception with
            {
                my $E = shift;
                my $message =
                  'Error detected while deserializing failure command: '
                  . $E->{'-text'};
                my $code = $E->{'-value'};
                &logAndWarn($message);
                $failureCommandResponse =
                  new MAM::Response()->failure($code, $message);
            };

            # Execute the sub request
            unless (defined $failureCommandResponse)
            {
                if ($log->is_info())
                {
                    $log->info("Executing failure command for event ($id)");
                }

                my $proxy = new MAM::Proxy(
                    database  => $database,
                    request   => $failureCommandRequest,
                    requestId => $requestId,
                );
                $failureCommandResponse = $proxy->execute();
                $failureCommandResponse->setChunkMax(0);
            }

            # Stick response details in the options
            my $status  = $failureCommandResponse->getStatus();
            my $code    = $failureCommandResponse->getCode();
            my $message = $failureCommandResponse->getMessage();
            my $count   = $status eq 'Success' ? 1 : 0;
            push @messages, ('Failure command message: ' . $message)
              if defined $message;

            # Isolate the transaction from the remaining operations
            if ($status eq 'Success' || $status eq 'Warning')
            {
                $database->getHandle()->commit();
            }
            else
            {
                $database->getHandle()->rollback();
            }

            # Send Failure Notification
            if (defined $notify)
            {
                my @notifyOptions = (
                    new MAM::Option(name => 'Event',  value => $id),
                    new MAM::Option(name => 'Type',   value => 'Failure'),
                    new MAM::Option(name => 'Status', value => $status),
                    new MAM::Option(name => 'Code',   value => $code),
                );
                push @notifyOptions,
                  new MAM::Option(
                    name  => 'Message',
                    value => join("\n", @messages)
                  ) if @messages;
                # Add keys
                my @keys                 = ();
                my $failureCommandObject = $failureCommandRequest->getObject();
                foreach my $attribute (
                    MAM::Cache->listAttributesWithPropertyValue(
                        $failureCommandObject, 'PrimaryKey', 'True'
                    )
                  )
                {
                    if (defined $failureCommandRequest->getConditionValue(
                            $attribute))
                    {
                        push @keys,
                          $failureCommandRequest->getConditionValue($attribute);
                    }
                    elsif (
                        defined $failureCommandRequest->getOptionValue(
                            $attribute))
                    {
                        push @keys,
                          $failureCommandRequest->getOptionValue($attribute);
                    }
                }
                if (@keys)
                {
                    push @notifyOptions,
                      new MAM::Option(
                        name  => 'Key',
                        value => join ':',
                        @keys
                      ),
                      ;
                }

                # Iterate over each notify expression
                foreach my $notification (split /,/, $notify)
                {
                    if ($notify =~ /^([+-=])?(?:(\w+):)?([^:\s]+)?$/)
                    {
                        my @notificationOptions = ();
                        my ($term, $deliveryMethod, $recipient) = ($1, $2, $3);

                        # Do not send notification if term is not met
                        if (defined $term)
                        {
                            next
                              if ( ($term eq '-' && $status ne 'Failure')
                                || ($term eq '+' && $status eq 'Failure'));
                        }

                        push @notificationOptions,
                          new MAM::Option(
                            name  => 'Recipient',
                            value => $recipient
                          ) if defined $recipient;
                        push @notificationOptions,
                          new MAM::Option(
                            name  => 'DeliveryMethod',
                            value => $deliveryMethod
                          ) if defined $deliveryMethod;

                        my $subRequest = new MAM::Request(
                            database => $database,
                            object   => 'Notification',
                            action   => 'Send',
                            actor    => $SUPER_USER,
                            options  => [@notifyOptions, @notificationOptions],
                        );
                        my $subResponse =
                          MAM::Base->send($subRequest, $requestId);
                        if ($subResponse->getStatus() eq 'Failure')
                        {
                            $status = 'Warning';
                            $code   = '100';
                            push @messages,
                              'Warning -- Unable to send failure notification: '
                              . $subResponse->getMessage();
                        }
                    }
                    else
                    {
                        push @messages,
                          "Warning -- Unable to send failure notification: Invalid notify option format ($notify)";
                    }
                }
            }
        }

        # Log the transaction
        my @logAssignments = ();
        if (defined $status)
        {
            push @logAssignments,
              new MAM::Assignment(name => 'Status', value => "$status");
            $fireCommandResponse->setStatus($status);
        }
        if (defined $code)
        {
            push @logAssignments,
              new MAM::Assignment(name => 'Code', value => "$code");
            $fireCommandResponse->setCode($code);
        }
        if (@messages)
        {
            push @logAssignments,
              new MAM::Assignment(
                name  => 'Message',
                value => join("\n", @messages)
              );
            $fireCommandResponse->setMessage(join("\n", @messages));
        }
        push @logAssignments,
          new MAM::Assignment(
            name  => 'Description',
            value => "$fireCommand"
          ) if defined $fireCommand;
        $database->logTransaction(
            requestId   => $requestId,
            txnId       => $txnId,
            object      => $object,
            action      => 'Fire',
            actor       => $actor,
            assignments => \@logAssignments,
            count       => $count,
        );

        # Commit the event wrapup and transaction recording
        $database->getHandle()->commit();

        # Return the response
        return $fireCommandResponse;
    }
    else
    {
        return new MAM::Response()->failure('314', "Invalid $object Id ($id)");
    }
}

# ----------------------------------------------------------------------------
# $response = send($request, $requestId);
# ----------------------------------------------------------------------------

# Send (Notification)
sub send
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object        = $request->getObject();
    my $actor         = $request->getActor();
    my $database      = $request->getDatabase();
    my $txnId         = $database->nextId('Transaction');
    my $event         = $request->getOptionValue('Event');
    my $type          = $request->getOptionValue('Type');
    my $statusOption  = $request->getOptionValue('Status');
    my $codeOption    = $request->getOptionValue('Code');
    my $messageOption = $request->getOptionValue('Message');
    my $key           = $request->getOptionValue('Key');
    my $recipient     = $request->getOptionValue('Recipient');
    my $deliveryMethod =
         $request->getOptionValue('DeliveryMethod')
      || $serverConfig->{'notification.deliverymethod'}
      || 'Store';
    my $endTime = $request->getOptionValue('EndTime');
    my @options = $request->getOptions();
    my $now     = $requestTime;

    if ($object ne 'Notification')
    {
        return new MAM::Response()
          ->failure('315', "$object does not implement the Send action");
    }

    unless (defined $event)
    {
        return new MAM::Response()
          ->failure('314', 'The Event option must be specified');
    }

    unless (defined $statusOption)
    {
        return new MAM::Response()
          ->failure('314', 'The Status option must be specified');
    }

    unless (defined $codeOption)
    {
        return new MAM::Response()
          ->failure('314', 'The Code option must be specified');
    }

    # Check delivery method
    if ($deliveryMethod =~ /^store$/i)
    {
        my $duration = $serverConfig->{'notification.duration'} || 1209600;

        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'Notification',
            action   => 'Create',
            actor    => $SUPER_USER,
        );
        $subRequest->addAssignment('Event', $event) if defined $event;
        $subRequest->addAssignment('Type',  $type)  if defined $type;
        $subRequest->addAssignment('Status', $statusOption)
          if defined $statusOption;
        $subRequest->addAssignment('Code', $codeOption)
          if defined $codeOption;
        $subRequest->addAssignment('Message', $messageOption)
          if defined $messageOption;
        $subRequest->addAssignment('Key', $key) if defined $key;
        $subRequest->addAssignment('Recipient', $recipient)
          if defined $recipient;

        if (defined $endTime)
        {
            $subRequest->addAssignment('EndTime', $endTime);
        }
        else
        {
            $subRequest->addAssignment('EndTime', $now + $duration);
        }

        my $subResponse = MAM::Base->create($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Failed sending notification: ' . $subResponse->getMessage());
        }
    }

    # No other delivery methods yet implemented
    else
    {
        return new MAM::Response()->failure('317',
            "Invalid DeliveryMethod option value ($deliveryMethod). 'Store' is the only delivery method that is currently implemented."
        );
    }

    # Log the transaction
    $database->logTransaction(
        requestId => $requestId,
        txnId     => $txnId,
        object    => $object,
        action    => 'Send',
        actor     => $actor,
        options   => \@options,
        count     => 1,
    );

    # Return response
    return new MAM::Response()
      ->success(1, 'Successfully sent the notification');
}

# ----------------------------------------------------------------------------
# $response = undo($request, $requestId);
# ----------------------------------------------------------------------------

# Undo (Transaction)
sub undo
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object   = $request->getObject();
    my $database = $request->getDatabase();
    my $dbh      = $database->getHandle();
    my $sql;
    my $count = 0;
    my $tmpCount;

    if ($object ne 'Transaction')
    {
        return new MAM::Response()
          ->failure('315', "$object does not implement the Undo action");
    }

# Use requestId as undoId if specified, otherwise grab the most recent requestId
    my $undoId = $request->getConditionValue('RequestId');
    if (! defined $undoId)
    {
        # undoId should be max(requestId) in Transaction
        # SELECT MAX(g_request_id) FROM g_transaction
        $sql = 'SELECT MAX(g_request_id) FROM g_transaction';
        if ($log->is_trace())
        {
            $log->trace("SQL Query: $sql");
        }
        my $sth = $dbh->prepare($sql);
        $sth->execute();
        my ($firstId) = $sth->fetchrow_array();
        if ($firstId)
        {
            $undoId = $firstId;
        }
        else
        {
            return new MAM::Response()
              ->failure('740', 'No transactions available to undo');
        }
    }

    # Iterate through all defined objects
    foreach my $obj (MAM::Cache->listObjects())
    {
        my $dbname = toDBC($obj);

        # Build primary keys
        my $firstOne      = 1;
        my $primaryKeys   = '';
        my $aPrimaryKeys  = '';
        my $abPrimaryKeys = '';
        foreach my $attribute (
            MAM::Cache->listAttributesWithPropertyValue(
                $obj, 'PrimaryKey', 'True'
            )
          )
        {
            if ($firstOne) { $firstOne = 0; }
            else
            {
                $primaryKeys   .= ',';
                $aPrimaryKeys  .= ',';
                $abPrimaryKeys .= ' AND ';
            }
            my $dbattr = toDBC($attribute);
            $primaryKeys   .= $dbattr;
            $aPrimaryKeys  .= "a.${dbattr}";
            $abPrimaryKeys .= "a.${dbattr}=b.${dbattr}";
        }

        # We cannot undo objects without a primary field
        next if $primaryKeys eq '';

        # Merge all the records with requestId>=undoId into journal
        # so redo can be performed later
        # INSERT INTO ${dbname}_log
        # SELECT * FROM $dbname WHERE requestId>=$undoId
        # Necessary for diff journaling
        $sql =
          "INSERT INTO ${dbname}_log SELECT * FROM $dbname WHERE g_request_id>=$undoId";
        if ($log->is_trace())
        {
            $log->trace("SQL Update: $sql");
        }
        $dbh->do($sql);

        # Delete all records from dbname with requestId>=undoId
        # DELETE FROM $dbname WHERE requestId>=$undoId
        # Necessary for all journaling
        $sql = "DELETE FROM $dbname WHERE g_request_id>=$undoId";
        if ($log->is_trace())
        {
            $log->trace("SQL Update: $sql");
        }
        $count += $dbh->do($sql);

        # Grab all highest requestId records from journal for which there are
        # records in the requestId<undoId and requestId>=requestId categories
        # Had to include txnId and requestId here to get the right records
        # INSERT INTO $dbname SELECT * FROM ${dbname}_log
        # WHERE ($primaryKeys,requestId,txnId) IN
        # ( SELECT a.$primaryKeys,a.requestId,a.txnId FROM
        # ( SELECT $primaryKeys,requestId,txnId FROM ${dbname}_log
        # WHERE ($primaryKeys,requestId,txnId) IN
        # ( SELECT $primaryKeys,MAX(requestId),MAX(txnId) FROM ${dbname}_log
        # WHERE requestId<$undoId GROUP BY $primaryKeys ))
        # AS a INNER JOIN
        # ( SELECT $primaryKeys,MAX(requestId),MAX(txnId) FROM ${dbname}_log
        # WHERE requestId>=$undoId GROUP BY $primaryKeys )
        # AS b ON a.$primaryKeys=b.$primaryKeys ) AND
        # ($primaryKeys) NOT IN ( SELECT $primaryKeys FROM $dbname );
        # Necessary for all journaling
        $sql =
          "INSERT INTO $dbname SELECT * FROM ${dbname}_log WHERE ($primaryKeys,g_request_id,g_transaction_id) IN ( SELECT $aPrimaryKeys,a.g_request_id,a.g_transaction_id FROM ( SELECT $primaryKeys,g_request_id,g_transaction_id FROM ${dbname}_log WHERE ($primaryKeys,g_request_id,g_transaction_id) IN ( SELECT $primaryKeys,MAX(g_request_id),MAX(g_transaction_id) FROM ${dbname}_log WHERE g_request_id<$undoId GROUP BY $primaryKeys)) AS a INNER JOIN ( SELECT $primaryKeys,MAX(g_request_id),MAX(g_transaction_id) FROM ${dbname}_log WHERE g_request_id>=$undoId GROUP BY $primaryKeys ) AS b ON $abPrimaryKeys ) AND ($primaryKeys) NOT IN ( SELECT $primaryKeys FROM $dbname )";
        if ($log->is_trace())
        {
            $log->trace("SQL Update: $sql");
        }
        $dbh->do($sql);

        # Delete the intersecting records
        # DELETE FROM ${dbname}_log
        # WHERE (requestId,txnId) IN
        # ( SELECT requestId,txnId FROM $dbname );
        # Necessary for diff journaling
        $sql =
          "DELETE FROM ${dbname}_log WHERE (g_request_id,g_transaction_id) IN ( SELECT g_request_id,g_transaction_id FROM $dbname )";
        if ($log->is_trace())
        {
            $log->trace("SQL Update: $sql");
        }
        $dbh->do($sql);
    }

    # Reset the Request Key Generator to the requestId we just undid
    $sql =
      "UPDATE g_key_generator set g_next_id=$undoId WHERE g_name='Request'";
    if ($log->is_trace())
    {
        $log->trace("SQL Update: $sql");
    }
    $tmpCount = $dbh->do($sql);
    if ($tmpCount != 1)
    {
        return new MAM::Response()
          ->failure('720',
            'Unable to update next generated request id after undo');
    }

    # Store the requestId in the undo table
    $sql = "INSERT INTO g_undo (g_request_id) VALUES ($undoId)";
    if ($log->is_trace())
    {
        $log->trace("SQL Update: $sql");
    }
    $tmpCount = $dbh->do($sql);
    if ($tmpCount != 1)
    {
        return new MAM::Response()
          ->failure('720', 'Unable to store the RequestId in the Undo table');
    }

    return new MAM::Response()
      ->success($count, "Successfully undid $count transactions");
}

# ----------------------------------------------------------------------------
# $response = redo($request, $requestId);
# ----------------------------------------------------------------------------

# Redo (Transaction)
sub redo
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object   = $request->getObject();
    my $database = $request->getDatabase();
    my $dbh      = $database->getHandle();
    my $sql;
    my $count = 0;
    my $tmpCount;

    if ($object ne 'Transaction')
    {
        return new MAM::Response()
          ->failure('315', "$object does not implement the Redo action");
    }

    my $ultUndoTargetId    = $requestId + 1;    # Initialize with "safe" value
    my $penultUndoTargetId = $requestId + 1;    # Initialize with "safe" value

    # Only allow redo if there is an undo. Grab the ultundoTargetId
    # stored in the requestId field of the undo table.
    # If there are more than one undo, get the next most recent one as
    # penultUndoTargetId so that we only undo what the last undo did.
    # SELECT RequestId FROM Undo ORDER BY RequestId;
    $sql = 'SELECT g_request_id FROM g_undo ORDER BY g_request_id';
    if ($log->is_trace())
    {
        $log->trace("SQL Query: $sql");
    }
    my $sth = $dbh->prepare($sql);
    $sth->execute();
    my ($firstId) = $sth->fetchrow_array();
    if ($firstId)
    {
        $ultUndoTargetId = $firstId;
    }
    else
    {
        return new MAM::Response()
          ->failure('740', 'No transactions available to redo');
    }
    my ($nextId) = $sth->fetchrow_array();
    if ($nextId)
    {
        $penultUndoTargetId = $nextId;
    }

    # Iterate through all defined objects
    foreach my $obj (MAM::Cache->listObjects())
    {
        my $dbname = toDBC($obj);

        # Build primary keys
        my $firstOne    = 1;
        my $primaryKeys = '';
        foreach my $attribute (
            MAM::Cache->listAttributesWithPropertyValue(
                $obj, 'PrimaryKey', 'True'
            )
          )
        {
            if ($firstOne) { $firstOne = 0; }
            else
            {
                $primaryKeys .= ',';
            }
            my $dbattr = toDBC($attribute);
            $primaryKeys .= $dbattr;
        }
        # We cannot redo objects without a primary field
        next if $primaryKeys eq '';

        # Have to repopulate the journal records from dbname
        # that we took out of there when undoing
        # INSERT INTO ${dbname}_log
        # SELECT * FROM $dbname
        # WHERE ($primaryKeys) IN
        # ( SELECT DISTINCT $primaryKeys FROM ${dbname}_log
        # WHERE requestId>=$ultUndoTargetId AND requestId<$penultUndoTargetId );
        # Necessary for diff journaling
        $sql =
          "INSERT INTO ${dbname}_log SELECT * FROM $dbname WHERE ($primaryKeys) IN ( SELECT DISTINCT $primaryKeys FROM ${dbname}_log WHERE g_request_id>=$ultUndoTargetId AND g_request_id<$penultUndoTargetId )";
        if ($log->is_trace())
        {
            $log->trace("SQL Update: $sql");
        }
        $dbh->do($sql);

        # Delete all records from dbname which will be replaced by
        # records in the journal, i.e. all records which have equivalents
        # the journal which are requestId>=undoId
        # DELETE FROM $dbname WHERE ($primaryKeys) IN
        # ( SELECT DISTINCT $primaryKeys FROM ${dbname}_log
        # WHERE requestId>=$ultUndoTargetId AND requestId<penultUndoTargetId );
        # Necessary for all journaling
        $sql =
          "DELETE FROM $dbname WHERE ($primaryKeys) IN ( SELECT DISTINCT $primaryKeys FROM ${dbname}_log WHERE g_request_id>=$ultUndoTargetId AND g_request_id<$penultUndoTargetId )";
        if ($log->is_trace())
        {
            $log->trace("SQL Update: $sql");
        }
        $dbh->do($sql);

        # Then to repopulate only the max records
        # INSERT INTO $dbname SELECT * FROM ${dbname}_log
        # WHERE ($primaryKeys,requestId,txnId) IN
        # ( SELECT $primaryKeys,MAX(requestId),MAX(txnId) FROM ${dbname}_log
        # WHERE requestId>=$ultUndoTargetId AND requestId<$penultUndoTargetId
        # GROUP BY $primaryKeys );
        # Necessary for all journaling
        $sql =
          "INSERT INTO $dbname SELECT * FROM ${dbname}_log WHERE ($primaryKeys,g_request_id,g_transaction_id) IN ( SELECT $primaryKeys,MAX(g_request_id),MAX(g_transaction_id) FROM ${dbname}_log WHERE g_request_id>=$ultUndoTargetId AND g_request_id<$penultUndoTargetId GROUP BY $primaryKeys )";
        if ($log->is_trace())
        {
            $log->trace("SQL Update: $sql");
        }
        $count += $dbh->do($sql);

        # Delete the intersecting records
        # DELETE FROM ${dbname}_log
        # WHERE ($primarykeys,requestId,txnId) IN
        # ( SELECT $primarykeys,MAX(requestId),MAX(txnId) FROM ${dbname}_log
        # WHERE requestId>=$ultUndoTargetId AND requestId<$penultUndoTargetId
        #  GROUP BY $primarykeys );
        # Necessary for diff journaling
        $sql =
          "DELETE FROM ${dbname}_log WHERE ($primaryKeys,g_request_id,g_transaction_id) IN ( SELECT $primaryKeys,MAX(g_request_id),MAX(g_transaction_id) FROM ${dbname}_log WHERE g_request_id>=$ultUndoTargetId AND g_request_id<$penultUndoTargetId GROUP BY $primaryKeys )";
        if ($log->is_trace())
        {
            $log->trace("SQL Update: $sql");
        }
        $dbh->do($sql);
    }

    # Delete the RequestId that we just redid from Undo table
    # DELETE FROM Undo WHERE RequestId=$ultUndoTargetId
    $sql = "DELETE FROM g_undo WHERE g_request_id=$ultUndoTargetId";
    if ($log->is_trace())
    {
        $log->trace("SQL Update: $sql");
    }
    $tmpCount = $dbh->do($sql);
    if ($tmpCount != 1)
    {
        return new MAM::Response()
          ->failure('720',
            'Unable to delete the RequestId from the Undo table');
    }

    # Reset the Request Key Generator to $ultUndoTargetId + 1
    $sql =
        'UPDATE g_key_generator set g_next_id='
      . ($ultUndoTargetId + 1)
      . " WHERE g_name='Request'";
    if ($log->is_trace())
    {
        $log->trace("SQL Update: $sql");
    }
    $tmpCount = $dbh->do($sql);
    if ($tmpCount != 1)
    {
        return new MAM::Response()
          ->failure('720',
            'Unable to update next generated request id after undo');
    }

    return new MAM::Response()
      ->success($count, "Successfully redid $count transactions");
}

# ----------------------------------------------------------------------------
# $response = prune($request, $requestId);
# ----------------------------------------------------------------------------

# Prune (System)
# When we get to perl 10, Date::Manip should be replaced with object methods
sub prune
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object          = $request->getObject();
    my $database        = $request->getDatabase();
    my $dbh             = $database->getHandle();
    my $confirmOption   = $request->getOptionValue('Confirm');
    my $retentionPeriod = $request->getOptionValue('RetentionPeriod');
    my $retentionTime   = $request->getOptionValue('RetentionTime');
    my $now             = $requestTime;
    my $objectCount     = 0;
    my $journalCount    = 0;
    my $sql;
    my $count;

    if ($object ne 'System')
    {
        return new MAM::Response()
          ->failure('315', "$object does not implement the Prune action");
    }

    # Must specify RetentionPeriod
    unless (defined $retentionTime || defined $retentionPeriod)
    {
        return new MAM::Response()
          ->failure('720',
            "A RetentionTime or RetentionPeriod option must be specified");
    }

    # RetentionTime and RetentionPeriod are mutually exclusive
    if (defined $retentionTime && defined $retentionPeriod)
    {
        return new MAM::Response()->failure('720',
            "The RetentionTime and RetentionPeriod options are mutually exclusive"
        );
    }

    # RetentionPeriod was specified
    unless (defined $retentionTime)
    {
        my $startTime = $now;
        my $baseTime  = $startTime;
        my $delta     = 0;

        # Adjust the base time and the delta as necessary
        if ($log->is_trace())
        {
            $log->trace(
                "Calling toDMF($retentionPeriod, \\\$baseTime[=$baseTime], \\\$delta]"
            );
        }
        my $recur = toDMF($retentionPeriod, \$baseTime, \$delta);
        if ($log->is_debug())
        {
            $log->debug("Calculated frequency description string ($recur)");
        }

        if ($recur =~ /^[\d\*:]+$/)
        {
            if ($log->is_trace())
            {
                $log->trace(
                    "Calling ParseRecur($recur, \"epoch $baseTime\", \"epoch "
                      . ($startTime - (2 * $delta) + 60)
                      . "\", \"epoch "
                      . ($startTime + 60)
                      . "\")");
            }
            my @retentionDates = ParseRecur(
                $recur,
                "epoch $baseTime",
                'epoch ' . ($startTime - (2 * $delta) + 60),
                'epoch ' . ($startTime + 60)
            );
            if (@retentionDates >= 2)
            {
                my $retentionDate = $retentionDates[-2];
                if ($retentionDate =~
                    /^(\d{4})(\d{2})(\d{2})(\d{2}):(\d{2}):(\d{2})$/)
                {
                    my ($year, $mon, $day, $hour, $min, $sec) =
                      ($1, $2, $3, $4, $5, $6);
                    $retentionTime =
                      timelocal($sec, $min, $hour, $day, $mon - 1,
                        $year - 1900);
                    if ($log->is_trace())
                    {
                        $log->trace(
                            "Calculated new retention time ($retentionTime)");
                    }
                }
                else
                {
                    return new MAM::Response()->failure('720',
                        "The calculated retention date is not in the expected format ($retentionDate)"
                    );
                }
            }
            else
            {
                return new MAM::Response()->failure('720',
                    'An insufficient retention time list was obtained for the pruning ('
                      . join(',', @retentionDates)
                      . ')');
            }
        }
        else
        {
            return new MAM::Response()
              ->failure('720',
                "Failed to calculate a retention time for the pruning");
        }
    }

    unless (defined $confirmOption && $confirmOption =~ /^True$/i)
    {
        my ($sec, $min, $hour, $day, $month, $year) =
          (localtime($retentionTime))[0 .. 5];
        my $pruneTime =
          sprintf('%04d-%02d-%02d', $year + 1900, $month + 1, $day);
        if ($hour || $min || $sec)
        {
            $pruneTime .= sprintf(' %02d:%02d:%02d', $hour, $min, $sec);
        }
        return new MAM::Response()->success(0,
            "This command will permanently delete historical data prior to $pruneTime from the database. You are advised to backup your database before running this command. If you are certain you want to take this action then rerun this command setting the Confirm option to True."
        );
    }

# Grab first request id after the retentionTime
# SELECT MIN(g_request_id) FROM g_transaction WHERE g_creation_time>=$retentionTime
    $sql =
      "SELECT MIN(g_request_id) FROM g_transaction WHERE g_creation_time>=$retentionTime";
    if ($log->is_trace())
    {
        $log->trace("SQL Query: $sql");
    }
    my $sth = $dbh->prepare($sql);
    $sth->execute();
    my %results = ();
    $results{cols}  = $sth->{NUM_OF_FIELDS};
    $results{names} = [map { toUCC($_) } @{$sth->{NAME_lc}}];
    $results{data}  = $sth->fetchall_arrayref();
    $results{rows}  = @{$results{data}};

    # Log the results
    if ($log->is_debug())
    {
        my $entry = '';
        foreach my $row (@{$results{data}})
        {
            $entry .= join('|', map { defined $_ ? $_ : '' } @$row) . "\n";
        }
        $log->debug("SQL Results: $entry");
    }

    my $minRequestId = $results{data}->[0]->[0];
    unless ($minRequestId)
    {
        $sth = $dbh->prepare(
            "SELECT g_next_id FROM g_key_generator WHERE g_name='Request'");
        $sth->execute();
        ($minRequestId) = $sth->fetchrow_array();
        unless ($minRequestId)
        {
            $log->error(
                "No minimum request id could be calculated for the prune operation"
            );
            return new MAM::Response()->failure('720',
                "No minimum request id could be calculated for the prune operation"
            );
        }
    }

    my %dependent = (
        'Lien'    => 'LienAllocation',
        'Quote'   => 'QuoteChargeRate',
        'Payment' => 'PaymentComponent'
    );

    # Iterate through all defined objects
    foreach my $obj (sort MAM::Cache->listObjects())
    {
        my $dbname = toDBC($obj);
        my @primaryKeys =
          MAM::Cache->listAttributesWithPropertyValue($obj, 'PrimaryKey',
            'True');

        # If it's a record
        if (grep { $_ eq $obj } ('Charge', 'Transaction', 'UsageRecord'))
        {
            # Delete all records last modified before retentionTime
            # Records are objects too voluminous to iterate over one at a time
            # We use modification time instead of creation time so we
            # will not delete long-running usage records that are still active

            # DELETE FROM $dbname WHERE g_modification_time<$retentionTime
            $sql =
              "DELETE FROM $dbname WHERE g_modification_time<$retentionTime";
            if ($log->is_trace())
            {
                $log->trace("SQL Update: $sql");
            }
            $count = $dbh->do($sql);
            if ($log->is_debug())
            {
                $log->debug("SQL Rows: $count");
            }
            $objectCount += $count;

            # DELETE FROM ${dbname}_log WHERE g_modification_time<$retentionTime
            $sql =
              "DELETE FROM ${dbname}_log WHERE g_modification_time<$retentionTime";
            if ($log->is_trace())
            {
                $log->trace("SQL Update: $sql");
            }
            $count = $dbh->do($sql);
            if ($log->is_debug())
            {
                $log->debug("SQL Rows: $count");
            }
            $journalCount += $count;
        }

        # If its persistence depends on its end time
        elsif (
            grep { $_ eq $obj } (
                'Event',          'Lien',
                'LienAllocation', 'Notification',
                'Quote',          'QuoteChargeRate',
                'Payment',        'PaymentComponent'
            )
          )
        {
            # Delete objects whose end time is older than retention time

            # Handle dependents with the primary object
            if (grep { $_ eq $obj }
                ('Event', 'Lien', 'Notification', 'Quote', 'Payment'))
            {
                # Delete the dependent objects and journals first
                if (defined $dependent{$obj})
                {
                    my $dependentdbname = toDBC($dependent{$obj});

# DELETE FROM $dependentdbname WHERE $dbname IN (SELECT g_id FROM $dbname WHERE g_end_time<$retentionTime)
                    $sql =
                      "DELETE FROM $dependentdbname WHERE $dbname IN (SELECT g_id FROM $dbname WHERE g_end_time<$retentionTime)";
                    if ($log->is_trace())
                    {
                        $log->trace("SQL Update: $sql");
                    }
                    $count = $dbh->do($sql);
                    if ($log->is_debug())
                    {
                        $log->debug("SQL Rows: $count");
                    }
                    $objectCount += $count;

# DELETE FROM ${dependentdbname}_log WHERE $dbname IN (SELECT g_id FROM $dbname WHERE g_end_time<$retentionTime)
                    $sql =
                      "DELETE FROM ${dependentdbname}_log WHERE $dbname IN (SELECT g_id FROM $dbname WHERE g_end_time<$retentionTime)";
                    if ($log->is_trace())
                    {
                        $log->trace("SQL Update: $sql");
                    }
                    $count = $dbh->do($sql);
                    if ($log->is_debug())
                    {
                        $log->debug("SQL Rows: $count");
                    }
                    $journalCount += $count;
                }

                # Then delete primary objects and journals

# DELETE FROM ${dbname}_log WHERE g_id IN (SELECT g_id FROM $dbname WHERE g_end_time<$retentionTime)
                $sql =
                  "DELETE FROM ${dbname}_log WHERE g_id IN (SELECT g_id FROM $dbname WHERE g_end_time<$retentionTime)";
                if ($log->is_trace())
                {
                    $log->trace("SQL Update: $sql");
                }
                $count = $dbh->do($sql);
                if ($log->is_debug())
                {
                    $log->debug("SQL Rows: $count");
                }
                $journalCount += $count;

                # DELETE FROM $dbname WHERE g_end_time<$retentionTime
                $sql = "DELETE FROM $dbname WHERE g_end_time<$retentionTime";
                if ($log->is_trace())
                {
                    $log->trace("SQL Update: $sql");
                }
                $count = $dbh->do($sql);
                if ($log->is_debug())
                {
                    $log->debug("SQL Rows: $count");
                }
                $objectCount += $count;
            }
        }

        # It is an object that does not grow per job
        else
        {
            my %results;

            # Delete old deleted entries

# Delete all objects and journals for objects deleted prior to retentionTime
# SELECT $primaryKeys FROM $dbname WHERE g_deleted='True' AND g_modification_time<$retentionTime

            if (@primaryKeys)
            {
                $sql = "SELECT ";
                $sql .= join ',', map { toDBC($_) } @primaryKeys;
                $sql .=
                  " FROM $dbname WHERE g_deleted='True' AND g_modification_time<$retentionTime";
                if ($log->is_trace())
                {
                    $log->trace("SQL Query: $sql");
                }
                $sth = $dbh->prepare($sql);
                $sth->execute();
                %results = ();
                $results{cols} = $sth->{NUM_OF_FIELDS};
                $results{names} = [map { toUCC($_) } @{$sth->{NAME_lc}}];
                $results{data}  = $sth->fetchall_arrayref();
                $results{rows}  = @{$results{data}};

                # Log the results
                if ($log->is_debug())
                {
                    my $entry = '';
                    foreach my $row (@{$results{data}})
                    {
                        $entry .=
                          join('|', map { defined $_ ? $_ : '' } @$row) . "\n";
                    }
                    $log->debug("SQL Results: $entry");
                }

                # Delete logs for each deleted object
                # DELETE FROM ${dbname}_log WHERE @primaryKeys=@values
                foreach my $row (@{$results{data}})
                {
                    my @conditions = ();
                    foreach my $index (0 .. $#primaryKeys)
                    {
                        my $primaryKey = $primaryKeys[$index];
                        my $dataType =
                          MAM::Cache->getAttributeProperty($obj,
                            $primaryKey, 'DataType');
                        my $delim = $dataType eq 'String' ? q{'} : q{};
                        my $condition = " AND " . toDBC($primaryKey);
                        $condition .=
                          defined ${$row}[$index]
                          ? "=${delim}${$row}[$index]${delim}"
                          : " IS NULL";
                        push @conditions, $condition;
                    }
                    $sql = "DELETE FROM ${dbname}_log WHERE 1=1";
                    $sql .= join '', @conditions;
                    if ($log->is_trace())
                    {
                        $log->trace("SQL Update: $sql");
                    }
                    $count = $dbh->do($sql);
                    if ($log->is_debug())
                    {
                        $log->debug("SQL Rows: $count");
                    }
                    $journalCount += $count;
                }

                # Delete each deleted object
                # DELETE FROM $dbname WHERE @primaryKeys=@values
                foreach my $row (@{$results{data}})
                {
                    my @conditions = ();
                    foreach my $index (0 .. $#primaryKeys)
                    {
                        my $primaryKey = $primaryKeys[$index];
                        my $dataType =
                          MAM::Cache->getAttributeProperty($obj,
                            $primaryKey, 'DataType');
                        my $delim = $dataType eq 'String' ? q{'} : q{};
                        my $condition = " AND " . toDBC($primaryKey);
                        $condition .=
                          defined ${$row}[$index]
                          ? "=${delim}${$row}[$index]${delim}"
                          : " IS NULL";
                        push @conditions, $condition;
                    }
                    $sql = "DELETE FROM $dbname WHERE 1=1";
                    $sql .= join '', @conditions;
                    if ($log->is_trace())
                    {
                        $log->trace("SQL Update: $sql");
                    }
                    $count = $dbh->do($sql);
                    if ($log->is_debug())
                    {
                        $log->debug("SQL Rows: $count");
                    }
                    $objectCount += $count;
                }
            }

            # Delete old journal entries

# For each current object (active or deleted), find the request id
# (in the union with the log) immediately prior to the reference
# transaction (call it the primordial id) and delete all previous to it
# SELECT MAX(combined.g_request_id),combined.$primaryKeys FROM (SELECT * FROM $dbname UNION ALL SELECT * FROM ${dbname}_log) AS combined WHERE combined.g_request_id<$minRequestId GROUP BY combined.$primaryKeys
            $sql = "SELECT MAX(combined.g_request_id)";
            $sql .=
              (',' . join(',', map { 'combined.' . toDBC($_) } @primaryKeys))
              if @primaryKeys;
            $sql .=
              " FROM (SELECT * FROM $dbname UNION ALL SELECT * FROM ${dbname}_log) AS combined WHERE combined.g_request_id<$minRequestId";
            $sql .=
              (' GROUP BY '
                  . join(',', map { 'combined.' . toDBC($_) } @primaryKeys))
              if @primaryKeys;
            if ($log->is_trace())
            {
                $log->trace("SQL Query: $sql");
            }
            $sth = $dbh->prepare($sql);
            $sth->execute();
            %results = ();
            $results{cols} = $sth->{NUM_OF_FIELDS};
            $results{names} = [map { toUCC($_) } @{$sth->{NAME_lc}}];
            $results{data}  = $sth->fetchall_arrayref();
            $results{rows}  = @{$results{data}};

            # Log the results
            if ($log->is_debug())
            {
                my $entry = '';
                foreach my $row (@{$results{data}})
                {
                    $entry .=
                      join('|', map { defined $_ ? $_ : '' } @$row) . "\n";
                }
                $log->debug("SQL Results: $entry");
            }

            # Delete all logs previous to the primordial id for each object
            # DELETE FROM ${dbname}_log WHERE g_request_id<$primordialId
            foreach my $row (@{$results{data}})
            {
                my $primordialId = shift @$row;
                my @conditions   = ();
                foreach my $index (0 .. $#primaryKeys)
                {
                    my $primaryKey = $primaryKeys[$index];
                    my $dataType =
                      MAM::Cache->getAttributeProperty($obj, $primaryKey,
                        'DataType');
                    my $delim = $dataType eq 'String' ? q{'} : q{};
                    my $condition = " AND " . toDBC($primaryKey);
                    $condition .=
                      defined ${$row}[$index]
                      ? "=${delim}${$row}[$index]${delim}"
                      : " IS NULL";
                    push @conditions, $condition;
                }
                $sql =
                  "DELETE FROM ${dbname}_log WHERE g_request_id<$primordialId";
                $sql .= join '', @conditions;
                if ($log->is_trace())
                {
                    $log->trace("SQL Update: $sql");
                }
                $count = $dbh->do($sql);
                if ($log->is_debug())
                {
                    $log->debug("SQL Rows: $count");
                }
                $journalCount += $count;
            }
        }
    }

    my $message = "Successfully pruned $objectCount object records";
    $message .= " and $journalCount journal records" if $journalCount;
    return new MAM::Response()->success($objectCount + $journalCount, $message);
}

# ----------------------------------------------------------------------------
# $response = usage($request);
# ----------------------------------------------------------------------------

# Display Usage
sub usage
{
    my ($class, $request) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $rawAction   = $request->getAction();
    my @objects     = $request->getObjects();
    my $firstObject = $objects[0]->getName();
    my $showHidden  = $request->getOptionValue('ShowHidden');

    # Peel off the scope resolution operator if existent
    $rawAction =~ /^(\w*::)?(\w+)/;
    my ($scope, $action) = ($1, $2);

    # Add the request beginning
    my $usage = "<Request action=\"$rawAction\">\n";

    # Add the objects
    foreach my $object (@objects)
    {
        my $name = $object->getName();
        $usage .= "    <Object>$name</Object>\n";
    }

    # Add the selections
    if ($action eq 'Query')
    {
        foreach my $object (@objects)
        {
            my $name = $object->getName();
            foreach my $attribute (MAM::Cache->listAttributes($name))
            {
                my $dataType =
                  MAM::Cache->getAttributeProperty($name, $attribute,
                    'DataType');
                my $hidden =
                  MAM::Cache->getAttributeProperty($name, $attribute, 'Hidden');

                # Only show hidden attributes if ShowHidden option is specified
                next
                  if ($hidden eq 'True'
                    && (! defined $showHidden || $showHidden eq 'False'));

                $usage .= "    [<Get name=\"$attribute\"";
                $usage .= " [op=\"Sort|Tros|Count|GroupBy";
                $usage .= '|Max|Min' if $dataType ne 'Boolean';
                $usage .= '|Sum|Average'
                  if ( $dataType eq 'Integer'
                    || $dataType eq 'Float'
                    || $dataType eq 'Currency');
                $usage .= "\"]";
                $usage .= " object=\"$name\"" if (@objects > 1);
                $usage .= "></Get>]\n";
            }
        }
    }

    # Add the assignments
    if ($action eq 'Create' || $action eq 'Modify')
    {
        foreach my $attribute (MAM::Cache->listAttributes($firstObject))
        {
            my %properties =
              MAM::Cache->getAttributeProperties($firstObject, $attribute);
            my $primaryKey   = $properties{'PrimaryKey'};
            my $required     = $properties{'Required'};
            my $fixed        = $properties{'Fixed'};
            my $dataType     = $properties{'DataType'};
            my $defaultValue = $properties{'DefaultValue'};
            my $description  = $properties{'Description'};
            my $hidden       = $properties{'Hidden'};

            # Only show hidden attributes if ShowHidden option is specified
            next
              if ($hidden eq 'True'
                && (! defined $showHidden || $showHidden eq 'False'));

            # Cannot make an assignment to an autogen attribute
            next if ($dataType eq 'AutoGen');

            # Cannot modify primary keys or fixed attributes
            next
              if ($action eq 'Modify'
                && ($primaryKey eq 'True' || $fixed eq 'True'));

            $usage .= '    ';
            $usage .= '['
              unless ($action eq 'Create' && $required eq 'True');
            $usage .= "<Set name=\"$attribute\"";
            $usage .= " [op=\"Assign";
            $usage .= '|Inc|Dec'
              if (
                (      $dataType eq 'Integer'
                    || $dataType eq 'Float'
                    || $dataType eq 'Currency'
                )
                && $action ne 'Create'
              );
            $usage .= " (Assign)\"]";
            $usage .= '>';
            $usage .= "{$description}" if ($dataType eq 'String');
            $usage .= 'True|False' if ($dataType eq 'Boolean');
            $usage .= '{Integer Number}' if ($dataType eq 'Integer');
            $usage .= '{Decimal Number}' if ($dataType eq 'Currency');
            $usage .= '{Floating Point Number}' if ($dataType eq 'Float');
            $usage .= 'YYYY-MM-DD[ hh:mm:ss]|-Infinity|Infinity|Now'
              if ($dataType eq 'TimeStamp');
            $usage .= " ($defaultValue)" if defined $defaultValue;
            $usage .= '</Set>';
            $usage .= ']'
              unless ($action eq 'Create' && $required eq 'True');
            $usage .= "\n";
        }
    }

    # Add the conditions
    if (   $action eq 'Query'
        || $action eq 'Modify'
        || $action eq 'Delete'
        || $action eq 'Undelete')
    {
        foreach my $object (@objects)
        {
            my $name = $object->getName();
            foreach my $attribute (MAM::Cache->listAttributes($name))
            {
                my %properties =
                  MAM::Cache->getAttributeProperties($name, $attribute);
                my $dataType    = $properties{'DataType'};
                my $description = $properties{'Description'};
                my $hidden      = $properties{'Hidden'};

                # Only show hidden attributes if ShowHidden option is specified
                next
                  if ($hidden eq 'True'
                    && (! defined $showHidden || $showHidden eq 'False'));

                $usage .= "    [<Where name=\"$attribute\"";
                $usage .= " [op=\"EQ|NE";
                $usage .= '|GT|GE|LT|LE' if ($dataType ne 'Boolean');
                $usage .= '|Match|NotMatch' if ($dataType eq 'String');
                $usage .= " (EQ)\"]";
                $usage .= " [conj=\"And|Or (And)\"]";
                $usage .= " [group=\"{Integer Number}\"]";
                $usage .= " object=\"$name\"" if (@objects > 1);
                $usage .= " [subject=\"{Object Name}\"]" if (@objects > 1);
                $usage .= '>';
                $usage .= "{$description}" if ($dataType eq 'String');
                $usage .= 'True|False' if ($dataType eq 'Boolean');
                $usage .= '{Integer Number}'
                  if ($dataType eq 'Integer' || $dataType eq 'AutoGen');
                $usage .= '{Decimal Number}' if ($dataType eq 'Currency');
                $usage .= '{Floating Point Number}'
                  if ($dataType eq 'Float');
                $usage .= 'YYYY-MM-DD[ hh:mm:ss]|-Infinity|Infinity|Now'
                  if ($dataType eq 'TimeStamp');
                $usage .= "</Where>]\n";
            }
            $usage .=
              "    [<Option name=\"Delete\">True|False (False)</Option>]\n"
              if $name eq 'Notification' && $action eq 'Query';
        }
    }
    if ($action eq 'Fire')
    {
        $usage .= "    <Option name=\"Id\">{Event Id}</Option>\n";
    }
    if ($action eq 'Prune')
    {
        $usage .=
          "    <Option name=\"RetentionTime\">YYYY-MM-DD[ hh:mm:ss]</Option>|<Option name=\"RetentionPeriod\">{Minimum Keep Period}[~|^|!]</Option>\n";
        $usage .= "    <Option name=\"Confirm\">True|False (False)</Option>\n";
    }
    if ($action eq 'Send')
    {
        $usage .= "    <Option name=\"Event\">{Event Id}</Option>\n";
        $usage .=
          "    <Option name=\"Status\">{Event Command Status}</Option>\n";
        $usage .=
          "    <Option name=\"Code\">{Event Command Exit Code}</Option>\n";
        $usage .=
          "    [<Option name=\"Message\">{Event Command Message}</Option>]\n";
        $usage .=
          "    [<Option name=\"Key\">{Object Primary Key Value}</Option>]\n";
        $usage .=
          "    [<Option name=\"Recipient\">{Recipient To Notify}</Option>]\n";
        $usage .=
          "    [<Option name=\"EndTime\">YYYY-MM-DD[ hh:mm:ss]|-Infinity|Infinity|Now</Option>]\n";
        $usage .=
          "    [<Option name=\"DeliveryMethod\">Store (Store)</Option>]\n";
    }

    # Add the options
    if ($action eq 'Query')
    {
        if ($objects[0]->getName() =~ /Attribute/)
        {
            $usage .=
              "    [<Option name=\"SortBySequence\">True|False (False)</Option>]\n";
        }
        $usage .=
          "    [<Option name=\"Time\">YYYY-MM-DD[ hh:mm:ss]</Option>]\n";
        $usage .= "    [<Option name=\"Unique\">True|False (False)</Option>]\n";
        $usage .= "    [<Option name=\"ChunkSize\">{Integer Number}</Option>\n";
        $usage .= "    [<Option name=\"Limit\">{Integer Number}</Option>\n";
        $usage .= "    [<Option name=\"Offset\">{Integer Number}</Option>]]\n";
        $usage .=
          "    [<Option name=\"ShowHidden\">True|False (False)</Option>]\n";
    }
    $usage .= "    [<Option name=\"ShowUsage\">True|False (False)</Option>]\n";

    # Add the request end
    $usage .= "</Request>\n";

    my $response = new MAM::Response(
        status  => 'Success',
        code    => '010',
        message => $usage
    );
    return $response;
}

1;
