#! /usr/bin/perl -wT
################################################################################
#
# MAM Cache object
#
# File   :  Cache.pm
#
################################################################################
#
#    This file includes parts which are copyrighted by Adaptive Computing
#    Enterprises, Inc., and parts which are copyrighted by Battelle Memorial
#    Institute.
#
#    The terms and conditions for the use and redistribution of these parts
#    are governed by the Moab Accounting Manager License and the BSD License
#    respectively. Refer to the LICENSE file for details.
#
#    Copyright (C) 2006 - 2016 by Adaptive Computing Enterprises, Inc.
#    All Rights Reserved.
#
#    The Moab Accounting Manager License specifies the terms and conditions
#    for use and redistribution.
#
#    The Moab Accounting Manager License applies to the Moab Accounting
#    Manager software offered by Adaptive Computing Enterprises, Inc.
#    By installing or using this software, Licensee accepts a non-exclusive
#    license from Adaptive Computing Enterprises, Inc. and is bound to accept
#    acknowledgement of and abide by the notices and conditions of the
#    Moab Accounting Manager License.
#
#    Copyright (C) 2003 - 2005 Pacific Northwest National Laboratory,
#    Battelle Memorial Institute. All rights reserved.
#
#    The BSD License specifies the terms and conditions for use and
#    redistribution.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions
#    are met:
#
#    - Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
#    - Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
#    - Neither the name of Battelle nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
#    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
#    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
#    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#    POSSIBILITY OF SUCH DAMAGE.
#
################################################################################

use strict;

=head1 NAME

MAM::Cache - an object metadata cache for quick object property checks

=head1 DESCRIPTION

The B<MAM::Cache> module acts as a metadata cache for properties of objects, actions, attributes, values, roles, passwords and other core objects needed for quick lookups. When any of the cached info is updated, the cache is marked as stale and repopulated from the database. This is a static class.

=head1 ACCESSORS

=over 4

=item $boolean = MAM::Cache->getStale();

=back

=head1 MUTATORS

=over 4

=item MAM::Cache->setStale($boolean);

=back

=head1 OTHER METHODS

=over 4

=item MAM::Cache->populate($database);

=item $boolean = MAM::Cache->objectExists($object)

=item @objects = MAM::Cache->associationLookup($parent,$child)

=item @objects = MAM::Cache->listObjects()

=item $value = MAM::Cache->getObjectProperty($object, $property)

=item %properties = MAM::Cache->getObjectProperties($object)

=item $boolean = MAM::Cache->attributeExists($object, $attribute)

=item @attributes = MAM::Cache->listAttributes($object)

=item @attributes = MAM::Cache->listAttributesWithPropertyValue($object, $property, $value)

=item @attributes = MAM::Cache->listAttributesWithoutPropertyValue($object, $property, $value)

=item $value = MAM::Cache->getAttributeProperty($object, $attribute, $property)

=item %properties = MAM::Cache->getAttributeProperties($object, $attribute)

=item $boolean = MAM::Cache->actionExists($object, $action)

=item @actions = MAM::Cache->listActions($object)

=item $value = MAM::Cache->getActionProperty($object, $action, $property)

=item $boolean = MAM::Cache->roleExists($role)

=item @roles = MAM::Cache->listRoles()

=item $value = MAM::Cache->getRoleProperty($role, $property)

=item $boolean = MAM::Cache->roleActionExists($role, $object, $action, $instance)

=item @roleActions = MAM::Cache->listRoleActions($role, $object, $instance)

=item @roleActionInstances = MAM::Cache->listRoleActionInstances($role, $object, $action, $instance)

=item $value = MAM::Cache->getRoleActionProperty($role, $object, $action, $instance, $property)

=item $boolean = MAM::Cache->roleUserExists($role, $user)

=item @roleUsers = MAM::Cache->listRoleUsers($role)

=item @userRoles = MAM::Cache->listUserRoles($user)

=item $value = MAM::Cache->getRoleUserProperty($role, $user, $property)

=item $boolean = MAM::Cache->passwordExists($password)

=item @passwords = MAM::Cache->listPasswords()

=item $value = MAM::Cache->getPasswordProperty($password, $property)

=item MAM::Cache->toString();

=back

=head1 EXAMPLES

use MAM::Cache;

MAM::Cache->populate($database);

=head1 REQUIRES

Perl 5.8

=head1 AUTHOR

Scott Jackson, scottmo@adaptivecomputing.com

=cut

##############################################################################

package MAM::Cache;

use vars qw($log);
use Data::Dumper;
use Error qw(:try);
use MAM::Database;
use MAM::Exception;
use MAM::Global;

# ----------------------------------------------------------------------------
# Static Variables
# ----------------------------------------------------------------------------

my $_stale = 1;
my %_cache;

# ----------------------------------------------------------------------------
# Accessors
# ----------------------------------------------------------------------------

# Get whether the cache is stale
sub getStale
{
    my ($class) = @_;
    return $_stale;
}

# ----------------------------------------------------------------------------
# Mutators
# ----------------------------------------------------------------------------

# Set whether the cache is stale
sub setStale
{
    my ($class, $stale) = @_;
    $_stale = $stale;
}

# ----------------------------------------------------------------------------
# $string = toString()
# ----------------------------------------------------------------------------

# Serialize message to printable string
sub toString
{
    my ($class) = @_;

    return Data::Dumper->Dump([\%_cache], ['Cache']);
}

# ----------------------------------------------------------------------------
# populate($database)
# ----------------------------------------------------------------------------

# Populates the metadata cache
# It should be passed a database object
sub populate
{
    my ($class, $database) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    # Try to obtain a database handle if we don't already have one
    unless ($database)
    {
        try
        {
            $database = new MAM::Database();
        }
        catch Error with
        {
            throw MAM::Exception('730',
                'Cache population failed obtaining database handle: ' . $_[0]);
        };
    }

    # Clear Previous Cache
    %_cache = ();

    # Select objects from database
    my $objectResults = $database->select(object => 'Object');
    my $objectCols = $objectResults->{cols};

    # Iterate over all objects
    foreach my $objectRow (@{$objectResults->{data}})
    {
        my $objectName = $objectRow->[0];

        # Populate Object Properties
        for (my $i = 0; $i < $objectCols; $i++)
        {
            my $propertyName  = $objectResults->{names}->[$i];
            my $propertyValue = $objectRow->[$i];

            $_cache{'Objects'}->{$objectName}->{'Properties'}->{$propertyName} =
              $propertyValue;
        }

        # Select attributes for this object from the database
        my $attributeResults = $database->select(
            object => 'Attribute',
            conditions =>
              [new MAM::Condition(name => 'Object', value => $objectName)],
            options =>
              [new MAM::Option(name => 'SortBySequence', value => 'True')]
        );
        my $attributeCols = $attributeResults->{cols};

        # Iterate over all attributes for this object
        foreach my $attributeRow (@{$attributeResults->{data}})
        {
            my $attributeName = $attributeRow->[1];

            # Populate Object Attributes
            for (my $i = 0; $i < $attributeCols; $i++)
            {
                my $propertyName  = $attributeResults->{names}->[$i];
                my $propertyValue = $attributeRow->[$i];

                $_cache{'Objects'}->{$objectName}->{'Attributes'}
                  ->{$attributeName}->{$propertyName} = $propertyValue;
            }
        }

        # Select actions for this object from the database
        my $actionResults = $database->select(
            object => 'Action',
            conditions =>
              [new MAM::Condition(name => 'Object', value => $objectName)]
        );
        my $actionCols = $actionResults->{cols};

        # Iterate over all actions for this object
        foreach my $actionRow (@{$actionResults->{data}})
        {
            my $actionName = $actionRow->[1];

            # Populate Object Actions
            for (my $i = 0; $i < $actionCols; $i++)
            {
                my $propertyName  = $actionResults->{names}->[$i];
                my $propertyValue = $actionRow->[$i];

                $_cache{'Objects'}->{$objectName}->{'Actions'}->{$actionName}
                  ->{$propertyName} = $propertyValue;
            }
        }
    }

    # Select roles from database
    my $roleResults = $database->select(object => 'Role');
    my $roleCols = $roleResults->{cols};

    # Iterate over all roles
    foreach my $roleRow (@{$roleResults->{data}})
    {
        my $roleName = $roleRow->[0];

        # Populate Role Properties
        for (my $i = 0; $i < $roleCols; $i++)
        {
            my $propertyName  = $roleResults->{names}->[$i];
            my $propertyValue = $roleRow->[$i];

            $_cache{'Roles'}->{$roleName}->{'Properties'}->{$propertyName} =
              $propertyValue;
        }

        # Select actions for this role from database
        my $roleActionResults = $database->select(
            object => 'RoleAction',
            conditions =>
              [new MAM::Condition(name => 'Role', value => $roleName)]
        );
        my $roleActionCols = $roleActionResults->{cols};

        # Iterate over all actions for this role
        foreach my $roleActionRow (@{$roleActionResults->{data}})
        {
            my $roleActionName = $roleActionRow->[1] . $roleActionRow->[2];

            # Populate Role Actions
            for (my $i = 0; $i < $roleActionCols; $i++)
            {
                my $propertyName  = $roleActionResults->{names}->[$i];
                my $propertyValue = $roleActionRow->[$i];

                $_cache{'Roles'}->{$roleName}->{'RoleActions'}
                  ->{$roleActionName}->{$propertyName} = $propertyValue;
            }
        }

        # Select users for this role from database
        my $roleUserResults = $database->select(
            object => 'RoleUser',
            conditions =>
              [new MAM::Condition(name => 'Role', value => $roleName)]
        );
        my $roleUserCols = $roleUserResults->{cols};

        # Iterate over all users for this role
        foreach my $roleUserRow (@{$roleUserResults->{data}})
        {
            my $roleUserName = $roleUserRow->[1];

            # Populate Role Users
            for (my $i = 0; $i < $roleUserCols; $i++)
            {
                my $propertyName  = $roleUserResults->{names}->[$i];
                my $propertyValue = $roleUserRow->[$i];

                $_cache{'Roles'}->{$roleName}->{'RoleUsers'}->{$roleUserName}
                  ->{$propertyName} = $propertyValue;
            }
        }
    }

    # Select passwords from database
    my $passwordResults = $database->select(object => 'Password');
    my $passwordCols = $passwordResults->{cols};

    # Iterate over all passwords in database
    foreach my $passwordRow (@{$passwordResults->{data}})
    {
        my $userName = $passwordRow->[0];

        # Populate Password Properties
        for (my $i = 0; $i < $passwordCols; $i++)
        {
            my $propertyName  = $passwordResults->{names}->[$i];
            my $propertyValue = $passwordRow->[$i];

            $_cache{'Passwords'}->{$userName}->{'Properties'}->{$propertyName} =
              $propertyValue;
        }
    }

    # Mark metadata as no longer stale
    $_stale = 0;

    # Log Cache DOM
    if ($log->is_trace())
    {
        $log->trace("Cache DOM:\n" . MAM::Cache->toString());
    }
}

# ----------------------------------------------------------------------------
# $boolean = objectExists($object)
# ----------------------------------------------------------------------------

# Check if Object Exists
sub objectExists
{
    my ($class, $object) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    return exists $_cache{'Objects'}->{$object};
}

# ----------------------------------------------------------------------------
# @objects = associationLookup($parent,$child)
# ----------------------------------------------------------------------------

# Perform Association Lookup
sub associationLookup
{
    my ($class, $parent, $child) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (',
            join(', ', map { defined $_ ? $_ : 'undef' } @_[1 .. $#_]), ')');
    }

    my @objects = ();
    foreach my $object (keys %{$_cache{'Objects'}})
    {
        my $objectHashRef = $_cache{'Objects'}->{$object};
        my $association   = $objectHashRef->{'Properties'}->{'Association'};
        my $parentAttr    = $objectHashRef->{'Properties'}->{'Parent'};
        my $childAttr     = $objectHashRef->{'Properties'}->{'Child'};
        if (
               defined $association
            && $association eq 'True'
            && (! defined $parent
                || (defined $parentAttr && $parentAttr eq $parent))
            && (! defined $child
                || (defined $childAttr && $childAttr eq $child))
          )
        {
            push @objects, $object;
        }
    }
    return @objects;
}

# ----------------------------------------------------------------------------
# @objects = listObjects()
# ----------------------------------------------------------------------------

# List Objects
sub listObjects
{
    my ($class) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    return sort keys %{$_cache{'Objects'}};
}

# ----------------------------------------------------------------------------
# $value = getObjectProperty($object, $property)
# ----------------------------------------------------------------------------

# Get Object Property
sub getObjectProperty
{
    my ($class, $object, $property) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    return (  exists $_cache{'Objects'}->{$object}
          and exists $_cache{'Objects'}->{$object}->{'Properties'}->{$property})
      ? $_cache{'Objects'}->{$object}->{'Properties'}->{$property}
      : undef;
}

# ----------------------------------------------------------------------------
# %properties = getObjectProperties($object)
# ----------------------------------------------------------------------------

# Get Object Properties
sub getObjectProperties
{
    my ($class, $object) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    return
      exists $_cache{'Objects'}->{$object}
      ? %{$_cache{'Objects'}->{$object}->{'Properties'}}
      : undef;
}

# ----------------------------------------------------------------------------
# $boolean = attributeExists($object, $attribute)
# ----------------------------------------------------------------------------

# Check if Attribute Exists
sub attributeExists
{
    my ($class, $object, $attribute) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    return (  exists $_cache{'Objects'}->{$object}
          and
          exists $_cache{'Objects'}->{$object}->{'Attributes'}->{$attribute});
}

# ----------------------------------------------------------------------------
# @attributes = listAttributes($object)
# ----------------------------------------------------------------------------

# List Attributes
sub listAttributes
{
    my ($class, $object) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    return
      exists $_cache{'Objects'}->{$object}
      ? sort keys %{$_cache{'Objects'}->{$object}->{'Attributes'}}
      : ();
}

# ----------------------------------------------------------------------------
# @attributes = listAttributesWithPropertyValue($object, $property, $value)
# ----------------------------------------------------------------------------

# List Attributes With Property Value
sub listAttributesWithPropertyValue
{
    my ($class, $object, $property, $value) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', map { defined $_ ? $_ : 'undef' } @_[1 .. $#_]), ')');
    }

    my @attributeList = ();
    if (exists $_cache{'Objects'}->{$object})
    {
        foreach my $attribute (
            sort keys %{$_cache{'Objects'}->{$object}->{'Attributes'}})
        {
            if (
                exists $_cache{'Objects'}->{$object}->{'Attributes'}
                ->{$attribute}->{$property})
            {
                if (defined $value)
                {
                    if ($_cache{'Objects'}->{$object}->{'Attributes'}
                        ->{$attribute}->{$property} eq $value)
                    {
                        push @attributeList, $attribute;
                    }
                }
                else
                {
                    unless (
                        defined $_cache{'Objects'}->{$object}->{'Attributes'}
                        ->{$attribute}->{$property})
                    {
                        push @attributeList, $attribute;
                    }
                }
            }
        }
    }
    return @attributeList;
}

# ----------------------------------------------------------------------------
# @attributes = listAttributesWithoutPropertyValue($object, $property, $value)
# ----------------------------------------------------------------------------

# List Attributes Without Property Value
sub listAttributesWithoutPropertyValue
{
    my ($class, $object, $property, $value) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', map { defined $_ ? $_ : 'undef' } @_[1 .. $#_]), ')');
    }

    my @attributeList = ();
    if (exists $_cache{'Objects'}->{$object})
    {
        foreach my $attribute (
            sort keys %{$_cache{'Objects'}->{$object}->{'Attributes'}})
        {
            if (
                exists $_cache{'Objects'}->{$object}->{'Attributes'}
                ->{$attribute}->{$property})
            {
                if (defined $value)
                {
                    if ($_cache{'Objects'}->{$object}->{'Attributes'}
                        ->{$attribute}->{$property} ne $value)
                    {
                        push @attributeList, $attribute;
                    }
                }
                else
                {
                    if (
                        defined $_cache{'Objects'}->{$object}->{'Attributes'}
                        ->{$attribute}->{$property})
                    {
                        push @attributeList, $attribute;
                    }
                }
            }
        }
    }
    return @attributeList;
}

# ----------------------------------------------------------------------------
# $value = getAttributeProperty($object, $attribute, $property)
# ----------------------------------------------------------------------------

# Get Attribute Property
sub getAttributeProperty
{
    my ($class, $object, $attribute, $property) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    return (  exists $_cache{'Objects'}->{$object}
          and exists $_cache{'Objects'}->{$object}->{'Attributes'}->{$attribute}
          and exists $_cache{'Objects'}->{$object}->{'Attributes'}->{$attribute}
          ->{$property})
      ? $_cache{'Objects'}->{$object}->{'Attributes'}->{$attribute}->{$property}
      : undef;
}

# ----------------------------------------------------------------------------
# %properties = getAttributeProperties($object, $attribute)
# ----------------------------------------------------------------------------

# Get Attribute Properties
sub getAttributeProperties
{
    my ($class, $object, $attribute) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    return (  exists $_cache{'Objects'}->{$object}
          and
          exists $_cache{'Objects'}->{$object}->{'Attributes'}->{$attribute})
      ? %{$_cache{'Objects'}->{$object}->{'Attributes'}->{$attribute}}
      : ();
}

# ----------------------------------------------------------------------------
# $boolean = actionExists($object, $action)
# ----------------------------------------------------------------------------

# Check if Action Exists
sub actionExists
{
    my ($class, $object, $action) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    return (  exists $_cache{'Objects'}->{$object}
          and exists $_cache{'Objects'}->{$object}->{'Actions'}->{$action});
}

# ----------------------------------------------------------------------------
# @actions = listActions($object)
# ----------------------------------------------------------------------------

# List Actions
sub listActions
{
    my ($class, $object) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    return
      exists $_cache{'Objects'}->{$object}
      ? sort keys %{$_cache{'Objects'}->{$object}->{'Actions'}}
      : ();
}

# ----------------------------------------------------------------------------
# $value = getActionProperty($object, $action, $property)
# ----------------------------------------------------------------------------

# Get Action Property
sub getActionProperty
{
    my ($class, $object, $action, $property) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    return (  exists $_cache{'Objects'}->{$object}
          and exists $_cache{'Objects'}->{$object}->{'Actions'}->{$action}
          and exists $_cache{'Objects'}->{$object}->{'Actions'}->{$action}
          ->{$property})
      ? $_cache{'Objects'}->{$object}->{'Actions'}->{$action}->{$property}
      : undef;
}

# ----------------------------------------------------------------------------
# $boolean = roleExists($role)
# ----------------------------------------------------------------------------

# Check if Role Exists
sub roleExists
{
    my ($class, $role) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    return exists $_cache{'Roles'}->{$role};
}

# ----------------------------------------------------------------------------
# @roles = listRoles()
# ----------------------------------------------------------------------------

# List Roles
sub listRoles
{
    my ($class) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    return sort keys %{$_cache{'Roles'}};
}

# ----------------------------------------------------------------------------
# $value = getRoleProperty($role, $property)
# ----------------------------------------------------------------------------

# Get Role Property
sub getRoleProperty
{
    my ($class, $role, $property) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    return (  exists $_cache{'Roles'}->{$role}
          and exists $_cache{'Roles'}->{$role}->{'Properties'}->{$property})
      ? $_cache{'Roles'}->{$role}->{'Properties'}->{$property}
      : undef;
}

# ----------------------------------------------------------------------------
# $boolean = roleActionExists($role, $object, $action, $instance)
# ----------------------------------------------------------------------------

# Check if Role Action Exists
sub roleActionExists
{
    my ($class, $role, $object, $action, $instance) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    return (  exists $_cache{'Roles'}->{$role}
          and
          exists $_cache{'Roles'}->{$role}->{'RoleActions'}->{$object . $action}
          and $_cache{'Roles'}->{$role}->{'RoleActions'}->{$object . $action}
          ->{'Instance'} eq $instance);
}

# ----------------------------------------------------------------------------
# @roleActions = listRoleActions($role, $object, $instance)
# ----------------------------------------------------------------------------

# List Role Actions
sub listRoleActions
{
    my ($class, $role, $object, $instance) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my @roleActionList = ();
    if (exists $_cache{'Roles'}->{$role})
    {
        foreach my $roleAction (
            sort keys %{$_cache{'Roles'}->{$role}->{'RoleActions'}})
        {
            if ($_cache{'Roles'}->{$role}->{'RoleActions'}->{$roleAction}
                ->{'Object'} eq $object
                && $_cache{'Roles'}->{$role}->{'RoleActions'}->{$roleAction}
                ->{'Instance'} eq $instance)
            {
                push @roleActionList,
                  $_cache{'Roles'}->{$role}->{'RoleActions'}->{$roleAction}
                  ->{'Name'};
            }
        }
    }
    return @roleActionList;
}

# ----------------------------------------------------------------------------
# @roleActionInstances = listRoleActionInstances($role, $object, $action, $instance)
# ----------------------------------------------------------------------------

# List Role Action Instances by Instance Type
sub listRoleActionInstances
{
    my ($class, $role, $object, $action, $instance) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my @roleActionInstanceList = ();
    if (exists $_cache{'Roles'}->{$role})
    {
        foreach my $roleAction (
            sort keys %{$_cache{'Roles'}->{$role}->{'RoleActions'}})
        {
            my $objectAttr =
              $_cache{'Roles'}->{$role}->{'RoleActions'}->{$roleAction}
              ->{'Object'};
            my $actionAttr =
              $_cache{'Roles'}->{$role}->{'RoleActions'}->{$roleAction}
              ->{'Name'};
            my $instanceAttr =
              $_cache{'Roles'}->{$role}->{'RoleActions'}->{$roleAction}
              ->{'Instance'};
            if (   ($objectAttr eq $object || $objectAttr eq 'ANY')
                && ($actionAttr eq $action || $actionAttr eq 'ANY'))
            {
                if (
                       ($instance eq 'ANY' && $instanceAttr eq 'ANY')
                    || ($instance eq 'SELF' && $instanceAttr eq 'SELF')
                    || (   $instance eq 'MEMBERS'
                        && $instanceAttr eq 'MEMBERS')
                    || (   $instance eq 'ADMIN'
                        && $instanceAttr eq 'ADMIN')
                    || (
                        $instance eq 'INSTANCE'
                        && (   $instanceAttr ne 'ANY'
                            && $instanceAttr ne 'SELF'
                            && $instanceAttr ne 'MEMBERS'
                            && $instanceAttr ne 'ADMIN')
                    )
                  )
                {
                    push(@roleActionInstanceList, $instanceAttr);
                }
            }
        }
    }
    return @roleActionInstanceList;
}

# ----------------------------------------------------------------------------
# $value = getRoleActionProperty($role, $object, $action, $instance, $property)
# ----------------------------------------------------------------------------

# Get Role Action Property
sub getRoleActionProperty
{
    my ($class, $role, $object, $action, $instance, $property) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    if (    exists $_cache{'Roles'}->{$role}
        and
        exists $_cache{'Roles'}->{$role}->{'RoleActions'}->{$object . $action}
        and $_cache{'Roles'}->{$role}->{'RoleActions'}->{$object . $action}
        ->{'Instance'} eq $instance)
    {
        return $_cache{'Roles'}->{$role}->{'RoleActions'}->{$object . $action}
          ->{$property};
    }
    return;
}

# ----------------------------------------------------------------------------
# $boolean = roleUserExists($role, $user)
# ----------------------------------------------------------------------------

# Check if Role User Exists
sub roleUserExists
{
    my ($class, $role, $user) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    return (  exists $_cache{'Roles'}->{$role}
          and exists $_cache{'Roles'}->{$role}->{'RoleUsers'}->{$user});
}

# ----------------------------------------------------------------------------
# @roleUsers = listRoleUsers($role)
# ----------------------------------------------------------------------------

# List Role Users
sub listRoleUsers
{
    my ($class, $role) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    return
      exists $_cache{'Roles'}->{$role}
      ? sort keys %{$_cache{'Roles'}->{$role}->{'RoleUsers'}}
      : undef;
}

# ----------------------------------------------------------------------------
# @userRoles = listUserRoles($user)
# ----------------------------------------------------------------------------

# List User Roles
sub listUserRoles
{
    my ($class, $user) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my @userRoleList = ();
    foreach my $role (sort keys %{$_cache{'Roles'}})
    {
        if (    exists $_cache{'Roles'}->{$role}
            and exists $_cache{'Roles'}->{$role}->{'RoleUsers'}->{$user})
        {
            push @userRoleList, $role;
        }
    }
    return @userRoleList;
}

# ----------------------------------------------------------------------------
# $value = getRoleUserProperty($role, $user, $property)
# ----------------------------------------------------------------------------

# Get Role User Property
sub getRoleUserProperty
{
    my ($class, $role, $user, $property) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    return (  exists $_cache{'Roles'}->{$role}
          and exists $_cache{'Roles'}->{$role}->{'RoleUsers'}->{$user}
          and
          exists $_cache{'Roles'}->{$role}->{'RoleUsers'}->{$user}->{$property})
      ? $_cache{'Roles'}->{$role}->{'RoleUsers'}->{$user}->{$property}
      : undef;
}

# ----------------------------------------------------------------------------
# $boolean = passwordExists($user)
# ----------------------------------------------------------------------------

# Check if Password Exists
sub passwordExists
{
    my ($class, $user) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    return exists $_cache{'Passwords'}->{$user};
}

# ----------------------------------------------------------------------------
# @passwords = listPasswords()
# ----------------------------------------------------------------------------

# List Passwords
sub listPasswords
{
    my ($class) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    return sort keys %{$_cache{'Passwords'}};
}

# ----------------------------------------------------------------------------
# $value = getPasswordProperty($password, $property)
# ----------------------------------------------------------------------------

# Get Password Property
sub getPasswordProperty
{
    my ($class, $user, $property) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    return (  exists $_cache{'Passwords'}->{$user}
          and exists $_cache{'Passwords'}->{$user}->{'Properties'}->{$property})
      ? $_cache{'Passwords'}->{$user}->{'Properties'}->{$property}
      : undef;
}

1;
