#! /usr/bin/perl -wT
################################################################################
#
# MAM Chunk object
#
# File   :  Chunk.pm
#
################################################################################
#
#    This file includes parts which are copyrighted by Adaptive Computing
#    Enterprises, Inc., and parts which are copyrighted by Battelle Memorial
#    Institute.
#
#    The terms and conditions for the use and redistribution of these parts
#    are governed by the Moab Accounting Manager License and the BSD License
#    respectively. Refer to the LICENSE file for details.
#
#    Copyright (C) 2006 - 2016 by Adaptive Computing Enterprises, Inc.
#    All Rights Reserved.
#
#    The Moab Accounting Manager License specifies the terms and conditions
#    for use and redistribution.
#
#    The Moab Accounting Manager License applies to the Moab Accounting
#    Manager software offered by Adaptive Computing Enterprises, Inc.
#    By installing or using this software, Licensee accepts a non-exclusive
#    license from Adaptive Computing Enterprises, Inc. and is bound to accept
#    acknowledgement of and abide by the notices and conditions of the
#    Moab Accounting Manager License.
#
#    Copyright (C) 2003 - 2005 Pacific Northwest National Laboratory,
#    Battelle Memorial Institute. All rights reserved.
#
#    The BSD License specifies the terms and conditions for use and
#    redistribution.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions
#    are met:
#
#    - Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
#    - Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
#    - Neither the name of Battelle nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
#    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
#    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
#    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#    POSSIBILITY OF SUCH DAMAGE.
#
################################################################################

use strict;

=head1 NAME

MAM::Chunk - services MAM chunks

=head1 DESCRIPTION

The B<MAM::Chunk> module defines functions to service MAM chunks. A chunk is an individual communication and may contain a request or a response (or part of a response when segmentation is required). A chunk (when rendered in XML) consists of the envelope with its body and optional signature. The chunk object handles authentication, encryption and the insertion/extraction of the request or response.

=head1 CONSTRUCTORS

 my $chunk = new MAM::Chunk(request => $request, $response => $response, host => $serverHost, port => $serverPort, backup => $backupHost, authentication => $boolean, encryption => $boolean, wireProtocol => $wireProtocol);

=head1 ACCESSORS

=over 4

=item $request = $chunk->getRequest();

=item $response = $chunk->getResponse();

=item $serverHost = $chunk->getHost();

=item $serverPort = $chunk->getPort();

=item $backupHost = $chunk->getBackup();

=item $boolean = $chunk->getAuthentication();

=item $boolean = $chunk->getEncryption();

=item $wireProtocol = $chunk->getWireProtocol();

=back

=head1 MUTATORS

=over 4

=item $chunk = $chunk->setRequest($request);

=item $chunk = $chunk->setResponse($response);

=item $chunk = $chunk->setHost($serverHost);

=item $chunk = $chunk->setPort($serverPort);

=item $chunk = $chunk->setBackup($backupHost);

=item $chunk = $chunk->setAuthentication($boolean);

=item $chunk = $chunk->setEncryption($boolean);

=item $chunk = $chunk->setWireProtocol($wireProtocol);

=back

=head1 OTHER METHODS

=over 4

=item $replyChunk = $messageChunk->getChunk();

=item $string = $chunk->toString();

=item sign($doc);

=item orient($doc);

=item authenticate($doc);

=item encrypt($doc);

=item decrypt($doc);

=back

=head1 EXAMPLES

use MAM::Chunk;

my $messageChunk = new MAM::Chunk();
$messageChunk->setRequest($request);
my $replyChunk = $messageChunk->getChunk();
my $response = $replyChunk->getResponse();

=head1 REQUIRES

Perl 5.8

=head1 AUTHOR

Scott Jackson, scottmo@adaptivecomputing.com

=cut

##############################################################################

package MAM::Chunk;

use vars qw($log $SBIN_DIR);
use Compress::Zlib;
use Config::Tiny;
use Crypt::CBC;
use Digest::MD5 qw(md5_hex);
use Digest::SHA qw(hmac_sha1 hmac_sha512 sha1 sha512);
use Encode qw(encode_utf8);
use Error qw(:try);
use IPC::Open3;
use MIME::Base64;
use Symbol;
use XML::LibXML;
use MAM::Exception;
use MAM::Global;
use MAM::Message;
use MAM::Request;
use MAM::Response;
use MAM::Reply;
use feature "switch";

# ----------------------------------------------------------------------------
# Constructor
# ----------------------------------------------------------------------------

sub new
{
    my ($class, %arg) = @_;

    # Instantiate the object
    my $self = {
        _request        => $arg{request},           # SCALAR
        _response       => $arg{response},          # SCALAR
        _host           => $arg{host},              # SCALAR
        _port           => $arg{port},              # SCALAR
        _backup         => $arg{backup},            # SCALAR
        _authentication => $arg{authentication},    # SCALAR
        _encryption     => $arg{encryption},        # SCALAR
        _wireProtocol   => $arg{wireProtocol},      # SCALAR
    };
    bless $self, $class;

    if (! defined($self->{_authentication}))
    {
        # Set authentication from config file
        if ((      $clientConfig->{'security.authentication'}
                || $SECURITY_AUTHENTICATION
            ) =~ /true/i
          )
        {
            $self->{_authentication} = 1;
        }
        else
        {
            $self->{_authentication} = 0;
        }
    }

    if (! defined($self->{_encrypted}))
    {
        # Set encryption from config file
        if (($clientConfig->{'security.encryption'} || $SECURITY_ENCRYPTION) =~
            /true/i)
        {
            $self->{_encryption} = 1;
        }
        else
        {
            $self->{_encryption} = 0;
        }
    }

    if (! defined($self->{_wireProtocol}))
    {
        # Set wireProtocol from config file
        $self->{_wireProtocol} = $clientConfig->{'wire.protocol'}
          || $WIRE_PROTOCOL;
    }

    return $self;
}

# ----------------------------------------------------------------------------
# $request = getRequest();
# ----------------------------------------------------------------------------

# Get the request
sub getRequest
{
    my ($self) = @_;
    return $self->{_request};
}

# ----------------------------------------------------------------------------
# $response = getResponse();
# ----------------------------------------------------------------------------

# Get the response
sub getResponse
{
    my ($self) = @_;
    return $self->{_response};
}

# ----------------------------------------------------------------------------
# $serverHost = getHost();
# ----------------------------------------------------------------------------

# Get the server host
sub getHost
{
    my ($self) = @_;
    return $self->{_host};
}

# ----------------------------------------------------------------------------
# $serverPort = getPort();
# ----------------------------------------------------------------------------

# Get the server port
sub getPort
{
    my ($self) = @_;
    return $self->{_port};
}

# ----------------------------------------------------------------------------
# $backupHost = getBackup();
# ----------------------------------------------------------------------------

# Get the backup host
sub getBackup
{
    my ($self) = @_;
    return $self->{_backup};
}

# ----------------------------------------------------------------------------
# $boolean = getAuthentication();
# ----------------------------------------------------------------------------

# Get authentication value
sub getAuthentication
{
    my ($self) = @_;
    return $self->{_authentication};
}

# ----------------------------------------------------------------------------
# $boolean = getEncryption();
# ----------------------------------------------------------------------------

# Get encryption value
sub getEncryption
{
    my ($self) = @_;
    return $self->{_encryption};
}

# ----------------------------------------------------------------------------
# $wireProtocol = getWireProtocol();
# ----------------------------------------------------------------------------

# Get wire protocol
sub getWireProtocol
{
    my ($self) = @_;
    return $self->{_wireProtocol};
}

# ----------------------------------------------------------------------------
# $chunk = setRequest($request);
# ----------------------------------------------------------------------------

# Set the request
sub setRequest
{
    my ($self, $request) = @_;
    $self->{_request} = $request if $request;
    return $self;
}

# ----------------------------------------------------------------------------
# $chunk = setResponse($response);
# ----------------------------------------------------------------------------

# Set the response
sub setResponse
{
    my ($self, $response) = @_;
    $self->{_response} = $response if $response;
    return $self;
}

# ----------------------------------------------------------------------------
# $chunk = setHost($serverHost);
# ----------------------------------------------------------------------------

# Set the server host
sub setHost
{
    my ($self, $host) = @_;
    $self->{_host} = $host if $host;
    return $self;
}

# ----------------------------------------------------------------------------
# $chunk = setPort($serverPort);
# ----------------------------------------------------------------------------

# Set the server port
sub setPort
{
    my ($self, $port) = @_;
    $self->{_port} = $port if $port;
    return $self;
}

# ----------------------------------------------------------------------------
# $chunk = setBackup($backupHost);
# ----------------------------------------------------------------------------

# Set the backup host
sub setBackup
{
    my ($self, $backup) = @_;
    $self->{_backup} = $backup if $backup;
    return $self;
}

# ----------------------------------------------------------------------------
# $chunk = setAuthentication($boolean);
# ----------------------------------------------------------------------------

# Set authentication value
sub setAuthentication
{
    my ($self, $authentication) = @_;
    $self->{_authentication} = $authentication;
    return $self;
}

# ----------------------------------------------------------------------------
# $chunk = setEncryption($boolean);
# ----------------------------------------------------------------------------

# Set encryption value
sub setEncryption
{
    my ($self, $encryption) = @_;
    $self->{_encryption} = $encryption;
    return $self;
}

# ----------------------------------------------------------------------------
# $chunk = setWireProtocol($wireProtocol);
# ----------------------------------------------------------------------------

# Set wire protocol
sub setWireProtocol
{
    my ($self, $wireProtocol) = @_;
    $self->{_wireProtocol} = $wireProtocol;
    return $self;
}

# ----------------------------------------------------------------------------
# $string = toString();
# ----------------------------------------------------------------------------

# Serialize chunk to printable string
sub toString
{
    my ($self) = @_;

    my $string = '[';

    $string .= $self->{_request} if $self->{_request};
    $string .= ', ';

    $string .= $self->{_response} if $self->{_response};
    $string .= ', ';

    $string .= $self->{_host} if $self->{_host};
    $string .= ', ';

    $string .= $self->{_port} if $self->{_port};
    $string .= ', ';

    $string .= $self->{_backup} if $self->{_backup};
    $string .= ', ';

    $string .= $self->{_authentication};
    $string .= ', ';

    $string .= $self->{_encryption};
    $string .= ', ';

    $string .= $self->{_wireProtocol};

    $string .= ']';

    return $string;
}

# ----------------------------------------------------------------------------
# sign($doc);
# ----------------------------------------------------------------------------

# Sign the chunk
sub sign
{
    my ($self, $doc) = @_;
    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    # Step 1) Validate the token type and name

    my $tokenType    = $self->{_request}->getTokenType() || $TOKEN_SYMMETRIC;
    my $tokenName    = $self->{_request}->getTokenName();
    my $tokenValue   = $self->{_request}->getTokenValue();
    my $wireProtocol = $self->{_wireProtocol};
    my $site         = $self->{_request}->getSite();
    my $envPrefix    = '';
    if ($wireProtocol eq 'SOAP')
    {
        $envPrefix = 'soap:';
    }
    my $promotionMethod = $clientConfig->{'security.promotion'}
      || $SECURITY_PROMOTION;

    # Password Token
    if ($tokenType eq $TOKEN_PASSWORD)
    {
        # Token name defaults to request actor
        if (! $tokenName)
        {
            $tokenName = $self->{_request}->getActor();
        }

        if (length($tokenValue) == 0)
        {
            throw MAM::Exception('422',
                'Security token cannot be zero length.');
        }
    }

    # Unknown Token Type
    elsif ($tokenType ne $TOKEN_SYMMETRIC)
    {
        throw MAM::Exception('422',
            "Unsupported security token type ($tokenType)");
    }

    if ($log->is_debug())
    {
        $log->debug("The security token type is ($tokenType).");
        $log->debug("The security token name is ($tokenName).")
          if $tokenName;
        # $log->debug("The security token value is ($tokenValue).")
        # if $tokenType eq $TOKEN_PASSWORD;
    }

    # Step 2) Canonicalize the body into a string

    my $root      = $doc->getDocumentElement();
    my @bodyNodes = $root->getChildrenByTagName("${envPrefix}Body");
    if (! @bodyNodes)
    {
        throw MAM::Exception('422', 'Message body cannot be empty');
    }
    local $XML::LibXML::setTagCompression = 1;
    my $body = $bodyNodes[0]->toString();
    if ($log->is_debug())
    {
        my $sanitizedBody = $body;
        if ($sanitizedBody =~ /<Set name="Password">/)
        {
            $sanitizedBody =~
              s|<Set name="Password">[^<]*</Set>|<Set name="Password">********</Set>|;
        }
        $log->debug("The canonicalized body text is ($sanitizedBody).");
    }

    # Step 3) Calculate the digest and MAC and derive the site config

    my ($digestString, $macString, $siteConfig);

    # Promotion method is mamauth
    if ($promotionMethod eq 'mamauth')
    {
        # Call mam-auth to compute the digest and signature based on token value
        my $cmd  = "$SBIN_DIR/mam-auth";
        my @args = ();
        push @args, '-k', "$tokenValue" if $tokenType eq $TOKEN_PASSWORD;
        if ($site)
        {
            if ($site =~ /^([-\/\w\.]+)$/)
            {
                $site = $1;    # Untaint $site
                push @args, '-s', "$site";
            }
            else
            {
                throw MAM::Exception('422',
                    "Illegal characters were found in site name ($site)\n");
            }
        }
        my ($childPid, $childIn, $childOut, $childErr);
        $childErr = gensym();    # We create a symbol for the errfh because
        # open3 does not do that for us

        eval {
            my $oldPath = $ENV{PATH};    # Save PATH
            $ENV{PATH} = '/bin:/usr/bin';    # Untaint PATH
            delete @ENV{'IFS', 'CDPATH', 'ENV', 'BASH_ENV'};
            $childPid = open3($childIn, $childOut, $childErr, $cmd, @args);
            $ENV{PATH} = $oldPath;           # Restore PATH
        };
        throw MAM::Exception('422', "Unable to open mam-auth $@\n")
          if $@;

        binmode $childIn, ':utf8';
        print $childIn $body;
        close $childIn;

        # Make output and error handles read all into scalars
        my $output = do { local $/; <$childOut> };
        my $error  = do { local $/; <$childErr> };
        close $childOut;
        close $childErr;

        waitpid($childPid, 0);
        throw MAM::Exception('422',
            "mam-auth exited with rc=" . ($? >> 8) . ", error=$error\n")
          if $?;

        # Stuff results into a dynamic site config
        my $configTiny = Config::Tiny->read_string($output);
        $siteConfig = $configTiny->{_};

        # Extract digest and signature from site config
        $digestString = $siteConfig->{'digest.value'};
        if ($log->is_debug())
        {
            $log->debug(
                "The base64-encoded message digest is ($digestString).");
        }
        $macString = $siteConfig->{'signature.value'};
        if ($log->is_debug())
        {
            $log->debug("The base64-encoded MAC is ($macString).");
        }
    }

    # Promotion method is suidperl
    elsif ($promotionMethod eq 'suidperl')
    {
        if ($site)
        {
            $siteConfig = $globalSiteConfig->{$site};
        }
        else
        {
            $siteConfig = $localSiteConfig;
        }

        # Obtain token value from site config if symmetric token
        if ($tokenType eq $TOKEN_SYMMETRIC)
        {
            # Token value defaults to authkey
            if (! $tokenValue)
            {
                $tokenValue = $siteConfig->{'token.value'};
            }
            if (! $tokenValue)
            {
                throw MAM::Exception('422',
                    'Unable to obtain symmetric token value.');
            }
            #$log->debug("The security token value is ($tokenValue).");
        }

        # Perform the digest
        my $digest;
        if   ($SHA_ALGORITHM eq '512') { $digest = sha512(encode_utf8($body)); }
        else                           { $digest = sha1(encode_utf8($body)); }
        chomp($digestString = encode_base64($digest, ''));
        if ($log->is_debug())
        {
            $log->debug(
                "The base64-encoded message digest is ($digestString).");
        }

        # Generate the MAC
        my $mac;
        if ($SHA_ALGORITHM eq '512')
        {
            $mac = hmac_sha512($digest, $tokenValue);
        }
        else { $mac = hmac_sha1($digest, $tokenValue); }
        chomp($macString = encode_base64($mac, ''));
        if ($log->is_debug())
        {
            $log->debug("The base64-encoded MAC is ($macString).");
        }
    }

    # Invalid promotion method
    else
    {
        throw MAM::Exception('422',
            "Invalid promotion method ($promotionMethod)");
    }

    # Step 4) Extract host, port & backup from site config unless overridden

    unless (defined $self->getHost())
    {
        $self->setHost($siteConfig->{'server.host'} || $SERVER_HOST);
    }
    unless (defined $self->getPort())
    {
        $self->setPort($siteConfig->{'server.port'} || $SERVER_PORT);
    }
    unless (defined $self->getBackup()
        || ! defined $siteConfig->{'backup.host'})
    {
        $self->setBackup($siteConfig->{'backup.host'});
    }

    # Step 5) Add the Signature to the message

    my $signature   = $doc->createElement('Signature');
    my $digestValue = $doc->createElement('DigestValue');
    $digestValue->appendText($digestString);
    $signature->appendChild($digestValue);
    my $signatureValue = $doc->createElement('SignatureValue');
    $signatureValue->appendText($macString);
    $signature->appendChild($signatureValue);
    my $securityToken = $doc->createElement('SecurityToken');
    $securityToken->setAttribute('type', $tokenType);
    $securityToken->setAttribute('name', $tokenName) if $tokenName;
    $signature->appendChild($securityToken);
    $root->appendChild($signature);

    if ($log->is_debug())
    {
        $log->debug('The signature is (' . $signature->toString() . ').');
    }
}

# ----------------------------------------------------------------------------
# orient();
#
# Orient is used to obtain the server host, port and backup for the connection
# when this information is not obtained via sign()
#
# ----------------------------------------------------------------------------

# Orient the chunk
sub orient
{
    my ($self) = @_;
    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $site            = $self->{_request}->getSite();
    my $promotionMethod = $clientConfig->{'security.promotion'}
      || $SECURITY_PROMOTION;
    my $siteConfig;

    # Promotion method is mam-auth
    if ($promotionMethod eq 'mamauth')
    {
        # Call mam-auth to obtain the server hostname, port and backup
        my $cmd  = "$SBIN_DIR/mam-auth";
        my @args = ();
        if ($site)
        {
            if ($site =~ /^([-\/\w\.]+)$/)
            {
                $site = $1;    # Untaint $site
                push @args, '-s', "$site";
            }
            else
            {
                throw MAM::Exception('410',
                    "Illegal characters were found in site name ($site)\n");
            }
        }
        my ($childPid, $childOut, $childErr);
        $childErr = gensym();    # We create a symbol for the errfh because
        # open3 does not do that for us

        eval {
            my $oldPath = $ENV{PATH};    # Save PATH
            $ENV{PATH} = '/bin:/usr/bin';    # Untaint PATH
            delete @ENV{'IFS', 'CDPATH', 'ENV', 'BASH_ENV'};
            $childPid = open3('</dev/null', $childOut, $childErr, $cmd, @args);
            $ENV{PATH} = $oldPath;           # Restore PATH
        };
        throw MAM::Exception('410', "Unable to open mam-auth $@\n")
          if $@;

        # Make output and error handles read all into scalars
        my $output = do { local $/; <$childOut> };
        my $error  = do { local $/; <$childErr> };
        close $childOut;
        close $childErr;

        waitpid($childPid, 0);
        throw MAM::Exception('410',
            "mam-auth exited with rc=" . ($? >> 8) . ", error=$error\n")
          if $?;

        # Stuff results into a dynamic site config
        my $configTiny = Config::Tiny->read_string($output);
        $siteConfig = $configTiny->{_};
    }

    # Promotion method is suidperl
    elsif ($promotionMethod eq 'suidperl')
    {
        if ($site)
        {
            $siteConfig = $globalSiteConfig->{$site};
        }
        else
        {
            $siteConfig = $localSiteConfig;
        }
    }

    # Invalid promotion method
    else
    {
        throw MAM::Exception('410',
            "Invalid promotion method ($promotionMethod)");
    }

    # Extract host, port and backup from site config
    unless (defined $self->getHost())
    {
        $self->setHost($siteConfig->{'server.host'} || $SERVER_HOST);
    }
    unless (defined $self->getPort())
    {
        $self->setPort($siteConfig->{'server.port'} || $SERVER_PORT);
    }
    unless (defined $self->getBackup()
        || ! defined $siteConfig->{'backup.host'})
    {
        $self->setBackup($siteConfig->{'backup.host'});
    }
}

# ----------------------------------------------------------------------------
# authenticate($doc);
# ----------------------------------------------------------------------------

# Authenticate the chunk
sub authenticate
{
    my ($self, $doc) = @_;
    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my ($tokenType, $tokenValue, $tokenName);
    my $wireProtocol = $self->{_wireProtocol};
    my $envPrefix    = '';
    my $appPrefix    = '';
    if ($wireProtocol eq 'SOAP')
    {
        $envPrefix = 'soap:';
        $appPrefix = 'gold:';
    }
    my $promotionMethod = $clientConfig->{'security.promotion'}
      || $SECURITY_PROMOTION;

    # Step 1) Extract the signature and digest from envelope

    my $root           = $doc->getDocumentElement();
    my @signatureNodes = $root->getChildrenByTagName('Signature');
    if (! @signatureNodes)
    {
        throw MAM::Exception('424', 'Message not signed -- No signature found');
    }
    my $signature = $signatureNodes[0];
    $self->{_authentication} = 1;
    my @digestValueNodes = $signature->getChildrenByTagName('DigestValue');
    if (! @digestValueNodes)
    {
        throw MAM::Exception('424', 'No digest value found');
    }
    my $digestValue = ($digestValueNodes[0])->textContent();
    my @signatureValueNodes =
      $signature->getChildrenByTagName('SignatureValue');
    if (! @signatureValueNodes)
    {
        throw MAM::Exception('424', 'No signature value found');
    }
    my $signatureValue = ($signatureValueNodes[0])->textContent();

    # Step 2) Validate the token type and name

    my @securityTokenNodes = $signature->getChildrenByTagName('SecurityToken');
    if (@securityTokenNodes)
    {
        $tokenType = ($securityTokenNodes[0])->getAttribute('type');
    }
    else
    {
        $tokenType = $TOKEN_SYMMETRIC;
    }
    if ($tokenType eq $TOKEN_PASSWORD)
    {
        $tokenName = ($securityTokenNodes[0])->getAttribute('name');

        if (! defined $tokenName)
        {
            throw MAM::Exception('424',
                'Token name must be specified in Password authentication');
        }

        # Extract actor from request
        my @bodyNodes    = $root->getChildrenByTagName("${envPrefix}Body");
        my $body         = $bodyNodes[0];
        my @requestNodes = $body->getChildrenByTagName("${appPrefix}Request");
        my $request      = $requestNodes[0];
        my $actor        = $request->getAttribute('actor');
        if ($tokenName ne $actor)
        {
            throw MAM::Exception('424',
                'Token name must match request actor in Password authentication'
            );
        }

        # Extract encrypted password from database
        my $encryptedPassword =
          MAM::Cache->getPasswordProperty($tokenName, 'Password');
        if (! defined $encryptedPassword)
        {
            throw MAM::Exception('424', "No password defined for $tokenName");
        }

        # Decrypt password with symmetric key

        ## Promotion method is mamauth
        #if ($promotionMethod eq 'mamauth')
        #{
        #}
        #
        ## Promotion method is suidperl
        #elsif ($promotionMethod eq 'suidperl')

        # Promotion method is suidperl or mamauth
        if ($promotionMethod eq 'suidperl' || $promotionMethod eq 'mamauth')
        {
            my $symmetricKey  = $localSiteConfig->{'token.value'};
            my $key           = pack('a24', $symmetricKey);
            my $cipherPayload = decode_base64($encryptedPassword);
            my $cipher        = new Crypt::CBC(
                {   key    => $key,
                    cipher => 'Crypt::DES_EDE3',
                    header => 'randomiv',         # No longer default as of 2.17
                    regenerate_key => 0,
                }
            );
            $tokenValue = $cipher->decrypt('RandomIV' . $cipherPayload);
            if (length($tokenValue) == 0)
            {
                throw MAM::Exception('424',
                    'Security token cannot be zero length.');
            }
        }
    }
    elsif ($tokenType ne $TOKEN_SYMMETRIC)
    {
        throw MAM::Exception('424',
            "Unsupported security token type ($tokenType)");
    }
    if ($log->is_debug())
    {
        $log->debug("The security token type is ($tokenType).");
        $log->debug("The security token name is ($tokenName).")
          if defined $tokenName;
        # $log->debug("The security token value is ($tokenValue).")
        # if $tokenType eq $TOKEN_PASSWORD;

    }

    # Step 3) Canonicalize the body into a string

    my @bodyNodes = $root->getChildrenByTagName("${envPrefix}Body");
    if (! @bodyNodes)
    {
        throw MAM::Exception('424', 'Message body cannot be empty');
    }
    local $XML::LibXML::setTagCompression = 1;
    my $body = $bodyNodes[0]->toString();
    if ($log->is_debug())
    {
        my $sanitizedBody = $body;
        if ($sanitizedBody =~ /<Set name="Password">/)
        {
            $sanitizedBody =~
              s|<Set name="Password">[^<]*</Set>|<Set name="Password">********</Set>|;
        }
        $log->debug("The canonicalized body text is ($sanitizedBody).");
    }

    # Step 4) Calculate the digest and MAC using the token value

    my ($digestString, $macString);

    # Promotion method is suidperl or mamauth
    if ($promotionMethod eq 'suidperl' || $promotionMethod eq 'mamauth')
    {
        # Obtain token value from site config if symmetric token
        if ($tokenType eq $TOKEN_SYMMETRIC)
        {
            # Token value defaults to authkey
            if (! $tokenValue)
            {
                $tokenValue = $localSiteConfig->{'token.value'};
            }
            if (! $tokenValue)
            {
                throw MAM::Exception('424',
                    'Unable to obtain symmetric token value.');
            }
            #$log->debug("The security token value is ($tokenValue).");
        }

        # Perform the digest
        my $digest;
        if   ($SHA_ALGORITHM eq '512') { $digest = sha512(encode_utf8($body)); }
        else                           { $digest = sha1(encode_utf8($body)); }
        chomp($digestString = encode_base64($digest, ''));
        if ($log->is_debug())
        {
            $log->debug(
                "The base64-encoded message digest is ($digestString).");
        }

        # Generate the MAC
        my $mac;
        if ($SHA_ALGORITHM eq '512')
        {
            $mac = hmac_sha512($digest, $tokenValue);
        }
        else { $mac = hmac_sha1($digest, $tokenValue); }
        chomp($macString = encode_base64($mac, ''));
        if ($log->is_debug())
        {
            $log->debug("The base64-encoded MAC is ($macString).");
        }
    }

    # Invalid promotion method
    else
    {
        throw MAM::Exception('424',
            "Invalid promotion method ($promotionMethod)");
    }

    # Step 4) Check the digest and the MAC

    if ($digestString ne $digestValue)
    {
        throw MAM::Exception('424',
            'Incoming digest does not match calculated digest');
    }
    if ($macString ne $signatureValue)
    {
        throw MAM::Exception('424',
            'Incoming MAC does not match calculated MAC');
    }
}

# ----------------------------------------------------------------------------
# encrypt($doc);
# ----------------------------------------------------------------------------

# Encrypt the chunk
sub encrypt
{
    my ($self, $doc) = @_;
    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    # Step 1) Validate the token type and name

    my $tokenType    = $self->{_request}->getTokenType() || $TOKEN_SYMMETRIC;
    my $tokenName    = $self->{_request}->getTokenName();
    my $tokenValue   = $self->{_request}->getTokenValue();
    my $wireProtocol = $self->{_wireProtocol};
    my $site         = $self->{_request}->getSite();
    my $envPrefix    = '';
    if ($wireProtocol eq 'SOAP')
    {
        $envPrefix = 'soap:';
    }
    my $promotionMethod = $clientConfig->{'security.promotion'}
      || $SECURITY_PROMOTION;

    # Password Token
    if ($tokenType eq $TOKEN_PASSWORD)
    {
        # Token name defaults to request actor
        if (! $tokenName)
        {
            $tokenName = $self->{_request}->getActor();
        }

        if (length($tokenValue) == 0)
        {
            throw MAM::Exception('432',
                'Security token cannot be zero length.');
        }
    }

    # Unknown Token Type
    elsif ($tokenType ne $TOKEN_SYMMETRIC)
    {
        throw MAM::Exception('432',
            "Unsupported security token type ($tokenType)");
    }

    if ($log->is_debug())
    {
        $log->debug("The security token type is ($tokenType).");
        $log->debug("The security token name is ($tokenName).")
          if $tokenName;
        # $log->debug("The security token value is ($tokenValue).")
        # if $tokenType eq $TOKEN_PASSWORD;
    }

    if ($promotionMethod eq 'mamauth')
    {
        throw MAM::Exception('432',
            "Encryption is not supported for the mamauth promotion method");
    }

    # Promotion method is suidperl
    elsif ($promotionMethod eq 'suidperl')
    {
        my $siteConfig;
        if ($site)
        {
            $siteConfig = $globalSiteConfig->{$site};
        }
        else
        {
            $siteConfig = $localSiteConfig;
        }

        # Obtain token value from site config if symmetric token
        if ($tokenType eq $TOKEN_SYMMETRIC)
        {
            # Token value defaults to authkey
            if (! $tokenValue)
            {
                $tokenValue = $siteConfig->{'token.value'};
            }
            if (! $tokenValue)
            {
                throw MAM::Exception('432',
                    'Unable to obtain symmetric token value.');
            }
            #$log->debug("The security token value is ($tokenValue).");
        }

        # Step 2) Generate 192-bit random session key (with 64-bit IV)

        # This is attempted implementation of ANSI X9.17 PRNG
        my $seed = pack('a24', md5_hex($$ . time() . int(rand(10000))));
        my $key  = pack('a24', $tokenValue);
        my $time = pack('a24', time());
        my $iv = Crypt::CBC->random_bytes(8);
       # We would rather use literal_key => 1, but this is not supported in 2.12
        my $cipher = new Crypt::CBC(
            {   cipher => 'Crypt::DES_EDE3',  # Triple DES (CBC)
                header => 'none',             # We don't want anything prepended
                iv  => $iv,     # Use own iv since we don't want it prepended
                key => $key,    # Must be at least 24 bytes
                regenerate_key => 0,         # Do not uses MD5 hash on key
                padding        => 'null',    # Null padding
                prepend_iv     => 0,         # Do not prepend iv
            }
        );
        my $encryptedTime = $cipher->encrypt($time);
        my $sessionKey    = $cipher->encrypt($encryptedTime ^ $seed);

        # Step 3) Encrypt the signature and body with the session key

        # Build the plaintext
        my $plainText = '';
        local $XML::LibXML::setTagCompression = 1;
        my $root           = $doc->getDocumentElement();
        my @signatureNodes = $root->getChildrenByTagName('Signature');
        if (@signatureNodes)
        {
            my $signature = $signatureNodes[0]->toString();
            $plainText .= $signature;
        }
        my @bodyNodes = $root->getChildrenByTagName("${envPrefix}Body");
        if (! @bodyNodes)
        {
            throw MAM::Exception('436', 'No message body found');
        }
        my $body = $bodyNodes[0]->toString();
        $plainText .= $body;
        if ($log->is_debug())
        {
            $log->debug("The plaintext is ($plainText).");
        }

        # Compress the plaintext
        my $compressedData = Compress::Zlib::memGzip($plainText);

        # Initialize the cipher
        $cipher = new Crypt::CBC(
            {   key    => $sessionKey,          # Must be at least 24 bytes
                cipher => 'Crypt::DES_EDE3',    # Triple DES (CBC)
                header => 'randomiv',           # No longer default as of 2.17
                regenerate_key => 0,            # Regenerate uses MD5 hash
                padding        => 'standard',   # PKCS#5
                prepend_iv     => 1,            # Prepends RandomIV.{8}
            }
        );

        # Encrypt and encode the CipherValue payload
        my $cipherPayload =
          substr($cipher->encrypt($compressedData), 8);    # Remove 'RandomIV'
        my $cipherText = encode_base64($cipherPayload, '');
        if ($log->is_debug())
        {
            $log->debug("The base64-encoded ciphertext is ($cipherText).");
        }

        # Step 4) Encrypt the session key with the security token

        # Compute the CMS key checksum and call this CKS
        my $md = Digest::SHA->new($SHA_ALGORITHM);
        $md->add($sessionKey);
        my $cks = $md->digest;

        # Let WKCKS = WK || CKS, where || is concatenation
        my $wkcks = $sessionKey . substr($cks, 0, 8);

        # Encrypt WKCKS in CBC mode using KEK as the key and IV as iv -> TEMP1
        # Let TEMP2 = IV || TEMP1
        $cipher = new Crypt::CBC(
            {   key    => $key,                 # Must be at least 24 bytes
                cipher => 'Crypt::DES_EDE3',    # Triple DES (CBC)
                header => 'randomiv',           # No longer default as of 2.17
                regenerate_key => 0,            # Regenerate uses MD5 hash
                padding        => 'null',       # No padding
                prepend_iv     => 1,            # Prepends RandomIV.{8}
            }
        );
        my $temp2 = substr($cipher->encrypt($wkcks), 8);    # Remove 'RandomIV'

        # Reverse the order of the octets in TEMP2 and call the result TEMP3
        my $temp3 = reverse $temp2;

       # Encrypt TEMP3 in CBC mode using the KEK and an IV of 0x4adda22c79e82105
        my $funkyIv =
          pack('C8', 0x4a, 0xdd, 0xa2, 0x2c, 0x79, 0xe8, 0x21, 0x05);
        $cipher = new Crypt::CBC(
            {   key    => $key,                 # Must be at least 24 bytes
                cipher => 'Crypt::DES_EDE3',    # Triple DES (CBC)
                header => 'none',               # No need to prepend an IV here
                iv     => $funkyIv,             # IV of 0x4adda22c79e82105
                regenerate_key => 0,            # Regenerate uses MD5 hash
                padding        => 'null',       # No padding
                prepend_iv     => 0,            # Does not prepend IV
            }
        );
        my $ek = $cipher->encrypt($temp3);
        my $wrappedText = encode_base64($ek, '');
        if ($log->is_debug())
        {
            $log->debug("The base64-encoded encrypted key is ($wrappedText).");
        }

        # Step 5) Replace envelope contents with encrypted data

        foreach my $node (@signatureNodes, @bodyNodes)
        {
            $root->removeChild($node);
        }

        my $encryptedData = $doc->createElement('EncryptedData');
        my $encryptedKey  = $doc->createElement('EncryptedKey');
        $encryptedKey->appendText($wrappedText);
        $encryptedData->appendChild($encryptedKey);
        my $cipherValue = $doc->createElement('CipherValue');
        $cipherValue->appendText($cipherText);
        $encryptedData->appendChild($cipherValue);
        my $securityToken = $doc->createElement('SecurityToken');
        $securityToken->setAttribute('type', $tokenType);
        $securityToken->setAttribute('name', $tokenName) if $tokenName;
        $encryptedData->appendChild($securityToken);
        $root->appendChild($encryptedData);

        if ($log->is_debug())
        {
            $log->debug(
                'The encrypted data is (' . $encryptedData->toString() . ').');
        }
    }

    # Invalid promotion method
    else
    {
        throw MAM::Exception('422',
            "Invalid promotion method ($promotionMethod)");
    }
}

# ----------------------------------------------------------------------------
# decrypt($doc);
# ----------------------------------------------------------------------------

# Decrypt the chunk
sub decrypt
{
    my ($self, $doc) = @_;
    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my ($tokenType, $tokenValue, $tokenName);
    my $wireProtocol = $self->{_wireProtocol};
    my $envPrefix    = '';
    my $appPrefix    = '';
    if ($wireProtocol eq 'SOAP')
    {
        $envPrefix = 'soap:';
        $appPrefix = 'gold:';
    }
    my $promotionMethod = $clientConfig->{'security.promotion'}
      || $SECURITY_PROMOTION;

    # Step 1)  Extract encrypted data from envelope

    my $root               = $doc->getDocumentElement();
    my @encryptedDataNodes = $root->getChildrenByTagName('EncryptedData');
    if (! @encryptedDataNodes)
    {
        throw MAM::Exception('434',
            'Message not encrypted -- No encrypted data found');
    }
    my $encryptedData = $encryptedDataNodes[0];
    $self->{_encryption} = 1;

    my @encryptedKeyNodes =
      $encryptedData->getChildrenByTagName('EncryptedKey');
    if (! @encryptedKeyNodes)
    {
        throw MAM::Exception('434', 'No encrypted key found');
    }
    my $wrappedText = ($encryptedKeyNodes[0])->textContent();

    my @cipherValueNodes = $encryptedData->getChildrenByTagName('CipherValue');
    if (! @cipherValueNodes)
    {
        throw MAM::Exception('434', 'No cipher value found');
    }
    my $cipherText = ($cipherValueNodes[0])->textContent();
    if ($log->is_debug())
    {
        $log->debug("The base64-encoded encrypted data is ($cipherText).");
    }

    # Obtain the security token

    my @securityTokenNodes =
      $encryptedData->getChildrenByTagName('SecurityToken');
    if (@securityTokenNodes)
    {
        $tokenType = ($securityTokenNodes[0])->getAttribute('type');
    }
    else
    {
        $tokenType = $TOKEN_SYMMETRIC;
    }

    # Symmetric Token
    if ($tokenType eq $TOKEN_SYMMETRIC)
    {
        $tokenValue = $localSiteConfig->{'token.value'};
    }

    # Password Token
    elsif ($tokenType eq $TOKEN_PASSWORD)
    {
        $tokenName = ($securityTokenNodes[0])->getAttribute('name');
        if (! defined $tokenName)
        {
            throw MAM::Exception('434',
                'Token name must be specified in Password authentication');
        }

        # Extract actor from request
        my @bodyNodes    = $root->getChildrenByTagName("${envPrefix}Body");
        my $body         = $bodyNodes[0];
        my @requestNodes = $body->getChildrenByTagName("${appPrefix}Request");
        my $request      = $requestNodes[0];
        my $actor        = $request->getAttribute('actor');
        if ($tokenName ne $actor)
        {
            throw MAM::Exception('434',
                'Token name must match request actor in Password authentication'
            );
        }

        # Extract encrypted password from database
        my $encryptedPassword =
          MAM::Cache->getPasswordProperty($tokenName, 'Password');
        if (! defined $encryptedPassword)
        {
            throw MAM::Exception('434', "No password defined for $tokenName");
        }

        # Decrypt password with symmetric key

        ## Promotion method is mamauth
       #if ($promotionMethod eq 'mamauth')
       #{
       #    throw MAM::Exception('434',
       #        "Decryption is not supported for the mamauth promotion method");
       #}
       #
        ## Promotion method is suidperl
        #elsif ($promotionMethod eq 'suidperl')

        # Promotion method is suidperl or mamauth
        if ($promotionMethod eq 'suidperl' || $promotionMethod eq 'mamauth')
        {
            my $symmetricKey  = $localSiteConfig->{'token.value'};
            my $key           = pack('a24', $symmetricKey);
            my $cipherPayload = decode_base64($encryptedPassword);
            my $cipher        = new Crypt::CBC(
                {   key    => $key,
                    header => 'randomiv',         # No longer default as of 2.17
                    cipher => 'Crypt::DES_EDE3',
                    regenerate_key => 0,
                }
            );
            $tokenValue = $cipher->decrypt('RandomIV' . $cipherPayload);
        }
    }

    # Unknown Token Type
    else
    {
        throw MAM::Exception('434',
            "Unsupported security token type ($tokenType)");
    }

    if (! $tokenValue)
    {
        throw MAM::Exception('434', 'Unable to obtain valid security token.');
    }
    if ($log->is_debug())
    {
        $log->debug("The security token type is ($tokenType).");
        #$log->debug("The security token value is ($tokenValue).");
    }

    my $plainText;

    ## Promotion method is mamauth
    #if ($promotionMethod eq 'mamauth')
    #{
    #    throw MAM::Exception('434',
    #        "Decryption is not supported for the mamauth promotion method");
    #}
    #
    ## Promotion method is suidperl
    #elsif ($promotionMethod eq 'suidperl')

    # Promotion method is suidperl or mamauth
    if ($promotionMethod eq 'suidperl' || $promotionMethod eq 'mamauth')
    {

        # Step 3) Decrypt the session key

        my $key = pack('a24', $tokenValue);
        my $ek = decode_base64($wrappedText);
        if (length($ek) != 40)
        {
            throw MAM::Exception('434',
                'Encrypted key is not 40 octets (' . length($ek) . ').');
        }

     # Decrypt text with 3DES in CBC mode using KEK and IV as 0x4adda22c79e82105
        my $funkyIv =
          pack('C8', 0x4a, 0xdd, 0xa2, 0x2c, 0x79, 0xe8, 0x21, 0x05);
        my $cipher = new Crypt::CBC(
            {   key            => $key,              # Must be at least 24 bytes
                cipher         => 'Crypt::DES_EDE3', # Triple DES (CBC)
                header         => 'none',            # No prepended IV here
                iv             => $funkyIv,          # IV of 0x4adda22c79e82105
                regenerate_key => 0,                 # Regenerate uses MD5 hash
                padding        => 'null',            # No padding
                prepend_iv     => 0,                 # Does not prepend IV
            }
        );
        my $temp3 = $cipher->decrypt($ek);

        # Reverse the order of the octets in TEMP3 and call the result TEMP2
        my $temp2 = reverse $temp3;

        # Decompose TEMP2 into IV, the first 8 octets, and TEMP1
        # Decrypt TEMP1 using 3DES in CBC mode using KEK and the IV -> WKCKS
        $cipher = new Crypt::CBC(
            {   key    => $key,                 # Must be at least 24 bytes
                cipher => 'Crypt::DES_EDE3',    # Triple DES (CBC)
                header => 'randomiv',           # No longer default as of 2.17
                regenerate_key => 0,            # Regenerate uses MD5 hash
                padding        => 'null',       # No padding
                prepend_iv     => 1,            # Prepends RandomIV.{8}
            }
        );
        my $wkcks = $cipher->decrypt('RandomIV' . $temp2);

       # Decompose WKCKS. CKS is the last 8 octets and WK are those preceding it
        my $wk = substr($wkcks, 0, 24);
        my $cks = substr($wkcks, 24);

       # Calculate a CMS key checksum over the WK and compare with extracted CKS
        my $md = Digest::SHA->new($SHA_ALGORITHM);
        $md->add($wk);
        my $ccks = $md->digest;
        if (substr($ccks, 0, 8) ne $cks)
        {
            throw MAM::Exception('434',
                'Key wrap checksum integrity check failed');
        }

        # WK is the wrapped key, extracted for use in data decryption
        if ($log->is_debug())
        {
            $log->debug('The base64-encoded session key is ('
                  . encode_base64($wk, '')
                  . ').');
        }

        # Step 4)  Decrypt the data with the session key

        # Recover the IV and the compressed data
        my $cipherPayload = decode_base64($cipherText);
        $cipher = new Crypt::CBC(
            {   key    => $wk,                  # Must be at least 24 bytes
                cipher => 'Crypt::DES_EDE3',    # Triple DES (CBC)
                header => 'randomiv',           # No longer default as of 2.17
                regenerate_key => 0,            # Regenerate uses MD5 hash
                padding        => 'standard',   # PKCS#5
                prepend_iv     => 1,            # Prepends RandomIV.{8}
            }
        );
        my $compressedData = $cipher->decrypt('RandomIV' . $cipherPayload);
        $plainText = Compress::Zlib::memGunzip($compressedData);
    }

    # Invalid promotion method
    else
    {
        throw MAM::Exception('424',
            "Invalid promotion method ($promotionMethod)");
    }

    # Step 5) Replace encrypted data with envelope contents

    foreach my $node (@encryptedDataNodes)
    {
        $root->removeChild($node);
    }
    local $XML::LibXML::setTagCompression = 1;
    my $emptyEnvelope = $doc->toString();
    my $stuffedEnvelope =
      substr($emptyEnvelope, 0,
        index($emptyEnvelope, "</${envPrefix}Envelope>"))
      . $plainText
      . '</Envelope>';
    my $parser        = new XML::LibXML();
    my $decipheredDoc = $parser->parse_string($stuffedEnvelope);
    foreach my $child ($decipheredDoc->getDocumentElement()->childNodes())
    {
        $doc->adoptNode($child);
        $root->appendChild($child);
    }
}

# ----------------------------------------------------------------------------
# $replyChunk = $messageChunk->getChunk();
# ----------------------------------------------------------------------------

# Sends the message chunk and receives the reply chunk
sub getChunk
{
    my ($self) = @_;
    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }
    my ($message, $reply, $chunk);
    my $backup = 0;
    my $caught = 0;

    # Send the message
    $message = new MAM::Message();
    try
    {
        $message->sendChunk($self);
    }
    catch MAM::Exception with
    {
        $log->error('Failed sending message to primary: (' . $_[0] . ').');
        $chunk = new MAM::Chunk()->setResponse(
            new MAM::Response()->failure(
                $_[0]->{-value},
                'Failed sending message: (' . $_[0]->{-text} . ').'
            )
        );
        $caught = $_[0]->{-value};
    };
    my $backupServer = $self->getBackup() || 'NONE';

    # Try sending to backup server if the primary connection fails
    if ($caught == 222 && $backupServer ne 'NONE')
    {
        $backup = 1;
        $caught = 0;
        try
        {
            $message->sendChunk($self, 'backup');
        }
        catch MAM::Exception with
        {
            $log->error('Failed sending message to backup: (' . $_[0] . ').');
            $chunk = new MAM::Chunk()->setResponse(
                new MAM::Response()->failure(
                    $_[0]->{-value},
                    'Failed sending message: (' . $_[0]->{-text} . ').'
                )
            );
            $caught = $_[0]->{-value};
        };
    }
    return $chunk if $caught;

    # Receive the reply
    try
    {
        $reply = $message->getReply();
        $chunk = $reply->receiveChunk();
    }
    catch MAM::Exception with
    {
        $log->error('Failed receiving reply (' . $_[0] . ').');
        $chunk = new MAM::Chunk()->setResponse(
            new MAM::Response()->failure(
                $_[0]->{-value},
                'Failed receiving reply: (' . $_[0]->{-text} . ').'
            )
        );
    };
    return $chunk if $caught;

   # Try sending to the backup server if the primary database connection is down
    if (! $backup && $backupServer ne 'NONE')
    {
        my $code = $chunk->getResponse()->getCode;
        if ($code == 730)    # Failed obtaining database connection
        {
            try
            {
                $message->sendChunk($self, 'backup');
            }
            catch MAM::Exception with
            {
                $log->error('Failed sending message: (' . $_[0] . ').');
                $chunk = new MAM::Chunk()->setResponse(
                    new MAM::Response()->failure(
                        $_[0]->{-value},
                        'Failed sending message: (' . $_[0]->{-text} . ').'
                    )
                );
                $caught = $_[0]->{-value};
            };
            return $chunk if $caught;

            # Receive the backup reply
            try
            {
                $reply = $message->getReply();
                $chunk = $reply->receiveChunk();
            }
            catch MAM::Exception with
            {
                $log->error('Failed receiving reply (' . $_[0] . ').');
                $chunk = new MAM::Chunk()->setResponse(
                    new MAM::Response()->failure(
                        $_[0]->{-value},
                        'Failed receiving reply: (' . $_[0]->{-text} . ').'
                    )
                );
            };
        }
    }

    return $chunk;
}

1;
