#! /usr/bin/perl -wT
################################################################################
#
#  MAM client helper routines
#
# File   :  Client.pm
#
################################################################################
#
#    This file includes parts which are copyrighted by Adaptive Computing
#    Enterprises, Inc., and parts which are copyrighted by Battelle Memorial
#    Institute.
#
#    The terms and conditions for the use and redistribution of these parts
#    are governed by the Moab Accounting Manager License and the BSD License
#    respectively. Refer to the LICENSE file for details.
#
#    Copyright (C) 2006 - 2016 by Adaptive Computing Enterprises, Inc.
#    All Rights Reserved.
#
#    The Moab Accounting Manager License specifies the terms and conditions
#    for use and redistribution.
#
#    The Moab Accounting Manager License applies to the Moab Accounting
#    Manager software offered by Adaptive Computing Enterprises, Inc.
#    By installing or using this software, Licensee accepts a non-exclusive
#    license from Adaptive Computing Enterprises, Inc. and is bound to accept
#    acknowledgement of and abide by the notices and conditions of the
#    Moab Accounting Manager License.
#
#    Copyright (C) 2003 - 2005 Pacific Northwest National Laboratory,
#    Battelle Memorial Institute. All rights reserved.
#
#    The BSD License specifies the terms and conditions for use and
#    redistribution.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions
#    are met:
#
#    - Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
#    - Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
#    - Neither the name of Battelle nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
#    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
#    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
#    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#    POSSIBILITY OF SUCH DAMAGE.
#
################################################################################

use strict;

=head1 NAME

MAM::Client - helper routines for MAM clients

=head1 DESCRIPTION

The B<MAM::Client> module defines routines for MAM clients

=head1 METHODS

=over 4

=item parseSupplement( $name, $value );

=item $buildSupplements( $request );

=item $displayResponse( $response );

=back 

=head1 EXAMPLES

use MAM::Client;

MAM::CLient::parseSupplement();

MAM::Client::buildSupplements( $request );

MAM::Client::displayResponse( $response );

=head1 REQUIRES

Perl 5.8

=head1 AUTHOR

Scott Jackson, scottmo@adaptivecomputing.com

=cut

##############################################################################

package MAM::Client;

use vars qw(@ARGV $format $log $quiet %supplement $time_division $verbose);
use MAM::Global;
use Log::Log4perl qw(get_logger);
use XML::LibXML;

# ----------------------------------------------------------------------------
# &parseSupplement( $name, $value )
# ----------------------------------------------------------------------------

sub parseSupplement
{
    my ($name, $value) = @_;
    my (%attributes);
    # Populate %attributes from option hash values
    while (@ARGV && $ARGV[0] =~ /^(\w+)=(([^\"]+)|\"([^\"]*)\")$/)
    {
        shift(@ARGV);
        $attributes{$1} = defined($3) ? $3 : $4;
    }
    push(@{$supplement{$name}}, \%attributes) if %attributes;
}

# ----------------------------------------------------------------------------
# $request = &buildSupplements( $request )
# ----------------------------------------------------------------------------

sub buildSupplements
{
    my ($request) = @_;

    # Process the --get options
    foreach my $hash (@{$supplement{get}})
    {
        if (keys %{$hash})
        {
            my $selection = new MAM::Selection(%{$hash});
            if ($log->is_debug())
            {
                $log->debug('Adding selection to request ('
                      . $selection->toString()
                      . ').');
            }
            $request->addSelection($selection);
        }
    }

    # Process the --set options
    foreach my $hash (@{$supplement{set}})
    {
        if (keys %{$hash})
        {
            my $assignment = new MAM::Assignment(%{$hash});
            if ($log->is_debug())
            {
                $log->debug('Adding assignment to request ('
                      . $assignment->toString()
                      . ').');
            }
            $request->addAssignment($assignment);
        }
    }

    # Process the --where options
    foreach my $hash (@{$supplement{where}})
    {
        if (keys %{$hash})
        {
            my $condition = new MAM::Condition(%{$hash});
            if ($log->is_debug())
            {
                $log->debug('Adding condition to request ('
                      . $condition->toString()
                      . ').');
            }
            $request->addCondition($condition);
        }
    }

    # Process the --option options
    foreach my $hash (@{$supplement{option}})
    {
        if (keys %{$hash})
        {
            my $option = new MAM::Option(%{$hash});
            if ($option->getName() eq 'ChunkSize')
            {
                my $value = $option->getValue();
                if ($log->is_trace())
                {
                    $log->trace(
                        "Setting chunking on request with chunk size ($value)."
                    );
                }
                $request->setChunking(1);
                $request->setChunkSize($value);
            }
            else
            {
                if ($log->is_debug())
                {
                    $log->debug('Adding option to request ('
                          . $option->toString()
                          . ').');
                }
                $request->addOption($option);
            }
        }
    }

    # Process the --usage-record data
    foreach my $hash (@{$supplement{'usage-record'}})
    {
        my $usageRecord = new MAM::Datum('UsageRecord');
        foreach my $property (keys %{$hash})
        {
            $usageRecord->setProperty($property, ${$hash}{$property});
        }
        if ($log->is_debug())
        {
            $log->debug('Adding usage record data to request ('
                  . $usageRecord->toString()
                  . ').');
        }
        $request->addDatum($usageRecord);
    }

    return $request;
}

# ----------------------------------------------------------------------------
# &displayResponse( $response[, $show])
# ----------------------------------------------------------------------------

sub displayResponse
{
    my ($response, $show) = @_;
    if ($log->is_debug())
    {
        $log->debug('Displaying response');
    }

    # Aggregate response data per specified operators if necessary
    if (defined $show && $show =~ /\(/)
    {
        $response = aggregateResponse($response, $show);
    }

    my $status      = $response->getStatus();
    my $message     = $response->getMessage();
    my $chunkNum    = $response->getChunkNum();
    my $doc         = XML::LibXML::Document->new();
    my $dataElement = $response->getDataElement();
    $doc->setDocumentElement($dataElement);
    my %metadata = $response->getMetadata();

    # Print data for query or if verbose
    if ($verbose)
    {
        if ($dataElement)
        {
            my @data = $dataElement->childNodes();
            if (@data)
            {
                my $currency_precision;
                if ($time_division == 3600)
                {
                    $currency_precision = 2;
                }
                else
                {
                    $currency_precision = $clientConfig->{'currency.precision'}
                      || 0;
                    if ($currency_precision =~ /^(\d+)$/)
                    {
                        $currency_precision = $1;
                    }
                    else
                    {
                        die
                          "Illegal characters were found in \$currency_precision ($currency_precision)\n";
                    }
                }

                # Raw or CSV format
                if ($format eq 'csv' || $format eq 'raw')
                {
                    my $firstField;
                    # Print header if not quiet
                    if (! $quiet && $chunkNum == 1)
                    {
                        my @fields = $data[0]->childNodes();
                        $firstField = 1;
                        foreach my $field (@fields)
                        {
                            if ($firstField) { $firstField = 0; }
                            elsif ($format eq 'csv') { print ','; }
                            else                     { print '|'; }
                            print $field->nodeName();
                        }
                        print "\n";
                    }
                    # Print data
                    foreach my $data (@data)
                    {
                        my $object = $data->nodeName();
                        my @fields = $data->childNodes();
                        $firstField = 1;
                        foreach my $field (@fields)
                        {
                            if ($firstField) { $firstField = 0; }
                            elsif ($format eq 'csv') { print ','; }
                            else                     { print '|'; }
                            my $name     = $field->nodeName();
                            my $value    = $field->textContent();
                            my $dataType = $metadata{$name} // 'String';
                            if ($dataType eq 'Currency' && $value ne '')
                            {

                                if ($value =~ /(-?)inf/i)
                                {
                                    $value = $1 . 'Infinity';
                                }
                                else
                                {
                                    $value = sprintf("%.${currency_precision}f",
                                        $value / $time_division);
                                }
                            }
                            if ($format eq 'csv' && $value =~ /[,\"]/)
                            {
                                $value =~ s/"/""/g;
                                print '"' . $value . '"';
                            }
                            else
                            {
                                print $value;
                            }
                        }
                        print "\n";
                    }
                }
                # Cooked format
                else
                {
                    my %size = ();
                    # Determine column widths
                    my @fields = $data[0]->childNodes();
                    foreach my $field (@fields)
                    {
                        my $name = $field->nodeName();
                        $size{$name} = length($name);
                    }
                    foreach my $data (@data)
                    {
                        my $object = $data->nodeName();
                        my @fields = $data->childNodes();
                        foreach my $field (@fields)
                        {
                            my $name     = $field->nodeName();
                            my $value    = $field->textContent();
                            my $dataType = $metadata{$name} // 'String';
                            if ($dataType eq 'Currency' && $value ne '')
                            {
                                if ($value =~ /(-?)inf/i)
                                {
                                    $value = $1 . 'Infinity';
                                }
                                else
                                {
                                    $value = sprintf("%.${currency_precision}f",
                                        $value / $time_division);
                                }
                            }
                            $size{$name} = length($value)
                              if length($value) > $size{$name};
                        }
                    }
                    # Print header if not quiet
                    unless ($quiet)
                    {
                        my @fields = $data[0]->childNodes();
                        foreach my $field (@fields)
                        {
                            my $name = $field->nodeName();
                            printf "%-$size{$name}s ", $name;
                        }
                        print "\n";
                        foreach my $field (@fields)
                        {
                            my $name = $field->nodeName();
                            printf "%-$size{$name}s ", '-' x $size{$name};
                        }
                        print "\n";
                    }

                    # Print data
                    my $rowIndex = 0;
                    foreach my $data (@data)
                    {
                        my $object = $data->nodeName();
                        my @fields = $data->childNodes();
                        # Separate totals if format is total
                        if ($format eq 'total' && ++$rowIndex == @data)
                        {
                            foreach my $field (@fields)
                            {
                                my $name = $field->nodeName();
                                printf "%-$size{$name}s ", '-' x $size{$name};
                            }
                            print "\n";
                        }
                        foreach my $field (@fields)
                        {
                            my $name     = $field->nodeName();
                            my $value    = $field->textContent();
                            my $dataType = $metadata{$name} // 'String';
                            $value =~ s/\n/\\n/g;    # Escape carriage returns
                            if ($dataType eq 'Currency' && $value ne '')
                            {
                                if ($value =~ /(-?)inf/i)
                                {
                                    printf("%-$size{$name}s ", $1 . 'Infinity');
                                }
                                else
                                {
                                    printf(
                                        "%$size{$name}.${currency_precision}f ",
                                        $value / $time_division);
                                }
                            }
                            elsif ($dataType eq 'Integer'
                                || $dataType eq 'AutoGen'
                                || $dataType eq 'Float')
                            {
                                printf("%$size{$name}s ", $value);
                            }
                            else
                            {
                                printf("%-$size{$name}s ", $value);
                            }
                        }
                        print "\n";
                    }
                }
            }
        }
    }

    # Print message if not quiet or not successfull
    if ($message && (! $quiet || $status ne 'Success')) { print "$message\n"; }
}

# ----------------------------------------------------------------------------
# &aggregateResponse( $response, $rawSelections )
# ----------------------------------------------------------------------------

sub aggregateResponse
{
    my ($response, $rawSelections) = @_;
    if ($log->is_debug())
    {
        $log->debug('Aggregating response');
    }

    # Return original response unless row count is positive
    return $response unless $response->getCount();

    # Return original response if there are no aggregation functions present
    return $response unless $rawSelections =~ /\(/;

    # Parse and check aggregate selections
    my @selectionNames     = ();
    my @selectionOperators = ();
    my %operatorIndices    = ();
    my %groupByIndices     = ();
    my @groupByIndices     = ();
    my $index              = 0;
    foreach my $selection (split /,/, $rawSelections)
    {

        if ($selection =~
            /^(?:(?:(\w+)\((\w+(?:\.\w+)?(?:\{[^\}]+\})?)?)\)|(\w+(?:\.\w+)?(?:\{[^\}]+\})?))(?:=(\w+))?$/
          )
        {
            my $op = $1 || 'GroupBy';
            my $complexName = $2 ? $2 : $3;
            my $alias = $4;
            my ($object, $name, $part);
            if ($complexName =~ /^(?:(\w+)\.)?(\w+)(?:\{([^\}]+)\})?/)
            {
                ($object, $name, $part) = ($1, $2, $3);
            }

            if    (defined $alias) { push @selectionNames, $alias; }
            elsif (defined $part)  { push @selectionNames, $part; }
            else                   { push @selectionNames, $name; }
            push @selectionOperators, $op;

            if ($op eq 'GroupBy')
            {
                push @groupByIndices, $index;
                $groupByIndices{$index} = 1;
            }
            elsif ($op eq 'Average'
                || $op eq 'Count'
                || $op eq 'Max'
                || $op eq 'Min'
                || $op eq 'Sum'
                || $op eq 'Unique')
            {
                $operatorIndices{$index} = 1;
            }
            else
            {
                return new MAM::Response()
                  ->failure('812',
                    "Invalid operator ($op) in selection string");
            }
        }
        else
        {
            push @selectionNames,     $selection;
            push @selectionOperators, 'GroupBy';
            push @groupByIndices,     $index;
            $groupByIndices{$index} = 1;
        }
        $index++;
    }

    # Prime the aggregate response
    my $aggregateResponse = new MAM::Response();

    $aggregateResponse->setStatus($response->getStatus());
    $aggregateResponse->setCode($response->getCode());
    $aggregateResponse->setMessage($response->getMessage());
    $aggregateResponse->setChunkNum($response->getChunkNum());

    # ------------------------------------------------------------------------
    # &buildAggregateValueRef( $aggregateValueRef, @groupByValues )
    # ------------------------------------------------------------------------

    local *buildAggregateValueRef = sub
    {
        my ($aggregateValueRef, @groupByValues) = @_;

        if (@groupByValues)
        {
            my $value = shift @groupByValues;
            if (@groupByValues)    # Still some values left after shifting
            {
                $aggregateValueRef->{$value} = {}
                  unless exists $aggregateValueRef->{$value};
                return buildAggregateValueRef($aggregateValueRef->{$value},
                    @groupByValues);
            }
            else    # Nothing left after shifting. This is the end
            {
                $aggregateValueRef->{$value} = []
                  unless exists $aggregateValueRef->{$value};
                return $aggregateValueRef->{$value};
            }
        }
        else        # There were no groupByValues from the beginning
        {
            return $aggregateValueRef;
        }
    };

    # Iterate over the data to build up an aggregate value hash
    # Notice the distinction between $aggregateValuesRef and $aggregateValueRef
    # $aggregateValuesRef is a reference to the head of the {HO}*L
    # while $aggregateValueRef is a reference to the leaf list.
    # There is a hash level for each groupby field.
    # At the leaf hash there is a list of the fields with the
    # groupby fields holding the group-by value and the operator fields
    # holding the aggregation value (sum, min, count, etc.). An extra list
    # element holds the count and is used for calculating the average.

    my $aggregateValuesRef;
    if   (@groupByIndices) { $aggregateValuesRef = {}; }
    else                   { $aggregateValuesRef = []; }
    my $datumName   = '';
    my $doc         = XML::LibXML::Document->new();
    my $dataElement = $response->getDataElement();
    $doc->setDocumentElement($dataElement);
    my %metadata = $response->getMetadata();

    foreach my $row ($dataElement->childNodes())
    {
        $datumName = $row->nodeName() unless $datumName;

        # Iterate the first time to build up a value tuple
        my @groupByValues = ();
        my $index         = 0;
        foreach my $col ($row->childNodes())
        {
            my $value = $col->textContent();

            if (defined $groupByIndices{$index})
            {
                push @groupByValues, $value;
            }
            $index++;
        }

       # Use the value tuple to build an aggregate value hash reference.
       # We may have referenced this tuple before or it may be the first time.
       # We need to use a recursive function because we do not know how many
       # levels we will have in the aggregate values hash (the group-by values).
        my $aggregateValueRef =
          buildAggregateValueRef($aggregateValuesRef, @groupByValues);

        # Iterate a second time to populate values into the aggregate value hash
        $index = 0;
        foreach my $col ($row->childNodes())
        {
            my $value = $col->textContent();

            if ($groupByIndices{$index})
            {
                $aggregateValueRef->[$index] = $value
                  unless defined $aggregateValueRef->[$index];
            }
            elsif ($operatorIndices{$index})
            {
                my $operator = $selectionOperators[$index];

                # Update metadata for aggregates from first row
                if ($index == 0)
                {
                    my $name = $col->nodeName();

                    if ($operator eq 'Average') { $metadata{$name} = 'Float'; }
                    elsif ($operator eq 'Count')
                    {
                        $metadata{$name} = 'Integer';
                    }
                }

                # Skip empty values
                if ($value eq '')
                {
                    if ($operator eq 'Count'
                        && ! defined $aggregateValueRef->[$index])
                    {
                        $aggregateValueRef->[$index] = 0;
                    }
                    $index++;
                    next;
                }

                if ($operator eq 'Average')
                {
                    if (defined $aggregateValueRef->[$index])
                    {
                        $aggregateValueRef->[$index] += $value;
                    }
                    else
                    {
                        $aggregateValueRef->[$index] = $value;
                    }

                    my $countIndex = @selectionNames + $index;
                    if (defined $aggregateValueRef->[$countIndex])
                    {
                        $aggregateValueRef->[$countIndex]++;
                    }
                    else
                    {
                        $aggregateValueRef->[$countIndex] = 1;
                    }
                }
                elsif ($operator eq 'Count')
                {
                    if (defined $aggregateValueRef->[$index])
                    {
                        $aggregateValueRef->[$index]++;
                    }
                    else
                    {
                        $aggregateValueRef->[$index] = 1;
                    }
                }
                elsif ($operator eq 'Max')
                {
                    if (! defined $aggregateValueRef->[$index]
                        || $value > $aggregateValueRef->[$index])
                    {
                        $aggregateValueRef->[$index] = $value;
                    }
                }
                elsif ($operator eq 'Min')
                {
                    if (! defined $aggregateValueRef->[$index]
                        || $value < $aggregateValueRef->[$index])
                    {
                        $aggregateValueRef->[$index] = $value;
                    }
                }
                elsif ($operator eq 'Sum')
                {
                    if (defined $aggregateValueRef->[$index])
                    {
                        $aggregateValueRef->[$index] += $value;
                    }
                    else
                    {
                        $aggregateValueRef->[$index] = $value;
                    }
                }
                elsif ($operator eq 'Unique')
                {
                    $aggregateValueRef->[$index]->{$value}++;
                }
            }
            $index++;
        }
    }

    # ------------------------------------------------------------------------
    # &buildResponseData( $aggregateValueRef, $newData )
    # ------------------------------------------------------------------------

    local *buildResponseData = sub
    {
        my ($aggregateValueRef, $newData) = @_;

        if (ref($aggregateValueRef) eq "HASH")
        {
            buildResponseData($aggregateValueRef->{$_}, $newData)
              foreach sort keys %$aggregateValueRef;
        }
        elsif (ref($aggregateValueRef) eq "ARRAY")
        {
            if (@selectionNames)
            {
                my @values = @{$aggregateValueRef};
                my $newRow = new XML::LibXML::Element($datumName);

                for (my $index = 0; $index < @selectionNames; $index++)
                {
                    my $name = $selectionNames[$index];
                    my $value;
                    if ($selectionOperators[$index] eq 'Unique')
                    {
                        $value = scalar keys %{$values[$index]};
                    }
                    else
                    {
                        $value = defined $values[$index] ? $values[$index] : '';
                        if ($selectionOperators[$index] eq 'Average')
                        {
                            my $count = $values[@selectionNames + $index];

                            $value =
                              $count == 0
                              ? 0.00
                              : sprintf("%.2f", $value / $count);
                        }
                    }
                    my $newCol = new XML::LibXML::Element($name);

                    $newCol->appendText($value);
                    $newRow->appendChild($newCol);
                }
                $newData->appendChild($newRow);
            }
        }
    };

    # Build up the aggregate response data
    # We use a recursive function since we do not know how many levels deep the
    # aggregateValueRef hash will be (based on the number of groupby elements).
    my $newData = new XML::LibXML::Element("Data");
    buildResponseData($aggregateValuesRef, $newData);
    $aggregateResponse->setDataElement($newData);
    $aggregateResponse->setMetadata(\%metadata);
    if (ref($aggregateValuesRef) eq 'HASH')
    {
        $aggregateResponse->setCount(scalar keys %$aggregateValuesRef);
    }
    elsif (ref($aggregateValuesRef) eq 'ARRAY')
    {
        $aggregateResponse->setCount(scalar @$aggregateValuesRef);
    }

    return $aggregateResponse;
}

# ----------------------------------------------------------------------------
# &enableDebug( $name, $value )
# ----------------------------------------------------------------------------

sub enableDebug
{
    my ($name, $value) = @_;

    my $screenAppender =
      new Log::Log4perl::Appender('Log::Log4perl::Appender::Screen',);
    my $layout = new Log::Log4perl::Layout::PatternLayout(
        '%d{yyyy-MM-dd HH:mm:ss.SSS} %-5p %M [%F:%L]  %m%n');
    $screenAppender->layout($layout);
    if ($value)
    {
        $screenAppender->threshold($value);
    }
    else
    {
        $screenAppender->threshold('TRACE');
    }

    $log->add_appender($screenAppender);
}

1;
