#! /usr/bin/perl -wT
################################################################################
#
# MAM Database object
#
# File   :  Database.pm
#
################################################################################
#
#    This file includes parts which are copyrighted by Adaptive Computing
#    Enterprises, Inc., and parts which are copyrighted by Battelle Memorial
#    Institute.
#
#    The terms and conditions for the use and redistribution of these parts
#    are governed by the Moab Accounting Manager License and the BSD License
#    respectively. Refer to the LICENSE file for details.
#
#    Copyright (C) 2006 - 2016 by Adaptive Computing Enterprises, Inc.
#    All Rights Reserved.
#
#    The Moab Accounting Manager License specifies the terms and conditions
#    for use and redistribution.
#
#    The Moab Accounting Manager License applies to the Moab Accounting
#    Manager software offered by Adaptive Computing Enterprises, Inc.
#    By installing or using this software, Licensee accepts a non-exclusive
#    license from Adaptive Computing Enterprises, Inc. and is bound to accept
#    acknowledgement of and abide by the notices and conditions of the
#    Moab Accounting Manager License.
#
#    Copyright (C) 2003 - 2005 Pacific Northwest National Laboratory,
#    Battelle Memorial Institute. All rights reserved.
#
#    The BSD License specifies the terms and conditions for use and
#    redistribution.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions
#    are met:
#
#    - Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
#    - Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
#    - Neither the name of Battelle nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
#    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
#    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
#    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#    POSSIBILITY OF SUCH DAMAGE.
#
################################################################################

use strict;

=head1 NAME

MAM::Database - performs database operations

=head1 DESCRIPTION

The B<MAM::Database> module obtains a handle and carries out database queries

=head1 CONSTRUCTORS

 my $db = new MAM::Database();
 my $db = new MAM::Database(handle => $handle);

=head1 ACCESSORS

=over 4

=item $handle = $database->getHandle();

=back

=head1 MUTATORS

=over 4

=item $handle = $database->setHandle($handle);

=back

=head1 OTHER METHODS

=over 4

=item $count = $database->insert(object => $object, actor => $actor, assignments => \@assignments, requestId => $requestId, txnId => $txnId, noLog => $noLog);

=item $results = $database->select(object => $object || objects => \@objects, selections => \@selections, conditions => \@conditions, options => \@options, chunkNum => $chunkNum, chunkSize => $chunkSize);

=item $count = $database->update(object => $object, actor => $actor, assignments => \@assignments, conditions => \@conditions, requestId => $requestId, txnId => $txnId, noLog => $noLog);

=item $count = $database->delete(object => $object, actor => $actor, conditions => \@conditions, requestId => $requestId, txnId => $txnId, noLog => $noLog);

=item $count = $database->undelete(object => $object, actor => $actor, conditions => \@conditions, requestId => $requestId, txnId => $txnId);

=item $count = $database->createTable(object => $object, requestId => $requestId, txnId => $txnId);

=item $count = $database->addColumn(request => $request, object => $object, attribute => $attribute, requestId => $requestId, txnId => $txnId);

=item $where = $database->buildWhere(object => $object || objects => \@objects, conditions => \@conditions);

=item $database->logTransaction(object => $object, action => $action, actor => $actor, assignments => \@assignments, conditions => \@conditions, options => \@options, data => \@data, count => $count, requestId => $requestId, txnId => $txnId);

=item $database->checkpoint($object, \@conditions);

=item $id = $database->nextId($object);

=item $string = $database->toString();

=back

=head1 EXAMPLES

use MAM::Database;

=head1 REQUIRES

Perl 5.8

=head1 AUTHOR

Scott Jackson, scottmo@adaptivecomputing.com

=cut

##############################################################################

package MAM::Database;

use vars qw($log);
use DBI;
use MAM::Cache;
use MAM::Condition;
use MAM::Exception;
use MAM::Global;
use MAM::Object;
use MAM::Selection;
use MAM::Assignment;

# ----------------------------------------------------------------------------
# Constructor
# ----------------------------------------------------------------------------

sub new
{
    my ($class, %arg) = @_;

    # Instantiate the object
    my $self = {
        _handle => $arg{handle},    # Database handle
    };
    bless $self, $class;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (',
            join(', ', map {"$_ => $arg{$_}"} keys %arg), ')');
    }

    # Obtain new database handle if necessary
    unless ($self->{_handle})
    {
        my $datasource = $serverConfig->{'database.datasource'}
          || $DB_DATASOURCE;
        my $dbuser = $serverConfig->{'database.user'} || (getpwuid($<))[0];
        my $dbpasswd = $serverConfig->{'database.password'};
        $self->{_handle} =
          DBI->connect($datasource, $dbuser, $dbpasswd,
            {AutoCommit => 0, RaiseError => 1});
    }

    return $self;
}

# ----------------------------------------------------------------------------
# Destructor
# ----------------------------------------------------------------------------

sub DESTROY
{
    my ($self) = @_;
    $self->{_handle}->disconnect() if $self->{_handle};
}

# ----------------------------------------------------------------------------
# Accessors
# ----------------------------------------------------------------------------

# Get the database handle
sub getHandle
{
    my ($self) = @_;
    return $self->{_handle};
}

# ----------------------------------------------------------------------------
# Mutators
# ----------------------------------------------------------------------------

# Set the database handle
sub setHandle
{
    my ($self, $handle) = @_;
    $self->{_handle} = $handle if $handle;
    return $self;
}

# ----------------------------------------------------------------------------
# $string = toString();
# ----------------------------------------------------------------------------

# Serialize message to printable string
sub toString
{
    my ($self) = @_;

    my $string = '[';

    $string .= $self->{_handle} if defined($self->{_handle});

    $string .= ']';

    return $string;
}

# ----------------------------------------------------------------------------
# $count = insert(object => $object,
#                   actor => $actor,
#                   assignments => \@assignments,
#                   requestId => $requestId,
#                   txnId => $txnId,
#                   noLog => $noLog,
#                   );
# ----------------------------------------------------------------------------

# perform insert
sub insert
{
    my ($self, %arg) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (',
            join(', ', map {"$_ => $arg{$_}"} keys %arg), ')');
    }

    # Declare and Initialize variables
    my $dbh = $self->{_handle};
    my $firstTime;
    my $object      = $arg{object};
    my $actor       = $arg{actor};
    my @assignments = $arg{assignments} ? @{$arg{assignments}} : ();
    my @options     = $arg{options} ? @{$arg{options}} : ();
    my $requestId   = $arg{requestId};
    my $txnId       = $arg{txnId} ? $arg{txnId} : $self->nextId('Transaction');
    my $noLog       = $arg{noLog};
    my $now         = $requestTime;

    # Obtain attribute datatypes
    my %dataTypes = ();
    my $results   = $self->select(
        object     => 'Attribute',
        selections => [
            new MAM::Selection(name => 'Name'),
            new MAM::Selection(name => 'DataType')
        ],
        conditions => [new MAM::Condition(name => 'Object', value => $object)]
    );
    foreach my $attributeRow (@{$results->{data}})
    {
        my $name     = $attributeRow->[0];
        my $dataType = $attributeRow->[1];
        $dataTypes{$name} = $dataType;
    }

    # Build SQL string
    my $sql = 'INSERT INTO ';

    # Add object to SQL
    $sql .= toDBC($object);

    # Add column list
    $sql .= ' (';
    $firstTime = 1;

    # Handle auto-generated name if one exists
    foreach my $attribute (keys %dataTypes)
    {
        if ($dataTypes{$attribute} eq 'AutoGen')
        {
            # Add the name
            if ($firstTime) { $firstTime = 0; }
            else            { $sql .= ','; }
            $sql .= toDBC($attribute);
        }
    }

    # Handle other specified names
    foreach my $assignment (@assignments)
    {
        # Add the name
        if ($firstTime) { $firstTime = 0; }
        else            { $sql .= ','; }
        my $name = $assignment->getName();
        $sql .= toDBC($name);
    }
    if ($firstTime) { $firstTime = 0; }
    else            { $sql .= ','; }
    $sql .=
      'g_creation_time,g_modification_time,g_request_id,g_transaction_id)';

    # Add insert items
    $sql .= ' VALUES (';
    $firstTime = 1;

    # Handle auto-generated values
    my @autoGenAssignments = ();
    foreach my $attribute (keys %dataTypes)
    {
        if ($dataTypes{$attribute} eq 'AutoGen')
        {
            # Add the value
            if ($firstTime) { $firstTime = 0; }
            else            { $sql .= ','; }
            # Get the next key generated value and assign it to the attribute
            my $value = $self->nextId($object);
            $sql .= $value;
            push @autoGenAssignments,
              new MAM::Assignment(name => $attribute, value => $value);
        }
    }

    # Handle other specified values
    foreach my $assignment (@assignments)
    {
        # Add the value
        if ($firstTime) { $firstTime = 0; }
        else            { $sql .= ','; }
        my $value = $assignment->getValue();
        $value =~ s/'/''/g;    # Escape quotes in values
        $sql .= "'$value'";
    }
    if ($firstTime) { $firstTime = 0; }
    else            { $sql .= ','; }
    $sql .= "'$now','$now','$requestId','$txnId')";

    # Perform the insert
    if ($log->is_debug())
    {
        $log->debug("SQL Update: $sql");
    }
    my $count = $dbh->do($sql);

    # Log the transaction
    $self->logTransaction(
        requestId   => $requestId,
        txnId       => $txnId,
        object      => $object,
        action      => 'Create',
        actor       => $actor,
        assignments => [@autoGenAssignments, @assignments],
        options     => \@options,
        count       => $count
    ) unless $noLog;

    # Return the number of objects/associations updated
    return $count;
}

# ----------------------------------------------------------------------------
# $results = select(object => $object || objects => \@objects,
#                   selections => \@selections,
#                   conditions => \@conditions,
#                   options => \@options,
#                   chunkNum => $chunkNum,
#                   chunkSize => $chunkSize
#                   );
# ----------------------------------------------------------------------------

# perform select
sub select
{
    my ($self, %arg) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (',
            join(', ', map {"$_ => $arg{$_}"} keys %arg), ')');
    }

    # Declare and Initialize variables
    my $dbh = $self->{_handle};
    my $sth;
    my $sql               = '';
    my $time              = '';
    my $unique            = '';
    my $limit             = '';
    my $offset            = '';
    my $dbname            = '';
    my $sortBySequence    = 0;
    my @sortSelections    = ();
    my @groupbySelections = ();
    my @objects =
        defined $arg{object} ? (new MAM::Object(name => $arg{object}))
      : defined $arg{objects} ? @{$arg{objects}}
      : throw MAM::Exception(
        "At least one object must be specified in \$database->select()");
    my @selections = defined $arg{selections} ? @{$arg{selections}} : ();
    my @conditions = defined $arg{conditions} ? @{$arg{conditions}} : ();
    my @options    = defined $arg{options}    ? @{$arg{options}}    : ();
    my $chunkNum   = defined $arg{chunkNum}   ? $arg{chunkNum}      : 0;

    # Build Select Command
    $sql .= 'SELECT ';
    my $tmpSQL = '';

    # Parse options
    foreach my $option (@options)
    {
        my $name  = $option->getName();
        my $value = $option->getValue();
        if ($name eq 'Time')
        {
            $time = $value;
        }
        elsif ($name eq 'Unique' && $value !~ /false/i)
        {
            $unique = $value;
        }
        elsif ($name eq 'Limit')
        {
            $limit = $value;
        }
        elsif ($name eq 'Offset')
        {
            $offset = $value;
        }
        elsif ($name eq 'SortBySequence'
            && $value !~ /false/i
            && $objects[0]->getName() eq 'Attribute')
        {
            $sortBySequence = 1;
        }
    }

    # Build selection item list
    # If joined object, we expect all names to be object qualified
    my %metadata = ();
    if (@selections)
    {
        my $index = 0;
        foreach my $selection (@selections)
        {
            my $name   = $selection->getName();
            my $part   = $selection->getPart();
            my $op     = $selection->getOperator();
            my $object = $selection->getObject();
            my $alias  = $selection->getAlias();
            $tmpSQL .= ',' if $index > 0;

            # Fix selection name
            $dbname = toDBC($name);

            # Fix object dbname for join selections
            if ($object) { $dbname = toDBC($object) . '.' . $dbname; }

            # Handle operators
            if ($op)
            {
                if ($op eq 'Count')
                {
                    $tmpSQL .= 'COUNT(';
                }
                elsif ($op eq 'Sum')
                {
                    $tmpSQL .= 'SUM(';
                }
                elsif ($op eq 'Average')
                {
                    $tmpSQL .= 'AVG(';
                }
                elsif ($op eq 'Min')
                {
                    $tmpSQL .= 'MIN(';
                }
                elsif ($op eq 'Max')
                {
                    $tmpSQL .= 'MAX(';
                }
                elsif ($op eq 'Sort')
                {
                    push @sortSelections, toDBC($dbname);
                }
                elsif ($op eq 'Tros')
                {
                    push @sortSelections, toDBC($dbname) . " DESC";
                }
                elsif ($op eq 'GroupBy')
                {
                    push @groupbySelections, $dbname;
                }
            }

            if ($time ne '')
            {
                $tmpSQL .= 'a.';
            }

            if ($part)
            {
                # Use part as the name if set and no alias specified
                $alias = $part unless $alias;

                $tmpSQL .= "substring(${dbname},E'\"${part}\":\"?(\\\\w+)\"?')";
            }
            else
            {
                $tmpSQL .= $dbname;
            }

            # Close aggregating operators
            if ($op
                && (   $op eq 'Sum'
                    || $op eq 'Count'
                    || $op eq 'Max'
                    || $op eq 'Min'
                    || $op eq 'Average')
              )
            {
                if ($alias)
                {
                    $tmpSQL .= ') AS ' . toDBC($alias);
                }
                else
                {
                    $tmpSQL .= ') AS ' . toDBC($name);
                }
            }
            else
            {
                if ($alias)
                {
                    $tmpSQL .= ' AS ' . toDBC($alias);
                }
                elsif ($object)
                {
                    $tmpSQL .= ' AS ' . toDBC($name);
                }
            }

            # Prepopulate metadata per selection details
            my %properties = MAM::Cache->getAttributeProperties($object
                  || $objects[0]->getName(), $name);
            if ($op eq 'Average') { $metadata{$alias || $name} = 'Float'; }
            elsif ($op eq 'Count')
            {
                $metadata{$alias || $name} = 'Integer';
            }
            elsif ($part)
            {
                if (defined $properties{'DataType'}
                    && $properties{'DataType'} =~ /JSON:(\w+)/)
                {
                    $metadata{$alias || $name} = $1;
                }
                else { $metadata{$alias || $name} = 'String'; }
            }
            elsif (defined $properties{'DataType'})
            {
                $metadata{$alias || $name} = $properties{'DataType'};
            }
            else { $metadata{$alias || $name} = 'String'; }

            $index++;
        }

        # Handle requests for unique results
        if (defined $unique && $unique =~ /True/i)
        {
            $sql .= 'DISTINCT ';
        }

        # Add selection item list to SQL
        $sql .= $tmpSQL;
    }

    # Default to * if selections are empty
    else
    {
        $sql .= '*';
    }

    # Add object dbname to sql
    $sql .= ' FROM ';

    # No time travel (but possible joined objects)
    if ($time eq '')
    {
        my $firstTime = 1;
        foreach my $object (@objects)
        {
            my $name  = $object->getName();
            my $join  = $object->getJoin();
            my $alias = $object->getAlias();
            if ($firstTime) { $firstTime = 0; }
            else            { $sql .= $join ? " $join " : ', '; }
            $sql .= toDBC($name);
            $sql .= " $alias" if $alias;
        }

        # Add search conditions
        $sql .=
          $self->buildWhere(objects => \@objects, conditions => \@conditions);
    }

    # Time travel requested (time travel not supported for joined objects)
    else
    {
        if (@objects != 1)
        {
            $log->error('Time travel not supported for multi-object queries');
            throw MAM::Exception('700',
                'Time travel not supported for multi-object queries');
        }

        my $name   = $objects[0]->getName();
        my $dbname = toDBC($name);
        $sql .=
          "( SELECT * FROM $dbname UNION SELECT * FROM ${dbname}_log) AS a, ";

        my $firstOne    = 1;
        my $primaryKeys = '';
        my @primaryKeys = ();
        foreach my $attribute (
            MAM::Cache->listAttributesWithPropertyValue(
                $name, 'PrimaryKey', 'True'
            )
          )
        {
            if ($firstOne) { $firstOne = 0; }
            else           { $primaryKeys .= ','; }
            $primaryKeys .= toDBC($attribute);
            push @primaryKeys, toDBC($attribute);
        }
        if ($primaryKeys eq '')
        {
            $log->error('Cannot timetravel. ' +
                  ucfirst($name) + ' does not have a primary field');
            throw MAM::Exception('700',
                'Cannot timetravel. ' +
                  ucfirst($name) +
                  ' does not have a primary field');
        }

        $sql .=
          "( SELECT $primaryKeys,MAX(g_transaction_id) AS g_transaction_id FROM ( SELECT $primaryKeys,g_transaction_id FROM $dbname WHERE g_modification_time<=$time UNION SELECT $primaryKeys,g_transaction_id FROM ${dbname}_log WHERE g_modification_time<=$time) as c GROUP BY $primaryKeys) AS b ";

        # Add search conditions
        $sql .= $self->buildWhere(
            objects    => \@objects,
            conditions => \@conditions,
            options    => \@options
        );

        # Add join conditions
        $sql .= ' AND a.g_transaction_id=b.g_transaction_id';
        foreach my $key (@primaryKeys)
        {
            $sql .= " AND a.$key=b.$key";
        }
    }

    # Handle groupby requests
    if (@groupbySelections)
    {
        $sql .= ' GROUP BY ' . join(',', @groupbySelections);
    }

    # Handle sort requests
    if ($sortBySequence) { push @sortSelections, 'g_sequence'; }
    if (@sortSelections)
    {
        $sql .= ' ORDER BY ' . join(',', @sortSelections);
    }

    # Handle limit (An express Limit overrides chunking)
    if ($limit)
    {
        $sql .= " LIMIT $limit";
        if ($offset)
        {
            $sql .= " OFFSET $offset";
        }
    }
    # Implement chunking unless chunkNum is set to zero
    elsif ($chunkNum > 0)
    {
        my $chunkSize =
            $arg{chunkSize}
          ? $arg{chunkSize}
          : $serverConfig->{'response.chunksize'} || 1000000000;
        $sql .= " LIMIT $chunkSize OFFSET " . ($chunkNum - 1) * $chunkSize;
        # Exchange for the line below for mysql versions prior to 4.0.6
        #$sql .= ' LIMIT ' . ($chunkNum - 1) * $chunkSize . ',' . $chunkSize;
    }

    # Perform Select
    if ($log->is_debug())
    {
        $log->debug("SQL Query: $sql");
    }
    $sth = $dbh->prepare($sql);
    $sth->execute();

    my %results = ();
    $results{cols}  = $sth->{NUM_OF_FIELDS};
    $results{names} = [map { toUCC($_) } @{$sth->{NAME_lc}}];
    $results{data}  = $sth->fetchall_arrayref();
    $results{rows}  = @{$results{data}};

    # Postpopulate metadata from names
    foreach my $name (@{$results{names}})
    {
        if (! defined $metadata{$name})
        {
            # This should only happen for the case where no selections are
            # specified and there should additionally only be a single object.
            my %properties =
              MAM::Cache->getAttributeProperties($objects[0]->getName(), $name);
            if (defined $properties{'DataType'})
            {
                $metadata{$name} = $properties{'DataType'};
            }
            else { $metadata{$name} = $properties{'String'}; }
        }
    }
    $results{metadata} = \%metadata;

    # Log the results
    if ($log->is_debug())
    {
        my $entry = '';
        foreach my $row (@{$results{data}})
        {
            $entry .= join('|', map { defined $_ ? $_ : '' } @$row) . "\n";
        }
        $log->debug("SQL Results: $entry");
    }

    return \%results;
}

# ----------------------------------------------------------------------------
# $count = update(object => $object,
#                   actor => $actor,
#                   assignments => \@assignments,
#                   conditions => \@conditions,
#                   requestId => $requestId,
#                   txnId => $txnId,
#                   noLog => $noLog,
#                   );
# ----------------------------------------------------------------------------

# perform update
sub update
{
    my ($self, %arg) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (',
            join(', ', map {"$_ => $arg{$_}"} keys %arg), ')');
    }

    # Declare and Initialize variables
    my $dbh         = $self->{_handle};
    my $firstTime   = 1;
    my $object      = $arg{object};
    my $actor       = $arg{actor};
    my @assignments = $arg{assignments} ? @{$arg{assignments}} : ();
    my @conditions  = $arg{conditions} ? @{$arg{conditions}} : ();
    my @options     = $arg{options} ? @{$arg{options}} : ();
    my $requestId   = $arg{requestId};
    my $txnId       = $arg{txnId} ? $arg{txnId} : $self->nextId('Transaction');
    my $noLog       = $arg{noLog};
    my $now         = $requestTime;

    # Build SQL string
    my $sql = 'UPDATE ';

    # Add table to SQL
    $sql .= toDBC($object);

    # Add assignment list
    $sql .= ' SET ';
    foreach my $assignment (@assignments)
    {
        if ($firstTime) { $firstTime = 0; }
        else            { $sql .= ','; }
        my $name   = $assignment->getName();
        my $dbname = toDBC($name);
        if (MAM::Cache->getAttributeProperty($object, $name, 'Fixed') eq 'True')
        {
            $log->error("$name is a fixed field and cannot be modified");
            throw MAM::Exception('740',
                "$name is a fixed field and cannot be modified");
        }
        my $value = $assignment->getValue();
        my $op    = $assignment->getOperator();
        $sql .= "$dbname=";
        if    ($op eq 'Inc') { $sql .= "$dbname+"; }
        elsif ($op eq 'Dec') { $sql .= "$dbname-"; }
        if    ($value ne 'NULL')
        {
            $value =~ s/'/''/g;    # Escape quotes in values
            $sql .= "'$value'";
        }
        else
        {
            $sql .= $value;
        }
    }
    if ($firstTime) { $firstTime = 0; }
    else            { $sql .= ','; }
    $sql .=
      "g_modification_time=$now,g_request_id=$requestId,g_transaction_id=$txnId";

    # Add search conditions
    $sql .= $self->buildWhere(object => $object, conditions => \@conditions);

    # Perform checkpoint
    $self->checkpoint($object, \@conditions);

    # Perform SQL Update
    if ($log->is_debug())
    {
        $log->debug("SQL Update: $sql");
    }
    my $count = $dbh->do($sql);
    $count = $count eq '0E0' ? 0 : $count;
    if ($log->is_debug())
    {
        $log->debug("SQL Rows: $count");
    }

    # Log the transaction
    $self->logTransaction(
        requestId   => $requestId,
        txnId       => $txnId,
        object      => $object,
        action      => 'Modify',
        actor       => $actor,
        assignments => \@assignments,
        conditions  => \@conditions,
        options     => \@options,
        count       => $count
    ) unless $noLog;

    # Return the number of objects/associations updated
    return $count;
}

# ----------------------------------------------------------------------------
# $count = delete(object => $object,
#                   actor => $actor,
#                   conditions => \@conditions,
#                   requestId => $requestId,
#                   txnId => $txnId,
#                   noLog => $noLog,
#                   );
# ----------------------------------------------------------------------------

# perform delete
sub delete
{
    my ($self, %arg) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (',
            join(', ', map {"$_ => $arg{$_}"} keys %arg), ')');
    }

    # Declare and Initialize variables
    my $dbh        = $self->{_handle};
    my $object     = $arg{object};
    my $actor      = $arg{actor};
    my @conditions = $arg{conditions} ? @{$arg{conditions}} : ();
    my @options    = $arg{options} ? @{$arg{options}} : ();
    my $requestId  = $arg{requestId};
    my $txnId      = $arg{txnId} ? $arg{txnId} : $self->nextId('Transaction');
    my $noLog      = $arg{noLog};
    my $now        = $requestTime;

    # Build SQL string
    my $sql = 'UPDATE ';
    $sql .= toDBC($object);

    # Change deleted = true for all records to be deleted
    $sql .=
      " SET g_deleted='True',g_modification_time=$now,g_request_id=$requestId,g_transaction_id=$txnId ";

    # Build where string
    my $where =
      $self->buildWhere(object => $object, conditions => \@conditions);
    $sql .= $where;

    # Perform checkpoint
    $self->checkpoint($object, \@conditions);

    # Perform SQL Update
    if ($log->is_debug())
    {
        $log->debug("SQL Update: $sql");
    }
    my $count = $dbh->do($sql);
    $count = $count eq '0E0' ? 0 : $count;
    if ($log->is_debug())
    {
        $log->debug("SQL Rows: $count");
    }

    # Log the transaction
    $self->logTransaction(
        requestId  => $requestId,
        txnId      => $txnId,
        object     => $object,
        action     => 'Delete',
        actor      => $actor,
        conditions => \@conditions,
        options    => \@options,
        count      => $count
    ) unless $noLog;

    # Return the number of objects/associations updated
    return $count;
}

# ----------------------------------------------------------------------------
# $count = undelete(object => $object,
#                   actor => $actor,
#                   conditions => \@conditions,
#                   requestId => $requestId,
#                   txnId => $txnId
#                   );
# ----------------------------------------------------------------------------

# perform undelete
sub undelete
{
    my ($self, %arg) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (',
            join(', ', map {"$_ => $arg{$_}"} keys %arg), ')');
    }

    # Declare and Initialize variables
    my $dbh          = $self->{_handle};
    my $object       = $arg{object};
    my $actor        = $arg{actor};
    my @inConditions = $arg{conditions} ? @{$arg{conditions}} : ();
    my @conditions = @inConditions;                           # Copies by value
    my @options    = $arg{options} ? @{$arg{options}} : ();
    my $requestId  = $arg{requestId};
    my $txnId = $arg{txnId} ? $arg{txnId} : $self->nextId('Transaction');
    my $now   = $requestTime;

    # Build SQL string
    my $sql = 'UPDATE ';
    $sql .= toDBC($object);

    # Change deleted = false for all records to be undeleted
    $sql .=
      " SET g_deleted='False',g_modification_time=$now,g_request_id=$requestId,g_transaction_id=$txnId ";

    # Build where string
    push @conditions, new MAM::Condition(name => 'Deleted', value => 'True');
    my $where =
      $self->buildWhere(object => $object, conditions => \@conditions);
    $sql .= $where;

    # Perform checkpoint
    $self->checkpoint($object, \@conditions);

    # Perform SQL Update
    if ($log->is_debug())
    {
        $log->debug("SQL Update: $sql");
    }
    my $count = $dbh->do($sql);
    $count = $count eq '0E0' ? 0 : $count;
    if ($log->is_debug())
    {
        $log->debug("SQL Rows: $count");
    }

    # Log the transaction
    $self->logTransaction(
        requestId  => $requestId,
        txnId      => $txnId,
        object     => $object,
        action     => 'Undelete',
        actor      => $actor,
        conditions => \@inConditions,
        options    => \@options,
        count      => $count
    );

    # Return the number of objects/associations updated
    return $count;
}

# ----------------------------------------------------------------------------
# $count = createTable(object => $object,
#                   requestId => $requestId,
#                   txnId => $txnId
#                   );
# ----------------------------------------------------------------------------

# create table
sub createTable
{
    my ($self, %arg) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (',
            join(', ', map {"$_ => $arg{$_}"} keys %arg), ')');
    }

    # Declare and Initialize variables
    my $dbh       = $self->{_handle};
    my $object    = $arg{object};
    my $requestId = $arg{requestId};
    my $txnId     = $arg{txnId};
    my $count     = 0;
    my $now       = $requestTime;
    my $sql;

    my $fields =
      " (\n\tg_creation_time int not null,\n\tg_modification_time int not null,\n\tg_deleted varchar(5) default 'False',\n\tg_request_id int not null,\n\tg_transaction_id int not null\n)";

    # Create new database table for object
    $sql = 'CREATE TABLE ' . toDBC($object) . $fields;
    if ($log->is_debug())
    {
        $log->debug("SQL Update: $sql");
    }
    $count += $dbh->do($sql);

    # Create new database table for object journal
    $sql = 'CREATE TABLE ' . toDBC($object) . '_log' . $fields;
    if ($log->is_debug())
    {
        $log->debug("SQL Update: $sql");
    }
    $count += $dbh->do($sql);

    # Add creationTime attribute
    $sql =
      "INSERT INTO g_attribute (g_object,g_name,g_data_type,g_primary_key,g_required,g_fixed,g_sequence,g_hidden,g_description,g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id) VALUES ('$object','CreationTime','TimeStamp','False','False','True',950,'True','First Created',$now,$now,'False',$requestId,$txnId)";
    if ($log->is_debug())
    {
        $log->debug("SQL Update: $sql");
    }
    $count += $dbh->do($sql);

    # Add modificationTime attribute
    $sql =
      "INSERT INTO g_attribute (g_object,g_name,g_data_type,g_primary_key,g_required,g_fixed,g_sequence,g_hidden,g_description,g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id) VALUES ('$object','ModificationTime','TimeStamp','False','False','False',960,'True','Last Updated',$now,$now,'False',$requestId,$txnId)";
    if ($log->is_debug())
    {
        $log->debug("SQL Update: $sql");
    }
    $count += $dbh->do($sql);

    # Add deleted attribute
    $sql =
      "INSERT INTO g_attribute (g_object,g_name,g_data_type,g_primary_key,g_required,g_fixed,g_default_value,g_sequence,g_hidden,g_description,g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id) VALUES ('$object','Deleted','Boolean','False','False','False','False',970,'True','Is this object deleted?',$now,$now,'False',$requestId,$txnId)";
    if ($log->is_debug())
    {
        $log->debug("SQL Update: $sql");
    }
    $count += $dbh->do($sql);

    # Add deleted index
    $sql =
        'CREATE INDEX '
      . toDBC($object)
      . '_deleted_idx ON '
      . toDBC($object)
      . ' (g_deleted)';
    if ($log->is_debug())
    {
        $log->debug("SQL Update: $sql");
    }
    $count += $dbh->do($sql);

    # Add requestId attribute
    $sql =
      "INSERT INTO g_attribute (g_object,g_name,g_data_type,g_primary_key,g_required,g_fixed,g_sequence,g_hidden,g_description,g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id) VALUES ('$object','RequestId','Integer','False','False','False',980,'True','Last Modifying Request Id',$now,$now,'False',$requestId,$txnId)";
    if ($log->is_debug())
    {
        $log->debug("SQL Update: $sql");
    }
    $count += $dbh->do($sql);

    # Add txnId attribute
    $sql =
      "INSERT INTO g_attribute (g_object,g_name,g_data_type,g_primary_key,g_required,g_fixed,g_sequence,g_hidden,g_description,g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id) VALUES ('$object','TransactionId','Integer','False','False','False',990,'True','Last Modifying Transaction Id',$now,$now,'False',$requestId,$txnId)";
    if ($log->is_debug())
    {
        $log->debug("SQL Update: $sql");
    }
    $count += $dbh->do($sql);

    # Add txnId index
    $sql =
        'CREATE INDEX '
      . toDBC($object)
      . '_txnid_idx ON '
      . toDBC($object)
      . ' (g_transaction_id)';
    if ($log->is_debug())
    {
        $log->debug("SQL Update: $sql");
    }
    $count += $dbh->do($sql);

    # Return the number of tables/columns updated
    $count = $count eq '0E0' ? 0 : $count;
    return $count;
}

# ----------------------------------------------------------------------------
# $count = addColumn(request => $request,
#                   object => $object,
#                   attribute => $attribute,
#                   requestId => $requestId,
#                   txnId => $txnId
#                   );
# ----------------------------------------------------------------------------

# Add column
sub addColumn
{
    my ($self, %arg) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (',
            join(', ', map {"$_ => $arg{$_}"} keys %arg), ')');
    }

    # Declare and Initialize variables
    my $dbh       = $self->{_handle};
    my $request   = $arg{request};
    my $object    = $arg{object};
    my $attribute = $arg{attribute};
    my $requestId = $arg{requestId};
    my $txnId     = $arg{txnId};
    my $count     = 0;

    my $dataType   = $request->getAssignmentValue('DataType');
    my $primaryKey = $request->getAssignmentValue('PrimaryKey');

    # Build SQL string
    my $sql = 'ALTER TABLE ' . toDBC($object);
    $sql .= ' ADD COLUMN ' . toDBC($attribute) . ' ';

    # Add datatype and size specifications
    if ($dataType eq 'String')
    {
        $sql .= 'varchar(1024)';
    }
    elsif ($dataType =~ /^JSON/)
    {
        $sql .= 'text';
    }
    elsif ($dataType eq 'Integer'
        || $dataType eq 'AutoGen'
        || $dataType eq 'TimeStamp')
    {
        $sql .= 'int';
    }
    elsif ($dataType eq 'Float' || $dataType eq 'Currency')
    {
        $sql .= 'double precision';
    }
    elsif ($dataType eq 'Boolean')
    {
        $sql .= 'varchar(5)';
    }
    else
    {
        $log->error(
            "Invalid column datatype ($dataType) being added for $object->$attribute"
        );
        throw MAM::Exception('317',
            "Invalid column datatype ($dataType) being added for $object->$attribute"
        );
    }

    # Perform SQL Update
    if ($log->is_debug())
    {
        $log->debug("SQL Update: $sql");
    }
    $count += $dbh->do($sql);

    # Return the number of tables/columns updated
    $count = $count eq '0E0' ? 0 : $count;

    # Add index for all new non-journal primary keys
    if ($object !~ /Log$/ && defined $primaryKey && $primaryKey =~ /True/i)
    {
        my $fieldName = toDBC($attribute);
        (my $shortFieldName = $fieldName) =~ s/^g_//;
        $sql =
            'CREATE INDEX '
          . toDBC($object) . '_'
          . $shortFieldName
          . '_idx ON '
          . toDBC($object)
          . " ($fieldName)";
        if ($log->is_debug())
        {
            $log->debug("SQL Update: $sql");
        }
        $count += $dbh->do($sql);
    }

    return $count;
}

# ----------------------------------------------------------------------------
# $where = buildWhere(object => $object || objects => \@objects,
#                   conditions =>\@conditions, options => \@options);
# ----------------------------------------------------------------------------

# Build where string
sub buildWhere
{
    my ($self, %arg) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (',
            join(', ', map {"$_ => $arg{$_}"} keys %arg), ')');
    }

    my @objects =
        defined $arg{object} ? (new MAM::Object(name => $arg{object}))
      : defined $arg{objects} ? @{$arg{objects}}
      :                         ();
    my @conditions = defined $arg{conditions} ? @{$arg{conditions}} : ();
    my @options    = defined $arg{options}    ? @{$arg{options}}    : ();
    my $where      = '';
    my $firstTime  = 1;
    my $deleted    = 0;
    my $dbname;
    my $time = '';

    # Parse options
    foreach my $option (@options)
    {
        my $name  = $option->getName();
        my $value = $option->getValue();
        if ($name eq 'Time')
        {
            $time = $value;
        }
    }

    # Add search conditions
    if (@conditions)
    {
        $where .= ' WHERE ( ';
        foreach my $condition (@conditions)
        {
            my $name    = $condition->getName();
            my $part    = $condition->getPart();
            my $value   = $condition->getValue();
            my $op      = $condition->getOperator();
            my $conj    = $condition->getConjunction();
            my $group   = $condition->getGroup();
            my $object  = $condition->getObject();
            my $subject = $condition->getSubject();

            # Ignore deleted unless deleted=true
            if ($name eq 'Deleted' && $value =~ /True/i)
            {
                $deleted = 1;
            }

            # Add the conjunction
            if ($firstTime) { $firstTime = 0; }
            else
            {
                if ($conj eq 'Or' || $conj eq 'OrNot')
                {
                    $where .= ' OR ';
                }
                else
                {
                    $where .= ' AND ';
                }
            }

            # Add grouping
            if ($group > 0)
            {
                $where .= ' (' x $group . ' ';
            }

            # Add not
            if ($conj eq 'OrNot' || $conj eq 'AndNot')
            {
                $where .= ' NOT ';
            }

            # Add the name
            if ($time ne '')
            {
                $where .= 'a.';
            }

            # Fix attribute name
            $dbname = toDBC($name);

            # Fix object dbname for join conditions
            if   ($object) { $dbname = toDBC($object) . '.' . $dbname; }
            else           { $object = $objects[0]->getName(); }

            if ($part)
            {
                $where .= "substring(${dbname},E'\"${part}\":\"?(\\\\w+)\"?')";
            }
            else
            {
                $where .= $dbname;
            }

            # Add the operator
            if ($value eq 'NULL' && (! $op || $op eq 'EQ'))
            {
                $where .= ' IS ';
            }
            elsif ($value eq 'NULL' && $op eq 'NE') { $where .= ' IS NOT '; }
            elsif (! $op || $op eq 'EQ') { $where .= '='; }
            elsif ($op eq 'GT') { $where .= '>'; }
            elsif ($op eq 'GE') { $where .= '>='; }
            elsif ($op eq 'LT') { $where .= '<'; }
            elsif ($op eq 'LE') { $where .= '<='; }
            elsif ($op eq 'NE') { $where .= '!='; }
            elsif ($op eq 'Match')
            {
                $where .= ' LIKE ';
                $value =~ s/\*/%/g;
                $value =~ s/\?/_/g;
            }
            elsif ($op eq 'NotMatch')
            {
                $where .= ' NOT LIKE ';
                $value =~ s/\*/%/g;
                $value =~ s/\?/_/g;
            }
            else { $where .= '='; }

            # If this is a NULL comparison
            if ($value eq 'NULL')
            {
                $where .= "$value";
            }

            # If this is not a joined value
            elsif (! $subject)
            {
                $where .= "'$value'";
            }

            # Else it is a joined value
            else
            {
                # Fix attribute name
                $where .= toDBC($subject) . '.' . toDBC($value);
            }

            # Add ungrouping
            if ($group < 0)
            {
                $where .= ' )' x -$group . ' ';
            }
        }
    }

    if (! $deleted)
    {
        # Avoid querying or updating deleted records
        if ($firstTime)
        {
            $where .= ' WHERE ';
            $firstTime = 0;
        }
        else
        {
            $where .= ' ) AND ';
        }

        # Split and handle joined objects
        if (@objects == 1)
        {
            $where .= "g_deleted!='True'";
        }
        else
        {
            my $anotherFirstTime = 1;
            foreach my $object (@objects)
            {
                if ($anotherFirstTime) { $anotherFirstTime = 0; }
                else                   { $where .= ' AND '; }

                $dbname = toDBC($object->getAlias());
                $dbname = toDBC($object->getName()) unless $dbname;
                $where .= "$dbname.g_deleted!='True'";
            }
        }
    }
    else    # include deleted
    {
        if (! $firstTime) { $where .= ' )'; }
    }

    return $where;
}

# ----------------------------------------------------------------------------
# logTransaction(object => $object,
#                   action => $action,
#                   actor => $actor,
#                   assignments => \@assignments,
#                   conditions => \@conditions,
#                   options => \@options,
#                   data => \@data,
#                   count => $count,
#                   requestId => $requestId,
#                   txnId => $txnId
#                   );
# ----------------------------------------------------------------------------

# Log Transaction
sub logTransaction
{
    my ($self, %arg) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (',
            join(', ', map {"$_ => $arg{$_}"} keys %arg), ')');
    }

    # Declare and Initialize variables
    my $dbh = $self->{_handle};
    my $firstTime;
    my $object      = $arg{object};
    my $action      = $arg{action} || 'Create';
    my $actor       = $arg{actor};
    my @assignments = $arg{assignments} ? @{$arg{assignments}} : ();
    my @conditions  = $arg{conditions} ? @{$arg{conditions}} : ();
    my @options     = $arg{options} ? @{$arg{options}} : ();
    my @data        = $arg{data} ? @{$arg{data}} : ();
    my $count       = $arg{count};
    my $requestId   = $arg{requestId};
    my $txnId       = $arg{txnId};
    my $names       = 'g_object,g_action';
    my $values      = "'$object','$action'";
    my %key         = ();
    my %child       = ();
    my %details     = ();
    my $description = '';
    my $now         = $requestTime;
    my $association = MAM::Cache->getObjectProperty($object, 'Association');
    my $parent      = MAM::Cache->getObjectProperty($object, 'Parent');

    # Log the user requesting the transaction
    $names  .= ',g_actor';
    $values .= ",'$actor'";

    # Iterate over assignments
    foreach my $assignment (@assignments)
    {
        my $name  = $assignment->getName();
        my $value = $assignment->getValue();
        $value =~ s/'/''/g;    # Escape quotes in values
        my $op = $assignment->getOperator();

        # Append key or child
        if ($name eq 'Name' || $name eq 'Id')
        {
            if ($association eq 'True')
            {
                $child{$value}++;
            }
            else
            {
                $key{$value}++;
            }
        }

        # Append parent to key if association
        elsif ($association eq 'True'
            && $name eq $parent)
        {
            $key{$value}++;
        }

        # Append as detail
        else
        {
            my $detail = $name;
            if (! $op)
            {
                $detail .= '=';    # Use = for assignments
            }
            else
            {
                if    ($op eq 'Assign') { $detail .= '='; }
                elsif ($op eq 'Inc')    { $detail .= '+='; }
                elsif ($op eq 'Dec')    { $detail .= '-='; }
                else                    { $detail .= "{$op}"; }
            }
            $detail .= $value;
            $details{$detail}++;
        }
    }

    # Iterate over conditions
    foreach my $condition (@conditions)
    {
        my $name  = $condition->getName();
        my $value = $condition->getValue();
        my $op    = $condition->getOperator();

        # Append key or child
        if ($name eq 'Name' || $name eq 'Id')
        {
            if ($association eq 'True')
            {
                $child{$value}++;
            }
            else
            {
                $key{$value}++;
            }
        }

        # Append parent to key if association
        elsif ($association eq 'True'
            && $name eq $parent)
        {
            $key{$value}++;
        }

        # Append as detail
        else
        {
            my $detail = $name;
            if (! $op)
            {
                $detail .= '==';
            }
            else
            {
                if    ($op eq 'EQ')       { $detail .= '=='; }
                elsif ($op eq 'GT')       { $detail .= '>'; }
                elsif ($op eq 'GE')       { $detail .= '>='; }
                elsif ($op eq 'LT')       { $detail .= '<'; }
                elsif ($op eq 'LE')       { $detail .= '<='; }
                elsif ($op eq 'NE')       { $detail .= '!='; }
                elsif ($op eq 'Match')    { $detail .= '~='; }
                elsif ($op eq 'NotMatch') { $detail .= '!~'; }
                else                      { $detail .= "{$op}"; }
            }
            $detail .= $value;
            $details{$detail}++;
        }
    }

    # Iterate over options
    foreach my $option (@options)
    {
        my $name  = $option->getName();
        my $value = $option->getValue();
        my $op    = $option->getOperator();

        # Append key or child
        if ($name eq 'Name' || $name eq 'Id')
        {
            if ($association eq 'True')
            {
                $child{$value}++;
            }
            else
            {
                $key{$value}++;
            }
        }

        # Append parent to key if association
        elsif ($association eq 'True'
            && $name eq $parent)
        {
            $key{$value}++;
        }

        # Append description
        elsif ($name eq 'Description')
        {
            $description = $value;
        }

        # Append as detail
        else
        {
            my $detail = $name;
            if (! $op)
            {
                $detail .= ':=';
            }
            else
            {
                if   ($op eq 'Not') { $detail .= ':!'; }
                else                { $detail .= "{$op}"; }
            }
            $detail .= $value;
            $details{$detail}++;
        }
    }

    # Iterate over data
    foreach my $datum (@data)
    {
        $details{$datum->toString()}++;
    }

    # Add fields to sql names and values

    # Log key
    if (keys %key)
    {
        $names .= ',g_key';
        $values .= ",'" . join(',', sort keys %key) . "'";
    }

    # Log child
    if (keys %child)
    {
        $names .= ',g_child';
        $values .= ",'" . join(',', sort keys %child) . "'";
    }

    # Log count
    if (defined $count)
    {
        $names  .= ',g_count';
        $values .= ",$count";
    }

    # Log description
    if ($description ne '')
    {
        $names  .= ',g_description';
        $values .= ",'$description'";
    }

    # Log details
    if (keys %details)
    {
        my $details = join ',', sort keys %details;
        # Truncate details value if necessary
        if (length $details >= 1024)
        {
            $details = substr($details, 0, 1024);
        }
        $names  .= ',g_details';
        $values .= ",'$details'";
    }

    # Log creationdate, modificationdate, requestId, txnId and id
    $names .=
      ',g_creation_time,g_modification_time,g_request_id,g_transaction_id,g_id';
    $values .= ",$now,$now,$requestId,$txnId,$txnId";

    # Add column list
    my $sql = "INSERT INTO g_transaction ($names) VALUES ($values)";

    # Perform SQL Update
    if ($log->is_debug())
    {
        $log->debug("SQL Update: $sql");
    }
    $dbh->do($sql);
}

# ----------------------------------------------------------------------------
# checkpoint($object, \@conditions);
# ----------------------------------------------------------------------------

# Checkpoint State (diff style -- before Update)
sub checkpoint
{
    my ($self, $object, $conditions) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $dbh = $self->{_handle};

    # Add table reference list to SQL
    my $dbname = toDBC($object);

    # Copy the records matched by the given conditions into journal
    my $sql = "INSERT INTO ${dbname}_log SELECT * FROM $dbname";
    $sql .= $self->buildWhere(object => $object, conditions => $conditions);
    $dbh->do($sql);
}

# ----------------------------------------------------------------------------
# cleanupAfterUndo();
# ----------------------------------------------------------------------------

# Cleanup After Undo
sub cleanupAfterUndo
{
    my ($self) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $dbh = $self->{_handle};

    # We assume there is a requestId in the undo table since we should only
    # get here if we were called from within the Proxy constructor

    # undoId should be max(requestId) in Transaction
    # SELECT MAX(g_request_id) FROM g_transaction
    my $sql = 'SELECT MIN(g_request_id) FROM g_undo';
    if ($log->is_debug())
    {
        $log->debug("SQL Query: $sql");
    }
    my $sth = $dbh->prepare($sql);
    $sth->execute();
    my ($lowestUndoId) = $sth->fetchrow_array();
    unless ($lowestUndoId)
    {
        $log->error('There is no undo to clean up');
        throw MAM::Exception('720', 'There is no undo to clean up');
    }

    foreach my $object (MAM::Cache->listObjects())
    {
        my $dbname = toDBC($object);
        # Delete any records left around for possible redo's
        # DELETE FROM ${dbname}_log WHERE g_request_id>=$lowestUndoId
        $sql = "DELETE FROM ${dbname}_log WHERE g_request_id>=$lowestUndoId";
        if ($log->is_debug())
        {
            $log->debug("SQL Query: $sql");
        }
        $dbh->do($sql);
    }

    # Also got to blow away undo table
    # DELETE FROM Undo
    $sql = 'DELETE FROM g_undo';
    if ($log->is_debug())
    {
        $log->debug("SQL Query: $sql");
    }
    $dbh->do($sql);
}

# ----------------------------------------------------------------------------
# $id = nextId($object);
# ----------------------------------------------------------------------------

# Return next auto-generated unique id
sub nextId
{
    my ($self, $object) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    # Declare and Initialize variables
    my $dbh = $self->{_handle};

    # Increment the nextid for the next seeker
    my $count = $dbh->do(
        "UPDATE g_key_generator set g_next_id=g_next_id+1 WHERE g_name='$object'"
    );
    if ($count != 1)
    {
        $log->error("Unable to generate next id for $object");
        throw MAM::Exception('720', "Unable to generate next id for $object");
    }

    # Obtain the unique id for the specified object
    my $sth = $dbh->prepare(
        "SELECT g_next_id-1 FROM g_key_generator WHERE g_name='$object'");
    $sth->execute();
    my ($id) = $sth->fetchrow_array();
    unless ($id)
    {
        $log->error("Unable to obtain unique id for $object");
        throw MAM::Exception('720', "Unable to obtain unique id for $object");
    }
    if ($log->is_trace())
    {
        $log->trace("generated new id $id for $object");
    }

    return $id;
}

1;
