#! /usr/bin/perl -wT
################################################################################
#
# MAM Datum object
#
# File   :  Datum.pm
#
################################################################################
#
#    This file includes parts which are copyrighted by Adaptive Computing
#    Enterprises, Inc., and parts which are copyrighted by Battelle Memorial
#    Institute.
#
#    The terms and conditions for the use and redistribution of these parts
#    are governed by the Moab Accounting Manager License and the BSD License
#    respectively. Refer to the LICENSE file for details.
#
#    Copyright (C) 2006 - 2016 by Adaptive Computing Enterprises, Inc.
#    All Rights Reserved.
#
#    The Moab Accounting Manager License specifies the terms and conditions
#    for use and redistribution.
#
#    The Moab Accounting Manager License applies to the Moab Accounting
#    Manager software offered by Adaptive Computing Enterprises, Inc.
#    By installing or using this software, Licensee accepts a non-exclusive
#    license from Adaptive Computing Enterprises, Inc. and is bound to accept
#    acknowledgement of and abide by the notices and conditions of the
#    Moab Accounting Manager License.
#
#    Copyright (C) 2003 - 2005 Pacific Northwest National Laboratory,
#    Battelle Memorial Institute. All rights reserved.
#
#    The BSD License specifies the terms and conditions for use and
#    redistribution.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions
#    are met:
#
#    - Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
#    - Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
#    - Neither the name of Battelle nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
#    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
#    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
#    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#    POSSIBILITY OF SUCH DAMAGE.
#
################################################################################

use strict;

=head1 NAME

MAM::Datum - represents an object within request or response data

=head1 DESCRIPTION

The B<MAM::Datum> module defines functions to build and parse Datum. A B<MAM::Request> or B<MAM::Response> may contain data which is a list of objects called datum. Normal query responses result in two-dimensional data, with the data containing one or more objects which have one or more properties with names and values. Some requests (such as UsageRecord Charge) may take a more complex data structure, however it is still considered a list of B<MAM::Datum> objects.

=head1 CONSTRUCTORS

my $datum = new MAM::Datum($name);
my $datum = new MAM::Datum($element);

=head1 ACCESSORS

=over 4

=item $element = $datum->getElement();

=item $name = $datum->getName();

=item @properties = $datum->getProperties();

=item $property = $datum->getProperty($name);

=item $value = $datum->getPropertyValue($name);

=back 

=head1 MUTATORS

=over 4

=item $datum = $datum->setElement($element);

=item $datum = $datum->setName($name);

=item $datum = $datum->setProperty($name, $value);

=item $datum = $datum->setProperty(\$property);

=item $datum = $datum->delProperty($name);

=item $datum = $datum->addChild($datum);

=back

=head1 OTHER METHODS

=over 4

=item $string = $datum->toString();

=back

=head1 EXAMPLES

use MAM::Datum;

my $datum = new MAM::Datum('User');
$datum->setProperty('Name', 'scottmo');
my $name = $datum->getPropertyValue('Name');

=head1 REQUIRES

Perl 5.8

=head1 AUTHOR

Scott Jackson, scottmo@adaptivecomputing.com

=cut

##############################################################################

package MAM::Datum;

use MAM::Datum;
use MAM::Property;
use XML::LibXML;

# ----------------------------------------------------------------------------
# Constructor
# ----------------------------------------------------------------------------

sub new
{
    my ($class, $name) = @_;
    my $element;

    if (! ref($name) && $name)
    {
        $element = new XML::LibXML::Element($name);
    }
    elsif (ref($name) eq 'XML::LibXML::Element')
    {
        $element = $name;
    }
    bless {
        _element => $element,    # Element
    }, $class;
}

# ----------------------------------------------------------------------------
# Accessors
# ----------------------------------------------------------------------------

# Get the datum element
sub getElement
{
    my ($self) = @_;
    return $self->{_element};
}

# Get the name of the datum object
sub getName
{
    my ($self) = @_;
    return $self->{_element}->nodeName();
}

# Get the properties of this datum
sub getProperties
{
    my ($self) = @_;
    my @properties = ();
    foreach my $propertyNode ($self->{_element}->childNodes())
    {
        my $name = $propertyNode->nodeName();
        my $value = $propertyNode->textContent();
        my $property = new MAM::Property(name => $name, value => $value);
        push @properties, $property;
    }
    return @properties;
}

# Get the named property
sub getProperty
{
    my ($self, $name) = @_;
    my @propertyNodes = $self->{_element}->getChildrenByTagName($name);
    if (@propertyNodes)
    {
        my $value = ($propertyNodes[0])->textContent();
        my $property = new MAM::Property(name => $name, value => $value);
        return $property;
    }
    return;
}

# Get the text value of the named property
sub getPropertyValue
{
    my ($self, $name) = @_;
    my @propertyNodes = $self->{_element}->getChildrenByTagName($name);
    if (@propertyNodes)
    {
        return ($propertyNodes[0])->textContent;
    }
    return;
}

# ----------------------------------------------------------------------------
# Mutators
# ----------------------------------------------------------------------------

# Set the datum from element
sub setElement
{
    my ($self, $element) = @_;
    $self->{_element} = $element if $element;
    return $self;
}

# Set the name of the datum
sub setName
{
    my ($self, $name) = @_;
    $self->{_element}->setNodeName($name) if $name;
    return $self;
}

# Add a datum property
sub addProperty
{
    my ($self, $name, $value) = @_;

    # By property
    if (ref($name) eq 'MAM::Property')
    {
        my $propertyName = $name->getName();
        my $propertyValue = $name->getValue();
        my %propertyAttributes = $name->getAttributes();
        my $element = new XML::LibXML::Element($propertyName);
        $element->appendText($propertyValue) if defined $propertyValue;
        foreach my $attributeName (sort keys %propertyAttributes)
        {
            my $attributeValue = $propertyAttributes{$attributeName};
            $element->setAttribute($attributeName, $attributeValue);
        }
        $self->{_element}->appendChild($element);
    }
    # By name, value
    elsif (! ref($name))
    {
        my $element = new XML::LibXML::Element($name);
        $element->appendText($value) if defined $value;
        $self->{_element}->appendChild($element);
    }
    return $self;
}

# Remove all datum properties with the specified name
sub delProperty
{
    my ($self, $name) = @_;
    # By property
    if (ref($name) eq 'MAM::Property')
    {
        $name = $name->getName();
    }
    # By name
    foreach my $child ($self->{_element}->getChildrenByTagName($name))
    {
        $self->{_element}->removeChild($child);
    }
    return $self;
}

# Set a datum property (replacing existing value)
sub setProperty
{
    my ($self, $name, $value) = @_;
    $self->delProperty($name);
    $self->addProperty($name, $value);
    return $self;
}

# Appends a new child datum
sub addChild
{
    my ($self, $child) = @_;
    $self->{_element}->appendChild($child->getElement()) if $child;
    return $self;
}

# ----------------------------------------------------------------------------
# Other Methods
# ----------------------------------------------------------------------------

# Convert option to printable string
sub toString
{
    my ($self) = @_;
    local $XML::LibXML::setTagCompression = 1;
    return $self->{_element}->toString();
}

1;
