#! /usr/bin/perl -wT
################################################################################
#
# MAM Message object
#
# File   :  Message.pm
#
################################################################################
#
#    This file includes parts which are copyrighted by Adaptive Computing
#    Enterprises, Inc., and parts which are copyrighted by Battelle Memorial
#    Institute.
#
#    The terms and conditions for the use and redistribution of these parts
#    are governed by the Moab Accounting Manager License and the BSD License
#    respectively. Refer to the LICENSE file for details.
#
#    Copyright (C) 2006 - 2016 by Adaptive Computing Enterprises, Inc.
#    All Rights Reserved.
#
#    The Moab Accounting Manager License specifies the terms and conditions
#    for use and redistribution.
#
#    The Moab Accounting Manager License applies to the Moab Accounting
#    Manager software offered by Adaptive Computing Enterprises, Inc.
#    By installing or using this software, Licensee accepts a non-exclusive
#    license from Adaptive Computing Enterprises, Inc. and is bound to accept
#    acknowledgement of and abide by the notices and conditions of the
#    Moab Accounting Manager License.
#
#    Copyright (C) 2003 - 2005 Pacific Northwest National Laboratory,
#    Battelle Memorial Institute. All rights reserved.
#
#    The BSD License specifies the terms and conditions for use and
#    redistribution.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions
#    are met:
#
#    - Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
#    - Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
#    - Neither the name of Battelle nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
#    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
#    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
#    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#    POSSIBILITY OF SUCH DAMAGE.
#
################################################################################

use strict;

=head1 NAME

MAM::Message - services MAM client messages

=head1 DESCRIPTION

The B<MAM::Message> module defines functions to service MAM messages. A message encapsulates a client transmission and is responsible for establishing the socket with the server and sending and receiving the message chunks. For packet economy, the header is sent with the first chunk and the tail with the last chunk. A message will have only one chunk.

=head1 CONSTRUCTORS

 my $message = new MAM::Message();
 my $message = new MAM::Message(connection => $connection);

=head1 METHODS

=over 4

=item $connection = $message->sendChunk($messageChunk[, $backup]);

=item $chunk = $message->receiveChunk();

=item $message->disconnect();

=item $reply = $message->getReply();

=item $string = $message->toString();

=back

=head1 EXAMPLES

use MAM::Message;
use MAM::Reply;

my $message = new MAM::Message();
$message->sendChunk($messageChunk);
my $reply = $message->getReply();
my $replyChunk = $reply->receiveChunk();

my $message = new MAM::Message(connection => $connection);
my $messageChunk = $message->receiveChunk();
my $reply = new MAM::Reply(connection => $connection);
$reply->sendChunk($replyChunk);

=head1 REQUIRES

Perl 5.8

=head1 AUTHOR

Scott Jackson, scottmo@adaptivecomputing.com

=cut

##############################################################################

package MAM::Message;

use vars qw($log);
use Error qw(:try);
use Socket;
use XML::LibXML;
use MAM::Chunk;
use MAM::Exception;
use MAM::Global;
use MAM::Reply;
use MAM::Request;

# ----------------------------------------------------------------------------
# Constructor
# ----------------------------------------------------------------------------

sub new
{
    my ($class, %arg) = @_;

    # Instantiate the object
    my $self = {
        _connection => $arg{connection},         # FILEHANDLE ref
        _chunk      => $arg{chunk},              # Chunk ref
        _persistent => $arg{persistent} || 0,    # SCALAR
        _chunkNum   => $arg{chunkNum} || 0,      # SCALAR
        _chunkMax   => $arg{chunkMax} || 0,      # SCALAR
    };
    bless $self, $class;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (',
            join(', ', map {"$_ => $arg{$_}"} keys %arg), ')');
    }

    return $self;
}

# ----------------------------------------------------------------------------
# $string = toString();
# ----------------------------------------------------------------------------

# Serialize message to printable string
sub toString
{
    my ($self) = @_;

    my $string = '[';

    $string .= $self->{_connection} if defined($self->{_connection});
    $string .= ', ';

    $string .= $self->{_chunk} if defined($self->{_chunk});
    $string .= ', ';

    #$string .= $self->{_persistent};
    #$string .= ', ';

    $string .= $self->{_chunkNum};
    $string .= ', ';

    $string .= $self->{_chunkMax};

    $string .= ']';

    return $string;
}

# ----------------------------------------------------------------------------
# $connection = sendChunk($messageChunk[, $backup]);
# ----------------------------------------------------------------------------

# Send the message chunk and return the connection
sub sendChunk
{
    my ($self, $chunk, $backup) = @_;
    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $wireProtocol = $chunk->getWireProtocol();

    # Set the message chunk
    $self->{_chunk} = $chunk;

    # Build payload and obtain content length
    # This also signs the message and embeds the host and port
    my $payload       = $self->marshallChunk($chunk);
    my $contentLength = length($payload);

    # There will always be only one message chunk
    # Connect to the server and send header

    # Obtain host and port and remove tainting
    my $serverHost = defined $backup ? $chunk->getBackup() : $chunk->getHost();
    my $serverPort = $chunk->getPort();
    if ($serverHost =~ /^([-\w.]+)$/) { $serverHost = $1; }
    else
    {
        throw MAM::Exception('212',
            "Invalid characters were found in server host ($serverHost)\n");
    }
    if ($serverPort =~ /^([\d]+)$/) { $serverPort = $1; }
    else
    {
        throw MAM::Exception('214',
            "server port must be an integer number ($serverPort)\n");
    }

    # Create a socket
    socket(SERVER, PF_INET, SOCK_STREAM, getprotobyname('tcp'))
      || throw MAM::Exception('216', "Unable to create socket ($!)");

    # Build the address of the remote machine
    my $ipAddress = inet_aton($serverHost)
      || throw MAM::Exception('212',
        "Unable to resolve server hostname $serverHost ($!)");
    my $socketAddress = sockaddr_in($serverPort, $ipAddress);

    # Connect to the socket
    connect(SERVER, $socketAddress)
      || throw MAM::Exception('222', "Unable to connect to socket ($!)");

    # Generate HTTP header
    my $header;
    if ($wireProtocol eq 'SOAP')
    {
        $header =
            "POST /mam HTTP/1.1\r\n"
          . "Content-Type: application/soap+xml; charset=\"utf-8\"\r\n"
          . "Content-Length: $contentLength\r\n";
    }
    elsif ($wireProtocol eq 'SSSRMAP')
    {
        $header =
            "POST /SSSRMAP3 HTTP/1.1\r\n"
          . "Content-Type: text/xml; charset=\"utf-8\"\r\n"
          . "Transfer-Encoding: chunked\r\n";
    }
    else
    {
        throw MAM::Exception('200', "Invalid wire protocol: $wireProtocol");
    }

    # Write out header to server
    if ($log->is_debug())
    {
        $log->debug("Writing message header ($header).");
    }
    print SERVER $header, "\r\n";

    # Write out payload to server
    if ($log->is_info())
    {
        my $sanitizedPayload = $payload;
        if ($sanitizedPayload =~ /<Set name="Password">/)
        {
            $sanitizedPayload =~
              s|<Set name="Password">[^<]*</Set>|<Set name="Password">********</Set>|;
        }
        $log->info(
            "Writing message payload ($contentLength, $sanitizedPayload).");
    }
    if ($wireProtocol eq 'SSSRMAP')
    {
        my $chunkSize = sprintf('%X', $contentLength);
        print SERVER $chunkSize, "\r\n";
    }
    print SERVER $payload;

    # This will always be the last chunk
    # Send final bytes
    if ($wireProtocol eq 'SSSRMAP')
    {
        print SERVER "0\r\n";
    }
    SERVER->flush();

    # Associate the connection with this message object
    $self->{_connection} = *SERVER{IO};

    # Return a reference to the filehandle
    return $self->{_connection};
}

# ----------------------------------------------------------------------------
# $messageChunk = receiveChunk();
# ----------------------------------------------------------------------------

# Receive the message chunk over connection
sub receiveChunk
{
    my ($self) = @_;
    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $client = $self->{_connection};

    # There will always be only one message chunk
    # Connect to the server and send header

    # Read initial header line from client
    my $headerLine = <$client>;
    unless (defined $headerLine)
    {
        throw MAM::Exception('243', "Null connection received.\n");
    }

    $headerLine =~ s/\r\n//;    # Remove trailing whitespace
    my $header = $headerLine . "\n";
    my ($method, $uri, $http) = split(/\s+/, $headerLine);
    if ($method ne 'POST')
    {
        throw MAM::Exception('242',
            "Invalid Method: $method (should be POST).\n");
    }
    if ($uri !~ /\/mam|\/SSSRMAP/)
    {
        throw MAM::Exception('242', "Invalid URI: $uri.\n");
    }

    # Read remaining header lines from client
    my $transferEncoding = 'UNDEFINED';
    my $contentType      = 'UNDEFINED';
    my $contentLength    = 0;
    while (1)
    {
        $headerLine = <$client>;
        $headerLine =~ s/\r\n//;    # Remove trailing whitespace
        $header .= $headerLine . "\n";
        my ($property, $value, $rest) = split(/[\s;]+/, $headerLine);
        last unless $property;      # Break out if encounter empty line
        if ($property =~ /content-type:/i)      { $contentType      = $value; }
        if ($property =~ /transfer-encoding:/i) { $transferEncoding = $value; }
        if ($property =~ /content-length:/i)    { $contentLength    = $value; }
    }

    if ($log->is_debug())
    {
        $log->debug("Read message header ($header).");
    }

    # Check for valid Content-Type
    my $wireProtocol;
    if ($contentType eq 'application/soap+xml')
    {
        $wireProtocol = 'SOAP';
    }
    elsif ($contentType eq 'text/xml')
    {
        $wireProtocol = 'SSSRMAP';
    }
    else
    {
        throw MAM::Exception('242', "Invalid Content-Type: $contentType.\n");
    }

    # Check for valid Transfer-Encoding
    if ($wireProtocol eq 'SSSRMAP' && $transferEncoding ne 'chunked')
    {
        throw MAM::Exception('242',
            "Invalid Transfer-Encoding: $transferEncoding (should be chunked).\n"
        );
    }

    # Render contentLength from chunksize if SSSRMAP
    if ($wireProtocol eq 'SSSRMAP')
    {
        my $chunkSize = <$client>;
        $chunkSize =~ s/\r\n//;    # Remove trailing whitespace
        $contentLength = hex($chunkSize);    # Convert from hex to decimal
    }
    throw MAM::Exception('244', "Invalid payload size ($contentLength)")
      unless $contentLength;

    # Read contentLength bytes from client
    my $payload;
    read $client, $payload, $contentLength;
    if ($log->is_info())
    {
        my $sanitizedPayload = $payload;
        if ($sanitizedPayload =~ /<Set name="Password">/)
        {
            $sanitizedPayload =~
              s|<Set name="Password">[^<]*</Set>|<Set name="Password">********</Set>|;
        }
        $log->info("Read message payload ($contentLength, $sanitizedPayload).");
    }

    # Warn if specified length does not match actual length
    if (length($payload) != $contentLength)
    {
        $log->error(
            "Specified content length ($contentLength) does not match length of actual XML payload ("
              . length($payload)
              . ').');
    }

    # We pass in the appropriate wire protocol by first setting
    # our chunk reference to have the correct wire protocol
    # Then unmarshallChunk will copy this in to the chunk it returns
    # Which replaces our chunk reference
    $self->{_chunk} = new MAM::Chunk()->setWireProtocol($wireProtocol);
    my $chunk = $self->unmarshallChunk($payload);
    $self->{_chunk} = $chunk;

    # This will always be the last chunk
    # Read final bytes and disconnect
    if ($wireProtocol eq 'SSSRMAP')
    {
        my $tail = <$client>;
        $tail =~ s/\r\n//;    # Remove trailing whitespace
        if ($log->is_debug())
        {
            $log->debug("Read message tail ($tail).");
        }
    }

    return $chunk;
}

# ----------------------------------------------------------------------------
# $reply = getReply();
# ----------------------------------------------------------------------------

# Returns a reply (with same connection as invoking message)
sub getReply
{
    my ($self) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $reply = new MAM::Reply(connection => $self->{_connection});

    return $reply;
}

# ----------------------------------------------------------------------------
# disconnect();
# ----------------------------------------------------------------------------

# Disconnect
sub disconnect
{
    my ($self) = @_;
    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $server = $self->{_connection};

    print $server "0\r\n";
    #$server->flush();
    close $server;
}

# ----------------------------------------------------------------------------
# $payload = marshallChunk($chunk);
# ----------------------------------------------------------------------------

# Get the payload (marshall the chunk object into an XML string)
sub marshallChunk
{
    my ($self, $chunk) = @_;
    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $wireProtocol = $chunk->getWireProtocol();
    my $envPrefix    = '';
    my $appPrefix    = '';
    if ($wireProtocol eq 'SOAP')
    {
        $envPrefix = 'soap:';
        $appPrefix = 'mam:';
    }

    # Create Envelope and Body
    my $doc = XML::LibXML::Document->new('1.0', 'UTF-8');
    my $root = $doc->createElement("${envPrefix}Envelope");
    if ($wireProtocol eq 'SOAP')
    {
        $root->setNamespace('http://www.w3.org/2003/05/soap-envelope',
            'soap', 0);
    }
    $doc->setDocumentElement($root);

    # Create the body
    my $body = $doc->createElement("${envPrefix}Body");
    $root->appendChild($body);

    # Append request if defined
    my $request = $chunk->{_request};
    if (defined($request))
    {
        my $requestElement = $doc->createElement("${appPrefix}Request");
        if ($wireProtocol eq 'SOAP')
        {
            $requestElement->setNamespace('http://www.clusterresources.com/mam',
                'mam', 0);
        }

        # Add action to request
        my $action = $request->getAction();
        if ($action)
        {
            $requestElement->setAttribute('action', $action);
        }
        else
        {
            throw MAM::Exception('312', 'Action not specified in request');
        }

        # Add actor to request
        my $actor = $request->getActor();
        if ($actor)
        {
            $requestElement->setAttribute('actor', $actor);
        }
        else
        {
            throw MAM::Exception('312', 'Actor not specified in request');
        }

        # Add chunking to request if necessary
        if ($request->getChunking()
            || ($clientConfig->{'response.chunking'} || $RESPONSE_CHUNKING) =~
            /True/i)
        {
            $requestElement->setAttribute('chunking', 'True');
        }

        # Add chunkSize to request if necessary
        my $chunkSize =
          $request->getChunkSize() || $clientConfig->{'response.chunksize'};
        if ($chunkSize)
        {
            $requestElement->setAttribute('chunkSize', $chunkSize);
        }

        # Add objects to request
        my @objects = $request->getObjects();
        unless (@objects)
        {
            throw MAM::Exception('311', 'Object not specified in request');
        }
        foreach my $object (@objects)
        {
            my $name          = $object->getName();
            my $alias         = $object->getAlias();
            my $join          = $object->getJoin();
            my $objectElement = $doc->createElement('Object');
            $objectElement->appendText($name);
            if ($alias)
            {
                $objectElement->setAttribute('alias', $alias);
            }
            if ($join)
            {
                $objectElement->setAttribute('join', $join);
            }
            $requestElement->appendChild($objectElement);
        }

        # Add selections to request
        foreach my $selection ($request->getSelections())
        {
            my $object           = $selection->getObject();
            my $name             = $selection->getName();
            my $part             = $selection->getPart();
            my $op               = $selection->getOperator();
            my $alias            = $selection->getAlias();
            my $selectionElement = $doc->createElement('Get');
            if ($object)
            {
                $selectionElement->setAttribute('object', $object);
            }
            if ($name)
            {
                $selectionElement->setAttribute('name', $name);
            }
            if ($part)
            {
                $selectionElement->setAttribute('part', $part);
            }
            if ($op)
            {
                $selectionElement->setAttribute('op', $op);
            }
            if ($alias)
            {
                $selectionElement->setAttribute('alias', $alias);
            }
            $requestElement->appendChild($selectionElement);
        }

        # Add assignments to request
        foreach my $assignment ($request->getAssignments())
        {
            my $name              = $assignment->getName();
            my $part              = $assignment->getPart();
            my $value             = $assignment->getValue();
            my $op                = $assignment->getOperator();
            my $assignmentElement = $doc->createElement('Set');
            if ($name)
            {
                $assignmentElement->setAttribute('name', $name);
            }
            if ($part)
            {
                $assignmentElement->setAttribute('part', $part);
            }
            if ($op && $op ne 'Assign')
            {
                $assignmentElement->setAttribute('op', $op);
            }
            if (defined $value)
            {
                $assignmentElement->appendText($value);
            }
            $requestElement->appendChild($assignmentElement);
        }

        # Add conditions to request
        foreach my $condition ($request->getConditions())
        {
            my $object           = $condition->getObject();
            my $name             = $condition->getName();
            my $part             = $condition->getPart();
            my $subject          = $condition->getSubject();
            my $value            = $condition->getValue();
            my $op               = $condition->getOperator();
            my $conj             = $condition->getConjunction();
            my $group            = $condition->getGroup();
            my $conditionElement = $doc->createElement('Where');
            if ($conj)
            {
                $conditionElement->setAttribute('conj', $conj);
            }
            if ($group && $group ne '0')
            {
                $conditionElement->setAttribute('group', $group);
            }
            if ($object)
            {
                $conditionElement->setAttribute('object', $object);
            }
            if ($name)
            {
                $conditionElement->setAttribute('name', $name);
            }
            if ($part)
            {
                $conditionElement->setAttribute('part', $part);
            }
            if ($op && $op ne 'eq')
            {
                $conditionElement->setAttribute('op', $op);
            }
            if ($subject)
            {
                $conditionElement->setAttribute('subject', $subject);
            }
            if (defined $value)
            {
                $value = toMFT($value) if $name =~ /(?<!CPU)Time$/;
                $conditionElement->appendText($value);
            }
            $requestElement->appendChild($conditionElement);
        }

        # Add options to request
        foreach my $option ($request->getOptions())
        {
            my $name          = $option->getName();
            my $value         = $option->getValue();
            my $op            = $option->getOperator();
            my $optionElement = $doc->createElement('Option');
            if ($name)
            {
                $optionElement->setAttribute('name', $name);
                if ($op)
                {
                    $optionElement->setAttribute('op', $op);
                }
                if (defined $value)
                {
                    $value = toMFT($value) if $name =~ /(?<!CPU)Time$/;
                    $optionElement->appendText($value);
                }
            }
            $requestElement->appendChild($optionElement);
        }

        # Add data to external request
        my $data = $request->getDataElement();
        if ($data->hasChildNodes())
        {
            # Transform timestamps to Message Format Time
            foreach my $objNode ($data->childNodes())
            {
                foreach my $attrNode ($objNode->childNodes())
                {
                    # We don't yet have metadata for requests so
                    # we look for *Time with exception of CPUTime
                    if ($attrNode->nodeName() =~ /(?<!CPU)Time$/)
                    {
                        my $textNode = ($attrNode->childNodes)[0];
                        if ($textNode)
                        {
                            my $dateTime = $textNode->getData();
                            $dateTime = toMFT($dateTime);
                            $textNode->setData($dateTime);
                        }
                    }
                }
            }
            $requestElement->appendChild($data);
        }

        $body->appendChild($requestElement);
    }

    # Sign the chunk if required
    if ($chunk->{_authentication})
    {
        try
        {
            $chunk->sign($doc);
        }
        catch MAM::Exception with
        {
            # Rethrow exception
            throw MAM::Exception('422',
                'Failed signing message: (' . $_[0] . ').');
        };
    }
    else
    {
        try
        {
            $chunk->orient($doc);
        }
        catch MAM::Exception with
        {
            # Rethrow exception
            throw MAM::Exception('422',
                'Failed orienting message: (' . $_[0] . ').');
        };
    }

    # Encrypt the chunk if required
    if ($chunk->{_encryption})
    {
        try
        {
            $chunk->encrypt($doc);
        }
        catch MAM::Exception with
        {
            # Rethrow exception
            throw MAM::Exception('432',
                'Failed encrypting message: (' . $_[0] . ').');
        };
    }

    local $XML::LibXML::setTagCompression = 1;
    return $doc->toString();
}

# ----------------------------------------------------------------------------
# $chunk = unmarshallChunk($payload);
# ----------------------------------------------------------------------------

# Set the payload (unmarshall the XML string into a chunk object)
sub unmarshallChunk
{
    my ($self, $payload) = @_;
    if ($log->is_trace())
    {
        my $sanitizedPayload = $payload;
        if ($sanitizedPayload =~ /<Set name="Password">/)
        {
            $sanitizedPayload =~
              s|<Set name="Password">[^<]*</Set>|<Set name="Password">********</Set>|;
        }
        $log->trace("Invoked with arguments: ($sanitizedPayload)");
    }

    my $wireProtocol = $self->{_chunk}->getWireProtocol();
    my $chunk        = new MAM::Chunk()->setWireProtocol($wireProtocol);
    my $parser       = new XML::LibXML();
    my $doc          = $parser->parse_string($payload);
    my $envPrefix    = '';
    my $appPrefix    = '';
    if ($wireProtocol eq 'SOAP')
    {
        $envPrefix = 'soap:';
        $appPrefix = 'mam:';
    }

    my $root = $doc->getDocumentElement();
    # Fail if root envelope is wrong
    if ($root->nodeName() !~ /Envelope/)
    {
        throw MAM::Exception('242',
                "The message is malformed\nThe root element is "
              . $root->nodeName()
              . ' but should be Envelope.');
    }

    # Decrypt the chunk if required
    my @encryptedDataNodes = $root->getChildrenByTagName('EncryptedData');
    if ($chunk->{_encryption} || @encryptedDataNodes)
    {
        try
        {
            $chunk->decrypt($doc);
        }
        catch MAM::Exception with
        {
            # Rethrow exception
            throw MAM::Exception('434',
                'Failed decrypting message: (' . $_[0] . ').');
        };
        $chunk->setEncryption(1) unless $chunk->getEncryption();
    }

    # Authenticate the chunk if required
    my @signatureNodes = $root->getChildrenByTagName('Signature');
    if ($chunk->{_authentication} || @signatureNodes)
    {
        try
        {
            $chunk->authenticate($doc);
        }
        catch MAM::Exception with
        {
            # Rethrow exception
            throw MAM::Exception('424',
                'Failed authenticating message: (' . $_[0] . ').');
        };
        $chunk->setAuthentication(1) unless $chunk->getAuthentication();
    }

    # Check for the body
    my @bodyNodes = $root->getChildrenByTagName("${envPrefix}Body");
    if (! @bodyNodes)
    {
        throw MAM::Exception('242',
            "No body found in the message\nEnvelope should have a Body element."
        );
    }
    my $body = $bodyNodes[0];

    # Extract the requests if any
    my @requestNodes = $body->getChildrenByTagName("${appPrefix}Request");
    if (@requestNodes)
    {
        my $requestElement = $requestNodes[0];

        # Instantiate a new request
        my $request = new MAM::Request();

        # Extract action from the request
        my $action = $requestElement->getAttribute('action');
        if ($action)
        {
            $request->setAction($action);
        }
        else
        {
            throw MAM::Exception('312', 'No action specified in the request.');
        }

        # Extract actor from the request
        my $actor = $requestElement->getAttribute('actor');
        if ($actor)
        {
            $request->setActor($actor);
        }
        else
        {
            throw MAM::Exception('312', 'No actor specified in the request.');
        }

        # Extract objects from the request
        my @objects = $requestElement->getChildrenByTagName('Object');
        unless (@objects)
        {
            throw MAM::Exception('311', 'No object specified in the request.');
        }
        foreach my $messageObject (@objects)
        {
            my $name          = $messageObject->textContent();
            my $alias         = $messageObject->getAttribute('alias');
            my $join          = $messageObject->getAttribute('join');
            my $requestObject = new MAM::Object(name => $name);
            $requestObject->setAlias($alias) if $alias;
            $requestObject->setJoin($join)   if $join;
            $request->addObject($requestObject);
        }

        # Extract chunking from the request
        my $chunking = $requestElement->getAttribute('chunking');
        if (defined $chunking && $chunking =~ /True/i)
        {
            $request->setChunking(1);
        }
        else
        {
            $request->setChunking(0);
        }

        # Extract chunkSize from the request
        my $clientChunkSize = $requestElement->getAttribute('chunkSize') || 0;
        my $serverChunkSize = $serverConfig->{'response.chunksize'}      || 0;
        my $chunkSize       = min(
            $clientChunkSize > 0 ? $clientChunkSize : 1000000000,
            $serverChunkSize > 0 ? $serverChunkSize : 1000000000
        );
        if ($chunkSize)
        {
            $request->setChunkSize($chunkSize);
        }

        # Extract selections if present
        foreach
          my $messageSelection ($requestElement->getChildrenByTagName('Get'))
        {
            my $object           = $messageSelection->getAttribute('object');
            my $name             = $messageSelection->getAttribute('name');
            my $part             = $messageSelection->getAttribute('part');
            my $op               = $messageSelection->getAttribute('op');
            my $alias            = $messageSelection->getAttribute('alias');
            my $requestSelection = new MAM::Selection(name => $name);
            $requestSelection->setObject($object) if $object;
            $requestSelection->setPart($part)     if $part;
            $requestSelection->setOperator($op)   if $op;
            $requestSelection->setAlias($alias)   if $alias;
            $request->addSelection($requestSelection);
        }

        # Extract assignments if present
        foreach
          my $messageAssignment ($requestElement->getChildrenByTagName('Set'))
        {
            my $name  = $messageAssignment->getAttribute('name');
            my $part  = $messageAssignment->getAttribute('part');
            my $op    = $messageAssignment->getAttribute('op');
            my $value = $messageAssignment->textContent();
            my $requestAssignment =
              new MAM::Assignment(name => $name, value => $value);
            $requestAssignment->setPart($part) if $part;
            $requestAssignment->setOperator($op) if $op;
            $request->addAssignment($requestAssignment);
        }

        # Extract conditions if present
        foreach
          my $messageCondition ($requestElement->getChildrenByTagName('Where'))
        {
            my $object  = $messageCondition->getAttribute('object');
            my $name    = $messageCondition->getAttribute('name');
            my $subject = $messageCondition->getAttribute('subject');
            my $part    = $messageCondition->getAttribute('part');
            my $op      = $messageCondition->getAttribute('op');
            my $value   = $messageCondition->textContent();
            my $conj    = $messageCondition->getAttribute('conj');
            my $group   = $messageCondition->getAttribute('group');
            my $requestCondition =
              new MAM::Condition(name => $name, value => $value);
            $requestCondition->setObject($object)    if $object;
            $requestCondition->setSubject($subject)  if $subject;
            $requestCondition->setPart($part)        if $part;
            $requestCondition->setOperator($op)      if $op;
            $requestCondition->setConjunction($conj) if $conj;
            $requestCondition->setGroup($group)      if $group;
            $request->addCondition($requestCondition);
        }

        # Extract options if present
        foreach
          my $messageOption ($requestElement->getChildrenByTagName('Option'))
        {
            my $name          = $messageOption->getAttribute('name');
            my $op            = $messageOption->getAttribute('op');
            my $value         = $messageOption->textContent();
            my $requestOption = new MAM::Option(name => $name, value => $value);
            $requestOption->setOperator($op) if $op;
            $request->addOption($requestOption);
        }

        # Extract data to the request
        my @dataNodes = $requestElement->getChildrenByTagName('Data');
        if (@dataNodes)
        {
            $request->setDataElement($dataNodes[0]);
        }

        # Link the chunk to the request
        $chunk->{_request} = $request;

        if ($log->is_debug())
        {
            my $sanitizedRequest = $request->toString();
            if ($sanitizedRequest =~ /\(Password, [^,]+,/)
            {
                $sanitizedRequest =~
                  s|\(Password, [^,]+,|\(Password, ********,|;
            }
            $log->debug("Extracted the request ($sanitizedRequest).");
        }
    }

    return $chunk;
}

1;
