#! /usr/bin/perl -wT
################################################################################
#
# MAM Object object
#
# File   :  Object.pm
#
################################################################################
#
#    This file includes parts which are copyrighted by Adaptive Computing
#    Enterprises, Inc., and parts which are copyrighted by Battelle Memorial
#    Institute.
#
#    The terms and conditions for the use and redistribution of these parts
#    are governed by the Moab Accounting Manager License and the BSD License
#    respectively. Refer to the LICENSE file for details.
#
#    Copyright (C) 2006 - 2016 by Adaptive Computing Enterprises, Inc.
#    All Rights Reserved.
#
#    The Moab Accounting Manager License specifies the terms and conditions
#    for use and redistribution.
#
#    The Moab Accounting Manager License applies to the Moab Accounting
#    Manager software offered by Adaptive Computing Enterprises, Inc.
#    By installing or using this software, Licensee accepts a non-exclusive
#    license from Adaptive Computing Enterprises, Inc. and is bound to accept
#    acknowledgement of and abide by the notices and conditions of the
#    Moab Accounting Manager License.
#
#    Copyright (C) 2003 - 2005 Pacific Northwest National Laboratory,
#    Battelle Memorial Institute. All rights reserved.
#
#    The BSD License specifies the terms and conditions for use and
#    redistribution.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions
#    are met:
#
#    - Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
#    - Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
#    - Neither the name of Battelle nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
#    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
#    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
#    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#    POSSIBILITY OF SUCH DAMAGE.
#
################################################################################

use strict;

=head1 NAME

MAM::Object - represents an object

=head1 DESCRIPTION

The B<MAM::Object> module defines functions to build and parse Objects

=head1 CONSTRUCTOR

my $object = new MAM::Object(name => $name, alias => $alias, join => $join);

=head1 ACCESSORS

=over 4

=item $name = $object->getName();

=item $alias = $object->getAlias();

=item $join = $object->getJoin();

=back 

=head1 MUTATORS

=over 4

=item $object->setName($name);

=item $object->setAlias($join);

=item $object->setJoin($join);

=back

=head1 OTHER METHODS

=over 4

=item $string = $object->toString();

=back

=head1 EXAMPLES

use MAM::Object;

my $object1 = new MAM::Object(name => 'Account');

my $object2 = new MAM::Object(name => 'AccountUser', alias => 'Member', join => 'LeftOuter');

=head1 REQUIRES

Perl 5.8

=head1 AUTHOR

Scott Jackson, scottmo@adaptivecomputing.com

=cut

##############################################################################

package MAM::Object;

# ----------------------------------------------------------------------------
# Constructor
# ----------------------------------------------------------------------------

sub new
{
    my ($class, %arg) = @_;
    bless {
        _name  => $arg{name}  || '',    # SCALAR
        _alias => $arg{alias} || '',    # SCALAR
        _join  => $arg{join}  || '',    # SCALAR
    }, $class;
}

# ----------------------------------------------------------------------------
# Accessors
# ----------------------------------------------------------------------------

# Get the object name
sub getName
{
    my ($self) = @_;
    return $self->{_name};
}

# Get the object alias
sub getAlias
{
    my ($self) = @_;
    return $self->{_alias};
}

# Get the object join
sub getJoin
{
    my ($self) = @_;
    return $self->{_join};
}

# ----------------------------------------------------------------------------
# Mutators
# ----------------------------------------------------------------------------

# Set the object name
sub setName
{
    my ($self, $name) = @_;
    $self->{_name} = $name if defined $name;
    return $self;
}

# Set the object alias
sub setAlias
{
    my ($self, $alias) = @_;
    $self->{_alias} = $alias if defined $alias;
    return $self;
}

# Set the object join
sub setJoin
{
    my ($self, $join) = @_;
    $self->{_join} = $join if defined $join;
    return $self;
}

# ----------------------------------------------------------------------------
# Other Methods
# ----------------------------------------------------------------------------

# Convert object to printable string
sub toString
{
    my ($self) = @_;
    my $string = '<Object';
    $string .= " alias=\"$self->{_alias}\"" if $self->{_alias} ne '';
    $string .= '>';
    $string .= $self->{_name} if $self->{_name} ne '';
    $string .= '</Object>';
    return $string;
}

1;
