#! /usr/bin/perl -wT
################################################################################
#
# MAM Property object
#
# File   :  Property.pm
#
################################################################################
#
#    This file includes parts which are copyrighted by Adaptive Computing
#    Enterprises, Inc., and parts which are copyrighted by Battelle Memorial
#    Institute.
#
#    The terms and conditions for the use and redistribution of these parts
#    are governed by the Moab Accounting Manager License and the BSD License
#    respectively. Refer to the LICENSE file for details.
#
#    Copyright (C) 2006 - 2016 by Adaptive Computing Enterprises, Inc.
#    All Rights Reserved.
#
#    The Moab Accounting Manager License specifies the terms and conditions
#    for use and redistribution.
#
#    The Moab Accounting Manager License applies to the Moab Accounting
#    Manager software offered by Adaptive Computing Enterprises, Inc.
#    By installing or using this software, Licensee accepts a non-exclusive
#    license from Adaptive Computing Enterprises, Inc. and is bound to accept
#    acknowledgement of and abide by the notices and conditions of the
#    Moab Accounting Manager License.
#
#    Copyright (C) 2003 - 2005 Pacific Northwest National Laboratory,
#    Battelle Memorial Institute. All rights reserved.
#
#    The BSD License specifies the terms and conditions for use and
#    redistribution.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions
#    are met:
#
#    - Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
#    - Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
#    - Neither the name of Battelle nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
#    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
#    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
#    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#    POSSIBILITY OF SUCH DAMAGE.
#
################################################################################

use strict;

=head1 NAME

MAM::Property - represents a property of a datum

=head1 DESCRIPTION

The B<MAM::Property> module defines functions to build and parse Datum Properties

=head1 CONSTRUCTOR

my $property = new MAM::Property(name => $name[, value => $value][, attributes => \%attributes]);

=head1 ACCESSORS

=over 4

=item $name = $property->getName();

=item $value = $property->getValue();

=item %attributes = $property->getAttributes();

=item $attributeValue = $property->getAttributeValue($attributeName);

=back 

=head1 MUTATORS

=over 4

=item $property = $property->setName($name);

=item $property = $property->setValue($value);

=item $property = $property->setAttributes(\%attributes);

=item $property = $property->setAttribute($attributeName, $attributeValue);

=item $property = $property->delAttribute($attributeName);

=back

=head1 OTHER METHODS

=over 4

=item $string = $property->toString();

=back

=head1 EXAMPLES

use MAM::Property;

my $property = new MAM::Property(name => 'IdleProcessorSeconds', value => '10', attributes=> { 'increment' => '5' });

=head1 REQUIRES

Perl 5.8

=head1 AUTHOR

Scott Jackson, scottmo@adaptivecomputing.com

=cut

##############################################################################

package MAM::Property;

# ----------------------------------------------------------------------------
# Constructor
# ----------------------------------------------------------------------------

sub new
{
    my ($class, %arg) = @_;
    bless {
        _name => $arg{name} || '',    # SCALAR
        _value => defined $arg{value} ? $arg{value} : '',    # SCALAR
        _attributes => $arg{attributes} || {},               # HASH REF
    }, $class;
}

# ----------------------------------------------------------------------------
# Accessors
# ----------------------------------------------------------------------------

# Get the property name
sub getName
{
    my ($self) = @_;
    return $self->{_name};
}

# Get the property value
sub getValue
{
    my ($self) = @_;
    return $self->{_value};
}

# Get the property attributes
sub getAttributes
{
    my ($self) = @_;
    return %{$self->{_attributes}};
}

# Get the property's attribute value
sub getAttributeValue
{
    my ($self, $name) = @_;
    if (exists ${$self->{_attributes}}{$name})
    {
        return ${$self->{_attributes}}{$name};
    }
    return;
}

# ----------------------------------------------------------------------------
# Mutators
# ----------------------------------------------------------------------------

# Set the property name
sub setName
{
    my ($self, $name) = @_;
    $self->{_name} = $name if defined $name;
    return $self;
}

# Set the property value
sub setValue
{
    my ($self, $value) = @_;
    $self->{_value} = $value if defined $value;
    return $self;
}

# Set the property attributes
sub setAttributes
{
    my ($self, $attributes) = @_;
    $self->{_attributes} = $attributes;
    return $self;
}

# Set the property attribute
sub setAttribute
{
    my ($self, $attributeName, $attributeValue) = @_;
    ${$self->{_attributes}}{$attributeName} = $attributeValue;
    return $self;
}

# Remove the property attribute
sub delAttribute
{
    my ($self, $attributeName) = @_;
    delete ${$self->{_attributes}}{$attributeName};
    return $self;
}

# ----------------------------------------------------------------------------
# Other Methods
# ----------------------------------------------------------------------------

# Convert property to printable string
sub toString
{
    my ($self) = @_;
    my $string = "<$self->{_name}";
    foreach my $attributeName (sort keys %{$self->{_attributes}})
    {
        $string .= " $attributeName=\"${$self->{_attributes}}{$attributeName}\"";
    }
    $string .= '>';
    $string .= $self->{_value} if $self->{_value} ne '';
    $string .= "</$self->{_name}>";
    return $string;
}

1;
