#! /usr/bin/perl -wT
################################################################################
#
# MAM Request object
#
# File   :  Request.pm
#
################################################################################
#
#    This file includes parts which are copyrighted by Adaptive Computing
#    Enterprises, Inc., and parts which are copyrighted by Battelle Memorial
#    Institute.
#
#    The terms and conditions for the use and redistribution of these parts
#    are governed by the Moab Accounting Manager License and the BSD License
#    respectively. Refer to the LICENSE file for details.
#
#    Copyright (C) 2006 - 2016 by Adaptive Computing Enterprises, Inc.
#    All Rights Reserved.
#
#    The Moab Accounting Manager License specifies the terms and conditions
#    for use and redistribution.
#
#    The Moab Accounting Manager License applies to the Moab Accounting
#    Manager software offered by Adaptive Computing Enterprises, Inc.
#    By installing or using this software, Licensee accepts a non-exclusive
#    license from Adaptive Computing Enterprises, Inc. and is bound to accept
#    acknowledgement of and abide by the notices and conditions of the
#    Moab Accounting Manager License.
#
#    Copyright (C) 2003 - 2005 Pacific Northwest National Laboratory,
#    Battelle Memorial Institute. All rights reserved.
#
#    The BSD License specifies the terms and conditions for use and
#    redistribution.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions
#    are met:
#
#    - Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
#    - Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
#    - Neither the name of Battelle nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
#    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
#    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
#    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#    POSSIBILITY OF SUCH DAMAGE.
#
################################################################################

use strict;

=head1 NAME

MAM::Request - build and parse a MAM request

=head1 DESCRIPTION

The B<MAM::Request> module defines functions to build and parse MAM requests

=head1 CONSTRUCTOR

 my $request = new MAM::Request(object => $name || objects => \@objects, action => $action, actor => $actor, objects => \@objects, selections => \@selections, assignments => \@assignments, conditions => \@conditions, options => \@options, data => \@data, chunking => $boolean, tokenType => $tokenType, tokenName => $tokenName, tokenValue => $tokenValue, site => $site, database => $database);

=head1 ACCESSORS

=over 4

=item $action = $request->getAction();

=item $actor = $request->getActor();

=item @objects = $request->getObjects();

=item $name = $request->getObject();

=item $object = $request->getObject($name);

=item @selections = $request->getSelections();

=item $selection = $request->getSelection($name);

=item @assignments = $request->getAssignments();

=item $assignment = $request->getAssignment($name);

=item $value = $request->getAssignmentValue($name);

=item @conditions = $request->getConditions();

=item $condition = $request->getCondition($name);

=item $value = $request->getConditionValue($name);

=item @options = $request->getOptions();

=item $option = $request->getOption($name);

=item $value = $request->getOptionValue($name);

=item @data = $request->getData();

=item $element = $request->getDataElement();

=item $value = $request->getDatumValue($name);

=item $boolean = $request->getChunking();

=item $size = $request->getChunkSize();

=item $boolean = $request->getOverride();

=item $tokenType = $request->getTokenType();

=item $tokenName = $request->getTokenName();

=item $tokenValue = $request->getTokenValue();

=item $site = $request->getSite();

=item $database = $request->getDatabase();

=back

=head1 MUTATORS

=over 4

=item $request = $request->setAction($action);

=item $request = $request->setActor($actor);

=item $request = $request->setObjects(\@objects);

=item $request = $request->addObject(\$object);

=item $request = $request->addObject($name[, $join]);

=item $request = $request->setSelections(\@selections);

=item $request = $request->addSelection(\$selections);

=item $request = $request->addSelection($name[, $op[, $object[, $alias[, $part]]]]);

=item $request = $request->setAssignments(\@assignments);

=item $request = $request->addAssignment(\$assignment);

=item $request = $request->addAssignment($name[, $value[, $op[, $part]]]);

=item $request = $request->setConditions(\@conditions);

=item $request = $request->addCondition(\$condition);

=item $request = $request->addCondition($name[, $value[, $op[, $conj[, $group[, $object[, $subject[, $part]]]]]]]);

=item $request = $request->addOption($name[, $value[, $op]]);

=item $request = $request->delOption($name);

=item $request = $request->setOption(\$option);

=item $request = $request->setOption($name[, $value[, $op]]);

=item $request = $request->setOptions(\@options);

=item $request = $request->addDatum($datum);

=item $request = $request->setData(\@data);

=item $request = $request->setDataElement($element);

=item $request = $request->setChunking($boolean);

=item $request = $request->setChunkSize($size);

=item $request = $request->setTokenType($tokenType);

=item $request = $request->setTokenName($tokenName);

=item $request = $request->setTokenValue($tokenValue);

=item $request = $request->setSite($site);

=item $request = $request->setOverride($boolean);

=item $request = $request->setDatabase($database);

=back

=head1 OTHER METHODS

=over 4

=item $string = $request->toString();

=item $response = $request->getResponse();

=back

=head1 EXAMPLES

use MAM::Request;

my $request = new MAM::Request(object => 'User', action => 'Query');
$request->addSelection('EmailAddress');
$request->addCondition('Name', 'scottmo'); 
my $response = $request->getResponse(); 

=head1 REQUIRES

Perl 5.8

=head1 AUTHOR

Scott Jackson, scottmo@adaptivecomputing.com

=cut

##############################################################################

package MAM::Request;

use vars qw($log);
use Error qw(:try);
use MAM::Selection;
use MAM::Assignment;
use MAM::Chunk;
use MAM::Condition;
use MAM::Exception;
use MAM::Object;
use MAM::Option;
use MAM::Response;
use MAM::Global;
use XML::LibXML;

# ----------------------------------------------------------------------------
# Constructor
# ----------------------------------------------------------------------------

sub new
{
    my ($class, %arg) = @_;

    my $self = {
        _action => $arg{action} || '',                   # SCALAR
        _actor  => $arg{actor}  || (getpwuid($<))[0],    # SCALAR
        _objects => $arg{object} ? [new MAM::Object(name => $arg{object})]
        : $arg{objects} ? $arg{objects}
        : [],                                            # ARRAY REF of Objects
        _selections  => $arg{selections}  || [],    # ARRAY REF of Selections
        _assignments => $arg{assignments} || [],    # ARRAY REF of Assignments
        _conditions  => $arg{conditions}  || [],    # ARRAY REF of Conditions
        _options     => $arg{options}     || [],    # ARRAY REF of Options
        _data        => $arg{data}        || [],    # ARRAY REF of Datums
        _chunking    => $arg{chunking}
          || ($clientConfig->{'response.chunking'} || $RESPONSE_CHUNKING) =~
          /True/i
        ? 1
        : 0,                                        # SCALAR
        _chunkSize  => $arg{chunkSize}  || 0,       # SCALAR
        _tokenType  => $arg{tokenType}  || '',      # SCALAR
        _tokenName  => $arg{tokenName}  || '',      # SCALAR
        _tokenValue => $arg{tokenValue} || '',      # SCALAR
        _site       => $arg{site}       || '',      # SCALAR
        _override   => $arg{override}   || 0,       # SCALAR
        _database => $arg{database},                # SCALAR
    };
    bless $self, $class;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (',
            join(', ', map {"$_ => $arg{$_}"} keys %arg), ')');
    }

    return $self;
}

# ----------------------------------------------------------------------------
# Accessors
# ----------------------------------------------------------------------------

# Get the request action
sub getAction
{
    my ($self) = @_;
    return $self->{_action};
}

# Get the request actor
sub getActor
{
    my ($self) = @_;
    return $self->{_actor};
}

# Get the object list
sub getObjects
{
    my ($self) = @_;
    return @{$self->{_objects}};
}

# Get an object by name or the first object name
sub getObject
{
    my ($self, $name) = @_;

    # Get an object object by name
    if (defined $name)
    {
        foreach my $object (@{$self->{_objects}})
        {
            if ($object->getName() eq $name)
            {
                return $object;
            }
        }
        return;
    }

    # Get the first object name
    else
    {
        if (@{$self->{_objects}})
        {
            return ${$self->{_objects}}[0]->getName();
        }
        return;
    }
}

# Get the selections list
sub getSelections
{
    my ($self) = @_;
    return @{$self->{_selections}};
}

# Get a selection by name
sub getSelection
{
    my ($self, $name) = @_;
    foreach my $selection (@{$self->{_selections}})
    {
        if ($selection->getName() eq $name)
        {
            return $selection;
        }
    }
    return;
}

# Get the assignments list
sub getAssignments
{
    my ($self) = @_;
    return @{$self->{_assignments}};
}

# Get an assignment by name
sub getAssignment
{
    my ($self, $name) = @_;
    foreach my $assignment (@{$self->{_assignments}})
    {
        if ($assignment->getName() eq $name)
        {
            return $assignment;
        }
    }
    return;
}

# Get an assignment value by name
sub getAssignmentValue
{
    my ($self, $name) = @_;
    foreach my $assignment (@{$self->{_assignments}})
    {
        if ($assignment->getName() eq $name)
        {
            return $assignment->getValue();
        }
    }
    return;
}

# Get the conditions list
sub getConditions
{
    my ($self) = @_;
    return @{$self->{_conditions}};
}

# Get a condition by name
sub getCondition
{
    my ($self, $name) = @_;
    foreach my $condition (@{$self->{_conditions}})
    {
        if ($condition->getName() eq $name)
        {
            return $condition;
        }
    }
    return;
}

# Get a condition value by name
sub getConditionValue
{
    my ($self, $name) = @_;
    foreach my $condition (@{$self->{_conditions}})
    {
        if ($condition->getName() eq $name)
        {
            return $condition->getValue();
        }
    }
    return;
}

# Get the options list
sub getOptions
{
    my ($self) = @_;
    return @{$self->{_options}};
}

# Get an option by name
sub getOption
{
    my ($self, $name) = @_;
    foreach my $option (@{$self->{_options}})
    {
        if ($option->getName() eq $name)
        {
            return $option;
        }
    }
    return;
}

# Get an option value by name
sub getOptionValue
{
    my ($self, $name) = @_;
    foreach my $option (@{$self->{_options}})
    {
        if ($option->getName() eq $name)
        {
            return $option->getValue();
        }
    }
    return;
}

# Get the request data (list of data)
sub getData
{
    my ($self) = @_;
    return @{$self->{_data}};
}

# Get the request data element
sub getDataElement
{
    my ($self) = @_;
    my $data = new XML::LibXML::Element('Data');
    foreach my $datum (@{$self->{_data}})
    {
        $data->appendChild($datum->getElement());
    }
    return $data;
}

# Get the value of the named property in the first datum
sub getDatumValue
{
    my ($self, $name) = @_;
    my $value;
    if (@{$self->{_data}})
    {
        my $datum      = (${$self->{_data}}[0])->getElement();
        my @properties = $datum->getChildrenByTagName($name);
        if ($properties[0])
        {
            $value = ($properties[0])->textContent();
        }
    }
    return $value;
}

# Get the chunking flag
sub getChunking
{
    my ($self) = @_;
    return $self->{_chunking};
}

# Get the chunk size
sub getChunkSize
{
    my ($self) = @_;
    return $self->{_chunkSize};
}

# Get the token type
sub getTokenType
{
    my ($self) = @_;
    return $self->{_tokenType};
}

# Get the token name
sub getTokenName
{
    my ($self) = @_;
    return $self->{_tokenName};
}

# Get the token value
sub getTokenValue
{
    my ($self) = @_;
    return $self->{_tokenValue};
}

# Get the site
sub getSite
{
    my ($self) = @_;
    return $self->{_site};
}

# Get whether authorization is overridden
sub getOverride
{
    my ($self) = @_;
    return $self->{_override};
}

# Get associated database object
sub getDatabase
{
    my ($self) = @_;
    return $self->{_database};
}

# ----------------------------------------------------------------------------
# Mutators
# ----------------------------------------------------------------------------

# Set the request action
sub setAction
{
    my ($self, $action) = @_;
    $self->{_action} = $action if $action;
    return $self;
}

# Set the request actor
sub setActor
{
    my ($self, $actor) = @_;
    $self->{_actor} = $actor if $actor;
    return $self;
}

# Set the objects from a list
sub setObjects
{
    my ($self, $objects) = @_;
    $self->{_objects} = $objects;
    return $self;
}

# Add a request object
sub addObject
{
    my ($self, $name, $join, $alias) = @_;
    # By object
    if (ref($name) eq 'MAM::Object')
    {
        push(@{$self->{_objects}}, $name);
    }
    # By name [,join [,alias]]
    elsif (! ref($name))
    {
        push(
            @{$self->{_objects}},
            new MAM::Object(name => $name, join => $join, alias => $alias)
        );
    }
    return $self;
}

# Set the selections from a list
sub setSelections
{
    my ($self, $selections) = @_;
    $self->{_selections} = $selections;
    return $self;
}

# Add a selection
sub addSelection
{
    my ($self, $name, $op, $object, $alias, $part) = @_;
    # By selection
    if (ref($name) eq 'MAM::Selection')
    {
        push(@{$self->{_selections}}, $name);
    }
    # By name [, op[, object[, alias[, part]]]]
    elsif (! ref($name))
    {
        push(
            @{$self->{_selections}},
            new MAM::Selection(
                name   => $name,
                part   => $part,
                op     => $op,
                object => $object,
                alias  => $alias
            )
        );
    }
    return $self;
}

# Set the assignments from a list
sub setAssignments
{
    my ($self, $assignments) = @_;
    $self->{_assignments} = $assignments;
    return $self;
}

# Add an assignment
sub addAssignment
{
    my ($self, $name, $value, $op, $part) = @_;
    # By assignment
    if (ref($name) eq 'MAM::Assignment')
    {
        push(@{$self->{_assignments}}, $name);
    }
    # By name [, value[, op[, part]]]
    elsif (! ref($name) && defined($value))
    {
        push(
            @{$self->{_assignments}},
            new MAM::Assignment(
                name  => $name,
                part  => $part,
                value => $value,
                op    => $op
            )
        );
    }
    return $self;
}

# Set the conditions from a list
sub setConditions
{
    my ($self, $conditions) = @_;
    $self->{_conditions} = $conditions;
    return $self;
}

# Add a condition
sub addCondition
{
    my ($self, $name, $value, $op, $conj, $group, $object, $subject, $part) =
      @_;
    # By condition
    if (ref($name) eq 'MAM::Condition')
    {
        push(@{$self->{_conditions}}, $name);
    }
    # By name [,value[, op[, conj[, group[, object[, subject[, part]]]]]]]
    elsif (! ref($name) && defined($value))
    {
        push(
            @{$self->{_conditions}},
            new MAM::Condition(
                name    => $name,
                part    => $part,
                value   => $value,
                op      => $op,
                conj    => $conj,
                group   => $group,
                object  => $object,
                subject => $subject
            )
        );
    }
    return $self;
}

# Add an option
sub addOption
{
    my ($self, $name, $value, $op) = @_;

    # By option
    if (ref($name) eq 'MAM::Option')
    {
        push(@{$self->{_options}}, $name);
    }
    # By name [,value[, op]]
    elsif (! ref($name) && defined($value))
    {
        push(
            @{$self->{_options}},
            new MAM::Option(name => $name, value => $value, op => $op)
        );
    }
    return $self;
}

# Remove an option
sub delOption
{
    my ($self, $name) = @_;
    if (ref($name) eq 'MAM::Option')
    {
        $name = $name->getName();
    }
    @{$self->{_options}} = grep { $_->getName() ne $name } @{$self->{_options}};
    return $self;
}

# Set an option (replacing existing values)
sub setOption
{
    my ($self, $name, $value, $op) = @_;
    $self->delOption($name);
    $self->addOption($name, $value, $op);
    return $self;
}

# Set the options from a list
sub setOptions
{
    my ($self, $options) = @_;
    $self->{_options} = $options;
    return $self;
}

# Set the request data (list of data)
sub setData
{
    my ($self, $data) = @_;
    $self->{_data} = $data;
    return $self;
}

# Set the request data (from element)
sub setDataElement
{
    my ($self, $data) = @_;
    foreach my $datum ($data->childNodes())
    {
        push(@{$self->{_data}}, new MAM::Datum($datum));
    }
    return $self;
}

# Add a request datum
sub addDatum
{
    my ($self, $datum) = @_;
    push(@{$self->{_data}}, $datum);
    return $self;
}

# Set the chunking flag
sub setChunking
{
    my ($self, $chunking) = @_;
    $self->{_chunking} = $chunking;
    return $self;
}

# Set the chunk size
sub setChunkSize
{
    my ($self, $chunkSize) = @_;
    $self->{_chunkSize} = $chunkSize;
    return $self;
}

# Set the token type
sub setTokenType
{
    my ($self, $tokenType) = @_;
    $self->{_tokenType} = $tokenType;
    return $self;
}

# Set the token name
sub setTokenName
{
    my ($self, $tokenName) = @_;
    $self->{_tokenName} = $tokenName;
    return $self;
}

# Set the token value
sub setTokenValue
{
    my ($self, $tokenValue) = @_;
    $self->{_tokenValue} = $tokenValue;
    return $self;
}

# Set the site
sub setSite
{
    my ($self, $site) = @_;
    $self->{_site} = $site;
    return $self;
}

# Set whether authorization is overridden
sub setOverride
{
    my ($self, $override) = @_;
    $self->{_override} = $override;
    return $self;
}

# Set database object
sub setDatabase
{
    my ($self, $database) = @_;
    $self->{_database} = $database;
    return $self;
}

# ----------------------------------------------------------------------------
# Other Methods
# ----------------------------------------------------------------------------

# Convert request to printable string
sub toString
{
    my ($self) = @_;

    my $string = '(';
    $string .= $self->{_action};
    $string .= ', ';
    $string .= $self->{_actor};
    $string .= ', ';
    $string .=
      '[' . join(', ', map($_->toString(), @{$self->{_objects}})) . ']';
    $string .= ', ';
    $string .=
      '[' . join(', ', map($_->toString(), @{$self->{_selections}})) . ']';
    $string .= ', ';
    $string .=
      '[' . join(', ', map($_->toString(), @{$self->{_assignments}})) . ']';
    $string .= ', ';
    $string .=
      '[' . join(', ', map($_->toString(), @{$self->{_conditions}})) . ']';
    $string .= ', ';
    $string .=
      '[' . join(', ', map($_->toString(), @{$self->{_options}})) . ']';
    $string .= ', ';
    $string .= '[' . join(', ', map($_->toString(), @{$self->{_data}})) . ']';
    $string .= ', ';
    $string .= $self->{_tokenType};
    $string .= ', ';
    $string .= $self->{_tokenName};
    $string .= ', ';
    $string .= $self->{_tokenValue} ? 'TOKEN_VALUE' : '';
    $string .= ', ';
    $string .= $self->{_site};
    $string .= ', ';
    $string .= $self->{_override};
    $string .= ', ';
    $string .= $self->{_chunking};
    $string .= ', ';
    $string .= $self->{_chunkSize};
    $string .= ', ';
    $string .= $self->{_database} if $self->{_database};
    $string .= ')';

    return $string;
}

# Deserialize from a commmand string
sub deserialize
{
    my ($self, $command) = @_;
    my ($action, $pattern);

    # Parse out the object
    if ($command =~ s/^\s*([\w,=]+)\s*//) { $pattern = $1; }
    else
    {
        throw MAM::Exception('311', 'An object must be specified.');
    }
    if ($pattern =~ /^([A-Z][\w,=]+)$/)
    {
        foreach my $obj (split /,/, $1)
        {
            $obj =~ /^(\w+)(=(\w+))?/;
            my ($name, $alias) = ($1, $3);
            my $object = new MAM::Object(name => $name);
            if ($alias)
            {
                $object->setAlias($alias);
            }
            if ($log->is_trace())
            {
                $log->trace(
                    'Adding object to request (' . $object->toString() . ').');
            }
            $self->addObject($object);
        }
    }
    else
    {
        throw MAM::Exception('311', "The object is malformed ($pattern).");
    }

    # Add the action to the request
    if ($command =~ s/^\s*((\w*::)?\w+)\s*//) { $pattern = $1; }
    else
    {
        throw MAM::Exception('312', 'An action must be specified.');
    }
    if ($pattern =~ /^((\w*::)?[A-Z]\w+)$/)
    {
        $action = $1;
        if ($log->is_trace())
        {
            $log->trace("Adding action to request ($action).");
        }
        $self->setAction($action);
    }
    else
    {
        throw MAM::Exception('312', "The action is malformed ($pattern).");
    }

    # Specify that we want chunking
    $self->setChunking(1);

    # Process the remaining predicates
    while (1)
    {
        my $quoted = 0;
        my $conj   = '';
        my $group  = 0;
        my $name   = '';
        my $part   = '';
        my $op     = '';
        my $value  = '';

        # Bail if we are at the end
        if ($command =~ s/^\s*$//) { last; }

        # Peel off any conjunction
        if ($command =~ s/^\s*(([\|&!]+\s*)+)\s*//)
        {
            $pattern = $1;
            if ($pattern =~ /^((&&)|(\|\|))?\s*(!)?\s*$/)
            {
                if    ($2 && ! $4) { $conj = 'And'; }
                elsif ($3 && ! $4) { $conj = 'Or'; }
                elsif ($2 && $4)   { $conj = 'AndNot'; }
                elsif ($3 && $4)   { $conj = 'OrNot'; }
                else               { $conj = 'AndNot'; }
                if ($log->is_trace())
                {
                    $log->trace("Located conjunction ($conj).");
                }
            }
            else
            {
                throw MAM::Exception('310',
                    "The conjunction is malformed ($pattern).");
            }
        }

        # Peel off any open grouping
        if ($command =~ s/^\s*(([\(\)]+\s*)+)\s*//)
        {
            $pattern = $1;
            if ($pattern =~ /^((\(\s*)+)$/)
            {
                my $parens = $1;
                $parens =~ s/\s//;    # Remove white space
                $group += length($parens);
                if ($log->is_trace())
                {
                    $log->trace("Located open grouping ($parens).");
                }
            }
            else
            {
                throw MAM::Exception('310',
                    "The open grouping is malformed ($pattern).");
            }
        }

        # Peel off the name
        if ($command =~ s/^\s*(\w+(?:\.\w+)?)\s*//) { $pattern = $1; }
        else
        {
            throw MAM::Exception('310',
                "An option name was expected ($command).");
        }
        if ($pattern =~ /^([A-Za-z]\w*(?:\.[A-Za-z]\w*)?)$/)
        {
            $name = $1;
            if ($log->is_trace())
            {
                $log->trace("Located name ($name).");
            }
        }
        else
        {
            throw MAM::Exception('310', "The name is malformed ($pattern).");
        }

        # Peel off the part
        if ($command =~ s/^\s*(?:\{([^\}]+)\})\s*//)
        {
            $pattern = $1;
            if ($pattern =~ /^\w+$/)
            {
                $part = $pattern;
                if ($log->is_trace())
                {
                    $log->trace("Located part ($part).");
                }
            }
            else
            {
                throw MAM::Exception('310',
                    "The part is malformed ($pattern).");
            }
        }

        # Peel off the operator
        if ($command =~ s/^\s*([-+<>!~:=]+)\s*//) { $pattern = $1; }
        else
        {
            throw MAM::Exception('310', "An operator was expected ($command).");
        }
        if ($pattern =~ /^(:!|[-+=<>!~:]=?|!~)$/)
        {
            if    ($1 eq '=')  { $op = 'Assign'; }
            elsif ($1 eq '==') { $op = 'EQ'; }
            elsif ($1 eq '>')  { $op = 'GT'; }
            elsif ($1 eq '<')  { $op = 'LT'; }
            elsif ($1 eq '>=') { $op = 'GE'; }
            elsif ($1 eq '<=') { $op = 'LE'; }
            elsif ($1 eq '!=') { $op = 'NE'; }
            elsif ($1 eq '~')  { $op = 'Match'; }
            elsif ($1 eq '!~') { $op = 'NotMatch'; }
            elsif ($1 eq '+=') { $op = 'Inc'; }
            elsif ($1 eq '-=') { $op = 'Dec'; }
            elsif ($1 eq ':=') { $op = 'Option'; }
            elsif ($1 eq ':!') { $op = 'NotOption'; }
            else
            {
                throw MAM::Exception('310', "Invalid opertor ($1).");
            }
            if ($log->is_trace())
            {
                $log->trace("Located operator ($op).");
            }
        }
        else
        {
            throw MAM::Exception('310',
                "The operator is malformed ($pattern).");
        }

        # Peel off the value
        if ($command =~
            s/^\s*([\"\']((?:[^\"\'\\]|\\.)*)[\"\']|([^\(\)\|& ]+))\s*//)
        {
            $pattern = $1;
        }
        else
        {
            throw MAM::Exception('310', "A value was expected ($command).");
        }
        if   (defined $3) { $value = $3; }
        else              { $value = $2; $value =~ s/\\\"/\"/g; $quoted = 1; }
        if ($log->is_trace())
        {
            $log->trace("Located value ($value).");
        }

        # Peel off any close grouping
        if ($command =~ s/^\s*((\(*\)+\s*)+)\s*//)
        {
            $pattern = $1;
            if ($pattern =~ /^((\)\s*)+)$/)
            {
                my $parens = $1;
                $parens =~ s/\s//;    # Remove white space
                $group -= length($parens);
                if ($log->is_trace())
                {
                    $log->trace("Located close grouping ($parens).");
                }
            }
            else
            {
                throw MAM::Exception('310',
                    "The close grouping is malformed ($pattern).");
            }
        }

        # Update option
        if ($op eq 'Option' || $op eq 'NotOption')
        {
            # Update selections
            if ($name eq 'Show')
            {
                # Populate Selections
                foreach my $select (split(/,/, $value))
                {
                    if ($select =~
                        /(?:(?:(\w+)\((\w+(?:\.\w+)?(?:\{[^\}]+\})?)?\))|(\w+(?:\.\w+)?(?:\{[^\}]+\})?))(?:=(\w+))?/
                      )
                    {
                        my $complexName = $2 ? $2 : $3;
                        my $op          = $1;
                        my $alias       = $4;
                        $complexName =~ /^(?:(\w+)\.)?(\w+)(?:\{([^\}]+)\})?/;
                        my ($object, $name, $part) = ($1, $2, $3);
                        my $selection = new MAM::Selection(name => $name);
                        if ($object)
                        {
                            $selection->setObject($object);
                        }
                        if ($part)
                        {
                            $selection->setPart($part);
                        }
                        if ($op)
                        {
                            $selection->setOperator($op);
                        }
                        if ($alias)
                        {
                            $selection->setAlias($alias);
                        }
                        if ($log->is_trace())
                        {
                            $log->trace('Adding selection to request ('
                                  . $selection->toString()
                                  . ').');
                        }
                        $self->addSelection($selection);
                    }
                    else
                    {
                        throw MAM::Exception('310',
                            "The selection is malformed ($select).");
                    }
                }
            }

            # Update Data
            elsif ($name eq 'Data')
            {
                # Populate Data
                if ($log->is_trace())
                {
                    $log->trace("Adding data to request ($value).");
                }
                my $parser = XML::LibXML->new();
                # Do we need to protect this area with a try/catch block?
                my $doc  = $parser->parse_string('<Data>' . $value . '</Data>');
                my $data = $doc->getDocumentElement();
                $self->setDataElement($data);
            }

            elsif ($name eq 'ChunkSize')
            {
                if ($log->is_trace())
                {
                    $log->trace(
                        "Setting chunking on request with chunk size ($value)."
                    );
                }
                $self->setChunking(1);
                $self->setChunkSize($value);
            }

            # Update not options
            elsif ($op eq 'NotOption')
            {
                my $option =
                  new MAM::Option(name => $name, value => $value, op => 'Not');
                if ($log->is_trace())
                {
                    $log->trace('Adding option to request ('
                          . $option->toString()
                          . ').');
                }
                $self->addOption($option);
            }

            # Update options
            else
            {
                my $option = new MAM::Option(name => $name, value => $value);
                if ($log->is_trace())
                {
                    $log->trace('Adding option to request ('
                          . $option->toString()
                          . ').');
                }
                $self->addOption($option);
            }
        }

        # Update assignment
        elsif ($op eq 'Assign' || $op eq 'Inc' || $op eq 'Dec')
        {
            my $assignment =
              new MAM::Assignment(name => $name, value => $value, op => $op);
            if ($log->is_trace())
            {
                $log->trace('Adding assignment to request ('
                      . $assignment->toString()
                      . ').');
            }
            $self->addAssignment($assignment);
        }

        # Update condition
        else
        {
            $name =~ /^(([A-Z]\w*)\.)?([A-Z]\w*)$/;
            my ($nameObject, $nameName) = ($2, $3);
            my $condition = new MAM::Condition(name => $nameName);
            if ($nameObject)
            {
                $condition->setObject($nameObject);
            }
            if ($quoted)
            {
                $condition->setValue($value);
            }
            else
            {
                $value =~ /^(([^\.]+)\.)?([^\.]+)$/;
                my ($valueObject, $valueName) = ($2, $3);
                $condition->setValue($valueName);
                if ($valueObject)
                {
                    $condition->setSubject($valueObject);
                }
            }
            if ($part)
            {
                $condition->setPart($part);
            }
            if ($op)
            {
                $condition->setOperator($op);
            }
            if ($conj)
            {
                $condition->setConjunction($conj);
            }
            if ($group)
            {
                $condition->setGroup($group);
            }
            if ($log->is_trace())
            {
                $log->trace('Adding condition to request ('
                      . $condition->toString()
                      . ').');
            }
            $self->addCondition($condition);
        }
    }

    $log->info('Built request: ', $self->toString());

    return $self;
}

# Seeks a response from the server
sub getResponse
{
    my ($self) = @_;
    my ($messageChunk, $replyChunk, $response, $caught);
    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    # Instantiate a new message
    $messageChunk = new MAM::Chunk();

    # Append the request to the message
    if ($log->is_debug())
    {
        $log->debug('Appending a request to the message chunk.');
    }
    try
    {
        $messageChunk->setRequest($self);
    }
    catch MAM::Exception with
    {
        $response =
          MAM::Response()->failure('300', 'Failed building message: ' . $_[0]);
        $caught = 1;
    };
    return $response if $caught;
    if ($log->is_trace())
    {
        $log->trace('Message chunk built.');
    }

    # Obtain the reply
    if ($log->is_debug())
    {
        $log->debug('Obtaining reply chunk from server.');
    }
    try
    {
        $replyChunk = $messageChunk->getChunk();
    }
    catch MAM::Exception with
    {
        $response =
          new Response()->failure('226', 'Failed obtaining reply: ' . $_[0]);
        $caught = 1;
    };
    return $response if $caught;
    if ($log->is_trace())
    {
        $log->trace('Obtained reply chunk.');
    }

    # Extract the response from the reply
    if ($log->is_debug())
    {
        $log->debug('Extracting response from the reply chunk.');
    }
    try
    {
        $response = $replyChunk->getResponse();
    }
    catch MAM::Exception with
    {
        $response =
          new Response()
          ->failure('300',
            'Failed extracting response from reply chunk: ' . $_[0]);
        $caught = 1;
    };
    return $response if $caught;
    if ($log->is_debug())
    {
        $log->debug(
            'Response extracted from reply chunk: ' . $response->toString());
    }

    return $response;
}

1;
