#! /usr/bin/perl -wT
################################################################################
#
# MAM Response object
#
# File   :  Response.pm
#
################################################################################
#
#    This file includes parts which are copyrighted by Adaptive Computing
#    Enterprises, Inc., and parts which are copyrighted by Battelle Memorial
#    Institute.
#
#    The terms and conditions for the use and redistribution of these parts
#    are governed by the Moab Accounting Manager License and the BSD License
#    respectively. Refer to the LICENSE file for details.
#
#    Copyright (C) 2006 - 2016 by Adaptive Computing Enterprises, Inc.
#    All Rights Reserved.
#
#    The Moab Accounting Manager License specifies the terms and conditions
#    for use and redistribution.
#
#    The Moab Accounting Manager License applies to the Moab Accounting
#    Manager software offered by Adaptive Computing Enterprises, Inc.
#    By installing or using this software, Licensee accepts a non-exclusive
#    license from Adaptive Computing Enterprises, Inc. and is bound to accept
#    acknowledgement of and abide by the notices and conditions of the
#    Moab Accounting Manager License.
#
#    Copyright (C) 2003 - 2005 Pacific Northwest National Laboratory,
#    Battelle Memorial Institute. All rights reserved.
#
#    The BSD License specifies the terms and conditions for use and
#    redistribution.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions
#    are met:
#
#    - Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
#    - Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
#    - Neither the name of Battelle nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
#    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
#    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
#    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#    POSSIBILITY OF SUCH DAMAGE.
#
################################################################################

use strict;

=head1 NAME

MAM::Response - build and parse a MAM response

=head1 DESCRIPTION

The B<MAM::Response> module defines functions to build and parse MAM responses

=head1 CONSTRUCTOR

my $response = new MAM::Response(status => $status, code => $code, message => $message, count => $code, data => \@data, metadata => \%metadata, actor => $actor);

=head1 ACCESSORS

=over 4

=item $actor = $response->getActor();

=item $status = $response->getStatus();

=item $code = $response->getCode();

=item $message = $response->getMessage();

=item $count = $response->getCount();

=item @data = $response->getData();

=item $element = $response->getDataElement();

=item $value = $response->getDatumValue($name);

=item %metadata = $response->getMetadata();

=item $dataType = $response->getDataType($name);

=item $number = $response->getChunkNum();

=item $maximum = $response->getChunkMax();

=back

=head1 MUTATORS

=over 4

=item $response = $response->setActor($actor);

=item $response = $response->setStatus($status);

=item $response = $response->setCode($code);

=item $response = $response->setMessage($message);

=item $response = $response->setCount($count);

=item $response = $response->setData(\@data);

=item $response = $response->setDataElement($element);

=item $response = $response->addDatum($datum);

=item $response = $response->setMetadata(\%metadata);

=item $response = $response->setDataType($name, $dataType);

=item $response = $response->setChunkNum($number);

=item $response = $response->setChunkMax($maximum);

=back

=head1 OTHER METHODS

=over 4

=item $string = $response->toString();

=item $response = $response->success($count, $message)
=item $response = $response->success($message)
=item $response = $response->success($count, \@data)
=item $response = $response->success($count, \@data, $message)

=item $response = $response->warning($count, $message)
=item $response = $response->warning($message)
=item $response = $response->warning($count, \@data, \%metadata)
=item $response = $response->warning($count, \@data, \%metadata, $message)

=item $response = $response->failure($code, \@data, \%metadata, $message)
=item $response = $response->failure($code, $message)
=item $response = $response->failure($message)
=item $response = $response->failure($exception)

=back

=head1 EXAMPLES

use MAM::Response;

my $response = new MAM::Response(status => 'Success', code => '000', message => 'Successfully Created 1 User');

=head1 REQUIRES

Perl 5.8

=head1 AUTHOR

Scott Jackson, scottmo@adaptivecomputing.com

=cut

##############################################################################

package MAM::Response;

use vars qw($log);
use XML::LibXML;
use MAM::Datum;
use MAM::Global;
use JSON;

# ----------------------------------------------------------------------------
# Constructor
# ----------------------------------------------------------------------------

sub new
{
    my ($class, %arg) = @_;
    my $self = {
        _actor   => $arg{actor}   || (getpwuid($<))[0],    # SCALAR
        _status  => $arg{status}  || 'Failure',            # SCALAR
        _code    => $arg{code}    || '999',                # SCALAR
        _message => $arg{message} || '',                   # SCALAR
        _count => defined $arg{count} ? $arg{count} : -1,  # SCALAR
        _data     => $arg{data}     || [],                 # ARRAY REF of Datums
        _metadata => $arg{metadata} || {},                 # HASH REF
        _chunkNum => $arg{chunkNum} || 1,                  # SCALAR
        _chunkMax => $arg{chunkMax} || 0,                  # SCALAR
    };
    bless $self, $class;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (',
            join(', ', map {"$_ => $arg{$_}"} keys %arg), ')');
    }

    return $self;
}

# ----------------------------------------------------------------------------
# Accessors
# ----------------------------------------------------------------------------

# Get the response actor
sub getActor
{
    my ($self) = @_;
    return $self->{_actor};
}

# Get the response status
sub getStatus
{
    my ($self) = @_;
    return $self->{_status};
}

# Get the response code
sub getCode
{
    my ($self) = @_;
    return $self->{_code};
}

# Get the response message
sub getMessage
{
    my ($self) = @_;
    return $self->{_message};
}

# Get the response count
sub getCount
{
    my ($self) = @_;
    return $self->{_count} + 0;
}

# Get the chunk number
sub getChunkNum
{
    my ($self) = @_;
    return $self->{_chunkNum} + 0;
}

# Get the chunk maximum
sub getChunkMax
{
    my ($self) = @_;
    return $self->{_chunkMax} + 0;
}

# Get the response data (list of data)
sub getData
{
    my ($self) = @_;
    return @{$self->{_data}};
}

# Get the response data element
sub getDataElement
{
    my ($self) = @_;
    my $data = new XML::LibXML::Element('Data');
    foreach my $datum (@{$self->{_data}})
    {
        $data->appendChild($datum->getElement());
    }
    return $data;
}

# Get the value of the named property in the first datum
sub getDatumValue
{
    my ($self, $name) = @_;
    my $value;
    if (@{$self->{_data}})
    {
        my $datum      = (${$self->{_data}}[0])->getElement();
        my @properties = $datum->getChildrenByTagName($name);
        if ($properties[0])
        {
            $value = ($properties[0])->textContent();
        }
    }
    return $value;
}

# Get the response metadata (hash of data types)
sub getMetadata
{
    my ($self) = @_;
    return %{$self->{_metadata}};
}

# Get a specific property's data type
sub getDataType
{
    my ($self, $name) = @_;
    if (exists ${$self->{_metadata}}{$name})
    {
        return ${$self->{_metadata}}{$name};
    }
    return;
}

# ----------------------------------------------------------------------------
# Mutators
# ----------------------------------------------------------------------------

# Set the response actor
sub setActor
{
    my ($self, $actor) = @_;
    $self->{_actor} = $actor;
    return $self;
}

# Set the response status
sub setStatus
{
    my ($self, $status) = @_;
    $self->{_status} = $status;
    return $self;
}

# Set the response code
sub setCode
{
    my ($self, $code) = @_;
    $self->{_code} = $code;
    return $self;
}

# Set the response message
sub setMessage
{
    my ($self, $message) = @_;
    $self->{_message} = $message;
    return $self;
}

# Set the response count
sub setCount
{
    my ($self, $count) = @_;
    $self->{_count} = $count;
    return $self;
}

# Set the chunk number
sub setChunkNum
{
    my ($self, $chunkNum) = @_;
    $self->{_chunkNum} = $chunkNum;
    return $self;
}

# Set the chunk maximum
sub setChunkMax
{
    my ($self, $chunkMax) = @_;
    $self->{_chunkMax} = $chunkMax;
    return $self;
}

# Set the response data (list of data)
sub setData
{
    my ($self, $data) = @_;
    $self->{_data} = $data;
    return $self;
}

# Set the response data (from element)
sub setDataElement
{
    my ($self, $data) = @_;

    foreach my $datum ($data->childNodes())
    {
        push(@{$self->{_data}}, new MAM::Datum($datum));
    }
    return $self;
}

# Add a response datum
sub addDatum
{
    my ($self, $datum) = @_;
    push(@{$self->{_data}}, $datum);
    return $self;
}

# Set the response metadata (hash of property data types)
sub setMetadata
{
    my ($self, $metadata) = @_;
    $self->{_metadata} = $metadata;
    return $self;
}

# Set a data type for a specific property
sub setDataType
{
    my ($self, $name, $dataType) = @_;
    ${$self->{_metadata}}{$name} = $dataType;
    return $self;
}

# ----------------------------------------------------------------------------
# Other Methods
# ----------------------------------------------------------------------------

# Convert response to printable string
sub toString
{
    my ($self) = @_;

    my $string = '(';
    $string .= $self->{_actor};
    $string .= ', ';
    $string .= $self->{_status};
    $string .= ', ';
    $string .= $self->{_code};
    $string .= ', ';
    $string .= $self->{_message};
    $string .= ', ';
    $string .= $self->{_count};
    $string .= ', ';
    $string .= '[' . join(', ', map($_->toString(), @{$self->{_data}})) . ']';
    $string .= ', ';
    $string .= encode_json $self->{_metadata};
    $string .= ', ';
    $string .= $self->{_chunkNum};
    $string .= ', ';
    $string .= $self->{_chunkMax};
    $string .= ')';

    return $string;
}

# Prepare a failure response
sub failure
{
    my ($self) = shift;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[0 .. $#_]), ')');
    }

    # failure($code, \@data, \%metadata, $message)
    if (   defined $_[0]
        && ! ref($_[0])
        && $_[1]
        && ref($_[1]) eq 'ARRAY'
        && $_[2]
        && ref($_[2]) eq 'HASH'
        && defined $_[3]
        && ! ref($_[3]))
    {
        $self->setStatus('Warning');
        $self->setCode('100');
        $self->setCount($_[0]);
        $self->setData($_[1]);
        $self->setMetadata($_[2]);
        $self->setMessage($_[3]);
        return $self;
    }

    # failure($code, $message)
    elsif (defined $_[0] && ! ref($_[0]) && defined $_[1] && ! ref($_[1]))
    {
        $self->setStatus('Failure');
        $self->setCode($_[0]);
        $self->setMessage($_[1]);
        return $self;
    }

    # failure($message)
    elsif (defined $_[0] && ! ref($_[0]) && ! $_[1])
    {
        $self->setStatus('Failure');
        $self->setCode('999');
        $self->setMessage($_[0]);
        return $self;
    }

    # failure($exception)
    elsif (ref($_[0]) eq 'Exception')
    {
        $self->setStatus('Failure');
        $self->setCode('999');
        $self->setMessage(($_[0])->{-text});
        return $self;
    }
}

# Prepare a warning response
sub warning
{
    my ($self) = shift;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[0 .. $#_]), ')');
    }

    # warning($count, \@data, \%metadata, $message)
    if (   defined $_[0]
        && ! ref($_[0])
        && $_[1]
        && ref($_[1]) eq 'ARRAY'
        && $_[2]
        && ref($_[2]) eq 'HASH'
        && defined $_[3]
        && ! ref($_[3]))
    {
        $self->setStatus('Warning');
        $self->setCode('100');
        $self->setCount($_[0]);
        $self->setData($_[1]);
        $self->setMetadata($_[2]);
        $self->setMessage($_[3]);
        return $self;
    }

    # warning($count, \@data, \%metadata)
    elsif (   defined $_[0]
        && ! ref($_[0])
        && $_[1]
        && ref($_[1]) eq 'ARRAY'
        && $_[2]
        && ref($_[2]) eq 'HASH')
    {
        $self->setStatus('Warning');
        $self->setCode('100');
        $self->setCount($_[0]);
        $self->setData($_[1]);
        $self->setMetadata($_[2]);
        return $self;
    }

    # warning($count, $message)
    elsif (defined $_[0] && ! ref($_[0]) && defined $_[1] && ! ref($_[1]))
    {
        $self->setStatus('Warning');
        $self->setCode('100');
        $self->setMessage($_[1]);
        $self->setCount($_[0]);
        return $self;
    }

    # warning($message)
    elsif (defined $_[0] && ! ref($_[0]))
    {
        $self->setStatus('Warning');
        $self->setCode('100');
        $self->setMessage($_[0]);
        return $self;
    }
}

# Prepare a successful response
sub success
{
    my ($self) = shift;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[0 .. $#_]), ')');
    }

    # success($count, \@data, \%metadata, $message)
    if (   defined $_[0]
        && ! ref($_[0])
        && $_[1]
        && ref($_[1]) eq 'ARRAY'
        && $_[2]
        && ref($_[2]) eq 'HASH'
        && defined $_[3]
        && ! ref($_[3]))
    {
        $self->setStatus('Success');
        $self->setCode('000');
        $self->setCount($_[0]);
        $self->setData($_[1]);
        $self->setMetadata($_[2]);
        $self->setMessage($_[3]);
        return $self;
    }

    # success($count, \@data, \%metadata)
    elsif (   defined $_[0]
        && ! ref($_[0])
        && $_[1]
        && ref($_[1]) eq 'ARRAY'
        && $_[2]
        && ref($_[2]) eq 'HASH')
    {
        $self->setStatus('Success');
        $self->setCode('000');
        $self->setCount($_[0]);
        $self->setData($_[1]);
        $self->setMetadata($_[2]);
        return $self;
    }

    # success($count, $message)
    elsif (defined $_[0] && ! ref($_[0]) && defined $_[1] && ! ref($_[1]))
    {
        $self->setStatus('Success');
        $self->setCode('000');
        $self->setMessage($_[1]);
        $self->setCount($_[0]);
        return $self;
    }

    # success($message)
    elsif (defined $_[0] && ! ref($_[0]))
    {
        $self->setStatus('Success');
        $self->setCode('000');
        $self->setMessage($_[0]);
        return $self;
    }
}

1;
