#! /usr/bin/perl -w
################################################################################
#
# Response Handler for MAM Web Services
#
# File   :  WSResponseHandler.pm
#
################################################################################
#
#    This file includes parts which are copyrighted by Adaptive Computing
#    Enterprises, Inc., and parts which are copyrighted by Battelle Memorial
#    Institute.
#
#    The terms and conditions for the use and redistribution of these parts
#    are governed by the Moab Accounting Manager License and the BSD License
#    respectively. Refer to the LICENSE file for details.
#
#    Copyright (C) 2006 - 2016 by Adaptive Computing Enterprises, Inc.
#    All Rights Reserved.
#
#    The Moab Accounting Manager License specifies the terms and conditions
#    for use and redistribution.
#
#    The Moab Accounting Manager License applies to the Moab Accounting
#    Manager software offered by Adaptive Computing Enterprises, Inc.
#    By installing or using this software, Licensee accepts a non-exclusive
#    license from Adaptive Computing Enterprises, Inc. and is bound to accept
#    acknowledgement of and abide by the notices and conditions of the
#    Moab Accounting Manager License.
#
#    Copyright (C) 2003 - 2005 Pacific Northwest National Laboratory,
#    Battelle Memorial Institute. All rights reserved.
#
#    The BSD License specifies the terms and conditions for use and
#    redistribution.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions
#    are met:
#
#    - Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
#    - Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
#    - Neither the name of Battelle nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
#    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
#    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
#    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#    POSSIBILITY OF SUCH DAMAGE.
#
################################################################################

package MAM::WSResponseHandler;

use vars qw($clientConfig $log);
use strict;
use lib qw(/opt/mam/lib /opt/mam/lib/perl5);
use JSON;
use Apache2::Const -compile =>
  qw(OK HTTP_BAD_REQUEST HTTP_CREATED HTTP_FORBIDDEN HTTP_INTERNAL_SERVER_ERROR HTTP_METHOD_NOT_ALLOWED HTTP_NOT_FOUND HTTP_OK HTTP_PARTIAL_CONTENT HTTP_UNAUTHORIZED);
use Apache2::RequestRec ();
use Apache2::RequestIO  ();
use Apache2::Access     ();
use Apache2::Connection ();
use Apache2::Response   ();
use MAM::WS;
use MAM::Global;
use MAM::Client;

my ($action,             $actor, $data,   $method,
    $instance_specified, $r,     %params, $path_info,
    %payload,            $http_code
);

# ----------------------------------------------------------------------------
# $http_status_code = handler($apache2_requestrec)
# ----------------------------------------------------------------------------

# Handle response (called by mod_ssl for each new request)
sub handler
{
    $r         = shift;
    %params    = ();
    %payload   = ();
    $http_code = '';
    $method    = $r->method();
    $path_info = $r->path_info();
    my $args = $r->args();
    $actor = $r->user();
    my $content;
    $data               = undef;
    $action             = 'Unspecified';
    $instance_specified = undef;

    # Build up params hash from args string
    foreach my $arg (split /&/, $args)
    {
        if ($arg =~ /^([-\w]+)(:!|!~|[=<>~](?!=)|[-+=<>!\@:%]=)(.*)$/)
        {
            my ($name, $op, $value) = ($1, $2, $3);

            if   ($op eq '=') { $params{$name}       = $value; }
            else              { $params{$name . $op} = $value; }
        }
        elsif ($arg =~ /^!(.*)$/)
        {
            $params{$1} = 'False';
        }
        elsif ($arg =~ /^[^=]+$/)
        {
            $params{$arg} = 'True';
        }
        else
        {
            return http_failure(
                http_code => Apache2::Const::HTTP_BAD_REQUEST,    # 400
                code      => '810',
                message   => "Invalid http parameter ($arg).",
            );
        }
    }

    # Read data from body
    my $contentLength = $r->headers_in->{'Content-length'};
    if ($contentLength > 0)
    {
        $r->read($content, $contentLength);
    }

    # Check that it is valid JSON
    if ($content)
    {
        eval { $data = from_json($content, {utf8 => 1}); };
        if ($@)
        {
            $@ =~ s% at /.*%%;    # Remove file name from message
            return http_failure(
                http_code => Apache2::Const::HTTP_BAD_REQUEST,    # 400
                code      => '812',
                message => "Invalid json in http request body ($content): $@",
            );
        }
    }

    if ($log->is_trace())
    {
        my $connection  = $r->connection();
        my $remote_host = $connection->get_remote_host();

        $log->trace("Handling new web service request: $method $path_info"
              . ($args ? '?' : '')
              . "$args from $actor at $remote_host");
    }

    #$log->debug("method: $method");
    #$log->debug("path-info $path_info");
    #$log->debug("params: " . Data::Dumper->Dump([\%params], ['params']));
    #$log->debug("data: " . Data::Dumper->Dump([$data], ['data']));

    # Check that we are using one of the supported http methods
    if (   $method eq 'GET'
        || $method eq 'POST'
        || $method eq 'PUT'
        || $method eq 'PATCH'
        || $method eq 'DELETE')
    {
        service_method();
    }

    # Return 405 Method Not Allowed if http method not supported
    else
    {
        http_failure(
            http_code => Apache2::Const::HTTP_METHOD_NOT_ALLOWED,    # 405
            code      => '810',
            message =>
              "Unsupported HTTP Method ($method) used in web services request",
        );
    }
}

# ----------------------------------------------------------------------------
# $http_status_code = http_response(code => $mam_code, count => $mam_count, data => $mam_data, http_code => $http_code, message => mam_message, status => mam_$status)
# ----------------------------------------------------------------------------

# Render the HTTP response
sub http_response
{
    my (%arg) = @_;

    # Incorporate args into response
    if (defined $arg{code})      { $payload{code}    = $arg{code}; }
    if (defined $arg{count})     { $payload{count}   = $arg{count}; }
    if (defined $arg{data})      { $payload{data}    = $arg{data}; }
    if (defined $arg{http_code}) { $http_code        = $arg{http_code}; }
    if (defined $arg{message})   { $payload{message} = $arg{message}; }
    if (defined $arg{status})    { $payload{status}  = $arg{status}; }

    # Derive status and code if absent

    # If we have neither a status nor a code, we default to success
    if (! defined $payload{status} && ! defined $payload{code})
    {
        $payload{status} = 'Success';
        $payload{code}   = '000';
    }
    # If we have a status but no code, derive code from status
    unless (defined($payload{code}))
    {
        if ($payload{status} eq 'Success')
        {
            $payload{code} = '000';
        }
        elsif ($payload{status} eq 'Warning')
        {
            $payload{code} = '100';
        }
        else { $payload{code} = '999'; }
    }
    # If we have a code but no status, derive status from code
    unless (defined $payload{status})
    {
        if ($payload{code} ge '000' && $payload{code} lt '100')
        {
            $payload{status} = 'Success';
        }
        elsif ($payload{code} ge '100'
            && $payload{code} lt '200')
        {
            $payload{status} = 'Warning';
        }
        else { $payload{status} = 'Failure'; }
    }

    # Print the payload in JSON if present
    if (%payload)
    {
        $r->content_type('application/json');

        my $json_flags = {canonical => 1};
        $json_flags->{pretty} = 1 unless ($params{pretty} // '') =~ /false/i;

        $r->print(to_json(\%payload, $json_flags));
    }

    # Derive HTTP status code if absent
    unless ($http_code)
    {
        if (   $instance_specified
            && $payload{code} eq '000'
            && $payload{count} == 0)
        {
            $http_code = Apache2::Const::HTTP_NOT_FOUND;    # 404
        }
        #elsif ($payload{code} eq '000' && $action eq 'Create')
        #{
        #    $http_code = Apache2::Const::HTTP_CREATED;                   # 201
        #}
        elsif ($payload{code} ge '000' && $payload{code} lt '200')
        {
            $http_code = Apache2::Const::HTTP_OK;    # 000
        }
        elsif ($payload{code} ge '310' && $payload{code} lt '320')
        {
            $http_code = Apache2::Const::HTTP_BAD_REQUEST;    # 400
        }
        elsif ($payload{code} ge '420' && $payload{code} lt '430')
        {
            $http_code = Apache2::Const::HTTP_UNAUTHORIZED;    # 401
        }
        elsif ($payload{code} ge '440' && $payload{code} lt '450')
        {
            $http_code = Apache2::Const::HTTP_FORBIDDEN;       # 403
        }
        else
        {
            $http_code = Apache2::Const::HTTP_BAD_REQUEST;     # 400
            #$http_code = Apache2::Const::HTTP_INTERNAL_SERVER_ERROR;     # 500
        }
    }

    # We have to do a little two-step shuffle here to remove mod_perl's output
    $r->status($http_code);
    $r->custom_response(Apache2::Const::HTTP_OK, '');
    #$r->custom_response($http_code, '');
    return Apache2::Const::HTTP_OK;
    #return $http_code;
}

# ----------------------------------------------------------------------------
# $http_response_string = http_success(body => $response_json)
# ----------------------------------------------------------------------------

# Return a successful HTTP response
sub http_success
{
    my (%arg) = @_;

    http_response(%arg);    # Pass through since we default to success
}

# ----------------------------------------------------------------------------
# $http_status_code = http_failure($mod_perl_request, $json_response[, $http_status_code])
# ----------------------------------------------------------------------------

# Return a failure HTTP response
sub http_failure
{
    my (%arg) = @_;

    $arg{status} = 'Failure';

    $arg{code} = '999' unless defined $arg{code};

    http_response(%arg);
}

# ----------------------------------------------------------------------------
# $http_status_code = http_warning($mod_perl_request, $json_response[, $http_status_code])
# ----------------------------------------------------------------------------

# Return a warning HTTP response
sub http_warning
{
    my (%arg) = @_;

    $arg{status} = 'Warning';

    $arg{code} = '100' unless defined $arg{code};

    http_response(%arg);
}

# ----------------------------------------------------------------------------
# $http_response_string = convert_mam_to_http_response($mam-response)
# ----------------------------------------------------------------------------

# Convert a MAM response to an HTTP response
sub convert_mam_to_http_response
{
    my ($response) = @_;

    my $code     = $response->getCode();
    my $count    = $response->getCount();
    my @data     = $response->getData();
    my $message  = $response->getMessage();
    my %metadata = $response->getMetadata();
    my $status   = $response->getStatus();

    my %arg = ();

    $arg{code} = $code;
    if ($count >= 0) { $arg{count} = $count; }
    if (@data)
    {
        my @dataLOH = ();
        foreach my $datum (@data)
        {
            my %datumH = ();
            foreach my $property ($datum->getProperties())
            {
                my $name     = $property->getName();
                my $value    = $property->getValue();
                my $dataType = $metadata{$name} // 'String';
                if ($value eq '')
                {
                    if ($params{'suppress-nulls'} =~ /true/i) { next; }
                    else { $value = undef; }
                }
                elsif ($dataType eq 'Currency')
                {
                    if   ($value =~ /(-?)inf/i) { $value = $1 . 'Infinity'; }
                    else                        { $value = $value + 0; }
                }
                elsif ($dataType eq 'AutoGen'
                    || $dataType eq 'Integer'
                    || $dataType eq 'Float')
                {
                    $value = $value + 0;
                }
                elsif ($dataType eq 'Boolean')
                {
                    if    ($value =~ /true/i)  { $value = JSON::true; }
                    elsif ($value =~ /false/i) { $value = JSON::false; }
                }
                $datumH{lhc($name)} = $value;
            }
            push @dataLOH, \%datumH;
        }
        $arg{data} = \@dataLOH;
    }
    if ($message ne '') { $arg{message} = $message; }
    $arg{status} = $status;

    return http_response(%arg);
}

# ----------------------------------------------------------------------------
# $query_response_string = service_method()
# ----------------------------------------------------------------------------

# Service an http request
sub service_method
{
    my @nodes;

    # Path-info is required (for object)
    if ((@nodes = ($path_info =~ m%(?:/([^/]+))%g)) && @nodes)
    {
        # Ensure we support and use the requested version
        my $version = 'v1';
        $version = shift @nodes if $nodes[0] =~ /^v\d$/;
        if ($version ne 'v1')
        {
            return http_failure(
                http_code => Apache2::Const::HTTP_BAD_REQUEST,    # 400
                code      => '810',
                message =>
                  "The specified api version ($version) is not supported.",
            );
        }

        # Extract the object
        my $object = shift @nodes
          || return http_failure(
            http_code => Apache2::Const::HTTP_BAD_REQUEST,                 # 400
            code      => '810',
            message   => "No object specified in path info ($path_info).",
          );
        $object =~ s/s$//;    # Make singular
        $object = ucc($object);    # Convert object to UCC for MAM

        # Instantiate the request
        my $request = new MAM::Request(
            object => $object,
            actor  => $actor
        );

        # Extract action from params if present
        if (defined $params{'action'})
        {
            $action = ucc($params{'action'});
        }

        # Set unspecified actions as appropriate for the http method
        if ($action eq 'Unspecified')
        {
            if ($method eq 'GET') { $action = 'Query'; }
            elsif ($method eq 'POST' || $method eq 'PUT')
            {
                $action = 'Create';
            }
            elsif ($method eq 'PATCH')  { $action = 'Modify'; }
            elsif ($method eq 'DELETE') { $action = 'Delete'; }
        }
        # Verify action is legal for http method
        if (   ($method eq 'GET' && $action ne 'Query')
            || ($method eq 'PUT'    && $action ne 'Create')
            || ($method eq 'PATCH'  && $action ne 'Modify')
            || ($method eq 'DELETE' && $action ne 'Delete'))
        {
            return http_failure(
                http_code => Apache2::Const::HTTP_BAD_REQUEST,    # 400
                code      => '810',
                message =>
                  "The $method http method cannot be used with the $action action parameter.",
            );
        }

        # Apply the action
        $request->setAction($action);

        # Add authentication bits
        if ($r->auth_type() eq 'Basic')
        {
            my ($status, $password) = $r->get_basic_auth_pw;

            if ($log->is_trace())
            {
                $log->trace("Basic authentication utilized by $actor");
            }

            $request->setTokenType($TOKEN_PASSWORD);
            $request->setTokenName($actor);
            $request->setTokenValue($password);
        }

        # Process additional path-info as primary keys
        my @primaryKeys;
        if (@nodes)
        {
            #$log->debug(Data::Dumper->Dump([\@nodes], ['nodes']));

            unless ($method eq 'GET'
                || $method eq 'PUT'
                || $method eq 'PATCH'
                || $method eq 'DELETE')
            {
                return http_failure(
                    http_code => Apache2::Const::HTTP_BAD_REQUEST,    # 400
                    code      => '810',
                    message =>
                      "Primary keys may not be specified via the path info when using the $method http method.",
                );
            }

            # Query metadata to obtain primary keys
            my $subRequest = new MAM::Request(
                object     => 'Attribute',
                action     => 'Query',
                actor      => $actor,
                selections => [new MAM::Selection(name => 'Name')],
                conditions => [
                    new MAM::Condition(name => 'Object',     value => $object),
                    new MAM::Condition(name => 'PrimaryKey', value => 'True')
                ],
                options =>
                  [new MAM::Option(name => 'SortBySequence', value => 'True')],
            );

            my $subResponse = $subRequest->getResponse();

            if ($subResponse->getStatus() eq 'Failure')
            {
                return http_failure(
                    code    => $subResponse->getCode(),
                    message => "Key data query failed with: "
                      . $subResponse->getMessage(),
                );
            }

            my @keyData = $subResponse->getData();
            @primaryKeys = map { $_->getPropertyValue('Name') } @keyData;

            if (@nodes == @primaryKeys)
            {
                $instance_specified = 1;
            }
            elsif (@nodes > @primaryKeys)
            {
                return http_failure(
                    http_code => Apache2::Const::HTTP_BAD_REQUEST,    # 400
                    code      => '810',
                    message =>
                      "Invalid path info in URL ($path_info). The specified object ($object) does not have "
                      . @nodes
                      . " primary keys",
                );
            }

            # Apply path elements as primary key conditions on request
            foreach my $index (0 .. $#nodes)
            {
                $nodes[$index] = '' if $nodes[$index] =~ /^null$/i;
                
                if ($method eq 'PUT'
                    && ($action eq 'Unspecified' || $action eq 'Create'))
                {
                    $request->addAssignment($primaryKeys[$index],
                        $nodes[$index]);
                }
                else
                {
                    $request->addCondition($primaryKeys[$index],
                        $nodes[$index]);
                }
            }
        }
        else
        {
            unless (
                   $method eq 'GET'
                || $method eq 'POST'
                || $method eq 'PATCH'    # Should we remove this?
                || $method eq 'DELETE'
              )                          # Should we remove this?
            {
                return http_failure(
                    http_code => Apache2::Const::HTTP_BAD_REQUEST,    # 400
                    code      => '810',
                    message =>
                      "The primary key(s) must be specified via the path info when using the $method http method.",
                );
            }
        }

        # Modify, Delete and Undelete should fully specify the object
        # unless the force option is used
        if (   ($method eq 'PATCH' || $method eq 'DELETE')
            && ((! @nodes) || (@nodes < @primaryKeys))
            && (! (    (exists $params{'force'} && ! defined $params{'force'})
                    || $params{'force'} =~ /^true$/i
                    || $params{'force:='} =~ /^true$/i
                    || $params{'options'} =~ /\bforce=true\b/i
                )
            )
          )
        {
            return http_failure(
                http_code => Apache2::Const::HTTP_BAD_REQUEST,    # 400
                code      => '810',
                message =>
                  "Did you really mean to specify the $action action against multiple objects (object not fully qualified in path-info)? Use the force option if you know what you are doing and want to override this protection.",
            );
        }

        # Process http params
        my $combinedSelections;
        my %dataProperties = ();
        foreach my $key (sort keys %params)
        {
            # Apply selections
            if (   $key eq 'select'
                || $key eq 'fields'
                || $key eq 'show'
                || $key eq 'get')
            {
                my $selectString = $params{$key};

                unless ($method eq 'GET')
                {
                    return http_failure(
                        http_code => Apache2::Const::HTTP_BAD_REQUEST,    # 400
                        code      => '810',
                        message =>
                          "The $key parameter may only be used with the GET http method.",
                    );
                }

         # Intercept partial+aggregate selection combos and extract aggregations
         # from backend request so we can handle aggregations later in frontend
                if ($selectString =~ /\(/ && $selectString =~ /\{/)
                {
                    # Remove quoting to simplify pattern matching
                    $selectString =~ s/^[\"\']((?:[^\"\'\\]|\\.)*)[\"\']$/$1/;
                    $combinedSelections = $selectString;

                    # Remove aggregations for backend
                    $selectString =~ s/[-\w]+\((\w+(?:\{[^\}]+\})?)\)/$1/g;

                    # Uppercase aggregations for frontend
                    $combinedSelections =~
                      s/([-\w]+)\(([-\w]+(?:\{[^\}]+\})?)\)/ucc($1)."($2)"/ge;
                }

                foreach my $select (split /,/, $selectString)
                {
                    if ($select =~
                        /(?:(?:([-\w]+)\(([-\w]+(?:\.[-\w]+)?(?:\{[^\}]+\})?)?\))|([-\w]+(?:\.[-\w]+)?(?:\{[^\}]+\})?))(?:=([-\w]+))?/
                      )
                    {
                        my $complexName = $2 ? $2 : $3;
                        my $op          = $1;
                        my $alias       = $4;
                        $complexName =~
                          /^(?:([-\w]+)\.)?([-\w]+)(?:\{([^\}]+)\})?/;
                        my ($object, $name, $part) = ($1, $2, $3);
                        my $selection = new MAM::Selection();
                        if ($object)
                        {
                            $selection->setObject(ucc($object));
                        }
                        $selection->setName(ucc($name));
                        if ($part)
                        {
                            $selection->setPart($part);
                        }
                        if ($op)
                        {
                            $selection->setOperator(ucc($op));
                        }
                        if ($alias)
                        {
                            $selection->setAlias($alias);
                        }

                        $request->addSelection($selection);
                    }
                    else
                    {
                        return http_failure(
                            http_code => Apache2::Const::HTTP_BAD_REQUEST, # 400
                            code      => '812',
                            message =>
                              "The field selection is malformed ($select).",
                        );
                    }
                }
            }

            # Apply conditions
            elsif ($key eq 'filter' || $key eq 'query' || $key eq 'where')
            {
                my $filterString = $params{$key};

                unless (
                       $method eq 'GET'
                    || $method eq 'PATCH'
                    || $method eq 'DELETE'
                    || ($method eq 'POST'
                        && (   $action eq 'Query'
                            || $action eq 'Modify'
                            || $action eq 'Delete'
                            || $action eq 'Undelete')
                    )
                  )
                {
                    return http_failure(
                        http_code => Apache2::Const::HTTP_BAD_REQUEST,    # 400
                        code      => '810',
                        message =>
                          "The $key parameter may not be used with the $method http method and $action action.",
                    );
                }

                while ($filterString =~ s/^([,\|])?([^,\|]+)//)
                {
                    my ($delimiter, $filter) = ($1, $2);
                    if ($filter =~
                        /^(\(*)(?:([-\w]+)\.)?([-\w]+)(?:\{([^\}]+)\})?([<>!~=]+)(?:([-\w]+)\.)?([^\)]+)(\)*)$/
                      )
                    {
                        my ($pregroup, $object, $name, $part, $op, $subject,
                            $value, $postgroup)
                          = ($1, $2, $3, $4, $5, $6, $7, $8);

                        my $condition = new MAM::Condition();
                        $condition->setObject(ucc($object)) if $object;
                        $condition->setName(ucc($name));
                        $condition->setPart($part) if $part;
                        if ($op eq '==' || $op eq '=')
                        {
                            $condition->setOperator('EQ');
                        }
                        elsif ($op eq '>')  { $condition->setOperator('GT'); }
                        elsif ($op eq '>=') { $condition->setOperator('GE'); }
                        elsif ($op eq '<')  { $condition->setOperator('LT'); }
                        elsif ($op eq '<=') { $condition->setOperator('LE'); }
                        elsif ($op eq '!=') { $condition->setOperator('NE'); }
                        elsif ($op eq '~') { $condition->setOperator('Match'); }
                        elsif ($op eq '!~')
                        {
                            $condition->setOperator('NotMatch');
                        }
                        else
                        {
                            return http_failure(
                                http_code =>
                                  Apache2::Const::HTTP_BAD_REQUEST,    # 400
                                code    => '812',
                                message => "Invalid filter operator ($op).",
                            );
                        }
                        if ($subject)
                        {
                            $condition->setSubject(ucc($subject));
                            $value = ucc($value);
                        }
                        $condition->setValue($value);

                        # Add conjunctions
                        if ($delimiter =~ /,/)
                        {
                            $condition->setConjunction('And');
                        }
                        elsif ($delimiter =~ /\|/)
                        {
                            $condition->setConjunction('Or');
                        }

                        # Add grouping
                        if ($pregroup || $postgroup)
                        {
                            $condition->setGroup(
                                length($pregroup) - length($postgroup));
                        }

                        # Add condition to request
                        $request->addCondition($condition);
                    }
                    else
                    {
                        return http_failure(
                            http_code => Apache2::Const::HTTP_BAD_REQUEST, # 400
                            code      => '812',
                            message =>
                              "The filter condition is malformed ($filter).",
                        );
                    }
                }
                if ($filterString)
                {
                    return http_failure(
                        http_code => Apache2::Const::HTTP_BAD_REQUEST,     # 400
                        code      => '812',
                        message =>
                          "The filter condition contains unparsed characters ($filterString).",
                    );
                }
            }

            # Apply direct simple explicit conditions (==, <, >=, !=, ~, etc.)
            elsif ($key =~ /^([-\w]+)([<>~]|[=<>!]=|!~)$/)
            {
                my ($name, $op) = ($1, $2);
                my $value = $params{$key};

                my $condition = new MAM::Condition();
                $condition->setName(ucc($name));
                $condition->setValue($value);
                if    ($op eq '==') { $condition->setOperator('EQ'); }
                elsif ($op eq '>')  { $condition->setOperator('GT'); }
                elsif ($op eq '>=') { $condition->setOperator('GE'); }
                elsif ($op eq '<')  { $condition->setOperator('LT'); }
                elsif ($op eq '<=') { $condition->setOperator('LE'); }
                elsif ($op eq '!=') { $condition->setOperator('NE'); }
                elsif ($op eq '~')  { $condition->setOperator('Match'); }
                elsif ($op eq '!~')
                {
                    $condition->setOperator('NotMatch');
                }

                $request->addCondition($condition);
            }

            # Apply assignments
            elsif ($key eq 'update' || $key eq 'assign' || $key eq 'set')
            {
                my $updateString = $params{$key};

                unless ($method eq 'POST' || $method eq 'PATCH')
                {
                    return http_failure(
                        http_code => Apache2::Const::HTTP_BAD_REQUEST,    # 400
                        code      => '810',
                        message =>
                          "The $key parameter may only be used with the POST or PATCH http methods.",
                    );
                }

                foreach my $update (split /,/, $updateString)
                {
                    if ($update =~ /^([-\w]+)([-+]?=)([^\)]+)$/)
                    {
                        my ($name, $op, $value) = ($1, $2, $3);

                        my $assignment = new MAM::Assignment();
                        $assignment->setName(ucc($name));
                        if ($op eq '=') { $assignment->setOperator('Assign'); }
                        elsif ($op eq '+=') { $assignment->setOperator('Inc'); }
                        elsif ($op eq '-=') { $assignment->setOperator('Dec'); }
                        else
                        {
                            return http_failure(
                                http_code =>
                                  Apache2::Const::HTTP_BAD_REQUEST,    # 400
                                code    => '812',
                                message => "Invalid update operator ($op).",
                            );
                        }
                        $value = 'NULL' if $value eq 'null';
                        $assignment->setValue($value);

                        $request->addAssignment($assignment);
                    }
                    else
                    {
                        return http_failure(
                            http_code => Apache2::Const::HTTP_BAD_REQUEST, # 400
                            code      => '812',
                            message =>
                              "The update assignment is malformed ($update).",
                        );
                    }
                }
            }

            # Apply direct explicit assignments (@=, +=, -=)
            elsif ($key =~ /^([-\w]+)([-+@]=)$/)
            {
                my ($name, $op) = ($1, $2);
                my $value = $params{$key};

                unless ($method eq 'POST' || $method eq 'PATCH')
                {
                    return http_failure(
                        http_code => Apache2::Const::HTTP_BAD_REQUEST,    # 400
                        code      => '810',
                        message =>
                          "Assignment operators ($op) may only be used with the POST or PATCH http methods.",
                    );
                }

                my $assignment = new MAM::Assignment();
                $assignment->setName(ucc($name));
                if    ($op eq '@=') { $assignment->setOperator('Assign'); }
                elsif ($op eq '+=') { $assignment->setOperator('Inc'); }
                elsif ($op eq '-=') { $assignment->setOperator('Dec'); }
                $value = 'NULL' if $value eq 'null';
                $assignment->setValue($value);

                $request->addAssignment($assignment);
            }

            # Apply data properties
            elsif ($key eq 'data' || $key eq 'properties')
            {
                my $propertyString = $params{$key};

                unless ($action eq 'Charge'
                    || $action eq 'Quote'
                    || $action eq 'Reserve')
                {
                    return http_failure(
                        http_code => Apache2::Const::HTTP_BAD_REQUEST,    # 400
                        code      => '810',
                        message =>
                          "The $key parameter may not be used with the $action action.",
                    );
                }

                foreach my $property (split /,/, $propertyString)
                {
                    if ($property =~ /^([-\w]+)=([^\)]+)$/)
                    {
                        my ($name, $value) = ($1, $2);
                        $dataProperties{$name} = $value;
                    }
                    else
                    {
                        return http_failure(
                            http_code => Apache2::Const::HTTP_BAD_REQUEST, # 400
                            code      => '812',
                            message =>
                              "The data property is malformed ($property).",
                        );
                    }
                }
            }

            # Apply direct data properties (%=)
            elsif ($key =~ /^([-\w]+)%=$/)
            {
                my ($name) = ($1);
                my $value = $params{$key};

                unless ($action eq 'Charge'
                    || $action eq 'Quote'
                    || $action eq 'Reserve')
                {
                    return http_failure(
                        http_code => Apache2::Const::HTTP_BAD_REQUEST,    # 400
                        code      => '810',
                        message =>
                          "The Data property operator (%=) may not be used with the $action action.",
                    );
                }

                $dataProperties{$name} = $value;
            }

            # Apply constraint filter option
            elsif ($key eq 'constraint-filter')
            {
                my $constraintFilterString = $params{$key};

                # Uppercase the filter names
                $constraintFilterString =~ s/([-\w]+)(=.*)/ucc($1).$2/ge;

                $request->addOption('Filter', $constraintFilterString);
            }

            # Apply options
            elsif ($key eq 'options')
            {
                my $optionString = $params{$key};

                foreach my $option (split /,/, $optionString)
                {
                    if ($option =~ /^([-\w]+)(!:|[:!]?=)(.+)$/)
                    {
                        my ($name, $op, $value) = ($1, $2, $3);

                        my $option = new MAM::Option();
                        $option->setName(ucc($name));
                        if ($op eq '!=' || $op eq ':!')
                        {
                            $option->setOperator('Not');
                        }
                        if ($value eq 'true' || $value eq 'false')
                        {
                            $value = ucfirst($value);
                        }
                        if ($name =~ /^constraint$/i)
                        {
                            $value =~ s/([-\w]+)(=.*)/ucc($1).$2/ge;
                        }

                        $option->setValue($value);

                        $request->addOption($option);
                    }
                    elsif ($option =~ /^(!?)([^=]+)$/)
                    {
                        my ($op, $name) = ($1, $2);

                        my $option = new MAM::Option();
                        $option->setName(ucc($name));
                        if   ($op eq '!') { $option->setValue('False'); }
                        else              { $option->setValue('True'); }

                        $request->addOption($option);
                    }
                    else
                    {
                        return http_failure(
                            http_code => Apache2::Const::HTTP_BAD_REQUEST, # 400
                            code      => '812',
                            message =>
                              "The option parameter is malformed ($option).",
                        );
                    }
                }
            }

            # Apply explicit options
            elsif ($key =~ /^([-\w]+)(:=|:!)$/)
            {
                my ($name, $op) = ($1, $2);
                my $value = $params{$key};

                my $option = new MAM::Option();
                $option->setName(ucc($name));
                if ($op eq ':!') { $option->setOp('Not'); }

                if ($key =~ /^Filter$/i || $key =~ /^Constraint$/i)
                {
                    $value =~ s/([-\w]+)(=.*)/ucc($1).$2/ge;
                }
                if ($value eq 'true' || $value eq 'false')
                {
                    $value = ucfirst($value);
                }

                $option->setValue($value);

                $request->addOption($option);
            }

            # Treat all other parameters as options
            else
            {
                my $optionString = $params{$key};

                if ($optionString eq 'true' || $optionString eq 'false')
                {
                    $optionString = ucfirst($optionString);
                }
                if ($key =~ /^Constraint$/i)
                {
                    $optionString =~ s/([-\w]+)(=.*)/ucc($1).$2/ge;
                }
                $request->addOption(ucc($key), $optionString);
            }
        }

        # Apply body content
        if ($data)
        {
            # Use data as assignments if present in create or modify
            if ($action eq 'Create' || $action eq 'Modify')
            {
                if (ref($data) eq 'HASH')
                {
                    while (my ($name, $value) = each %$data)
                    {
                        if (JSON::is_bool($value))
                        {
                            $value = ucfirst(to_json($value));
                        }
                        elsif (ref($value))
                        {
                            $value = to_json($value);
                        }
                        elsif (! defined $value)
                        {
                            $value = 'NULL';
                        }
                        $request->addAssignment(ucc($name), $value);
                    }
                }
                else
                {
                    return http_failure(
                        http_code => Apache2::Const::HTTP_BAD_REQUEST,    # 400
                        code      => '810',
                        message =>
                          "Root data type must be a JSON object for the $action action.",
                    );
                }
            }

            # Convert request data to MAM data for charge, reserve, quote
            elsif ($action eq 'Charge'
                || $action eq 'Quote'
                || $action eq 'Reserve')
            {
                if (ref($data) eq 'HASH')
                {
                    while (my ($name, $value) = each %$data)
                    {
                        $value = to_json($value) if ref($value);
                        $dataProperties{$name} = $value;
                    }
                }
                else
                {
                    return http_failure(
                        http_code => Apache2::Const::HTTP_BAD_REQUEST,    # 400
                        code      => '810',
                        message =>
                          "Root data type must be a JSON object for the $action action.",
                    );
                }
            }

            # JSON data not supported for other actions
            else
            {
                return http_failure(
                    http_code => Apache2::Const::HTTP_BAD_REQUEST,        # 400
                    code      => '810',
                    message =>
                      "JSON data is not supported for the $action action.",
                );
            }
        }

        # Add data properties to request
        if (%dataProperties)
        {
            my $datum = new MAM::Datum('UsageRecord');
            while (my ($name, $value) = each %dataProperties)
            {
                $datum->setProperty(ucc($name), $value);
            }
            $request->addDatum($datum);
        }

        my $response = $request->getResponse();

        # Return 404 if asking for specific resource and it was not found
        #if (   scalar @nodes
        #    && @nodes == @primaryKeys
        #    && $response->getCount() == 0)
        #{
        #    return http_warning(
        #        http_code => Apache2::Const::HTTP_NOT_FOUND,        # 404
        #        code      => '142',
        #        message   => "The specified item does not exist",
        #    );
        #}

        # Handle combined partial-aggregation selections in frontend
        if ($combinedSelections)
        {
            $response =
              MAM::Client::aggregateResponse($response, $combinedSelections);
        }

        return convert_mam_to_http_response($response);
    }

    else
    {
        return http_failure(
            http_code => Apache2::Const::HTTP_BAD_REQUEST,                 # 400
            code      => '810',
            message   => "No object specified in path info ($path_info).",
        );
    }
}

# ----------------------------------------------------------------------------
# $ucname = ucc($name)
# ----------------------------------------------------------------------------

# To Upper Camel Case
# Converts from the form compound-name to CompoundName
sub ucc
{
    my ($name) = @_;

    return unless defined $name;

    $name = ucfirst $name;         # Uppercase first letter
    $name =~ s/-([a-z])/\U$1/g;    # Replace all -a with A

    return $name;
}

# ----------------------------------------------------------------------------
# $lhname = lhc($name)
# ----------------------------------------------------------------------------

# To Lower Hyphenated Case
# Converts from the form CompoundName to compound-name
sub lhc
{
    my ($name) = @_;

    return unless defined $name;

    $name = lcfirst $name;         # Lowercase first letter
    $name =~ s/([A-Z])/-\L$1/g;    # Replace all A with -a

    return $name;
}

1;
