#! /usr/bin/perl -w
################################################################################
#
# Allocation Summary Report
#
# File   :  allocation-summary
#
################################################################################

use strict;
use vars qw(@ARGV $format $log $quiet %supplement $time_division $verbose);
use lib qw(/opt/mam/lib /opt/mam/lib/perl5);
use Getopt::Long 2.24 qw(:config no_ignore_case);
use autouse 'Pod::Usage' => qw(pod2usage);
use MAM;
use Data::Dumper;

Main:
{
    $log->debug("Command line arguments: @ARGV");

    my $fallbackQOS = 'windfall';

    # Parse Command Line Arguments
    my ($active,  $account,   $allocation,   $class,      $cmd,
        $debug,   $endTime,   %filters,      $filterType, $full,
        $fund,    $group,     $help,         $machine,    $man,
        %members, $message,   $organization, $output,     $rc,
        $show,    $startTime, $user,         $version,
    );
    $verbose = 1;
    GetOptions(
        'A'             => \$active,
        'a=s'           => \$account,
        'c=s'           => \$class,
        'e=s'           => \$endTime,
        'f=i'           => \$fund,
        'filter=s'      => \%filters,
        'filter-type=s' => \$filterType,
        'g=s'           => \$group,
        'i=i'           => \$allocation,
        'm=s'           => \$machine,
        'o=s'           => \$organization,
        's=s'           => \$startTime,
        'u=s'           => \$user,
        'show=s'        => \$show,
        'full'          => \$full,
        'debug'         => \$debug,
        'help'          => \$help,
        'man'           => \$man,
        'quiet'         => \$quiet,
        'format=s'      => \$format,
    ) or pod2usage(2);

    # No arguments are allowed
    pod2usage(2) if @ARGV;

    # Display usage if necessary
    pod2usage(2) if $help;
    if ($man)
    {
        if ($< == 0)    # Cannot invoke perldoc as root
        {
            my $id = eval { getpwnam("nobody") };
            $id = eval { getpwnam("nouser") } unless defined $id;
            $id = -2                          unless defined $id;
            $<  = $id;
        }
        $> = $<;                         # Disengage setuid
        $ENV{PATH} = "/bin:/usr/bin";    # Untaint PATH
        delete @ENV{'IFS', 'CDPATH', 'ENV', 'BASH_ENV'};
        if ($0 =~ /^([-\/\w\.]+)$/) { $0 = $1; }    # Untaint $0
        else { die "Illegal characters were found in \$0 ($0)\n"; }
        pod2usage(-exitstatus => 0, -verbose => 2);
    }

    # Augment the path to find MAM command line clients as necessary
    #$ENV{PATH} = "/opt/mam/bin:/opt/mam/sbin:$ENV{PATH}";

    #########################################################################
    # Query Funds to determine applicable funds and constraints
    #########################################################################

    $cmd = 'mam-list-funds';
    $cmd .= " -a $account"               if defined $account;
    $cmd .= " -c $class"                 if defined $class;
    $cmd .= " -f $fund"                  if defined $fund;
    $cmd .= " -g $group"                 if defined $group;
    $cmd .= " -m $machine"               if defined $machine;
    $cmd .= " -o $organization"          if defined $organization;
    $cmd .= " -u $user"                  if defined $user;
    $cmd .= " --filter-type $filterType" if defined $filterType;
    foreach my $key (keys %filters)
    {
        $cmd .= " --filter $key=$filters{$key}";
    }
    $cmd .= ' --show=Id,Constraints --format=raw -q';
    print STDERR Data::Dumper->Dump([$cmd], ['cmd']), "\n" if $debug;
    $output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
    print STDERR Data::Dumper->Dump([$output], ['output']), "\n" if $debug;
    $rc      = $? >> 8;
    $message = "Subcommand ($cmd) failed with rc=$rc";
    $message .= ": $output" if $output;
    die $message if $rc;

    my %constraintNames = ();
    my %fundConstraints = ();
    foreach my $line (split /\n/, $output)
    {
        my ($id, $constraints) = split /\|/, $line;
        %{$fundConstraints{$id}} = ();
        foreach my $constraint (split /,/, $constraints)
        {
            my ($name, $value) = split /=/, $constraint;
            $constraintNames{$name} //= 1;
            if (defined $fundConstraints{$id}->{$name})
            {
                $fundConstraints{$id}->{$name} .= ",$value";
            }
            else { $fundConstraints{$id}->{$name} = $value; }
        }
    }
    print STDERR Data::Dumper->Dump([\%constraintNames], ['constraintNames']),
      "\n"
      if $debug;
    print STDERR Data::Dumper->Dump([\%fundConstraints], ['fundConstraints']),
      "\n"
      if $debug;

    #########################################################################
    # Build up selections based on available constraints
    #########################################################################

    my @constraints = sort keys %constraintNames;
    print STDERR Data::Dumper->Dump([\@constraints], ['constraints']), "\n"
      if $debug;
    my $constraintSelections = join '', map { $_ . ',' } @constraints;

    # Use a hard-coded selection list if no --show option specified
    unless ($show)
    {
        if ($full)
        {
            $show =
              "${constraintSelections}Allocation,Fund,StartTime,EndTime,InitialDeposit,Adjustments,Allocated,Used,Remaining,PercentUsed";
        }
        elsif ($active)
        {
            $show = "${constraintSelections}Allocated,Balance,PercentUsed";
        }
        else
        {
            $show =
              "${constraintSelections}Allocated,Used,Remaining,PercentUsed";
        }
    }
    my @selections = split(/,/, $show);
    print STDERR Data::Dumper->Dump([\@selections], ['selections']), "\n"
      if $debug;

    #########################################################################
    # Query Allocations for qualifying funds and intersecting time period
    #########################################################################

    # Specify the selection and date range. Post-filter by fund. Get xml back
    $cmd = 'mam-list-allocations';
    $cmd .= ' -A' if ($active || (! defined $startTime && ! defined $endTime));
    $cmd .= " -e '$endTime'"   if defined $endTime;
    $cmd .= " -f $fund"        if defined $fund;
    $cmd .= " -i $allocation"  if defined $allocation;
    $cmd .= " -s '$startTime'" if defined $startTime;
    my @allocationSelections = intersection(
        [   'FundName',       'Active',
            'InitialDeposit', 'Adjustments',
            'Allocated',      'Balance',
            'CreditLimit',    'Capacity',
            'Used',           'PercentUsed',
            'Remaining',      'PercentRemaining',
            'Reserved',       'Effective',
            'Available',      'Description',
            'CreationTime'
        ],
        \@selections
    );
    unshift @allocationSelections, 'Id', 'Fund', 'StartTime', 'EndTime';
    print STDERR Data::Dumper->Dump([\@allocationSelections],
        ['allocationSelections']), "\n"
      if $debug;
    $cmd .= " --show=" . join ',', @allocationSelections;
    $cmd .= ' --format=raw -q';
    print STDERR Data::Dumper->Dump([$cmd], ['cmd']), "\n" if $debug;
    $output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
    print STDERR Data::Dumper->Dump([$output], ['output']), "\n" if $debug;
    $rc      = $? >> 8;
    $message = "Subcommand ($cmd) failed with rc=$rc";
    $message .= ": $output" if $output;
    die $message if $rc;

    my %allocations                 = ();
    my $earliestAllocationStartTime = toMFT('Infinity');
    my $latestAllocationEndTime     = toMFT('-Infinity');
    foreach my $line (split /\n/, $output)
    {
        my @values    = split /\|/, $line;
        my $id        = $values[0];
        my $fund      = $values[1];
        my $startTime = toMFT($values[2]);
        my $endTime   = toMFT($values[3]);

        # Restrict to the filtered funds
        next unless exists $fundConstraints{$fund};

        # Compute widest range of allocation expanse
        $earliestAllocationStartTime = $startTime
          if $startTime < $earliestAllocationStartTime;
        $latestAllocationEndTime = $endTime
          if $endTime > $latestAllocationEndTime;

        $allocations{$id}->{'Allocation'} = $id;

        for (my $i = 1; $i < @allocationSelections; $i++)
        {
            $allocations{$id}->{$allocationSelections[$i]} = $values[$i] // '';
        }

        # Merge in fund constraints
        foreach my $name (@constraints)
        {
            $allocations{$id}->{$name} = $fundConstraints{$fund}->{$name} // '';
        }
    }
    print STDERR Data::Dumper->Dump([\%allocations], ['allocations']), "\n"
      if $debug;
    print STDERR Data::Dumper->Dump(
        [$earliestAllocationStartTime],
        ['earliestAllocationStartTime']
      ),
      "\n"
      if $debug;
    print STDERR Data::Dumper->Dump(
        [$latestAllocationEndTime],
        ['latestAllocationEndTime']
      ),
      "\n"
      if $debug;

    #########################################################################
    # Query Transactions for ancillary information
    #########################################################################

    my @transactionOrientedSelections = intersection(
        [   'FallbackCharged',        'FallbackProcHours',
            'Jobs',                   'JobNodeHours',
            'PercentReservationUsed', 'JobProcHours',
            'ReservationIdleCharged', 'ReservationIdleProcHours',
            'ReservationProcHours',   'Reservations',
        ],
        \@selections
    );

    if (@transactionOrientedSelections)
    {
        $cmd = 'mam-list-transactions';
        $cmd .= " -e $latestAllocationEndTime";
        $cmd .= " -s $earliestAllocationStartTime";
        $cmd .= ' --where name=Fund op=NE value=NULL';
        my @transactionSelections = ('Allocation', 'Fund');
        my ($deltaSpecified,       $idleProcessorSecondsSpecified,
            $instanceSpecified,    $nodesSpecified,
            $procSecondsSpecified, $reservedProcessorSecondsSpecified,
            $qosSpecified,         $typeSpecified
        );
        foreach my $selection (@transactionOrientedSelections)
        {

            if ($selection eq 'FallbackCharged')
            {
                unless ($deltaSpecified)
                {
                    push @transactionSelections, 'Delta';
                    $deltaSpecified = 1;
                }
                unless ($qosSpecified)
                {
                    push @transactionSelections, 'QualityOfService';
                    $qosSpecified = 1;
                }
                unless ($typeSpecified)
                {
                    push @transactionSelections, 'Type';
                    $typeSpecified = 1;
                }
            }
            elsif ($selection eq 'FallbackProcHours')
            {
                unless ($procSecondsSpecified)
                {
                    push @transactionSelections, 'ProcSeconds';
                    $procSecondsSpecified = 1;
                }
                unless ($qosSpecified)
                {
                    push @transactionSelections, 'QualityOfService';
                    $qosSpecified = 1;
                }
                unless ($typeSpecified)
                {
                    push @transactionSelections, 'Type';
                    $typeSpecified = 1;
                }
            }
            elsif ($selection eq 'Jobs' || $selection eq 'Reservations')
            {
                unless ($instanceSpecified)
                {
                    push @transactionSelections, 'Instance';
                    $instanceSpecified = 1;
                }
                unless ($typeSpecified)
                {
                    push @transactionSelections, 'Type';
                    $typeSpecified = 1;
                }
            }
            elsif ($selection eq 'JobNodeHours')
            {
                push @transactionSelections, 'NodeSeconds';
                unless ($typeSpecified)
                {
                    push @transactionSelections, 'Type';
                    $typeSpecified = 1;
                }
            }
            elsif ($selection eq 'JobProcHours'
                || $selection eq 'ReservationProcHours')
            {
                unless ($procSecondsSpecified)
                {
                    push @transactionSelections, 'ProcSeconds';
                    $procSecondsSpecified = 1;
                }
                unless ($typeSpecified)
                {
                    push @transactionSelections, 'Type';
                    $typeSpecified = 1;
                }
            }
            elsif ($selection eq 'ReservationIdleProcHours')
            {
                unless ($idleProcessorSecondsSpecified)
                {
                    push @transactionSelections, 'IdleProcessorSeconds';
                    $idleProcessorSecondsSpecified = 1;
                }
                unless ($typeSpecified)
                {
                    push @transactionSelections, 'Type';
                    $typeSpecified = 1;
                }
            }
            elsif ($selection eq 'PercentReservationUsed')
            {
                unless ($idleProcessorSecondsSpecified)
                {
                    push @transactionSelections, 'IdleProcessorSeconds';
                    $idleProcessorSecondsSpecified = 1;
                }
                unless ($reservedProcessorSecondsSpecified)
                {
                    push @transactionSelections, 'ReservedProcessorSeconds';
                    $reservedProcessorSecondsSpecified = 1;
                }
                unless ($typeSpecified)
                {
                    push @transactionSelections, 'Type';
                    $typeSpecified = 1;
                }
            }
            elsif ($selection eq 'ReservationIdleCharged')
            {
                unless ($deltaSpecified)
                {
                    push @transactionSelections, 'Delta';
                    $deltaSpecified = 1;
                }
                unless ($typeSpecified)
                {
                    push @transactionSelections, 'Type';
                    $typeSpecified = 1;
                }
            }
            elsif ($selection eq 'ReservationProcHours')
            {
                unless ($reservedProcessorSecondsSpecified)
                {
                    push @transactionSelections, 'ReservedProcessorSeconds';
                    $reservedProcessorSecondsSpecified = 1;
                }
                unless ($typeSpecified)
                {
                    push @transactionSelections, 'Type';
                    $typeSpecified = 1;
                }
            }
        }
        print STDERR Data::Dumper->Dump(
            [\@transactionSelections],
            ['transactionSelections']
          ),
          "\n"
          if $debug;
        $cmd .= " --show=" . join ',', @transactionSelections;
        $cmd .= ' --format=raw -q';
        print STDERR Data::Dumper->Dump([$cmd], ['cmd']), "\n" if $debug;
        $output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
        print STDERR Data::Dumper->Dump([$output], ['output']), "\n" if $debug;
        $rc      = $? >> 8;
        $message = "Subcommand ($cmd) failed with rc=$rc";
        $message .= ": $output" if $output;
        die $message if $rc;

        my @transactions = ();
        foreach my $line (split /\n/, $output)
        {
            my @values     = split /\|/, $line;
            my $allocation = $values[0];
            my $fund       = $values[1];

          # Restrict to transactions referencing an eligible fund and allocation
            next unless defined $fund       && exists $fundConstraints{$fund};
            next unless defined $allocation && exists $allocations{$allocation};

            my %transaction = ();
            for (my $i = 0; $i < @transactionSelections; $i++)
            {
                $transaction{$transactionSelections[$i]} = $values[$i] // '';
            }
            push @transactions, \%transaction;
        }
        print STDERR Data::Dumper->Dump([\@transactions], ['transactions']),
          "\n"
          if $debug;

        # Iterate over transactions to populate allocation hash
        foreach my $transaction (@transactions)
        {
            my %transaction = %{$transaction};
            my $allocation  = $transaction{'Allocation'};

            foreach my $selection (@transactionOrientedSelections)
            {

                if ($selection eq 'FallbackCharged')
                {
                    if (   defined $fallbackQOS
                        && $transaction{'QualityOfService'} eq $fallbackQOS
                        && $transaction{'Type'} eq 'Job'
                        && $transaction{'Delta'} ne '')
                    {
                        if (defined $allocations{$allocation}
                            ->{'FallbackCharged'})
                        {
                            $allocations{$allocation}->{'FallbackCharged'} -=
                              $transaction{'Delta'};
                        }
                        else
                        {
                            $allocations{$allocation}->{'FallbackCharged'} =
                              0 - $transaction{'Delta'};
                        }
                    }
                }
                elsif ($selection eq 'FallbackProcHours')
                {
                    if (   defined $fallbackQOS
                        && $transaction{'QualityOfService'} eq $fallbackQOS
                        && $transaction{'Type'} eq 'Job'
                        && $transaction{'ProcSeconds'} ne '')
                    {
                        if (defined $allocations{$allocation}
                            ->{'FallbackProcSeconds'})
                        {
                            $allocations{$allocation}->{'FallbackProcSeconds'}
                              += $transaction{'ProcSeconds'};
                        }
                        else
                        {
                            $allocations{$allocation}->{'FallbackProcSeconds'}
                              = 0 + $transaction{'ProcSeconds'};
                        }
                    }
                }
                elsif ($selection eq 'Jobs')
                {
                    if (   $transaction{'Type'} eq 'Job'
                        && $transaction{'Instance'} ne '')
                    {
                        $allocations{$allocation}->{'Jobs'}
                          ->{$transaction{'Instance'}}++;
                    }
                }
                elsif ($selection eq 'Reservations')
                {
                    if (   $transaction{'Type'} eq 'Reservation'
                        && $transaction{'Instance'} ne '')
                    {
                        $allocations{$allocation}->{'Reservations'}
                          ->{$transaction{'Instance'}}++;
                    }
                }
                elsif ($selection eq 'JobNodeHours')
                {
                    if (   $transaction{'Type'} eq 'Job'
                        && $transaction{'NodeSeconds'} ne '')
                    {
                        if (defined $allocations{$allocation}
                            ->{'JobNodeSeconds'})
                        {
                            $allocations{$allocation}->{'JobNodeSeconds'} +=
                              $transaction{'NodeSeconds'};
                        }
                        else
                        {
                            $allocations{$allocation}->{'JobNodeSeconds'} =
                              0 + $transaction{'NodeSeconds'};
                        }
                    }
                }
                elsif ($selection eq 'JobProcHours')
                {
                    if (   $transaction{'Type'} eq 'Job'
                        && $transaction{'ProcSeconds'} ne '')
                    {
                        if (defined $allocations{$allocation}
                            ->{'JobProcSeconds'})
                        {
                            $allocations{$allocation}->{'JobProcSeconds'} +=
                              $transaction{'ProcSeconds'};
                        }
                        else
                        {
                            $allocations{$allocation}->{'JobProcSeconds'} =
                              0 + $transaction{'ProcSeconds'};
                        }
                    }
                }
                elsif ($selection eq 'ReservationIdleCharged')
                {
                    if (   $transaction{'Type'} eq 'Reservation'
                        && $transaction{'Delta'} ne '')
                    {
                        if (defined $allocations{$allocation}
                            ->{'ReservationIdleCharged'})
                        {
                            $allocations{$allocation}
                              ->{'ReservationIdleCharged'} -=
                              $transaction{'Delta'};
                        }
                        else
                        {
                            $allocations{$allocation}
                              ->{'ReservationIdleCharged'} =
                              0 - $transaction{'Delta'};
                        }
                    }
                }
                elsif ($selection eq 'ReservationIdleProcHours'
                    || $selection eq 'PercentReservationUsed')
                {
                    if (   $transaction{'Type'} eq 'Reservation'
                        && $transaction{'IdleProcessorSeconds'} ne ''
                        && $transaction{'Instance'} ne '')
                    {
                        $allocations{$allocation}
                          ->{'ReservationIdleProcSeconds'}
                          ->{$transaction{'Instance'}} =
                          0 + $transaction{'IdleProcessorSeconds'};
                    }
                }
                elsif ($selection eq 'ReservationProcHours'
                    || $selection eq 'PercentReservationUsed')
                {
                    if (   $transaction{'Type'} eq 'Reservation'
                        && $transaction{'ReservedProcessorSeconds'} ne ''
                        && $transaction{'Instance'} ne '')
                    {
                        $allocations{$allocation}->{'ReservationProcSeconds'}
                          ->{$transaction{'Instance'}} =
                          0 + $transaction{'ReservedProcessorSeconds'};
                    }
                }
            }
        }
    }
    print STDERR Data::Dumper->Dump(
        [\%allocations], ['allocations_with_transaction_info_merged_in']
      ),
      "\n"
      if $debug;

    #########################################################################
    # Build up the response data from the %allocations hash
    #########################################################################

    my @specifiedConstraints = intersection(\@constraints, \@selections);
    print STDERR Data::Dumper->Dump([\@specifiedConstraints],
        ['specifiedConstraints']), "\n"
      if $debug;

    # Subroutine to sort by specified constraints, then allocation id
    sub bySpecifiedConstraints
    {
        foreach my $constraint (@specifiedConstraints)
        {
            my $cmp =
              $allocations{$a}->{$constraint}
              cmp $allocations{$b}->{$constraint};
            return $cmp if $cmp;
        }
        return $a <=> $b;
    }

    # Build up a new data element from the %allocations hash
    my $newData     = new XML::LibXML::Element("Data");
    my %totals      = ();
    my $rowCount    = 0;
    foreach my $id (sort bySpecifiedConstraints keys %allocations)
    {
        my $newRow = new XML::LibXML::Element("Allocation");
        $rowCount++;

        # Walk through selections
        foreach my $selection (@selections)
        {
            my $newCol = new XML::LibXML::Element($selection);

            if ($selection eq 'PercentReservationUsed')
            {
                my ($idleProcessorSeconds, $reservedProcessorSeconds) = (0, 0);
                $idleProcessorSeconds += $_
                  for
                  values %{$allocations{$id}->{'ReservationIdleProcSeconds'}};
                $reservedProcessorSeconds += $_
                  for values %{$allocations{$id}->{'ReservationProcSeconds'}};
                if (   defined $idleProcessorSeconds
                    && defined $idleProcessorSeconds)
                {
                    if ($reservedProcessorSeconds == 0)
                    {
                        $newCol->appendText(sprintf("%.2f", 100));
                    }
                    else
                    {
                        $newCol->appendText(
                            sprintf(
                                "%.2f",
                                (   $reservedProcessorSeconds -
                                      $idleProcessorSeconds
                                ) * 100 / $reservedProcessorSeconds
                            )
                        );
                    }
                }
                else { $newCol->appendText(''); }
            }
            elsif ($selection eq 'FallbackProcHours'
                || $selection eq 'JobNodeHours'
                || $selection eq 'JobProcHours')
            {
                (my $morphedSelection = $selection) =~ s/Hours/Seconds/;
                my $procSeconds = $allocations{$id}->{$morphedSelection};
                if (defined $procSeconds)
                {
                    my $procHours = sprintf("%.2f", $procSeconds / 3600);
                    $newCol->appendText($procHours);
                    $totals{$selection} += $procHours;
                }
                else { $newCol->appendText(''); }
            }
            elsif ($selection eq 'ReservationIdleProcHours'
                || $selection eq 'ReservationProcHours')
            {
                (my $morphedSelection = $selection) =~ s/Hours/Seconds/;
                my $procSeconds = 0;
                $procSeconds += $_
                  for values %{$allocations{$id}->{$morphedSelection}};
                if (defined $procSeconds)
                {
                    my $procHours = sprintf("%.2f", $procSeconds / 3600);
                    $newCol->appendText($procHours);
                    $totals{$selection} += $procHours;
                }
                else { $newCol->appendText(''); }
            }
            elsif ($selection eq 'Jobs' || $selection eq 'Reservations')
            {
                my $count = scalar keys %{$allocations{$id}->{$selection}};
                $newCol->appendText($count);
                $totals{$selection} += $count;
            }
            else
            {
                my $value = $allocations{$id}->{$selection};
                if (defined $value)
                {
                    $newCol->appendText($value);
                    if (   $selection eq 'Adjustments'
                        || $selection eq 'Allocated'
                        || $selection eq 'Available'
                        || $selection eq 'Balance'
                        || $selection eq 'Capacity'
                        || $selection eq 'CreditLimit'
                        || $selection eq 'Effective'
                        || $selection eq 'FallbackCharged'
                        || $selection eq 'InitialDeposit'
                        || $selection eq 'Remaining'
                        || $selection eq 'ReservationIdleCharged'
                        || $selection eq 'Reserved'
                        || $selection eq 'Used')
                    {
                        $totals{$selection} += $value;
                    }
                }
                else { $newCol->appendText(''); }
            }
            # Add the new column to the row
            $newRow->appendChild($newCol);
        }
        # Append the row into the new data element
        $newData->appendChild($newRow);
    }
    # Add in totals unless csv or raw format
    unless ($format eq 'csv' || $format eq 'raw' || $quiet)
    {
        $format = 'total';
        my $totalRow = new XML::LibXML::Element("Allocation");

        # Walk through selections
        foreach my $selection (@selections)
        {
            my $totalCol = new XML::LibXML::Element($selection);
            $totalCol->appendText($totals{$selection} // '');
            $totalRow->appendChild($totalCol);
        }
        $newData->appendChild($totalRow);
    }

    #print STDERR $newData->toString(), "\n" if $debug;

    # Create a new response with the merged data
    my $response = new MAM::Response();
    $response->setDataElement($newData);

    #########################################################################
    # Print out the response
    #########################################################################

    # Print out the response
    &MAM::Client::displayResponse($response);
}

##############################################################################

__END__

=head1 NAME

allocation-summary - report allocation status

=head1 SYNOPSIS

B<allocation-summary> [B<-A> | {[B<-s> I<start_time>] [B<-e> I<end_time>]}] [B<-a> I<account_name>] [B<-u> I<user_name>] [B<-g> I<group_name>] [B<-o> I<organization_name>] [B<-c> I<class_name>] [B<-m> I<machine_name>] [B<-f> I<fund_id>] [B<-i> I<allocation_id>] [B<--filter> I<filter_name>=I<filter_value>]... [B<--filter-type> ExactMatch|Exclusive|NonExclusive] [B<--full>] [B<--show> I<attribute_name>,...] [B<--format> csv|raw|standard] [B<--debug>] [B<--help>] [B<--man>] [B<--quiet>]

=head1 DESCRIPTION

B<allocation-summary> is used to summarize the status of allocations. By default it will display applicable fund constraints and allocation and usage details for the currently active allocations. If a timeframe is given, both active and expired allocations overlapping the specified time period will be displayed. Filters can be specified to narrow the set of allocations displayed. Specific fields can be specified for display. A totals line is displayed unless using the csv or raw format.

=head1 OPTIONS

=over 4

=item B<-a> I<account_name>

displays only allocations usable by the specified account

=item B<-A>

displays only active allocations. Additionally, the default displayed fields include Balance instead of Remaining and Used.

=item B<-c> I<class_name>

displays only allocations usable by the specified class

=item B<-e> I<end_time>

displays only allocations that start before the specified end time in the format: YYYY-MM-DD[ hh:mm:ss]|-Infinity|Infinity|Now

=item B<-f> I<fund_id>

displays only the allocations associated with the specified fund

=item B<--filter> I<filter_name>=I<filter_value>

displays allocations whose fund constraints comply with the specified filters. For example mam-list-funds --filter User=amy will display funds usable by the user amy. Multiple filter options may be specified which are logically anded together.

=item B<--filter-type> ExactMatch|Exclusive|NonExclusive

selects the filtering type. If the exact-match filter type is used, a fund will only be matched if the specified filters exactly match the fund constraints. If the exclusive filter type is used, a fund will only be matched if the specified filters meet all constraints (not only must the filters be a non-conflicting superset of the fund constraints, but all constraint association dependencies must also be satisfied). If the non-exclusive filter type is used, a fund will be matched as long as the specified filters do not conflict with the constraints. The default filter type is NonExclusive.

=item B<-g> I<group_name>

displays only allocations usable by the specified group

=item [B<-i>] I<allocation_id>

displays only the allocation with the given id

=item B<-m> I<machine_name>

displays only allocations usable by the specified machine

=item B<-o> I<organization_name>

displays only allocations usable by the specified organization

=item B<-s> I<start_time>

displays only allocations that end after the specified start time in the format: YYYY-MM-DD[ hh:mm:ss]|-Infinity|Infinity|Now

=item B<-u> I<user_name>

displays only allocations usable by the specified user

=item B<--full>

displays additional attributes. This can be configured by customizing this script.

=item B<--show> I<attribute_name>[,I<attribute_name>...]

displays only the specified attributes in the given order. Valid attributes include: Allocation, Fund, FundName, Active, StartTime, EndTime, InitialDeposit, Adjustments, Allocated, CreditLimit, Capacity, Used, PercentUsed, Remaining, PercentRemaining, Balance, Reserved, Effective, Available, Description, CreationTime, Jobs, JobProcHours, JobNodeHours, FallbackCharged, FallbackProcHours, Reservations, ReservationIdleCharged, ReservationIdleProcHours, ReservationProcHours and PercentReservationUsed. Additionally, fund constraint names may be specified, such as Account, User, Group, Organization or Machine.

=over 4

=item Active

boolean indicating whether this allocation is active or not

=item Adjustments

total of subsequent adjustments to the initial deposit via deposits, withdrawals or transfers (Allocated - InitialDeposit)

=item Allocated

adjusted allocation. This value stores the effective allocated amount based on the initial deposit and subsequent allocation adjustments via deposits, withdrawals or transfers.

=item Available

amount currently available for charging. If the allocation is active, this is Remaining - Reserved + CreditLimit. If the allocation is inactive, this is zero.

=item Balance

active allocation balance. If the allocation is active, this is the remaining allocation amount (Remaining). If the allocation is inactive, this is zero.

=item Capacity

total expendable amount (Allocated + CreditLimit)

=item CreationTime

time this allocation was created

=item CreditLimit

credit limit. Determines how far in the negative this allocation is permitted to be used (enforced in quotes and liens)

=item Description

allocation description

=item Effective

effective balance not blocked by liens. If the allocation is active, this is Remaining - Reserved. If the allocation is inactive, this is zero.

=item EndTime

time this allocation becomes inactive

=item FallbackCharged

amount charged to this allocation for the fallback quality of service. The $fallbackQOS variable must be configured in this script.

=item FallbackProcHours

processor hours used against this allocation for the fallback quality of service. The $fallbackQOS variable must be configured in this script.

=item Fund

fund id

=item FundName

fund name

=item Id

allocation id

=item InitialDeposit

amount of the first deposit into this allocation

=item Jobs

number of jobs charged against this allocation

=item JobProcHours

number of processor hours used associated with job charges against this allocation

=item JobProcHours

number of node hours used associated with job charges against this allocation

=item PercentRemaining

percentage of allocation remaining (Remaining * 100 / Capacity)

=item PercentReservationUsed

percentage of reservation processor hours used that were charged against this allocation ((ReservationProcHours - ReservationIdleProcHours) * 100 / ReservationProcHours)

=item PercentUsed

percentage of allocation used (Used * 100 / Capacity)

=item Remaining

remaining allocation amount

=item ReservationIdleCharged

amount charged to this allocation for idle reservation cycles

=item ReservationIdleProcHours

idle processor hours for reservations charged against this allocation

=item ReservationProcHours

total (idle and busy) processor hours for reservations charged against this allocation

=item Reservations

number of reservations charged against this allocation

=item Reserved

sum of active lien amounts against this allocation

=item StartTime

time this allocation becomes active

=item Used

amount used from this allocation (Allocated - Remaining)

=back

=item B<--format> I<output_format>
 
data output format. Valid values include standard, raw and csv.

=item B<--quiet>

suppress headers and totals

=item B<--debug>

log debug info to screen

=item B<--help>

brief help message

=item B<--man>

full documentation

=back

=head1 AUTHOR

Scott Jackson, scottmo@adaptivecomputing.com

=cut

