#!/usr/bin/perl -w
use strict;

use lib qw(/opt/mam/lib /opt/mam/lib/perl5);
use Config::Tiny;
use DBI;
use MAM;
use MAM::Global;

my ($cmd, $output, $rc, $sql, $count);
my ($successes, $failures, $ignores) = (0, 0, 0);
my @messages   = ();
$ENV{PATH} = "/opt/mam/bin:/opt/mam/sbin:$ENV{PATH}";
my $now = time;

# Read the Config File
my $configFile = "/opt/mam/etc/mam-server.conf";
my $configTiny = Config::Tiny->new;
$configTiny = Config::Tiny->read($configFile)
  or die(
    "Unable to open config file ($configFile) for reading: $Config::Tiny::errstr.\nYou may need to set the \$configFile variable in this script.\n"
  );
$serverConfig = $configTiny->{_};
my $datasource = $serverConfig->{'database.datasource'};
my $dbuser     = $serverConfig->{'database.user'} || (getpwuid($<))[0];
my $dbpasswd   = $serverConfig->{'database.password'};

################################################################################
# Display warning and prompt to proceed
################################################################################

my $warning = <<"ENDOFWARNING";
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! IMPORTANT !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

This script will attempt to convert an existing Gold Allocation Manager 2.2 database (using the datasource ($datasource) referenced in /opt/mam/etc/mam-server.conf) into a Moab Accounting Manager 7.0 database. It will make additions, modifications and deletions to your Gold Allocation Manager 2.2 schema and data to match the new Moab Accounting Manager 7.0 schema and data values.

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! BACKUP YOUR DATA !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

You should MAKE A BACKUP of your Gold Allocation Manager 2.2 database before running this script. If you proceed without a backup, a failure in this script could irreversibly corrupt the integrity and data in your Gold Allocation Manager 2.2 database.

ENDOFWARNING

print $warning;
print "Do you want to proceed? [y/N] ";
my $answer = <STDIN>;
exit 0 unless ($answer =~ /^y(es)?$/i);

# ----------------------------------------------------------------------------
# update($description, $test, $comparison, $action);
# ----------------------------------------------------------------------------

# Update the object metadata via mam-shell
sub update
{
    my ($description, $test, $comparison, $action) = @_;
    print "\n***$description***\n";
    if (chomp($_ = qx{mam-shell -q $test}) && eval "\$_ $comparison 0")
    {
        $cmd    = "mam-shell -v $action";
        $output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
        $rc     = $? >> 8;
        print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
        if ($rc)
        {
            $failures++;
            my $message = "Subcommand ($cmd) failed with rc=$rc";
            $message .= ": $output" if $output;
            warn($message);
            push @messages, $message;
        }
        else
        {
            $successes++;
        }
    }
    else
    {
        print "Already applied\n";
        $ignores++;
    }
}

################################################################################
# Perform Precursor Base Changes
################################################################################

print "\n***Creating new Transaction Key Attribute***\n";
if (
    chomp(
        $_ =
          qx{mam-shell -q Attribute Query Object==Transaction Name==Key Show:=\"\\\"Count(Name)\\\"\"}
    )
    && $_ == 0
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});
    my $sth;

    $sql = "ALTER TABLE g_transaction ADD COLUMN g_key varchar(1024)";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_transaction_log ADD COLUMN g_key varchar(1024)";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql =
      "UPDATE g_key_generator set g_next_id=g_next_id+1 WHERE g_name='Request'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sth = $dbh->prepare(
        "SELECT g_next_id-1 FROM g_key_generator WHERE g_name='Request'");
    $sth->execute();
    my ($requestId) = $sth->fetchrow_array();
    $sql =
      "UPDATE g_key_generator set g_next_id=g_next_id+1 WHERE g_name='Transaction'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sth = $dbh->prepare(
        "SELECT g_next_id-1 FROM g_key_generator WHERE g_name='Transaction'");
    $sth->execute();
    my ($transactionId) = $sth->fetchrow_array();

    if (defined $requestId && defined $transactionId)
    {
        $sql =
          qq{INSERT INTO g_attribute (g_object, g_name, g_data_type, g_primary_key, g_required, g_fixed, g_values, g_default_value, g_sequence, g_hidden, g_description, g_deleted, g_creation_time, g_modification_time, g_request_id, g_transaction_id) VALUES ('Transaction', 'Key', 'String', 'False', 'False', 'False', NULL, NULL, 50, 'False', 'Object Primary Key Value', 'False', $now, $now, $requestId, $transactionId)};
        print "Executing sql: $sql\n";
        $count = $dbh->do($sql);
        print "Updated $count lines\n";
        $sql =
          qq{INSERT INTO g_transaction (g_id, g_object, g_action, g_actor, g_name, g_child, g_count, g_details, g_description, g_deleted, g_creation_time, g_modification_time, g_request_id, g_transaction_id, g_job_id, g_amount, g_delta, g_user, g_project, g_machine, g_account, g_allocation, g_key) VALUES (1, 'Attribute', 'Create', 'scottmo', NULL, NULL, 1, 'DataType=String,Deleted=False,Description="Object Primary Key Value",Fixed=False,Hidden=False,NoRefresh:=True,Object=Transaction,PrimaryKey=False,Required=False,Sequence=50', NULL, 'False', $now, $now, 1, 1, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 'Key')};
        print "Executing sql: $sql\n";
        $count = $dbh->do($sql);
        print "Inserted $count lines\n";
    }

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

################################################################################
# Create New Base Objects and Attributes
################################################################################

update(
    q{Creating new Object DefaultValue Attribute},
    q{Attribute Query Object==Object Name==DefaultValue Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Object Name=DefaultValue DataType=String Description="\"Default Value for this object\""}
);

update(
    q{Creating new Object AutoGen Attribute},
    q{Attribute Query Object==Object Name==AutoGen Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Object Name=AutoGen DataType=Boolean DefaultValue=False Description="\"Should this object be Auto-Generated?\""}
);

update(
    q{Modifying Objects to set AutoGen to False},
    q{Object Query AutoGen==False Show:="\"Count(Name)\""},
    q{==},
    q{Object Modify AutoGen=False NoRefresh:=True}
);

################################################################################
# Create New Accounting Objects and Attributes
################################################################################

update(
    q{Creating new UsageRecord Object},
    q{Object Query Name==UsageRecord Show:="\"Count(Name)\""},
    q{==},
    q{Object Create Name=UsageRecord Description="\"Usage Record\""}
);

update(
    q{Creating new UsageRecord Id Attribute},
    q{Attribute Query Object==UsageRecord Name==Id Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Id PrimaryKey=True DataType=AutoGen Description="\"Usage Record Id\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Type Attribute},
    q{Attribute Query Object==UsageRecord Name==Type Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Type DataType=String Description="\"Usage Record Type\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Instance Attribute},
    q{Attribute Query Object==UsageRecord Name==Instance Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Instance DataType=String Description="\"Instance Name\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Charge Attribute},
    q{Attribute Query Object==UsageRecord Name==Charge Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Charge DataType=Currency DefaultValue=0 Description="\"Total Amount Charged for Usage\""}
);

update(
    q{Creating new UsageRecord Stage Attribute},
    q{Attribute Query Object==UsageRecord Name==Stage Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Stage DataType=String Values=\"\(Charge,Create,Quote,Reserve\)\" Description="\"Latest Usage Stage\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Quote Attribute},
    q{Attribute Query Object==UsageRecord Name==Quote Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Quote DataType=Integer Values=@Quote Description="\"Quote Id\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord User Attribute},
    q{Attribute Query Object==UsageRecord Name==User Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=User DataType=String Description="\"User Name\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Group Attribute},
    q{Attribute Query Object==UsageRecord Name==Group Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Group DataType=String Description="\"Group Name\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Project Attribute},
    q{Attribute Query Object==UsageRecord Name==Project Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Project DataType=String Description="\"Project Name\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Organization Attribute},
    q{Attribute Query Object==UsageRecord Name==Organization Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Organization DataType=String Description="\"Organization Name\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Class Attribute},
    q{Attribute Query Object==UsageRecord Name==Class Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Class DataType=String Description="\"Class or Queue\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord QualityOfService Attribute},
    q{Attribute Query Object==UsageRecord Name==QualityOfService Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=QualityOfService DataType=String Description="\"Quality of Service\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Machine Attribute},
    q{Attribute Query Object==UsageRecord Name==Machine Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Machine DataType=String Description="\"Machine Name\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Nodes Attribute},
    q{Attribute Query Object==UsageRecord Name==Nodes Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Nodes DataType=Integer Description="\"Number of Nodes Used\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Processors Attribute},
    q{Attribute Query Object==UsageRecord Name==Processors Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Processors DataType=Integer Description="\"Number of Processors Used\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Memory Attribute},
    q{Attribute Query Object==UsageRecord Name==Memory Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Memory DataType=String Description="\"Memory Used\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Disk Attribute},
    q{Attribute Query Object==UsageRecord Name==Disk Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Disk DataType=String Description="\"Disk or Storage Used\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Network Attribute},
    q{Attribute Query Object==UsageRecord Name==Network Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Network DataType=String Description="\"Network Used\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Duration Attribute},
    q{Attribute Query Object==UsageRecord Name==Duration Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Duration DataType=Integer Description="\"Usage duration\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord StartTime Attribute},
    q{Attribute Query Object==UsageRecord Name==StartTime Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=StartTime DataType=TimeStamp Description="\"Start Time\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord EndTime Attribute},
    q{Attribute Query Object==UsageRecord Name==EndTime Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=EndTime DataType=TimeStamp Description="\"End Time\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Description Attribute},
    q{Attribute Query Object==UsageRecord Name==Description Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Description Sequence=900 DataType=String Description="\"Description\"" NoRefresh:=True}
);

update(
    q{Modifying UsageRecord Actions to set Display Attribute to True},
    q{Action Query Object==UsageRecord "(Name==Create || Name==Query || Name==Modify || Name==Delete)" Display!=True Show:="\"Count(Name)\""},
    q{>},
    q{Action Modify Object==UsageRecord "(Name==Create || Name==Query || Name==Modify || Name==Delete)" Display=True NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Charge Action},
    q{Action Query Object==UsageRecord Name==Charge Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=UsageRecord Name=Charge Description="\"Charge for Usage\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Reserve Action},
    q{Action Query Object==UsageRecord Name==Reserve Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=UsageRecord Name=Reserve Description="\"Make a Usage Reservation\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Quote Action},
    q{Action Query Object==UsageRecord Name==Quote Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=UsageRecord Name=Quote Description="\"Request a Quote for Usage\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Refund Action},
    q{Action Query Object==UsageRecord Name==Refund Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=UsageRecord Name=Refund Display=True Description="\"Refund Usage\"" NoRefresh:=True}
);

update(
    q{Creating new Quote Object},
    q{Object Query Name==Quote Show:="\"Count(Name)\""},
    q{==},
    q{Object Create Name=Quote Description="\"Quote\""}
);

update(
    q{Creating new Quote Id Attribute},
    q{Attribute Query Object==Quote Name==Id Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Quote Name=Id PrimaryKey=True DataType=AutoGen Description="\"Quote Id\"" NoRefresh:=True}
);

update(
    q{Creating new Quote Amount Attribute},
    q{Attribute Query Object==Quote Name==Amount Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Quote Name=Amount DataType=Currency Description="\"Charge Estimate\"" NoRefresh:=True}
);

update(
    q{Creating new Quote Pinned Attribute},
    q{Attribute Query Object==Quote Name==Pinned Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Quote Name=Pinned DataType=Boolean DefaultValue=True Description="\"Pinned to a single instance?\""}
);

update(
    q{Creating new Quote Instance Attribute},
    q{Attribute Query Object==Quote Name==Instance Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Quote Name=Instance DataType=String Description="\"Instance Name\"" NoRefresh:=True}
);

update(
    q{Creating new Quote UsageRecord Attribute},
    q{Attribute Query Object==Quote Name==UsageRecord Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Quote Name=UsageRecord DataType=Integer Values=@UsageRecord Description="\"Usage Record Id\"" NoRefresh:=True}
);

update(
    q{Creating new Quote StartTime Attribute},
    q{Attribute Query Object==Quote Name==StartTime Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Quote Name=StartTime DataType=TimeStamp Required=True Description="\"When does this Quote start?\"" NoRefresh:=True}
);

update(
    q{Creating new Quote EndTime Attribute},
    q{Attribute Query Object==Quote Name==EndTime Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Quote Name=EndTime DataType=TimeStamp Required=True Description="\"When does this Quote expire?\"" NoRefresh:=True}
);

update(
    q{Creating new Quote Duration Attribute},
    q{Attribute Query Object==Quote Name==Duration Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Quote Name=Duration DataType=Integer Description="\"WallTime Estimate\"" NoRefresh:=True}
);

update(
    q{Creating new Quote Description Attribute},
    q{Attribute Query Object==Quote Name==Description Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Quote Name=Description Sequence=900 DataType=String Description="\"Description\"" NoRefresh:=True}
);

update(
    q{Modifying Quote Actions to set Display Attribute to True},
    q{Action Query Object==Quote "(Name==Query || Name==Delete)" Display!=True Show:="\"Count(Name)\""},
    q{>},
    q{Action Modify Object==Quote "(Name==Query || Name==Delete)" Display=True NoRefresh:=True}
);

update(
    q{Creating new QuoteChargeRate Object},
    q{Object Query Name==QuoteChargeRate Show:="\"Count(Name)\""},
    q{==},
    q{Object Create Name=QuoteChargeRate Association=True Parent=Quote Child=ChargeRate Description="\"Charge Rate guaranteed by the associated Quote\""}
);

update(
    q{Creating new QuoteChargeRate Quote Attribute},
    q{Attribute Query Object==QuoteChargeRate Name==Quote Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=QuoteChargeRate Name=Quote PrimaryKey=True Values=@Quote DataType=Integer Description="\"Parent Quote Id\"" NoRefresh:=True}
);

update(
    q{Creating new QuoteChargeRate Type Attribute},
    q{Attribute Query Object==QuoteChargeRate Name==Type Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=QuoteChargeRate Name=Type DataType=String PrimaryKey=True Description="\"Charge Rate Type\"" NoRefresh:=True}
);

update(
    q{Creating new QuoteChargeRate Name Attribute},
    q{Attribute Query Object==QuoteChargeRate Name==Name Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=QuoteChargeRate Name=Name PrimaryKey=True DataType=String Description="\"Charge Rate Name\"" NoRefresh:=True}
);

update(
    q{Creating new QuoteChargeRate Instance Attribute},
    q{Attribute Query Object==QuoteChargeRate Name==Instance Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=QuoteChargeRate Name=Instance PrimaryKey=True DataType=String DefaultValue=\"\" Description="\"Charge Rate Instance\"" NoRefresh:=True}
);

update(
    q{Creating new QuoteChargeRate Rate Attribute},
    q{Attribute Query Object==QuoteChargeRate Name==Rate Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=QuoteChargeRate Name=Rate DataType=Float Required=True Description="\"Charge Rate\"" NoRefresh:=True}
);

update(
    q{Modifying QuoteChargeRate Actions to set Display Attribute to True},
    q{Action Query Object==QuoteChargeRate Name==Query Display!=True Show:="\"Count(Name)\""},
    q{>},
    q{Action Modify Object==QuoteChargeRate Name==Query Display=True NoRefresh:=True}
);

update(
    q{Creating new Transaction Instance Attribute},
    q{Attribute Query Object==Transaction Name==Instance Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Transaction Name=Instance DataType=String Description=\""Instance Name"\" NoRefresh:=True}
);

update(
    q{Creating new Transaction UsageRecord Attribute},
    q{Attribute Query Object==Transaction Name==UsageRecord Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Transaction Name=UsageRecord DataType=Integer Description=\""Usage Record Id"\" NoRefresh:=True}
);

update(
    q{Creating new Transaction Duration Attribute},
    q{Attribute Query Object==Transaction Name==Duration Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Transaction Name=Duration DataType=Integer Description=\""Event Duration"\" NoRefresh:=True}
);

update(
    q{Creating new Reservation UsageRecord Attribute},
    q{Attribute Query Object==Reservation Name==UsageRecord Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Reservation Name=UsageRecord DataType=Integer Values=@UsageRecord Sequence=30 Description=\""Usage Record Id"\" NoRefresh:=True}
);

update(
    q{Creating new Reservation Instance Attribute},
    q{Attribute Query Object==Reservation Name==Instance Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Reservation Name=Instance DataType=String Sequence=20 Description=\""Instance Name"\" NoRefresh:=True}
);

update(
    q{Creating new Reservation Duration Attribute},
    q{Attribute Query Object==Reservation Name==Duration Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Reservation Name=Duration DataType=Integer Sequence=90 Description=\""WallTime Estimate"\" NoRefresh:=True}
);

update(
    q{Creating new Constraint Object},
    q{Object Query Name==Constraint Show:="\"Count(Name)\""},
    q{==},
    q{Object Create Name=Constraint Description="\"Account Constraint\""}
);

update(
    q{Creating new Constraint Id Attribute},
    q{Attribute Query Object==Constraint Name==Id Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Constraint Name=Id PrimaryKey=True DataType=AutoGen Description="\"Constraint Id\"" NoRefresh:=True}
);

update(
    q{Creating new Constraint Account Attribute},
    q{Attribute Query Object==Constraint Name==Account Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Constraint Name=Account Required=True Fixed=True Values=@Account DataType=Integer Description="\"Account Id\"" NoRefresh:=True}
);

update(
    q{Creating new Constraint Name Attribute},
    q{Attribute Query Object==Constraint Name==Name Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Constraint Name=Name Required=True DataType=String Description="\"Constraint Name\"" NoRefresh:=True}
);

update(
    q{Creating new Constraint Value Attribute},
    q{Attribute Query Object==Constraint Name==Value Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Constraint Name=Value Required=True DataType=String Description="\"Constraint Value\"" NoRefresh:=True}
);

update(
    q{Creating new Account Priority Attribute},
    q{Attribute Query Object==Account Name==Priority Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Account Name=Priority DataType=Integer DefaultValue=0 Description="\"Account Priority\""}
);

update(
    q{Modifying Accounts to set Priority to 0},
    q{Account Query Priority==NULL Show:="\"Count(Id)\""},
    q{>},
    q{Account Modify Priority=0 NoRefresh:=True}
);

update(
    q{Creating new Scheduler UsageRecord Create RoleAction},
    q{RoleAction Query Role==Scheduler Object==UsageRecord Name==Create Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=Scheduler Object=UsageRecord Name=Create Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new Scheduler UsageRecord Quote RoleAction},
    q{RoleAction Query Role==Scheduler Object==UsageRecord Name==Quote Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=Scheduler Object=UsageRecord Name=Quote Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new Scheduler UsageRecord Reserve RoleAction},
    q{RoleAction Query Role==Scheduler Object==UsageRecord Name==Reserve Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=Scheduler Object=UsageRecord Name=Reserve Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new Scheduler UsageRecord Charge RoleAction},
    q{RoleAction Query Role==Scheduler Object==UsageRecord Name==Charge Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=Scheduler Object=UsageRecord Name=Charge Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new Scheduler ANY Query RoleAction},
    q{RoleAction Query Role==Scheduler Object==ANY Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=Scheduler Object=ANY Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new UserServices UsageRecord Refund RoleAction},
    q{RoleAction Query Role==UserServices Object==UsageRecord Name==Refund Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=UserServices Object=UsageRecord Name=Refund Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new Anonymous AccountAccount Query RoleAction},
    q{RoleAction Query Role==Anonymous Object==AccountAccount Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=Anonymous Object=AccountAccount Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new Anonymous Action Query RoleAction},
    q{RoleAction Query Role==Anonymous Object==Action Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=Anonymous Object=Action Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new Anonymous Attribute Query RoleAction},
    q{RoleAction Query Role==Anonymous Object==Attribute Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=Anonymous Object=Attribute Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new Anonymous ChargeRate Query RoleAction},
    q{RoleAction Query Role==Anonymous Object==ChargeRate Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=Anonymous Object=ChargeRate Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new Anonymous Constraint Query RoleAction},
    q{RoleAction Query Role==Anonymous Object==Constraint Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=Anonymous Object=Constraint Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new Anonymous Machine Query RoleAction},
    q{RoleAction Query Role==Anonymous Object==Machine Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=Anonymous Object=Machine Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new Anonymous Object Query RoleAction},
    q{RoleAction Query Role==Anonymous Object==Object Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=Anonymous Object=Object Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new Anonymous Organization Query RoleAction},
    q{RoleAction Query Role==Anonymous Object==Organization Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=Anonymous Object=Organization Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new Anonymous ProjectUser Query RoleAction},
    q{RoleAction Query Role==Anonymous Object==ProjectUser Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=Anonymous Object=ProjectUser Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new Anonymous QuoteChargeRate Query RoleAction},
    q{RoleAction Query Role==Anonymous Object==QuoteChargeRate Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=Anonymous Object=QuoteChargeRate Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new Anonymous ReservationAllocation Query RoleAction},
    q{RoleAction Query Role==Anonymous Object==ReservationAllocation Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=Anonymous Object=ReservationAllocation Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new Anonymous Role Query RoleAction},
    q{RoleAction Query Role==Anonymous Object==Role Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=Anonymous Object=Role Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new Anonymous RoleAction Query RoleAction},
    q{RoleAction Query Role==Anonymous Object==RoleAction Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=Anonymous Object=RoleAction Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new Anonymous RoleUser Query RoleAction},
    q{RoleAction Query Role==Anonymous Object==RoleUser Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=Anonymous Object=RoleUser Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new Anonymous System Query RoleAction},
    q{RoleAction Query Role==Anonymous Object==System Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=Anonymous Object=System Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new OVERRIDE Account Query RoleAction},
    q{RoleAction Query Role==OVERRIDE Object==Account Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=OVERRIDE Object=Account Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new OVERRIDE Allocation Query RoleAction},
    q{RoleAction Query Role==OVERRIDE Object==Allocation Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=OVERRIDE Object=Allocation Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new OVERRIDE Project Query RoleAction},
    q{RoleAction Query Role==OVERRIDE Object==Project Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=OVERRIDE Object=Project Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new OVERRIDE Quote Query RoleAction},
    q{RoleAction Query Role==OVERRIDE Object==Quote Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=OVERRIDE Object=Quote Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new OVERRIDE Reservation Query RoleAction},
    q{RoleAction Query Role==OVERRIDE Object==Reservation Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=OVERRIDE Object=Reservation Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new OVERRIDE Transaction Query RoleAction},
    q{RoleAction Query Role==OVERRIDE Object==Transaction Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=OVERRIDE Object=Transaction Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new OVERRIDE UsageRecord Query RoleAction},
    q{RoleAction Query Role==OVERRIDE Object==UsageRecord Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=OVERRIDE Object=UsageRecord Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Creating new OVERRIDE User Query RoleAction},
    q{RoleAction Query Role==OVERRIDE Object==User Name==Query Show:="\"Count(Name)\""},
    q{==},
    q{RoleAction Create Role=OVERRIDE Object=User Name=Query Instance=ANY NoRefresh:=True}
);

update(
    q{Modifying Organization Object AutoGen Attribute},
    q{Object Query Name==Organization AutoGen==True Show:="\"Count(Name)\""},
    q{==},
    q{Object Modify Name==Organization AutoGen=True NoRefresh:=True}
);

update(
    q{Modifying User DefaultProject Values Attribute},
    q{Attribute Query Object==User Name==DefaultProject Values=="\"@?=Project\"" Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Modify Object==User Name==DefaultProject Values="\"@?=Project\"" NoRefresh:=True}
);

update(
    q{Modifying Project Organization Values Attribute},
    q{Attribute Query Object==Project Name==Organization Values=="\"@!=Organization\"" Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Modify Object==Project Name==Organization Values="\"@!=Organization\"" NoRefresh:=True}
);

update(
    q{Modifying User Object AutoGen Attribute},
    q{Object Query Name==User AutoGen==True Show:="\"Count(Name)\""},
    q{==},
    q{Object Modify Name==User AutoGen=True NoRefresh:=True}
);

update(
    q{Modifying Project Object AutoGen Attribute},
    q{Object Query Name==Project AutoGen==True Show:="\"Count(Name)\""},
    q{==},
    q{Object Modify Name==Project AutoGen=True NoRefresh:=True}
);

update(
    q{Modifying Machine Object AutoGen Attribute},
    q{Object Query Name==Machine AutoGen==True Show:="\"Count(Name)\""},
    q{==},
    q{Object Modify Name==Machine AutoGen=True NoRefresh:=True}
);

update(
    q{Modifying ProjectUser Actions to set Display Attribute to False},
    q{Action Query Object==ProjectUser "(Name==Create || Name==Modify || Name==Delete)" Display!=True Show:="\"Count(Name)\""},
    q{==},
    q{Action Modify Object==ProjectUser "(Name==Create || Name==Modify || Name==Delete)" Display=False NoRefresh:=True}
);

update(
    q{Modifying CreationTime Attributes to set Fixed Attribute to True},
    q{Attribute Query Name==CreationTime Fixed!=True Show:="\"Count(Name)\""},
    q{>},
    q{Attribute Modify Name==CreationTime Fixed!=True Fixed=True NoRefresh:=True}
);

update(
    q{Modifying ModificationTime Attributes to set Fixed Attribute to False},
    q{Attribute Query Name==ModificationTime Fixed!=False Show:="\"Count(Name)\""},
    q{>},
    q{Attribute Modify Name==ModificationTime Fixed!=False Fixed=False NoRefresh:=True}
);

update(
    q{Modifying Deleted Attributes to set Fixed Attribute to False},
    q{Attribute Query Name==Deleted Fixed!=False Show:="\"Count(Name)\""},
    q{>},
    q{Attribute Modify Name==Deleted Fixed!=False Fixed=False NoRefresh:=True}
);

update(
    q{Modifying RequestId Attributes to set Fixed Attribute to False},
    q{Attribute Query Name==RequestId Fixed!=False Show:="\"Count(Name)\""},
    q{>},
    q{Attribute Modify Name==RequestId Fixed!=False Fixed=False NoRefresh:=True}
);

update(
    q{Modifying TransactionId Attributes to set Fixed Attribute to False},
    q{Attribute Query Name==TransactionId Fixed!=False Show:="\"Count(Name)\""},
    q{>},
    q{Attribute Modify Name==TransactionId Fixed!=False Fixed=False NoRefresh:=True}
);

update(
    q{Modifying Deleted Attributes to set DefaultValue to be False},
    q{Attribute Query Name==Deleted DefaultValue==NULL Show:="\"Count(Name)\""},
    q{>},
    q{Attribute Modify Name==Deleted DefaultValue==NULL DefaultValue=False NoRefresh:=True}
);

print "\n***Refreshing Cache***\n";
$cmd    = "mam-shell -v System Refresh";
$output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
$rc     = $? >> 8;
print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
if ($rc)
{
    $failures++;
    my $message = "Subcommand ($cmd) failed with rc=$rc";
    $message .= ": $output" if $output;
    warn($message);
    push @messages, $message;
}
else
{
    $successes++;
}

################################################################################
# Transform Accounting Tables
################################################################################

print
  "\n***Transforming old Account{User,Project,Machine}s to Constraints***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Object Query Name==AccountProject Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    my (%constraints, $request, $response, $code, $message, $doc, $data);

    my $queryFailures = 0;
    foreach my $entity ("User", "Project", "Machine")
    {
        print "\n***Reading Account${entity}s***\n";
        $request = new MAM::Request(
            object => "Account$entity",
            action => "Query",
        );
        $request->addSelection("Account");
        $request->addSelection("Name");
        $request->addSelection("Access");
        $response = $request->getResponse();
        &MAM::Client::displayResponse($response);
        if ($response->getStatus() eq "Failure")
        {
            $code    = $response->getCode();
            $message = "Account$entity Query failed with return code ($code):\n"
              . $response->getMessage() . "\n";
            $queryFailures++;
            warn($message);
            push @messages, $message;
        }
        else
        {
            $doc  = XML::LibXML::Document->new();
            $data = $response->getDataElement();
            $doc->setDocumentElement($data);
            foreach my $row ($data->childNodes())
            {
                my $account =
                  ($row->getChildrenByTagName("Account"))[0]->textContent();
                my $name =
                  ($row->getChildrenByTagName("Name"))[0]->textContent();
                my $access =
                  ($row->getChildrenByTagName("Access"))[0]->textContent();
                push(
                    @{$constraints{$account}},
                    {
                        name  => "$entity",
                        value => $access eq "False" ? "!$name" : "$name"
                    }
                );
            }
        }
    }

    if ($queryFailures)
    {
        $failures++;
    }
    else
    {
        print "\n***Writing Constraints***\n";
      CONSTRAINT:
        foreach my $account (sort keys %constraints)
        {
            foreach my $constraint (@{$constraints{$account}})
            {
                my $name  = $constraint->{name};
                my $value = $constraint->{value};
                next if ($value eq "ANY" || $value eq "MEMBERS");
                $request = new MAM::Request(
                    object => "Constraint",
                    action => "Create",
                );
                $request->addAssignment("Account", $account);
                $request->addAssignment("Name",    $name);
                $request->addAssignment("Value",   $value);
                $response = $request->getResponse();
                &MAM::Client::displayResponse($response);

                if ($response->getStatus() eq "Failure")
                {
                    $code = $response->getCode();
                    $message =
                      "Constraint Create failed with return code ($code):\n"
                      . $response->getMessage() . "\n";
                    $queryFailures++;
                    warn($message);
                    push @messages, $message;
                    last CONSTRAINT;
                }
            }
        }

        if ($queryFailures)
        {
            $failures++;
        }
        else
        {
            $successes++;
        }
    }
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Populating new UsageRecord table from old Job table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Object Query Name==UsageRecord Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Object Query Name==Job Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});
    my $sth;

    $sql =
      "INSERT INTO g_usage_record (g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_id,g_type,g_instance,g_charge,g_stage,g_quote,g_user,g_project,g_class,g_quality_of_service,g_machine,g_nodes,g_processors,g_duration,g_start_time,g_end_time,g_description) SELECT g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_id,g_type,g_job_id,g_charge,g_stage,cast(g_quote_id as integer),g_user,g_project,g_queue,g_quality_of_service,g_machine,g_nodes,g_processors,g_wall_duration,g_start_time,g_end_time,g_description FROM g_job";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Inserted $count lines\n";
    $sql =
      "INSERT INTO g_usage_record_log (g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_id,g_type,g_instance,g_charge,g_stage,g_quote,g_user,g_project,g_class,g_quality_of_service,g_machine,g_nodes,g_processors,g_duration,g_start_time,g_end_time,g_description) SELECT g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_id,g_type,g_job_id,g_charge,g_stage,cast(g_quote_id as integer),g_user,g_project,g_queue,g_quality_of_service,g_machine,g_nodes,g_processors,g_wall_duration,g_start_time,g_end_time,g_description FROM g_job_log";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Inserted $count lines\n";

    print "\n***Updating key generator for UsageRecord***\n";
    $sth =
      $dbh->prepare("SELECT g_next_id FROM g_key_generator WHERE g_name='Job'");
    $sth->execute();
    my ($nextId) = $sth->fetchrow_array();
    if (defined $nextId)
    {
        $sql =
          "UPDATE g_key_generator set g_next_id=$nextId WHERE g_name='UsageRecord'";
        print "Executing sql: $sql\n";
        $count = $dbh->do($sql);
        print "Updated $count lines\n";
        $sql = "DELETE from g_key_generator where g_name='Job'";
        print "Executing sql: $sql\n";
        $count = $dbh->do($sql);
        print "Updated $count lines\n";
    }

    print "\n***Setting default UsageRecord Type***\n";
    $sth = $dbh->prepare(
        "SELECT g_default_value FROM g_attribute WHERE g_object='Job' AND g_name='Type'"
    );
    $sth->execute();
    my ($defaultValue) = $sth->fetchrow_array();
    if (defined $defaultValue)
    {
        $sql =
          "UPDATE g_attribute set g_default_value=$defaultValue WHERE g_object='UsageRecord' AND g_name='Type'";
        print "Executing sql: $sql\n";
        $count = $dbh->do($sql);
        print "Updated $count lines\n";
    }

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Updating Transaction table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Transaction Name==Instance Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Transaction Name==JobId Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});

    $sql = "UPDATE g_transaction SET g_instance=g_job_id";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql = "UPDATE g_transaction_log SET g_instance=g_job_id";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Updating Reservation table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Reservation Name==UsageRecord Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Reservation Name==Job Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});

    $sql =
      "UPDATE g_reservation SET g_usage_record=cast(g_job as integer),g_instance=g_name";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql =
      "UPDATE g_reservation_log SET g_usage_record=cast(g_job as integer),g_instance=g_name";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Populating new Quote table from old Quotation table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Object Query Name==Quote Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Object Query Name==Quotation Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});
    my $sth;

    $sql =
      "INSERT INTO g_quote (g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_id,g_amount,g_pinned,g_usage_record,g_start_time,g_end_time,g_duration,g_description) SELECT g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_id,g_amount,'True',cast(g_job as integer),g_start_time,g_end_time,g_wall_duration,g_description FROM g_quotation";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Inserted $count lines\n";
    $sql =
      "INSERT INTO g_quote_log (g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_id,g_amount,g_pinned,g_usage_record,g_start_time,g_end_time,g_duration,g_description) SELECT g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_id,g_amount,'True',cast(g_job as integer),g_start_time,g_end_time,g_wall_duration,g_description FROM g_quotation";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Inserted $count lines\n";

    print "\n***Updating key generator for Quote***\n";
    $sth = $dbh->prepare(
        "SELECT g_next_id FROM g_key_generator WHERE g_name='Quotation'");
    $sth->execute();
    my ($nextId) = $sth->fetchrow_array();
    if (defined $nextId)
    {
        $sql =
          "UPDATE g_key_generator set g_next_id=$nextId WHERE g_name='Quote'";
        print "Executing sql: $sql\n";
        $count = $dbh->do($sql);
        print "Updated $count lines\n";
        $sql = "DELETE from g_key_generator where g_name='Quotation'";
        print "Executing sql: $sql\n";
        $count = $dbh->do($sql);
        print "Updated $count lines\n";
    }

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

print
  "\n***Populating new QuoteChargeRate table from old QuotationChargeRate table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Object Query Name==QuoteChargeRate Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Object Query Name==QuotationChargeRate Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});

    $sql =
      "INSERT INTO g_quote_charge_rate (g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_quote,g_type,g_name,g_rate,g_instance) SELECT g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_quotation,g_type,g_name,g_rate,g_instance FROM g_quotation_charge_rate";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Inserted $count lines\n";
    $sql =
      "INSERT INTO g_quote_charge_rate_log (g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_quote,g_type,g_name,g_rate,g_instance) SELECT g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_quotation,g_type,g_name,g_rate,g_instance FROM g_quotation_charge_rate_log";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Inserted $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Updating ChargeRate table names***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q ChargeRate Query Name==UserId Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    || (
        chomp(
            $_ =
              qx{mam-shell -q ChargeRate Query Name==ProjectId Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    || (
        chomp(
            $_ =
              qx{mam-shell -q ChargeRate Query Name==MachineName Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});

    $sql = "UPDATE g_charge_rate SET g_name='User' WHERE g_name='UserId'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql = "UPDATE g_charge_rate SET g_name='Project' WHERE g_name='ProjectId'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql =
      "UPDATE g_charge_rate SET g_name='Machine' WHERE g_name='MachineName'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql = "UPDATE g_charge_rate_log SET g_name='User' WHERE g_name='UserId'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql =
      "UPDATE g_charge_rate_log SET g_name='Project' WHERE g_name='ProjectId'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql =
      "UPDATE g_charge_rate_log SET g_name='Machine' WHERE g_name='MachineName'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

################################################################################
# Delete and Update old Accounting Objects and Attributes
################################################################################

update(
    q{Modifying UsageRecords to set NULL Charges to 0},
    q{UsageRecord Query Charge==NULL Show:="\"Count(Id)\""},
    q{>},
    q{UsageRecord Modify Charge==NULL Charge=0 NoRefresh:=True}
);

update(
    q{Deleting old Job RoleActions},
    q{RoleAction Query Object==Job Show:="\"Count(Name)\""},
    q{>},
    q{RoleAction Delete Object==Job NoRefresh:=True}
);

update(
    q{Deleting old ProjectMachine RoleActions},
    q{RoleAction Query Object==ProjectMachine Show:="\"Count(Name)\""},
    q{>},
    q{RoleAction Delete Object==ProjectMachine NoRefresh:=True}
);

update(
    q{Deleting old Anonymous Account Balance RoleAction},
    q{RoleAction Query Role==Anonymous Object==Account Name==Balance Instance==ANY Show:="\"Count(Name)\""},
    q{>},
    q{RoleAction Delete Role==Anonymous Object==Account Name==Balance Instance==ANY NoRefresh:=True}
);

update(
    q{Deleting old Anonymous ANY Query RoleAction},
    q{RoleAction Query Role==Anonymous Object==ANY Name==Query Instance==ANY Show:="\"Count(Name)\""},
    q{>},
    q{RoleAction Delete Role==Anonymous Object==ANY Name==Query Instance==ANY NoRefresh:=True}
);

update(
    q{Disabling Undelete Action From GUI},
    q{Action Query Name==Undelete Display==True Show:="\"Count(Name)\""},
    q{>},
    q{Action Modify Name==Undelete Display=False NoRefresh:=True}
);

update(
    q{Deleting old System Information},
    q{System Query Name==Gold Show:="\"Count(Name)\""},
    q{>},
    q{System Delete Name==Gold NoRefresh:=True}
);

################################################################################
# Cleaning up old Metadata and Updating Indexes
################################################################################

print "\n***Cleaning up old Metadata and Updating Indexes***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q System Query Name==\\\"Moab Accounting Manager\\\" Version==\\\"7.0\\\" Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ == 0
    )
    && ! $failures
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 0});

    # AccountMachine
    $sql = "DELETE FROM g_attribute WHERE g_object='AccountMachine'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_attribute_log WHERE g_object='AccountMachine'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action WHERE g_object='AccountMachine'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action_log WHERE g_object='AccountMachine'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object WHERE g_name='AccountMachine'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object_log WHERE g_name='AccountMachine'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DROP TABLE g_account_machine";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "DROP TABLE g_account_machine_log";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # AccountOrganization
    $sql = "DELETE FROM g_attribute WHERE g_object='AccountOrganization'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_attribute_log WHERE g_object='AccountOrganization'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action WHERE g_object='AccountOrganization'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action_log WHERE g_object='AccountOrganization'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object WHERE g_name='AccountOrganization'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object_log WHERE g_name='AccountOrganization'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DROP TABLE g_account_organization";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "DROP TABLE g_account_organization_log";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # AccountProject
    $sql = "DELETE FROM g_attribute WHERE g_object='AccountProject'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_attribute_log WHERE g_object='AccountProject'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action WHERE g_object='AccountProject'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action_log WHERE g_object='AccountProject'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object WHERE g_name='AccountProject'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object_log WHERE g_name='AccountProject'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DROP TABLE g_account_project";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "DROP TABLE g_account_project_log";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # AccountUser
    $sql = "DELETE FROM g_attribute WHERE g_object='AccountUser'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_attribute_log WHERE g_object='AccountUser'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action WHERE g_object='AccountUser'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action_log WHERE g_object='AccountUser'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object WHERE g_name='AccountUser'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object_log WHERE g_name='AccountUser'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DROP TABLE g_account_user";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "DROP TABLE g_account_user_log";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Action
    $sql = "DELETE FROM g_action WHERE g_special='True'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action_log WHERE g_special='True'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Action' AND g_name='Special'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Action' AND g_name='Special'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_action DROP COLUMN g_special";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_action_log DROP COLUMN g_special";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Allocation
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Allocation' AND g_name='CallType'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Allocation' AND g_name='CallType'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_allocation DROP COLUMN g_call_type";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_allocation_log DROP COLUMN g_call_type";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Attribute
    $sql = "DELETE FROM g_attribute WHERE g_special='True'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_attribute_log WHERE g_special='True'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Attribute' AND g_name='Special'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Attribute' AND g_name='Special'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_attribute DROP COLUMN g_special";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_attribute_log DROP COLUMN g_special";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Job
    $sql = "DELETE FROM g_attribute WHERE g_object='Job'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_attribute_log WHERE g_object='Job'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action WHERE g_object='Job'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action_log WHERE g_object='Job'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object WHERE g_name='Job'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object_log WHERE g_name='Job'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DROP TABLE g_job";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "DROP TABLE g_job_log";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Machine
    $sql = "DELETE FROM g_machine WHERE g_special='True'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_machine_log WHERE g_special='True'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Machine' AND g_name='Special'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Machine' AND g_name='Special'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_machine DROP COLUMN g_special";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_machine_log DROP COLUMN g_special";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Object
    $sql = "DELETE FROM g_object WHERE g_special='True'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object_log WHERE g_special='True'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Object' AND g_name='Special'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Object' AND g_name='Special'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_object DROP COLUMN g_special";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_object_log DROP COLUMN g_special";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Organization
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Organization' AND g_name='Host'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Organization' AND g_name='Host'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Organization' AND g_name='Port'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Organization' AND g_name='Port'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Organization' AND g_name='Special'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Organization' AND g_name='Special'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_organization DROP COLUMN g_host";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_organization_log DROP COLUMN g_host";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_organization DROP COLUMN g_port";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_organization_log DROP COLUMN g_port";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Organization' AND g_name='Special'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Organization' AND g_name='Special'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_organization DROP COLUMN g_special";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_organization_log DROP COLUMN g_special";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Project
    $sql = "DELETE FROM g_project WHERE g_special='True'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_project_log WHERE g_special='True'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Project' AND g_name='Special'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Project' AND g_name='Special'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_project DROP COLUMN g_special";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_project_log DROP COLUMN g_special";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # ProjectMachine
    $sql = "DELETE FROM g_attribute WHERE g_object='ProjectMachine'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_attribute_log WHERE g_object='ProjectMachine'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action WHERE g_object='ProjectMachine'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action_log WHERE g_object='ProjectMachine'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object WHERE g_name='ProjectMachine'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object_log WHERE g_name='ProjectMachine'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DROP TABLE g_project_machine";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "DROP TABLE g_project_machine_log";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Quotation
    $sql = "DELETE FROM g_attribute WHERE g_object='Quotation'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_attribute_log WHERE g_object='Quotation'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action WHERE g_object='Quotation'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action_log WHERE g_object='Quotation'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object WHERE g_name='Quotation'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object_log WHERE g_name='Quotation'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DROP TABLE g_quotation";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "DROP TABLE g_quotation_log";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # QuotationChargeRate
    $sql = "DELETE FROM g_attribute WHERE g_object='QuotationChargeRate'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_attribute_log WHERE g_object='QuotationChargeRate'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action WHERE g_object='QuotationChargeRate'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action_log WHERE g_object='QuotationChargeRate'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object WHERE g_name='QuotationChargeRate'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object_log WHERE g_name='QuotationChargeRate'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DROP TABLE g_quotation_charge_rate";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "DROP TABLE g_quotation_charge_rate_log";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Reservation
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Reservation' AND g_name='Job'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Reservation' AND g_name='Job'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Reservation' AND g_name='Name'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Reservation' AND g_name='Name'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Reservation' AND g_name='User'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Reservation' AND g_name='User'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Reservation' AND g_name='Project'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Reservation' AND g_name='Project'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Reservation' AND g_name='Machine'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Reservation' AND g_name='Machine'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Reservation' AND g_name='CallType'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Reservation' AND g_name='CallType'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_reservation DROP COLUMN g_job";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_reservation_log DROP COLUMN g_job";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_reservation DROP COLUMN g_name";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_reservation_log DROP COLUMN g_name";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_reservation DROP COLUMN g_user";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_reservation_log DROP COLUMN g_user";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_reservation DROP COLUMN g_project";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_reservation_log DROP COLUMN g_project";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_reservation DROP COLUMN g_machine";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_reservation_log DROP COLUMN g_machine";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_reservation DROP COLUMN g_call_type";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_reservation_log DROP COLUMN g_call_type";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # System
    $sql =
      "DELETE FROM g_attribute WHERE g_object='System' AND g_name='Organization'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='System' AND g_name='Organization'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_system DROP COLUMN g_organization";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_system_log DROP COLUMN g_organization";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Transaction
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Transaction' AND g_name='JobId'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Transaction' AND g_name='JobId'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_transaction DROP COLUMN g_job_id";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_transaction_log DROP COLUMN g_job_id";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # User
    $sql = "DELETE FROM g_user WHERE g_special='True'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_user_log WHERE g_special='True'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_attribute WHERE g_object='User' AND g_name='Special'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='User' AND g_name='Special'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_user DROP COLUMN g_special";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_user_log DROP COLUMN g_special";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
}
else
{
    print "Already applied\n";
    $ignores++;
}

################################################################################
# Refresh Cache and Create New System Information
################################################################################

print "\n***Refreshing Cache***\n";
$cmd    = "mam-shell -v System Refresh";
$output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
$rc     = $? >> 8;
print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
if ($rc)
{
    $failures++;
    my $message = "Subcommand ($cmd) failed with rc=$rc";
    $message .= ": $output" if $output;
    warn($message);
    push @messages, $message;
}
else
{
    $successes++;
}

unless ($failures)
{
    update(
        q{Creating new System Information},
        q{System Query Name==\"Moab Accounting Manager\" Version==\"7.0\" Show:="\"Count(Name)\""},
        q{==},
        q{System Create Name="\"Moab Accounting Manager\"" Version=\"7.0\" Description="\"Commercial Release\""}
    );
}

################################################################################
# Display Results
################################################################################

print "\n";
print "$successes updates were successfully applied.\n";
print "$ignores updates were skipped having already been applied.\n";
print "$failures updates failed.\n";
if (@messages)
{
    print "The following is a list of failures:\n";
    print join "\n", @messages;
}

exit 0;
