#!/usr/bin/perl -w
use strict;

use lib qw(/opt/mam/lib /opt/mam/lib/perl5);
use Config::Tiny;
use DBI;
use MAM;
use MAM::Global;

my ($cmd, $output, $rc, $sql, $count);
my ($successes, $failures, $ignores) = (0, 0, 0);
my @messages   = ();
$ENV{PATH} = "/opt/mam/bin:/opt/mam/sbin:$ENV{PATH}";
my $now = time;

# Read the Config File
my $configFile = "/opt/mam/etc/mam-server.conf";
my $configTiny = Config::Tiny->new;
$configTiny = Config::Tiny->read($configFile)
  or die(
    "Unable to open config file ($configFile) for reading: $Config::Tiny::errstr.\nYou may need to set the \$configFile variable in this script.\n"
  );
$serverConfig = $configTiny->{_};
my $datasource = $serverConfig->{'database.datasource'};
my $dbuser     = $serverConfig->{'database.user'} || (getpwuid($<))[0];
my $dbpasswd   = $serverConfig->{'database.password'};

################################################################################
# Display warning and prompt to proceed
################################################################################

my $warning = <<"ENDOFWARNING";
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! IMPORTANT !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

This script will attempt to convert an existing Moab Accounting Manager 7.0 database (using the datasource ($datasource) referenced in /opt/mam/etc/mam-server.conf) into a Moab Accounting Manager 7.1 database. It will make additions, modifications and deletions to your Moab Accounting Manager 7.0 schema and data to match the new base Moab Accounting Manager 7.1 schema and data values.

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! BACKUP YOUR DATA !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

You should MAKE A BACKUP of your Moab Accounting Manager database before running this script. If you proceed without a backup, a failure in this script could irreversibly corrupt the integrity and data of your Moab Accounting Manager database.

ENDOFWARNING

print $warning;
print "Do you want to proceed? [y/N] ";
my $answer = <STDIN>;
exit 0 unless ($answer =~ /^y(es)?$/i);

# ----------------------------------------------------------------------------
# update($description, $test, $comparison, $action);
# ----------------------------------------------------------------------------

# Update the object metadata via mam-shell
sub update
{
    my ($description, $test, $comparison, $action) = @_;
    print "\n***$description***\n";
    if (chomp($_ = qx{mam-shell -q $test}) && eval "\$_ $comparison 0")
    {
        $cmd    = "mam-shell -v $action";
        $output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
        $rc     = $? >> 8;
        print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
        if ($rc)
        {
            $failures++;
            my $message = "Subcommand ($cmd) failed with rc=$rc";
            $message .= ": $output" if $output;
            warn($message);
            push @messages, $message;
        }
        else
        {
            $successes++;
        }
    }
    else
    {
        print "Already applied\n";
        $ignores++;
    }
}

################################################################################
# Perform Precursor Base Changes
################################################################################

print "\n***Creating new Transaction Key Attribute***\n";
if (
    chomp(
        $_ =
          qx{mam-shell -q Attribute Query Object==Transaction Name==Key Show:=\"\\\"Count(Name)\\\"\"}
    )
    && $_ == 0
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});
    my $sth;

    $sql = "ALTER TABLE g_transaction ADD COLUMN g_key varchar(1024)";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_transaction_log ADD COLUMN g_key varchar(1024)";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql =
      "UPDATE g_key_generator set g_next_id=g_next_id+1 WHERE g_name='Request'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sth = $dbh->prepare(
        "SELECT g_next_id-1 FROM g_key_generator WHERE g_name='Request'");
    $sth->execute();
    my ($requestId) = $sth->fetchrow_array();
    $sql =
      "UPDATE g_key_generator set g_next_id=g_next_id+1 WHERE g_name='Transaction'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sth = $dbh->prepare(
        "SELECT g_next_id-1 FROM g_key_generator WHERE g_name='Transaction'");
    $sth->execute();
    my ($transactionId) = $sth->fetchrow_array();
    $sql =
      qq{INSERT INTO g_attribute (g_object, g_name, g_data_type, g_primary_key, g_required, g_fixed, g_values, g_default_value, g_sequence, g_hidden, g_description, g_deleted, g_creation_time, g_modification_time, g_request_id, g_transaction_id) VALUES ('Transaction', 'Key', 'String', 'False', 'False', 'False', NULL, NULL, 50, 'False', 'Object Primary Key Value', 'False', $now, $now, $requestId, $transactionId)};
    print " Executing sql : $sql \n ";
    $count = $dbh->do($sql);
    print " Updated $count lines \n ";
    $sql =
      qq{INSERT INTO g_transaction (g_id, g_object, g_action, g_actor, g_name, g_child, g_count, g_details, g_description, g_deleted, g_creation_time, g_modification_time, g_request_id, g_transaction_id, g_instance, g_amount, g_delta, g_user, g_project, g_machine, g_account, g_allocation, g_usage_record, g_duration, g_key) VALUES (1, 'Attribute', 'Create', 'mam', NULL, NULL, 1, 'DataType=String,Deleted=False,Description="Object Primary Key Value",Fixed=False,Hidden=False,NoRefresh:=True,Object=Transaction,PrimaryKey=False,Required=False,Sequence=50', NULL, 'False', $now, $now, $requestId, $transactionId, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 'Key')};
    print " Executing sql : $sql \n ";
    $count = $dbh->do($sql);
    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

################################################################################
# Create New Base Objects and Attributes
################################################################################

update(
    q{Creating new Event Object},
    q{Object Query Name==Event Show:="\"Count(Name)\""},
    q{==},
    q{Object Create Name=Event Description="\"Scheduled Event\""}
);

update(
    q{Creating new Event Id Attribute},
    q{Attribute Query Object==Event Name==Id Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Event Name=Id PrimaryKey=True DataType=AutoGen Description="\"Event Id\"" NoRefresh:=True}
);

update(
    q{Creating new Event FireCommand Attribute},
    q{Attribute Query Object==Event Name==FireCommand Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Event Name=FireCommand Required=True Description="\"Command To Execute When Fired\"" NoRefresh:=True}
);

update(
    q{Creating new Event ArmTime Attribute},
    q{Attribute Query Object==Event Name==ArmTime Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Event Name=ArmTime DataType=TimeStamp Description="\"Lap Start Time\"" NoRefresh:=True}
);

update(
    q{Creating new Event FireTime Attribute},
    q{Attribute Query Object==Event Name==FireTime Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Event Name=FireTime DataType=TimeStamp Description="\"Target Fire Time\"" NoRefresh:=True}
);

update(
    q{Creating new Event RearmPeriod Attribute},
    q{Attribute Query Object==Event Name==RearmPeriod Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Event Name=RearmPeriod DataType=String Description="\"Repeat Frequency\"" NoRefresh:=True}
);

update(
    q{Creating new Event EndTime Attribute},
    q{Attribute Query Object==Event Name==EndTime Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Event Name=EndTime DataType=TimeStamp Description="\"End Time\"" NoRefresh:=True}
);

update(
    q{Creating new Event Notify Attribute},
    q{Attribute Query Object==Event Name==Notify Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Event Name=Notify DataType=String Description="\"Notification URL\"" NoRefresh:=True}
);

update(
    q{Creating new Event RearmOnFailure Attribute},
    q{Attribute Query Object==Event Name==RearmOnFailure Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Event Name=RearmOnFailure DataType=Boolean DefaultValue=False Description="\"Rearm Event On Failure?\"" NoRefresh:=True}
);

update(
    q{Creating new Event FailureCommand Attribute},
    q{Attribute Query Object==Event Name==FailureCommand Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Event Name=FailureCommand Description="\"Command To Execute On Failure\"" NoRefresh:=True}
);

update(
    q{Creating new Event CatchUp Attribute},
    q{Attribute Query Object==Event Name==CatchUp Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Event Name=CatchUp DataType=Boolean DefaultValue=True Description="\"Progressively Fire Past Due Events?\"" NoRefresh:=True}
);

update(
    q{Creating new Event Description Attribute},
    q{Attribute Query Object==Event Name==Description Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Event Name=Description DataType=String Sequence=900 Description="\"Event Description\"" NoRefresh:=True}
);

update(
    q{Modifying Event Actions to set Display Attribute to True},
    q{Action Query Object==Event "(Name==Query || Name==Delete)" Display!=True Show:="\"Count(Name)\""},
    q{>},
    q{Action Modify Object==Event "(Name==Query || Name==Delete)" Display=True NoRefresh:=True}
);

update(
    q{Creating new Event Fire Action},
    q{Action Query Object==Event Name==Fire Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=Event Name=Fire Display=False Description="\"Trigger an event\"" NoRefresh:=True}
);

update(
    q{Creating new Event Refresh Action},
    q{Action Query Object==Event Name==Refresh Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=Event Name=Refresh Display=False Description="\"Delete stale events\"" NoRefresh:=True}
);

update(
    q{Creating new Notification Object},
    q{Object Query Name==Notification Show:="\"Count(Name)\""},
    q{==},
    q{Object Create Name=Notification Description="\"Event Notification\""}
);

update(
    q{Creating new Notification Id Attribute},
    q{Attribute Query Object==Notification Name==Id Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Notification Name=Id PrimaryKey=True DataType=AutoGen Description="\"Notification Id\"" NoRefresh:=True}
);

update(
    q{Creating new Notification Event Attribute},
    q{Attribute Query Object==Notification Name==Event Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Notification Name=Event DataType=Integer Required=True Fixed=True Values=@Event Description="\"Event Id\"" NoRefresh:=True}
);

update(
    q{Creating new Notification Type Attribute},
    q{Attribute Query Object==Notification Name==Type Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Notification Name=Type DataType=String Description="\"Notification Type\"" NoRefresh:=True}
);

update(
    q{Creating new Notification Status Attribute},
    q{Attribute Query Object==Notification Name==Status Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Notification Name=Status DataType=String Required=True Description="\"Event Command Status\"" NoRefresh:=True}
);

update(
    q{Creating new Notification Code Attribute},
    q{Attribute Query Object==Notification Name==Code Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Notification Name=Code DataType=String Required=True Description="\"Event Command Exit Code\"" NoRefresh:=True}
);

update(
    q{Creating new Notification Message Attribute},
    q{Attribute Query Object==Notification Name==Message Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Notification Name=Message DataType=String Description="\"Event Command Message\"" NoRefresh:=True}
);

update(
    q{Creating new Notification Key Attribute},
    q{Attribute Query Object==Notification Name==Key Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Notification Name=Key DataType=String Description="\"Object Primary Key Value\"" NoRefresh:=True}
);

update(
    q{Creating new Notification Recipient Attribute},
    q{Attribute Query Object==Notification Name==Recipient Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Notification Name=Recipient DataType=String Description="\"Recipient to Notify\"" NoRefresh:=True}
);

update(
    q{Creating new Notification EndTime Attribute},
    q{Attribute Query Object==Notification Name==EndTime Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Notification Name=EndTime DataType=TimeStamp Description="\"When does this Notification Expire?\"" NoRefresh:=True}
);

update(
    q{Modifying Notification Actions to set Display Attribute to True},
    q{Action Query Object==Notification "(Name==Query || Name==Delete)" Display!=True Show:="\"Count(Name)\""},
    q{>},
    q{Action Modify Object==Notification "(Name==Query || Name==Delete)" Display=True NoRefresh:=True}
);

update(
    q{Creating new Notification Send Action},
    q{Action Query Object==Notification Name==Send Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=Notification Name=Send Display=False Description="\"Send Notification\"" NoRefresh:=True}
);

update(
    q{Creating new Notification Refresh Action},
    q{Action Query Object==Notification Name==Refresh Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=Notification Name=Refresh Display=False Description="\"Delete Stale Notifications\"" NoRefresh:=True}
);

update(
    q{Modifying CreationTime Attributes to set Fixed Attribute to True},
    q{Attribute Query Name==CreationTime Fixed!=True Show:="\"Count(Name)\""},
    q{>},
    q{Attribute Modify Name==CreationTime Fixed!=True Fixed=True NoRefresh:=True}
);

################################################################################
# Transform Base Tables
################################################################################

print "\n***Updating Transaction table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Transaction Name==Key Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Transaction Name==Name Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});

    $sql = "UPDATE g_transaction SET g_key=g_name";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql = "UPDATE g_transaction_log SET g_key=g_name";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

################################################################################
# Create New Accounting Objects and Attributes
################################################################################

update(
    q{Creating new UsageRecord Payment Attribute},
    q{Attribute Query Object==UsageRecord Name==Payment Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Payment DataType=Integer Values=@Payment Hidden=True Description="\"Payment Id\"" NoRefresh:=True}
);

update(
    q{Modifying UsageRecord Quote Attribute to set Hidden to True},
    q{Attribute Query Object==UsageRecord Name==Quote Hidden==True Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Modify Object==UsageRecord Name==Quote Hidden=True}
);

update(
    q{Modifying UsageRecord Duration Attribute to set DefaultValue to 0},
    q{Attribute Query Object==UsageRecord Name==Duration DefaultValue==0 Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Modify Object==UsageRecord Name==Duration DefaultValue=0}
);

update(
    q{Modifying UsageRecord Type Attribute to set DefaultValue to Job if NULL},
    q{Attribute Query Object==UsageRecord Name==Type DefaultValue==NULL Show:="\"Count(Name)\""},
    q{!=},
    q{Attribute Modify Object==UsageRecord Name==Type DefaultValue==NULL DefaultValue=Job}
);

update(
    q{Creating new ChargeRate Value Attribute},
    q{Attribute Query Object==ChargeRate Name==Value Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=ChargeRate Name=Value PrimaryKey=True DataType=String DefaultValue="\"\"" Sequence=30 Description="\"Charge Rate Value\"" NoRefresh:=True}
);

update(
    q{Modifying ChargeRate Type Attribute},
    q{Attribute Query Object==ChargeRate Name==Type PrimaryKey==False Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Modify Object==ChargeRate Name==Type PrimaryKey=False Fixed=False Required=True Values="\"(VBR,NBR,VBU,NBU,VBM,NBM,VBF,NBF)\"" Sequence=35 NoRefresh:=True}
);

update(
    q{Creating new QuoteChargeRate Value Attribute},
    q{Attribute Query Object==QuoteChargeRate Name==Value Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=QuoteChargeRate Name=Value PrimaryKey=True DataType=String Sequence=40 Description="\"Charge Rate Value\"" NoRefresh:=True}
);

update(
    q{Modifying QuoteChargeRate Type Attribute},
    q{Attribute Query Object==QuoteChargeRate Name==Type PrimaryKey==False Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Modify Object==QuoteChargeRate Name==Type PrimaryKey=False Fixed=False Required=True Values="\"(VBR,NBR,VBU,NBU,VBM,NBM,VBF,NBF)\"" Sequence=45 NoRefresh:=True}
);

update(
    q{Creating new Charge Object},
    q{Object Query Name==Charge Show:="\"Count(Name)\""},
    q{==},
    q{Object Create Name=Charge Description="\"Charge Information\""}
);

update(
    q{Creating new Charge UsageRecord Attribute},
    q{Attribute Query Object==Charge Name==UsageRecord Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Charge Name=UsageRecord DataType=Integer Required=True Description="\"Usage Record Id\"" NoRefresh:=True}
);

update(
    q{Creating new Charge Instance Attribute},
    q{Attribute Query Object==Charge Name==Instance Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Charge Name=Instance DataType=String Description="\"Instance Name\"" NoRefresh:=True}
);

update(
    q{Creating new Charge Name Attribute},
    q{Attribute Query Object==Charge Name==Name Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Charge Name=Name DataType=String Required=True Description="\"Usage Property Name\"" NoRefresh:=True}
);

update(
    q{Creating new Charge Value Attribute},
    q{Attribute Query Object==Charge Name==Value Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Charge Name=Value DataType=String Required=True Description="\"Usage Property Value\"" NoRefresh:=True}
);

update(
    q{Creating new Charge Duration Attribute},
    q{Attribute Query Object==Charge Name==Duration Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Charge Name=Duration DataType=Integer Description="\"Usage Duration\"" NoRefresh:=True}
);

update(
    q{Creating new Charge Rate Attribute},
    q{Attribute Query Object==Charge Name==Rate Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Charge Name=Rate DataType=Float Description="\"Base Charge Rate\"" NoRefresh:=True}
);

update(
    q{Creating new Charge ScalingFactor Attribute},
    q{Attribute Query Object==Charge Name==ScalingFactor Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Charge Name=ScalingFactor DataType=Float Description="\"Product of Discounts and Premiums\"" NoRefresh:=True}
);

update(
    q{Creating new Charge Amount Attribute},
    q{Attribute Query Object==Charge Name==Amount Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Charge Name=Amount DataType=Currency Required=True Description="\"Amount Charged\"" NoRefresh:=True}
);

update(
    q{Creating new Charge Details Attribute},
    q{Attribute Query Object==Charge Name==Details Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Charge Name=Details DataType=String Description="\"Charge Calculation Details\"" NoRefresh:=True}
);

update(
    q{Creating new Charge Description Attribute},
    q{Attribute Query Object==Charge Name==Description Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Charge Name=Description DataType=String Description="\"Charge Description\"" NoRefresh:=True}
);

update(
    q{Modifying Charge Query Action to set Display Attribute to True},
    q{Action Query Object==Charge Name==Query Display!=True Show:="\"Count(Name)\""},
    q{>},
    q{Action Modify Object==Charge Name==Query Display=True NoRefresh:=True}
);

update(
    q{Creating new Payment Object},
    q{Object Query Name==Payment Show:="\"Count(Name)\""},
    q{==},
    q{Object Create Name=Payment Description="\"Scheduled Payment Plan\""}
);

update(
    q{Creating new Payment Id Attribute},
    q{Attribute Query Object==Payment Name==Id Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Payment Name=Id PrimaryKey=True DataType=AutoGen Description="\"Payment Id\"" NoRefresh:=True}
);

update(
    q{Creating new Payment BeginEvent Attribute},
    q{Attribute Query Object==Payment Name==BeginEvent Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Payment Name=BeginEvent DataType=Integer Values=@Event Description="\"Begin Event\"" NoRefresh:=True}
);

update(
    q{Creating new Payment UseEvent Attribute},
    q{Attribute Query Object==Payment Name==UseEvent Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Payment Name=UseEvent DataType=Integer Values=@Event Description="\"Use Event\"" NoRefresh:=True}
);

update(
    q{Creating new Payment EndEvent Attribute},
    q{Attribute Query Object==Payment Name==EndEvent Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Payment Name=EndEvent DataType=Integer Values=@Event Description="\"End Event\"" NoRefresh:=True}
);

update(
    q{Creating new Payment BeginTime Attribute},
    q{Attribute Query Object==Payment Name==BeginTime Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Payment Name=BeginTime DataType=TimeStamp Description="\"Begin Time\"" NoRefresh:=True}
);

update(
    q{Creating new Payment StartTime Attribute},
    q{Attribute Query Object==Payment Name==StartTime Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Payment Name=StartTime DataType=TimeStamp Description="\"Last Start Time\"" NoRefresh:=True}
);

update(
    q{Creating new Payment StopTime Attribute},
    q{Attribute Query Object==Payment Name==StopTime Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Payment Name=StopTime DataType=TimeStamp Description="\"Last Stop Time\"" NoRefresh:=True}
);

update(
    q{Creating new Payment EndTime Attribute},
    q{Attribute Query Object==Payment Name==EndTime Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Payment Name=EndTime DataType=TimeStamp Description="\"End Time\"" NoRefresh:=True}
);

update(
    q{Creating new Payment Duration Attribute},
    q{Attribute Query Object==Payment Name==Duration Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Payment Name=Duration DataType=Integer Description="\"Payment Duration\"" NoRefresh:=True}
);

update(
    q{Creating new Payment ChargePeriod Attribute},
    q{Attribute Query Object==Payment Name==ChargePeriod Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Payment Name=ChargePeriod DataType=String Description="\"Charge Interval\"" NoRefresh:=True}
);

update(
    q{Creating new Payment ReserveMode Attribute},
    q{Attribute Query Object==Payment Name==ReserveMode Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Payment Name=ReserveMode DataType=String Values="\"(Combined,None,Periodic)\"" DefaultValue=Periodic Description="\"Payment Reservation Model\"" NoRefresh:=True}
);

update(
    q{Creating new Payment ChargeMode Attribute},
    q{Attribute Query Object==Payment Name==ChargeMode Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Payment Name=ChargeMode DataType=String Values="\"(Combined,None,Periodic)\"" DefaultValue=Periodic Description="\"Payment Charge Model\"" NoRefresh:=True}
);

update(
    q{Creating new Payment Notify Attribute},
    q{Attribute Query Object==Payment Name==Notify Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Payment Name=Notify DataType=String Description="\"Notification URL\"" NoRefresh:=True}
);

update(
    q{Creating new Payment Description Attribute},
    q{Attribute Query Object==Payment Name==Description Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Payment Name=Description DataType=String Description="\"Payment Description\"" NoRefresh:=True}
);

#update(
#    q{Modifying Payment Actions to set Display Attribute to True},
#    q{Action Query Object==Payment "(Name==Query || Name==Delete)" Display!=True Show:="\"Count(Name)\""},
#    q{==},
#    q{Action Modify Object==Payment "(Name==Query || Name==Delete)" Display=True NoRefresh:=True}
#);

update(
    q{Creating new Payment Quote Action},
    q{Action Query Object==Payment Name==Quote Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=Payment Name=Quote Display=False Description="\"Request a Quote for a Payment\"" NoRefresh:=True}
);

update(
    q{Creating new Payment Begin Action},
    q{Action Query Object==Payment Name==Begin Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=Payment Name=Begin Display=False Description="\"Begin Payment Charges\"" NoRefresh:=True}
);

update(
    q{Creating new Payment Start Action},
    q{Action Query Object==Payment Name==Start Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=Payment Name=Start Display=False Description="\"Resume Payment Charges\"" NoRefresh:=True}
);

update(
    q{Creating new Payment Use Action},
    q{Action Query Object==Payment Name==Use Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=Payment Name=Use Display=False Description="\"Payment Use Charge\"" NoRefresh:=True}
);

update(
    q{Creating new Payment Stop Action},
    q{Action Query Object==Payment Name==Stop Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=Payment Name=Stop Display=False Description="\"Pause Payment Charges\"" NoRefresh:=True}
);

update(
    q{Creating new Payment End Action},
    q{Action Query Object==Payment Name==End Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=Payment Name=End Display=False Description="\"End Payment Charges\"" NoRefresh:=True}
);

update(
    q{Creating new PaymentComponent Object},
    q{Object Query Name==PaymentComponent Show:="\"Count(Name)\""},
    q{==},
    q{Object Create Name=PaymentComponent Association=True Parent=Payment Child=UsageRecord Description="\"Payment Usage Record Association\""}
);

update(
    q{Creating new PaymentComponent Payment Attribute},
    q{Attribute Query Object==PaymentComponent Name==Payment Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=PaymentComponent Name=Payment PrimaryKey=True Values=@Payment DataType=Integer Description="\"Parent Payment Id\"" NoRefresh:=True}
);

update(
    q{Creating new PaymentComponent Id Attribute},
    q{Attribute Query Object==PaymentComponent Name==Id Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=PaymentComponent Name=Id PrimaryKey=True Values=@UsageRecord DataType=Integer Description="\"Child Usage Record Id\"" NoRefresh:=True}
);

update(
    q{Creating new PaymentComponent Type Attribute},
    q{Attribute Query Object==PaymentComponent Name==Type Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=PaymentComponent Name=Type DataType=String Required=True Values="\"(Begin,End,Use)\"" Description="\"Payment Component Type\"" NoRefresh:=True}
);

#update(
#    q{Modifying PaymentComponent Query Action to set Display Attribute to True},
#    q{Action Query Object==PaymentComponent Name==Query Display!=True Show:="\"Count(Name)\""},
#    q{==},
#    q{Action Modify Object==PaymentComponent Name==Query Display=True NoRefresh:=True}
#);

update(
    q{Modifying Allocation Amount to set Required Attribute to False and DefaultValue Attribute to 0},
    q{Attribute Query Object==Allocation Name==Amount Required==False Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Modify Object==Allocation Name==Amount Required=False DefaultValue=0 NoRefresh:=True}
);

update(
    q{Creating new Account Reset Action},
    q{Action Query Object==Account Name==Reset Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=Account Name=Reset Display=True Description="\"Reset\"" NoRefresh:=True}
);

update(
    q{Creating new Allocation Reset Action},
    q{Action Query Object==Allocation Name==Reset Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=Allocation Name=Reset Display=False Description="\"Reset\"" NoRefresh:=True}
);

update(
    q{Capitalizing Infinity in Attribute DefaultValues},
    q{Attribute Query DefaultValue==infinity Show:="\"Count(Name)\""},
    q{>},
    q{Attribute Modify DefaultValue==infinity DefaultValue=Infinity NoRefresh:=True}
);

update(
    q{Capitalizing -Infinity in Attribute DefaultValues},
    q{Attribute Query DefaultValue=="\"-infinity\"" Show:="\"Count(Name)\""},
    q{>},
    q{Attribute Modify DefaultValue=="\"-infinity\"" DefaultValue="\"-Infinity\"" NoRefresh:=True}
);

update(
    q{Capitalizing -Infinity in Allocation StartTime},
    q{Allocation Base::Query StartTime=="\"-infinity\"" Show:="\"Count(Id)\""},
    q{>},
    q{Allocation Base::Modify StartTime=="\"-infinity\"" StartTime="\"-Infinity\"" NoRefresh:=True}
);

update(
    q{Capitalizing Infinity in Allocation EndTime},
    q{Allocation Base::Query EndTime==infinity Show:="\"Count(Id)\""},
    q{>},
    q{Allocation Base::Modify EndTime==infinity EndTime=Infinity NoRefresh:=True}
);

update(
    q{Capitalizing -Infinity in Reservation StartTime},
    q{Reservation Query StartTime=="\"-infinity\"" Show:="\"Count(Id)\""},
    q{>},
    q{Reservation Modify StartTime=="\"-infinity\"" StartTime="\"-Infinity\"" NoRefresh:=True}
);

update(
    q{Capitalizing Infinity in Reservation EndTime},
    q{Reservation Query EndTime==infinity Show:="\"Count(Id)\""},
    q{>},
    q{Reservation Modify EndTime==infinity EndTime=Infinity NoRefresh:=True}
);

update(
    q{Modifying Allocation CreditLimit DefaultValue to be Infinity},
    q{Attribute Query Object==Allocation Name==CreditLimit DefaultValue==Infinity Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Modify Object==Allocation Name==CreditLimit DefaultValue=Infinity NoRefresh:=True}
);

update(
    q{Creating new UserServices RoleAction for Organization},
    q{RoleAction Query Role==UserServices Object==Organization Show:="\"Count(Role)\""},
    q{==},
    q{RoleAction Create Role=UserServices Object=Organization Name=ANY Instance=ANY NoRefresh:=True}
);

update(
    q{Modifying Deleted Attributes' DefaultValue to be False},
    q{Attribute Query Name==Deleted DefaultValue==NULL Show:="\"Count(Name)\""},
    q{>},
    q{Attribute Modify Name==Deleted DefaultValue==NULL DefaultValue=False NoRefresh:=True}
);

print "\n***Refreshing Cache***\n";
$cmd    = "mam-shell -v System Refresh";
$output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
$rc     = $? >> 8;
print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
if ($rc)
{
    $failures++;
    my $message = "Subcommand ($cmd) failed with rc=$rc";
    $message .= ": $output" if $output;
    warn($message);
    push @messages, $message;
}
else
{
    $successes++;
}

################################################################################
# Transform Accounting Tables
################################################################################

print "\n***Updating ChargeRate table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==ChargeRate Name==Value Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==ChargeRate Name==Instance Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});

    $sql = "UPDATE g_charge_rate SET g_value=g_instance";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql = "UPDATE g_charge_rate_log SET g_value=g_instance";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Updating QuoteChargeRate table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==QuoteChargeRate Name==Value Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==QuoteChargeRate Name==Instance Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});

    $sql = "UPDATE g_quote_charge_rate SET g_value=g_instance";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql = "UPDATE g_quote_charge_rate_log SET g_value=g_instance";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

################################################################################
# Cleaning up old Metadata and Updating Indexes
################################################################################

print "\n***Cleaning up old Metadata and Updating Indexes***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q System Query Name==\\\"Moab Accounting Manager\\\" Version==\\\"7.1\\\" Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ == 0
    )
    && ! $failures
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 0});

    # ChargeRate
    $sql =
      "DELETE FROM g_attribute WHERE g_object='ChargeRate' AND g_name='Instance'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='ChargeRate' AND g_name='Instance'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_charge_rate DROP COLUMN g_instance";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_charge_rate_log DROP COLUMN g_instance";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # QuoteChargeRate
    $sql =
      "DELETE FROM g_attribute WHERE g_object='QuoteChargeRate' AND g_name='Instance'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='QuoteChargeRate' AND g_name='Instance'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_quote_charge_rate DROP COLUMN g_instance";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_quote_charge_rate_log DROP COLUMN g_instance";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Transaction
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Transaction' AND g_name='Name'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Transaction' AND g_name='Name'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_transaction DROP COLUMN g_name";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_transaction_log DROP COLUMN g_name";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # User
    $sql =
      "DELETE FROM g_attribute WHERE g_object='User' AND g_name='Organization'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='User' AND g_name='Organization'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_user DROP COLUMN g_organization";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_user_log DROP COLUMN g_organization";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Indexes
    print "\n***Updating Indexes***\n";
    $sql =
      "CREATE INDEX g_reservation_instance_idx ON g_reservation (g_instance)";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql =
      "CREATE INDEX g_quote_time_idx ON g_quote (g_start_time, g_end_time)";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "DROP INDEX g_charge_rate_type_idx";
    $sql .= ' ON g_charge_rate' if 'Pg' eq 'mysql';
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "DROP INDEX g_quote_charge_rate_type_idx";
    $sql .= ' ON g_quote_charge_rate' if 'Pg' eq 'mysql';
    print "Executing sql: $sql\n";
    $dbh->do($sql);
}

################################################################################
# Refresh Cache and Update System Information
################################################################################

print "\n***Refreshing Cache***\n";
$cmd    = "mam-shell -v System Refresh";
$output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
$rc     = $? >> 8;
print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
if ($rc)
{
    $failures++;
    my $message = "Subcommand ($cmd) failed with rc=$rc";
    $message .= ": $output" if $output;
    warn($message);
    push @messages, $message;
}
else
{
    $successes++;
}

unless ($failures)
{
    update(
        q{Updating System Information},
        q{System Query Version==\"7.1\" Show:="\"Count(Name)\""},
        q{==}, q{System Modify Version=\"7.1\"}
    );
}

################################################################################
# Display Results
################################################################################

print "\n";
print "$successes updates were successfully applied.\n";
print "$ignores updates were skipped having been already applied.\n";
print "$failures updates failed.\n";
if (@messages)
{
    print "The following is a list of failures:\n";
    print join "\n", @messages;
}

exit 0;
