#!/usr/bin/perl -w
use strict;

use lib qw(/opt/mam/lib /opt/mam/lib/perl5);
use Config::Tiny;
use DBI;
use MAM;
use MAM::Global;

my ($cmd, $output, $rc, $sql, $count);
my ($successes, $failures, $ignores) = (0, 0, 0);
my @messages   = ();
$ENV{PATH} = "/opt/mam/bin:/opt/mam/sbin:$ENV{PATH}";

# Read the Config File
my $configFile = "/opt/mam/etc/mam-server.conf";
my $configTiny = Config::Tiny->new;
$configTiny = Config::Tiny->read($configFile)
  or die(
    "Unable to open config file ($configFile) for reading: $Config::Tiny::errstr.\nYou may need to set the \$configFile variable in this script.\n"
  );
$serverConfig = $configTiny->{_};
my $datasource = $serverConfig->{'database.datasource'};
my $dbuser     = $serverConfig->{'database.user'} || (getpwuid($<))[0];
my $dbpasswd   = $serverConfig->{'database.password'};

################################################################################
# Display warning and prompt to proceed
################################################################################

my $warning = <<"ENDOFWARNING";
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! IMPORTANT !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

This script will attempt to convert an existing Moab Accounting Manager 7.1 database (using the datasource ($datasource) referenced in /opt/mam/etc/mam-server.conf) into a Moab Accounting Manager 7.2 database. It will make additions, modifications and deletions to your Moab Accounting Manager 7.1 schema and data to match the new base Moab Accounting Manager 7.2 schema and data values.

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! BACKUP YOUR DATA !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

You should MAKE A BACKUP of your Moab Accounting Manager database before running this script. If you proceed without a backup, a failure in this script could irreversibly corrupt the integrity and data of your Moab Accounting Manager database.

ENDOFWARNING

print $warning;
print "Do you want to proceed? [y/N] ";
my $answer = <STDIN>;
exit 0 unless ($answer =~ /^y(es)?$/i);

# ----------------------------------------------------------------------------
# update($description, $test, $comparison, $action);
# ----------------------------------------------------------------------------

# Update the object metadata via mam-shell
sub update
{
    my ($description, $test, $comparison, $action) = @_;
    print "\n***$description***\n";
    if (chomp($_ = qx{mam-shell -q $test}) && eval "\$_ $comparison 0")
    {
        $cmd    = "mam-shell -v $action";
        $output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
        $rc     = $? >> 8;
        print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
        if ($rc)
        {
            $failures++;
            my $message = "Subcommand ($cmd) failed with rc=$rc";
            $message .= ": $output" if $output;
            warn($message);
            push @messages, $message;
        }
        else
        {
            $successes++;
        }
    }
    else
    {
        print "Already applied\n";
        $ignores++;
    }
}

################################################################################
# Create New Base Action
################################################################################

update(
    q{Creating new System Prune Action},
    q{Action Query Object==System Name==Prune Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=System Name=Prune Display=False NoRefresh:=True}
);

################################################################################
# Create and Update Accounting Objects and Attributes -- Stage 1
################################################################################

update(
    q{Modifying Transaction Delta Attribute to adjust Description},
    q{Attribute Query Object==Transaction Name==Delta Description=="\"Fund Delta\"" Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Modify Object==Transaction Name==Delta Description="\"Fund Delta\"" NoRefresh:=True}
);

update(
    q{Creating new Transaction Fund Attribute},
    q{Attribute Query Object==Transaction Name==Fund Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Transaction Name=Fund DataType=Integer Sequence=240 Description="\"Fund Id\"" NoRefresh:=True}
);

update(
    q{Creating new Fund Object},
    q{Object Query Name==Fund Show:="\"Count(Name)\""},
    q{==},
    q{Object Create Name=Fund Description="\"Fund\""}
);

update(
    q{Creating new Fund Id Attribute},
    q{Attribute Query Object==Fund Name==Id Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Fund Name=Id PrimaryKey=True DataType=AutoGen Description="\"Fund Id\"" NoRefresh:=True}
);

update(
    q{Creating new Fund Name Attribute},
    q{Attribute Query Object==Fund Name==Name Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Fund Name=Name DataType=String Description="\"Fund Name\"" NoRefresh:=True}
);

update(
    q{Creating new Fund Priority Attribute},
    q{Attribute Query Object==Fund Name==Priority Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Fund Name=Priority DataType=Integer DefaultValue=0 Description="\"Fund Priority\"" NoRefresh:=True}
);

update(
    q{Creating new Fund Description Attribute},
    q{Attribute Query Object==Fund Name==Description Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Fund Name=Description Sequence=900 DataType=String Description="\"Description\"" NoRefresh:=True}
);

update(
    q{Creating new Fund Withdraw Action},
    q{Action Query Object==Fund Name==Withdraw Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=Fund Name=Withdraw Display=True Description=Withdraw NoRefresh:=True}
);

update(
    q{Creating new Fund Balance Action},
    q{Action Query Object==Fund Name==Balance Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=Fund Name=Balance Display=True Description=Balance NoRefresh:=True}
);

update(
    q{Creating new Fund Deposit Action},
    q{Action Query Object==Fund Name==Deposit Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=Fund Name=Deposit Display=True Description=Deposit NoRefresh:=True}
);

update(
    q{Creating new Fund Transfer Action},
    q{Action Query Object==Fund Name==Transfer Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=Fund Name=Transfer Display=True Description=Transfer NoRefresh:=True}
);

update(
    q{Creating new Fund Reset Action},
    q{Action Query Object==Fund Name==Reset Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=Fund Name=Reset Display=True Description=Reset NoRefresh:=True}
);

update(
    q{Modifying Fund Actions to set Display Attribute to True},
    q{Action Query Object==Fund "(Name==Create || Name==Query || Name==Modify || Name==Delete)" Display!=True Show:="\"Count(Name)\""},
    q{>},
    q{Action Modify Object==Fund "(Name==Create || Name==Query || Name==Modify || Name==Delete)" Display=True NoRefresh:=True}
);

update(
    q{Creating new FundFund Object},
    q{Object Query Name==FundFund Show:="\"Count(Name)\""},
    q{==},
    q{Object Create Name=FundFund Association=True Parent=Fund Child=Fund Description="\"Fund Deposit Linkage\""}
);

update(
    q{Creating new FundFund Fund Attribute},
    q{Attribute Query Object==FundFund Name==Fund Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=FundFund Name=Fund PrimaryKey=True Values=@Fund DataType=Integer Description="\"Parent Fund Id\"" NoRefresh:=True}
);

update(
    q{Creating new FundFund Id Attribute},
    q{Attribute Query Object==FundFund Name==Id Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=FundFund Name=Id PrimaryKey=True Values=@Fund DataType=Integer Description="\"Child Fund Id\"" NoRefresh:=True}
);

update(
    q{Creating new FundFund DepositShare Attribute},
    q{Attribute Query Object==FundFund Name==DepositShare Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=FundFund Name=DepositShare Required=True DataType=Integer DefaultValue=0 Description="\"Deposit Share\"" NoRefresh:=True}
);

update(
    q{Creating new FundFund Overflow Attribute},
    q{Attribute Query Object==FundFund Name==Overflow Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=FundFund Name=Overflow DataType=Boolean DefaultValue=False Description="\"Do descendant charges overflow into this Fund?\"" NoRefresh:=True}
);

update(
    q{Modifying Constraint Description},
    q{Object Query Name==Constraint Description=="\"Fund Constraint\"" Show:="\"Count(Name)\""},
    q{==},
    q{Object Modify Name==Constraint Description="\"Fund Constraint\"" NoRefresh:=True}
);

update(
    q{Creating new Constraint Fund Attribute},
    q{Attribute Query Object==Constraint Name==Fund Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Constraint Name=Fund Required=True Fixed=True Values=@Fund DataType=Integer Sequence=20 Description="\"Fund Id\"" NoRefresh:=True}
);

update(
    q{Creating new Allocation Fund Attribute},
    q{Attribute Query Object==Allocation Name==Fund Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Allocation Name=Fund Required=True Fixed=True Values=@Fund DataType=Integer Sequence=20 Description="\"Fund Id\"" NoRefresh:=True}
);

update(
    q{Creating new ReservationAllocation Fund Attribute},
    q{Attribute Query Object==ReservationAllocation Name==Fund Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=ReservationAllocation Name=Fund Required=True Values=@Fund DataType=Integer Description="\"Fund Id\"" NoRefresh:=True}
);

update(
    q{Creating new Payment UpdateEvent Attribute},
    q{Attribute Query Object==Payment Name==UpdateEvent Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Payment Name=UpdateEvent DataType=Integer Values=@Event Description="\"Update Event\"" NoRefresh:=True}
);

update(
    q{Creating new Payment Update Action},
    q{Action Query Object==Payment Name==Update Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=Payment Name=Update Display=False Description="\"Update Payment Charges\"" NoRefresh:=True}
);

update(
    q{Modifying PaymentComponent Type Values to Begin,End,Update},
    q{Attribute Query Object==PaymentComponent Name==Type Values=="\"(Begin,End,Update)\"" Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Modify Object==PaymentComponent Name==Type Values="\"(Begin,End,Update)\"" NoRefresh:=True}
);

update(
    q{Modifying PaymentComponent Use Type to Update},
    q{PaymentComponent Query Type==Use Show:="\"Count(Id)\""},
    q{>},
    q{PaymentComponent Modify Type==Use Type=Update NoRefresh:=True}
);

#update(
#    q{Creating new Anonymous FundFund Query RoleAction},
#    q{RoleAction Query Role==Anonymous Object==FundFund Name==Query Show:="\"Count(Name)\""},
#    q{==},
#    q{RoleAction Create Role=Anonymous Object=FundFund Name=Query Instance=ANY NoRefresh:=True}
#);

#update(
#    q{Creating new OVERRIDE Fund Balance RoleAction},
#    q{RoleAction Query Role==OVERRIDE Object==Fund Name==Balance Show:="\"Count(Name)\""},
#    q{==},
#    q{RoleAction Create Role=OVERRIDE Object=Fund Name=Balance Instance=ANY NoRefresh:=True}
#);

#update(
#    q{Creating new OVERRIDE Fund Query RoleAction},
#    q{RoleAction Query Role==OVERRIDE Object==Fund Name==Query Show:="\"Count(Name)\""},
#    q{==},
#    q{RoleAction Create Role=OVERRIDE Object=Fund Name=Query Instance=ANY NoRefresh:=True}
#);

print "\n***Refreshing Cache***\n";
$cmd    = "mam-shell -v System Refresh";
$output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
$rc     = $? >> 8;
print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
if ($rc)
{
    $failures++;
    my $message = "Subcommand ($cmd) failed with rc=$rc";
    $message .= ": $output" if $output;
    warn($message);
    push @messages, $message;
}
else
{
    $successes++;
}

################################################################################
# Transform Accounting Tables -- Stage 1
################################################################################

print "\n***Updating Transaction table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Transaction Name==Account DataType==Integer Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Transaction Name==Fund Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});

    $sql = "UPDATE g_transaction SET g_fund=g_account";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql = "UPDATE g_transaction_log SET g_fund=g_account";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Populating new Fund table from old Account table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Object Query Name==Fund Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Object Query Name==Account Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Account Name==Id Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});
    my $sth;

    $sql =
      "INSERT INTO g_fund (g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_id,g_name,g_priority,g_description) SELECT g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_id,g_name,g_priority,g_description FROM g_account";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Inserted $count lines\n";
    $sql =
      "INSERT INTO g_fund_log (g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_id,g_name,g_priority,g_description) SELECT g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_id,g_name,g_priority,g_description FROM g_account_log";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Inserted $count lines\n";

    print "\n***Updating key generator for Fund***\n";
    $sth = $dbh->prepare(
        "SELECT g_next_id FROM g_key_generator WHERE g_name='Account'");
    $sth->execute();
    my ($nextId) = $sth->fetchrow_array();
    if (defined $nextId)
    {
        $sql =
          "UPDATE g_key_generator set g_next_id=$nextId WHERE g_name='Fund'";
        print "Executing sql: $sql\n";
        $count = $dbh->do($sql);
        print "Updated $count lines\n";
        $sql = "DELETE from g_key_generator where g_name='Account'";
        print "Executing sql: $sql\n";
        $count = $dbh->do($sql);
        print "Updated $count lines\n";
    }

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Populating new FundFund table from old AccountAccount table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Object Query Name==FundFund Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Object Query Name==AccountAccount Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});
    my $sth;

    $sql =
      "INSERT INTO g_fund_fund (g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_fund,g_id,g_deposit_share,g_overflow) SELECT g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,cast(g_account AS integer),cast(g_id AS integer),g_deposit_share,g_overflow FROM g_account_account";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Inserted $count lines\n";
    $sql =
      "INSERT INTO g_fund_fund_log (g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_fund,g_id,g_deposit_share,g_overflow) SELECT g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,cast(g_account AS integer),cast(g_id AS integer),g_deposit_share,g_overflow FROM g_account_account_log";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Inserted $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Updating Constraint table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Constraint Name==Account Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Constraint Name==Fund Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});

    $sql = "UPDATE g_constraint SET g_fund=g_account";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql = "UPDATE g_constraint_log SET g_fund=g_account";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Updating Allocation table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Allocation Name==Account Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Allocation Name==Fund Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});

    $sql = "UPDATE g_allocation SET g_fund=g_account";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql = "UPDATE g_allocation_log SET g_fund=g_account";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Updating ReservationAllocation table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==ReservationAllocation Name==Account Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==ReservationAllocation Name==Fund Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});

    $sql = "UPDATE g_reservation_allocation SET g_fund=g_account";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql = "UPDATE g_reservation_allocation_log SET g_fund=g_account";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Updating RoleAction table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q RoleAction Query Object==Account Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Account Name==Id Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});

    $sql = "UPDATE g_role_action SET g_object='Fund' WHERE g_object='Account'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql =
      "UPDATE g_role_action_log SET g_object='Fund' WHERE g_object='Account'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql =
      "UPDATE g_role_action SET g_object='FundFund' WHERE g_object='AccountAccount'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql =
      "UPDATE g_role_action_log SET g_object='FundFund' WHERE g_object='AccountAccount'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Updating Payment table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Payment Name==UseEvent Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Payment Name==UpdateEvent Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});

    $sql = "UPDATE g_payment SET g_update_event=g_use_event";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql = "UPDATE g_payment_log SET g_update_event=g_use_event";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

################################################################################
# Delete old Accounting Objects and Attributes -- Stage 1
################################################################################

# Don't do deletions if there are failures above or we have
# already changed to new version (so we don't unnecessarily repeat)
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q System Query Version==\\\"7.2\\\" Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ == 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Account Name==Id Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && ! $failures
  )
{
    update(
        q{Deleting old Account RoleActions},
        q{RoleAction Query Object==Account Show:="\"Count(Name)\""},
        q{>},
        q{RoleAction Delete Object==Account NoRefresh:=True}
    );

    update(
        q{Deleting old AccountAccount RoleActions},
        q{RoleAction Query Object==AccountAccount Show:="\"Count(Name)\""},
        q{>},
        q{RoleAction Delete Object==AccountAccount NoRefresh:=True}
    );
}

################################################################################
# Cleaning up old Metadata and Updating Indexes -- Stage 1
################################################################################

# Don't do deletions if there are failures above or we have
# already changed to new version (so we don't unnecessarily repeat)
print "\n***Cleaning up old Metadata and Updating Indexes***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q System Query Version==\\\"7.2\\\" Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ == 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Account Name==Id Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && ! $failures
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 0});

    # Account
    $sql = "DELETE FROM g_attribute WHERE g_object='Account'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_attribute_log WHERE g_object='Account'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action WHERE g_object='Account'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action_log WHERE g_object='Account'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object WHERE g_name='Account'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object_log WHERE g_name='Account'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DROP TABLE g_account";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "DROP TABLE g_account_log";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # AccountAccount
    $sql = "DELETE FROM g_attribute WHERE g_object='AccountAccount'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_attribute_log WHERE g_object='AccountAccount'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action WHERE g_object='AccountAccount'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action_log WHERE g_object='AccountAccount'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object WHERE g_name='AccountAccount'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object_log WHERE g_name='AccountAccount'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DROP TABLE g_account_account";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "DROP TABLE g_account_account_log";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # AccountUser (in case remnants from Gold were not previously cleaned up)
    $sql = "DELETE FROM g_attribute WHERE g_object='AccountUser'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_attribute_log WHERE g_object='AccountUser'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action WHERE g_object='AccountUser'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action_log WHERE g_object='AccountUser'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object WHERE g_name='AccountUser'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object_log WHERE g_name='AccountUser'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DROP TABLE g_account_user";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "DROP TABLE g_account_user_log";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Allocation
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Allocation' AND g_name='Account'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Allocation' AND g_name='Account'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_allocation DROP COLUMN g_account";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_allocation_log DROP COLUMN g_account";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Constraint
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Constraint' AND g_name='Account'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Constraint' AND g_name='Account'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_constraint DROP COLUMN g_account";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_constraint_log DROP COLUMN g_account";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

  #    # Machine (We should not delete machine if customer might already use it)
  #    $sql = "DELETE FROM g_attribute WHERE g_object='Machine'";
  #    print "Executing sql: $sql\n";
  #    $successes += $dbh->do($sql);
  #    $sql = "DELETE FROM g_attribute_log WHERE g_object='Machine'";
  #    print "Executing sql: $sql\n";
  #    $successes += $dbh->do($sql);
  #    $sql = "DELETE FROM g_action WHERE g_object='Machine'";
  #    print "Executing sql: $sql\n";
  #    $successes += $dbh->do($sql);
  #    $sql = "DELETE FROM g_action_log WHERE g_object='Machine'";
  #    print "Executing sql: $sql\n";
  #    $successes += $dbh->do($sql);
  #    $sql = "DELETE FROM g_object WHERE g_name='Machine'";
  #    print "Executing sql: $sql\n";
  #    $successes += $dbh->do($sql);
  #    $sql = "DELETE FROM g_object_log WHERE g_name='Machine'";
  #    print "Executing sql: $sql\n";
  #    $successes += $dbh->do($sql);
  #    $sql = "DROP TABLE g_machine";
  #    print "Executing sql: $sql\n";
  #    $dbh->do($sql);
  #    $sql = "DROP TABLE g_machine_log";
  #    print "Executing sql: $sql\n";
  #    $dbh->do($sql);

    # Payment
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Payment' AND g_name='UseEvent'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Payment' AND g_name='UseEvent'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action WHERE g_object='Payment' AND g_name='Use'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action_log WHERE g_object='Payment' AND g_name='Use'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_payment DROP COLUMN g_use_event";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_payment_log DROP COLUMN g_use_event";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # ReservationAllocation
    $sql =
      "DELETE FROM g_attribute WHERE g_object='ReservationAllocation' AND g_name='Account'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='ReservationAllocation' AND g_name='Account'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_reservation_allocation DROP COLUMN g_account";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_reservation_allocation_log DROP COLUMN g_account";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Transaction
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Transaction' AND g_name='Account'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Transaction' AND g_name='Account'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_transaction DROP COLUMN g_account";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_transaction_log DROP COLUMN g_account";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    print "\n***Updating Indexes***\n";
    $sql = "CREATE INDEX g_allocation_fund_idx ON g_allocation (g_fund)";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "CREATE INDEX g_transaction_fund_idx ON g_transaction (g_fund)";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    print "\n***Refreshing Cache***\n";
    $cmd    = "mam-shell -v System Refresh";
    $output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
    $rc     = $? >> 8;
    print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
    if ($rc)
    {
        $failures++;
        my $message = "Subcommand ($cmd) failed with rc=$rc";
        $message .= ": $output" if $output;
        warn($message);
        push @messages, $message;
    }
    else
    {
        $successes++;
    }
}

################################################################################
# Create and Update Accounting Objects and Attributes -- Stage 2
################################################################################

update(
    q{Creating new Transaction Account Attribute},
    q{Attribute Query Object==Transaction Name==Account DataType==String Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Transaction Name=Account DataType=String Sequence=220 Description="\"Account Name\"" NoRefresh:=True}
);

update(
    q{Creating new Account Object},
    q{Object Query Name==Account Show:="\"Count(Name)\""},
    q{==},
    q{Object Create Name=Account AutoGen=True Description="\"Account\""}
);

update(
    q{Creating new Account Name Attribute},
    q{Attribute Query Object==Account Name==Name Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Account Name=Name PrimaryKey=True DataType=String Description="\"Account Name\"" NoRefresh:=True}
);

update(
    q{Creating new Account Active Attribute},
    q{Attribute Query Object==Account Name==Active Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Account Name=Active DataType=Boolean DefaultValue=True Description="\"Is the Account Active?\"" NoRefresh:=True}
);

update(
    q{Creating new Account Organization Attribute},
    q{Attribute Query Object==Account Name==Organization Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Account Name=Organization DataType=String Values="\"@!=Organization\"" Description="\"Organization\"" NoRefresh:=True}
);

update(
    q{Creating new Account Description Attribute},
    q{Attribute Query Object==Account Name==Description Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Account Name=Description Sequence=900 DataType=String Description="\"Description\"" NoRefresh:=True}
);

update(
    q{Modifying Account Actions to set Display Attribute to True},
    q{Action Query Object==Account "(Name==Create || Name==Query || Name==Modify || Name==Delete)" Display!=True Show:="\"Count(Name)\""},
    q{>},
    q{Action Modify Object==Account "(Name==Create || Name==Query || Name==Modify || Name==Delete)" Display=True NoRefresh:=True}
);

update(
    q{Creating new User DefaultAccount Attribute},
    q{Attribute Query Object==User Name==DefaultAccount Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=User Name=DefaultAccount DataType=String Values="\"@?=Account\"" Description="\"Default Account\"" NoRefresh:=True}
);

update(
    q{Creating new AccountUser Object},
    q{Object Query Name==AccountUser Show:="\"Count(Name)\""},
    q{==},
    q{Object Create Name=AccountUser Association=True Parent=Account Child=User Description="\"Membership mapping Users to Accounts\""}
);

update(
    q{Creating new AccountUser Account Attribute},
    q{Attribute Query Object==AccountUser Name==Account Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=AccountUser Name=Account PrimaryKey=True Values=@Account DataType=String Description="\"Parent Account Name\"" NoRefresh:=True}
);

update(
    q{Creating new AccountUser Name Attribute},
    q{Attribute Query Object==AccountUser Name==Name Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=AccountUser Name=Name PrimaryKey=True Values=@User DataType=String Description="\"Member User Name\"" NoRefresh:=True}
);

update(
    q{Creating new AccountUser Active Attribute},
    q{Attribute Query Object==AccountUser Name==Active Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=AccountUser Name=Active DataType=Boolean DefaultValue=True Description="\"Is this sub-account Active?\"" NoRefresh:=True}
);

update(
    q{Creating new AccountUser Admin Attribute},
    q{Attribute Query Object==AccountUser Name==Admin Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=AccountUser Name=Admin DataType=Boolean DefaultValue=False Description="\"Is this user an Account Administrator?\"" NoRefresh:=True}
);

update(
    q{Modifying AccountUser Query Action to set Display Attribute to True},
    q{Action Query Object==AccountUser Name==Query Display!=True Show:="\"Count(Name)\""},
    q{>},
    q{Action Modify Object==AccountUser Name==Query Display=True NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Account Attribute},
    q{Attribute Query Object==UsageRecord Name==Account Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Account DataType=String Description="\"Account Name\"" NoRefresh:=True}
);

#update(
#    q{Creating new UserServices AccountUser ANY RoleAction},
#    q{RoleAction Query Role==UserServices Object==AccountUser Name==ANY Show:="\"Count(Name)\""},
#    q{==},
#    q{RoleAction Create Role=UserServices Object=AccountUser Name=ANY Instance=ANY NoRefresh:=True}
#);

update(
    q{Creating new AccountAdmin Role},
    q{Role Query Name==AccountAdmin Show:="\"Count(Name)\""},
    q{==},
    q{Role Create Name=AccountAdmin Description="\"Can update or view an account they are admin for\"" NoRefresh:=True}
);

#update(
#    q{Creating new AccountAdmin Account ANY RoleAction},
#    q{RoleAction Query Role==AccountAdmin Object==Account Name==ANY Show:="\"Count(Name)\""},
#    q{==},
#    q{RoleAction Create Role=AccountAdmin Object=Account Name=ANY Instance=ADMIN NoRefresh:=True}
#);

#update(
#    q{Creating new Anonymous AccountUser Query RoleAction},
#    q{RoleAction Query Role==Anonymous Object==AccountUser Name==Query Show:="\"Count(Name)\""},
#    q{==},
#    q{RoleAction Create Role=Anonymous Object=AccountUser Name=Query Instance=ANY NoRefresh:=True}
#);

#update(
#    q{Creating new OVERRIDE Account Query RoleAction},
#    q{RoleAction Query Role==OVERRIDE Object==Account Name==Query Show:="\"Count(Name)\""},
#    q{==},
#    q{RoleAction Create Role=OVERRIDE Object=Account Name=Query Instance=ANY NoRefresh:=True}
#);

update(
    q{Modifying Constraint Names to replace Project with Account},
    q{Constraint Query Name==Project Show:="\"Count(Name)\""},
    q{>},
    q{Constraint Modify Name==Project Name=Account NoRefresh:=True}
);

print "\n***Refreshing Cache***\n";
$cmd    = "mam-shell -v System Refresh";
$output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
$rc     = $? >> 8;
print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
if ($rc)
{
    $failures++;
    my $message = "Subcommand ($cmd) failed with rc=$rc";
    $message .= ": $output" if $output;
    warn($message);
    push @messages, $message;
}
else
{
    $successes++;
}

################################################################################
# Transform Accounting Tables -- Stage 2
################################################################################

print "\n***Updating Transaction table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Transaction Name==Project Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Transaction Name==Account Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});

    $sql = "UPDATE g_transaction SET g_account=g_project";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql = "UPDATE g_transaction_log SET g_account=g_project";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Populating new Account table from old Project table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Object Query Name==Account Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Object Query Name==Project Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});
    my $sth;

    $sql =
      "INSERT INTO g_account (g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_name,g_active,g_organization,g_description) SELECT g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_name,g_active,g_organization,g_description FROM g_project";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Inserted $count lines\n";
    $sql =
      "INSERT INTO g_account_log (g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_name,g_active,g_organization,g_description) SELECT g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_name,g_active,g_organization,g_description FROM g_project_log";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Inserted $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Updating User table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==User Name==DefaultAccount Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==User Name==DefaultProject Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});

    $sql = "UPDATE g_user SET g_default_account=g_default_project";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql = "UPDATE g_user_log SET g_default_account=g_default_project";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Populating new AccountUser table from old ProjectUser table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Object Query Name==AccountUser Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Object Query Name==ProjectUser Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});
    my $sth;

    $sql =
      "INSERT INTO g_account_user (g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_account,g_name,g_active,g_admin) SELECT g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_project,g_name,g_active,g_admin FROM g_project_user";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Inserted $count lines\n";
    $sql =
      "INSERT INTO g_account_user_log (g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_account,g_name,g_active,g_admin) SELECT g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_project,g_name,g_active,g_admin FROM g_project_user_log";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Inserted $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Updating UsageRecord table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==UsageRecord Name==Project Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==UsageRecord Name==Account Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});

    $sql = "UPDATE g_usage_record SET g_account=g_project";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql = "UPDATE g_usage_record_log SET g_account=g_project";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Updating RoleAction table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q RoleAction Query Object==Project Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Project Name==Name Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});

    $sql =
      "UPDATE g_role_action SET g_role='AccountAdmin' WHERE g_role='ProjectAdmin'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql =
      "UPDATE g_role_action_log SET g_role='AccountAdmin' WHERE g_role='ProjectAdmin'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql =
      "UPDATE g_role_action SET g_object='Account' WHERE g_object='Project'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql =
      "UPDATE g_role_action_log SET g_object='Account' WHERE g_object='Project'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql =
      "UPDATE g_role_action SET g_object='AccountUser' WHERE g_object='ProjectUser'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql =
      "UPDATE g_role_action_log SET g_object='AccountUser' WHERE g_object='ProjectUser'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Updating RoleUser table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q RoleUser Query Role==AccountAdmin Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q RoleUser Query Role==ProjectAdmin Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});

    $sql =
      "UPDATE g_role_user SET g_role='AccountAdmin' WHERE g_role='ProjectAdmin'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql =
      "UPDATE g_role_user_log SET g_role='AccountAdmin' WHERE g_role='ProjectAdmin'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

################################################################################
# Delete old Accounting Objects and Attributes -- Stage 2
################################################################################

# Don't do deletions if there are failures above or we have
# already changed to new version (so we don't unnecessarily repeat)
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q System Query Version==\\\"7.2\\\" Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ == 0
    )
    && ! $failures
  )
{
    update(
        q{Deleting old ProjectAdmin RoleActions},
        q{RoleAction Query Role==ProjectAdmin Show:="\"Count(Name)\""},
        q{>},
        q{RoleAction Delete Role==ProjectAdmin NoRefresh:=True}
    );

    update(
        q{Deleting old ProjectAdmin RoleUsers},
        q{RoleUser Query Role==ProjectAdmin Show:="\"Count(Name)\""},
        q{>},
        q{RoleUser Delete Role==ProjectAdmin NoRefresh:=True}
    );

    update(
        q{Deleting old ProjectAdmin Role},
        q{Role Query Name==ProjectAdmin Show:="\"Count(Name)\""},
        q{>},
        q{Role Delete Name==ProjectAdmin NoRefresh:=True}
    );

    update(
        q{Deleting old Project RoleActions},
        q{RoleAction Query Object==Project Show:="\"Count(Name)\""},
        q{>},
        q{RoleAction Delete Object==Project NoRefresh:=True}
    );

    update(
        q{Deleting old ProjectUser RoleActions},
        q{RoleAction Query Object==ProjectUser Show:="\"Count(Name)\""},
        q{>},
        q{RoleAction Delete Object==ProjectUser NoRefresh:=True}
    );
}

################################################################################
# Cleaning up old Metadata and Updating Indexes -- Stage 2
################################################################################

# Don't do deletions if there are failures above or we have
# already changed to new version (so we don't unnecessarily repeat)
print "\n***Cleaning up old Metadata and Updating Indexes***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q System Query Version==\\\"7.2\\\" Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ == 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Project Name==Name Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && ! $failures
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 0});

    # Project
    $sql = "DELETE FROM g_attribute WHERE g_object='Project'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_attribute_log WHERE g_object='Project'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action WHERE g_object='Project'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action_log WHERE g_object='Project'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object WHERE g_name='Project'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object_log WHERE g_name='Project'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DROP TABLE g_project";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "DROP TABLE g_project_log";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # ProjectUser
    $sql = "DELETE FROM g_attribute WHERE g_object='ProjectUser'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_attribute_log WHERE g_object='ProjectUser'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action WHERE g_object='ProjectUser'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action_log WHERE g_object='ProjectUser'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object WHERE g_name='ProjectUser'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object_log WHERE g_name='ProjectUser'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DROP TABLE g_project_user";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "DROP TABLE g_project_user_log";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Transaction
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Transaction' AND g_name='Project'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Transaction' AND g_name='Project'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_transaction DROP COLUMN g_project";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_transaction_log DROP COLUMN g_project";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # UsageRecord
    $sql =
      "DELETE FROM g_attribute WHERE g_object='UsageRecord' AND g_name='Project'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='UsageRecord' AND g_name='Project'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_usage_record DROP COLUMN g_project";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_usage_record_log DROP COLUMN g_project";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # User
    $sql =
      "DELETE FROM g_attribute WHERE g_object='User' AND g_name='DefaultProject'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='User' AND g_name='DefaultProject'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_user DROP COLUMN g_default_project";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_user_log DROP COLUMN g_default_project";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    print "\n***Refreshing Cache***\n";
    $cmd    = "mam-shell -v System Refresh";
    $output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
    $rc     = $? >> 8;
    print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
    if ($rc)
    {
        $failures++;
        my $message = "Subcommand ($cmd) failed with rc=$rc";
        $message .= ": $output" if $output;
        warn($message);
        push @messages, $message;
    }
    else
    {
        $successes++;
    }
}

################################################################################
# Create and Update Accounting Objects and Attributes -- Stage 3
################################################################################

update(
    q{Creating new Lien Object},
    q{Object Query Name==Lien Show:="\"Count(Name)\""},
    q{==},
    q{Object Create Name=Lien Description="\"Lien\""}
);

update(
    q{Creating new Lien Id Attribute},
    q{Attribute Query Object==Lien Name==Id Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Lien Name=Id PrimaryKey=True DataType=AutoGen Description="\"Lien Id\"" NoRefresh:=True}
);

update(
    q{Creating new Lien Instance Attribute},
    q{Attribute Query Object==Lien Name==Instance Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Lien Name=Instance DataType=String Description="\"Instance Name\"" NoRefresh:=True}
);

update(
    q{Creating new Lien UsageRecord Attribute},
    q{Attribute Query Object==Lien Name==UsageRecord Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Lien Name=UsageRecord DataType=Integer Values=@UsageRecord Description="\"Usage Record Id\"" NoRefresh:=True}
);

update(
    q{Creating new Lien StartTime Attribute},
    q{Attribute Query Object==Lien Name==StartTime Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Lien Name=StartTime DataType=TimeStamp DefaultValue="\"-Infinity\"" Description="\"When does this Lien start?\"" NoRefresh:=True}
);

update(
    q{Creating new Lien EndTime Attribute},
    q{Attribute Query Object==Lien Name==EndTime Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Lien Name=EndTime DataType=TimeStamp DefaultValue="Infinity" Description="\"When does this Lien expire?\"" NoRefresh:=True}
);

update(
    q{Creating new Lien Duration Attribute},
    q{Attribute Query Object==Lien Name==Duration Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Lien Name=Duration DataType=Integer Description="\"WallTime Estimate\"" NoRefresh:=True}
);

update(
    q{Creating new Lien Description Attribute},
    q{Attribute Query Object==Lien Name==Description Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Lien Name=Description Sequence=900 DataType=String Description="\"Description\"" NoRefresh:=True}
);

update(
    q{Modifying Lien Actions to set Display Attribute to True},
    q{Action Query Object==Lien "(Name==Query || Name==Delete)" Display!=True Show:="\"Count(Name)\""},
    q{>},
    q{Action Modify Object==Lien "(Name==Query || Name==Delete)" Display=True NoRefresh:=True}
);

update(
    q{Creating new LienAllocation Object},
    q{Object Query Name==LienAllocation Show:="\"Count(Name)\""},
    q{==},
    q{Object Create Name=LienAllocation Association=True Parent=Lien Child=Allocation Description="\"Lien Allocation Association\""}
);

update(
    q{Creating new LienAllocation Lien Attribute},
    q{Attribute Query Object==LienAllocation Name==Lien Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=LienAllocation Name=Lien PrimaryKey=True Values=@Lien DataType=Integer Description="\"Parent Lien Id\"" NoRefresh:=True}
);

update(
    q{Creating new LienAllocation Id Attribute},
    q{Attribute Query Object==LienAllocation Name==Id Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=LienAllocation Name=Id PrimaryKey=True Values=@Allocation DataType=Integer Description="\"Child Allocation Id\"" NoRefresh:=True}
);

update(
    q{Creating new LienAllocation Fund Attribute},
    q{Attribute Query Object==LienAllocation Name==Fund Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=LienAllocation Name=Fund Required=True Values=@Fund DataType=Integer Description="\"Fund Id\"" NoRefresh:=True}
);

update(
    q{Creating new LienAllocation Amount Attribute},
    q{Attribute Query Object==LienAllocation Name==Amount Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=LienAllocation Name=Amount DataType=Currency Required=True Description="\"Amount\"" NoRefresh:=True}
);

update(
    q{Modifying LienAllocation Query Action to set Display Attribute to True},
    q{Action Query Object==LienAllocation Name==Query Display!=True Show:="\"Count(Name)\""},
    q{>},
    q{Action Modify Object==LienAllocation Name==Query Display=True NoRefresh:=True}
);

update(
    q{Modifying UsageRecord Reserve Action to adjust Description},
    q{Action Query Object==UsageRecord Name==Reserve Description=="\"Make a Usage Lien\"" Show:="\"Count(Name)\""},
    q{==},
    q{Action Modify Object==UsageRecord Name==Reserve Description="\"Make a Usage Lien\"" NoRefresh:=True}
);

print "\n***Refreshing Cache***\n";
$cmd    = "mam-shell -v System Refresh";
$output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
$rc     = $? >> 8;
print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
if ($rc)
{
    $failures++;
    my $message = "Subcommand ($cmd) failed with rc=$rc";
    $message .= ": $output" if $output;
    warn($message);
    push @messages, $message;
}
else
{
    $successes++;
}

################################################################################
# Transform Accounting Tables -- Stage 3
################################################################################

print "\n***Populating new Lien table from old Reservation table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Object Query Name==Lien Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Object Query Name==Reservation Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});
    my $sth;

    $sql =
      "INSERT INTO g_lien (g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_id,g_instance,g_usage_record,g_start_time,g_end_time,g_duration,g_description) SELECT g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_id,g_instance,g_usage_record,g_start_time,g_end_time,g_duration,g_description FROM g_reservation";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Inserted $count lines\n";
    $sql =
      "INSERT INTO g_lien_log (g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_id,g_instance,g_usage_record,g_start_time,g_end_time,g_duration,g_description) SELECT g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_id,g_instance,g_usage_record,g_start_time,g_end_time,g_duration,g_description FROM g_reservation_log";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Inserted $count lines\n";

    print "\n***Updating key generator for Lien***\n";
    $sth = $dbh->prepare(
        "SELECT g_next_id FROM g_key_generator WHERE g_name='Reservation'");
    $sth->execute();
    my ($nextId) = $sth->fetchrow_array();
    if (defined $nextId)
    {
        $sql =
          "UPDATE g_key_generator set g_next_id=$nextId WHERE g_name='Lien'";
        print "Executing sql: $sql\n";
        $count = $dbh->do($sql);
        print "Updated $count lines\n";
        $sql = "DELETE from g_key_generator where g_name='Reservation'";
        print "Executing sql: $sql\n";
        $count = $dbh->do($sql);
        print "Updated $count lines\n";
    }

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

print
  "\n***Populating new LienAllocation table from old ReservationAllocation table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Object Query Name==LienAllocation Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Object Query Name==ReservationAllocation Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});
    my $sth;

    $sql =
      "INSERT INTO g_lien_allocation (g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_lien,g_id,g_fund,g_amount) SELECT g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_reservation,g_id,g_fund,g_amount FROM g_reservation_allocation";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Inserted $count lines\n";
    $sql =
      "INSERT INTO g_lien_allocation_log (g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_lien,g_id,g_fund,g_amount) SELECT g_creation_time,g_modification_time,g_deleted,g_request_id,g_transaction_id,g_reservation,g_id,g_fund,g_amount FROM g_reservation_allocation_log";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Inserted $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Updating RoleAction table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q RoleAction Query Object==Reservation Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Reservation Name==Id Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});

    $sql =
      "UPDATE g_role_action SET g_object='Lien' WHERE g_object='Reservation'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql =
      "UPDATE g_role_action_log SET g_object='Lien' WHERE g_object='Reservation'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql =
      "UPDATE g_role_action SET g_object='LienAllocation' WHERE g_object='ReservationAllocation'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql =
      "UPDATE g_role_action_log SET g_object='LienAllocation' WHERE g_object='ReservationAllocation'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

################################################################################
# Delete old Accounting Objects and Attributes -- Stage 3
################################################################################

# Don't do deletions if there are failures above or we have
# already changed to new version (so we don't unnecessarily repeat)
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q System Query Version==\\\"7.2\\\" Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ == 0
    )
    && ! $failures
  )
{
    update(
        q{Deleting old Reservation RoleActions},
        q{RoleAction Query Object==Reservation Show:="\"Count(Name)\""},
        q{>},
        q{RoleAction Delete Object==Reservation NoRefresh:=True}
    );

    update(
        q{Deleting old ReservationAllocation RoleActions},
        q{RoleAction Query Object==ReservationAllocation Show:="\"Count(Name)\""},
        q{>},
        q{RoleAction Delete Object==ReservationAllocation NoRefresh:=True}
    );
}

################################################################################
# Cleaning up old Metadata and Updating Indexes -- Stage 3
################################################################################

# Don't do deletions if there are failures above or we have
# already changed to new version (so we don't unnecessarily repeat)
print "\n***Cleaning up old Metadata and Updating Indexes***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q System Query Version==\\\"7.2\\\" Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ == 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Reservation Name==Id Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && ! $failures
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 0});

    # Reservation
    $sql = "DELETE FROM g_attribute WHERE g_object='Reservation'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_attribute_log WHERE g_object='Reservation'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action WHERE g_object='Reservation'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action_log WHERE g_object='Reservation'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object WHERE g_name='Reservation'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object_log WHERE g_name='Reservation'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DROP TABLE g_reservation";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "DROP TABLE g_reservation_log";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # ReservationAllocation
    $sql = "DELETE FROM g_attribute WHERE g_object='ReservationAllocation'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_attribute_log WHERE g_object='ReservationAllocation'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action WHERE g_object='ReservationAllocation'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_action_log WHERE g_object='ReservationAllocation'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object WHERE g_name='ReservationAllocation'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DELETE FROM g_object_log WHERE g_name='ReservationAllocation'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "DROP TABLE g_reservation_allocation";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "DROP TABLE g_reservation_allocation_log";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    print "\n***Updating Indexes***\n";
    $sql =
      "CREATE INDEX g_id_not_deleted_idx ON g_lien (g_deleted) WHERE NOT g_deleted='True'";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql =
      "CREATE INDEX g_lien_alloc_fund_where_idx ON g_lien_allocation (g_fund) WHERE g_deleted!='True'";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "CREATE INDEX g_lien_instance_idx ON g_lien (g_instance)";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "CREATE INDEX g_lien_time_idx ON g_lien (g_start_time, g_end_time)";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql =
      "CREATE INDEX g_lien_allocation_fund_idx ON g_lien_allocation (g_fund)";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    print "\n***Refreshing Cache***\n";
    $cmd    = "mam-shell -v System Refresh";
    $output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
    $rc     = $? >> 8;
    print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
    if ($rc)
    {
        $failures++;
        my $message = "Subcommand ($cmd) failed with rc=$rc";
        $message .= ": $output" if $output;
        warn($message);
        push @messages, $message;
    }
    else
    {
        $successes++;
    }
}

################################################################################
# Refresh Cache and Update System Information
################################################################################

print "\n***Refreshing Cache***\n";
$cmd    = "mam-shell -v System Refresh";
$output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
$rc     = $? >> 8;
print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
if ($rc)
{
    $failures++;
    my $message = "Subcommand ($cmd) failed with rc=$rc";
    $message .= ": $output" if $output;
    warn($message);
    push @messages, $message;
}
else
{
    $successes++;
}

unless ($failures)
{
    update(
        q{Updating System Information},
        q{System Query Version==\"7.2\" Show:="\"Count(Name)\""},
        q{==},
        q{System Modify Version=\"7.2\"}
    );
}

################################################################################
# Display Results
################################################################################

print "\n";
print "$successes updates were successfully applied.\n";
print "$ignores updates were skipped having been already applied.\n";
print "$failures updates failed.\n";
if (@messages)
{
    print "The following is a list of failures:\n";
    print join "\n", @messages;
}

exit 0;
