#!/usr/bin/perl -w
use strict;

use lib qw(/opt/mam/lib /opt/mam/lib/perl5);
use Config::Tiny;
use DBI;
use MAM;
use MAM::Global;

my ($cmd, $output, $rc, $sql, $count);
my ($successes, $failures, $ignores) = (0, 0, 0);
my @messages   = ();
$ENV{PATH} = "/opt/mam/bin:/opt/mam/sbin:$ENV{PATH}";

# Read the Config File
my $configFile = "/opt/mam/etc/mam-server.conf";
my $configTiny = Config::Tiny->new;
$configTiny = Config::Tiny->read($configFile)
  or die(
    "Unable to open config file ($configFile) for reading: $Config::Tiny::errstr.\nYou may need to set the \$configFile variable in this script.\n"
  );
$serverConfig = $configTiny->{_};
my $datasource = $serverConfig->{'database.datasource'};
my $dbuser     = $serverConfig->{'database.user'} || (getpwuid($<))[0];
my $dbpasswd   = $serverConfig->{'database.password'};

################################################################################
# Display warning and prompt to proceed
################################################################################

my $warning = <<"ENDOFWARNING";
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! IMPORTANT !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

This script will attempt to convert an existing Moab Accounting Manager 7.2 database (using the datasource ($datasource) referenced in /opt/mam/etc/mam-server.conf) into a Moab Accounting Manager 7.3 database. It will make additions, modifications and deletions to your Moab Accounting Manager 7.2 schema and data to match the new base Moab Accounting Manager 7.3 schema and data values.

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! BACKUP YOUR DATA !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

You should MAKE A BACKUP of your Moab Accounting Manager database before running this script. If you proceed without a backup, a failure in this script could irreversibly corrupt the integrity and data of your Moab Accounting Manager database.

ENDOFWARNING

print $warning;
print "Do you want to proceed? [y/N] ";
my $answer = <STDIN>;
exit 0 unless ($answer =~ /^y(es)?$/i);

# ----------------------------------------------------------------------------
# update($description, $test, $comparison, $action);
# ----------------------------------------------------------------------------

# Update the object metadata via mam-shell
sub update
{
    my ($description, $test, $comparison, $action) = @_;
    print "\n***$description***\n";
    if (chomp($_ = qx{mam-shell -q $test}) && eval "\$_ $comparison 0")
    {
        $cmd    = "mam-shell -v $action";
        $output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
        $rc     = $? >> 8;
        print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
        if ($rc)
        {
            $failures++;
            my $message = "Subcommand ($cmd) failed with rc=$rc";
            $message .= ": $output" if $output;
            warn($message);
            push @messages, $message;
        }
        else
        {
            $successes++;
        }
    }
    else
    {
        print "Already applied\n";
        $ignores++;
    }
}

################################################################################
# Create and Update Accounting Objects and Attributes
################################################################################

print "\n***Creating new Accounting Objects and Attributes***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==ChargeRate Name==Amount Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ == 0
    )
    || (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==QuoteChargeRate Name==Amount Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ == 0
    )
  )
{
    update(
        q{Creating new ChargeRate Amount Attribute},
        q{Attribute Query Object==ChargeRate Name==Amount Show:="\"Count(Name)\""},
        q{==},
        q{Attribute Create Object=ChargeRate Name=Amount Required=True DataType=String Sequence=30 Description="\"Charge Rate Amount\"" NoRefresh:=True}
    );

    update(
        q{Creating new QuoteChargeRate Amount Attribute},
        q{Attribute Query Object==QuoteChargeRate Name==Amount Show:="\"Count(Name)\""},
        q{==},
        q{Attribute Create Object=QuoteChargeRate Name=Amount Required=True DataType=String Sequence=40 Description="\"Charge Rate Amount\"" NoRefresh:=True}
    );

    print "\n***Refreshing Cache***\n";
    $cmd    = "mam-shell -v System Refresh";
    $output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
    $rc     = $? >> 8;
    print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
    if ($rc)
    {
        $failures++;
        my $message = "Subcommand ($cmd) failed with rc=$rc";
        $message .= ": $output" if $output;
        warn($message);
        push @messages, $message;
    }
    else
    {
        $successes++;
    }
}

update(
    q{Modifying UsageRecord Stage Attribute Values to include ChargeFail},
    q{Attribute Query Object==UsageRecord Name==Stage Values~"*ChargeFail*" Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Modify Object==UsageRecord Name==Stage Values=\"\(Charge,ChargeFail,Create,Quote,Reserve\)\" NoRefresh:=True}
);

################################################################################
# Transform Accounting Tables
################################################################################

print
  "\n***Modifying the ChargeRate Amount and QuoteChargeRate Amount to be consistent with the former Type and Rate***\n";
if (
    (
        (
            chomp(
                $_ =
                  qx{mam-shell -q Attribute Query Object==ChargeRate Name==Amount Show:=\"\\\"Count(Name)\\\"\"}
            )
            && $_ > 0
        )
        && (
            chomp(
                $_ =
                  qx{mam-shell -q Attribute Query Object==ChargeRate Name==Type Show:=\"\\\"Count(Name)\\\"\"}
            )
            && $_ > 0
        )
        && (
            chomp(
                $_ =
                  qx{mam-shell -q Attribute Query Object==ChargeRate Name==Rate Show:=\"\\\"Count(Name)\\\"\"}
            )
            && $_ > 0
        )
    )
    || (
        (
            chomp(
                $_ =
                  qx{mam-shell -q Attribute Query Object==ChargeRate Name==Amount Show:=\"\\\"Count(Name)\\\"\"}
            )
            && $_ > 0
        )
        && (
            chomp(
                $_ =
                  qx{mam-shell -q Attribute Query Object==ChargeRate Name==Type Show:=\"\\\"Count(Name)\\\"\"}
            )
            && $_ > 0
        )
        && (
            chomp(
                $_ =
                  qx{mam-shell -q Attribute Query Object==ChargeRate Name==Rate Show:=\"\\\"Count(Name)\\\"\"}
            )
            && $_ > 0
        )
    )
  )
{
    my $queryFailures = 0;
    my ($request, $response, $code, $message, $doc, $data);
    print "\n***Transforming ChargeRates***\n";
    $request = new MAM::Request(
        object => "ChargeRate",
        action => "Query",
    );
    $request->addSelection("Type");
    $request->addSelection("Name");
    $request->addSelection("Value");
    $request->addSelection("Rate");
    $response = $request->getResponse();
    &MAM::Client::displayResponse($response);

    if ($response->getStatus() eq "Failure")
    {
        $code    = $response->getCode();
        $message = "ChargeRate Query failed with return code ($code):\n"
          . $response->getMessage() . "\n";
        $queryFailures++;
        warn($message);
        push @messages, $message;
    }
    else
    {
        $doc  = XML::LibXML::Document->new();
        $data = $response->getDataElement();
        $doc->setDocumentElement($data);
        foreach my $row ($data->childNodes())
        {
            my $type  = ($row->getChildrenByTagName("Type"))[0]->textContent();
            my $name  = ($row->getChildrenByTagName("Name"))[0]->textContent();
            my $value = ($row->getChildrenByTagName("Value"))[0]->textContent();
            my $rate  = ($row->getChildrenByTagName("Rate"))[0]->textContent();

            $request = new MAM::Request(
                object => "ChargeRate",
                action => "Modify",
            );
            $request->addCondition("Name",  $name);
            $request->addCondition("Value", $value);
            if ($type =~ /^[NV]BR$/)
            {
                $request->addAssignment("Amount", $rate . '/s');
            }
            elsif ($type =~ /^[NV]BM$/)
            {
                $request->addAssignment("Amount", '*' . $rate);
            }
            elsif ($type =~ /^[NV]BF$/)
            {
                $request->addAssignment("Amount", $rate . '+');
            }
            else
            {
                $request->addAssignment("Amount", $rate);
            }
            $response = $request->getResponse();
            &MAM::Client::displayResponse($response);

            if ($response->getStatus() eq "Failure")
            {
                $code = $response->getCode();
                $message =
                  "ChargeRate Modify failed with return code ($code):\n"
                  . $response->getMessage() . "\n";
                $queryFailures++;
                warn($message);
                push @messages, $message;
            }
        }
    }

    if ($queryFailures)
    {
        $failures++;
    }
    else
    {
        $successes++;
    }

    $queryFailures = 0;
    print "\n***Transforming QuoteChargeRates***\n";
    $request = new MAM::Request(
        object => "QuoteChargeRate",
        action => "Query",
    );
    $request->addSelection("Quote");
    $request->addSelection("Type");
    $request->addSelection("Name");
    $request->addSelection("Value");
    $request->addSelection("Rate");
    $response = $request->getResponse();
    &MAM::Client::displayResponse($response);

    if ($response->getStatus() eq "Failure")
    {
        $code    = $response->getCode();
        $message = "QuoteChargeRate Query failed with return code ($code):\n"
          . $response->getMessage() . "\n";
        $queryFailures++;
        warn($message);
        push @messages, $message;
    }
    else
    {
        $doc  = XML::LibXML::Document->new();
        $data = $response->getDataElement();
        $doc->setDocumentElement($data);
        foreach my $row ($data->childNodes())
        {
            my $quote = ($row->getChildrenByTagName("Quote"))[0]->textContent();
            my $type  = ($row->getChildrenByTagName("Type"))[0]->textContent();
            my $name  = ($row->getChildrenByTagName("Name"))[0]->textContent();
            my $value = ($row->getChildrenByTagName("Value"))[0]->textContent();
            my $rate  = ($row->getChildrenByTagName("Rate"))[0]->textContent();

            $request = new MAM::Request(
                object => "QuoteChargeRate",
                action => "Modify",
            );
            $request->addCondition("Quote", $quote);
            $request->addCondition("Name",  $name);
            $request->addCondition("Value", $value);
            if ($type =~ /^[NV]BR$/)
            {
                $request->addAssignment("Amount", $rate . '/s');
            }
            elsif ($type =~ /^[NV]BM$/)
            {
                $request->addAssignment("Amount", '*' . $rate);
            }
            elsif ($type =~ /^[NV]BF$/)
            {
                $request->addAssignment("Amount", $rate . '+');
            }
            else
            {
                next;
            }
            $response = $request->getResponse();
            &MAM::Client::displayResponse($response);

            if ($response->getStatus() eq "Failure")
            {
                $code = $response->getCode();
                $message =
                  "QuoteChargeRate Modify failed with return code ($code):\n"
                  . $response->getMessage() . "\n";
                $queryFailures++;
                warn($message);
                push @messages, $message;
            }
        }
    }

    if ($queryFailures)
    {
        $failures++;
    }
    else
    {
        $successes++;
    }
}
else
{
    print "Already applied\n";
    $ignores++;
}

################################################################################
# Change Attribute Datatypes
################################################################################

print "\n***Changing Charge Rate DataType from Float to String***\n";
if (
    chomp(
        $_ =
          qx{mam-shell -q Attribute Query Object==Charge Name==Rate DataType==Float Show:=\"\\\"Count(Name)\\\"\"}
    )
    && $_ > 0
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 0});

    # Create placeholder column
    $sql = "ALTER TABLE g_charge ADD COLUMN g_rate_tmp character varying(1024)";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql =
      "ALTER TABLE g_charge_log ADD COLUMN g_rate_tmp character varying(1024)";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Copy old values into placeholder
    $sql = "UPDATE g_charge set g_rate_tmp=g_rate";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql = "UPDATE g_charge_log set g_rate_tmp=g_rate";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    # Delete old column
    $sql = "ALTER TABLE g_charge DROP COLUMN g_rate";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_charge_log DROP COLUMN g_rate";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Create new column
    $sql = "ALTER TABLE g_charge ADD COLUMN g_rate character varying(1024)";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_charge_log ADD COLUMN g_rate character varying(1024)";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Copy placeholder values into new
    $sql = "UPDATE g_charge set g_rate=g_rate_tmp";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql = "UPDATE g_charge_log set g_rate=g_rate_tmp";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    # Delete placeholder column
    $sql = "ALTER TABLE g_charge DROP COLUMN g_rate_tmp";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_charge_log DROP COLUMN g_rate_tmp";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Change DataType in Attribute
    $sql =
      "UPDATE g_attribute set g_data_type='String' WHERE g_object='Charge' AND g_name='Rate'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $successes += $count;
    $sql =
      "UPDATE g_attribute_log set g_data_type='String' WHERE g_object='Charge' AND g_name='Rate'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Changing UsageRecord Memory DataType from String to Float***\n";
if (
    chomp(
        $_ =
          qx{mam-shell -q Attribute Query Object==UsageRecord Name==Memory DataType==String Show:=\"\\\"Count(Name)\\\"\"}
    )
    && $_ > 0
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 0});

    # Create placeholder column
    $sql =
      "ALTER TABLE g_usage_record ADD COLUMN g_memory_tmp double precision";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql =
      "ALTER TABLE g_usage_record_log ADD COLUMN g_memory_tmp double precision";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Copy old values into placeholder
    $sql =
      "UPDATE g_usage_record set g_memory_tmp=cast(g_memory AS double precision)";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql =
      "UPDATE g_usage_record_log set g_memory_tmp=cast(g_memory AS double precision)";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    # Delete old column
    $sql = "ALTER TABLE g_usage_record DROP COLUMN g_memory";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_usage_record_log DROP COLUMN g_memory";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Create new column
    $sql = "ALTER TABLE g_usage_record ADD COLUMN g_memory double precision";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql =
      "ALTER TABLE g_usage_record_log ADD COLUMN g_memory double precision";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Copy placeholder values into new
    $sql = "UPDATE g_usage_record set g_memory=g_memory_tmp";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql = "UPDATE g_usage_record_log set g_memory=g_memory_tmp";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    # Delete placeholder column
    $sql = "ALTER TABLE g_usage_record DROP COLUMN g_memory_tmp";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_usage_record_log DROP COLUMN g_memory_tmp";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Change DataType in Attribute
    $sql =
      "UPDATE g_attribute set g_data_type='Float' WHERE g_object='UsageRecord' AND g_name='Memory'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $successes += $count;
    $sql =
      "UPDATE g_attribute_log set g_data_type='Float' WHERE g_object='UsageRecord' AND g_name='Memory'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
}
else
{
    print "Already applied\n";
    $ignores++;
}

print "\n***Changing UsageRecord Disk DataType from String to Float***\n";
if (
    chomp(
        $_ =
          qx{mam-shell -q Attribute Query Object==UsageRecord Name==Disk DataType==String Show:=\"\\\"Count(Name)\\\"\"}
    )
    && $_ > 0
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 0});

    # Create placeholder column
    $sql =
      "ALTER TABLE g_usage_record ADD COLUMN g_disk_tmp double precision";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql =
      "ALTER TABLE g_usage_record_log ADD COLUMN g_disk_tmp double precision";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Copy old values into placeholder
    $sql =
      "UPDATE g_usage_record set g_disk_tmp=cast(g_disk AS double precision)";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql =
      "UPDATE g_usage_record_log set g_disk_tmp=cast(g_disk AS double precision)";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    # Delete old column
    $sql = "ALTER TABLE g_usage_record DROP COLUMN g_disk";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_usage_record_log DROP COLUMN g_disk";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Create new column
    $sql = "ALTER TABLE g_usage_record ADD COLUMN g_disk double precision";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql =
      "ALTER TABLE g_usage_record_log ADD COLUMN g_disk double precision";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Copy placeholder values into new
    $sql = "UPDATE g_usage_record set g_disk=g_disk_tmp";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql = "UPDATE g_usage_record_log set g_disk=g_disk_tmp";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    # Delete placeholder column
    $sql = "ALTER TABLE g_usage_record DROP COLUMN g_disk_tmp";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_usage_record_log DROP COLUMN g_disk_tmp";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # Change DataType in Attribute
    $sql =
      "UPDATE g_attribute set g_data_type='Float' WHERE g_object='UsageRecord' AND g_name='Disk'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $successes += $count;
    $sql =
      "UPDATE g_attribute_log set g_data_type='Float' WHERE g_object='UsageRecord' AND g_name='Disk'";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
}
else
{
    print "Already applied\n";
    $ignores++;
}

################################################################################
# Delete old Accounting Objects and Attributes
################################################################################

if (
    (
        (
            chomp(
                $_ =
                  qx{mam-shell -q Attribute Query Object==ChargeRate Name==Type Show:=\"\\\"Count(Name)\\\"\"}
            )
            && $_ > 0
        )
        || (
            chomp(
                $_ =
                  qx{mam-shell -q Attribute Query Object==QuoteChargeRate Name==Type Show:=\"\\\"Count(Name)\\\"\"}
            )
            && $_ > 0
        )
        || (
            chomp(
                $_ =
                  qx{mam-shell -q Attribute Query Object==ChargeRate Name==Rate Show:=\"\\\"Count(Name)\\\"\"}
            )
            && $_ > 0
        )
        || (
            chomp(
                $_ =
                  qx{mam-shell -q Attribute Query Object==QuoteChargeRate Name==Rate Show:=\"\\\"Count(Name)\\\"\"}
            )
            && $_ > 0
        )
    )
    && ! $failures
  )
{
    update(
        q{Deleting old ChargeRate Type Attribute},
        q{Attribute Query Object==ChargeRate Name==Type Show:="\"Count(Name)\""},
        q{>},
        q{Attribute Delete Object==ChargeRate Name==Type NoRefresh:=True}
    );

    update(
        q{Deleting old ChargeRate Rate Attribute},
        q{Attribute Query Object==ChargeRate Name==Rate Show:="\"Count(Name)\""},
        q{>},
        q{Attribute Delete Object==ChargeRate Name==Rate NoRefresh:=True}
    );

    update(
        q{Deleting old QuoteChargeRate Type Attribute},
        q{Attribute Query Object==QuoteChargeRate Name==Type Show:="\"Count(Name)\""},
        q{>},
        q{Attribute Delete Object==QuoteChargeRate Name==Type NoRefresh:=True}
    );

    update(
        q{Deleting old QuoteChargeRate Rate Attribute},
        q{Attribute Query Object==QuoteChargeRate Name==Rate Show:="\"Count(Name)\""},
        q{>},
        q{Attribute Delete Object==QuoteChargeRate Name==Rate NoRefresh:=True}
    );

    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 0});

    # ChargeRate
    $sql =
      "DELETE FROM g_attribute WHERE g_object='ChargeRate' AND g_name='Type'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='ChargeRate' AND g_name='Type'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute WHERE g_object='ChargeRate' AND g_name='Rate'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='ChargeRate' AND g_name='Rate'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_charge_rate DROP COLUMN g_type";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_charge_rate_log DROP COLUMN g_type";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_charge_rate DROP COLUMN g_rate";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_charge_rate_log DROP COLUMN g_rate";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    # QuoteChargeRate
    $sql =
      "DELETE FROM g_attribute WHERE g_object='QuoteChargeRate' AND g_name='Type'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='QuoteChargeRate' AND g_name='Type'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute WHERE g_object='QuoteChargeRate' AND g_name='Rate'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='QuoteChargeRate' AND g_name='Rate'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_quote_charge_rate DROP COLUMN g_type";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_quote_charge_rate_log DROP COLUMN g_type";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_quote_charge_rate DROP COLUMN g_rate";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_quote_charge_rate_log DROP COLUMN g_rate";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    $successes++;
}
else
{
    print "Already applied or dependencies not met\n";
    $ignores++;
}

################################################################################
# Refresh Cache and Update System Information
################################################################################

print "\n***Refreshing Cache***\n";
$cmd    = "mam-shell -v System Refresh";
$output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
$rc     = $? >> 8;
print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
if ($rc)
{
    $failures++;
    my $message = "Subcommand ($cmd) failed with rc=$rc";
    $message .= ": $output" if $output;
    warn($message);
    push @messages, $message;
}
else
{
    $successes++;
}

unless ($failures)
{
    update(
        q{Updating System Information},
        q{System Query Version==\"7.3\" Show:="\"Count(Name)\""},
        q{==}, q{System Modify Version=\"7.3\"}
    );
}

################################################################################
# Display Results
################################################################################

print "\n";
print "$successes updates were successfully applied.\n";
print "$ignores updates were skipped having been already applied.\n";
print "$failures updates failed.\n";
if (@messages)
{
    print "The following is a list of failures:\n";
    print join "\n", @messages;
}

exit 0;
