#!/usr/bin/perl -w
use strict;

use lib qw(/opt/mam/lib /opt/mam/lib/perl5);
use Config::Tiny;
use DBI;
use MAM;
use MAM::Global;

my ($cmd, $output, $rc, $sql, $count);
my ($successes, $failures, $ignores) = (0, 0, 0);
my @messages   = ();
$ENV{PATH} = "/opt/mam/bin:/opt/mam/sbin:$ENV{PATH}";

# Read the Config File
my $configFile = "/opt/mam/etc/mam-server.conf";
my $configTiny = Config::Tiny->new;
$configTiny = Config::Tiny->read($configFile)
  or die(
    "Unable to open config file ($configFile) for reading: $Config::Tiny::errstr.\nYou may need to set the \$configFile variable in this script.\n"
  );
$serverConfig = $configTiny->{_};
my $datasource = $serverConfig->{'database.datasource'};
my $dbuser     = $serverConfig->{'database.user'} || (getpwuid($<))[0];
my $dbpasswd   = $serverConfig->{'database.password'};

################################################################################
# Display warning and prompt to proceed
################################################################################

my $warning = <<"ENDOFWARNING";
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! IMPORTANT !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

This script will attempt to convert an existing Moab Accounting Manager 7.3 database (using the datasource ($datasource) referenced in /opt/mam/etc/mam-server.conf) into a Moab Accounting Manager 7.5 database. It will make additions, modifications and deletions to your Moab Accounting Manager 7.3 schema and data to match the new base Moab Accounting Manager 7.5 schema and data values.

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! BACKUP YOUR DATA !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

You should MAKE A BACKUP of your Moab Accounting Manager database before running this script. If you proceed without a backup, a failure in this script could irreversibly corrupt the integrity and data of your Moab Accounting Manager database.

ENDOFWARNING

print $warning;
print "Do you want to proceed? [y/N] ";
my $answer = <STDIN>;
exit 0 unless ($answer =~ /^y(es)?$/i);

# ----------------------------------------------------------------------------
# update($description, $test, $comparison, $action);
# ----------------------------------------------------------------------------

# Update the object metadata via mam-shell
sub update
{
    my ($description, $test, $comparison, $action) = @_;
    print "\n***$description***\n";
    if (chomp($_ = qx{mam-shell -q $test}) && eval "\$_ $comparison 0")
    {
        $cmd    = "mam-shell -v $action";
        $output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
        $rc     = $? >> 8;
        print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
        if ($rc)
        {
            $failures++;
            my $message = "Subcommand ($cmd) failed with rc=$rc";
            $message .= ": $output" if $output;
            warn($message);
            push @messages, $message;
        }
        else
        {
            $successes++;
        }
    }
    else
    {
        print "Already applied\n";
        $ignores++;
    }
}

################################################################################
# Create and Update Accounting Objects and Attributes
################################################################################

update(
    q{Creating new Allocation InitialDeposit Attribute},
    q{Attribute Query Object==Allocation Name==InitialDeposit Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Allocation Name=InitialDeposit DataType=Currency DefaultValue=0 Sequence=65 Description="\"Initial Deposit\"" NoRefresh:=True}
);

update(
    q{Creating new Allocation Allocated Attribute},
    q{Attribute Query Object==Allocation Name==Allocated Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Allocation Name=Allocated DataType=Currency DefaultValue=0 Sequence=70 Description="\"Adjusted Allocation\"" NoRefresh:=True}
);

update(
    q{Creating new Fund DefaultDeposit Attribute},
    q{Attribute Query Object==Fund Name==DefaultDeposit Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Fund Name=DefaultDeposit DataType=Currency Sequence=35 Description="\"Default Deposit Amount\"" NoRefresh:=True}
);

update(
    q{Creating new Payment ReservePattern Attribute},
    q{Attribute Query Object==Payment Name==ReservePattern Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Payment Name=ReservePattern DataType=String Values=\"\(Combined,None,Periodic\)\" DefaultValue=Periodic Sequence=110 Description="\"Payment Lien Pattern\"" NoRefresh:=True}
);

update(
    q{Creating new Payment ChargePattern Attribute},
    q{Attribute Query Object==Payment Name==ChargePattern Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=Payment Name=ChargePattern DataType=String Values=\"\(Combined,None,Periodic\)\" DefaultValue=Periodic Sequence=120 Description="\"Payment Charge Pattern\"" NoRefresh:=True}
);

update(
    q{Creating new UsageRecord CPUTime Attribute},
    q{Attribute Query Object==UsageRecord Name==CPUTime Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=CPUTime DataType=Float Sequence=165 Description="\"CPU Time Used\"" NoRefresh:=True}
);

update(
    q{Creating new Fund Stop Action},
    q{Action Query Object==Fund Name==Stop Show:="\"Count(Name)\""},
    q{==},
    q{Action Create Object=Fund Name=Stop Description="\"End current Allocation\"" NoRefresh:=True}
);

update(
    q{Modifying Fund Reset Action Description},
    q{Action Query Object==Fund Name==Reset Description=="\"Create new default Allocation\"" Show:="\"Count(Name)\""},
    q{==},
    q{Action Modify Object==Fund Name==Reset Description="\"Create new default Allocation\"" NoRefresh:=True}
);


################################################################################
# Transform Accounting Tables
################################################################################

print "\n***Updating Allocation table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Allocation Name==Deposited Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Allocation Name==Allocated Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});

    $sql = "UPDATE g_allocation SET g_allocated=g_deposited";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql = "UPDATE g_allocation_log SET g_allocated=g_deposited";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

################################################################################
# Delete old Accounting Objects and Attributes
################################################################################

# Delete Allocation Deposited Attribute
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==Allocation Name==Deposited Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && ! $failures
  )
{
    update(
        q{Deleting old Allocation Deposited Attribute},
        q{Attribute Query Object==Allocation Name==Deposited Show:="\"Count(Name)\""},
        q{>},
        q{Attribute Delete Object==Allocation Name==Deposited NoRefresh:=True}
    );

    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 0});

    # Allocation
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Allocation' AND g_name='Deposited'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Allocation' AND g_name='Deposited'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_allocation DROP COLUMN g_deposited";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_allocation_log DROP COLUMN g_deposited";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    $successes++;
}
else
{
    print "Already applied or dependencies not met\n";
    $ignores++;
}

# Delete Payment {Charge,Reserve}Mode Attributes
if (
    (
        (
            chomp(
                $_ =
                  qx{mam-shell -q Attribute Query Object==Payment Name==ReserveMode Show:=\"\\\"Count(Name)\\\"\"}
            )
            && $_ > 0
        )
        || (
            chomp(
                $_ =
                  qx{mam-shell -q Attribute Query Object==Payment Name==ChargeMode Show:=\"\\\"Count(Name)\\\"\"}
            )
            && $_ > 0
        )
    )
    && ! $failures
  )
{
    update(
        q{Deleting old Payment ReserveMode Attribute},
        q{Attribute Query Object==Payment Name==ReserveMode Show:="\"Count(Name)\""},
        q{>},
        q{Attribute Delete Object==Payment Name==ReserveMode NoRefresh:=True}
    );

    update(
        q{Deleting old Payment ChargeMode Attribute},
        q{Attribute Query Object==Payment Name==ChargeMode Show:="\"Count(Name)\""},
        q{>},
        q{Attribute Delete Object==Payment Name==ChargeMode NoRefresh:=True}
    );

    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 0});

    # Payment
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Payment' AND g_name='ReserveMode'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute WHERE g_object='Payment' AND g_name='ChargeMode'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Payment' AND g_name='ReserveMode'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='Payment' AND g_name='ChargeMode'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_payment DROP COLUMN g_reserve_mode";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_payment DROP COLUMN g_charge_mode";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_payment_log DROP COLUMN g_reserve_mode";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_payment_log DROP COLUMN g_charge_mode";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    $successes++;
}
else
{
    print "Already applied or dependencies not met\n";
    $ignores++;
}

# Delete UsageRecord Network Attribute
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==UsageRecord Name==Network Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && ! $failures
  )
{
    update(
        q{Deleting old UsageRecord Network Attribute},
        q{Attribute Query Object==UsageRecord Name==Network Show:="\"Count(Name)\""},
        q{>},
        q{Attribute Delete Object==UsageRecord Name==Network NoRefresh:=True}
    );

    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 0});

    # UsageRecord
    $sql =
      "DELETE FROM g_attribute WHERE g_object='UsageRecord' AND g_name='Network'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='UsageRecord' AND g_name='Network'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_usage_record DROP COLUMN g_network";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_usage_record_log DROP COLUMN g_network";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    $successes++;
}
else
{
    print "Already applied or dependencies not met\n";
    $ignores++;
}

# Delete Allocation Reset Action
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Action Query Object==Allocation Name==Reset Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && ! $failures
  )
{
    update(
        q{Deleting old Allocation Reset Action},
        q{Action Query Object==Allocation Name==Reset Show:="\"Count(Name)\""},
        q{>},
        q{Action Delete Object==Allocation Name==Reset NoRefresh:=True}
    );

    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 0});

    # Allocation
    $sql =
      "DELETE FROM g_action WHERE g_object='Allocation' AND g_name='Reset'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_action_log WHERE g_object='Allocation' AND g_name='Reset'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
}
else
{
    print "Already applied or dependencies not met\n";
    $ignores++;
}

################################################################################
# Refresh Cache and Update System Information
################################################################################

print "\n***Refreshing Cache***\n";
$cmd    = "mam-shell -v System Refresh";
$output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
$rc     = $? >> 8;
print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
if ($rc)
{
    $failures++;
    my $message = "Subcommand ($cmd) failed with rc=$rc";
    $message .= ": $output" if $output;
    warn($message);
    push @messages, $message;
}
else
{
    $successes++;
}

unless ($failures)
{
    update(
        q{Updating System Information},
        q{System Query Version==\"7.5\" Show:="\"Count(Name)\""},
        q{==}, q{System Modify Version=\"7.5\"}
    );
}

################################################################################
# Display Results
################################################################################

print "\n";
print "$successes updates were successfully applied.\n";
print "$ignores updates were skipped having been already applied.\n";
print "$failures updates failed.\n";
if (@messages)
{
    print "The following is a list of failures:\n";
    print join "\n", @messages;
}

exit 0;
