#!/usr/bin/perl -w
use strict;

use lib qw(/opt/mam/lib /opt/mam/lib/perl5);
use Config::Tiny;
use DBI;
use MAM;
use MAM::Global;

my ($cmd, $output, $rc, $sql, $count);
my ($successes, $failures, $ignores) = (0, 0, 0);
my @messages   = ();
$ENV{PATH} = "/opt/mam/bin:/opt/mam/sbin:$ENV{PATH}";

# Read the Config File
my $configFile = "/opt/mam/etc/mam-server.conf";
my $configTiny = Config::Tiny->new;
$configTiny = Config::Tiny->read($configFile)
  or die(
    "Unable to open config file ($configFile) for reading: $Config::Tiny::errstr.\nYou may need to set the \$configFile variable in this script.\n"
  );
$serverConfig = $configTiny->{_};
my $datasource = $serverConfig->{'database.datasource'};
my $dbuser     = $serverConfig->{'database.user'} || (getpwuid($<))[0];
my $dbpasswd   = $serverConfig->{'database.password'};

################################################################################
# Display warning and prompt to proceed
################################################################################

my $warning = <<"ENDOFWARNING";
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! IMPORTANT !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

This script will attempt to convert an existing Moab Accounting Manager 7.5 database (using the datasource ($datasource) referenced in /opt/mam/etc/mam-server.conf) into a Moab Accounting Manager 8.0 database. It will make additions, modifications and deletions to your Moab Accounting Manager 7.5 schema and data to match the new base Moab Accounting Manager 8.0 schema and data values.

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! BACKUP YOUR DATA !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

You should MAKE A BACKUP of your Moab Accounting Manager database before running this script. If you proceed without a backup, a failure in this script could irreversibly corrupt the integrity and data of your Moab Accounting Manager database.

ENDOFWARNING

print $warning;
print "Do you want to proceed? [y/N] ";
my $answer = <STDIN>;
exit 0 unless ($answer =~ /^y(es)?$/i);

# ----------------------------------------------------------------------------
# update($description, $test, $comparison, $action);
# ----------------------------------------------------------------------------

# Update the object metadata via mam-shell
sub update
{
    my ($description, $test, $comparison, $action) = @_;
    print "\n***$description***\n";
    if (chomp($_ = qx{mam-shell -q $test}) && eval "\$_ $comparison 0")
    {
        $cmd    = "mam-shell -v $action";
        $output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
        $rc     = $? >> 8;
        print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
        if ($rc)
        {
            $failures++;
            my $message = "Subcommand ($cmd) failed with rc=$rc";
            $message .= ": $output" if $output;
            warn($message);
            push @messages, $message;
        }
        else
        {
            $successes++;
        }
    }
    else
    {
        print "Already applied\n";
        $ignores++;
    }
}

################################################################################
# Delete old Accounting Objects and Attributes
################################################################################

update(
    q{Deleting old OVERRIDE Fund Balance RoleAction},
    q{RoleAction Query Role==OVERRIDE Object==Fund Name==Balance Instance==ANY Show:="\"Count(Name)\""},
    q{>},
    q{RoleAction Delete Role==OVERRIDE Object==Fund Name==Balance Instance==ANY NoRefresh:=True}
);

# Delete Fund Balance Action
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Action Query Object==Fund Name==Balance Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    update(
        q{Deleting old Fund Balance Action},
        q{Action Query Object==Fund Name==Balance Show:="\"Count(Name)\""},
        q{>},
        q{Action Delete Object==Fund Name==Balance NoRefresh:=True}
    );

    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 0});

    # Allocation
    $sql =
      "DELETE FROM g_action WHERE g_object='Fund' AND g_name='Balance'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_action_log WHERE g_object='Fund' AND g_name='Balance'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
}
else
{
    print "Already applied or dependencies not met\n";
    $ignores++;
}

################################################################################
# Refresh Cache and Update System Information
################################################################################

print "\n***Refreshing Cache***\n";
$cmd    = "mam-shell -v System Refresh";
$output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
$rc     = $? >> 8;
print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
if ($rc)
{
    $failures++;
    my $message = "Subcommand ($cmd) failed with rc=$rc";
    $message .= ": $output" if $output;
    warn($message);
    push @messages, $message;
}
else
{
    $successes++;
}

unless ($failures)
{
    update(
        q{Updating System Information},
        q{System Query Version==\"8.0\" Show:="\"Count(Name)\""},
        q{==}, q{System Modify Version=\"8.0\"}
    );
}

################################################################################
# Display Results
################################################################################

print "\n";
print "$successes updates were successfully applied.\n";
print "$ignores updates were skipped having been already applied.\n";
print "$failures updates failed.\n";
if (@messages)
{
    print "The following is a list of failures:\n";
    print join "\n", @messages;
}

exit 0;
