#!/usr/bin/perl -w
use strict;

use lib qw(/opt/mam/lib /opt/mam/lib/perl5);
use Config::Tiny;
use DBI;
use MAM;
use MAM::Global;

my ($cmd, $output, $rc, $sql, $count);
my ($successes, $failures, $ignores) = (0, 0, 0);
my @messages   = ();
$ENV{PATH} = "/opt/mam/bin:/opt/mam/sbin:$ENV{PATH}";

# Read the Config File
my $configFile = "/opt/mam/etc/mam-server.conf";
my $configTiny = Config::Tiny->new;
$configTiny = Config::Tiny->read($configFile)
  or die(
    "Unable to open config file ($configFile) for reading: $Config::Tiny::errstr.\nYou may need to set the \$configFile variable in this script.\n"
  );
$serverConfig = $configTiny->{_};
my $datasource = $serverConfig->{'database.datasource'};
my $dbuser     = $serverConfig->{'database.user'} || (getpwuid($<))[0];
my $dbpasswd   = $serverConfig->{'database.password'};

################################################################################
# Display warning and prompt to proceed
################################################################################

my $warning = <<"ENDOFWARNING";
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! IMPORTANT !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

This script will attempt to convert an existing Moab Accounting Manager 9.0 database (using the datasource ($datasource) referenced in /opt/mam/etc/mam-server.conf) into a Moab Accounting Manager 9.1 database. It will make additions, modifications and deletions to your Moab Accounting Manager 9.0 schema and data to match the new base Moab Accounting Manager 9.1 schema and data values.

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! BACKUP YOUR DATA !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

You should MAKE A BACKUP of your Moab Accounting Manager database before running this script. If you proceed without a backup, a failure in this script could irreversibly corrupt the integrity and data of your Moab Accounting Manager database.

ENDOFWARNING

print $warning;
print "Do you want to proceed? [y/N] ";
my $answer = <STDIN>;
exit 0 unless ($answer =~ /^y(es)?$/i);

# ----------------------------------------------------------------------------
# update($description, $test, $comparison, $action);
# ----------------------------------------------------------------------------

# Update the object metadata via mam-shell
sub update
{
    my ($description, $test, $comparison, $action) = @_;
    print "\n***$description***\n";
    if (chomp($_ = qx{mam-shell -q $test}) && eval "\$_ $comparison 0")
    {
        $cmd    = "mam-shell -v $action";
        $output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
        $rc     = $? >> 8;
        print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
        if ($rc)
        {
            $failures++;
            my $message = "Subcommand ($cmd) failed with rc=$rc";
            $message .= ": $output" if $output;
            warn($message);
            push @messages, $message;
        }
        else
        {
            $successes++;
        }
    }
    else
    {
        print "Already applied\n";
        $ignores++;
    }
}

################################################################################
# Create and Update Accounting Objects and Attributes
################################################################################

update(
    q{Modifying Attribute DataType Values to include JSON},
    q{Attribute Query Object==Attribute Name==DataType Values~"*JSON?:?w+*" Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Modify Object==Attribute Name==DataType Values=\"\(AutoGen,Boolean,Currency,Float,Integer,JSON\(:\\\w+\)?,String,TimeStamp\)\"}
);

update(
    q{Creating new UsageRecord Resources Attribute},
    q{Attribute Query Object==UsageRecord Name==Resources Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Resources DataType=JSON:Integer Description="\"Generic Resources\"" Sequence=180 NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Licenses Attribute},
    q{Attribute Query Object==UsageRecord Name==Licenses Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Licenses DataType=JSON:Integer Description="\"Licenses Used\"" Sequence=190 NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Metrics Attribute},
    q{Attribute Query Object==UsageRecord Name==Metrics Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Metrics DataType=JSON:Float Description="\"Generic Metrics\"" Sequence=200 NoRefresh:=True}
);

update(
    q{Creating new UsageRecord Variables Attribute},
    q{Attribute Query Object==UsageRecord Name==Variables Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=Variables DataType=JSON:String Description="\"Job Variables\"" Sequence=210 NoRefresh:=True}
);

update(
    q{Creating new UsageRecord RequestedDuration Attribute},
    q{Attribute Query Object==UsageRecord Name==RequestedDuration Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=RequestedDuration DataType=Integer Description="\"Requested Wallclock Time in seconds\"" Sequence=230 NoRefresh:=True}
);

update(
    q{Creating new UsageRecord ExitCode Attribute},
    q{Attribute Query Object==UsageRecord Name==ExitCode Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Create Object=UsageRecord Name=ExitCode DataType=Integer Description="\"Exit Code\"" Sequence=270 NoRefresh:=True}
);

update(
    q{Modifying Event Notify Attribute to default to Store:},
    q{Attribute Query Object==Event Name==Notify DefaultValue==Store: Show:="\"Count(Name)\""},
    q{==},
    q{Attribute Modify Object==Event Name==Notify DefaultValue=Store: NoRefresh:=True}
);

update(
    q{Creating new Notification Refresh Event},
    q{Event Query FireCommand=="\"Notification Refresh\"" Show:="\"Count(Id)\""},
    q{==},
    q{Event Create FireCommand="\"Notification Refresh\"" FireTime=Now RearmPeriod="\"1 day @ hour 2\"" RearmOnFailure=True CatchUp=False Description="\"Delete Stale Notifications\""}
);

update(
    q{Creating new Allocation Refresh Event},
    q{Event Query FireCommand=="\"Allocation Refresh\"" Show:="\"Count(Id)\""},
    q{==},
    q{Event Create FireCommand="\"Allocation Refresh\"" FireTime=Now RearmPeriod="\"15 minutes^\"" RearmOnFailure=True CatchUp=False Description="\"Activate/Deactivate Stale Allocations\""}
);

################################################################################
# Transform Accounting Tables
################################################################################

print "\n***Updating UsageRecord table***\n";
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==UsageRecord Name==ExitCode Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==UsageRecord Name==CompletionCode Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
  )
{
    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 1});

    $sql = "UPDATE g_usage_record SET g_exit_code=g_completion_code";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";
    $sql = "UPDATE g_usage_record_log SET g_exit_code=g_completion_code";
    print "Executing sql: $sql\n";
    $count = $dbh->do($sql);
    print "Updated $count lines\n";

    $successes++;
}
else
{
    print "Already applied\n";
    $ignores++;
}

################################################################################
# Delete old Accounting Objects and Attributes
################################################################################

# Delete UsageRecord CompletionCode Attribute
if (
    (
        chomp(
            $_ =
              qx{mam-shell -q Attribute Query Object==UsageRecord Name==CompletionCode Show:=\"\\\"Count(Name)\\\"\"}
        )
        && $_ > 0
    )
    && ! $failures
  )
{
    update(
        q{Deleting old UsageRecord CompletionCode Attribute},
        q{Attribute Query Object==UsageRecord Name==CompletionCode Show:="\"Count(Name)\""},
        q{>},
        q{Attribute Delete Object==UsageRecord Name==CompletionCode NoRefresh:=True}
    );

    # Obtain a database handle
    my $dbh =
      DBI->connect($datasource, $dbuser, $dbpasswd,
        {AutoCommit => 1, RaiseError => 0});

    # UsageRecord
    $sql =
      "DELETE FROM g_attribute WHERE g_object='UsageRecord' AND g_name='CompletionCode'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql =
      "DELETE FROM g_attribute_log WHERE g_object='UsageRecord' AND g_name='CompletionCode'";
    print "Executing sql: $sql\n";
    $successes += $dbh->do($sql);
    $sql = "ALTER TABLE g_usage_record DROP COLUMN g_completion_code";
    print "Executing sql: $sql\n";
    $dbh->do($sql);
    $sql = "ALTER TABLE g_usage_record_log DROP COLUMN g_completion_code";
    print "Executing sql: $sql\n";
    $dbh->do($sql);

    $successes++;
}
else
{
    print "Already applied or dependencies not met\n";
    $ignores++;
}

################################################################################
# Refresh Cache and Update System Information
################################################################################

print "\n***Refreshing Cache***\n";
$cmd    = "mam-shell -v System Refresh";
$output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
$rc     = $? >> 8;
print("Subcommand ($cmd) returned with rc=$rc:\n$output") if $output;
if ($rc)
{
    $failures++;
    my $message = "Subcommand ($cmd) failed with rc=$rc";
    $message .= ": $output" if $output;
    warn($message);
    push @messages, $message;
}
else
{
    $successes++;
}

unless ($failures)
{
    update(
        q{Updating System Information},
        q{System Query Version==\"9.1\" Show:="\"Count(Name)\""},
        q{==}, q{System Modify Version=\"9.1\"}
    );
}

################################################################################
# Display Results
################################################################################

print "\n";
print "$successes updates were successfully applied.\n";
print "$ignores updates were skipped having been already applied.\n";
print "$failures updates failed.\n";
if (@messages)
{
    print "The following is a list of failures:\n";
    print join "\n", @messages;
}

exit 0;
