#! /usr/bin/perl -w
################################################################################
#
# MAM Custom Bank Class
#
# File   :  Bank.pm
#
################################################################################
#
#    This file includes parts which are copyrighted by Adaptive Computing
#    Enterprises, Inc., and parts which are copyrighted by Battelle Memorial
#    Institute.
#
#    The terms and conditions for the use and redistribution of these parts
#    are governed by the Moab Accounting Manager License and the BSD License
#    respectively. Refer to the LICENSE file for details.
#
#    Copyright (C) 2006 - 2016 by Adaptive Computing Enterprises, Inc.
#    All Rights Reserved.
#
#    The Moab Accounting Manager License specifies the terms and conditions
#    for use and redistribution.
#
#    The Moab Accounting Manager License applies to the Moab Accounting
#    Manager software offered by Adaptive Computing Enterprises, Inc.
#    By installing or using this software, Licensee accepts a non-exclusive
#    license from Adaptive Computing Enterprises, Inc. and is bound to accept
#    acknowledgement of and abide by the notices and conditions of the
#    Moab Accounting Manager License.
#
#    Copyright (C) 2003 - 2005 Pacific Northwest National Laboratory,
#    Battelle Memorial Institute. All rights reserved.
#
#    The BSD License specifies the terms and conditions for use and
#    redistribution.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions
#    are met:
#
#    - Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
#    - Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
#    - Neither the name of Battelle nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
#    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
#    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
#    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#    POSSIBILITY OF SUCH DAMAGE.
#
################################################################################

use strict;

=head1 NAME

MAM::Bank - implements custom bank actions

=head1 DESCRIPTION

The B<MAM::Bank> module handles custom bank actions through class methods such as begin, charge,end, deposit, quote, refund, reserve, start, stop, transfer, withdraw, update and overriding base methods.

=head1 METHODS

=over 4

=item $response = MAM::Bank->execute($proxy);

=item $response = MAM::Bank->begin($request, $requestId);

=item $response = MAM::Bank->charge($request, $requestId);

=item $response = MAM::Bank->create($request, $requestId);

=item $response = MAM::Bank->delete($request, $requestId);

=item $response = MAM::Bank->deposit($request, $requestId);

=item $response = MAM::Bank->end($request, $requestId);

=item $response = MAM::Bank->modify($request, $requestId);

=item $response = MAM::Bank->quote($request, $requestId);

=item $response = MAM::Bank->query($request, $requestId);

=item $response = MAM::Bank->refund($request, $requestId);

=item $response = MAM::Bank->refresh($request, $requestId);

=item $response = MAM::Bank->reserve($request, $requestId);

=item $response = MAM::Bank->reset($request, $requestId);

=item $response = MAM::Bank->start($request, $requestId);

=item $response = MAM::Bank->stop($request, $requestId);

=item $response = MAM::Bank->transfer($request, $requestId);

=item $response = MAM::Bank->withdraw($request, $requestId);

=item $response = MAM::Bank->undelete($request, $requestId);

=item $response = MAM::Bank->update($request, $requestId);

=item $response = MAM::Bank->usage($request, $requestId);

=item MAM::Bank->logTransaction(database => $database, object => $object, action => $action, actor => $actor, assignments => \@assignments, conditions => \@conditions, options => \@options, data => \@data, count => $count, fund => $fund, delta => $delta, balance => $balance, remaining => $remaining, allocation => $allocation, requestId => $requestId, txnId => $txnId);

=item ($charge, $itemizedCharges, $message, \@data) = MAM::Bank->calculateCharge(request => $request, requestId => $requestId, usageRecordProperties => \%usageRecordProperties, quote => $quoteId, duration => $duration, charge_ref => \$charge, itemizedCharges_ref => \$itemizedCharges, data_ref => \@data, metadata_ref => \%metadata, message_ref => \$chargeMessage);

=back

=head1 EXAMPLES

use MAM::Base;

my $response = MAM::Bank->charge($request, $requestId);

=head1 REQUIRES

Perl 5.8

=head1 AUTHOR

Scott Jackson, scottmo@adaptivecomputing.com

=cut

##############################################################################

package MAM::Bank;

use vars qw($log);
use Date::Manip;
use DBI qw( :sql_types );
use Time::Local;
use XML::LibXML;
use MAM::Cache;
use MAM::Global;
use MAM::Request;
use MAM::Response;

# ----------------------------------------------------------------------------
# $response = execute($proxy);
# ----------------------------------------------------------------------------

# Performs a custom bank action
sub execute
{
    my ($class, $proxy) = @_;

    if ($log->is_trace())
    {
        $log->trace('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $request      = $proxy->getRequest();
    my $action       = $request->getAction();
    my $requestId    = $proxy->getRequestId();
    my $lastResponse = $proxy->getResponse();
    my $response;

    # Provide usage if help flag specified
    my $showUsage = $request->getOptionValue('ShowUsage');
    if (defined $showUsage && $showUsage =~ /^true$/i)
    {
        # Call the usage for the custom bank action
        $response = MAM::Bank->usage($request, $requestId);
    }

    # Otherwise try to perform the custom action

    # Begin
    elsif ($action eq 'Begin')
    {
        $response = MAM::Bank->begin($request, $requestId);
    }

    # Charge
    elsif ($action eq 'Charge')
    {
        $response = MAM::Bank->charge($request, $requestId);
    }

    # Create
    elsif ($action eq 'Create')
    {
        $response = MAM::Bank->create($request, $requestId);
    }

    # Delete
    elsif ($action eq 'Delete')
    {
        $response = MAM::Bank->delete($request, $requestId);
    }

    # Deposit
    elsif ($action eq 'Deposit')
    {
        $response = MAM::Bank->deposit($request, $requestId);
    }

    # End
    elsif ($action eq 'End')
    {
        $response = MAM::Bank->end($request, $requestId);
    }

    # Modify
    elsif ($action eq 'Modify')
    {
        $response = MAM::Bank->modify($request, $requestId);
    }

    # Quote
    elsif ($action eq 'Quote')
    {
        $response = MAM::Bank->quote($request, $requestId);
    }

    # Query
    elsif ($action eq 'Query')
    {
     # If this is a continuation of a chunked response pass in next chunk number
        if ($lastResponse && $lastResponse->getChunkMax() == -1)
        {
            $response =
              MAM::Bank->query($request, $requestId,
                $lastResponse->getChunkNum() + 1);
        }
        else
        {
            $response = MAM::Bank->query($request, $requestId);
        }
    }

    # Refund
    elsif ($action eq 'Refund')
    {
        $response = MAM::Bank->refund($request, $requestId);
    }

    # Refresh
    elsif ($action eq 'Refresh')
    {
        $response = MAM::Bank->refresh($request, $requestId);
    }

    # Reserve
    elsif ($action eq 'Reserve')
    {
        $response = MAM::Bank->reserve($request, $requestId);
    }

    # Reset
    elsif ($action eq 'Reset')
    {
        $response = MAM::Bank->reset($request, $requestId);
    }

    # Start
    elsif ($action eq 'Start')
    {
        $response = MAM::Bank->start($request, $requestId);
    }

    # Stop
    elsif ($action eq 'Stop')
    {
        $response = MAM::Bank->stop($request, $requestId);
    }

    # Transfer
    elsif ($action eq 'Transfer')
    {
        $response = MAM::Bank->transfer($request, $requestId);
    }

    # Withdraw
    elsif ($action eq 'Withdraw')
    {
        $response = MAM::Bank->withdraw($request, $requestId);
    }

    # Undelete
    elsif ($action eq 'Undelete')
    {
        $response = MAM::Bank->undelete($request, $requestId);
    }

    # Update
    elsif ($action eq 'Update')
    {
        $response = MAM::Bank->update($request, $requestId);
    }

    else
    {
        $log->error("Unsupported action: ($action)");
        $response = new MAM::Response()
          ->failure('313', "Unsupported bank action: ($action)");
    }

    return $response;
}

# ----------------------------------------------------------------------------
# $response = begin($request, $requestId);
# ----------------------------------------------------------------------------

# Begin (Payment)
sub begin
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object   = $request->getObject();
    my $database = $request->getDatabase();
    my $actor    = $request->getActor();
    my @options  = $request->getOptions();
    my $now      = $requestTime;
    my $warnings = 0;
    my @messages = ();
    my $response;

    if ($object ne 'Payment')
    {
        return new MAM::Response()
          ->failure('315', "The Bank class does not implement $object Begin");
    }

    # Id
    my $id = $request->getOptionValue('Id');
    if (! defined $id)
    {
        return new MAM::Response()
          ->failure('314', 'An Id option must be specified');
    }

# SELECT BeginEvent,Duration,EndTime,Notify,ReservePattern FROM Payment WHERE Id=$id
    my $results = $database->select(
        object     => 'Payment',
        selections => [
            new MAM::Selection(name => 'BeginEvent'),
            new MAM::Selection(name => 'Duration'),
            new MAM::Selection(name => 'EndTime'),
            new MAM::Selection(name => 'Notify'),
            new MAM::Selection(name => 'ReservePattern'),
        ],
        conditions => [
            new MAM::Condition(
                name  => 'Id',
                value => $id,
            ),
        ],
    );

    unless (@{$results->{data}})
    {
        return new MAM::Response()
          ->failure('317', "Payment Begin: Invalid payment id ($id)");
    }

    my $beginEvent     = ${$results->{data}}[0]->[0];
    my $duration       = ${$results->{data}}[0]->[1];
    my $endTime        = ${$results->{data}}[0]->[2];
    my $notify         = ${$results->{data}}[0]->[3];
    my $reservePattern = ${$results->{data}}[0]->[4];
    unless (defined $reservePattern)
    {
        return new MAM::Response()
          ->failure('740',
            'Payment Begin: Unable to determine payment reserve mode');
    }

    # Invoke begin charges
    # SELECT Id FROM PaymentComponent WHERE Payment=$id AND Type='Begin'
    $results = $database->select(
        object     => 'PaymentComponent',
        selections => [new MAM::Selection(name => 'Id'),],
        conditions => [
            new MAM::Condition(
                name  => 'Payment',
                value => $id,
            ),
            new MAM::Condition(
                name  => 'Type',
                value => 'Begin',
            ),
        ],
    );

    # Iterate over begin components to issue setup charges
    my $chargeAmount    = 0;
    my @beginComponents = @{$results->{data}};
    foreach my $row (@{$results->{data}})
    {
        my $usageRecord = $row->[0];

        # UsageRecord Charge
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'UsageRecord',
            action   => 'Charge',
            actor    => $SUPER_USER,
            options  => [new MAM::Option(name => 'Id', value => $usageRecord)],
        );
        my $subResponse = MAM::Bank->charge($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Payment Begin failed in beginning payment charge: '
                  . $subResponse->getMessage());
        }
        $chargeAmount += $subResponse->getCount();
        push @messages, $subResponse->getMessage();
    }

    # Create new Payment EndEvent if necessary
    my $endEvent;
    if (defined $duration && ! defined $endTime)
    {
        my @eventAssignments = (
            new MAM::Assignment(
                name  => 'FireTime',
                value => $now + $duration,
            ),
            new MAM::Assignment(
                name  => 'FireCommand',
                value => "Payment End Id:=$id"
            ),
            new MAM::Assignment(
                name  => 'Description',
                value => 'End Payment'
            )
        );
        push @eventAssignments,
          new MAM::Assignment(name => 'Notify', value => $notify)
          if defined $notify;
        my $subRequest = new MAM::Request(
            database    => $database,
            object      => 'Event',
            action      => 'Create',
            actor       => $SUPER_USER,
            assignments => \@eventAssignments,
        );
        # Invoke the proxy so it prepares the subRequest by setting defaults
        MAM::Proxy->new(
            database  => $database,
            request   => $subRequest,
            requestId => $requestId
        );
        my $subResponse = MAM::Base->create($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Success')
        {
            # Figure out the id of the event we just created
            $endEvent = $subResponse->getDatumValue('Id');
            push @messages, $subResponse->getMessage();
        }
        else
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Payment Begin: Unable to create payment end event: '
                  . $subResponse->getMessage());
        }
    }

# UPDATE Payment SET BeginTime=$now,BeginEvent=NULL[,EndTime=$now + $duration][,EndEvent=$endEvent] WHERE Id=$id
    {
        my @paymentAssignments = (
            new MAM::Assignment(
                name  => 'BeginTime',
                value => $now,
            ),
            new MAM::Assignment(
                name  => 'BeginEvent',
                value => 'NULL',
            )
        );
        push @paymentAssignments,
          new MAM::Assignment(name => 'EndTime', value => $now + $duration)
          if (defined $duration && ! defined $endTime);
        push @paymentAssignments,
          new MAM::Assignment(name => 'EndEvent', value => $endEvent)
          if defined $endEvent;
        my $subRequest = new MAM::Request(
            database    => $database,
            object      => 'Payment',
            action      => 'Modify',
            actor       => $SUPER_USER,
            assignments => \@paymentAssignments,
            conditions  => [
                new MAM::Condition(
                    name  => 'Id',
                    value => $id
                )
            ],
        );
        my $subResponse = MAM::Base->modify($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            push @messages, 'Warning -- Unable to modify payment begin time';
            $warnings++;
        }
    }

    # Update StartTime of update component usage records
    # SELECT Id FROM PaymentComponent WHERE Payment=$id AND Type='Update'
    $results = $database->select(
        object     => 'PaymentComponent',
        selections => [new MAM::Selection(name => 'Id'),],
        conditions => [
            new MAM::Condition(
                name  => 'Payment',
                value => $id,
            ),
            new MAM::Condition(
                name  => 'Type',
                value => 'Update',
            ),
        ],
    );

    # Iterate over update components
    foreach my $row (@{$results->{data}})
    {
        my $usageRecord = $row->[0];

        # UsageRecord Modify
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'UsageRecord',
            action   => 'Modify',
            actor    => $SUPER_USER,
            conditions =>
              [new MAM::Condition(name => 'Id', value => $usageRecord)],
            assignments =>
              [new MAM::Assignment(name => 'StartTime', value => $now)],
        );
        my $subResponse = MAM::Base->modify($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Payment Begin failed setting start time in payment update component usage record: '
                  . $subResponse->getMessage());
        }
    }

    # Delete Payment BeginEvent
    if (defined $beginEvent)
    {
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'Event',
            action   => 'Delete',
            actor    => $SUPER_USER,
            conditions =>
              [new MAM::Condition(name => 'Id', value => $beginEvent)],
        );
        my $subResponse = MAM::Base->delete($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            push @messages, 'Warning -- Unable to delete begin event: '
              . $subResponse->getMessage();
            $warnings++;
        }
    }

    # Delete Begin PaymentComponents
    if (@beginComponents)
    {
        my $subRequest = new MAM::Request(
            database   => $database,
            object     => 'PaymentComponent',
            action     => 'Delete',
            actor      => $SUPER_USER,
            conditions => [
                new MAM::Condition(
                    name  => 'Payment',
                    value => $id,
                ),
                new MAM::Condition(
                    name  => 'Type',
                    value => 'Begin',
                ),
            ],
        );
        my $subResponse = MAM::Base->delete($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            push @messages,
              'Warning -- Unable to delete begin payment components: '
              . $subResponse->getMessage();
            $warnings++;
        }
    }

    # Log the transaction
    MAM::Bank->logTransaction(
        database  => $database,
        requestId => $requestId,
        txnId     => $database->nextId('Transaction'),
        object    => 'Payment',
        action    => 'Begin',
        actor     => $actor,
        options   => \@options,
        count     => 1,
    );

    # Call Payment Start
    my $startOption = $request->getOptionValue('Start');
    unless (defined $startOption && $startOption =~ /^false$/i)
    {
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'Payment',
            action   => 'Start',
            actor    => $SUPER_USER,
            options  => [new MAM::Option(name => 'Id', value => $id)],
        );
        my $subResponse = MAM::Bank->start($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Payment Begin failed starting payment: '
                  . $subResponse->getMessage());
        }
        push @messages, $subResponse->getMessage();
    }

    # Return response
    push @messages, 'No action taken' unless @messages;
    if ($warnings)
    {
        $response = new MAM::Response()
          ->warning($chargeAmount, 'Payment Begin: ' . join("\n", @messages));
    }
    else
    {
        $response = new MAM::Response()
          ->success($chargeAmount, 'Payment Begin: ' . join("\n", @messages));
    }
    return $response;
}

# ----------------------------------------------------------------------------
# $response = charge($request, $requestId);
# ----------------------------------------------------------------------------

# Charge (UsageRecord)
sub charge
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object          = $request->getObject();
    my $actor           = $request->getActor();
    my $database        = $request->getDatabase();
    my @options         = $request->getOptions();
    my $incremental     = $request->getOptionValue('Incremental') || 'False';
    my $accounting_mode = $serverConfig->{'accounting.mode'}
      || $ACCOUNTING_MODE;
    my $accountingModeOption = $request->getOptionValue('AccountingMode');
    my $currency_precision   = $serverConfig->{'currency.precision'} || 0;
    my $chargeFund           = $request->getOptionValue('Fund');
    my $time_division =
      defined $request->getOptionValue('ShowHours') ? 3600 : 1;
    my $now                = $requestTime;
    my $balance            = 0;
    my $createdUsageRecord = 0;
    my %allocations        = ();
    my %charges            = ();
    my %liens              = ();
    my %fundAllocations    = ();
    my %lienAllocations    = ();
    my %chargeChunk        = ();
    my $infinity           = 2147483647;
    my $message            = '';
    my $quoteMessage       = '';
    my $reserveMessage     = '';
    my $chargeMessage      = '';
    my $warningMessage     = '';
    my $warnings           = 0;
    my $failureCode        = '';
    my $failureMessage     = '';
    my @data               = ();
    my %metadata           = ();
    my $results;
    my $response;

    if ($object ne 'UsageRecord')
    {
        return new MAM::Response()
          ->failure('315', "The Bank class does not implement $object Charge");
    }

    my $balanceFeedback;
    my $balanceFeedbackOption = $request->getOptionValue('BalanceFeedback');
    if (defined $balanceFeedbackOption && $balanceFeedbackOption =~ /^true$/i)
    {
        $balanceFeedback = 1;
    }
    else
    {
        $balanceFeedback = 0;
    }

    my $itemize;
    my $itemizeOption = $request->getOptionValue('Itemize');
    if (defined $itemizeOption && $itemizeOption =~ /^true$/i)
    {
        $itemize = 1;
    }
    else
    {
        $itemize = 0;
    }

    if ($balanceFeedback && $itemize)
    {
        return new MAM::Response()
          ->failure('318',
            'The Itemize and BalanceFeedback options are mutually exclusive');
    }

    # Demote accounting mode if option is lower than configuration
    if (defined $accountingModeOption)
    {
        if ((   $accountingModeOption eq 'usage-tracking'
                && (   $accounting_mode eq 'usage-tracking'
                    || $accounting_mode eq 'notional-charging'
                    || $accounting_mode eq 'fast-allocation'
                    || $accounting_mode eq 'strict-allocation')
            )
            || ($accountingModeOption eq 'notional-charging'
                && (   $accounting_mode eq 'notional-charging'
                    || $accounting_mode eq 'fast-allocation'
                    || $accounting_mode eq 'strict-allocation')
            )
            || ($accountingModeOption eq 'fast-allocation'
                && (   $accounting_mode eq 'fast-allocation'
                    || $accounting_mode eq 'strict-allocation')
            )
            || (   $accountingModeOption eq 'strict-allocation'
                && $accounting_mode eq 'strict-allocation')
          )
        {
            $accounting_mode = $accountingModeOption;
        }
        else
        {
            $warningMessage .=
              "\nWarning -- The specified accounting mode ($accountingModeOption) is being ignored because it is more strict than the accounting mode configured in Moab Accounting Manager ($accounting_mode).";
            $warnings++;
        }
    }

    # Check if data is present and parse out data properties
    my $data = $request->getDataElement();
    my $doc  = XML::LibXML::Document->new();
    $doc->setDocumentElement($data);
    my @rows = $data->childNodes();

    # We currently only support a solitary usage record
    if (@rows == 1)
    {
        return new MAM::Response()
          ->failure('314', 'The Data must be a Usage Record')
          unless $rows[0]->nodeName eq 'UsageRecord';
    }
    elsif (@rows > 1)
    {
        return new MAM::Response()
          ->failure('314', 'Only one UsageRecord may be specified in the Data');
    }

    # Build up a list of usage record properties from the data
    my %dataProperties       = ();
    my %chargeDataProperties = ();
    if (@rows > 0)
    {
        foreach my $col ($rows[0]->childNodes())
        {
            my $propertyName = $col->nodeName();
            my $increment    = $col->getAttribute('increment');
            if ($col->hasChildNodes())    # Simple properties only
            {
                my $propertyValue   = $col->textContent();
                my $consumptionRate = $col->getAttribute('consumptionRate');
                $propertyValue .= "*$consumptionRate"
                  if defined $consumptionRate;
                $dataProperties{$propertyName} = $propertyValue;
                if (defined $increment)
                {
                    $chargeDataProperties{$propertyName} = $increment;
                }
                else
                {
                    $chargeDataProperties{$propertyName} = $propertyValue;
                }

            }
            elsif (defined $increment)
            {
                $chargeDataProperties{$propertyName} = $increment;
            }
        }
    }

    # Id (usage record id)
    my $id;
    my $idOption = $request->getOptionValue('Id');
    if (! defined $idOption && ! keys %dataProperties)
    {
        return new MAM::Response()
          ->failure('314',
            'Either an Id option or usage record data is required');
    }
    my $idProperty = $dataProperties{Id};
    if (defined $idOption)
    {
        return new MAM::Response()->failure('318',
            "The specified Id option ($idOption) differs from the specified Id usage record property ($idProperty)"
        ) if (defined $idProperty && $idProperty ne $idOption);
        $id = $idOption;
    }
    else
    {
        $id = $idProperty;
    }

    # Look up existing usage record properties if id is known
    my %usageRecordProperties = ();
    my %chargeProperties      = ();
    if (defined $id)
    {
        my $subRequest = new MAM::Request(
            database   => $database,
            object     => 'UsageRecord',
            action     => 'Query',
            actor      => $SUPER_USER,
            conditions => [new MAM::Condition(name => 'Id', value => $id)],
        );
        # Invoke the proxy so it prepares the subRequest by setting defaults
        MAM::Proxy->new(
            database  => $database,
            request   => $subRequest,
            requestId => $requestId
        );
        # But use select so we can parse the data results easier
        my @preparedSelections = $subRequest->getSelections();
        # SELECT * FROM UsageRecord WHERE Id=$id
        my $subResults = $database->select(
            object     => 'UsageRecord',
            selections => \@preparedSelections,
            conditions => [new MAM::Condition(name => 'Id', value => $id)],
        );
        if (@{$subResults->{data}})
        {
            my $row   = ${$subResults->{data}}[0];
            my $cols  = $subResults->{cols};
            my $names = $subResults->{names};
            for (my $i = 0; $i < $cols; $i++)
            {
                my $name = $names->[$i];
                next if $name eq 'Duration';  # Don't pass in extant UR duration
                my $value = $row->[$i];       # Null comes in as undef
                if (defined $value)
                {
                    $usageRecordProperties{$name} = $value;
                    $chargeProperties{$name} = $value if $value =~ /^\D/;
                }
            }
        }
        else
        {
            return new MAM::Response()
              ->failure('740', "No usage record found for specified id ($id)");
        }
    }

    # Merge data properties in with existing usage record and charge properties
    foreach my $name (keys %dataProperties)
    {
        $usageRecordProperties{$name} = $dataProperties{$name};
    }
    foreach my $name (keys %chargeDataProperties)
    {
        $chargeProperties{$name} = $chargeDataProperties{$name};
    }

    # Employ object instance overrides for specified usage record properties
    my @overrideCheckProperties = keys %usageRecordProperties;
    while (my $propertyName = shift @overrideCheckProperties)
    {
        # If the usage record property has an object associated with it
        if (MAM::Cache->objectExists($propertyName))
        {
            # Iterate over its attributes
            foreach my $attribute (
                MAM::Cache->listAttributesWithoutPropertyValue(
                    $propertyName, 'Values', undef
                )
              )
            {
                my $values =
                  MAM::Cache->getAttributeProperty($propertyName, $attribute,
                    'Values');
                # If an attribute is an override attribute
                if ($values =~ /^@([:?!]=)(\w+)/)
                {
                    my ($overrideType, $overrideName) = ($1, $2);
# See if the instance has a value set for this attribute
# Assume the primary key is Name
# SELECT $attribute FROM $propertyName WHERE Name=$usageRecordProperties{$propertyName}
                    my $subResults = $database->select(
                        object     => "$propertyName",
                        selections => [new MAM::Selection(name => $attribute)],
                        conditions => [
                            new MAM::Condition(
                                name  => 'Name',
                                value => $usageRecordProperties{$propertyName}
                            )
                        ]
                    );
                    # If the instance override has a value
                    if (@{$subResults->{data}}
                        && defined ${$subResults->{data}}[0]->[0])
                    {
                        my $overrideValue = ${$subResults->{data}}[0]->[0];
                        # Override it with the value if the usage record
                        # property is undefined or if it has a different
                        # value and this is an assign-always override
                        if ((! defined $usageRecordProperties{$overrideName})
                            || (   $overrideType eq ':='
                                && $usageRecordProperties{$overrideName} ne
                                $overrideValue)
                          )
                        {
                            $usageRecordProperties{$overrideName} =
                              $overrideValue;
                            $chargeProperties{$overrideName} = $overrideValue;
                            push @overrideCheckProperties, $overrideName;
                        }
                        # Or return with a failure if the usage record
                        # property is defined but has a different
                        # value and this is an assign-verify override
                        elsif ($overrideType eq '!='
                            && defined $usageRecordProperties{$overrideName}
                            && $usageRecordProperties{$overrideName} ne
                            $overrideValue)
                        {
                 # Since the job has actually run, we do not want to completely
                 # bail out here (rollback). We will still log the usage, record
                 # the transaction and cleanup liens. We mark it as a failure so
                 # that the charge does not occur.
                 # We will also override with the correct value so that usage is
                 # tracked according to the administrator's intentions.

                            $failureCode = '740';
                            $failureMessage =
                              "Specified $overrideName ($usageRecordProperties{$overrideName}) conflicts with ${propertyName}-designated $overrideName ($overrideValue)";

                            $usageRecordProperties{$overrideName} =
                              $overrideValue;
                            $chargeProperties{$overrideName} = $overrideValue;
                            push @overrideCheckProperties, $overrideName;
                        }
                    }
                }
            }
        }
    }

    # Build up a list of usage record attributes and set defaults
    my %usageRecordAttributes = ();
    foreach my $attributeName (
        MAM::Cache->listAttributesWithPropertyValue(
            'UsageRecord', 'Hidden', 'False'
        )
      )
    {
        $usageRecordAttributes{$attributeName} = 1;
        next if $attributeName eq 'Duration';   # Don't set default for duration

        # Set default values for all undefined usage record properties that are
        # first class objects with a default value
        my $defaultValue =
          MAM::Cache->getObjectProperty($attributeName, 'DefaultValue');
        if (! defined $usageRecordProperties{$attributeName}
            && defined $defaultValue)
        {
            $usageRecordProperties{$attributeName} = $defaultValue;
            $chargeProperties{$attributeName}      = $defaultValue;
        }
    }

    # Build up list of usage record assignments
    # Verify that every specified usage record property exists if it is a value-
    # constrained object and that it is active if it has an active field
    my @usageRecordAssignments = ();
    foreach my $propertyName (sort keys %usageRecordProperties)
    {
        if ($usageRecordAttributes{$propertyName})
        {
            my $propertyValue = $usageRecordProperties{$propertyName};
            # Remove coefficients and consumption rates
            $propertyValue =~ s/\*(\d*\.?\d+)$//;
            push @usageRecordAssignments,
              new MAM::Assignment(
                name  => $propertyName,
                value => $propertyValue
              )
              unless ($propertyName eq 'Id'
                || $propertyName eq 'Charge');

            my $values =
              MAM::Cache->getAttributeProperty('UsageRecord', $propertyName,
                'Values');
            if (   defined $values
                && $values =~ /^@(?:[:?!]=)?(\w+)/
                && MAM::Cache->objectExists($1))
            {
                my $propertyObject = $1;
                my @primaryKeys =
                  MAM::Cache->listAttributesWithPropertyValue($propertyObject,
                    'PrimaryKey', 'True');

                # We can only do this if it has a single primary key
                if (@primaryKeys == 1)
                {
# SELECT $propertyKey[,Active] FROM $filterObject WHERE $propertyKey=$propertyValue
                    my $propertyKey   = $primaryKeys[0];
                    my $propertyValue = $usageRecordProperties{$propertyName};
                    my @validitySelections =
                      (new MAM::Selection(name => $propertyKey));
                    push @validitySelections,
                      new MAM::Selection(name => 'Active')
                      if MAM::Cache->attributeExists($propertyObject, 'Active');

                    my @validityConditions = (
                        new MAM::Condition(
                            name  => $propertyKey,
                            value => $propertyValue
                        )
                    );
                    my $subResults = $database->select(
                        object     => $propertyObject,
                        selections => \@validitySelections,
                        conditions => \@validityConditions,
                    );
                    unless (@{$subResults->{data}})
                    {
                        # Check to see if the instance should be auto-generated
                        if (MAM::Cache->getObjectProperty($propertyObject,
                                'AutoGen') eq 'True'
                          )
                        {
                            # Create the instance
                            my @autoGenAssignments = (
                                new MAM::Assignment(
                                    name  => $propertyKey,
                                    value => $propertyValue
                                )
                            );
                            if (MAM::Cache->attributeExists(
                                    $propertyObject, 'Description'
                                )
                              )
                            {
                                push @autoGenAssignments,
                                  new MAM::Assignment(
                                    name  => 'Description',
                                    value => 'Auto-Generated'
                                  );
                            }
                            my $subRequest = new MAM::Request(
                                database    => $database,
                                object      => $propertyObject,
                                action      => 'Create',
                                actor       => $SUPER_USER,
                                assignments => \@autoGenAssignments,
                            );
            # Invoke the proxy so it prepares the subRequest by setting defaults
                            MAM::Proxy->new(
                                database  => $database,
                                request   => $subRequest,
                                requestId => $requestId
                            );
                            my $subResponse =
                              MAM::Base->create($subRequest, $requestId);
                            if ($subResponse->getStatus() eq 'Failure')
                            {
                                my $message =
                                  "Unable to auto-gen $propertyObject $propertyValue: "
                                  . $subResponse->getMessage();
                                $log->error($message);
                                throw MAM::Exception($subResponse->getCode(),
                                    $message);
                            }
                        }
                        else
                        {
                 # Since the job has actually run, we do not want to completely
                 # bail out here (rollback). We will still log the usage, record
                 # the transaction and cleanup liens. We mark it as a failure so
                 # that the charge does not occur.
                            $failureCode = '740';
                            $failureMessage =
                              "$propertyObject ($propertyName) $propertyValue does not exist";
                        }
                    }
                    elsif (exists ${$subResults->{data}}[0]->[1]
                        && ${$subResults->{data}}[0]->[1] ne 'True')
                    {
                 # Since the job has actually run, we do not want to completely
                 # bail out here (rollback). We will still log the usage, record
                 # the transaction and cleanup liens. We mark it as a failure so
                 # that the charge does not occur.
                        $failureCode = '740';
                        $failureMessage =
                          "$propertyObject $propertyValue is not active";
                    }
                }
            }
        }
    }

    # Start Time, End Time and Duration
    my ($chargeStartTime, $chargeEndTime, $chargeDuration);
    my $startTimeOption   = $request->getOptionValue('StartTime');
    my $startTimeProperty = $usageRecordProperties{StartTime};
    my $endTimeOption     = $request->getOptionValue('EndTime');
    my $endTimeProperty   = $usageRecordProperties{EndTime};
    my $durationOption    = $request->getOptionValue('Duration');
    my $durationProperty  = $usageRecordProperties{Duration};

    if (defined $endTimeOption)
    {
        $chargeEndTime = $endTimeOption;
    }
    else    # Use default if duration was not obtained via option
    {
        $chargeEndTime = $now;
    }

    if (defined $durationOption)
    {
        $chargeDuration = $durationOption;
        if (defined $startTimeOption)
        {
            $chargeStartTime = $startTimeOption;
        }
        else
        {
            $chargeStartTime = $chargeEndTime - $chargeDuration;
        }
    }
    elsif (defined $startTimeOption)
    {
        $chargeStartTime = $startTimeOption;
        $chargeDuration  = $chargeEndTime - $chargeStartTime;
    }
    else
    {
        # Override current end time with end time property if specified
        if ((! defined $endTimeOption) && defined $endTimeProperty)
        {
            $chargeEndTime = $endTimeProperty;
        }

        if (defined $durationProperty)
        {
            $chargeDuration = $durationProperty;
            if (defined $startTimeProperty)
            {
                $chargeStartTime = $startTimeProperty;
            }
            else
            {
                $chargeStartTime = $chargeEndTime - $chargeDuration;
            }
        }
        elsif (defined $startTimeProperty)
        {
            $chargeStartTime = $startTimeProperty;
            $chargeDuration  = $chargeEndTime - $chargeStartTime;
        }
        else
        {
            $chargeStartTime = $chargeEndTime;
            $chargeDuration  = 0;
        }
    }
    if ($chargeDuration !~ /^\d+$/)
    {
        return new MAM::Response()->failure('317',
            "The duration ($chargeDuration) must be a non-negative integer number of seconds"
        );
    }

    # Build up list of chargeable funds if enforcing allocations
    if (   $accounting_mode ne 'usage-tracking'
        && $accounting_mode ne 'notional-charging')
    {
        # Refresh Allocations
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'Allocation',
            action   => 'Refresh'
        );
        my $subResponse = MAM::Bank->refresh($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Failed while refreshing allocations: '
                  . $subResponse->getMessage());
        }

# First, we build a list of funds along with their eligible allocations
# We need to do this separately from the constraint and lien queries
# because the trivial fund might have no constraints nor liens
# SELECT DISTINCT Allocation.Id,Allocation.Fund,Allocation.Amount,Allocation.CreditLimit,Allocation.StartTime,Allocation.EndTime,Fund.Priority FROM Allocation,Fund WHERE Allocation.Fund=Fund.Id AND Allocation.StartTime<=$chargeEndTime AND Allocation.EndTime>$chargeEndTime
        my @allocationConditions = ();
        if (($serverConfig->{'debit.activeallocationonly'} || 'false') =~
            /^true$/i)
        {
            push @allocationConditions,
              ( new MAM::Condition(
                    object => 'Allocation',
                    name   => 'StartTime',
                    value  => $now,
                    op     => 'LE',
                ),
                new MAM::Condition(
                    object => 'Allocation',
                    name   => 'EndTime',
                    value  => $now,
                    op     => 'GT',
                )
              );
        }
        elsif (
            ($serverConfig->{'debit.pointintimeonly'} || 'false') =~ /^true$/i)
        {
            push @allocationConditions,
              ( new MAM::Condition(
                    object => 'Allocation',
                    name   => 'StartTime',
                    value  => $chargeEndTime,
                    op     => 'LE',
                ),
                new MAM::Condition(
                    object => 'Allocation',
                    name   => 'EndTime',
                    value  => $chargeEndTime,
                    op     => 'GT',
                )
              );
        }
        else
        {
            push @allocationConditions,
              ( new MAM::Condition(
                    object => 'Allocation',
                    name   => 'StartTime',
                    value  => $chargeEndTime,
                    op     => 'LE',
                ),
                new MAM::Condition(
                    object => 'Allocation',
                    name   => 'EndTime',
                    value  => $chargeStartTime,
                    op     => 'GT',
                )
              );
        }
        my @fundConditions = (
            new MAM::Condition(
                object  => 'Allocation',
                name    => 'Fund',
                subject => 'Fund',
                value   => 'Id',
            ),
        );
        push @fundConditions,
          new MAM::Condition(
            object => 'Fund',
            name   => 'Id',
            value  => $chargeFund
          ) if defined $chargeFund;
        $results = $database->select(
            objects => [
                new MAM::Object(name => 'Allocation'),
                new MAM::Object(name => 'Fund'),
            ],
            selections => [
                new MAM::Selection(object => 'Allocation', name => 'Id'),
                new MAM::Selection(object => 'Allocation', name => 'Fund'),
                new MAM::Selection(object => 'Allocation', name => 'Amount'),
                new MAM::Selection(
                    object => 'Allocation',
                    name   => 'CreditLimit'
                ),
                new MAM::Selection(
                    object => 'Allocation',
                    name   => 'StartTime'
                ),
                new MAM::Selection(object => 'Allocation', name => 'EndTime'),
                new MAM::Selection(object => 'Fund',       name => 'Priority'),
            ],
            conditions => [@fundConditions, @allocationConditions,],
            options => [new MAM::Option(name => 'Unique', value => 'True')]
        );
        foreach my $row (@{$results->{data}})
        {
            my $id          = $row->[0];
            my $fund        = $row->[1];
            my $amount      = $row->[2] || 0;
            my $creditLimit = $row->[3] || 0;
            my $startTime   = $row->[4];
            my $endTime     = $row->[5];
            my $priority    = $row->[6] || 0;

            push @{$fundAllocations{$fund}}, {
                id        => $id,
                fund      => $fund,
                balance   => $amount + $creditLimit,
                startTime => $startTime,
                endTime   => $endTime,
                # If you want it to take precedence over endTime, use millions
                weight => 100 * int(($infinity - $endTime) / 86400) +
                  10 * $priority,
            };
        }

       # Next, we prune off the funds that violate the constraints to limbo
       # In order to limit the list of constraints we must loop on, we do a join
       # on eligible allocations instead of constraining it to a huge fund list
        my %fundConstraints = ();
        my %limbo           = ();

        # Build up an association table for quick checks
        my %associations = ();
        foreach my $association (MAM::Cache->associationLookup(undef, undef))
        {
            my $parent = MAM::Cache->getObjectProperty($association, 'Parent');
            next unless defined $usageRecordProperties{$parent};
            my $child = MAM::Cache->getObjectProperty($association, 'Child');
            $associations{$parent}{$child} = $association;
        }

# SELECT DISTINCT Constraint.Id,Constraint.Fund,Constraint.Name,Constraint.Value FROM Constraint,Allocation WHERE Constraint.Fund=Allocation.Fund AND Allocation.StartTime<=$chargeEndTime AND Allocation.EndTime>$chargeEndTime
        $results = $database->select(
            objects => [
                new MAM::Object(name => 'Constraint'),
                new MAM::Object(name => 'Allocation'),
            ],
            selections => [
                new MAM::Selection(object => 'Constraint', name => 'Id'),
                new MAM::Selection(object => 'Constraint', name => 'Fund'),
                new MAM::Selection(object => 'Constraint', name => 'Name'),
                new MAM::Selection(object => 'Constraint', name => 'Value'),
            ],
            conditions => [
                new MAM::Condition(
                    object  => 'Constraint',
                    name    => 'Fund',
                    subject => 'Allocation',
                    value   => 'Fund',
                ),
                @allocationConditions,
            ],
            options => [new MAM::Option(name => 'Unique', value => 'True')]
        );
        foreach my $row (@{$results->{data}})
        {
            my $id     = $row->[0];
            my $fundId = $row->[1];
            my $name   = $row->[2];
            my $value  = $row->[3];

            push @{$fundConstraints{$fundId}{$name}}, $value;
        }

        my %accepted    = ();
        my %rejected    = ();
        my %memberships = ();
      FUNDID:
        foreach my $fundId (sort { $a <=> $b } keys %fundConstraints)
        {
            foreach my $constraintName (sort keys %{$fundConstraints{$fundId}})
            {
                foreach my $constraintVal (
                    @{$fundConstraints{$fundId}{$constraintName}})
                {
                    my $constraintValue = $constraintVal;    # Make a copy
                    my $accessAllowed = $constraintValue !~ s/^!//;

                   # If there is a usage record property matching the constraint
                    if (defined $usageRecordProperties{$constraintName}
                        && $usageRecordProperties{$constraintName} eq
                        $constraintValue)
                    {
                        if ($accessAllowed)
                        {
                            # Check if the usage record property is a
                            # parent in an association
                            if (defined $associations{$constraintName})
                            {
                                foreach my $childName (
                                    keys %{$associations{$constraintName}})
                                {
                                    my $association =
                                      $associations{$constraintName}
                                      {$childName};

                                    # If so, the child should be specified
                                    if (defined
                                        $usageRecordProperties{$childName})
                                    {
                             # If there is a matching member found, add the fund

                                        # Check for a match unless the
                                        # fund has a matching constraint
                                        unless (
                                            exists $fundConstraints{$fundId}
                                            {$childName})
                                        {

                                            # Use cache for membership results
                                            unless (
                                                exists
                                                $memberships{$association})
                                            {
                           # e.g. Account==chemistry && (Name==amy || Name==ANY)
                                                my @membershipConditions = (
                                                    new MAM::Condition(
                                                        name => $constraintName,
                                                        value =>
                                                          $usageRecordProperties{
                                                            $constraintName},
                                                    ),
                                                    new MAM::Condition(
                                                        name => 'Name'
                                                        , # Assume simple association
                                                        value =>
                                                          $usageRecordProperties{
                                                            $childName},
                                                        conj  => 'And',
                                                        group => '+1',
                                                    ),
                                                    new MAM::Condition(
                                                        name => 'Name'
                                                        , # Assume simple association
                                                        value => 'ANY',
                                                        conj  => 'Or',
                                                        group => '-1',
                                                    ),
                                                );
                                                push @membershipConditions,
                                                  new MAM::Condition(
                                                    name  => 'Active',
                                                    value => 'True'
                                                  )
                                                  if
                                                  MAM::Cache->attributeExists(
                                                    $association, 'Active');
                                                my $subResults =
                                                  $database->select(
                                                    object => $association,
                                                    conditions =>
                                                      \@membershipConditions,
                                                  );
                                                $memberships{$association} =
                                                  scalar @{$subResults->{data}};
                                            }
                                            unless ($memberships{$association})
                                            {
                            # Since the job has actually run, we do not want to
                            # completely bail out here (rollback). We will still
                            # log the usage, record the transaction and cleanup
                            # liens. We mark it as a failure so that the charge
                            # does not occur.
                                                $failureCode = '740';
                                                $failureMessage =
                                                  "$childName $usageRecordProperties{$childName} is not a valid member of $constraintName $usageRecordProperties{$constraintName}";
                                            }
                                        }
                                    }
                                    else
                                    {
                    # Since the job has actually run, we do not want to
                    # completely bail out here (rollback). We will still log
                    # the usage, record the transaction and cleanup liens.
                    # We mark it as a failure so that the charge does not occur.
                                        $failureCode = '740';
                                        $failureMessage =
                                          "A $childName must be specified when specifying a $constraintName";
                                    }
                                }
                            }
                            $accepted{$fundId}{$constraintName} = 1;
                            # Add to the allocation weight for each constraint
                            foreach
                              my $allocation (@{$fundAllocations{$fundId}})
                            {
                                $allocation->{weight} += 1;
                            }
                        }
                        else
                        {
                            if (defined $fundAllocations{$fundId})
                            {
                                # Move them to limbo
                                @{$limbo{$fundId}} =
                                  @{$fundAllocations{$fundId}};
                                delete $fundAllocations{$fundId};
                                next FUNDID;
                            }
                        }
                    }
         # There is not a usage record property matching the positive constraint
         # or the filter does not exist or the constraint is negative
                    else
                    {
                        if ($accessAllowed
                            || !
                            defined $usageRecordProperties{$constraintName})
                        {
                            if (defined $fundAllocations{$fundId})
                            {
                                # Provisionally Reject the constraint
                                $rejected{$fundId}{$constraintName} = 1;
                            }
                        }
                    }
                }
            }
        }

        # Move all funds to limbo that have rejected constraints and that
        # do not have accepted constraints
        foreach my $fundId (sort { $a <=> $b } keys %rejected)
        {
            foreach my $constraintName (sort keys %{$rejected{$fundId}})
            {
                unless (exists $accepted{$fundId}{$constraintName}
                    || ! defined $fundAllocations{$fundId})
                {
                    # Move them to limbo
                    @{$limbo{$fundId}} = @{$fundAllocations{$fundId}};
                    delete $fundAllocations{$fundId};
                }
            }
        }

        # Then, we build up fund linkages for ancestor funds
        my %ancestors = ();
        # SELECT Fund,Id FROM FundFund WHERE Overflow=True
        $results = $database->select(
            object     => 'FundFund',
            selections => [
                new MAM::Selection(name => 'Fund'),
                new MAM::Selection(name => 'Id'),
            ],
            conditions =>
              [new MAM::Condition(name => 'Overflow', value => 'True')],
        );
        foreach my $row (@{$results->{data}})
        {
            my $parent = $row->[0];
            my $child  = $row->[1];
            $ancestors{$child}{$parent} = 1;
        }

        # Add ancestors a generation at a time until there is no more growth
        my $distance = 0;
        while (1)
        {
            $distance++;
            my $count = scalar keys %fundAllocations;
            foreach my $fund (keys %fundAllocations)
            {
                foreach my $parent (keys %{$ancestors{$fund}})
                {
                    if ($limbo{$parent})
                    {
                      # Move it back from limbo with negative distance as weight
                        foreach my $allocation (@{$limbo{$parent}})
                        {
                            $allocation->{weight} =
                              $distance * 100000 - ($allocation->{endTime});
                        }
                        @{$fundAllocations{$parent}} = @{$limbo{$parent}};
                        delete $limbo{$parent};
                    }
                }
            }
            last if $count == scalar keys %fundAllocations;
        }

        if ($log->is_debug())
        {
            $log->debug('Feasible fund list is: ' . join ',',
                sort keys %fundAllocations);
        }

        # Begin building allocations list
        foreach my $fundId (sort { $a <=> $b } keys %fundAllocations)
        {
            foreach my $allocation (@{$fundAllocations{$fundId}})
            {
                $allocations{$allocation->{id}} = $allocation;
            }
        }
    }

    # Instance
    my $instance = $usageRecordProperties{Instance};

   # Look up the quote from the usage record so we can use it for quoted rates
   # Postpone for now since this could be quite expensive for large usage tables

    # Quote Id
    my $quoteId = $usageRecordProperties{Quote};

    # Enforce quotes if notional charging or enforcing allocations
    if ($accounting_mode ne 'usage-tracking' && defined $quoteId)
    {
# Check to see if the quote exists and is not expired or claimed by a different instance
# SELECT Pinned,Instance,UsageRecord FROM Quote WHERE Id=$quoteId AND StartTime<=$chargeEndTime AND EndTime>$chargeEndTime
        my $subResults = $database->select(
            object     => 'Quote',
            selections => [
                new MAM::Selection(name => 'Pinned'),
                new MAM::Selection(name => 'Instance'),
                new MAM::Selection(name => 'UsageRecord')
            ],
            conditions => [
                new MAM::Condition(name => 'Id', value => $quoteId),
                new MAM::Condition(
                    name  => 'StartTime',
                    value => $chargeEndTime,
                    op    => 'LE'
                ),
                new MAM::Condition(
                    name  => 'EndTime',
                    value => $chargeEndTime,
                    op    => 'GT'
                )
            ]
        );
        if (@{$subResults->{data}})
        {
            my $row              = ${$subResults->{data}}[0];
            my $quotePinned      = $row->[0];
            my $quoteInstance    = $row->[1];
            my $quoteUsageRecord = $row->[2];

            # Reject all pinned quotes when either the usage instance is not
            # specified or it is specified but does not match the quote instance
            if (   defined $quotePinned
                && $quotePinned eq 'True'
                && (   (! defined $instance)
                    || (defined $quoteInstance && $quoteInstance ne $instance))
              )
            {
                # Nullify the quote if absent, expired or used
                $quoteMessage .= "\nWarning -- Quote ($quoteId) is pinned";
                $quoteMessage .= " by instance ($quoteInstance)"
                  if defined $quoteInstance;
                $quoteMessage .= ' and cannot be used';
                $quoteMessage .= ' anonymously' if ! defined $quoteInstance;
                $warnings++;

                undef $quoteId;
                @usageRecordAssignments =
                  grep { $_->getName() ne 'Quote' } @usageRecordAssignments;
            }
# Accept all unpinned quotes or
# ( pinned quotes where a usage instance has been specified
# in the charge and ( where the quote instance is either
# uninitialized or matches the usage instance ))
# if ((! defined $quotePinned) || $quotePinned ne 'True' || (defined $instance &&
# (!(defined $quoteInstance) || $quoteInstance eq $instance)))
            else
            {
                # Set usage record id from quote
                $id = $quoteUsageRecord
                  if defined $quoteUsageRecord && ! defined $id;
                $usageRecordProperties{Id} = $id;
                $chargeProperties{Id}      = $id;

                # Set the quote instance if pinned and uninitialized
                if (   defined $quotePinned
                    && $quotePinned eq 'True'
                    && defined $instance
                    && ! defined $quoteInstance)
                {
                    my $subRequest = new MAM::Request(
                        database    => $database,
                        object      => 'Quote',
                        action      => 'Modify',
                        actor       => $SUPER_USER,
                        assignments => [
                            new MAM::Assignment(
                                name  => 'Instance',
                                value => $instance
                            )
                        ],
                        conditions =>
                          [new MAM::Condition(name => 'Id', value => $quoteId)]
                    );
                    my $subResponse =
                      MAM::Base->modify($subRequest, $requestId);
                    if ($subResponse->getStatus() eq 'Failure')
                    {
                        return new MAM::Response()->failure(
                            $subResponse->getCode(),
                            'Unable to modify quote: '
                              . $subResponse->getMessage()
                        );
                    }
                }
            }
        }
        else
        {
            # Nullify the quote if absent or expired
            $quoteMessage .=
              "\nWarning -- Quote ($quoteId) is invalid and cannot be used";
            $warnings++;

            undef $quoteId;
            @usageRecordAssignments =
              grep { $_->getName() ne 'Quote' } @usageRecordAssignments;
        }
    }

    # Lien Id
    my $lienId = $usageRecordProperties{Lien};

    # Factor in liens if performing lien-based allocation enforcement
    if (   $accounting_mode ne 'usage-tracking'
        && $accounting_mode ne 'notional-charging'
        && $accounting_mode ne 'fast-allocation')
    {
# Construct a list of (conflicting) liens associated with the lien id,
# the instance or the usage record id and populate the other values as
# able. We are also constructing a list of debitable or removable liens
# SELECT Id,UsageRecord,Instance FROM Lien WHERE [Id=$lienId|UsageRecord=$id|Instance=$instance]
        if (defined $lienId || defined $instance || defined $id)
        {
            my %instances      = ();
            my %ids            = ();
            my @lienConditions = ();
            if (defined $lienId)
            {
                push @lienConditions,
                  new MAM::Condition(
                    name  => 'Id',
                    value => $lienId
                  );
            }
            elsif (defined $id)
            {
                push @lienConditions,
                  new MAM::Condition(
                    name  => 'UsageRecord',
                    value => $id
                  );
            }
            elsif (defined $instance)
            {
                push @lienConditions,
                  new MAM::Condition(
                    name  => 'Instance',
                    value => $instance
                  );
            }
            $results = $database->select(
                object     => 'Lien',
                selections => [
                    new MAM::Selection(name => 'Id'),
                    new MAM::Selection(name => 'UsageRecord'),
                    new MAM::Selection(name => 'Instance')
                ],
                conditions => \@lienConditions,
            );
            foreach my $row (@{$results->{data}})
            {
                my $lienId = $row->[0];
                $liens{$lienId}++;
                $ids{$row->[1]}++       if defined $row->[1];
                $instances{$row->[2]}++ if defined $row->[2];
            }
            $lienId = ${[keys %liens]}[0]
              if (! defined $lienId && scalar keys %liens == 1);
            if (! defined $id && scalar keys %ids == 1)
            {
                $id                        = ${[keys %ids]}[0];
                $usageRecordProperties{Id} = $id;
                $chargeProperties{Id}      = $id;
            }
            $instance = ${[keys %instances]}[0]
              if (! defined $instance && scalar keys %instances == 1);

# Add lienallocation amounts to the balance. The active allocations
# will be subtracted back out later and the inactive ones need to persist.
# SELECT LienAllocation.Lien,LienAllocation.Id,LienAllocation.Amount,Allocation.Fund,Allocation.StartTime,Allocation.EndTime FROM LienAllocation,Allocation WHERE LienAllocation.Id=Allocation.Id AND [LienAllocation.Lien IN %liens]
# At the moment we will not worry about whether the lien is active or not
# (AND Lien.StartTime<=$now AND Lien.EndTime>$now)
            if (scalar keys %liens)
            {
                my @objects = (
                    new MAM::Object(name => 'LienAllocation'),
                    new MAM::Object(name => 'Allocation'),
                );

                my @selections = (
                    new MAM::Selection(
                        object => 'LienAllocation',
                        name   => 'Lien'
                    ),
                    new MAM::Selection(
                        object => 'LienAllocation',
                        name   => 'Id'
                    ),
                    new MAM::Selection(
                        object => 'LienAllocation',
                        name   => 'Amount'
                    ),
                    new MAM::Selection(object => 'Allocation', name => 'Fund'),
                    new MAM::Selection(
                        object => 'Allocation',
                        name   => 'StartTime'
                    ),
                    new MAM::Selection(
                        object => 'Allocation',
                        name   => 'EndTime'
                    ),
                );
                my @lienAllocationConditions = (
                    new MAM::Condition(
                        object  => 'LienAllocation',
                        name    => 'Id',
                        subject => 'Allocation',
                        value   => 'Id',
                    )
                );
                my $firstTime = 1;
                my $counter   = 0;
                foreach my $id (keys %liens)
                {
                    $counter++;
                    my $lastTime = ($counter == scalar keys %liens) ? 1 : 0;

                    if ($firstTime && $lastTime)
                    {
                        push @lienAllocationConditions,
                          new MAM::Condition(
                            object => 'LienAllocation',
                            name   => 'Lien',
                            value  => $id,
                            conj   => 'And',
                            group  => '0'
                          );
                    }
                    elsif ($firstTime)
                    {
                        $firstTime = 0;
                        push @lienAllocationConditions,
                          new MAM::Condition(
                            object => 'LienAllocation',
                            name   => 'Lien',
                            value  => $id,
                            conj   => 'And',
                            group  => '+1'
                          );
                    }
                    elsif ($lastTime)
                    {
                        push @lienAllocationConditions,
                          new MAM::Condition(
                            object => 'LienAllocation',
                            name   => 'Lien',
                            value  => $id,
                            conj   => 'Or',
                            group  => '-1'
                          );
                    }
                    else
                    {
                        push @lienAllocationConditions,
                          new MAM::Condition(
                            object => 'LienAllocation',
                            name   => 'Lien',
                            value  => $id,
                            conj   => 'Or',
                            group  => '0'
                          );
                    }
                }
                $results = $database->select(
                    objects    => \@objects,
                    selections => \@selections,
                    conditions => \@lienAllocationConditions
                );
                foreach my $row (@{$results->{data}})
                {
                    my $lienId       = $row->[0];
                    my $allocationId = $row->[1];
                    my $lienAmount   = $row->[2] || 0;
                    my $fundId       = $row->[3];
                    my $startTime    = $row->[4];
                    my $endTime      = $row->[5];

                    if ($allocations{$allocationId})
                    {
                        $allocations{$allocationId}{balance} += $lienAmount;
                        $lienAllocations{$lienId, $allocationId} = $lienAmount;
                    }
                    elsif (
                        (   $serverConfig->{'debit.ignorelienallocations'}
                            || 'false'
                        ) !~ /^true$/i
                      )
                    {
                        # Adding to define its attributes for the transaction
                        # but excluding weight so excluded from non-lien pass
                        $allocations{$allocationId} = {
                            id        => $allocationId,
                            fund      => $fundId,
                            balance   => $lienAmount,
                            startTime => $startTime,
                            endTime   => $endTime,
                        };
                        $lienAllocations{$lienId, $allocationId} = $lienAmount;
                    }
                }
            }
        }

        if ($log->is_debug())
        {
            $log->debug(
                'Lien allocation list is: ' . join ',',
                map { '(' . (join ',', split $;, $_) . ')' }
                  keys %lienAllocations
            );
        }
    }

    # Make sure the specified usage record exists
    if (defined $id)
    {
        my $results = $database->select(
            object     => 'UsageRecord',
            selections => [new MAM::Selection(name => 'Id')],
            conditions => [
                new MAM::Condition(
                    name  => 'Id',
                    value => $id
                )
            ]
        );
        unless (@{$results->{data}})
        {
            $quoteMessage .=
              "\nWarning -- The specified usage record ($id) does not exist. Creating new usage record.";
            undef $id;
            $warnings++;
        }
    }

    # Create a new usage record if one does not already exist
    if (! defined $id)
    {
        # Create a new usage record
        my @usageRecordOptions = ();
        push @usageRecordOptions,
          new MAM::Option(name => 'Instance', value => $instance)
          if defined $instance;
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'UsageRecord',
            action   => 'Create',
            actor    => $SUPER_USER,
            options  => \@usageRecordOptions,
        );
        # Invoke the proxy so it prepares the subRequest by setting defaults
        MAM::Proxy->new(
            database  => $database,
            request   => $subRequest,
            requestId => $requestId
        );
        my $subResponse = MAM::Base->create($subRequest, $requestId, 'NoLog');
        if ($subResponse->getStatus() eq 'Success')
        {
            # Figure out the id of the usage record we just created
            $id                        = $subResponse->getDatumValue('Id');
            $usageRecordProperties{Id} = $id;
            $chargeProperties{Id}      = $id;

            # Log the transaction
            MAM::Bank->logTransaction(
                database    => $database,
                requestId   => $requestId,
                txnId       => $database->nextId('Transaction'),
                object      => 'UsageRecord',
                action      => 'Create',
                actor       => $actor,
                assignments => [
                    $subRequest->getAssignments(),
                    new MAM::Assignment(name => 'Id', value => $id),
                    new MAM::Assignment(
                        name  => 'UsageRecord',
                        value => $id
                    ),
                ],
                options => \@options,
                count   => 1,
            );

            $createdUsageRecord = 1;
        }
        else
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Unable to create usage record: ' . $subResponse->getMessage());
        }
    }

    my $charge = $request->getOptionValue('Charge');
    my $itemizedCharges = $request->getOptionValue('ItemizedCharges') || '';

    # Impute charge if notional charging or enforcing allocations
    if ($accounting_mode eq 'usage-tracking') { $charge = 0; }
    else
    {
        # Charge must be a non-negative real number
        if (defined $charge && $charge !~ /^\d*\.?\d+(e\+\d+)?$/)
        {
            return new MAM::Response()->failure('317',
                'The specified charge amount must be a non-negative real number'
            );
        }

        # Calculate charge if not already specified
        MAM::Bank->calculateCharge(
            request               => $request,
            requestId             => $requestId,
            usageRecordProperties => \%chargeProperties,
            quote                 => $quoteId,
            duration              => $chargeDuration,
            charge_ref            => \$charge,
            itemizedCharges_ref   => \$itemizedCharges,
            data_ref              => \@data,
            metadata_ref          => \%metadata,
            message_ref           => \$chargeMessage,
        );

        if ($log->is_debug())
        {
            $log->debug("The amount to be charged is: $charge");
        }
    }

    # Fail if there are no allocations to withdraw from
    if (   $accounting_mode ne 'usage-tracking'
        && $accounting_mode ne 'notional-charging'
        && ! scalar keys %allocations
        && ! $failureCode)
    {
        $failureCode = '740';
        $failureMessage =
          'Insufficient funds: There are no valid allocations against which to issue the charge';
        $failureMessage .= " (Instance $instance)" if defined $instance;
    }

# Charge the fund(s) if charge is positive
# We have marked all *-allocation accounting modes with no valid allocations as failed,
# so the following case can only be entered if valid allocations exist
    if (   $accounting_mode ne 'usage-tracking'
        && $accounting_mode ne 'notional-charging'
        && ! $failureCode)
    {
        my $remaining = $charge;

        if ($accounting_mode ne 'fast-allocation')
        {
# Subtract all liens from allocation balances (including my own)
# The lien balance associated with this usage has been added previously
# This excludes the allocations protected by liens from others
# while including the allocations protected by this job's lien
# SELECT GroupBy(LienAllocation.Id),Sum(LienAllocation.Amount) FROM Lien,LienAllocation WHERE Lien.Id=LienAllocation.Lien [AND LienAllocation.Id=<Id> ...]
            my @subSelections = (
                new MAM::Selection(
                    object => 'LienAllocation',
                    name   => 'Id',
                    op     => 'GroupBy',
                ),
                new MAM::Selection(
                    object => 'LienAllocation',
                    name   => 'Amount',
                    op     => 'Sum'
                ),
            );

            my @subConditions = (
                new MAM::Condition(
                    object  => 'Lien',
                    name    => 'Id',
                    subject => 'LienAllocation',
                    value   => 'Lien',
                ),
            );

            # Constrain to list of accepted allocation ids
            my $firstTime = 1;
            my $counter   = 0;
            foreach my $id (sort keys %allocations)
            {
                $counter++;
                my $lastTime = ($counter == scalar keys %allocations) ? 1 : 0;

                if ($firstTime && $lastTime)
                {
                    push @subConditions,
                      new MAM::Condition(
                        object => 'LienAllocation',
                        name   => 'Id',
                        value  => $id,
                        conj   => 'And',
                        group  => '0'
                      );
                }
                elsif ($firstTime)
                {
                    $firstTime = 0;
                    push @subConditions,
                      new MAM::Condition(
                        object => 'LienAllocation',
                        name   => 'Id',
                        value  => $id,
                        conj   => 'And',
                        group  => '+1'
                      );
                }
                elsif ($lastTime)
                {
                    push @subConditions,
                      new MAM::Condition(
                        object => 'LienAllocation',
                        name   => 'Id',
                        value  => $id,
                        conj   => 'Or',
                        group  => '-1'
                      );
                }
                else
                {
                    push @subConditions,
                      new MAM::Condition(
                        object => 'LienAllocation',
                        name   => 'Id',
                        value  => $id,
                        conj   => 'Or',
                        group  => '0'
                      );
                }
            }

            $results = $database->select(
                objects => [
                    new MAM::Object(name => 'Lien'),
                    new MAM::Object(name => 'LienAllocation'),
                ],
                selections => \@subSelections,
                conditions => \@subConditions
            );
            foreach my $row (@{$results->{data}})
            {
                my $allocationId = $row->[0];
                my $lienAmount   = $row->[1];

                if (exists $allocations{$allocationId})
                {
                    $allocations{$allocationId}{balance} -= $lienAmount;
                }
            }
        }

        # Sort the allocations by weight, excluding those with no weight
        my @allocations =
          sort { $allocations{$b}{weight} <=> $allocations{$a}{weight} }
          grep { defined $allocations{$_}{weight} } keys %allocations;

        if ($log->is_debug())
        {
            $log->debug('Sorted allocation list is: ' . join ',', @allocations);
        }

        # First, we debit from any reserved allocations to avoid double booking
        if ((   ($serverConfig->{'debit.ignorelienallocations'} || 'false') !~
                /^true$/i
            )
            && (scalar keys %lienAllocations)
          )
        {
            # Iterate through the lien allocation list
            foreach my $key (sort keys %lienAllocations)
            {
                my ($lienId, $allocationId) = split $;, $key;

                # Break out if we are done
                last if $remaining <= 0;

                my $fundId = $allocations{$allocationId}{fund};
                $balance = $lienAllocations{$lienId, $allocationId};
          # The amount I subtract from this allocation should be the minimum of:
          # 1. $remaining (the amount left to debit)
          # 2. $balance (the amount available within this lien)
                my $subCharge = min($remaining, $balance);
                next unless $subCharge > 0;
                $chargeChunk{$allocationId} += $subCharge;
                $charges{$fundId}           += $subCharge;
                $remaining -= $subCharge;
                $allocations{$allocationId}{balance} -= $subCharge;

                if ($log->is_trace())
                {
                    $log->trace(
                        "Charging $subCharge credits from allocation $allocationId due to lien $lienId"
                    );
                }

                # Break out if we are done
                last if $remaining <= 0;
            }
        }

       # Use simple iteration if using active or point-in-time allocation debits
       # or if charge is zero
        if ((   (   ($serverConfig->{'debit.activeallocationonly'} || 'false')
                    =~ /^true$/i
                )
                || (($serverConfig->{'debit.pointintimeonly'} || 'false') =~
                    /^true$/i)
            )
            || ($charge == 0)
          )
        {
            # Iterate through the allocation list
            foreach my $allocationId (@allocations)
            {
                # Zero charge case (should only get here if originally zero)
                if ($charge == 0)
                {
                    $chargeChunk{$allocationId} = 0;
                }
                else
                {
# The amount I subtract from this allocation should be the minimum of:
# 1. $remaining (the amount left to withdraw)
# 2. $allocations{$allocationId}{balance} (the amount available within this allocation (considering liens and credit))
                    my $fundId = $allocations{$allocationId}{fund};
                    $balance = $allocations{$allocationId}{balance};
                    my $subCharge = min($remaining, $balance);
                    next unless $subCharge > 0;
                    $chargeChunk{$allocationId} += $subCharge;
                    $charges{$fundId}           += $subCharge;
                    $remaining -= $subCharge;
                    $allocations{$allocationId}{balance} -= $subCharge;

                    if ($log->is_trace())
                    {
                        $log->trace(
                            "Charging $subCharge credits from allocation $allocationId"
                        );
                    }
                }

                # Break out if we are done
                last if $remaining <= 0;
            }
        }

        # Iterate over segment boundaries if using uniform allocation debits
        else
        {
            # Build up the charge segment edges
            my %edge = ();
            $edge{$chargeStartTime}++;
            $edge{$chargeEndTime}++;
            foreach my $allocationId (@allocations)
            {
                my $allocationStartTime =
                  $allocations{$allocationId}{startTime};
                my $allocationEndTime = $allocations{$allocationId}{endTime};
                $edge{$allocationStartTime}++
                  if $allocationStartTime > $chargeStartTime;
                $edge{$allocationEndTime}++
                  if $allocationEndTime < $chargeEndTime;
            }

            # Iterate over charge segments, building up charge chunks
            my ($allocationFound, $trailingEdge);
            my @continuityMessages = ();
            # First pass going for uniform proportional distribution
            if ($remaining > 0)
            {
                my $uniformPart = $remaining;

                foreach my $leadingEdge (reverse sort keys %edge)
                {
                    # Handle case where charge start time equals charge end time
                    if (scalar keys %edge == 1)
                    {
                        $trailingEdge = $leadingEdge;
                    }

                    if (defined $trailingEdge)
                    {
                        my $proportion =
                          ($chargeStartTime == $chargeEndTime)
                          ? sprintf("%.${currency_precision}f",
                            ($uniformPart / $time_division))
                          : sprintf(
                            "%.${currency_precision}f",
                            (   ($uniformPart * ($trailingEdge - $leadingEdge))
                                / ( $time_division *
                                      ($chargeEndTime - $chargeStartTime)
                                )
                            )
                          );
                        my $subRemaining = min($proportion, $remaining);
                        # Final segment should incorporate loss due to rounding
                        if (($leadingEdge == $chargeStartTime)
                            && ($remaining - $subRemaining <=
                                10**(-$currency_precision))
                          )
                        {
                            $subRemaining = $remaining;
                        }
                        next unless $subRemaining > 0;
                        $allocationFound = 0;
                        foreach my $allocationId (@allocations)
                        {
                            if ((   $allocations{$allocationId}{startTime} <=
                                    $leadingEdge
                                )
                                && ($allocations{$allocationId}{endTime} >=
                                    $trailingEdge)
                              )
                            {
                                $balance = $allocations{$allocationId}{balance};
                                my $subCharge = min($subRemaining, $balance);
                                next unless $subCharge > 0;
                                $chargeChunk{$allocationId} += $subCharge;
                                $subRemaining -= $subCharge;
                                $remaining    -= $subCharge;
                                $allocations{$allocationId}{balance} -=
                                  $subCharge;
                                $allocationFound = 1;

                                if ($log->is_trace())
                                {
                                    $log->trace(
                                        "Charging $subCharge credits from allocation $allocationId for interval ($leadingEdge, $trailingEdge)"
                                    );
                                }
                                last if $subRemaining <= 0;
                            }
                        }
                        if (! $allocationFound)
                        {
                            my $continuityMessage = sprintf(
                                "Allocation discontinuity of %.${currency_precision}f credits detected between %s and %s.",
                                $subRemaining, toHRT($leadingEdge),
                                toHRT($trailingEdge));
                            if ((   $serverConfig->{
                                        'debit.suppresscontinuitywarnings'}
                                    || 'false'
                                ) !~ /^true$/i
                              )
                            {
                                push @continuityMessages, $continuityMessage;
                            }
                            $log->trace($continuityMessage)
                              if $log->is_trace();
                        }
                        elsif ($subRemaining > 0)
                        {
                            my $continuityMessage = sprintf(
                                "Allocation deficiency of %.${currency_precision}f credits detected between %s and %s.",
                                $subRemaining, toHRT($leadingEdge),
                                toHRT($trailingEdge));
                            if ((   $serverConfig->{
                                        'debit.suppresscontinuitywarnings'}
                                    || 'false'
                                ) !~ /^true$/i
                              )
                            {
                                push @continuityMessages, $continuityMessage;
                            }
                            $log->trace($continuityMessage)
                              if $log->is_trace();
                        }
                    }

                    last if $remaining <= 0;
                    $trailingEdge = $leadingEdge;
                }
            }

            # If there were gaps in the first pass allocation, initiate a second
            # pass going from latest to earliest and allocating the maximum
            # possible in each allocation segment
            if (($remaining > 0)
                && ((   $serverConfig->{'debit.requirecontinuousallocations'}
                        || 'false'
                    ) !~ /^true$/i
                )
              )
            {
                undef $trailingEdge;
                foreach my $leadingEdge (reverse sort keys %edge)
                {
                    if (defined $trailingEdge)
                    {
                        foreach my $allocationId (@allocations)
                        {
                            if ((   $allocations{$allocationId}{startTime} <=
                                    $leadingEdge
                                )
                                && ($allocations{$allocationId}{endTime} >=
                                    $trailingEdge)
                              )
                            {
                                $balance = $allocations{$allocationId}{balance};
                                my $subCharge = min($remaining, $balance);
                                next unless $subCharge > 0;
                                $chargeChunk{$allocationId} += $subCharge;
                                $remaining -= $subCharge;
                                $allocations{$allocationId}{balance} -=
                                  $subCharge;
                                my $continuityMessage = sprintf(
                                    "Continuity deficit of %.${currency_precision}f credits satisfied from allocation $allocationId.",
                                    $subCharge);
                                if ((   $serverConfig->{
                                            'debit.suppresscontinuitywarnings'}
                                        || 'false'
                                    ) !~ /^true$/i
                                  )
                                {
                                    push @continuityMessages,
                                      $continuityMessage;
                                }
                                $log->trace($continuityMessage)
                                  if $log->is_trace();
                                if ($log->is_trace())
                                {
                                    $log->trace(
                                        "Charging $subCharge credits from allocation $allocationId to satisfy continuity deficit"
                                    );
                                }
                                last if $remaining <= 0;
                            }
                        }
                    }

                    last if $remaining <= 0;
                    $trailingEdge = $leadingEdge;
                }
            }

            if (@continuityMessages)
            {
                $warningMessage .= "\nWarning -- " . join "\n",
                  @continuityMessages;
                $warnings++;
            }
        }

        # If charges remaining
        if ($remaining > 0)
        {
            # Warn if preventing negative allocations
            if (($serverConfig->{'debit.preventnegativeallocations'} || 'false')
                =~ /^true$/i)
            {
                $warningMessage .=
                  "\nWarning -- Insufficient funds exist to maintain a positive balance. A charge of $remaining credits has been prevented in order to avoid inducing a negative allocation balance.";
                $warnings++;
            }

            # Otherwise allow first allocation to go negative
            else
            {
                my $allocationId = $allocations[0];
                my $fundId       = $allocations{$allocationId}{fund};
                my $subCharge    = $remaining;
                $chargeChunk{$allocationId} += $subCharge;
                $remaining -= $subCharge;
                $charges{$fundId} += $subCharge;
                $allocations{$allocationId}{balance} -= $subCharge;
                $warningMessage .=
                  "\nWarning -- Insufficient funds exist to maintain a positive balance. However, an allocation will still be charged in order to account for the chargeable usage.";
                $warnings++;
            }
        }

# use Data::Dumper;
# print Data::Dumper->Dump([\%chargeChunk], ['chargeChunk']), "\n";
# print Data::Dumper->Dump([\@continuityMessages], ['continuityMessages']), "\n";

        # Iterate through the charge chunks
        foreach my $allocationId (sort keys %chargeChunk)
        {
            my $fundId    = $allocations{$allocationId}{fund};
            my $subCharge = $chargeChunk{$allocationId};

            if ($subCharge > 0)
            {
                # Debit the allocation
                my $subRequest = new MAM::Request(
                    database    => $database,
                    object      => 'Allocation',
                    action      => 'Modify',
                    actor       => $SUPER_USER,
                    assignments => [
                        new MAM::Assignment(
                            name  => 'Amount',
                            value => $subCharge,
                            op    => 'Dec'
                        )
                    ],
                    conditions => [
                        new MAM::Condition(
                            name  => 'Id',
                            value => $allocationId
                        )
                    ],
                );
                my $subResponse =
                  MAM::Base->modify($subRequest, $requestId, 'NoLog');
                if ($subResponse->getStatus() eq 'Failure')
                {
                    return new MAM::Response()->failure($subResponse->getCode(),
                        'Unable to debit allocation: '
                          . $subResponse->getMessage());
                }

                # Log the transaction
                MAM::Bank->logTransaction(
                    database    => $database,
                    requestId   => $requestId,
                    txnId       => $database->nextId('Transaction'),
                    object      => 'Allocation',
                    action      => 'Modify',
                    actor       => $actor,
                    assignments => [
                        new MAM::Assignment(
                            name  => 'Amount',
                            value => $subCharge,
                        )
                    ],
                    conditions => [
                        new MAM::Condition(
                            name  => 'Id',
                            value => $allocationId
                        )
                    ],
                    count      => 1,
                    fund       => $fundId,
                    allocation => $allocationId
                );
            }

            # Obtain remaining allocation amount
            my $results = $database->select(
                object     => 'Allocation',
                selections => [new MAM::Selection(name => 'Amount')],
                conditions => [
                    new MAM::Condition(
                        name  => 'Id',
                        value => $allocationId
                    )
                ]
            );
            my $remainingAllocation = ${$results->{data}}[0]->[0];

            my $delta;
            if (   $allocations{$allocationId}{startTime} <= $now
                && $allocations{$allocationId}{endTime} > $now)
            {
                $delta   = 0 - $subCharge;
                $balance = $remainingAllocation;
            }
            else
            {
                $delta   = 0;
                $balance = 0;
            }

            my @chargeOptions = (
                @options,
                new MAM::Option(
                    name  => 'Amount',
                    value => $subCharge
                ),
                new MAM::Option(name => 'Id',          value => $id),
                new MAM::Option(name => 'UsageRecord', value => $id),
                new MAM::Option(
                    name  => 'ItemizedCharges',
                    value => $itemizedCharges
                )
            );
            push @chargeOptions,
              new MAM::Option(name => 'Child', value => $instance)
              if defined $instance;
            MAM::Bank->logTransaction(
                database    => $database,
                requestId   => $requestId,
                txnId       => $database->nextId('Transaction'),
                object      => 'UsageRecord',
                action      => 'Charge',
                actor       => $actor,
                options     => \@chargeOptions,
                assignments => \@usageRecordAssignments,
                count       => 1,
                fund        => $fundId,
                delta       => $delta,
                balance     => $balance,
                remaining   => $remainingAllocation,
                allocation  => $allocationId
            );
        }
    }

    # Log the transaction for the case where accounting mode is usage-tracking
    # or notional-charging or a failure has occurred
    else
    {
        my @chargeOptions = (
            @options,
            new MAM::Option(name => 'Id',          value => $id),
            new MAM::Option(name => 'UsageRecord', value => $id),
        );
        push @chargeOptions,
          new MAM::Option(name => 'Child', value => $instance)
          if defined $instance;
        push @chargeOptions,
          new MAM::Option(name => 'Amount', value => $charge),
          new MAM::Option(
            name  => 'ItemizedCharges',
            value => $itemizedCharges
          ) if ($accounting_mode ne 'usage-tracking' && ! $failureCode);
        push @chargeOptions,
          new MAM::Option(name => 'Description', value => $failureMessage)
          if $failureCode;
        MAM::Bank->logTransaction(
            database    => $database,
            requestId   => $requestId,
            txnId       => $database->nextId('Transaction'),
            object      => 'UsageRecord',
            action      => 'Charge',
            options     => \@chargeOptions,
            assignments => \@usageRecordAssignments,
            actor       => $actor,
            count       => 1,
        );
    }

    # If enforcing allocations via liens, adjust or remove the liens
    if (   $accounting_mode ne 'usage-tracking'
        && $accounting_mode ne 'notional-charging'
        && $accounting_mode ne 'fast-allocation')
    {
        # If incremental charging, decrement the liens
        if ($incremental eq 'True' && scalar keys %lienAllocations)
        {
           # Debit the liens -- Do not delete the liens
           # We will only debit active liens, since there may be stale ones left
           # over from another usage record.
            my $remaining = $charge;

            if ($charge > 0)
            {
                # Iterate through the allocation list
                foreach my $key (sort keys %lienAllocations)
                {
                    my ($lienId, $allocationId) = split $;, $key;

                    # Break out if we are done
                    last if $remaining <= 0;

     # The amount I subtract from this lien allocation should be the minimum of:
     # 1. $remaining (the amount left to debit)
     # 2. $liens{$lienId, $allocationId} (the amount available within this lien)
                    my $debitAmount =
                      min($remaining, $lienAllocations{$lienId, $allocationId});
                    next unless $debitAmount > 0;

                    # Debit the lien allocation
                    my $subRequest = new MAM::Request(
                        database    => $database,
                        object      => 'LienAllocation',
                        action      => 'Modify',
                        actor       => $SUPER_USER,
                        assignments => [
                            new MAM::Assignment(
                                name  => 'Amount',
                                value => $debitAmount,
                                op    => 'Dec'
                            )
                        ],
                        conditions => [
                            new MAM::Condition(
                                name  => 'Lien',
                                value => $lienId
                            ),
                            new MAM::Condition(
                                name  => 'Id',
                                value => $allocationId
                            )
                        ]
                    );
                    my $subResponse =
                      MAM::Base->modify($subRequest, $requestId);
                    if ($subResponse->getStatus() eq 'Failure')
                    {
                        return new MAM::Response()->failure(
                            $subResponse->getCode(),
                            'Unable to debit lien: '
                              . $subResponse->getMessage()
                        );
                    }

                    $remaining -= $debitAmount;
                }
            }

            # Log it if we are charging more than the lien has
            if ($remaining > 0)
            {
                if ($log->is_warn())
                {
                    $log->warn(
                        "The charge is larger than the lien by $remaining credits"
                    );
                }
            }
        }

        # Otherwise remove the liens
        elsif (scalar keys %liens)
        {
            # Remove the liens -- this should also remove LienAllocations
            # DELETE FROM Lien WHERE Id=<id>...
            # Constrain to list of accepted allocation ids
            my $firstTime        = 1;
            my $counter          = 0;
            my @deleteConditions = ();
            foreach my $id (keys %liens)
            {
                $counter++;
                my $lastTime = ($counter == scalar keys %liens) ? 1 : 0;

                if ($firstTime && $lastTime)
                {
                    push @deleteConditions,
                      new MAM::Condition(
                        name  => 'Id',
                        value => $id,
                        conj  => 'And',
                        group => '0'
                      );
                }
                elsif ($firstTime)
                {
                    $firstTime = 0;
                    push @deleteConditions,
                      new MAM::Condition(
                        name  => 'Id',
                        value => $id,
                        conj  => 'And',
                        group => '+1'
                      );
                }
                elsif ($lastTime)
                {
                    push @deleteConditions,
                      new MAM::Condition(
                        name  => 'Id',
                        value => $id,
                        conj  => 'Or',
                        group => '-1'
                      );
                }
                else
                {
                    push @deleteConditions,
                      new MAM::Condition(
                        name  => 'Id',
                        value => $id,
                        conj  => 'Or',
                        group => '0'
                      );
                }
            }
            my $subRequest = new MAM::Request(
                database   => $database,
                object     => 'Lien',
                action     => 'Delete',
                actor      => $SUPER_USER,
                conditions => \@deleteConditions,
            );
            my $subResponse = MAM::Base->delete($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Failure')
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    'Unable to delete liens: ' . $subResponse->getMessage());
            }
            my $count = $subResponse->getCount();
            if ($count)
            {
                $reserveMessage .=
                    "\n$count lien"
                  . ($count == 1 ? ' was' : 's were')
                  . ' removed';
            }
        }
    }

    # Modify the usage record with charge and other attributes
    my @usageRecordOptions = @options;
    push @usageRecordOptions,
      new MAM::Option(name => 'Instance', value => $instance)
      if defined $instance;
    push @usageRecordAssignments,
      new MAM::Assignment(name => 'Charge', value => $charge, op => 'Inc')
      if ($accounting_mode ne 'usage-tracking' && ! $failureCode);
    # By putting @usageRecordAssignments last, we allow the usage record
    # property Duration to override the increment of the duration option.
    # Duration should not be in data properties unless you want this effect
    my $subRequest = new MAM::Request(
        database    => $database,
        object      => 'UsageRecord',
        action      => 'Modify',
        actor       => $SUPER_USER,
        conditions  => [new MAM::Condition(name => 'Id', value => $id)],
        options     => \@usageRecordOptions,
        assignments => [
            new MAM::Assignment(
                name  => 'Duration',
                value => $chargeDuration,
                op    => 'Inc'
            ),
            @usageRecordAssignments,
        ],
    );
    my $subResponse = MAM::Base->modify($subRequest, $requestId);

    if (   $subResponse->getStatus() eq 'Failure'
        || $subResponse->getCount() < 1)
    {
        return new MAM::Response()->failure($subResponse->getCode(),
            'Unable to modify usage record: ' . $subResponse->getMessage());
    }

    # Append balance info to data if balance feedback requested
    if (   $balanceFeedback
        && $accounting_mode ne 'usage-tracking'
        && $accounting_mode ne 'notional-charging'
        && scalar @data == 1
        && ! $failureCode)
    {
        my $datumElement   = $data[0]->{_element};
        my $lastChild      = $datumElement->lastChild;
        my $balanceElement = new XML::LibXML::Element('Balance');
        $balanceElement->appendText($balance);
        $datumElement->insertAfter($balanceElement, $lastChild);
        $metadata{'Balance'} = 'Currency';
    }

    $charge = sprintf("%.${currency_precision}f", $charge / $time_division);

    if ($accounting_mode eq 'usage-tracking')
    {
        $message = "Successfully recorded usage information";
    }
    else
    {
        if ($failureCode)
        {
            $message = "Failed charging $charge credits";
        }
        else
        {
            $message = "Successfully charged $charge credits";
        }
    }
    $message .= " for instance $instance" if defined $instance;
    $message .= " and created usage record $id"
      if $createdUsageRecord && $id;
    $message .=
      $warningMessage . $quoteMessage . $reserveMessage . $chargeMessage;
    if ($failureCode)
    {
        # Commit so the usage record, transaction, and liens will be updated
        $database->getHandle()->commit();

        $response = new MAM::Response()
          ->failure($failureCode, $message . "\n" . $failureMessage);
    }
    elsif ($warnings)
    {
        $response =
          new MAM::Response()->warning($charge, \@data, \%metadata, $message);
    }
    else
    {
        $response =
          new MAM::Response()->success($charge, \@data, \%metadata, $message);
    }
    return $response;
}

# ----------------------------------------------------------------------------
# $response = create($request, $requestId);
# ----------------------------------------------------------------------------

# Create (Account, Accountuser, Allocation, Fund, FundFund, Payment, UsageRecord)
sub create
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object   = $request->getObject();
    my $database = $request->getDatabase();
    my $actor    = $request->getActor();
    my @options  = $request->getOptions();
    my $now      = $requestTime;
    my $infinity = 2147483647;

    # Account
    if ($object eq 'Account')
    {
        # Invoke base Account Create
        my $baseResponse = MAM::Base->create($request, $requestId);
        my $baseMessage = $baseResponse->getMessage();

        if ($baseResponse->getStatus() eq 'Success')
        {
            # Determine if we should auto-generate a fund
            my $createFund       = 0;
            my $createFundOption = $request->getOptionValue('CreateFund');
            if (MAM::Cache->getObjectProperty('Fund', 'AutoGen') eq 'True')
            {
                if (! defined $createFundOption
                    || $createFundOption !~ /^false$/i)
                {
                    # Autogen and CreateFund is not False
                    $createFund = 1;
                }
            }
            else
            {
                if (defined $createFundOption
                    && $createFundOption =~ /^true$/i)
                {
                    # No autogen but CreateFund is True
                    $createFund = 1;
                }
            }

            # Auto-generate a fund if we resolved to do so
            if ($createFund)
            {
                my $account = $request->getAssignmentValue('Name');

                # Create the fund
                my $subRequest = new MAM::Request(
                    database    => $database,
                    object      => 'Fund',
                    action      => 'Create',
                    actor       => $SUPER_USER,
                    assignments => [
                        new MAM::Assignment(
                            name  => 'Name',
                            value => "Account=$account"
                        ),
                        new MAM::Assignment(
                            name  => 'Priority',
                            value => '0'
                        ),
                        new MAM::Assignment(
                            name  => 'Description',
                            value => 'Auto-generated'
                        )
                    ]
                );
                my $subResponse = MAM::Bank->create($subRequest, $requestId);

                if ($subResponse->getStatus() eq 'Success')
                {
                    # Figure out the id of the fund we just created
                    my $fund = $subResponse->getDatumValue('Id');
                    $baseMessage .= "\n" . $subResponse->getMessage();

                    # Add Account Constraint
                    $subRequest = new MAM::Request(
                        database    => $database,
                        object      => 'Constraint',
                        action      => 'Create',
                        actor       => $SUPER_USER,
                        assignments => [
                            new MAM::Assignment(
                                name  => 'Fund',
                                value => $fund,
                            ),
                            new MAM::Assignment(
                                name  => 'Name',
                                value => 'Account',
                            ),
                            new MAM::Assignment(
                                name  => 'Value',
                                value => $account,
                            ),
                        ]
                    );
                    $subResponse = MAM::Base->create($subRequest, $requestId);
                    if ($subResponse->getStatus() eq 'Failure')
                    {
                        return new MAM::Response()->failure(
                            $subResponse->getCode(),
                            'Unable to create account constraint: '
                              . $subResponse->getMessage()
                        );
                    }
                }
                else
                {
                    return new MAM::Response()->failure($subResponse->getCode(),
                        'Unable to create fund: ' . $subResponse->getMessage());
                }
            }
        }

        return $baseResponse->setMessage($baseMessage);
    }

    # AccountUser
    if ($object eq 'AccountUser')
    {
        # If the user.firstaccountdefault parameter is true and this is the
        # user's first account, set this as the user's default account
        # unless overridden by the DefaultParent option.
        if (($serverConfig->{'user.firstaccountdefault'} || 'true') =~ /^true$/i
            && (! defined $request->getOptionValue('DefaultParent')))
        {
            my %assignments =
              map { ${_}->getName(), $_ } $request->getAssignments();
            my $user = $assignments{'Name'}->getValue();

            # Select Account FROM AccountUser WHERE Name=$user
            my $results = $database->select(
                object     => 'AccountUser',
                selections => [new MAM::Selection(name => 'Account')],
                conditions =>
                  [new MAM::Condition(name => 'Name', value => $user),]
            );
            unless (@{$results->{data}})
            {
                $request->setOption('DefaultParent', 'True');
            }
        }

        # Invoke base AccountUser Create
        return MAM::Base->create($request, $requestId);
    }

    # Allocation
    elsif ($object eq 'Allocation')
    {
        # Allocations should only be created via Fund Deposit
        return new MAM::Response()->failure('740',
            'Invoking Allocation Create directly is not allowed because it bypasses accounting. Use Fund Deposit instead.'
        );
    }

    # Fund
    elsif ($object eq 'Fund')
    {
        # If no fund name is passed in, concoct a default name
        unless (defined $request->getAssignmentValue('Name'))
        {
            my @constraints = ();
            foreach my $option (@options)
            {
                if ($option->getName() eq 'Constraint')
                {
                    my $constraint = $option->getValue();
                    if ($constraint =~ /^([^\s=!]+)=(!?[^\s=](?:[^=]*[^\s=]))$/)
                    {
                        my ($constraintName, $constraintValue) = ($1, $2);
                        push @constraints, "$constraintName=$constraintValue";
                    }
                }
            }
            $request->addAssignment('Name', join ',', @constraints)
              if @constraints;
        }

        # Invoke base Fund Create
        my $baseResponse = MAM::Base->create($request, $requestId);
        my $baseMessage = $baseResponse->getMessage();

        if ($baseResponse->getStatus() eq 'Success')
        {
            # Figure out the id of the fund we just created
            my $fund = $baseResponse->getDatumValue('Id');
            $baseMessage .= " with id $fund";
            my @messages = ();
            my $warnings = 0;

            # Create constraints
            my $memberCount = 0;
            foreach my $option (@options)
            {
                next unless $option->getName() eq 'Constraint';
                my ($constraintName, $constraintValue);
                my $constraint = $option->getValue();
                if ($constraint =~ /^([^\s=!]+)=(!?[^\s=](?:[^=]*[^\s=]))$/)
                {
                    ($constraintName, $constraintValue) = ($1, $2);
                }
                unless (defined $constraintName && defined $constraintValue)
                {
                    return new MAM::Response()
                      ->failure('317', "Invalid constraint ($constraint)");
                }

                my @constraintAssignments = (
                    new MAM::Assignment(
                        name  => 'Fund',
                        value => $fund,
                    ),
                    new MAM::Assignment(
                        name  => 'Name',
                        value => $constraintName,
                    ),
                    new MAM::Assignment(
                        name  => 'Value',
                        value => $constraintValue,
                    ),
                );
                my $subRequest = new MAM::Request(
                    database    => $database,
                    actor       => $SUPER_USER,
                    object      => 'Constraint',
                    action      => 'Create',
                    assignments => \@constraintAssignments,
                );
                # Invoke the proxy so it prepares the subRequest by setting
                # defaults
                MAM::Proxy->new(
                    database  => $database,
                    request   => $subRequest,
                    requestId => $requestId
                );
                my $subResponse =
                  MAM::Base->create($subRequest, $requestId, 'NoLog');
                if ($subResponse->getStatus() eq 'Failure')
                {
                    return new MAM::Response()->failure($subResponse->getCode(),
                        'Unable to create constraint: '
                          . $subResponse->getMessage());
                }
                my $count = $subResponse->getCount();
                $memberCount += $count if $count > 0;

                # Check to see if the constraint should be auto-generated
                # If it is an object, see if the instance exists
                # else see if the object supports auto-generation
                if (MAM::Cache->objectExists($constraintName))
                {
                    # Figure out its primary keys
                    my @primaryKeys =
                      MAM::Cache->listAttributesWithPropertyValue(
                        $constraintName, 'PrimaryKey', 'True');

                    # We can only do this if it has a single primary key
                    if (@primaryKeys == 1)
                    {
                        my $constraintKey = $primaryKeys[0];

# Look up the constraint instance
# SELECT $constraintKey FROM $constraintName WHERE $constraintName=$constraintValue
                        my $subResults = $database->select(
                            object => $constraintName,
                            selections =>
                              [new MAM::Selection(name => $constraintKey)],
                            conditions => [
                                new MAM::Condition(
                                    name  => $constraintKey,
                                    value => $constraintValue
                                )
                            ],
                        );
                        unless (@{$subResults->{data}})
                        {
                         # Check to see if the instance should be auto-generated
                            if (MAM::Cache->getObjectProperty($constraintName,
                                    'AutoGen') eq 'True'
                              )
                            {
                                # Create the instance
                                my @autoGenAssignments = (
                                    new MAM::Assignment(
                                        name  => $constraintKey,
                                        value => $constraintValue
                                    )
                                );
                                push @autoGenAssignments,
                                  new MAM::Condition(
                                    name  => 'Active',
                                    value => 'True'
                                  )
                                  if MAM::Cache->attributeExists(
                                    $constraintName, 'Active');
                                push @autoGenAssignments,
                                  new MAM::Assignment(
                                    name  => 'Description',
                                    value => 'Auto-Generated'
                                  )
                                  if MAM::Cache->attributeExists(
                                    $constraintName, 'Description');
                                my $subRequest = new MAM::Request(
                                    database    => $database,
                                    object      => $constraintName,
                                    action      => 'Create',
                                    actor       => $SUPER_USER,
                                    assignments => \@autoGenAssignments,
                                );
            # Invoke the proxy so it prepares the subRequest by setting defaults
                                MAM::Proxy->new(
                                    database  => $database,
                                    request   => $subRequest,
                                    requestId => $requestId
                                );
                                my $subResponse =
                                  MAM::Base->create($subRequest, $requestId);
                                if ($subResponse->getStatus() eq 'Failure')
                                {
                                    push @messages,
                                      "Warning -- Unable to auto-generate $constraintName $constraintValue: "
                                      . $subResponse->getMessage();
                                    $warnings++;
                                }
                                else
                                {
                                    push @messages, $subResponse->getMessage();
                                }
                            }
                        }
                    }
                    else
                    {
                        # Push a warning message
                        push @messages,
                          "Warning -- Unable to auto-generate $constraintName $constraintValue: Multiple primary keys exist for the object";
                        $warnings++;
                    }
                }
            }

            $baseMessage .=
              " and $memberCount constraint" . ($memberCount == 1 ? '' : 's')
              if $memberCount;

            # Auto-generate allocations if configured to do so
            if (MAM::Cache->getObjectProperty('Allocation', 'AutoGen') eq
                'True')
            {
                # Make a blank deposit using default amount and credit limit
                my $defaultAmount =
                  MAM::Cache->getAttributeProperty('Allocation', 'Amount',
                    'DefaultValue');
                my $defaultCreditLimit =
                  MAM::Cache->getAttributeProperty('Allocation', 'CreditLimit',
                    'DefaultValue');
                my @depositOptions = (
                    new MAM::Option(
                        name  => 'Id',
                        value => $fund
                    )
                );
                push @depositOptions,
                  new MAM::Option(name => 'Amount', value => $defaultAmount)
                  if defined $defaultAmount;
                push @depositOptions,
                  new MAM::Option(
                    name  => 'CreditLimit',
                    value => $defaultCreditLimit
                  ) if defined $defaultCreditLimit;

                my $subRequest = new MAM::Request(
                    database => $database,
                    object   => 'Fund',
                    action   => 'Deposit',
                    actor    => $SUPER_USER,
                    options  => \@depositOptions,
                );
                my $subResponse = MAM::Bank->deposit($subRequest, $requestId);
                if ($subResponse->getStatus() eq 'Success')
                {
                    $baseMessage .= ' and auto-generated 1 allocation';
                }
                else
                {
                    return new MAM::Response()->failure($subResponse->getCode(),
                        'Unable to create allocation: '
                          . $subResponse->getMessage());
                }
            }

            $baseResponse->setMessage(join("\n", $baseMessage, @messages));
            if ($warnings)
            {
                $baseResponse->setStatus('Warning');
                $baseResponse->setCode('100');
            }
            return $baseResponse;
        }
        else
        {
            return new MAM::Response()->failure($baseResponse->getCode(),
                'Unable to create fund: ' . $baseMessage);
        }
    }

    # FundFund
    elsif ($object eq 'FundFund')
    {
        my $depositShare = $request->getAssignmentValue('DepositShare');

        if (defined $depositShare)
        {
            # Make sure deposit shares do not exceed 100 for a given fund
            my $fund              = $request->getAssignmentValue('Fund');
            my $depositShareTotal = $depositShare;

            my $results = $database->select(
                object     => 'FundFund',
                selections => [
                    new MAM::Selection(
                        name => 'DepositShare',
                        op   => 'Sum'
                    )
                ],
                conditions => [
                    new MAM::Condition(
                        name  => 'Fund',
                        value => $fund
                    )
                ]
            );
            my $depositShareSum = ${$results->{data}}[0]->[0];
            if (defined $depositShareSum)
            {
                $depositShareTotal += $depositShareSum;
            }
            if ($depositShareTotal > 100)
            {
                return new MAM::Response()->failure('740',
                    "Total deposit shares for a fund cannot exceed 100% ($depositShareTotal)"
                );
            }
        }

        # Invoke base FundFund Create
        return MAM::Base->create($request, $requestId);
    }

    # Payment
    elsif ($object eq 'Payment')
    {
        my $beginTime = $request->getAssignmentValue('BeginTime');
        my $duration  = $request->getAssignmentValue('Duration');
        my $endTime   = $request->getAssignmentValue('EndTime');
        my $notify    = $request->getAssignmentValue('Notify');

        # Derive duration from begin time and end time if applicable
        if (defined $beginTime && defined $endTime && ! defined $duration)
        {
            $duration = $endTime - $beginTime;
            $request->addAssignment('Duration', $duration);
        }

        # Derive end time from begin time and duration if applicable
        if (defined $beginTime && defined $duration && ! defined $endTime)
        {
            $endTime = $beginTime + $duration;
            $request->addAssignment('EndTime', $endTime);
        }

        # Do not derive begin time from end time and duration

        # Invoke base Payment Create
        my $baseResponse = MAM::Base->create($request, $requestId);

        if ($baseResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($baseResponse->getCode(),
                'Unable to create payment: ' . $baseResponse->getMessage());
        }
        my $message      = $baseResponse->getMessage();
        my @baseData     = $baseResponse->getData();
        my %baseMetadata = $baseResponse->getMetadata();

        # Figure out the id, charge period and reserve mode of the payment
        # we just created
        my $id             = $baseResponse->getDatumValue('Id');
        my $chargePeriod   = $baseResponse->getDatumValue('ChargePeriod');
        my $reservePattern = $baseResponse->getDatumValue('ReservePattern');

        # Disallow unfulfillable property combinations
        return new MAM::Response()->failure('314',
            'The payment charge period must be specified when using the Periodic reservation model'
        ) if ($reservePattern eq 'Periodic' && $chargePeriod eq '');
        return new MAM::Response()->failure('314',
            'The payment end time must be specified when using the Combined reservation model'
        ) if ($reservePattern eq 'Combined' && ! defined $endTime);

        # Create PaymentComponents
        my $componentCount       = 0;
        my $usageRecordCount     = 0;
        my $updateComponentCount = 0;
        my $data                 = $request->getDataElement();
        my @components           = $data->childNodes();
        if (@components)
        {
            my $doc = XML::LibXML::Document->new();
            $doc->setDocumentElement($data);

            # Build up a list of usage record attributes and set defaults
            my %usageRecordAttributes = ();
            foreach my $attributeName (
                MAM::Cache->listAttributesWithPropertyValue(
                    'UsageRecord', 'Hidden', 'False'
                )
              )
            {
                $usageRecordAttributes{$attributeName} = 1;
            }

            # Parse out usage record components
            # Iterate over usage records (components)
            # Sort components in order of Begin, Update, End
            foreach my $component (
                sort {
                    if ($a->nodeName eq $b->nodeName) {0}
                    elsif ($a->nodeName eq 'End' || $b->nodeName eq 'Begin')
                    {
                        1;
                    }
                    elsif ($a->nodeName eq 'Begin' || $b->nodeName eq 'End')
                    {
                        -1;
                    }
                } @components
              )
            {
                # The type will be checked in the PaymentComponent creation
                my $type = $component->nodeName;

                # Build up a list of usage record properties
                my %componentProperties = ();
                foreach my $col ($component->childNodes())
                {
                    if ($col->hasChildNodes())    # Simple properties only
                    {
                        my $propertyName  = $col->nodeName();
                        my $propertyValue = $col->textContent();
                        $componentProperties{$propertyName} = $propertyValue;
                    }
                }
                my $usageRecordId = $componentProperties{Id};
                delete $componentProperties{Id} if $usageRecordId;

                # Build up usage record assignments
                my @usageRecordAssignments = (
                    new MAM::Assignment(name => 'Payment', value => $id),
                    new MAM::Assignment(name => 'Stage',   value => 'Create')
                );
                foreach my $propertyName (sort keys %componentProperties)
                {
                    if ($usageRecordAttributes{$propertyName})
                    {
                        push @usageRecordAssignments,
                          new MAM::Assignment(
                            name  => $propertyName,
                            value => $componentProperties{$propertyName}
                          );
                    }
                }

                # Create or Modify Usage Record if needed
                if (keys %componentProperties)
                {
                    my $assignmentAction;
                    if   ($usageRecordId) { $assignmentAction = 'Modify'; }
                    else                  { $assignmentAction = 'Create'; }
                    my $subRequest = new MAM::Request(
                        database    => $database,
                        object      => 'UsageRecord',
                        action      => $assignmentAction,
                        actor       => $SUPER_USER,
                        assignments => \@usageRecordAssignments,
                    );
                    $subRequest->addCondition(
                        new MAM::Condition(
                            name  => 'Id',
                            value => $usageRecordId
                        )
                    ) if $usageRecordId;
                    # Invoke the proxy so it prepares the subRequest by setting
                    # defaults
                    MAM::Proxy->new(
                        database  => $database,
                        request   => $subRequest,
                        requestId => $requestId
                    );
                    my $subResponse;
                    if ($usageRecordId)
                    {
                        $subResponse =
                          MAM::Base->modify($subRequest, $requestId);
                    }
                    else
                    {
                        $subResponse =
                          MAM::Base->create($subRequest, $requestId, 'NoLog');
                    }

                    if ($subResponse->getStatus() eq 'Success')
                    {
                        # Figure out the id of the usage record we just created
                        unless ($usageRecordId)
                        {
                            $usageRecordId = $subResponse->getDatumValue('Id');
                            $usageRecordCount++;

                            # Log the transaction
                            MAM::Bank->logTransaction(
                                database    => $database,
                                requestId   => $requestId,
                                txnId       => $database->nextId('Transaction'),
                                object      => 'UsageRecord',
                                action      => $assignmentAction,
                                actor       => $actor,
                                assignments => [
                                    $subRequest->getAssignments(),
                                    new MAM::Assignment(
                                        name  => 'Id',
                                        value => $usageRecordId
                                    ),
                                    new MAM::Assignment(
                                        name  => 'UsageRecord',
                                        value => $usageRecordId
                                    ),
                                ],
                                count => 1,
                            );
                        }
                    }
                    else
                    {
                        return new MAM::Response()->failure(
                            $subResponse->getCode(),
                            'Unable to create usage record: '
                              . $subResponse->getMessage()
                        );
                    }
                }

                # Create PaymentComponent
                if ($usageRecordId)
                {
                    my $subRequest = new MAM::Request(
                        database    => $database,
                        object      => 'PaymentComponent',
                        action      => 'Create',
                        actor       => $SUPER_USER,
                        assignments => [
                            new MAM::Assignment(
                                name  => 'Payment',
                                value => $id
                            ),
                            new MAM::Assignment(
                                name  => 'Type',
                                value => $type
                            ),
                            new MAM::Assignment(
                                name  => 'Id',
                                value => $usageRecordId
                            ),
                        ],
                    );
                    # Invoke the proxy so it will check for valid type values
                    MAM::Proxy->new(
                        database  => $database,
                        request   => $subRequest,
                        requestId => $requestId
                    );
                    my $subResponse =
                      MAM::Base->create($subRequest, $requestId);
                    if ($subResponse->getStatus() eq 'Failure')
                    {
                        return new MAM::Response()->failure(
                            $subResponse->getCode(),
                            'Unable to create payment component: '
                              . $subResponse->getMessage()
                        );
                    }
                    $componentCount++;
                    $updateComponentCount++ if $type eq 'Update';
                }
            }
        }

        # Create Events
        my $eventCount = 0;

        # Create Begin Event
        my $beginEvent;
        if (defined $beginTime)
        {
            my @eventAssignments = (
                new MAM::Assignment(
                    name  => 'FireTime',
                    value => $beginTime,
                ),
                new MAM::Assignment(
                    name  => 'FireCommand',
                    value => "Payment Begin Id:=$id"
                ),
                new MAM::Assignment(
                    name  => 'Description',
                    value => 'Begin Payment'
                )
            );
            push @eventAssignments,
              new MAM::Assignment(name => 'Notify', value => $notify)
              if defined $notify;
            my $subRequest = new MAM::Request(
                database    => $database,
                object      => 'Event',
                action      => 'Create',
                actor       => $SUPER_USER,
                assignments => \@eventAssignments,
            );
            # Invoke the proxy so it prepares the subRequest by setting defaults
            MAM::Proxy->new(
                database  => $database,
                request   => $subRequest,
                requestId => $requestId
            );
            my $subResponse = MAM::Base->create($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Success')
            {
                # Figure out the id of the event we just created
                $beginEvent = $subResponse->getDatumValue('Id');
                $eventCount++;
            }
            else
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    'Unable to create payment begin event: '
                      . $subResponse->getMessage());
            }
        }

        # Create Update Event
        my $updateEvent;
        if ($updateComponentCount)
        {
            my @eventAssignments = (
                new MAM::Assignment(
                    name  => 'FireCommand',
                    value => "Payment Update Id:=$id"
                ),
                new MAM::Assignment(
                    name  => 'Description',
                    value => 'Update Payment'
                )
            );
            push @eventAssignments,
              new MAM::Assignment(name => 'Notify', value => $notify)
              if defined $notify;
            my $subRequest = new MAM::Request(
                database    => $database,
                object      => 'Event',
                action      => 'Create',
                actor       => $SUPER_USER,
                assignments => \@eventAssignments,
            );
            # Invoke the proxy so it prepares the subRequest by setting defaults
            MAM::Proxy->new(
                database  => $database,
                request   => $subRequest,
                requestId => $requestId
            );
            my $subResponse = MAM::Base->create($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Success')
            {
                # Figure out the id of the event we just created
                $updateEvent = $subResponse->getDatumValue('Id');
                $eventCount++;
            }
            else
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    'Unable to create payment update event: '
                      . $subResponse->getMessage());
            }
        }

        # Create End Event
        my $endEvent;
        if (defined $endTime)
        {
            my $endCommand   = "Payment End Id:=$id";
            my $deleteOption = $request->getOptionValue('Delete');
            $endCommand .= ' Delete:=True'
              if (defined $deleteOption && $deleteOption =~ /^true$/i);

            my @eventAssignments = (
                new MAM::Assignment(
                    name  => 'FireTime',
                    value => $endTime,
                ),
                new MAM::Assignment(
                    name  => 'FireCommand',
                    value => $endCommand
                ),
                new MAM::Assignment(
                    name  => 'Description',
                    value => 'End Payment'
                )
            );
            push @eventAssignments,
              new MAM::Assignment(name => 'Notify', value => $notify)
              if defined $notify;
            my $subRequest = new MAM::Request(
                database    => $database,
                object      => 'Event',
                action      => 'Create',
                actor       => $SUPER_USER,
                assignments => \@eventAssignments,
            );
            # Invoke the proxy so it prepares the subRequest by setting defaults
            MAM::Proxy->new(
                database  => $database,
                request   => $subRequest,
                requestId => $requestId
            );
            my $subResponse = MAM::Base->create($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Success')
            {
                # Figure out the id of the event we just created
                $endEvent = $subResponse->getDatumValue('Id');
                $eventCount++;
            }
            else
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    'Unable to create payment end event: '
                      . $subResponse->getMessage());
            }
        }

        # Update payment with begin, end and update events
        my @paymentAssignments = ();
        push @paymentAssignments,
          new MAM::Assignment(
            name  => 'BeginEvent',
            value => $beginEvent
          ) if $beginEvent;
        push @paymentAssignments,
          new MAM::Assignment(name => 'EndEvent', value => $endEvent)
          if $endEvent;
        push @paymentAssignments,
          new MAM::Assignment(name => 'UpdateEvent', value => $updateEvent)
          if $updateEvent;
        if (@paymentAssignments)
        {
            # Modify Payment
            my $subRequest = new MAM::Request(
                database    => $database,
                object      => 'Payment',
                action      => 'Modify',
                actor       => $SUPER_USER,
                assignments => \@paymentAssignments,
                conditions  => [
                    new MAM::Condition(
                        name  => 'Id',
                        value => $id
                    )
                ],
            );
            my $subResponse = MAM::Base->modify($subRequest, $requestId);

            if ($subResponse->getStatus() eq 'Failure')
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    'Unable to modify payment: ' . $subResponse->getMessage());
            }
        }

        # Build up message
        $message .= ", $eventCount event" . ($eventCount == 1 ? '' : 's')
          if $eventCount;
        $message .=
          ", $usageRecordCount usage record"
          . ($usageRecordCount == 1 ? '' : 's')
          if $usageRecordCount;
        $message .=
          " and $componentCount component" . ($componentCount == 1 ? '' : 's')
          if $componentCount;

        # Invoke Payment Quote if Quote option is True
        my $quoteOption = $request->getOptionValue('Quote');
        if (defined $quoteOption
            && $quoteOption =~ /^true$/i)
        {
            my @quoteOptions      = ();
            my $seenQuoteDuration = 0;
            foreach my $option (@options)
            {
                my $name  = $option->getName();
                my $value = $option->getValue();

                if ($name eq 'QuoteDuration')
                {
                    push @quoteOptions,
                      new MAM::Option(
                        name  => 'Duration',
                        value => $value
                      );
                    $seenQuoteDuration++;
                }
                else
                {
                    push @quoteOptions, $option;
                    $seenQuoteDuration++ if $name eq 'Duration';
                }
            }
            push @quoteOptions,
              new MAM::Option(name => 'Duration', value => $duration)
              if (defined $duration && ! $seenQuoteDuration);
            push @quoteOptions, new MAM::Option(name => 'Id', value => $id);

            # Call Payment Quote
            my $subRequest = new MAM::Request(
                database => $database,
                object   => 'Payment',
                action   => 'Quote',
                actor    => $SUPER_USER,
                options  => \@quoteOptions,
            );
            my $subResponse = MAM::Bank->quote($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Failure')
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    'Payment Create -- Failed in payment quote: '
                      . $subResponse->getMessage());
            }
            $message .= "\n" . $subResponse->getMessage();
            $baseData[0]->setProperty('QuoteAmount', $subResponse->getCount());
            $baseMetadata{'QuoteAmount'} = 'Currency';
        }

        # Combine with Payment Begin if requested
        my $beginOption = $request->getOptionValue('Begin');
        if (defined $beginOption
            && $beginOption =~ /^true$/i)
        {
            # Call Payment Begin
            my $subRequest = new MAM::Request(
                database => $database,
                object   => 'Payment',
                action   => 'Begin',
                actor    => $SUPER_USER,
                options  => [new MAM::Option(name => 'Id', value => $id)],
            );
            my $subResponse = MAM::Bank->begin($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Failure')
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    'Payment Create -- Failed beginning payment: '
                      . $subResponse->getMessage());
            }
            $message .= "\n" . $subResponse->getMessage();
        }

        # Return response
        return new MAM::Response()
          ->success(1, \@baseData, \%baseMetadata, $message);
    }

    # UsageRecord
    elsif ($object eq 'UsageRecord')
    {
        # Set Stage = Create unless it is already being set
        $request->addAssignment(
            new MAM::Assignment(name => 'Stage', value => 'Create'))
          unless defined $request->getAssignmentValue('Stage');

        # Build up Usage Record Properties hash
        my %usageRecordProperties = ();
        foreach my $assignment ($request->getAssignments())
        {
            $usageRecordProperties{$assignment->getName()} =
              $assignment->getValue();
        }

        # Employ object instance overrides for specified usage record properties
        my @overrideCheckProperties = keys %usageRecordProperties;
        while (my $propertyName = shift @overrideCheckProperties)
        {
            # If the usage record property has an object associated with it
            if (MAM::Cache->objectExists($propertyName))
            {
                # Iterate over its attributes
                foreach my $attribute (
                    MAM::Cache->listAttributesWithoutPropertyValue(
                        $propertyName, 'Values', undef
                    )
                  )
                {
                    my $values = MAM::Cache->getAttributeProperty($propertyName,
                        $attribute, 'Values');
                    # If an attribute is an override attribute
                    if ($values =~ /^@([:?!]=)(\w+)/)
                    {
                        my ($overrideType, $overrideName) = ($1, $2);
# See if the instance has a value set for this attribute
# Assume the primary key is Name
# SELECT $attribute FROM $propertyName WHERE Name=$usageRecordProperties{$propertyName}
                        my $subResults = $database->select(
                            object => "$propertyName",
                            selections =>
                              [new MAM::Selection(name => $attribute)],
                            conditions => [
                                new MAM::Condition(
                                    name => 'Name',
                                    value =>
                                      $usageRecordProperties{$propertyName}
                                )
                            ]
                        );
                        # If the instance override has a value
                        if (@{$subResults->{data}}
                            && defined ${$subResults->{data}}[0]->[0])
                        {
                            my $overrideValue = ${$subResults->{data}}[0]->[0];
                            # Override it with the value if the usage record
                            # property is undefined or if it has a different
                            # value and this is an assign-always override
                            if ((   ! defined
                                    $usageRecordProperties{$overrideName}
                                )
                                || (   $overrideType eq ':='
                                    && $usageRecordProperties{$overrideName} ne
                                    $overrideValue)
                              )
                            {
                                $request->addAssignment($overrideName,
                                    $overrideValue);
                                $usageRecordProperties{$overrideName} =
                                  $overrideValue;
                                push @overrideCheckProperties, $overrideName;
                            }
                            # Or return with a failure if the usage record
                            # property is defined but has a different
                            # value and this is an assign-verify override
                            elsif ($overrideType eq '!='
                                && defined $usageRecordProperties{$overrideName}
                                && $usageRecordProperties{$overrideName} ne
                                $overrideValue)
                            {
                                return new MAM::Response()->failure('740',
                                    "Specified $overrideName ($usageRecordProperties{$overrideName}) conflicts with ${propertyName}-designated $overrideName ($overrideValue)"
                                );
                            }
                        }
                    }
                }
            }
        }

        # Invoke base Create
        my $response = MAM::Base->create($request, $requestId, 'NoLog');

        if ($response->getStatus() eq 'Success')
        {
            my $usageRecordId = $response->getDatumValue('Id');
            $response->setMessage(
                $response->getMessage() . " with id $usageRecordId");

            # Log the transaction
            MAM::Bank->logTransaction(
                database    => $database,
                requestId   => $requestId,
                txnId       => $database->nextId('Transaction'),
                object      => 'UsageRecord',
                action      => 'Create',
                actor       => $actor,
                assignments => [
                    $request->getAssignments(),
                    new MAM::Assignment(
                        name  => 'Id',
                        value => $usageRecordId
                    ),
                    new MAM::Assignment(
                        name  => 'UsageRecord',
                        value => $usageRecordId
                    ),
                ],
                options => \@options,
                count   => 1,
            );
        }

        return $response;
    }

    else
    {
        return new MAM::Response()
          ->failure('315', "The Bank class does not implement $object Create");
    }
}

# ----------------------------------------------------------------------------
# $response = delete($request, $requestId);
# ----------------------------------------------------------------------------

# Delete (Allocation, Fund, Payment, Transaction, UsageRecord)
sub delete
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object     = $request->getObject();
    my $actor      = $request->getActor();
    my $database   = $request->getDatabase();
    my @conditions = $request->getConditions();
    my @options    = $request->getOptions();
    my $now        = $requestTime;

    if ($object eq 'Fund')
    {
        my $associationCount = 0;

        # Get a list of funds to be deleted
        # SELECT Id FROM Fund WHERE <Conditions>
        my $results = $database->select(
            object     => 'Fund',
            selections => [new MAM::Selection(name => 'Id')],
            conditions => \@conditions
        );
        foreach my $row (@{$results->{data}})
        {
            my $id = $row->[0];

            # Delete all allocations associated with this fund
            # DELETE FROM Allocation WHERE Fund=$id
            my $subRequest = new MAM::Request(
                database   => $database,
                object     => 'Allocation',
                action     => 'Delete',
                actor      => $SUPER_USER,
                conditions => [new MAM::Condition(name => 'Fund', value => $id)]
            );
            my $subResponse = MAM::Bank->delete($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Failure')
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    'Unable to delete allocation: '
                      . $subResponse->getMessage());
            }
            else
            {
                $associationCount += $subResponse->getCount();
            }

            # Delete all constraints associated with this fund
            # DELETE FROM Constraint WHERE Fund=$id
            $subRequest = new MAM::Request(
                database   => $database,
                object     => 'Constraint',
                action     => 'Delete',
                actor      => $SUPER_USER,
                conditions => [new MAM::Condition(name => 'Fund', value => $id)]
            );
            $subResponse = MAM::Base->delete($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Failure')
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    'Unable to delete constraint: '
                      . $subResponse->getMessage());
            }
            else
            {
                $associationCount += $subResponse->getCount();
            }
        }

        # Delete funds by calling base Fund Delete
        my $baseResponse = MAM::Base->delete($request, $requestId);
        my $baseMessage = $baseResponse->getMessage();
        $baseMessage .= " and $associationCount associations"
          if $associationCount;
        $baseResponse->setMessage($baseMessage);

        # Return response
        return $baseResponse;
    }

    elsif ($object eq 'Allocation')
    {
        # Refresh Allocations
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'Allocation',
            action   => 'Refresh'
        );
        my $subResponse = MAM::Bank->refresh($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Failed while refreshing allocations: '
                  . $subResponse->getMessage());
        }

    # Get a list of allocations to be deleted
    # SELECT Id,Fund,Amount,StartTime,EndTime FROM Allocation WHERE <Conditions>
        my $results = $database->select(
            object     => 'Allocation',
            selections => [
                new MAM::Selection(name => 'Id'),
                new MAM::Selection(name => 'Fund'),
                new MAM::Selection(name => 'Amount'),
                new MAM::Selection(name => 'StartTime'),
                new MAM::Selection(name => 'EndTime')
            ],
            conditions => \@conditions
        );
        foreach my $row (@{$results->{data}})
        {
            my $id        = $row->[0];
            my $fund      = $row->[1];
            my $amount    = $row->[2];
            my $startTime = $row->[3];
            my $endTime   = $row->[4];

            my $delta;
            if ($startTime <= $now && $endTime > $now) { $delta = 0 - $amount; }
            else                                       { $delta = 0; }

            # Log the transaction (includes the new fund id and delta)
            MAM::Bank->logTransaction(
                database   => $database,
                requestId  => $requestId,
                txnId      => $database->nextId('Transaction'),
                object     => 'Allocation',
                action     => 'Delete',
                actor      => $actor,
                conditions => [
                    new MAM::Condition(name => 'Id', value => $id),
                    new MAM::Condition(
                        name  => 'Description',
                        value => 'Redundant entry includes fund and delta'
                    )
                ],
                options    => \@options,
                count      => 1,
                fund       => $fund,
                delta      => $delta,
                balance    => 0,
                remaining  => 0,
                allocation => $id
            );
        }

        # Delete allocations by calling base Allocation Delete
        my $baseResponse = MAM::Base->delete($request, $requestId);

        return $baseResponse;
    }

    elsif ($object eq 'Payment')
    {
        # Id
        my $id       = $request->getConditionValue('Id');
        my $warnings = '';

        # Clean up associated events
        my $eventCount = 0;
        # SELECT BeginEvent,UpdateEvent,EndEvent FROM Payment WHERE Id=$id
        my $results = $database->select(
            object     => 'Payment',
            selections => [
                new MAM::Selection(name => 'BeginEvent'),
                new MAM::Selection(name => 'UpdateEvent'),
                new MAM::Selection(name => 'EndEvent'),
            ],
            conditions => \@conditions,
        );

        foreach my $row (@{$results->{data}})
        {
            my $beginEvent  = $row->[0];
            my $updateEvent = $row->[1];
            my $endEvent    = $row->[2];

            foreach my $eventId ($beginEvent, $updateEvent, $endEvent)
            {
                if (defined $eventId)
                {
                    my $subRequest = new MAM::Request(
                        database   => $database,
                        object     => 'Event',
                        action     => 'Delete',
                        actor      => $SUPER_USER,
                        conditions => [
                            new MAM::Condition(
                                name  => 'Id',
                                value => $eventId
                            )
                        ],
                    );
                    my $subResponse =
                      MAM::Base->delete($subRequest, $requestId);
                    if ($subResponse->getStatus() eq 'Failure')
                    {
                        $warnings .=
                          "\nWarning -- Unable to delete payment event: "
                          . $subResponse->getMessage();
                    }
                    $eventCount += $subResponse->getCount();
                }
            }
        }

        # Invoke base Payment Delete
        my $baseResponse = MAM::Base->delete($request, $requestId);
        my $baseMessage = $baseResponse->getMessage();
        $baseMessage .= " and $eventCount events" if $eventCount;
        $baseResponse->setMessage($baseMessage . $warnings);
        if ($warnings)
        {
            $baseResponse->setStatus('Warning');
            $baseResponse->setCode('100');
        }

        # Return response
        return $baseResponse;
    }

    elsif ($object eq 'Transaction' || $object eq 'UsageRecord')
    {
        # Call base delete but use MAM::Bank->logTransaction
        my $baseResponse = MAM::Base->delete($request, $requestId, 'NoLog');
        MAM::Bank->logTransaction(
            database   => $database,
            requestId  => $requestId,
            txnId      => $database->nextId('Transaction'),
            object     => $object,
            action     => 'Delete',
            actor      => $actor,
            conditions => \@conditions,
            options    => \@options,
            count      => $baseResponse->getCount()
        );
        return $baseResponse;
    }

    else
    {
        return new MAM::Response()
          ->failure('315', "The Bank class does not implement $object Delete");
    }
}

# ----------------------------------------------------------------------------
# $response = deposit($request, $requestId);
# ----------------------------------------------------------------------------

# Deposit (Fund)
sub deposit
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object             = $request->getObject();
    my $actor              = $request->getActor();
    my $database           = $request->getDatabase();
    my $fund               = $request->getOptionValue('Id');
    my $allocation         = $request->getOptionValue('Allocation');
    my $amount             = $request->getOptionValue('Amount');
    my $creditLimit        = $request->getOptionValue('CreditLimit');
    my $startTime          = $request->getOptionValue('StartTime');
    my $endTime            = $request->getOptionValue('EndTime');
    my $ancestors          = $request->getOptionValue('Ancestors') || '';
    my @options            = $request->getOptions();
    my $depositSum         = 0;
    my $currency_precision = $serverConfig->{'currency.precision'} || 0;
    my $showHours          = $request->getOptionValue('ShowHours');
    my $useHours           = $request->getOptionValue('UseHours');
    my $time_division = (defined $showHours && $showHours eq 'True') ? 3600 : 1;
    my $reset         = $request->getOptionValue('Reset');
    my $now           = $requestTime;
    my $infinity      = 2147483647;
    my @messages      = ();
    my ($depositStartTime, $depositEndTime);

    # Scale amount into seconds if useHours was specified
    $amount = $amount * 3600
      if (defined $amount && defined $useHours && $useHours =~ /^true$/i);
    $creditLimit = $creditLimit * 3600
      if (defined $creditLimit && defined $useHours && $useHours =~ /^true$/i);

    # Check filters
    my $filterType = $request->getOptionValue('FilterType')
      || 'NonExclusive';
    if (   $filterType ne 'ExactMatch'
        && $filterType ne 'Exclusive'
        && $filterType ne 'NonExclusive')
    {
        return new MAM::Response()->failure('317',
            'The FilterType option must be one of ExactMatch, Exclusive or NonExclusive'
        );
    }

    if ($object ne 'Fund')
    {
        return new MAM::Response()
          ->failure('315', "The Bank class does not implement $object Deposit");
    }

    # Amount must be a non-negative number
    if (defined $amount && $amount !~ /^(\d*\.?\d+(e\+\d+)?|inf(inity)?)$/i)
    {
        return new MAM::Response()
          ->failure('317', 'The deposit amount must be a non-negative number');
    }

    # CreditLimit must be a non-negative number
    if (defined $creditLimit
        && $creditLimit !~ /^(\d*\.?\d+(e\+\d+)?|inf(inity)?)$/i)
    {
        return new MAM::Response()
          ->failure('317', 'The credit limit must be a non-negative number');
    }

    # Credit Limit and Allocation may not both be specified
    if (defined $creditLimit && defined $allocation)
    {
        return new MAM::Response()->failure('740',
            'A credit limit cannot be specifed when an allocation is also specified'
        );
    }

    # Parse out filters
    my %filters = ();
    foreach my $option (@options)
    {
        if ($option->getName() eq 'Filter')
        {
            my $filter = $option->getValue();
            if ($filter =~ /^([^\s=!]+)=([^\s=]+)$/)
            {
                $filters{$1} = $2;
            }
        }
    }

    # Refresh Allocations
    my $subRequest = new MAM::Request(
        database => $database,
        object   => 'Allocation',
        action   => 'Refresh'
    );
    my $subResponse = MAM::Bank->refresh($subRequest, $requestId);
    if ($subResponse->getStatus() eq 'Failure')
    {
        return new MAM::Response()->failure($subResponse->getCode(),
            'Failed while refreshing allocations: '
              . $subResponse->getMessage());
    }

# First, we build a list of funds
# We need to do this separately from the constraint query
# because the trivial fund might have no constraints
# SELECT [DISTINCT] [Fund.]Id,[Fund.]Name,[Fund.]DefaultDeposit FROM Fund[,Allocation] [WHERE Allocation.Fund=Fund.Id AND Allocation.Id=$allocation] [AND [Fund.]Id=$fund]
    my %funds          = ();
    my @fundObjects    = (new MAM::Object(name => 'Fund'));
    my @fundSelections = (
        new MAM::Selection(name => 'Id'),
        new MAM::Selection(name => 'Name'),
        new MAM::Selection(name => 'DefaultDeposit')
    );
    my @fundConditions = ();
    push @fundConditions,
      new MAM::Condition(
        name  => 'Id',
        value => $fund
      ) if defined $fund;
    my @fundOptions = ();
    if (defined $allocation)
    {
        push @fundObjects, new MAM::Object(name => 'Allocation');
        foreach my $selection (@fundSelections)
        {
            $selection->setObject('Fund');
        }
        foreach my $condition (@fundConditions)
        {
            $condition->setObject('Fund');
        }
        push @fundConditions,
          ( new MAM::Condition(
                object  => 'Fund',
                name    => 'Id',
                subject => 'Allocation',
                value   => 'Fund'
            ),
            new MAM::Condition(
                object => 'Allocation',
                name   => 'Id',
                value  => $allocation,
            )
          );
        push @fundOptions, new MAM::Option(name => 'Unique', value => 'True');
    }
    my $results = $database->select(
        objects    => \@fundObjects,
        selections => \@fundSelections,
        conditions => \@fundConditions,
        options    => \@fundOptions,
    );
    foreach my $row (@{$results->{data}})
    {
        my $id             = $row->[0];
        my $name           = $row->[1];
        my $defaultDeposit = $row->[2];
        $funds{$id}{name} = $name || '';
        $funds{$id}{defaultDeposit} = $defaultDeposit;
    }

    # Next, we prune off the funds that violate the constraints
    my %fundConstraints = ();

    # SELECT Id,Fund,Name,Value FROM Constraint
    $results = $database->select(
        object     => 'Constraint',
        selections => [
            new MAM::Selection(name => 'Id'),
            new MAM::Selection(name => 'Fund'),
            new MAM::Selection(name => 'Name'),
            new MAM::Selection(name => 'Value'),
        ],
    );
    foreach my $row (@{$results->{data}})
    {
        my $id     = $row->[0];
        my $fundId = $row->[1];
        my $name   = $row->[2];
        my $value  = $row->[3];

        push @{$fundConstraints{$fundId}{$name}}, $value;
    }

    if ($filterType eq 'ExactMatch')
    {
      FUNDID:
        foreach my $fundId (sort { $a <=> $b } keys %funds)
        {
            # Reject if there are any filters without corresponding constraints
            foreach my $filterName (sort keys %filters)
            {
                if (   (! defined $fundConstraints{$fundId})
                    || (! defined $fundConstraints{$fundId}{$filterName})
                    || (@{$fundConstraints{$fundId}{$filterName}} != 1)
                    || (${$fundConstraints{$fundId}{$filterName}}[0] ne
                        $filters{$filterName})
                  )
                {
                    delete $funds{$fundId};
                    next FUNDID;
                }
            }

            # Reject if there are any constraints without corresponding filters
            next unless exists $fundConstraints{$fundId};
            foreach my $constraintName (sort keys %{$fundConstraints{$fundId}})
            {
                if (   (! defined $filters{$constraintName})
                    || (@{$fundConstraints{$fundId}{$constraintName}} != 1)
                    || (${$fundConstraints{$fundId}{$constraintName}}[0] ne
                        $filters{$constraintName})
                  )
                {
                    delete $funds{$fundId};
                    next FUNDID;
                }
            }
        }
    }

    elsif ($filterType eq 'NonExclusive')
    {
      FUNDID:
        foreach my $fundId (sort { $a <=> $b } keys %fundConstraints)
        {
            my %constraintRejected = ();
            foreach my $filterName (sort keys %filters)
            {
                my $filterValue = $filters{$filterName};
                my $accepted    = 0;
                my $rejected    = 0;

                foreach
                  my $constraintName (sort keys %{$fundConstraints{$fundId}})
                {
                    foreach my $constraintVal (
                        @{$fundConstraints{$fundId}{$constraintName}})
                    {
                        my $constraintValue = $constraintVal;    # Make a copy
                        my $accessAllowed = $constraintValue !~ s/^!//;
                        if (exists $constraintRejected{$constraintName}
                            {$constraintValue})
                        {
                            $rejected = 1;    # This constraint was rejected
                            next;             # Don't allow it to be accepted
                        }

                   # Same Name  |-1|-1|-1| 0| 0| 0|+1|+1|+1|
                   # Other Name |-1| 0|+1|-1| 0|+1|-1| 0|+1|
                   # Accpt/Rjct |R |R |R |  |  |R | A| A|RA|
                   # Result     |X |X |X |/ |/ |X |/ |/ |/ |
                   # A filter should be accepted unless rejected unless accepted

                        # We have a matching constraint name
                        if ($filterName eq $constraintName)
                        {
                            # Mark as accepted if it is a positive
                            # constraint with a matching value
                            if (   $accessAllowed
                                && $filterValue eq $constraintValue)
                            {
                                $accepted = 1;
                            }
                            # Mark as rejected if it is a non-matching value
                            elsif (

                                $accessAllowed
                                && $filterValue ne $constraintValue
                              )
                            {
                                $rejected = 1;
                                $constraintRejected{$constraintName}
                                  {$constraintValue} = 1;
                            }
                            # Immediately delete if it is explicitly rejected
                            elsif (! $accessAllowed
                                && $filterValue eq $constraintValue)
                            {
                                delete $funds{$fundId};
                                next FUNDID;
                            }
                        }

                 # If the constraint is in a parent association with this filter
                 # check it for membership
                 # We could do some caching here but we are not doing it yet
                        elsif (
                            $accessAllowed
                            && MAM::Cache->associationLookup($constraintName,
                                $filterName)
                            && (! exists $fundConstraints{$fundId}{$filterName}
                                || ! grep ! /!/,
                                @{$fundConstraints{$fundId}{$filterName}}
                            )
                          )
                        {
                            my ($association) =
                              MAM::Cache->associationLookup($constraintName,
                                $filterName);
                            # If there is not a matching member found
                            # reject the fund
                            my @membershipConditions = (
                                new MAM::Condition(
                                    name => 'Name',  # Assume simple association
                                    value => $filterValue,
                                    conj  => 'And',
                                    group => '+1',
                                ),
                                new MAM::Condition(
                                    name  => 'Name', # Assume simple association
                                    value => 'ANY',
                                    conj  => 'Or',
                                    group => '-1',
                                )
                            );
                            push @membershipConditions,
                              new MAM::Condition(
                                name  => 'Active',
                                value => 'True'
                              )
                              if MAM::Cache->attributeExists($association,
                                'Active');
                            my $subResults = $database->select(
                                object => $association,
                                selections =>
                                  [new MAM::Selection(name => $constraintName)],
                                conditions => \@membershipConditions,
                            );
                            my $keep = 0;
                            foreach my $row (@{$subResults->{data}})
                            {
                                my $parent = $row->[0];
                                if ($parent eq $constraintValue)
                                {
                                    $keep = 1;
                                }
                            }
                            if ($keep) { $accepted = 1; }
                            else
                            {
                                $rejected = 1;
                                $constraintRejected{$constraintName}
                                  {$constraintValue} = 1;
                            }
                        }

                  # If the constraint is in a child association with this filter
                  # check it for membership
                  # We could do some caching here but we are not doing it yet
                        elsif (
                            $accessAllowed
                            && MAM::Cache->associationLookup($filterName,
                                $constraintName)
                            && ! exists $fundConstraints{$fundId}{$filterName}
                          )
                        {
                            my ($association) =
                              MAM::Cache->associationLookup($filterName,
                                $constraintName);
                            # If there is not a matching member found
                            # reject the fund
                            my @membershipConditions = (
                                new MAM::Condition(
                                    name  => $filterName,
                                    value => $filterValue,
                                )
                            );
                            push @membershipConditions,
                              new MAM::Condition(
                                name  => 'Active',
                                value => 'True'
                              )
                              if MAM::Cache->attributeExists($association,
                                'Active');
                            my $subResults = $database->select(
                                object => $association,
                                selections =>
                                  [new MAM::Selection(name => 'Name')]
                                ,    # Simple name
                                conditions => \@membershipConditions,
                            );
                            my $keep = 0;
                            foreach my $row (@{$subResults->{data}})
                            {
                                my $child = $row->[0];
                                if ($child eq $constraintValue)
                                {
                                    $keep = 1;
                                }
                            }
                            if ($keep) { $accepted = 1; }
                            else
                            {
                                $rejected = 1;
                                $constraintRejected{$constraintName}
                                  {$constraintValue} = 1;
                            }
                        }
                    }
                }

                if ($rejected && ! $accepted)
                {
                    # Delete them
                    delete $funds{$fundId};
                    next FUNDID;
                }
            }
        }
    }

    elsif ($filterType eq 'Exclusive')
    {
        foreach my $fundId (sort { $a <=> $b } keys %fundConstraints)
        {
            foreach my $constraintName (sort keys %{$fundConstraints{$fundId}})
            {
                my %accepted = ();
                my %rejected = ();
                foreach my $constraintValue (
                    @{$fundConstraints{$fundId}{$constraintName}})
                {
                    my $accessAllowed = $constraintValue !~ s/^!//;

                    # If there is a filter matching the constraint
                    if (defined $filters{$constraintName}
                        && $filters{$constraintName} eq $constraintValue)
                    {
                        # It's a matching positive constraint
                        if ($accessAllowed)
                        {
                            my @associations =
                              MAM::Cache->associationLookup($constraintName,
                                undef);

                            # There are associations
                            if (@associations)
                            {
                                foreach my $association (@associations)
                                {
                                    my $child = MAM::Cache->getObjectProperty(
                                        $association, 'Child');

                                    # Check if the usage record property is a
                                    # parent in an association
                                    if (defined $filters{$child})
                                    {
                                        # Check for a match unless the
                                        # fund has a matching constraint
                                        unless (
                                            exists $fundConstraints{$fundId}
                                            {$child})
                                        {
                             # If there is a matching member found, add the fund
                             # We could improve upon this by caching the results
                                            my @membershipConditions = (
                                                new MAM::Condition(
                                                    name => $constraintName,
                                                    value =>
                                                      $filters{$constraintName},
                                                ),
                                                new MAM::Condition(
                                                    name => 'Name'
                                                    , # Assume simple association
                                                    value => $filters{$child},
                                                    conj  => 'And',
                                                    group => '+1',
                                                ),
                                                new MAM::Condition(
                                                    name => 'Name'
                                                    , # Assume simple association
                                                    value => 'ANY',
                                                    conj  => 'Or',
                                                    group => '-1',
                                                ),
                                            );
                                            push @membershipConditions,
                                              new MAM::Condition(
                                                name  => 'Active',
                                                value => 'True'
                                              )
                                              if MAM::Cache->attributeExists(
                                                $association, 'Active');
                                            my $subResults = $database->select(
                                                object => $association,
                                                conditions =>
                                                  \@membershipConditions,
                                            );
                                            if (@{$subResults->{data}})
                                            {
                                                $accepted{$constraintName} = 1;
                                            }
                                            else
                                            {
                                                $rejected{$constraintName} = 1;
                                            }
                                        }
                                        $accepted{$constraintName} = 1;
                                    }
                                    # Child associations must be specified
                                    # for an Exclusive filter
                                    else
                                    {
                                        return new MAM::Response()->failure(
                                            '314',
                                            "A $child must be specified when specifying a $constraintName"
                                        );
                                    }
                                }
                            }
                            # Accept it if there are no associations
                            else
                            {
                                $accepted{$constraintName} = 1;
                            }
                        }
                        # Reject the matching negative constraint
                        else
                        {
                            $rejected{$constraintName} = 1;
                        }
                    }
                    # There is not a filter matching the positive constraint
                    # or the filter does not exist or the constraint is negative
                    else
                    {
                        if ($accessAllowed
                            || ! defined $filters{$constraintName})
                        {
                            $rejected{$constraintName} = 1;
                        }
                    }
                }

                # Constraints of the same name are Ored while
                # Constraints of different names are Anded
                # Delete if there is a constraint name
                # that is rejected and not accepted
                foreach my $name (sort keys %rejected)
                {
                    if (! $accepted{$name} && exists $funds{$fundId})
                    {
                        # Delete them
                        delete $funds{$fundId};
                    }
                }
            }
        }
    }

    if ($log->is_debug())
    {
        $log->debug('Feasible fund list is: ' . join ',', sort keys %funds);
    }

    # If there are no matching funds, see if we should autogen one
    if (scalar keys %funds == 0)
    {
        if (defined $fund)
        {
            return new MAM::Response()
              ->failure('740', 'No funds could be matched for the deposit');
        }

        my $createFund       = 0;
        my $createFundOption = $request->getOptionValue('CreateFund');
        if (MAM::Cache->getObjectProperty('Fund', 'AutoGen') eq 'True')
        {
            if (! defined $createFundOption
                || $createFundOption !~ /^false$/i)
            {
                # Autogen and CreateFund is not False
                $createFund = 1;
            }
        }
        else
        {
            if (defined $createFundOption
                && $createFundOption =~ /^true$/i)
            {
                # No autogen but CreateFund is True
                $createFund = 1;
            }
        }

        # Create a default fund if we resolved to do so
        if ($createFund)
        {
            # Create the fund
            my @fundAutoGenOptions = ();
            foreach my $filterName (sort keys %filters)
            {
                my $filterValue = $filters{$filterName};
                push @fundAutoGenOptions,
                  new MAM::Option(
                    name  => 'Constraint',
                    value => "$filterName=$filterValue"
                  );
            }
            my $subRequest = new MAM::Request(
                database    => $database,
                object      => 'Fund',
                action      => 'Create',
                actor       => $SUPER_USER,
                assignments => [
                    new MAM::Assignment(
                        name  => 'Priority',
                        value => '0'
                    ),
                    new MAM::Assignment(
                        name  => 'Description',
                        value => 'Auto-generated'
                    )
                ],
                options => \@fundAutoGenOptions,
            );
            my $subResponse = MAM::Bank->create($subRequest, $requestId);

            if ($subResponse->getStatus() eq 'Success')
            {
                # Figure out the id of the fund we just created
                $fund = $subResponse->getDatumValue('Id');
                push @messages, "Auto-generated Fund $fund";
            }
            else
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    "No funds could be matched for the deposit.\nUnable to auto-generate fund: "
                      . $subResponse->getMessage());
            }
        }
        else
        {
            return new MAM::Response()
              ->failure('740', 'No funds could be matched for the deposit');
        }
    }

    # List options if multiple funds were matched
    elsif (scalar keys %funds > 1)
    {
        my $message =
          "Multiple funds were matched for the deposit.\nPlease respecify using one of the following fund ids:\n"
          . join("\n", map {"$_\t[$funds{$_}{name}]"} sort keys %funds);
        return new MAM::Response()->failure('740', $message);
    }

    # We should have narrowed the fund down to a single id
    $fund = ${[keys %funds]}[0] unless $fund;

    # Use the default deposit amount if amount was not specified
    my $defaultDeposit = $funds{$fund}{defaultDeposit};
    my $depositAmount =
      defined $amount ? $amount : $defaultDeposit ? $defaultDeposit : 0;
    my $remaining = $depositAmount;

    # Perform a fund stop if Reset option was specified
    # If there is no amount and no default deposit, do not reset
    if (   defined $reset
        && $reset =~ /^true$/i
        && (defined $amount || defined $defaultDeposit))
    {
        # Invoke Fund Stop
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'Fund',
            action   => 'Stop',
            actor    => $SUPER_USER,
            options  => [new MAM::Option(name => 'Id', value => $fund)],
        );
        my $subResponse = MAM::Bank->stop($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                "Failed stopping fund $fund: " . $subResponse->getMessage());
        }
        push @messages, $subResponse->getMessage()
          if $subResponse->getCount() > 0;
    }

    # Prevent cyclic recursive deposits
    if ($ancestors =~ /:${fund}:/)
    {
        return new MAM::Response()->failure('740',
            "Cycle detected. The fund for this deposit ($fund) has already been involved in an upstream deposit. Aborting deposit."
        );
    }

    # Look for deposit shares
    if (($depositAmount > 0) && (! defined $allocation))
    {
      # SELECT Id,DepositShare FROM FundFund WHERE Fund=$fund AND DepositShare>0
        $results = $database->select(
            object     => 'FundFund',
            selections => [
                new MAM::Selection(name => 'Id'),
                new MAM::Selection(name => 'DepositShare')
            ],
            conditions => [
                new MAM::Condition(name => 'Fund', value => $fund),
                new MAM::Condition(
                    name  => 'DepositShare',
                    value => 0,
                    op    => 'GT'
                )
            ]
        );

        # Deposit shares were found -- break up the deposit
        if (@{$results->{data}})
        {
            my %depositShare = ();
            my $shareTotal   = 0;
            foreach my $row (@{$results->{data}})
            {
                my $shareId = $row->[0];
                my $share   = $row->[1];

                $depositShare{$shareId} = $share;
                $shareTotal += $share;
            }

            # Treat the shares as a percentage
            # Fail if the sum of the percentage exceeds 100
            if ($shareTotal > 100)
            {
                return new MAM::Response()
                  ->failure('740',
                    "Deposit shares cannot exceed 100% ($shareTotal)");
            }

            my $shareSum = 0;
            foreach my $shareId (keys %depositShare)
            {
                $shareSum += $depositShare{$shareId};
                my $shareAmount =
                    $shareSum == 100
                  ? $remaining
                  : int($amount * $depositShare{$shareId} / 100);

                my @descendantOptions = (
                    new MAM::Option(
                        name  => 'Id',
                        value => $shareId
                    ),
                    new MAM::Option(
                        name  => 'Amount',
                        value => $shareAmount
                    ),
                    new MAM::Option(
                        name  => 'Ancestors',
                        value => $ancestors . ":$fund:"
                    ),
                );
                if (defined $startTime)
                {
                    push @descendantOptions,
                      new MAM::Option(
                        name  => 'StartTime',
                        value => $startTime
                      );
                }
                if (defined $endTime)
                {
                    push @descendantOptions,
                      new MAM::Option(
                        name  => 'EndTime',
                        value => $endTime
                      );
                }
                # Make a descendant deposit
                my $subRequest = new MAM::Request(
                    database => $database,
                    object   => 'Fund',
                    action   => 'Deposit',
                    options  => \@descendantOptions,
                );
                my $subResponse = MAM::Bank->deposit($subRequest, $requestId);
                if ($subResponse->getStatus() eq 'Failure')
                {
                    return new MAM::Response()->failure($subResponse->getCode(),
                        'Failed while making a descendant deposit: '
                          . $subResponse->getMessage());
                }
                my $count = $subResponse->getCount();
                $depositSum += $count;
                $remaining -= $count;
            }
        }
    }

    # Deposit the remainder into this fund if we have specified an amount
    # (even if it is zero so that an empty allocation gets created), or we
    # have a non-negative default deposit, or we have defined a credit limit.
    if (   ((defined $amount) && ($amount >= 0))
        || ((defined $defaultDeposit) && ($defaultDeposit >= 0))
        || ((defined $creditLimit) && ($creditLimit > 0)))
    {
        # Obtain all allocations for this fund and store in a hash
        # SELECT Id,StartTime,EndTime FROM Allocation WHERE Fund=$fund

        my %preexistingAllocation = ();
        my $results               = $database->select(
            object     => 'Allocation',
            selections => [
                new MAM::Selection(name => 'Id'),
                new MAM::Selection(name => 'StartTime'),
                new MAM::Selection(name => 'EndTime'),
            ],
            conditions => [new MAM::Condition(name => 'Fund', value => $fund)],
        );

        # Populate the preexisting allocation hash with the results
        foreach my $row (@{$results->{data}})
        {
            my $id        = $row->[0];
            my $startTime = $row->[1];
            my $endTime   = $row->[2];

            $preexistingAllocation{$id}{startTime} = $startTime;
            $preexistingAllocation{$id}{endTime}   = $endTime;
        }

        my $allocationId = 0;

        # If allocation specified
        if (defined $allocation)
        {
            if (exists $preexistingAllocation{$allocation})
            {
                $allocationId = $allocation;
            }
            else
            {
                return new MAM::Response()
                  ->failure('740', 'The specified allocation is not valid');
            }
        }

        # If a start time or end time is specified, look for a matching
        # allocation
        elsif (defined $startTime || defined $endTime)
        {
            foreach my $id (reverse sort keys %preexistingAllocation)
            {
                if (defined $startTime && defined $endTime)
                {
                    if (   $preexistingAllocation{$id}{startTime} == $startTime
                        && $preexistingAllocation{$id}{endTime} == $endTime)
                    {
                        $allocationId = $id;
                        last;
                    }
                }
                elsif (defined $startTime)
                {
                    if ($preexistingAllocation{$id}{startTime} == $startTime)
                    {
                        $allocationId = $id;
                        last;
                    }
                }
                elsif (defined $endTime)
                {
                    if ($preexistingAllocation{$id}{endTime} == $endTime)
                    {
                        $allocationId = $id;
                        last;
                    }
                }
            }
        }

        # Look for an active allocation
        else
        {
            foreach my $id (reverse sort keys %preexistingAllocation)
            {
                if (   $preexistingAllocation{$id}{startTime} <= $now
                    && $preexistingAllocation{$id}{endTime} > $now)
                {
                    $allocationId = $id;
                    last;
                }
            }
        }

        # A matching allocation already exists
        if ($allocationId)
        {
            $depositStartTime =
              $preexistingAllocation{$allocationId}{startTime};
            $depositEndTime = $preexistingAllocation{$allocationId}{endTime};

            if ($remaining > 0)
            {
                # Credit the allocation
                my @assignments = (
                    new MAM::Assignment(
                        name  => 'Amount',
                        value => $remaining,
                        op    => 'Inc'
                    ),
                    new MAM::Assignment(
                        name  => 'Allocated',
                        value => $remaining,
                        op    => 'Inc'
                    )
                );
                push @assignments,
                  new MAM::Assignment(
                    name  => 'InitialDeposit',
                    value => $remaining
                  ) if (defined $reset && $reset =~ /^true$/i);
                my $subRequest = new MAM::Request(
                    database    => $database,
                    object      => 'Allocation',
                    action      => 'Modify',
                    actor       => $SUPER_USER,
                    assignments => \@assignments,
                    conditions  => [
                        new MAM::Condition(
                            name  => 'Fund',
                            value => $fund
                        ),
                        new MAM::Condition(
                            name  => 'Id',
                            value => $allocationId
                        )
                    ]
                );
                my $subResponse =
                  MAM::Base->modify($subRequest, $requestId, 'NoLog');
                if ($subResponse->getStatus() eq 'Failure')
                {
                    return new MAM::Response()->failure($subResponse->getCode(),
                        'Unable to credit allocation: '
                          . $subResponse->getMessage());
                }

                # Log the transaction
                MAM::Bank->logTransaction(
                    database    => $database,
                    requestId   => $requestId,
                    txnId       => $database->nextId('Transaction'),
                    object      => 'Allocation',
                    action      => 'Modify',
                    actor       => $actor,
                    assignments => [
                        new MAM::Assignment(
                            name  => 'Amount',
                            value => $remaining,
                        )
                    ],
                    conditions => [
                        new MAM::Condition(
                            name  => 'Id',
                            value => $allocationId
                        )
                    ],
                    count      => 1,
                    fund       => $fund,
                    allocation => $allocationId
                );
            }

            if (defined $creditLimit && $creditLimit > 0)
            {
                # Extend the credit limit
                my $subRequest = new MAM::Request(
                    database    => $database,
                    object      => 'Allocation',
                    action      => 'Modify',
                    actor       => $SUPER_USER,
                    assignments => [
                        new MAM::Assignment(
                            name  => 'CreditLimit',
                            value => $creditLimit,
                            op    => 'Inc'
                        )
                    ],
                    conditions => [
                        new MAM::Condition(
                            name  => 'Fund',
                            value => $fund
                        ),
                        new MAM::Condition(
                            name  => 'Id',
                            value => $allocationId
                        )
                    ]
                );
                my $subResponse =
                  MAM::Base->modify($subRequest, $requestId, 'NoLog');
                if ($subResponse->getStatus() eq 'Failure')
                {
                    return new MAM::Response()->failure($subResponse->getCode(),
                        'Unable to extend credit limit: '
                          . $subResponse->getMessage());
                }

                # Log the transaction
                MAM::Bank->logTransaction(
                    database    => $database,
                    requestId   => $requestId,
                    txnId       => $database->nextId('Transaction'),
                    object      => 'Allocation',
                    action      => 'Modify',
                    actor       => $actor,
                    assignments => [
                        new MAM::Assignment(
                            name  => 'CreditLimit',
                            value => $creditLimit,
                        )
                    ],
                    conditions => [
                        new MAM::Condition(
                            name  => 'Id',
                            value => $allocationId
                        )
                    ],
                    count      => 1,
                    fund       => $fund,
                    allocation => $allocationId
                );
            }
        }

        # An Allocation does not exist
        else
        {
            my $enforceDiscreteOption =
              $request->getOptionValue('EnforceDiscrete');
            my $enforceDiscreteValue =
              defined $enforceDiscreteOption
              ? $enforceDiscreteOption
              : ($serverConfig->{'allocation.enforcediscrete'} || 'true');
            my $enforceDiscreteAllocations =
              ($enforceDiscreteValue =~ /^false$/i) ? 0 : 1;

            # Create a new allocation
            my $activeness;
            $depositStartTime = defined $startTime ? $startTime : $now;
            $depositEndTime   = defined $endTime   ? $endTime   : $infinity;

            # Ensure the end time is not before the start time
            if ($depositStartTime > $depositEndTime)
            {
                return new MAM::Response()
                  ->failure('740',
                    "The start time cannot be after the end time");
            }

            # If we are enforcing discrete allocations, make sure this new
            # allocation does not conflict with an existing one
            if ($enforceDiscreteAllocations)
            {
                foreach my $id (sort keys %preexistingAllocation)
                {
                    my $startTime = $preexistingAllocation{$id}{startTime};
                    my $endTime   = $preexistingAllocation{$id}{endTime};
                    # If an existing allocation does not (end before or
                    # start after) our proposed one, it conflicts
                    if (! (    ($endTime <= $depositStartTime)
                            || ($startTime >= $depositEndTime)
                        )
                      )
                    {
                        my $message = sprintf(
                            "The resulting new allocation with start time (%s) and end time (%s) would overlap with an existing allocation (%d) in the same fund (%d) with start time (%s) and end time (%s)",
                            toHRT($depositStartTime),
                            toHRT($depositEndTime), $id, $fund,
                            toHRT($startTime), toHRT($endTime)
                        );
                        return new MAM::Response()->failure('740', $message);
                    }
                }
            }

            if ($depositStartTime <= $now && $depositEndTime > $now)
            {
                $activeness = 'True';
            }
            else { $activeness = 'False'; }
            my $subRequest = new MAM::Request(
                database    => $database,
                object      => 'Allocation',
                action      => 'Create',
                actor       => $SUPER_USER,
                assignments => [
                    new MAM::Assignment(
                        name  => 'Fund',
                        value => $fund
                    ),
                    new MAM::Assignment(
                        name  => 'StartTime',
                        value => $depositStartTime
                    ),
                    new MAM::Assignment(
                        name  => 'EndTime',
                        value => $depositEndTime
                    ),
                    new MAM::Assignment(
                        name  => 'Amount',
                        value => $remaining
                    ),
                    new MAM::Assignment(
                        name  => 'CreditLimit',
                        value => defined $creditLimit ? $creditLimit : 0
                    ),
                    new MAM::Assignment(
                        name  => 'InitialDeposit',
                        value => $remaining
                    ),
                    new MAM::Assignment(
                        name  => 'Allocated',
                        value => $remaining
                    ),
                    new MAM::Assignment(
                        name  => 'Active',
                        value => $activeness
                    )
                ]
            );
            my $subResponse = MAM::Base->create($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Failure')
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    'Unable to create allocation: '
                      . $subResponse->getMessage());
            }
            push @messages, $subResponse->getMessage();

            # Figure out the id of the allocation we just created
            $allocationId = $subResponse->getDatumValue('Id');
        }

        # Log the transaction
        if ($depositStartTime <= $now && $depositEndTime > $now)
        {
            MAM::Bank->logTransaction(
                database   => $database,
                requestId  => $requestId,
                txnId      => $database->nextId('Transaction'),
                object     => 'Fund',
                action     => 'Deposit',
                actor      => $actor,
                options    => \@options,
                count      => 1,
                fund       => $fund,
                amount     => $remaining,
                balance    => $remaining,
                delta      => $remaining,
                remaining  => $remaining,
                allocation => $allocationId
            );
        }
        else
        {
            MAM::Bank->logTransaction(
                database   => $database,
                requestId  => $requestId,
                txnId      => $database->nextId('Transaction'),
                object     => 'Fund',
                action     => 'Deposit',
                actor      => $actor,
                options    => \@options,
                count      => 1,
                fund       => $fund,
                amount     => $remaining,
                balance    => 0,
                delta      => 0,
                remaining  => $remaining,
                allocation => $allocationId
            );
        }
        $depositSum += $remaining;
    }

    $depositSum =
      sprintf("%.${currency_precision}f", $depositSum / $time_division);
    if ($depositSum > 0)
    {
        unshift @messages,
          "Successfully deposited $depositSum credits into fund $fund";
    }
    else
    {
        unshift @messages, "No credits were deposited into fund $fund";
    }
    return new MAM::Response()->success($depositSum, join("\n", @messages));
}

# ----------------------------------------------------------------------------
# $response = end($request, $requestId);
# ----------------------------------------------------------------------------

# End (Payment)
sub end
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object   = $request->getObject();
    my $database = $request->getDatabase();
    my $actor    = $request->getActor();
    my @options  = $request->getOptions();
    my $now      = $requestTime;
    my @messages = ();
    my $warnings = 0;
    my $response;

    if ($object ne 'Payment')
    {
        return new MAM::Response()
          ->failure('315', "The Bank class does not implement $object End");
    }

    # Id
    my $id = $request->getOptionValue('Id');
    if (! defined $id)
    {
        return new MAM::Response()
          ->failure('314', 'An Id option must be specified');
    }

    # Call Payment Stop
    my $stopOption = $request->getOptionValue('Stop');
    unless (defined $stopOption && $stopOption =~ /^false$/i)
    {
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'Payment',
            action   => 'Stop',
            actor    => $SUPER_USER,
            options  => [new MAM::Option(name => 'Id', value => $id)],
        );
        my $subResponse = MAM::Bank->stop($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Payment End failed in stopping payment: '
                  . $subResponse->getMessage());
        }
        push @messages, $subResponse->getMessage();
    }

# SELECT BeginTime,ChargePattern,Duration,EndEvent,ReservePattern,UpdateEvent FROM Payment WHERE Id=$id
    my $results = $database->select(
        object     => 'Payment',
        selections => [
            new MAM::Selection(name => 'BeginTime'),
            new MAM::Selection(name => 'ChargePattern'),
            new MAM::Selection(name => 'Duration'),
            new MAM::Selection(name => 'EndEvent'),
            new MAM::Selection(name => 'ReservePattern'),
            new MAM::Selection(name => 'UpdateEvent'),
        ],
        conditions => [
            new MAM::Condition(
                name  => 'Id',
                value => $id,
            ),
        ],
    );

    unless (@{$results->{data}})
    {
        return new MAM::Response()
          ->failure('317', "Payment End: Invalid payment id ($id)");
    }

    my $beginTime      = ${$results->{data}}[0]->[0];
    my $chargePattern  = ${$results->{data}}[0]->[1];
    my $duration       = ${$results->{data}}[0]->[2];
    my $endEvent       = ${$results->{data}}[0]->[3];
    my $reservePattern = ${$results->{data}}[0]->[4];
    my $updateEvent    = ${$results->{data}}[0]->[5];
    unless (defined $reservePattern)
    {
        return new MAM::Response()
          ->failure('740',
            'Payment End: Unable to determine payment reserve mode');
    }
    unless (defined $chargePattern)
    {
        return new MAM::Response()
          ->failure('740',
            'Payment End: Unable to determine payment charge mode');
    }

    # Invoke end charges
    # SELECT Id FROM PaymentComponent WHERE Payment=$id AND Type='End'
    $results = $database->select(
        object     => 'PaymentComponent',
        selections => [new MAM::Selection(name => 'Id'),],
        conditions => [
            new MAM::Condition(
                name  => 'Payment',
                value => $id,
            ),
            new MAM::Condition(
                name  => 'Type',
                value => 'End',
            ),
        ],
    );

    # Iterate over end components
    my $chargeAmount  = 0;
    my @endComponents = @{$results->{data}};
    foreach my $row (@{$results->{data}})
    {
        my $usageRecord = $row->[0];

        # UsageRecord Charge
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'UsageRecord',
            action   => 'Charge',
            actor    => $SUPER_USER,
            options  => [new MAM::Option(name => 'Id', value => $usageRecord)],
        );
        my $subResponse = MAM::Bank->charge($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Payment End failed in ending payment charge: '
                  . $subResponse->getMessage());
        }
        $chargeAmount += $subResponse->getCount();
        push @messages, $subResponse->getMessage();
    }

    # Remove outstanding liens
    if ($reservePattern eq 'Combined')
    {
        # SELECT Id FROM PaymentComponent WHERE Payment=$id AND Type='Update'
        my $results = $database->select(
            object     => 'PaymentComponent',
            selections => [new MAM::Selection(name => 'Id'),],
            conditions => [
                new MAM::Condition(
                    name  => 'Payment',
                    value => $id,
                ),
                new MAM::Condition(
                    name  => 'Type',
                    value => 'Update',
                ),
            ],
        );

        # Iterate over update components
        foreach my $row (@{$results->{data}})
        {
            my $usageRecord = $row->[0];

            # Lien Delete
            my $subRequest = new MAM::Request(
                database   => $database,
                object     => 'Lien',
                action     => 'Delete',
                actor      => $SUPER_USER,
                conditions => [
                    new MAM::Condition(
                        name  => 'UsageRecord',
                        value => $usageRecord
                    )
                ],
            );
            my $subResponse = MAM::Base->delete($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Failure')
            {
                push @messages,
                  'Warning -- Unable to remove update component lien: '
                  . $subResponse->getMessage();
                $warnings++;
            }
            push @messages, $subResponse->getMessage();
        }
    }

# UPDATE Payment SET EndTime=$now,EndEvent=NULL,UpdateEvent=NULL[,Duration=$endTime-$beginTime] WHERE Id=$id
    {
        my @paymentAssignments = (
            new MAM::Assignment(
                name  => 'UpdateEvent',
                value => 'NULL',
            ),
            new MAM::Assignment(
                name  => 'EndTime',
                value => $now,
            ),
            new MAM::Assignment(
                name  => 'EndEvent',
                value => 'NULL',
            ),
        );
        push @paymentAssignments,
          new MAM::Assignment(name => 'Duration', value => $now - $beginTime)
          if (defined $beginTime && ! defined $duration);
        my $subRequest = new MAM::Request(
            database    => $database,
            object      => 'Payment',
            action      => 'Modify',
            actor       => $SUPER_USER,
            assignments => \@paymentAssignments,
            conditions  => [
                new MAM::Condition(
                    name  => 'Id',
                    value => $id
                )
            ],
        );
        my $subResponse = MAM::Base->modify($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            push @messages, 'Warning -- Unable to modify payment end time';
            $warnings++;
        }
    }

    # Update EndTime of update component usage records
    # SELECT Id FROM PaymentComponent WHERE Payment=$id AND Type='Update'
    $results = $database->select(
        object     => 'PaymentComponent',
        selections => [new MAM::Selection(name => 'Id'),],
        conditions => [
            new MAM::Condition(
                name  => 'Payment',
                value => $id,
            ),
            new MAM::Condition(
                name  => 'Type',
                value => 'Update',
            ),
        ],
    );

    # Iterate over update components
    my @updateComponents = @{$results->{data}};
    foreach my $row (@{$results->{data}})
    {
        my $usageRecord = $row->[0];

        # UsageRecord Modify
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'UsageRecord',
            action   => 'Modify',
            actor    => $SUPER_USER,
            conditions =>
              [new MAM::Condition(name => 'Id', value => $usageRecord)],
            assignments =>
              [new MAM::Assignment(name => 'EndTime', value => $now)],
        );
        my $subResponse = MAM::Base->modify($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Payment End failed setting end time in payment update component usage record: '
                  . $subResponse->getMessage());
        }
    }

    # Delete Payment updateEvent
    if (defined $updateEvent)
    {
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'Event',
            action   => 'Delete',
            actor    => $SUPER_USER,
            conditions =>
              [new MAM::Condition(name => 'Id', value => $updateEvent)],
        );
        my $subResponse = MAM::Base->delete($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            push @messages, 'Warning -- Unable to delete update event: '
              . $subResponse->getMessage();
            $warnings++;
        }
    }

    # Delete Payment EndEvent
    if (defined $endEvent)
    {
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'Event',
            action   => 'Delete',
            actor    => $SUPER_USER,
            conditions =>
              [new MAM::Condition(name => 'Id', value => $endEvent)],
        );
        my $subResponse = MAM::Base->delete($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            push @messages, 'Warning -- Unable to delete end event: '
              . $subResponse->getMessage();
            $warnings++;
        }
    }

    # Delete End and Update PaymentComponents
    if (@updateComponents || @endComponents)
    {
        my $subRequest = new MAM::Request(
            database   => $database,
            object     => 'PaymentComponent',
            action     => 'Delete',
            actor      => $SUPER_USER,
            conditions => [
                new MAM::Condition(
                    name  => 'Payment',
                    value => $id,
                ),
                new MAM::Condition(
                    name  => 'Type',
                    value => 'Update',
                    conj  => 'And',
                    group => '+1'
                ),
                new MAM::Condition(
                    name  => 'Type',
                    value => 'End',
                    conj  => 'Or',
                    group => '-1'
                ),
            ],
        );
        my $subResponse = MAM::Base->delete($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            push @messages,
              'Warning -- Unable to delete update and/or end payment components: '
              . $subResponse->getMessage();
            $warnings++;
        }
    }

    # Log the transaction
    MAM::Bank->logTransaction(
        database  => $database,
        requestId => $requestId,
        txnId     => $database->nextId('Transaction'),
        object    => 'Payment',
        action    => 'End',
        actor     => $actor,
        options   => \@options,
        count     => 1,
    );

    # Call Payment Delete if Delete option specified
    my $deleteOption = $request->getOptionValue('Delete');
    if (defined $deleteOption && $deleteOption =~ /^true$/i)
    {
        my $subRequest = new MAM::Request(
            database   => $database,
            object     => 'Payment',
            action     => 'Delete',
            actor      => $SUPER_USER,
            conditions => [new MAM::Condition(name => 'Id', value => $id)],
        );
        my $subResponse = MAM::Bank->delete($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Payment End failed in deleting payment: '
                  . $subResponse->getMessage());
        }
        push @messages, $subResponse->getMessage();
    }

    # Return response
    push @messages, 'No action taken' unless @messages;
    if ($warnings)
    {
        $response = new MAM::Response()
          ->warning($chargeAmount, 'Payment End: ' . join("\n", @messages));
    }
    else
    {
        $response = new MAM::Response()
          ->success($chargeAmount, 'Payment End: ' . join("\n", @messages));
    }
    return $response;
}

# ----------------------------------------------------------------------------
# $response = modify($request, $requestId);
# ----------------------------------------------------------------------------

# Modify (Allocation, FundFund)
sub modify
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object   = $request->getObject();
    my $actor    = $request->getActor();
    my $database = $request->getDatabase();
    my @options  = $request->getOptions();
    my $now      = $requestTime;

    if ($object eq 'Allocation')
    {
        my $enforceDiscreteOption = $request->getOptionValue('EnforceDiscrete');
        my $enforceDiscreteValue =
          defined $enforceDiscreteOption
          ? $enforceDiscreteOption
          : ($serverConfig->{'allocation.enforcediscrete'} || 'true');
        my $enforceDiscreteAllocations =
          ($enforceDiscreteValue =~ /^false$/i) ? 0 : 1;

        # The Amount property cannot be modified directly
        if (defined $request->getAssignmentValue('Amount'))
        {
            return new MAM::Response()->failure('740',
                'The Amount cannot be modified directly because it bypasses funding. Use Fund Deposit or Fund Withdraw instead.'
            );
        }

        # The Active property cannot be modified directly
        if (defined $request->getAssignmentValue('Active'))
        {
            return new MAM::Response()->failure('740',
                'The Active property cannot be modified directly. Modify the StartTime and/or EndTime instead.'
            );
        }

        # CreditLimit must be a non-negative number
        my $creditLimit = $request->getAssignmentValue('CreditLimit');
        if (defined $creditLimit
            && $creditLimit !~ /^(\d*\.?\d+(e\+\d+)?|inf(inity)?)$/i)
        {
            return new MAM::Response()
              ->failure('317',
                'The credit limit must be a non-negative number');
        }

        # Obtain a list of allocation start and end times in the same fund
        my $allocationId = $request->getConditionValue('Id');
        my $newStartTime = $request->getAssignmentValue('StartTime');
        my $newEndTime   = $request->getAssignmentValue('EndTime');
        if ((defined $newStartTime) || (defined $newEndTime))
        {
            if (! defined $allocationId)
            {
                return new MAM::Response()->failure('314',
                    'An allocation id must be specified when modifying an allocation start time or end time'
                );
            }

          # Obtain fund id, start time and end time from the allocation
          # being modified
          # SELECT Fund,StartTime,EndTime FROM Allocation WHERE Id=$allocationId
            my $results = $database->select(
                object     => 'Allocation',
                selections => [
                    new MAM::Selection(name => 'Fund'),
                    new MAM::Selection(name => 'StartTime'),
                    new MAM::Selection(name => 'EndTime'),
                ],
                conditions =>
                  [new MAM::Condition(name => 'Id', value => $allocationId),],
            );

            my ($fundId, $origStartTime, $origEndTime);
            if (@{$results->{data}})
            {
                $fundId        = ${$results->{data}}[0]->[0];
                $origStartTime = ${$results->{data}}[0]->[1];
                $origEndTime   = ${$results->{data}}[0]->[2];
            }
            else
            {
                return new MAM::Response()
                  ->failure('317', "Invalid allocation id ($allocationId)");
            }

            # Populate a hash of start and end time from other allocations in
            # the fund
            # SELECT Id,StartTime,EndTime FROM Allocation WHERE Fund=$fundId
            # AND Id!=$allocationId
            $results = $database->select(
                object     => 'Allocation',
                selections => [
                    new MAM::Selection(name => 'Id'),
                    new MAM::Selection(name => 'StartTime'),
                    new MAM::Selection(name => 'EndTime'),
                ],
                conditions => [
                    new MAM::Condition(name => 'Fund', value => $fundId),
                    new MAM::Condition(
                        name  => 'Id',
                        value => $allocationId,
                        op    => 'NE'
                    ),
                ],
            );

            # Populate the allocation hash with the results
            my %allocation = ();
            foreach my $row (@{$results->{data}})
            {
                my $id        = $row->[0];
                my $startTime = $row->[1];
                my $endTime   = $row->[2];

                $allocation{$id}{startTime} = $startTime;
                $allocation{$id}{endTime}   = $endTime;
            }

            # StartTime can be changed to any time in past or future as long as:
            # 1) the StartTime is not after the EndTime
            # 2) it does not overlap with an existing allocation
            # 3) it can be moved forward only if it has not already started
            #    (i.e. the source and target start times are both in the future)
            if (defined $newStartTime)
            {
                if ($newStartTime >
                    (defined $newEndTime ? $newEndTime : $origEndTime))
                {
                    return new MAM::Response()->failure('740',
                        "The allocation start time cannot be moved after the end time"
                    );
                }

                if (   $enforceDiscreteAllocations
                    && $origStartTime <= $now
                    && $newStartTime > $origStartTime)
                {
                    return new MAM::Response()->failure('740',
                        "The allocation start time cannot be moved forward if it has already started"
                    );
                }
            }

            # EndTime can be changed to any time in past or future as long as:
            # 1) the EndTime is not before the StartTime
            # 2) it does not overlap with an existing allocation
            # 3) it can be moved backward only if the end time is in the future
            #    (i.e. the source and target end times are both in the future)
            if (defined $newEndTime)
            {
                if ($newEndTime <
                    (defined $newStartTime ? $newStartTime : $origStartTime))
                {
                    return new MAM::Response()->failure('740',
                        "The allocation end time cannot be moved before the start time"
                    );
                }

                if (   $enforceDiscreteAllocations
                    && $newEndTime < $now
                    && $newEndTime < $origEndTime)
                {
                    return new MAM::Response()->failure('740',
                        "The allocation end time may not be moved backward if the result is in the past"
                    );
                }
            }

            my $resultingEndTime =
              defined $newEndTime ? $newEndTime : $origEndTime;
            my $resultingStartTime =
              defined $newStartTime ? $newStartTime : $origStartTime;

            # If we are enforcing discrete allocations, make sure the
            # modified allocation does not conflict with an existing one
            if ($enforceDiscreteAllocations)
            {
                foreach my $id (sort keys %allocation)
                {
                    my $startTime = $allocation{$id}{startTime};
                    my $endTime   = $allocation{$id}{endTime};
                    # If an existing allocation does not (end before or
                    # start after) our proposed one, it conflicts
                    if (! (    ($endTime <= $resultingStartTime)
                            || ($startTime >= $resultingEndTime)
                        )
                      )
                    {
                        my $message = sprintf(
                            "The modified allocation (%d) with start time (%s) and end time (%s) would overlap with an existing allocation (%d) in the same fund (%d) with start time (%s) and end time (%s)",
                            $allocationId,
                            toHRT($resultingStartTime),
                            toHRT($resultingEndTime),
                            $id,
                            $fundId,
                            toHRT($startTime),
                            toHRT($endTime)
                        );
                        return new MAM::Response()->failure('740', $message);
                    }
                }
            }
        }

        # Invoke base Allocation Modify
        my $baseResponse = MAM::Base->modify($request, $requestId);

        # Refresh Allocations
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'Allocation',
            action   => 'Refresh'
        );
        my $subResponse = MAM::Bank->refresh($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Failed while refreshing allocations: '
                  . $subResponse->getMessage());
        }

        return $baseResponse;
    }

    elsif ($object eq 'FundFund')
    {
        my $depositShare = $request->getAssignmentValue('DepositShare');

        if (defined $depositShare)
        {
            # Make sure deposit shares do not exceed 100 for a given fund
            my $fund = $request->getConditionValue('Fund');
            my $id   = $request->getConditionValue('Id');
            unless (defined $fund && defined $id)
            {
                return new MAM::Response()->failure('740',
                    'Deposit shares cannot be modified for multiple subfunds simultaneously'
                );
            }
            my $depositShareTotal = $depositShare;

            my $results = $database->select(
                object     => 'FundFund',
                selections => [
                    new MAM::Selection(
                        name => 'DepositShare',
                        op   => 'Sum'
                    )
                ],
                conditions => [
                    new MAM::Condition(
                        name  => 'Fund',
                        value => $fund
                    ),
                    new MAM::Condition(
                        name  => 'Id',
                        value => $id,
                        op    => 'NE'
                    )
                ]
            );
            my $depositShareSum = ${$results->{data}}[0]->[0];
            if (defined $depositShareSum)
            {
                $depositShareTotal += $depositShareSum;
            }
            if ($depositShareTotal > 100)
            {
                return new MAM::Response()->failure('740',
                    "Total deposit shares for a fund cannot exceed 100% ($depositShareTotal)"
                );
            }
        }

        # Modify fund funds by calling base FundFund Modify
        my $baseResponse = MAM::Base->modify($request, $requestId);

        return $baseResponse;
    }

    else
    {
        return new MAM::Response()
          ->failure('315', "The Bank class does not implement $object Modify");
    }
}

# ----------------------------------------------------------------------------
# $response = query($request, $requestId[, $chunkNum]);
# ----------------------------------------------------------------------------

# Query (Account, Allocation, Fund, Payment, Quote, Lien, Transaction, UsageRecord, User)
sub query
{
    my ($class, $request, $requestId, $chunkNum) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object     = $request->getObject();
    my $actor      = $request->getActor();
    my $database   = $request->getDatabase();
    my @selections = $request->getSelections();
    my @conditions = $request->getConditions();
    my @options    = $request->getOptions();
    my @objects    = $request->getObjects();
    my $now        = $requestTime;
    $chunkNum = 1 unless defined $chunkNum;

    if ($object eq 'Account')
    {
        my $name = $request->getConditionValue('Name');

        # Parse out filters
        my %filters = ();
        foreach my $option (@options)
        {
            if ($option->getName() eq 'Filter')
            {
                my $filter = $option->getValue();
                if ($filter =~ /^([^\s=!]+)=([^\s=]+)$/)
                {
                    $filters{$1} = $2;
                }
            }
        }

        # Override Authorization
        my $authorized = 1;
        if ($request->getOverride())
        {
            $authorized = 0;
        }
        if (   ! $authorized
            && defined $filters{User}
            && $filters{User} eq $actor)
        {
            $authorized = 1;
        }
        if (   ! $authorized
            && defined $name
            && MAM::Cache->objectExists('AccountUser'))
        {
# See if actor is account admin
# SELECT Admin FROM AccountUser WHERE Account=$account AND Name=$actor AND Admin=True
            my $subResults = $database->select(
                object     => 'AccountUser',
                selections => [new MAM::Selection(name => 'Admin')],
                conditions => [
                    new MAM::Condition(
                        name  => 'Account',
                        value => $name
                    ),
                    new MAM::Condition(
                        name  => 'Name',
                        value => $actor
                    ),
                    new MAM::Condition(
                        name  => 'Admin',
                        value => 'True'
                    )
                ]
            );
            if (@{$subResults->{data}})
            {
                $authorized = 1;
            }
        }
        if (! $authorized)
        {
            if (! defined $filters{User})
            {
                push @options,
                  new MAM::Option(
                    name  => 'Filter',
                    value => "User=$actor"
                  );
                $filters{User} = $actor;
            }
            else
            {
                return new MAM::Response()
                  ->failure('444',
                    "$actor is not authorized to Query the Accounts");
            }
        }

        # Quick access to base query if no options are specified
        # This has to occur after the override authorization check
        if (! @options || ! MAM::Cache->objectExists('AccountUser'))
        {
            # Invoke base Account Query
            return MAM::Base->query($request, $requestId, $chunkNum);
        }

# Issue Account Query with appropriate join conditions
# SELECT [DISTINCT] Account.@selections FROM Account[,AccountUser] WHERE [Account.]@conditions [AND Account.Name=AccountUser.Account]
        my @accountObjects    = (new MAM::Object(name => 'Account'));
        my @accountSelections = @selections;
        my @accountConditions = @conditions;
        my @accountOptions    = @options;
        if (defined $filters{User})
        {
            push @accountObjects, (new MAM::Object(name => 'AccountUser'),);
            foreach my $selection (@accountSelections)
            {
                $selection->setObject('Account');
            }
            foreach my $condition (@accountConditions)
            {
                $condition->setObject('Account');
            }
            push @accountConditions,
              ( new MAM::Condition(
                    object  => 'Account',
                    name    => 'Name',
                    subject => 'AccountUser',
                    value   => 'Account',
                ),
                new MAM::Condition(
                    object => 'AccountUser',
                    name   => 'Name',
                    value  => $filters{User},
                ),
              );
            push @accountOptions,
              new MAM::Option(name => 'Unique', value => 'True');
        }
        my $results = $database->select(
            objects    => \@accountObjects,
            selections => \@accountSelections,
            conditions => \@accountConditions,
            options    => \@accountOptions,
        );
        my $count    = 0;
        my @data     = ();
        my $data     = $results->{data};
        my $cols     = $results->{cols};
        my $names    = $results->{names};
        my %metadata = %{$results->{metadata}};

        # Populate Data
        foreach my $row (@{$data})
        {
            $count++;
            my $datum = new MAM::Datum($object);
            for (my $i = 0; $i < $cols; $i++)
            {
                my $name  = $names->[$i];
                my $value = $row->[$i];     # Null comes in as undef
                $datum->addProperty($name, $value);
            }
            push @data, $datum;
        }

        my $response = new MAM::Response()->success($count, \@data, \%metadata);

        return $response;
    }

    elsif ($object eq 'Allocation')
    {
        my $includeAncestors = $request->getOptionValue('IncludeAncestors')
          || 'False';
        my $filterType = $request->getOptionValue('FilterType')
          || 'NonExclusive';

        if (   $filterType ne 'ExactMatch'
            && $filterType ne 'Exclusive'
            && $filterType ne 'NonExclusive')
        {
            return new MAM::Response()->failure('317',
                'The FilterType option must be one of ExactMatch, Exclusive or NonExclusive'
            );
        }

        # Parse out filters
        my %filters = ();
        foreach my $option (@options)
        {
            if ($option->getName() eq 'Filter')
            {
                my $filter = $option->getValue();
                if ($filter =~ /^([^\s=!]+)=([^\s=]+)$/)
                {
                    $filters{$1} = $2;
                }
            }
        }

        # Override Authorization
        my $authorized = 1;
        if ($request->getOverride())
        {
            $authorized = 0;
        }
        if (   ! $authorized
            && defined $filters{User}
            && $filters{User} eq $actor)
        {
            $authorized = 1;
        }
        if (   ! $authorized
            && defined $filters{Account}
            && MAM::Cache->objectExists('AccountUser'))
        {
# See if actor is account admin
# SELECT Admin FROM AccountUser WHERE Account=$account AND Name=$actor AND Admin=True
            my $subResults = $database->select(
                object     => 'AccountUser',
                selections => [new MAM::Selection(name => 'Admin')],
                conditions => [
                    new MAM::Condition(
                        name  => 'Account',
                        value => $filters{Account},
                    ),
                    new MAM::Condition(
                        name  => 'Name',
                        value => $actor
                    ),
                    new MAM::Condition(
                        name  => 'Admin',
                        value => 'True'
                    )
                ]
            );
            if (@{$subResults->{data}})
            {
                $authorized = 1;
            }
        }
        if (! $authorized)
        {
            if (! defined $filters{User})
            {
                push @options,
                  new MAM::Option(
                    name  => 'Filter',
                    value => "User=$actor"
                  );
                $filters{User} = $actor;
            }
            else
            {
                return new MAM::Response()
                  ->failure('444',
                    "$actor is not authorized to Query the Allocations");
            }
        }

        # Quick access to base query if no options are specified
        if (! @options)
        {
            # Invoke base Allocation Query
            return MAM::Base->query($request, $requestId, $chunkNum);
        }

        # Verify that every specified usage record filter option exists if it is
        # a value-constrained object and that it is active if it has an active
        # field
        foreach my $filterName (sort keys %filters)
        {
            my $values =
              MAM::Cache->getAttributeProperty('UsageRecord', $filterName,
                'Values');
            if (   defined $values
                && $values =~ /^@(?:[:?!]=)?(\w+)/
                && MAM::Cache->objectExists($1))
            {
                my $filterObject = $1;
                my @primaryKeys =
                  MAM::Cache->listAttributesWithPropertyValue($filterObject,
                    'PrimaryKey', 'True');

                # We can only do this if it has a single primary key
                if (@primaryKeys == 1)
                {
   # SELECT $filterKey[,Active] FROM $filterObject WHERE $filterKey=$filterValue
                    my $filterKey   = $primaryKeys[0];
                    my $filterValue = $filters{$filterName};
                    my @validitySelections =
                      (new MAM::Selection(name => $filterKey));
                    push @validitySelections,
                      new MAM::Selection(name => 'Active')
                      if MAM::Cache->attributeExists($filterObject, 'Active');

                    my @validityConditions = (
                        new MAM::Condition(
                            name  => $filterKey,
                            value => $filterValue
                        )
                    );
                    my $subResults = $database->select(
                        object     => $filterObject,
                        selections => \@validitySelections,
                        conditions => \@validityConditions,
                    );
                    unless (@{$subResults->{data}})
                    {
                        return new MAM::Response()->failure('740',
                            "$filterObject ($filterName) $filterValue does not exist"
                        );
                    }
                    elsif (exists ${$subResults->{data}}[0]->[1]
                        && ${$subResults->{data}}[0]->[1] ne 'True')
                    {
                        return new MAM::Response()->failure('740',
                            "$filterObject ($filterName) $filterValue is not active"
                        );
                    }
                }
            }
        }

        # First, we build a list of allocations from the supplied conditions
        # We need to do this separately from the constraint query
        # because the trivial fund might have no constraints
        # SELECT Id FROM Allocation WHERE @conditions
        my %funds   = ();
        my $results = $database->select(
            object     => 'Allocation',
            selections => [
                new MAM::Selection(name => 'Id'),
                new MAM::Selection(name => 'Fund'),
            ],
            conditions => \@conditions,
            options    => \@options,
        );
        foreach my $row (@{$results->{data}})
        {
            my $id     = $row->[0];
            my $fundId = $row->[1];
            $funds{$fundId}{$id}++;
        }

        # Next, we prune off the funds that violate the constraints to limbo
        my %fundConstraints = ();
        my %limbo           = ();

        # SELECT Id,Fund,Name,Value FROM Constraint
        $results = $database->select(
            object     => 'Constraint',
            selections => [
                new MAM::Selection(name => 'Id'),
                new MAM::Selection(name => 'Fund'),
                new MAM::Selection(name => 'Name'),
                new MAM::Selection(name => 'Value'),
            ],
        );
        foreach my $row (@{$results->{data}})
        {
            my $id     = $row->[0];
            my $fundId = $row->[1];
            my $name   = $row->[2];
            my $value  = $row->[3];

            push @{$fundConstraints{$fundId}{$name}}, $value;
        }

        if ($filterType eq 'ExactMatch')
        {
          FUNDID:
            foreach my $fundId (sort { $a <=> $b } keys %funds)
            {
             # Reject if there are any filters without corresponding constraints
                foreach my $filterName (sort keys %filters)
                {
                    if (   (! defined $fundConstraints{$fundId})
                        || (! defined $fundConstraints{$fundId}{$filterName})
                        || (@{$fundConstraints{$fundId}{$filterName}} != 1)
                        || (${$fundConstraints{$fundId}{$filterName}}[0] ne
                            $filters{$filterName})
                      )
                    {
                        delete $funds{$fundId};
                        next FUNDID;
                    }
                }

             # Reject if there are any constraints without corresponding filters
                next unless exists $fundConstraints{$fundId};
                foreach
                  my $constraintName (sort keys %{$fundConstraints{$fundId}})
                {
                    if (   (! defined $filters{$constraintName})
                        || (@{$fundConstraints{$fundId}{$constraintName}} != 1)
                        || (${$fundConstraints{$fundId}{$constraintName}}[0] ne
                            $filters{$constraintName})
                      )
                    {
                        delete $funds{$fundId};
                        next FUNDID;
                    }
                }
            }
        }

        elsif ($filterType eq 'NonExclusive')
        {
          FUNDID:
            foreach my $fundId (sort { $a <=> $b } keys %fundConstraints)
            {
                my %constraintRejected = ();
                foreach my $filterName (sort keys %filters)
                {
                    my $filterValue = $filters{$filterName};
                    my $accepted    = 0;
                    my $rejected    = 0;

                    foreach my $constraintName (
                        sort keys %{$fundConstraints{$fundId}})
                    {
                        foreach my $constraintVal (
                            @{$fundConstraints{$fundId}{$constraintName}})
                        {
                            my $constraintValue = $constraintVal;  # Make a copy
                            my $accessAllowed = $constraintValue !~ s/^!//;
                            if (exists $constraintRejected{$constraintName}
                                {$constraintValue})
                            {
                                $rejected = 1;   # This constraint was rejected
                                next;            # Don't allow it to be accepted
                            }

                   # Same Name  |-1|-1|-1| 0| 0| 0|+1|+1|+1|
                   # Other Name |-1| 0|+1|-1| 0|+1|-1| 0|+1|
                   # Accpt/Rjct |R |R |R |  |  |R | A| A|RA|
                   # Result     |X |X |X |/ |/ |X |/ |/ |/ |
                   # A filter should be accepted unless rejected unless accepted

                            # We have a matching constraint name
                            if ($filterName eq $constraintName)
                            {
                                # Mark as accepted if it is a positive
                                # constraint with a matching value
                                if (   $accessAllowed
                                    && $filterValue eq $constraintValue)
                                {
                                    $accepted = 1;
                                }
                                # Mark as rejected if it is a non-matching value
                                elsif (

                                    $accessAllowed
                                    && $filterValue ne $constraintValue
                                  )
                                {
                                    $rejected = 1;
                                    $constraintRejected{$constraintName}
                                      {$constraintValue} = 1;
                                }
                        # Immediately move to limbo if it is explicitly rejected
                                elsif (! $accessAllowed
                                    && $filterValue eq $constraintValue)
                                {
                                    # Move them to limbo
                                    $limbo{$fundId} = $funds{$fundId};
                                    delete $funds{$fundId};
                                    next FUNDID;
                                }
                            }

                 # If the constraint is in a parent association with this filter
                 # check it for membership
                 # We could do some caching here but we are not doing it yet
                            elsif (
                                $accessAllowed
                                && MAM::Cache->associationLookup(
                                    $constraintName, $filterName)
                                && (! exists $fundConstraints{$fundId}
                                    {$filterName} || ! grep ! /!/,
                                    @{$fundConstraints{$fundId}{$filterName}}
                                )
                              )
                            {
                                my ($association) =
                                  MAM::Cache->associationLookup($constraintName,
                                    $filterName);
                                # If there is not a matching member found
                                # reject the fund
                                my @membershipConditions = (
                                    new MAM::Condition(
                                        name =>
                                          'Name',    # Assume simple association
                                        value => $filterValue,
                                        conj  => 'And',
                                        group => '+1',
                                    ),
                                    new MAM::Condition(
                                        name =>
                                          'Name',    # Assume simple association
                                        value => 'ANY',
                                        conj  => 'Or',
                                        group => '-1',
                                    )
                                );
                                push @membershipConditions,
                                  new MAM::Condition(
                                    name  => 'Active',
                                    value => 'True'
                                  )
                                  if MAM::Cache->attributeExists($association,
                                    'Active');
                                my $subResults = $database->select(
                                    object     => $association,
                                    selections => [
                                        new MAM::Selection(
                                            name => $constraintName
                                        )
                                    ],
                                    conditions => \@membershipConditions,
                                );
                                my $keep = 0;
                                foreach my $row (@{$subResults->{data}})
                                {
                                    my $parent = $row->[0];
                                    if ($parent eq $constraintValue)
                                    {
                                        $keep = 1;
                                    }
                                }
                                if ($keep) { $accepted = 1; }
                                else
                                {
                                    $rejected = 1;
                                    $constraintRejected{$constraintName}
                                      {$constraintValue} = 1;
                                }
                            }

                  # If the constraint is in a child association with this filter
                  # check it for membership
                  # We could do some caching here but we are not doing it yet
                            elsif (
                                $accessAllowed
                                && MAM::Cache->associationLookup($filterName,
                                    $constraintName)
                                && !
                                exists $fundConstraints{$fundId}{$filterName}
                              )
                            {
                                my ($association) =
                                  MAM::Cache->associationLookup($filterName,
                                    $constraintName);
                                # If there is not a matching member found
                                # reject the fund
                                my @membershipConditions = (
                                    new MAM::Condition(
                                        name  => $filterName,
                                        value => $filterValue,
                                    )
                                );
                                push @membershipConditions,
                                  new MAM::Condition(
                                    name  => 'Active',
                                    value => 'True'
                                  )
                                  if MAM::Cache->attributeExists($association,
                                    'Active');
                                my $subResults = $database->select(
                                    object => $association,
                                    selections =>
                                      [new MAM::Selection(name => 'Name')]
                                    ,    # Simple name
                                    conditions => \@membershipConditions,
                                );
                                my $keep = 0;
                                foreach my $row (@{$subResults->{data}})
                                {
                                    my $child = $row->[0];
                                    if ($child eq $constraintValue)
                                    {
                                        $keep = 1;
                                    }
                                }
                                if ($keep) { $accepted = 1; }
                                else
                                {
                                    $rejected = 1;
                                    $constraintRejected{$constraintName}
                                      {$constraintValue} = 1;
                                }
                            }
                        }
                    }

                    if ($rejected && ! $accepted)
                    {
                        # Move them to limbo
                        $limbo{$fundId} = $funds{$fundId};
                        delete $funds{$fundId};
                        next FUNDID;
                    }
                }
            }
        }

        elsif ($filterType eq 'Exclusive')
        {
            foreach my $fundId (sort { $a <=> $b } keys %fundConstraints)
            {
                foreach
                  my $constraintName (sort keys %{$fundConstraints{$fundId}})
                {
                    my %accepted = ();
                    my %rejected = ();
                    foreach my $constraintValue (
                        @{$fundConstraints{$fundId}{$constraintName}})
                    {
                        my $accessAllowed = $constraintValue !~ s/^!//;

                        # If there is a filter matching the constraint
                        if (defined $filters{$constraintName}
                            && $filters{$constraintName} eq $constraintValue)
                        {
                            # It's a matching positive constraint
                            if ($accessAllowed)
                            {
                                my @associations =
                                  MAM::Cache->associationLookup($constraintName,
                                    undef);

                                # There are associations
                                if (@associations)
                                {
                                    foreach my $association (@associations)
                                    {
                                        my $child =
                                          MAM::Cache->getObjectProperty(
                                            $association, 'Child');

                                       # Check if the usage record property is a
                                       # parent in an association
                                        if (defined $filters{$child})
                                        {
                                            # Check for a match unless the
                                            # fund has a matching constraint
                                            unless (
                                                exists $fundConstraints{$fundId}
                                                {$child})
                                            {
                             # If there is a matching member found, add the fund
                             # We could improve upon this by caching the results
                                                my @membershipConditions = (
                                                    new MAM::Condition(
                                                        name => $constraintName,
                                                        value => $filters{
                                                            $constraintName},
                                                    ),
                                                    new MAM::Condition(
                                                        name => 'Name'
                                                        , # Assume simple association
                                                        value =>
                                                          $filters{$child},
                                                        conj  => 'And',
                                                        group => '+1',
                                                    ),
                                                    new MAM::Condition(
                                                        name => 'Name'
                                                        , # Assume simple association
                                                        value => 'ANY',
                                                        conj  => 'Or',
                                                        group => '-1',
                                                    ),
                                                );
                                                push @membershipConditions,
                                                  new MAM::Condition(
                                                    name  => 'Active',
                                                    value => 'True'
                                                  )
                                                  if
                                                  MAM::Cache->attributeExists(
                                                    $association, 'Active');
                                                my $subResults =
                                                  $database->select(
                                                    object => $association,
                                                    conditions =>
                                                      \@membershipConditions,
                                                  );
                                                if (@{$subResults->{data}})
                                                {
                                                    $accepted{$constraintName}
                                                      = 1;
                                                }
                                                else
                                                {
                                                    $rejected{$constraintName}
                                                      = 1;
                                                }
                                            }
                                            $accepted{$constraintName} = 1;
                                        }
                                        # Child associations must be specified
                                        # for an Exclusive filter
                                        else
                                        {
                                            return new MAM::Response()
                                              ->failure(
                                                '314',
                                                "A $child must be specified when specifying a $constraintName"
                                              );
                                        }
                                    }
                                }
                                # Accept it if there are no associations
                                else
                                {
                                    $accepted{$constraintName} = 1;
                                }
                            }
                            # Reject the matching negative constraint
                            else
                            {
                                $rejected{$constraintName} = 1;
                            }
                        }
                    # There is not a filter matching the positive constraint
                    # or the filter does not exist or the constraint is negative
                        else
                        {
                            if ($accessAllowed
                                || ! defined $filters{$constraintName})
                            {
                                $rejected{$constraintName} = 1;
                            }
                        }
                    }

                    # Constraints of the same name are Ored while
                    # Constraints of different names are Anded
                    # Send to limbo if there is a constraint name
                    # that is rejected and not accepted
                    foreach my $name (sort keys %rejected)
                    {
                        if (! $accepted{$name} && $funds{$fundId})
                        {
                            # Move them to limbo
                            $limbo{$fundId} = $funds{$fundId};
                            delete $funds{$fundId};
                        }
                    }
                }
            }
        }

        # Then, we build up fund linkages for ancestor funds
        if ($includeAncestors eq 'True')
        {
            my %ancestors = ();
            # SELECT Fund,Id FROM FundFund WHERE Overflow=True
            $results = $database->select(
                object     => 'FundFund',
                selections => [
                    new MAM::Selection(name => 'Fund'),
                    new MAM::Selection(name => 'Id'),
                ],
                conditions => [
                    new MAM::Condition(
                        name  => 'Overflow',
                        value => 'True'
                    )
                ],
            );
            foreach my $row (@{$results->{data}})
            {
                my $parent = $row->[0];
                my $child  = $row->[1];
                $ancestors{$child}{$parent} = 1;
            }

            # Add ancestors a generation at a time until there is no more growth
            while (1)
            {
                my $count = scalar keys %funds;
                foreach my $fundId (sort { $a <=> $b } keys %funds)
                {
                    foreach my $parent (keys %{$ancestors{$fundId}})
                    {
                        if ($limbo{$parent})
                        {
                            # Move it back from limbo
                            $funds{$parent} = $limbo{$parent};
                            delete $limbo{$parent};
                        }
                    }
                }
                last if $count == scalar keys %funds;
            }
        }

        if ($log->is_debug())
        {
            $log->debug('Feasible fund list is: ' . join ',', sort keys %funds);
        }

       # Reissue Allocation Query with selected funds as conditions
       # SELECT @selections FROM Allocation WHERE @conditions AND [Fund=<Id>...]

        # Constrain to list of accepted funds if there are any
        if (scalar keys %funds)
        {
            my $firstTime = 1;
            my $counter   = 0;
            foreach my $id (keys %funds)
            {
                $counter++;
                my $lastTime = ($counter == scalar keys %funds) ? 1 : 0;

                if ($firstTime && $lastTime)
                {
                    push @conditions,
                      new MAM::Condition(
                        name  => 'Fund',
                        value => $id,
                        conj  => 'And',
                        group => '0'
                      );
                }
                elsif ($firstTime)
                {
                    $firstTime = 0;
                    push @conditions,
                      new MAM::Condition(
                        name  => 'Fund',
                        value => $id,
                        conj  => 'And',
                        group => '+1'
                      );
                }
                elsif ($lastTime)
                {
                    push @conditions,
                      new MAM::Condition(
                        name  => 'Fund',
                        value => $id,
                        conj  => 'Or',
                        group => '-1'
                      );
                }
                else
                {
                    push @conditions,
                      new MAM::Condition(
                        name  => 'Fund',
                        value => $id,
                        conj  => 'Or',
                        group => '0'
                      );
                }
            }
        }

        # Otherwise constrain to a bogus fund so that aggregations will work
        else
        {
            push @conditions, new MAM::Condition(name => 'Fund', value => 0);
        }

        $results = $database->select(
            object     => 'Allocation',
            selections => \@selections,
            conditions => \@conditions,
            options    => \@options,
        );
        my $count    = 0;
        my @data     = ();
        my $data     = $results->{data};
        my $cols     = $results->{cols};
        my $names    = $results->{names};
        my %metadata = %{$results->{metadata}};

        # Populate Data
        foreach my $row (@{$data})
        {
            $count++;
            my $datum = new MAM::Datum($object);
            for (my $i = 0; $i < $cols; $i++)
            {
                my $name  = $names->[$i];
                my $value = $row->[$i];     # Null comes in as undef
                $datum->addProperty($name, $value);
            }
            push @data, $datum;
        }

        my $response = new MAM::Response()->success($count, \@data, \%metadata);

        return $response;
    }

    elsif ($object eq 'Fund')
    {
        my $active           = $request->getOptionValue('Active');
        my $includeAncestors = $request->getOptionValue('IncludeAncestors')
          || 'False';
        my $includeDescendants = $request->getOptionValue('IncludeDescendants')
          || 'False';
        my $filterType = $request->getOptionValue('FilterType')
          || 'NonExclusive';

        # There are three fund filter types: Exclusive, NonExclusive and
        # ExactMatch.
        # If an exclusive filter type is used, the query will return only the
        # funds for which the specified filters meet all constraints.
        # For example, Fund Query FilterType:=Exclusive Filter:=User=scottmo
        # would not return a fund with the sole constraint Machine=blue.
        # If a non-exclusive filter type is used, the query will return all
        # funds for which the filters do not specifically exclude the
        # constraints. The query assumes that if constraints are not specified
        # within the filters, they can be assumed as a wildcard and will return
        # all funds that are not specifically excluded by the filter. For
        # example, Fund Query FilterType:=NonExclusive Filter:=User=scottmo
        # would return a fund whose only constraint was Machine=blue
        # but would not return a fund with the constraint User=bob.
        # If an exact-match filter type is used the query will return all
        # funds whose constraints exactly match the specified filters. For
        # example, Fund Query FilterType:=ExactMatch Filter:=Account=chemistry
        #  Filter:=User=amy will only match a fund with the exact constraints
        # Account=chemistry and User=amy.
        if (   $filterType ne 'ExactMatch'
            && $filterType ne 'Exclusive'
            && $filterType ne 'NonExclusive')
        {
            return new MAM::Response()->failure('317',
                'The FilterType option must be one of ExactMatch, Exclusive or NonExclusive'
            );
        }

        # Parse out filters
        my %filters = ();
        foreach my $option (@options)
        {
            if ($option->getName() eq 'Filter')
            {
                my $filter = $option->getValue();
                if ($filter =~ /^([^\s=!]+)=([^\s=]+)$/)
                {
                    $filters{$1} = $2;
                }
            }
        }

        # Override Authorization
        my $authorized = 1;
        if ($request->getOverride())
        {
            $authorized = 0;
        }
        if (   ! $authorized
            && defined $filters{User}
            && $filters{User} eq $actor)
        {
            $authorized = 1;
        }
        if (   ! $authorized
            && defined $filters{Account}
            && MAM::Cache->objectExists('AccountUser'))
        {
# See if actor is account admin
# SELECT Admin FROM AccountUser WHERE Account=$account AND Name=$actor AND Admin=True
            my $subResults = $database->select(
                object     => 'AccountUser',
                selections => [new MAM::Selection(name => 'Admin')],
                conditions => [
                    new MAM::Condition(
                        name  => 'Account',
                        value => $filters{Account},
                    ),
                    new MAM::Condition(
                        name  => 'Name',
                        value => $actor
                    ),
                    new MAM::Condition(
                        name  => 'Admin',
                        value => 'True'
                    )
                ]
            );
            if (@{$subResults->{data}})
            {
                $authorized = 1;
            }
        }
        if (! $authorized)
        {
            if (! defined $filters{User})
            {
                push @options,
                  new MAM::Option(
                    name  => 'Filter',
                    value => "User=$actor"
                  );
                $filters{User} = $actor;
            }
            else
            {
                return new MAM::Response()
                  ->failure('444',
                    "$actor is not authorized to Query the Funds");
            }
        }

        # Quick access to base query if no options are specified
        # This has to occur after the override authorization check
        if (! @options)
        {
            # Invoke base Fund Query
            return MAM::Base->query($request, $requestId, $chunkNum);
        }

        # Verify that every specified usage record filter option exists if it is
        # a value-constrained object and that it is active if it has an active
        # field
        foreach my $filterName (sort keys %filters)
        {
            my $values =
              MAM::Cache->getAttributeProperty('UsageRecord', $filterName,
                'Values');
            if (   defined $values
                && $values =~ /^@(?:[:?!]=)?(\w+)/
                && MAM::Cache->objectExists($1))
            {
                my $filterObject = $1;
                my @primaryKeys =
                  MAM::Cache->listAttributesWithPropertyValue($filterObject,
                    'PrimaryKey', 'True');

                # We can only do this if it has a single primary key
                if (@primaryKeys == 1)
                {
   # SELECT $filterKey[,Active] FROM $filterObject WHERE $filterKey=$filterValue
                    my $filterKey   = $primaryKeys[0];
                    my $filterValue = $filters{$filterName};
                    my @validitySelections =
                      (new MAM::Selection(name => $filterKey));
                    push @validitySelections,
                      new MAM::Selection(name => 'Active')
                      if MAM::Cache->attributeExists($filterObject, 'Active');

                    my @validityConditions = (
                        new MAM::Condition(
                            name  => $filterKey,
                            value => $filterValue
                        )
                    );
                    my $subResults = $database->select(
                        object     => $filterObject,
                        selections => \@validitySelections,
                        conditions => \@validityConditions,
                    );
                    unless (@{$subResults->{data}})
                    {
                        return new MAM::Response()->failure('740',
                            "$filterObject ($filterName) $filterValue does not exist"
                        );
                    }
                    elsif (exists ${$subResults->{data}}[0]->[1]
                        && ${$subResults->{data}}[0]->[1] ne 'True')
                    {
                        return new MAM::Response()->failure('740',
                            "$filterObject ($filterName) $filterValue is not active"
                        );
                    }
                }
            }
        }

# First, we build a list of funds
# If the active option is specified we filter them according to their
# association with active or inactive allocations
# We need to do this separately from the constraint query
# because the trivial fund might have no constraints
# SELECT [DISTINCT] [Fund.]Id FROM Fund[,Allocation] [WHERE Allocation.Fund=Fund.Id AND Allocation.Active=$active]
        my %funds          = ();
        my @fundObjects    = (new MAM::Object(name => 'Fund'));
        my @fundSelections = (new MAM::Selection(name => 'Id'));
        my @fundConditions = @conditions;
        my @fundOptions    = ();
        if (defined $active)
        {
            push @fundObjects, new MAM::Object(name => 'Allocation');
            foreach my $selection (@fundSelections)
            {
                $selection->setObject('Fund');
            }
            foreach my $condition (@fundConditions)
            {
                $condition->setObject('Fund');
            }
            push @fundConditions,
              ( new MAM::Condition(
                    object  => 'Fund',
                    name    => 'Id',
                    subject => 'Allocation',
                    value   => 'Fund'
                ),
                new MAM::Condition(
                    object => 'Allocation',
                    name   => 'Active',
                    value  => $active,
                )
              );
            push @fundOptions,
              new MAM::Option(name => 'Unique', value => 'True');
        }
        my $results = $database->select(
            objects    => \@fundObjects,
            selections => \@fundSelections,
            conditions => \@fundConditions,
            options    => \@fundOptions,
        );
        foreach my $row (@{$results->{data}})
        {
            my $id = $row->[0];
            $funds{$id}++;
        }

        # Then, we prune off the funds that violate the constraints to limbo
        my %fundConstraints = ();
        my %limbo           = ();

        # SELECT Id,Fund,Name,Value FROM Constraint
        $results = $database->select(
            object     => 'Constraint',
            selections => [
                new MAM::Selection(name => 'Id'),
                new MAM::Selection(name => 'Fund'),
                new MAM::Selection(name => 'Name'),
                new MAM::Selection(name => 'Value'),
            ],
        );
        foreach my $row (@{$results->{data}})
        {
            my $id     = $row->[0];
            my $fundId = $row->[1];
            my $name   = $row->[2];
            my $value  = $row->[3];

            push @{$fundConstraints{$fundId}{$name}}, $value;
        }

        if ($filterType eq 'ExactMatch')
        {
          FUNDID:
            foreach my $fundId (sort { $a <=> $b } keys %funds)
            {
             # Reject if there are any filters without corresponding constraints
                foreach my $filterName (sort keys %filters)
                {
                    if (   (! defined $fundConstraints{$fundId})
                        || (! defined $fundConstraints{$fundId}{$filterName})
                        || (@{$fundConstraints{$fundId}{$filterName}} != 1)
                        || (${$fundConstraints{$fundId}{$filterName}}[0] ne
                            $filters{$filterName})
                      )
                    {
                        delete $funds{$fundId};
                        next FUNDID;
                    }
                }

             # Reject if there are any constraints without corresponding filters
                next unless exists $fundConstraints{$fundId};
                foreach
                  my $constraintName (sort keys %{$fundConstraints{$fundId}})
                {
                    if (   (! defined $filters{$constraintName})
                        || (@{$fundConstraints{$fundId}{$constraintName}} != 1)
                        || (${$fundConstraints{$fundId}{$constraintName}}[0] ne
                            $filters{$constraintName})
                      )
                    {
                        delete $funds{$fundId};
                        next FUNDID;
                    }
                }
            }
        }

        elsif ($filterType eq 'NonExclusive')
        {
          FUNDID:
            foreach my $fundId (sort { $a <=> $b } keys %fundConstraints)
            {
                my %constraintRejected = ();
                foreach my $filterName (sort keys %filters)
                {
                    my $filterValue = $filters{$filterName};
                    my $accepted    = 0;
                    my $rejected    = 0;

                    foreach my $constraintName (
                        sort keys %{$fundConstraints{$fundId}})
                    {
                        foreach my $constraintVal (
                            @{$fundConstraints{$fundId}{$constraintName}})
                        {
                            my $constraintValue = $constraintVal;  # Make a copy
                            my $accessAllowed = $constraintValue !~ s/^!//;
                            if (exists $constraintRejected{$constraintName}
                                {$constraintValue})
                            {
                                $rejected = 1;   # This constraint was rejected
                                next;            # Don't allow it to be accepted
                            }

                   # Same Name  |-1|-1|-1| 0| 0| 0|+1|+1|+1|
                   # Other Name |-1| 0|+1|-1| 0|+1|-1| 0|+1|
                   # Accpt/Rjct |R |R |R |  |  |R | A| A|RA|
                   # Result     |X |X |X |/ |/ |X |/ |/ |/ |
                   # A filter should be accepted unless rejected unless accepted

                            # We have a matching constraint name
                            if ($filterName eq $constraintName)
                            {
                                # Mark as accepted if it is a positive
                                # constraint with a matching value
                                if (   $accessAllowed
                                    && $filterValue eq $constraintValue)
                                {
                                    $accepted = 1;
                                }
                                # Mark as rejected if it is a non-matching value
                                elsif (

                                    $accessAllowed
                                    && $filterValue ne $constraintValue
                                  )
                                {
                                    $rejected = 1;
                                    $constraintRejected{$constraintName}
                                      {$constraintValue} = 1;
                                }
                        # Immediately move to limbo if it is explicitly rejected
                                elsif (! $accessAllowed
                                    && $filterValue eq $constraintValue)
                                {
                                    # Move them to limbo
                                    $limbo{$fundId} = $funds{$fundId};
                                    delete $funds{$fundId};
                                    next FUNDID;
                                }
                            }

                 # If the constraint is in a parent association with this filter
                 # check it for membership
                 # We could do some caching here but we are not doing it yet
                            elsif (
                                $accessAllowed
                                && MAM::Cache->associationLookup(
                                    $constraintName, $filterName)
                                && (! exists $fundConstraints{$fundId}
                                    {$filterName} || ! grep ! /!/,
                                    @{$fundConstraints{$fundId}{$filterName}}
                                )
                              )
                            {
                                my ($association) =
                                  MAM::Cache->associationLookup($constraintName,
                                    $filterName);
                                # If there is not a matching member found
                                # reject the fund
                                my @membershipConditions = (
                                    new MAM::Condition(
                                        name =>
                                          'Name',    # Assume simple association
                                        value => $filterValue,
                                        conj  => 'And',
                                        group => '+1',
                                    ),
                                    new MAM::Condition(
                                        name =>
                                          'Name',    # Assume simple association
                                        value => 'ANY',
                                        conj  => 'Or',
                                        group => '-1',
                                    )
                                );
                                push @membershipConditions,
                                  new MAM::Condition(
                                    name  => 'Active',
                                    value => 'True'
                                  )
                                  if MAM::Cache->attributeExists($association,
                                    'Active');
                                my $subResults = $database->select(
                                    object     => $association,
                                    selections => [
                                        new MAM::Selection(
                                            name => $constraintName
                                        )
                                    ],
                                    conditions => \@membershipConditions,
                                );
                                my $keep = 0;
                                foreach my $row (@{$subResults->{data}})
                                {
                                    my $parent = $row->[0];
                                    if ($parent eq $constraintValue)
                                    {
                                        $keep = 1;
                                    }
                                }
                                if ($keep) { $accepted = 1; }
                                else
                                {
                                    $rejected = 1;
                                    $constraintRejected{$constraintName}
                                      {$constraintValue} = 1;
                                }
                            }

                  # If the constraint is in a child association with this filter
                  # check it for membership
                  # We could do some caching here but we are not doing it yet
                            elsif (
                                $accessAllowed
                                && MAM::Cache->associationLookup($filterName,
                                    $constraintName)
                                && !
                                exists $fundConstraints{$fundId}{$filterName}
                              )
                            {
                                my ($association) =
                                  MAM::Cache->associationLookup($filterName,
                                    $constraintName);
                                # If there is not a matching member found
                                # reject the fund
                                my @membershipConditions = (
                                    new MAM::Condition(
                                        name  => $filterName,
                                        value => $filterValue,
                                    )
                                );
                                push @membershipConditions,
                                  new MAM::Condition(
                                    name  => 'Active',
                                    value => 'True'
                                  )
                                  if MAM::Cache->attributeExists($association,
                                    'Active');
                                my $subResults = $database->select(
                                    object => $association,
                                    selections =>
                                      [new MAM::Selection(name => 'Name')]
                                    ,    # Simple name
                                    conditions => \@membershipConditions,
                                );
                                my $keep = 0;
                                foreach my $row (@{$subResults->{data}})
                                {
                                    my $child = $row->[0];
                                    if ($child eq $constraintValue)
                                    {
                                        $keep = 1;
                                    }
                                }
                                if ($keep) { $accepted = 1; }
                                else
                                {
                                    $rejected = 1;
                                    $constraintRejected{$constraintName}
                                      {$constraintValue} = 1;
                                }
                            }
                        }
                    }

                    if ($rejected && ! $accepted)
                    {
                        # Move them to limbo
                        $limbo{$fundId} = $funds{$fundId};
                        delete $funds{$fundId};
                        next FUNDID;
                    }
                }
            }
        }

        elsif ($filterType eq 'Exclusive')
        {
            foreach my $fundId (sort { $a <=> $b } keys %fundConstraints)
            {
                foreach
                  my $constraintName (sort keys %{$fundConstraints{$fundId}})
                {
                    my %accepted = ();
                    my %rejected = ();
                    foreach my $constraintValue (
                        @{$fundConstraints{$fundId}{$constraintName}})
                    {
                        my $accessAllowed = $constraintValue !~ s/^!//;

                        # If there is a filter matching the constraint
                        if (defined $filters{$constraintName}
                            && $filters{$constraintName} eq $constraintValue)
                        {
                            # It's a matching positive constraint
                            if ($accessAllowed)
                            {
                                my @associations =
                                  MAM::Cache->associationLookup($constraintName,
                                    undef);

                                # There are associations
                                if (@associations)
                                {
                                    foreach my $association (@associations)
                                    {
                                        my $child =
                                          MAM::Cache->getObjectProperty(
                                            $association, 'Child');
                                       # Check if the usage record property is a
                                       # parent in an association
                                        if (defined $filters{$child})
                                        {
                                            # Check for a match unless the
                                            # fund has a matching constraint
                                            unless (
                                                exists $fundConstraints{$fundId}
                                                {$child})
                                            {
                             # If there is a matching member found, add the fund
                             # We could improve upon this by caching the results
                                                my @membershipConditions = (
                                                    new MAM::Condition(
                                                        name => $constraintName,
                                                        value => $filters{
                                                            $constraintName},
                                                    ),
                                                    new MAM::Condition(
                                                        name => 'Name'
                                                        , # Assume simple association
                                                        value =>
                                                          $filters{$child},
                                                        conj  => 'And',
                                                        group => '+1',
                                                    ),
                                                    new MAM::Condition(
                                                        name => 'Name'
                                                        , # Assume simple association
                                                        value => 'ANY',
                                                        conj  => 'Or',
                                                        group => '-1',
                                                    ),
                                                );
                                                push @membershipConditions,
                                                  new MAM::Condition(
                                                    name  => 'Active',
                                                    value => 'True'
                                                  )
                                                  if
                                                  MAM::Cache->attributeExists(
                                                    $association, 'Active');
                                                my $subResults =
                                                  $database->select(
                                                    object => $association,
                                                    conditions =>
                                                      \@membershipConditions,
                                                  );
                                                if (@{$subResults->{data}})
                                                {
                                                    $accepted{$constraintName}
                                                      = 1;
                                                }
                                                else
                                                {
                                                    $rejected{$constraintName}
                                                      = 1;
                                                }
                                            }
                                            $accepted{$constraintName} = 1;
                                        }
                                        # Child associations must be specified
                                        # for an Exclusive filter
                                        else
                                        {
                                            return new MAM::Response()
                                              ->failure(
                                                '314',
                                                "A $child must be specified when specifying a $constraintName"
                                              );
                                        }
                                    }
                                }
                                # Accept it if there are no associations
                                else
                                {
                                    $accepted{$constraintName} = 1;
                                }
                            }
                            # Reject the matching negative constraint
                            else
                            {
                                $rejected{$constraintName} = 1;
                            }
                        }
                    # There is not a filter matching the positive constraint
                    # or the filter does not exist or the constraint is negative
                        else
                        {
                            if ($accessAllowed
                                || ! defined $filters{$constraintName})
                            {
                                $rejected{$constraintName} = 1;
                            }
                        }
                    }

                    # Constraints of the same name are Ored while
                    # Constraints of different names are Anded
                    # Send to limbo if there is a constraint name
                    # that is rejected and not accepted
                    foreach my $name (sort keys %rejected)
                    {
                        if (! $accepted{$name} && $funds{$fundId})
                        {
                            # Move them to limbo
                            $limbo{$fundId} = $funds{$fundId};
                            delete $funds{$fundId};
                        }
                    }
                }
            }
        }

        # Add descendants a generation at a time until there is no more growth
        if ($includeDescendants eq 'True')
        {
            my %descendants = ();

            # Build up fund linkages for descendant funds
            # SELECT Fund,Id FROM FundFund WHERE Overflow=True
            $results = $database->select(
                object     => 'FundFund',
                selections => [
                    new MAM::Selection(name => 'Fund'),
                    new MAM::Selection(name => 'Id'),
                ],
                conditions => [
                    new MAM::Condition(
                        name  => 'Overflow',
                        value => 'True'
                    )
                ],
            );
            foreach my $row (@{$results->{data}})
            {
                my $parent = $row->[0];
                my $child  = $row->[1];
                $descendants{$parent}{$child} = 1;
            }

            while (1)
            {
                my $count = scalar keys %funds;
                foreach my $fundId (sort { $a <=> $b } keys %funds)
                {
                    foreach my $child (keys %{$descendants{$fundId}})
                    {
                        if ($limbo{$child})
                        {
                            $funds{$child} = $limbo{$child};
                            delete $limbo{$child};
                        }
                    }
                }
                last if $count == scalar keys %funds;
            }
        }

        # Add ancestors a generation at a time until there is no more growth
        if ($includeAncestors eq 'True')
        {
            my %ancestors = ();

            # Build up fund linkages for ancestor funds
            # SELECT Fund,Id FROM FundFund WHERE Overflow=True
            $results = $database->select(
                object     => 'FundFund',
                selections => [
                    new MAM::Selection(name => 'Fund'),
                    new MAM::Selection(name => 'Id'),
                ],
                conditions => [
                    new MAM::Condition(
                        name  => 'Overflow',
                        value => 'True'
                    )
                ],
            );
            foreach my $row (@{$results->{data}})
            {
                my $parent = $row->[0];
                my $child  = $row->[1];
                $ancestors{$child}{$parent} = 1;
            }

            while (1)
            {
                my $count = scalar keys %funds;
                foreach my $fundId (sort { $a <=> $b } keys %funds)
                {
                    foreach my $parent (keys %{$ancestors{$fundId}})
                    {
                        if ($limbo{$parent})
                        {
                            $funds{$parent} = $limbo{$parent};
                            delete $limbo{$parent};
                        }
                    }
                }
                last if $count == scalar keys %funds;
            }
        }

        if ($log->is_debug())
        {
            $log->debug('Feasible fund list is: ' . join ',', sort keys %funds);
        }

        # Perform selected query over list of funds
        # SELECT <selections> FROM Fund WHERE (Id=<Id> ...)
        my @subConditions = ();

        # Constrain to list of accepted ids if there are any
        if (scalar keys %funds)
        {
            my $firstTime = 1;
            my $counter   = 0;
            foreach my $id (keys %funds)
            {
                $counter++;
                my $lastTime = ($counter == scalar keys %funds) ? 1 : 0;

                if ($firstTime && $lastTime)
                {
                    push @subConditions,
                      new MAM::Condition(
                        name  => 'Id',
                        value => $id,
                        conj  => 'And',
                        group => '0'
                      );
                }
                elsif ($firstTime)
                {
                    $firstTime = 0;
                    push @subConditions,
                      new MAM::Condition(
                        name  => 'Id',
                        value => $id,
                        conj  => 'And',
                        group => '+1'
                      );
                }
                elsif ($lastTime)
                {
                    push @subConditions,
                      new MAM::Condition(
                        name  => 'Id',
                        value => $id,
                        conj  => 'Or',
                        group => '-1'
                      );
                }
                else
                {
                    push @subConditions,
                      new MAM::Condition(
                        name  => 'Id',
                        value => $id,
                        conj  => 'Or',
                        group => '0'
                      );
                }
            }
        }

        # Otherwise constrain to a bogus id so that aggregations will work
        else
        {
            push @subConditions, new MAM::Condition(name => 'Id', value => 0);
        }

        $results = $database->select(
            object     => 'Fund',
            selections => \@selections,
            conditions => \@subConditions,
            options    => \@options
        );
        my $count    = 0;
        my @data     = ();
        my $data     = $results->{data};
        my $cols     = $results->{cols};
        my $names    = $results->{names};
        my %metadata = %{$results->{metadata}};

        # Populate Data
        foreach my $row (@{$data})
        {
            $count++;
            my $datum = new MAM::Datum($object);
            for (my $i = 0; $i < $cols; $i++)
            {
                my $name  = $names->[$i];
                my $value = $row->[$i];     # Null comes in as undef
                $datum->addProperty($name, $value);
            }
            push @data, $datum;
        }

        my $response = new MAM::Response()->success($count, \@data, \%metadata);

        return $response;
    }

    elsif ($object eq 'Lien')
    {
        my $active          = $request->getOptionValue('Active');
        my $ignoreAncestors = $request->getOptionValue('IgnoreAncestors')
          || 'False';
        my $filterType = $request->getOptionValue('FilterType')
          || 'AttributedTo';

        # There are two filter types: AttributedTo and ImpingesUpon
        # If ImpingesUpon is used, the query will return all liens
        # associated with funds satisfying the filters. For example,
        # Lien Query FilterType:=ImpingesUpon Filter:=User=scottmo
        # will return all liens impinging on Funds usable by scottmo.
        # If AttributedTo is used, the query will return all liens
        # associated with usage records satisfying the filters. For example,
        # Lien Query FilterType:=AttributedTo Filter:=User=scottmo will
        # return all liens with usage records associated with scottmo.
        if (   $filterType ne 'AttributedTo'
            && $filterType ne 'ImpingesUpon')
        {
            return new MAM::Response()->failure('317',
                'The FilterType option must be either AttributedTo or ImpingesUpon'
            );
        }

        # Parse out filters
        my %filters = ();
        foreach my $option (@options)
        {
            if ($option->getName() eq 'Filter')
            {
                my $filter = $option->getValue();
                if ($filter =~ /^([^\s=!]+)=([^\s=]+)$/)
                {
                    $filters{$1} = $2;
                }
            }
        }

        # Override Authorization
        my $authorized = 1;
        if ($request->getOverride())
        {
            $authorized = 0;
        }
        if (   ! $authorized
            && defined $filters{User}
            && $filters{User} eq $actor)
        {
            $authorized = 1;
        }
        if (   ! $authorized
            && defined $filters{Account}
            && MAM::Cache->objectExists('AccountUser'))
        {
# See if actor is account admin
# SELECT Admin FROM AccountUser WHERE Account=$filters{Account} AND Name=$actor AND Admin=True
            my $subResults = $database->select(
                object     => 'AccountUser',
                selections => [new MAM::Selection(name => 'Admin')],
                conditions => [
                    new MAM::Condition(
                        name  => 'Account',
                        value => $filters{Account},
                    ),
                    new MAM::Condition(
                        name  => 'Name',
                        value => $actor
                    ),
                    new MAM::Condition(
                        name  => 'Admin',
                        value => 'True'
                    )
                ]
            );
            if (@{$subResults->{data}})
            {
                $authorized = 1;
            }
        }
        if (! $authorized)
        {
            if (! defined $filters{User})
            {
                push @options,
                  new MAM::Option(
                    name  => 'Filter',
                    value => "User=$actor"
                  );
                $filters{User} = $actor;
            }
            else
            {
                return new MAM::Response()
                  ->failure('444',
                    "$actor is not authorized to Query the Liens");
            }
        }

        # Quick access to base query if no options are specified
        # Also pass all join queries through since they cannot be handled here
        if (@objects > 1 || ! @options)
        {
            # Invoke base Lien Query
            return MAM::Base->query($request, $requestId, $chunkNum);
        }

        # If filters are specified and filterType is ImpingesUpon, we need
        # to calculate all liens that impinge on the entities' balance
        my %funds = ();
        if ($filterType eq 'ImpingesUpon' && scalar keys %filters)
        {
# First, we build a list of funds from a join of Liens,
# LienAllocations and Allocations so we narrow this down to just the
# feasible list of funds to filter further by constraint
# We need to do this separately from the constraint query
# because the trivial fund might have no constraints
# SELECT DISTINCT Allocation.Fund FROM Lien,LienAllocation,Allocation WHERE Lien.Id=LienAllocation.Lien AND LienAllocation.Id=Allocation.Id
            my $results = $database->select(
                objects => [
                    new MAM::Object(name => 'Lien'),
                    new MAM::Object(name => 'LienAllocation'),
                    new MAM::Object(name => 'Allocation'),
                ],
                selections => [
                    new MAM::Selection(
                        object => 'Allocation',
                        name   => 'Fund'
                    )
                ],
                conditions => [
                    new MAM::Condition(
                        object  => 'Lien',
                        name    => 'Id',
                        subject => 'LienAllocation',
                        value   => 'Lien',
                    ),
                    new MAM::Condition(
                        object  => 'LienAllocation',
                        name    => 'Id',
                        subject => 'Allocation',
                        value   => 'Id',
                    ),
                ],
                options => [new MAM::Option(name => 'Unique', value => 'True')],
            );
            foreach my $row (@{$results->{data}})
            {
                my $id = $row->[0];
                $funds{$id}++;
            }

            # Then, we prune off the funds that violate the constraints to limbo
            my %fundConstraints = ();
            my %limbo           = ();

            # SELECT Id,Fund,Name,Value FROM Constraint WHERE [Fund=<Id>...]
            # Constrain to list of accepted funds
            my @constraintConditions = ();
            my $firstTime            = 1;
            my $counter              = 0;
            foreach my $id (keys %funds)
            {
                $counter++;
                my $lastTime = ($counter == scalar keys %funds) ? 1 : 0;

                if ($firstTime && $lastTime)
                {
                    push @constraintConditions,
                      new MAM::Condition(
                        name  => 'Fund',
                        value => $id,
                        conj  => 'And',
                        group => '0'
                      );
                }
                elsif ($firstTime)
                {
                    $firstTime = 0;
                    push @constraintConditions,
                      new MAM::Condition(
                        name  => 'Fund',
                        value => $id,
                        conj  => 'And',
                        group => '+1'
                      );
                }
                elsif ($lastTime)
                {
                    push @constraintConditions,
                      new MAM::Condition(
                        name  => 'Fund',
                        value => $id,
                        conj  => 'Or',
                        group => '-1'
                      );
                }
                else
                {
                    push @constraintConditions,
                      new MAM::Condition(
                        name  => 'Fund',
                        value => $id,
                        conj  => 'Or',
                        group => '0'
                      );
                }
            }

            $results = $database->select(
                object     => 'Constraint',
                selections => [
                    new MAM::Selection(name => 'Id'),
                    new MAM::Selection(name => 'Fund'),
                    new MAM::Selection(name => 'Name'),
                    new MAM::Selection(name => 'Value'),
                ],
                conditions => \@constraintConditions,
            );
            foreach my $row (@{$results->{data}})
            {
                my $id     = $row->[0];
                my $fundId = $row->[1];
                my $name   = $row->[2];
                my $value  = $row->[3];

                push @{$fundConstraints{$fundId}{$name}}, $value;
            }

          FUNDID:
            foreach my $fundId (sort { $a <=> $b } keys %fundConstraints)
            {
                my %constraintRejected = ();
                foreach my $filterName (sort keys %filters)
                {
                    my $filterValue = $filters{$filterName};
                    my $accepted    = 0;
                    my $rejected    = 0;

                    foreach my $constraintName (
                        sort keys %{$fundConstraints{$fundId}})
                    {
                        foreach my $constraintVal (
                            @{$fundConstraints{$fundId}{$constraintName}})
                        {
                            my $constraintValue = $constraintVal;  # Make a copy
                            my $accessAllowed = $constraintValue !~ s/^!//;
                            if (exists $constraintRejected{$constraintName}
                                {$constraintValue})
                            {
                                $rejected = 1;   # This constraint was rejected
                                next;            # Don't allow it to be accepted
                            }

                   # Same Name  |-1|-1|-1| 0| 0| 0|+1|+1|+1|
                   # Other Name |-1| 0|+1|-1| 0|+1|-1| 0|+1|
                   # Accpt/Rjct |R |R |R |  |  |R | A| A|RA|
                   # Result     |X |X |X |/ |/ |X |/ |/ |/ |
                   # A filter should be accepted unless rejected unless accepted

                            # We have a matching constraint name
                            if ($filterName eq $constraintName)
                            {
                                # Mark as accepted if it is a positive
                                # constraint with a matching value
                                if (   $accessAllowed
                                    && $filterValue eq $constraintValue)
                                {
                                    $accepted = 1;
                                }
                                # Mark as rejected if it is a non-matching value
                                elsif (

                                    $accessAllowed
                                    && $filterValue ne $constraintValue
                                  )
                                {
                                    $rejected = 1;
                                    $constraintRejected{$constraintName}
                                      {$constraintValue} = 1;
                                }
                        # Immediately move to limbo if it is explicitly rejected
                                elsif (! $accessAllowed
                                    && $filterValue eq $constraintValue)
                                {
                                    # Move them to limbo
                                    $limbo{$fundId} = $funds{$fundId};
                                    delete $funds{$fundId};
                                    next FUNDID;
                                }
                            }

                 # If the constraint is in a parent association with this filter
                 # check it for membership
                 # We could do some caching here but we are not doing it yet
                            elsif (
                                $accessAllowed
                                && MAM::Cache->associationLookup(
                                    $constraintName, $filterName)
                                && (! exists $fundConstraints{$fundId}
                                    {$filterName} || ! grep ! /!/,
                                    @{$fundConstraints{$fundId}{$filterName}}
                                )
                              )
                            {
                                my ($association) =
                                  MAM::Cache->associationLookup($constraintName,
                                    $filterName);
                                # If there is not a matching member found
                                # reject the fund
                                my @membershipConditions = (
                                    new MAM::Condition(
                                        name =>
                                          'Name',    # Assume simple association
                                        value => $filterValue,
                                        conj  => 'And',
                                        group => '+1',
                                    ),
                                    new MAM::Condition(
                                        name =>
                                          'Name',    # Assume simple association
                                        value => 'ANY',
                                        conj  => 'Or',
                                        group => '-1',
                                    )
                                );
                                push @membershipConditions,
                                  new MAM::Condition(
                                    name  => 'Active',
                                    value => 'True'
                                  )
                                  if MAM::Cache->attributeExists($association,
                                    'Active');
                                my $subResults = $database->select(
                                    object     => $association,
                                    selections => [
                                        new MAM::Selection(
                                            name => $constraintName
                                        )
                                    ],
                                    conditions => \@membershipConditions,
                                );
                                my $keep = 0;
                                foreach my $row (@{$subResults->{data}})
                                {
                                    my $parent = $row->[0];
                                    if ($parent eq $constraintValue)
                                    {
                                        $keep = 1;
                                    }
                                }
                                if ($keep) { $accepted = 1; }
                                else
                                {
                                    $rejected = 1;
                                    $constraintRejected{$constraintName}
                                      {$constraintValue} = 1;
                                }
                            }

                  # If the constraint is in a child association with this filter
                  # check it for membership
                  # We could do some caching here but we are not doing it yet
                            elsif (
                                $accessAllowed
                                && MAM::Cache->associationLookup($filterName,
                                    $constraintName)
                                && !
                                exists $fundConstraints{$fundId}{$filterName}
                              )
                            {
                                my ($association) =
                                  MAM::Cache->associationLookup($filterName,
                                    $constraintName);
                                # If there is not a matching member found
                                # reject the fund
                                my @membershipConditions = (
                                    new MAM::Condition(
                                        name  => $filterName,
                                        value => $filterValue,
                                    )
                                );
                                push @membershipConditions,
                                  new MAM::Condition(
                                    name  => 'Active',
                                    value => 'True'
                                  )
                                  if MAM::Cache->attributeExists($association,
                                    'Active');
                                my $subResults = $database->select(
                                    object => $association,
                                    selections =>
                                      [new MAM::Selection(name => 'Name')]
                                    ,    # Simple name
                                    conditions => \@membershipConditions,
                                );
                                my $keep = 0;
                                foreach my $row (@{$subResults->{data}})
                                {
                                    my $child = $row->[0];
                                    if ($child eq $constraintValue)
                                    {
                                        $keep = 1;
                                    }
                                }
                                if ($keep) { $accepted = 1; }
                                else
                                {
                                    $rejected = 1;
                                    $constraintRejected{$constraintName}
                                      {$constraintValue} = 1;
                                }
                            }
                        }
                    }

                    if ($rejected && ! $accepted)
                    {
                        # Move them to limbo
                        $limbo{$fundId} = $funds{$fundId};
                        delete $funds{$fundId};
                        next FUNDID;
                    }
                }
            }

            # Then, we build up fund linkages for ancestor funds
            if ($ignoreAncestors ne 'True')
            {
                my %ancestors = ();

                # Build up fund linkages for ancestor funds
                # SELECT Fund,Id FROM FundFund WHERE Overflow=True
                $results = $database->select(
                    object     => 'FundFund',
                    selections => [
                        new MAM::Selection(name => 'Fund'),
                        new MAM::Selection(name => 'Id'),
                    ],
                    conditions => [
                        new MAM::Condition(
                            name  => 'Overflow',
                            value => 'True'
                        )
                    ],
                );
                foreach my $row (@{$results->{data}})
                {
                    my $parent = $row->[0];
                    my $child  = $row->[1];
                    $ancestors{$child}{$parent} = 1;
                }

                while (1)
                {
                    my $count = scalar keys %funds;
                    foreach my $fundId (sort { $a <=> $b } keys %funds)
                    {
                        foreach my $parent (keys %{$ancestors{$fundId}})
                        {
                            if ($limbo{$parent})
                            {
                                $funds{$parent} = $limbo{$parent};
                                delete $limbo{$parent};
                            }
                        }
                    }
                    last if $count == scalar keys %funds;
                }
            }

            if ($log->is_debug())
            {
                $log->debug('Feasible fund list is: ' . join ',',
                    sort keys %funds);
            }

            # Return immediately if there are no matching funds
            if (! scalar keys %funds)
            {
                return new MAM::Response()->success(0, [], {});
            }
        }

# Issue Lien Query with appropriate join conditions
# SELECT [DISTINCT] Lien.@selections FROM Lien[,LienAllocation,Allocation] WHERE [Lien.Id=LienAllocation.Lien AND LienAllocation.Id=Allocation.Id] [AND Allocation.Fund=<Id>...]
        my @lienObjects    = (new MAM::Object(name => 'Lien'));
        my @lienSelections = @selections;
        my @lienConditions = @conditions;
        my @lienOptions    = @options;
        if (defined $active)
        {
            if ($active eq 'True')
            {
                push @lienConditions,
                  new MAM::Condition(
                    name  => 'StartTime',
                    value => $now,
                    op    => 'LE'
                  );
                push @lienConditions,
                  new MAM::Condition(
                    name  => 'EndTime',
                    value => $now,
                    op    => 'GT'
                  );
            }
            elsif ($active eq 'False')
            {
                push @lienConditions,
                  new MAM::Condition(
                    name  => 'StartTime',
                    value => $now,
                    op    => 'GT',
                    conj  => 'And',
                    group => '+1'
                  );
                push @lienConditions,
                  new MAM::Condition(
                    name  => 'EndTime',
                    value => $now,
                    op    => 'LE',
                    conj  => 'Or',
                    group => '-1'
                  );
            }
        }
        if ($filterType eq 'ImpingesUpon' && scalar keys %funds)
        {
            push @lienObjects,
              ( new MAM::Object(name => 'LienAllocation'),
                new MAM::Object(name => 'Allocation')
              );
            foreach my $selection (@lienSelections)
            {
                $selection->setObject('Lien');
            }
            foreach my $condition (@lienConditions)
            {
                $condition->setObject('Lien');
            }
            push @lienConditions,
              ( new MAM::Condition(
                    object  => 'Lien',
                    name    => 'Id',
                    subject => 'LienAllocation',
                    value   => 'Lien',
                ),
                new MAM::Condition(
                    object  => 'LienAllocation',
                    name    => 'Id',
                    subject => 'Allocation',
                    value   => 'Id',
                ),
              );
            push @lienOptions,
              new MAM::Option(name => 'Unique', value => 'True');

            # Constrain to list of accepted funds
            my $firstTime = 1;
            my $counter   = 0;
            foreach my $id (keys %funds)
            {
                $counter++;
                my $lastTime = ($counter == scalar keys %funds) ? 1 : 0;

                if ($firstTime && $lastTime)
                {
                    push @lienConditions,
                      new MAM::Condition(
                        object => 'Allocation',
                        name   => 'Fund',
                        value  => $id,
                        conj   => 'And',
                        group  => '0'
                      );
                }
                elsif ($firstTime)
                {
                    $firstTime = 0;
                    push @lienConditions,
                      new MAM::Condition(
                        object => 'Allocation',
                        name   => 'Fund',
                        value  => $id,
                        conj   => 'And',
                        group  => '+1'
                      );
                }
                elsif ($lastTime)
                {
                    push @lienConditions,
                      new MAM::Condition(
                        object => 'Allocation',
                        name   => 'Fund',
                        value  => $id,
                        conj   => 'Or',
                        group  => '-1'
                      );
                }
                else
                {
                    push @lienConditions,
                      new MAM::Condition(
                        object => 'Allocation',
                        name   => 'Fund',
                        value  => $id,
                        conj   => 'Or',
                        group  => '0'
                      );
                }
            }
        }
        elsif ($filterType eq 'AttributedTo' && scalar keys %filters)
        {
            push @lienObjects, (new MAM::Object(name => 'UsageRecord'));
            foreach my $selection (@lienSelections)
            {
                $selection->setObject('Lien');
            }
            foreach my $condition (@lienConditions)
            {
                $condition->setObject('Lien');
            }
            push @lienConditions,
              ( new MAM::Condition(
                    object  => 'Lien',
                    name    => 'UsageRecord',
                    subject => 'UsageRecord',
                    value   => 'Id',
                )
              );
            push @lienOptions,
              new MAM::Option(name => 'Unique', value => 'True');

            # Constrain to filter conditions
            foreach my $filterName (keys %filters)
            {
                my $filterValue = $filters{$filterName};
                push @lienConditions,
                  new MAM::Condition(
                    object => 'UsageRecord',
                    name   => $filterName,
                    value  => $filterValue,
                  );
            }
        }

        my $results = $database->select(
            objects    => \@lienObjects,
            selections => \@lienSelections,
            conditions => \@lienConditions,
            options    => \@lienOptions,
        );
        my $count    = 0;
        my @data     = ();
        my $data     = $results->{data};
        my $cols     = $results->{cols};
        my $names    = $results->{names};
        my %metadata = %{$results->{metadata}};

        # Populate Data
        foreach my $row (@{$data})
        {
            $count++;
            my $datum = new MAM::Datum($object);
            for (my $i = 0; $i < $cols; $i++)
            {
                my $name  = $names->[$i];
                my $value = $row->[$i];     # Null comes in as undef
                $datum->addProperty($name, $value);
            }
            push @data, $datum;
        }

        my $response = new MAM::Response()->success($count, \@data, \%metadata);

        return $response;

        # Former return style
        # If the results come back looking funny because they are joins,
        # we can come back and do this style via a list of lien ids
        ## Invoke base Lien Query
        #return MAM::Base->query($request, $requestId, $chunkNum);
    }

    elsif ($object eq 'Payment')
    {
        my $active = $request->getOptionValue('Active');

        # Parse out filters
        my %filters = ();
        foreach my $option (@options)
        {
            if ($option->getName() eq 'Filter')
            {
                my $filter = $option->getValue();
                if ($filter =~ /^([^\s=!]+)=([^\s=]+)$/)
                {
                    $filters{$1} = $2;
                }
            }
        }

        # Override Authorization
        my $authorized = 1;
        if ($request->getOverride())
        {
            $authorized = 0;
        }
        if (   ! $authorized
            && defined $filters{User}
            && $filters{User} eq $actor)
        {
            $authorized = 1;
        }
        if (   ! $authorized
            && defined $filters{Account}
            && MAM::Cache->objectExists('AccountUser'))
        {
# See if actor is account admin
# SELECT Admin FROM AccountUser WHERE Account=$filters{Account} AND Name=$actor AND Admin=True
            my $subResults = $database->select(
                object     => 'AccountUser',
                selections => [new MAM::Selection(name => 'Admin')],
                conditions => [
                    new MAM::Condition(
                        name  => 'Account',
                        value => $filters{Account},
                    ),
                    new MAM::Condition(
                        name  => 'Name',
                        value => $actor
                    ),
                    new MAM::Condition(
                        name  => 'Admin',
                        value => 'True'
                    )
                ]
            );
            if (@{$subResults->{data}})
            {
                $authorized = 1;
            }
        }
        if (! $authorized)
        {
            if (! defined $filters{User})
            {
                push @options,
                  new MAM::Option(
                    name  => 'Filter',
                    value => "User=$actor"
                  );
                $filters{User} = $actor;
            }
            else
            {
                return new MAM::Response()
                  ->failure('444',
                    "$actor is not authorized to Query the Payments");
            }
        }

        # Quick access to base query if no options are specified
        if (! @options)
        {
            # Invoke base Payment Query
            return MAM::Base->query($request, $requestId, $chunkNum);
        }

        # If filters are specified, we need to display all payments that
        # have usage records with the filter condition

# Issue Payment Query with appropriate join conditions
# SELECT [DISTINCT] [Payment.]@selections FROM Payment[,UsageRecord] WHERE [Payment.Id=UsageRecord.Payment] AND [Payment.]@conditions
        my @paymentObjects    = (new MAM::Object(name => 'Payment'));
        my @paymentSelections = @selections;
        my @paymentConditions = @conditions;
        my @paymentOptions    = @options;
        if (defined $active)
        {
            if ($active eq 'True')
            {
                push @paymentConditions,
                  new MAM::Condition(
                    name  => 'BeginTime',
                    value => $now,
                    op    => 'LE'
                  );
                push @paymentConditions,
                  new MAM::Condition(
                    name  => 'EndTime',
                    value => $now,
                    op    => 'GT',
                    conj  => 'And',
                    group => '+1'
                  );
                push @paymentConditions,
                  new MAM::Condition(
                    name  => 'EndTime',
                    value => 'NULL',
                    op    => 'EQ',
                    conj  => 'Or',
                    group => '-1'
                  );
            }
            elsif ($active eq 'False')
            {
                push @paymentConditions,
                  new MAM::Condition(
                    name  => 'BeginTime',
                    value => $now,
                    op    => 'GT',
                    conj  => 'And',
                    group => '+1'
                  );
                push @paymentConditions,
                  new MAM::Condition(
                    name  => 'BeginTime',
                    value => 'NULL',
                    op    => 'EQ',
                    conj  => 'Or'
                  );
                push @paymentConditions,
                  new MAM::Condition(
                    name  => 'EndTime',
                    value => $now,
                    op    => 'LE',
                    conj  => 'Or',
                    group => '-1'
                  );
            }
        }
        if (scalar keys %filters)
        {
            push @paymentObjects, new MAM::Object(name => 'UsageRecord');
            foreach my $selection (@paymentSelections)
            {
                $selection->setObject('Payment');
            }
            foreach my $condition (@paymentConditions)
            {
                $condition->setObject('Payment');
            }
            push @paymentConditions,
              ( new MAM::Condition(
                    object  => 'Payment',
                    name    => 'Id',
                    subject => 'UsageRecord',
                    value   => 'Payment',
                )
              );
            push @paymentOptions,
              new MAM::Option(name => 'Unique', value => 'True');

            # Constrain to filter conditions
            foreach my $filterName (keys %filters)
            {
                my $filterValue = $filters{$filterName};
                push @paymentConditions,
                  ( new MAM::Condition(
                        object => 'UsageRecord',
                        name   => $filterName,
                        value  => $filterValue,
                    )
                  );
            }
        }

        my $results = $database->select(
            objects    => \@paymentObjects,
            selections => \@paymentSelections,
            conditions => \@paymentConditions,
            options    => \@paymentOptions,
        );
        my $count    = 0;
        my @data     = ();
        my $data     = $results->{data};
        my $cols     = $results->{cols};
        my $names    = $results->{names};
        my %metadata = %{$results->{metadata}};

        # Populate Data
        foreach my $row (@{$data})
        {
            $count++;
            my $datum = new MAM::Datum($object);
            for (my $i = 0; $i < $cols; $i++)
            {
                my $name  = $names->[$i];
                my $value = $row->[$i];     # Null comes in as undef
                $datum->addProperty($name, $value);
            }
            push @data, $datum;
        }

        my $response = new MAM::Response()->success($count, \@data, \%metadata);

        return $response;

        # Former return style
        # If the results come back looking funny because they are joins,
        # we can come back and do this style via a list of payment ids
        ## Invoke base Payment Query
        #return MAM::Base->query($request, $requestId, $chunkNum);
    }

    elsif ($object eq 'Quote')
    {
        my $active = $request->getOptionValue('Active');

        # Parse out filters
        my %filters = ();
        foreach my $option (@options)
        {
            if ($option->getName() eq 'Filter')
            {
                my $filter = $option->getValue();
                if ($filter =~ /^([^\s=!]+)=([^\s=]+)$/)
                {
                    $filters{$1} = $2;
                }
            }
        }

        # Override Authorization
        my $authorized = 1;
        if ($request->getOverride())
        {
            $authorized = 0;
        }
        if (   ! $authorized
            && defined $filters{User}
            && $filters{User} eq $actor)
        {
            $authorized = 1;
        }
        if (   ! $authorized
            && defined $filters{Account}
            && MAM::Cache->objectExists('AccountUser'))
        {
# See if actor is account admin
# SELECT Admin FROM AccountUser WHERE Account=$filters{Account} AND Name=$actor AND Admin=True
            my $subResults = $database->select(
                object     => 'AccountUser',
                selections => [new MAM::Selection(name => 'Admin')],
                conditions => [
                    new MAM::Condition(
                        name  => 'Account',
                        value => $filters{Account},
                    ),
                    new MAM::Condition(
                        name  => 'Name',
                        value => $actor
                    ),
                    new MAM::Condition(
                        name  => 'Admin',
                        value => 'True'
                    )
                ]
            );
            if (@{$subResults->{data}})
            {
                $authorized = 1;
            }
        }
        if (! $authorized)
        {
            if (! defined $filters{User})
            {
                push @options,
                  new MAM::Option(
                    name  => 'Filter',
                    value => "User=$actor"
                  );
                $filters{User} = $actor;
            }
            else
            {
                return new MAM::Response()
                  ->failure('444',
                    "$actor is not authorized to Query the Quotes");
            }
        }

        # Quick access to base query if no options are specified
        if (! @options)
        {
            # Invoke base Quote Query
            return MAM::Base->query($request, $requestId, $chunkNum);
        }

        # If filters are specified, we need to display all quotes that
        # have usage records with the filter condition

# Issue Quote Query with appropriate join conditions
# SELECT [DISTINCT] [Quote.]@selections FROM Quote[,UsageRecord] WHERE [Quote.UsageRecord=UsageRecord.Id] AND [Quote.]@conditions
        my @quoteObjects    = (new MAM::Object(name => 'Quote'));
        my @quoteSelections = @selections;
        my @quoteConditions = @conditions;
        my @quoteOptions    = @options;
        if (defined $active)
        {
            if ($active eq 'True')
            {
                push @quoteConditions,
                  new MAM::Condition(
                    name  => 'StartTime',
                    value => $now,
                    op    => 'LE'
                  );
                push @quoteConditions,
                  new MAM::Condition(
                    name  => 'EndTime',
                    value => $now,
                    op    => 'GT'
                  );
            }
            elsif ($active eq 'False')
            {
                push @quoteConditions,
                  new MAM::Condition(
                    name  => 'StartTime',
                    value => $now,
                    op    => 'GT',
                    conj  => 'And',
                    group => '+1'
                  );
                push @quoteConditions,
                  new MAM::Condition(
                    name  => 'EndTime',
                    value => $now,
                    op    => 'LE',
                    conj  => 'Or',
                    group => '-1'
                  );
            }
        }
        if (scalar keys %filters)
        {
            push @quoteObjects, (new MAM::Object(name => 'UsageRecord'),);
            foreach my $selection (@quoteSelections)
            {
                $selection->setObject('Quote');
            }
            foreach my $condition (@quoteConditions)
            {
                $condition->setObject('Quote');
            }
            push @quoteConditions,
              ( new MAM::Condition(
                    object  => 'Quote',
                    name    => 'UsageRecord',
                    subject => 'UsageRecord',
                    value   => 'Id',
                )
              );
            push @quoteOptions,
              new MAM::Option(name => 'Unique', value => 'True');

            # Constrain to filter conditions
            foreach my $filterName (keys %filters)
            {
                my $filterValue = $filters{$filterName};
                push @quoteConditions,
                  ( new MAM::Condition(
                        object => 'UsageRecord',
                        name   => $filterName,
                        value  => $filterValue,
                    )
                  );
            }
        }

        my $results = $database->select(
            objects    => \@quoteObjects,
            selections => \@quoteSelections,
            conditions => \@quoteConditions,
            options    => \@quoteOptions,
        );
        my $count    = 0;
        my @data     = ();
        my $data     = $results->{data};
        my $cols     = $results->{cols};
        my $names    = $results->{names};
        my %metadata = %{$results->{metadata}};

        # Populate Data
        foreach my $row (@{$data})
        {
            $count++;
            my $datum = new MAM::Datum($object);
            for (my $i = 0; $i < $cols; $i++)
            {
                my $name  = $names->[$i];
                my $value = $row->[$i];     # Null comes in as undef
                $datum->addProperty($name, $value);
            }
            push @data, $datum;
        }

        my $response = new MAM::Response()->success($count, \@data, \%metadata);

        return $response;

        # Former return style
        # If the results come back looking funny because they are joins,
        # we can come back and do this style via a list of quote ids
        ## Invoke base Quote Query
        #return MAM::Base->query($request, $requestId, $chunkNum);
    }

    elsif ($object eq 'Transaction')
    {
        my $account   = $request->getConditionValue('Account');
        my $user      = $request->getConditionValue('User');
        my $actorCond = $request->getConditionValue('Actor');

        # Override Authorization
        my $authorized = 1;
        if ($request->getOverride())
        {
            $authorized = 0;
        }
        if (! $authorized
            && (   (defined $user && $actor eq $user)
                || (defined $actorCond && $actor eq $actorCond))
          )
        {
            $authorized = 1;
        }
        if (   ! $authorized
            && defined $account
            && MAM::Cache->objectExists('AccountUser'))
        {
# See if actor is account admin
# SELECT Admin FROM AccountUser WHERE Account=$account AND Name=$actor AND Admin=True
            my $subResults = $database->select(
                object     => 'AccountUser',
                selections => [new MAM::Selection(name => 'Admin')],
                conditions => [
                    new MAM::Condition(
                        name  => 'Account',
                        value => $account
                    ),
                    new MAM::Condition(
                        name  => 'Name',
                        value => $actor
                    ),
                    new MAM::Condition(
                        name  => 'Admin',
                        value => 'True'
                    )
                ]
            );
            if (@{$subResults->{data}})
            {
                $authorized = 1;
            }
        }
        if (! $authorized)
        {
            if (! defined $user && ! defined $actorCond)
            {
                $request->setConditions(
                    [   @conditions,
                        new MAM::Condition(
                            name  => 'User',
                            value => $actor,
                            conj  => 'And',
                            group => '+1'
                        ),
                        new MAM::Condition(
                            name  => 'Actor',
                            value => $actor,
                            conj  => 'Or',
                            group => '-1'
                        )
                    ]
                );
            }
            else
            {
                return new MAM::Response()
                  ->failure('444',
                    "$actor is not authorized to Query the Transactions");
            }
        }

        # Invoke base Transaction Query
        return MAM::Base->query($request, $requestId, $chunkNum);
    }

    elsif ($object eq 'UsageRecord')
    {
        my $account = $request->getConditionValue('Account');
        my $user    = $request->getConditionValue('User');

        # Override Authorization
        my $authorized = 1;
        if ($request->getOverride())
        {
            $authorized = 0;
        }
        if (! $authorized && defined $user && $user eq $actor)
        {
            $authorized = 1;
        }
        if (   ! $authorized
            && defined $account
            && MAM::Cache->objectExists('AccountUser'))
        {
# See if actor is account admin
# SELECT Admin FROM AccountUser WHERE Account=$account AND Name=$actor AND Admin=True
            my $subResults = $database->select(
                object     => 'AccountUser',
                selections => [new MAM::Selection(name => 'Admin')],
                conditions => [
                    new MAM::Condition(
                        name  => 'Account',
                        value => $account
                    ),
                    new MAM::Condition(
                        name  => 'Name',
                        value => $actor
                    ),
                    new MAM::Condition(
                        name  => 'Admin',
                        value => 'True'
                    )
                ]
            );
            if (@{$subResults->{data}})
            {
                $authorized = 1;
            }
        }
        if (! $authorized)
        {
            if (! defined $user)
            {
                $request->addCondition(
                    new MAM::Condition(
                        name  => 'User',
                        value => $actor,
                    )
                );
            }
            else
            {
                return new MAM::Response()
                  ->failure('444',
                    "$actor is not authorized to Query the Usage Records");
            }
        }

        # Invoke base UsageRecord Query
        return MAM::Base->query($request, $requestId, $chunkNum);
    }

    elsif ($object eq 'User')
    {
        my $name = $request->getConditionValue('Name');

        # Parse out filters
        my %filters = ();
        foreach my $option (@options)
        {
            if ($option->getName() eq 'Filter')
            {
                my $filter = $option->getValue();
                if ($filter =~ /^([^\s=!]+)=([^\s=]+)$/)
                {
                    $filters{$1} = $2;
                }
            }
        }

        # Override Authorization
        my $authorized = 1;
        if ($request->getOverride())
        {
            $authorized = 0;
        }
        if (! $authorized && defined $name && $name eq $actor)
        {
            $authorized = 1;
        }
        if (! $authorized)
        {
            if (! defined $name)
            {
                $request->addCondition(
                    new MAM::Condition(
                        name  => 'Name',
                        value => $actor,
                    )
                );
            }
            else
            {
                return new MAM::Response()
                  ->failure('444',
                    "$actor is not authorized to Query the Users");
            }
        }

        # Quick access to base query if no options are specified
        # This has to occur after the override authorization check
        if (! @options || ! MAM::Cache->objectExists('AccountUser'))
        {
            # Invoke base User Query
            return MAM::Base->query($request, $requestId, $chunkNum);
        }

# Issue User Query with appropriate join conditions
# SELECT [DISTINCT] User.@selections FROM User[,AccountUser] WHERE [User.]@conditions [AND User.Name=AccountUser.Name]
        my @userObjects    = (new MAM::Object(name => 'User'));
        my @userSelections = @selections;
        my @userConditions = @conditions;
        my @userOptions    = @options;
        if (defined $filters{Account})
        {
            push @userObjects, (new MAM::Object(name => 'AccountUser'),);
            foreach my $selection (@userSelections)
            {
                $selection->setObject('User');
            }
            foreach my $condition (@userConditions)
            {
                $condition->setObject('User');
            }
            push @userConditions,
              ( new MAM::Condition(
                    object  => 'User',
                    name    => 'Name',
                    subject => 'AccountUser',
                    value   => 'Name',
                ),
                new MAM::Condition(
                    object => 'AccountUser',
                    name   => 'Account',
                    value  => $filters{Account},
                ),
              );
            push @userOptions,
              new MAM::Option(name => 'Unique', value => 'True');
        }
        my $results = $database->select(
            objects    => \@userObjects,
            selections => \@userSelections,
            conditions => \@userConditions,
            options    => \@userOptions,
        );
        my $count    = 0;
        my @data     = ();
        my $data     = $results->{data};
        my $cols     = $results->{cols};
        my $names    = $results->{names};
        my %metadata = %{$results->{metadata}};

        # Populate Data
        foreach my $row (@{$data})
        {
            $count++;
            my $datum = new MAM::Datum($object);
            for (my $i = 0; $i < $cols; $i++)
            {
                my $name  = $names->[$i];
                my $value = $row->[$i];     # Null comes in as undef
                $datum->addProperty($name, $value);
            }
            push @data, $datum;
        }

        my $response = new MAM::Response()->success($count, \@data, \%metadata);

        return $response;
    }

    else
    {
        return new MAM::Response()
          ->failure('315', "The Bank class does not implement $object Query");
    }
}

# ----------------------------------------------------------------------------
# $response = quote($request, $requestId);
# ----------------------------------------------------------------------------

# Quote (Payment, UsageRecord)
sub quote
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object          = $request->getObject();
    my $actor           = $request->getActor();
    my $database        = $request->getDatabase();
    my @options         = $request->getOptions();
    my $now             = $requestTime;
    my $balance         = 0;
    my $accounting_mode = $serverConfig->{'accounting.mode'}
      || $ACCOUNTING_MODE;
    my $accountingModeOption = $request->getOptionValue('AccountingMode');
    my $currency_precision = $serverConfig->{'currency.precision'} || 0;
    my $time_division =
      defined $request->getOptionValue('ShowHours') ? 3600 : 1;
    my $quoteMessage   = '';
    my $warningMessage = '';
    my $warnings       = 0;

    my $balanceFeedback;
    my $balanceFeedbackOption = $request->getOptionValue('BalanceFeedback');
    if (defined $balanceFeedbackOption && $balanceFeedbackOption =~ /^true$/i)
    {
        $balanceFeedback = 1;
    }
    else
    {
        $balanceFeedback = 0;
    }

    my $itemize;
    my $itemizeOption = $request->getOptionValue('Itemize');
    if (defined $itemizeOption && $itemizeOption =~ /^true$/i)
    {
        $itemize = 1;
    }
    else
    {
        $itemize = 0;
    }

    if ($balanceFeedback && $itemize)
    {
        return new MAM::Response()
          ->failure('318',
            'The Itemize and BalanceFeedback options are mutually exclusive');
    }

    # Demote accounting mode if option is lower than configuration
    if (defined $accountingModeOption)
    {
        if ((   $accountingModeOption eq 'usage-tracking'
                && (   $accounting_mode eq 'usage-tracking'
                    || $accounting_mode eq 'notional-charging'
                    || $accounting_mode eq 'fast-allocation'
                    || $accounting_mode eq 'strict-allocation')
            )
            || ($accountingModeOption eq 'notional-charging'
                && (   $accounting_mode eq 'notional-charging'
                    || $accounting_mode eq 'fast-allocation'
                    || $accounting_mode eq 'strict-allocation')
            )
            || ($accountingModeOption eq 'fast-allocation'
                && (   $accounting_mode eq 'fast-allocation'
                    || $accounting_mode eq 'strict-allocation')
            )
            || (   $accountingModeOption eq 'strict-allocation'
                && $accounting_mode eq 'strict-allocation')
          )
        {
            $accounting_mode = $accountingModeOption;
        }
        else
        {
            $warningMessage .=
              "\nWarning -- The specified accounting mode ($accountingModeOption) is being ignored because it is more strict than the accounting mode configured in Moab Accounting Manager ($accounting_mode).";
            $warnings++;
        }
    }

    if ($object eq 'Payment')
    {
        my $code     = 0;
        my @data     = ();
        my %metadata = ();
        my @messages = ();
        my $response;

        # We must pass either a payment id option
        my $id = $request->getOptionValue('Id');

        # Or payment component data
        my $data       = $request->getDataElement();
        my @components = $data->childNodes();

        # Quote Duration
        my $duration = $request->getOptionValue('Duration');

        # Payment Id was specified
        if (defined $id)
        {
            # SELECT Id from PaymentComponent WHERE Payment=$id
            my $results = $database->select(
                object     => 'PaymentComponent',
                selections => [
                    new MAM::Selection(name => 'Id', op => 'Sort'),
                    new MAM::Selection(name => 'Type'),
                ],
                conditions => [
                    new MAM::Condition(
                        name  => 'Payment',
                        value => $id,
                    ),
                ],
            );

            unless (@{$results->{data}})
            {
                return new MAM::Response(
                    status => 'Warning',
                    code   => '100',
                    message =>
                      "Warning -- No payment components were found for payment id ($id)",
                );
            }

            # Iterate over all components
            my $quote        = 0;
            my @usageRecords = @{$results->{data}};
            foreach my $row (@usageRecords)
            {
                my $usageRecord = $row->[0];
                my $type        = $row->[1];

                my @quoteOptions =
                  (new MAM::Option(name => 'Id', value => $usageRecord),);
                push @quoteOptions,
                  new MAM::Option(name => 'Duration', value => $duration)
                  if defined $duration;
                foreach my $option (@options)
                {
                    my $name = $option->getName();
                    if (   $name eq 'CostOnly'
                        || $name eq 'Itemize'
                        || $name eq 'ShowHours')
                    {
                        push @quoteOptions, $option;
                    }
                }

                # UsageRecord Quote
                my $subRequest = new MAM::Request(
                    database => $database,
                    object   => 'UsageRecord',
                    action   => 'Quote',
                    actor    => $SUPER_USER,
                    options  => \@quoteOptions,
                );
                my $subResponse = MAM::Bank->quote($subRequest, $requestId);
                if ($subResponse->getStatus() eq 'Failure')
                {
                    return new MAM::Response()->failure($subResponse->getCode(),
                        'Payment Quote failed in usage record quote: '
                          . $subResponse->getMessage());
                }
                $quote += $subResponse->getCount();
                push @data, $subResponse->getData();
                my %subMetadata = $subResponse->getMetadata();
                @metadata{keys %subMetadata} = values %subMetadata;
                push @messages,
                  'Warning -- No duration was specified for Update component quote'
                  if ($type eq 'Update' && ! defined $duration);
                $warnings++;
            }
            $quote =
              sprintf("%.${currency_precision}f", $quote / $time_division);
            my $message = join "\n", "Successfully quoted $quote credits",
              @messages;
            if ($warnings)
            {
                $response =
                  new MAM::Response()
                  ->warning($quote, \@data, \%metadata, $message);
            }
            else
            {
                $response =
                  new MAM::Response()
                  ->success($quote, \@data, \%metadata, $message);
            }
            return $response;
        }

        # Payment Component Data was specified
        elsif (@components)
        {
            # Iterate over usage records (components)
            # Sort components in order of Begin, Update, End
            my $quote = 0;
            foreach my $component (
                sort {
                    if ($a->nodeName eq $b->nodeName) {0}
                    elsif ($a->nodeName eq 'End' || $b->nodeName eq 'Begin')
                    {
                        1;
                    }
                    elsif ($a->nodeName eq 'Begin' || $b->nodeName eq 'End')
                    {
                        -1;
                    }
                } @components
              )
            {
                my $type = $component->nodeName;
                $component->setNodeName('UsageRecord');
                my $doc = XML::LibXML::Document->new();
                $doc->setDocumentElement($component);

                my @quoteOptions = ();
                push @quoteOptions,
                  new MAM::Option(name => 'Duration', value => $duration)
                  if defined $duration;
                foreach my $option (@options)
                {
                    my $name = $option->getName();
                    if (   $name eq 'CostOnly'
                        || $name eq 'Itemize'
                        || $name eq 'ShowHours')
                    {
                        push @quoteOptions, $option;
                    }
                }

                # UsageRecord Quote
                my $subRequest = new MAM::Request(
                    database => $database,
                    object   => 'UsageRecord',
                    action   => 'Quote',
                    actor    => $SUPER_USER,
                    options  => \@quoteOptions,
                );
                $subRequest->setDataElement($doc);
                my $subResponse = MAM::Bank->quote($subRequest, $requestId);
                if ($subResponse->getStatus() eq 'Failure')
                {
                    return new MAM::Response()->failure($subResponse->getCode(),
                        'Payment Quote failed in usage record quote: '
                          . $subResponse->getMessage());
                }
                $quote += $subResponse->getCount();
                push @data, $subResponse->getData();
                my %subMetadata = $subResponse->getMetadata();
                @metadata{keys %subMetadata} = values %subMetadata;
                push @messages,
                  'Warning -- No duration was specified for Update component quote'
                  if ($type eq 'Update' && ! defined $duration);
            }
            $quote =
              sprintf("%.${currency_precision}f", $quote / $time_division);
            my $message = join "\n", "Successfully quoted $quote credits",
              @messages;
            if ($warnings)
            {
                $response =
                  new MAM::Response()
                  ->warning($quote, \@data, \%metadata, $message);
            }
            else
            {
                $response =
                  new MAM::Response()
                  ->success($quote, \@data, \%metadata, $message);
            }
            return $response;
        }

        # Else fail
        else
        {
            return new MAM::Response()
              ->failure('314',
                'Either an Id option or payment component data is required');
        }
    }

    elsif ($object eq 'UsageRecord')
    {
        my $newQuoteId         = 0;
        my $createdUsageRecord = 0;
        my %quotes             = ();
        my $infinity           = 2147483647;
        my %fundAllocations    = ();
        my %allocations        = ();
        my %quoteChunk         = ();
        my @data               = ();
        my %metadata           = ();
        my $results;
        my $response;
        my $message;

        # A cost-only quote does not do any balance or validity checking
        my $costOnly;
        my $costOnlyOption = $request->getOptionValue('CostOnly');
        if (defined $costOnlyOption && $costOnlyOption =~ /^true$/i)
        {
            $costOnly = 1;
        }
        else
        {
            $costOnly = 0;
        }

        # A guaranteed quote will return a quote id with saved chargerates
        # as well as instantiate a usage record.
        my $guarantee;
        my $guaranteeOption = $request->getOptionValue('Guarantee');
        if (defined $guaranteeOption && $guaranteeOption =~ /^true$/i)
        {
            $guarantee = 1;
        }
        else
        {
            $guarantee = 0;
        }

        if ($costOnly && $guarantee)
        {
            return new MAM::Response()
              ->failure('318',
                'The CostOnly and Guarantee options are mutually exclusive');
        }

        if ($accounting_mode eq 'usage-tracking')
        {
            if ($costOnly)
            {
                return new MAM::Response()->failure('318',
                    'The CostOnly option may not be used when the accounting mode is usage-tracking'
                );
            }

            if ($guarantee)
            {
                return new MAM::Response()->failure('318',
                    'The Guarantee option may not be used when the accounting mode is usage-tracking'
                );
            }
        }

        # Check if data is present and parse out data properties
        my $data = $request->getDataElement();
        my $doc  = XML::LibXML::Document->new();
        $doc->setDocumentElement($data);
        my @rows = $data->childNodes();

        # We currently only support a solitary usage record
        if (@rows == 1)
        {
            return new MAM::Response()
              ->failure('314', 'The Data must be a Usage Record')
              unless $rows[0]->nodeName eq 'UsageRecord';
        }
        elsif (@rows > 1)
        {
            return new MAM::Response()
              ->failure('314',
                'Only one UsageRecord may be specified in the Data');
        }

        # Build up a list of usage record properties from the data
        my %dataProperties       = ();
        my %chargeDataProperties = ();
        if (@rows > 0)
        {
            foreach my $col ($rows[0]->childNodes())
            {
                my $propertyName = $col->nodeName();
                my $increment    = $col->getAttribute('increment');
                if ($col->hasChildNodes())    # Simple properties only
                {
                    my $propertyValue   = $col->textContent();
                    my $consumptionRate = $col->getAttribute('consumptionRate');
                    $propertyValue .= "*$consumptionRate"
                      if defined $consumptionRate;
                    $dataProperties{$propertyName} = $propertyValue;
                    if (defined $increment)
                    {
                        $chargeDataProperties{$propertyName} = $increment;
                    }
                    else
                    {
                        $chargeDataProperties{$propertyName} = $propertyValue;
                    }
                }
                elsif (defined $increment)
                {
                    $chargeDataProperties{$propertyName} = $increment;
                }
            }
        }

        # Id (usage record id)
        my $id;
        my $idOption = $request->getOptionValue('Id');
        if (! defined $idOption && ! keys %dataProperties)
        {
            return new MAM::Response()
              ->failure('314',
                'Either an Id option or usage record data is required');
        }
        my $idProperty = $dataProperties{Id};
        if (defined $idOption)
        {
            return new MAM::Response()->failure('318',
                "The specified Id option ($idOption) differs from the specified Id usage record property ($idProperty)"
            ) if (defined $idProperty && $idProperty ne $idOption);
            $id = $idOption;
        }
        else
        {
            $id = $idProperty;
        }

        # Look up existing usage record properties if id is known
        my %usageRecordProperties = ();
        my %chargeProperties      = ();
        if (defined $id)
        {
            my $subRequest = new MAM::Request(
                database   => $database,
                object     => 'UsageRecord',
                action     => 'Query',
                actor      => $SUPER_USER,
                conditions => [new MAM::Condition(name => 'Id', value => $id)],
            );
            # Invoke the proxy so it prepares the subRequest by setting defaults
            MAM::Proxy->new(
                database  => $database,
                request   => $subRequest,
                requestId => $requestId
            );
            # But use select so we can parse the data results easier
            # SELECT * FROM UsageRecord WHERE Id=$id
            my $subResults = $database->select(
                object     => 'UsageRecord',
                selections => [$subRequest->getSelections()],
                conditions => [new MAM::Condition(name => 'Id', value => $id)],
            );
            if (@{$subResults->{data}})
            {
                my $row   = ${$subResults->{data}}[0];
                my $cols  = $subResults->{cols};
                my $names = $subResults->{names};
                for (my $i = 0; $i < $cols; $i++)
                {
                    my $name  = $names->[$i];
                    my $value = $row->[$i];     # Null comes in as undef
                    if (defined $value)
                    {
                        $usageRecordProperties{$name} = $value;
                        $chargeProperties{$name} = $value if $value =~ /^\D/;
                    }
                }
            }
            else
            {
                return new MAM::Response()
                  ->failure('740',
                    "No usage record found for specified id ($id)");
            }
        }

     # Merge data properties in with existing usage record and charge properties
        foreach my $name (keys %dataProperties)
        {
            $usageRecordProperties{$name} = $dataProperties{$name};
        }
        foreach my $name (keys %chargeDataProperties)
        {
            $chargeProperties{$name} = $chargeDataProperties{$name};
        }

        # Employ object instance overrides for specified usage record properties
        my @overrideCheckProperties = keys %usageRecordProperties;
        while (my $propertyName = shift @overrideCheckProperties)
        {
            # If the usage record property has an object associated with it
            if (MAM::Cache->objectExists($propertyName))
            {
                # Iterate over its attributes
                foreach my $attribute (
                    MAM::Cache->listAttributesWithoutPropertyValue(
                        $propertyName, 'Values', undef
                    )
                  )
                {
                    my $values = MAM::Cache->getAttributeProperty($propertyName,
                        $attribute, 'Values');
                    # If an attribute is an override attribute
                    if ($values =~ /^@([:?!]=)(\w+)/)
                    {
                        my ($overrideType, $overrideName) = ($1, $2);
# See if the instance has a value set for this attribute
# Assume the primary key is Name
# SELECT $attribute FROM $propertyName WHERE Name=$usageRecordProperties{$propertyName}
                        my $subResults = $database->select(
                            object => "$propertyName",
                            selections =>
                              [new MAM::Selection(name => $attribute)],
                            conditions => [
                                new MAM::Condition(
                                    name => 'Name',
                                    value =>
                                      $usageRecordProperties{$propertyName}
                                )
                            ]
                        );
                        # If the instance override has a value
                        if (@{$subResults->{data}}
                            && defined ${$subResults->{data}}[0]->[0])
                        {
                            my $overrideValue = ${$subResults->{data}}[0]->[0];
                            # Override it with the value if the usage record
                            # property is undefined or if it has a different
                            # value and this is an assign-always override
                            if ((   ! defined
                                    $usageRecordProperties{$overrideName}
                                )
                                || (   $overrideType eq ':='
                                    && $usageRecordProperties{$overrideName} ne
                                    $overrideValue)
                              )
                            {
                                $usageRecordProperties{$overrideName} =
                                  $overrideValue;
                                $chargeProperties{$overrideName} =
                                  $overrideValue;
                                push @overrideCheckProperties, $overrideName;
                            }
                            # Or return with a failure if the usage record
                            # property is defined but has a different
                            # value and this is an assign-verify override
                            elsif ($overrideType eq '!='
                                && defined $usageRecordProperties{$overrideName}
                                && $usageRecordProperties{$overrideName} ne
                                $overrideValue)
                            {
                                return new MAM::Response()->failure('740',
                                    "Specified $overrideName ($usageRecordProperties{$overrideName}) conflicts with ${propertyName}-designated $overrideName ($overrideValue)"
                                );
                            }
                        }
                    }
                }
            }
        }

        # Build up a list of usage record attributes and set defaults
        my %usageRecordAttributes = ();
        foreach my $attributeName (
            MAM::Cache->listAttributesWithPropertyValue(
                'UsageRecord', 'Hidden', 'False'
            )
          )
        {
            $usageRecordAttributes{$attributeName} = 1;

         # Set default values for all undefined usage record properties that are
         # first class objects with a default value
            my $defaultValue =
              MAM::Cache->getObjectProperty($attributeName, 'DefaultValue');
            if (! defined $usageRecordProperties{$attributeName}
                && defined $defaultValue)
            {
                $usageRecordProperties{$attributeName} = $defaultValue;
                $chargeProperties{$attributeName}      = $defaultValue;
            }
        }

        # Build up list of usage record assignments
        my @usageRecordAssignments = ();
        foreach my $propertyName (sort keys %usageRecordProperties)
        {
            if ($usageRecordAttributes{$propertyName})
            {
                my $propertyValue = $usageRecordProperties{$propertyName};
                # Remove coefficients and consumption rates
                $propertyValue =~ s/\*(\d*\.?\d+)$//;
                push @usageRecordAssignments,
                  new MAM::Assignment(
                    name  => $propertyName,
                    value => $propertyValue
                  ) unless ($propertyName eq 'Id' || $propertyName eq 'Charge');
            }
        }

        # Start Time, End Time and Duration
        my ($quoteStartTime, $quoteEndTime, $quoteDuration);
        my $startTimeOption     = $request->getOptionValue('StartTime');
        my $startTimeProperty   = $usageRecordProperties{StartTime};
        my $endTimeOption       = $request->getOptionValue('EndTime');
        my $endTimeProperty     = $usageRecordProperties{EndTime};
        my $durationOption      = $request->getOptionValue('Duration');
        my $graceDurationOption = $request->getOptionValue('GraceDuration');
        my $requestedDurationProperty =
          $usageRecordProperties{RequestedDuration};

        if (defined $startTimeOption)
        {
            $quoteStartTime = $startTimeOption;
        }
        else    # Use default if duration was not obtained via option
        {
            $quoteStartTime = $now;
        }

        if (defined $durationOption)
        {
            $quoteDuration = $durationOption;
            if (defined $endTimeOption)
            {
                $quoteEndTime = $endTimeOption;
            }
            elsif (defined $graceDurationOption)
            {
                if ($graceDurationOption !~ /^\d+$/)
                {
                    return new MAM::Response()->failure('314',
                        'The grace duration must be an integer number of seconds'
                    );
                }
                $quoteEndTime =
                  $quoteStartTime + $quoteDuration + $graceDurationOption;
            }
            else
            {
                $quoteEndTime = $quoteStartTime + $quoteDuration;
            }
        }
        elsif (defined $endTimeOption)
        {
            $quoteEndTime  = $endTimeOption;
            $quoteDuration = $quoteEndTime - $quoteStartTime;
        }
        else
        {
            # Override current start time with start time property if specified
            if ((! defined $startTimeOption) && defined $startTimeProperty)
            {
                $quoteStartTime = $startTimeProperty;
            }

            if (defined $requestedDurationProperty)
            {
                $quoteDuration = $requestedDurationProperty;
                if (defined $endTimeProperty)
                {
                    $quoteEndTime = $endTimeProperty;
                }
                else
                {
                    $quoteEndTime = $quoteStartTime + $quoteDuration;
                }
            }
            elsif (defined $endTimeProperty)
            {
                $quoteEndTime  = $endTimeProperty;
                $quoteDuration = $quoteEndTime - $quoteStartTime;
            }
            else
            {
                $quoteEndTime  = $quoteStartTime;
                $quoteDuration = 0;
            }
        }
        if ($quoteDuration !~ /^\d+$/)
        {
            return new MAM::Response()->failure('317',
                "The duration ($quoteDuration) must be a non-negative integer number of seconds"
            );
        }

        unless ($costOnly)
        {
    # Verify that every specified usage record property exists if it is a value-
    # constrained object and that it is active if it has an active field
            foreach my $propertyName (sort keys %usageRecordProperties)
            {
                my $values =
                  MAM::Cache->getAttributeProperty('UsageRecord', $propertyName,
                    'Values');
                if (   defined $values
                    && $values =~ /^@(?:[:?!]=)?(\w+)/
                    && MAM::Cache->objectExists($1))
                {
                    my $propertyObject = $1;
                    my @primaryKeys =
                      MAM::Cache->listAttributesWithPropertyValue(
                        $propertyObject, 'PrimaryKey', 'True');

                    # We can only do this if it has a single primary key
                    if (@primaryKeys == 1)
                    {
   # SELECT $filterKey[,Active] FROM $filterObject WHERE $filterKey=$filterValue
                        my $propertyKey = $primaryKeys[0];
                        my $propertyValue =
                          $usageRecordProperties{$propertyName};
                        my @validitySelections =
                          (new MAM::Selection(name => $propertyKey));
                        push @validitySelections,
                          new MAM::Selection(name => 'Active')
                          if MAM::Cache->attributeExists($propertyObject,
                            'Active');

                        my @validityConditions = (
                            new MAM::Condition(
                                name  => $propertyKey,
                                value => $propertyValue
                            )
                        );
                        my $subResults = $database->select(
                            object     => $propertyObject,
                            selections => \@validitySelections,
                            conditions => \@validityConditions,
                        );
                        unless (@{$subResults->{data}})
                        {
                         # Check to see if the instance should be auto-generated
                            if (MAM::Cache->getObjectProperty($propertyObject,
                                    'AutoGen') eq 'True'
                              )
                            {
                                # Create the instance
                                my @autoGenAssignments = (
                                    new MAM::Assignment(
                                        name  => $propertyKey,
                                        value => $propertyValue
                                    )
                                );
                                if (MAM::Cache->attributeExists(
                                        $propertyObject, 'Description'
                                    )
                                  )
                                {
                                    push @autoGenAssignments,
                                      new MAM::Assignment(
                                        name  => 'Description',
                                        value => 'Auto-Generated'
                                      );
                                }
                                my $subRequest = new MAM::Request(
                                    database    => $database,
                                    object      => $propertyObject,
                                    action      => 'Create',
                                    actor       => $SUPER_USER,
                                    assignments => \@autoGenAssignments,
                                );
            # Invoke the proxy so it prepares the subRequest by setting defaults
                                MAM::Proxy->new(
                                    database  => $database,
                                    request   => $subRequest,
                                    requestId => $requestId
                                );
                                my $subResponse =
                                  MAM::Base->create($subRequest, $requestId);
                                if ($subResponse->getStatus() eq 'Failure')
                                {
                                    my $message =
                                      "Unable to auto-gen $propertyObject $propertyValue: "
                                      . $subResponse->getMessage();
                                    $log->error($message);
                                    throw MAM::Exception(
                                        $subResponse->getCode(), $message);
                                }
                            }
                            else
                            {

                                return new MAM::Response()->failure('740',
                                    "$propertyObject ($propertyName) $propertyValue does not exist"
                                );
                            }
                        }
                        elsif (exists ${$subResults->{data}}[0]->[1]
                            && ${$subResults->{data}}[0]->[1] ne 'True')
                        {
                            return new MAM::Response()->failure('740',
                                "$propertyObject $propertyValue is not active");
                        }
                    }
                }
            }

            if (   $accounting_mode ne 'usage-tracking'
                && $accounting_mode ne 'notional-charging')
            {
                # Refresh Allocations
                my $subRequest = new MAM::Request(
                    database => $database,
                    object   => 'Allocation',
                    action   => 'Refresh'
                );
                my $subResponse = MAM::Bank->refresh($subRequest, $requestId);
                if ($subResponse->getStatus() eq 'Failure')
                {
                    return new MAM::Response()->failure($subResponse->getCode(),
                        'Failed while refreshing allocations: '
                          . $subResponse->getMessage());
                }

# First, we build a list of funds along with their eligible allocations
# We need to do this because the trivial fund might have no constraints
# SELECT DISTINCT Allocation.Id,Allocation.Fund,Allocation.Amount,Allocation.CreditLimit,Allocation.StartTime,Allocation.EndTime,Fund.Priority FROM Allocation,Fund WHERE Allocation.Fund=Fund.Id AND Allocation.StartTime<=$quoteStartTime AND Allocation.EndTime>$quoteStartTime
                my @allocationConditions = ();
                if (($serverConfig->{'debit.activeallocationonly'} || 'false')
                    =~ /^true$/i)
                {
                    push @allocationConditions,
                      ( new MAM::Condition(
                            object => 'Allocation',
                            name   => 'StartTime',
                            value  => $now,
                            op     => 'LE',
                        ),
                        new MAM::Condition(
                            object => 'Allocation',
                            name   => 'EndTime',
                            value  => $now,
                            op     => 'GT',
                        )
                      );
                }
                elsif (($serverConfig->{'debit.pointintimeonly'} || 'false') =~
                    /^true$/i)
                {
                    push @allocationConditions,
                      ( new MAM::Condition(
                            object => 'Allocation',
                            name   => 'StartTime',
                            value  => $quoteStartTime,
                            op     => 'LE',
                        ),
                        new MAM::Condition(
                            object => 'Allocation',
                            name   => 'EndTime',
                            value  => $quoteStartTime,
                            op     => 'GT',
                        )
                      );
                }
                else
                {
                    push @allocationConditions,
                      ( new MAM::Condition(
                            object => 'Allocation',
                            name   => 'StartTime',
                            value  => $quoteEndTime,
                            op     => 'LE',
                        ),
                        new MAM::Condition(
                            object => 'Allocation',
                            name   => 'EndTime',
                            value  => $quoteStartTime,
                            op     => 'GT',
                        )
                      );
                }
                $results = $database->select(
                    objects => [
                        new MAM::Object(name => 'Allocation'),
                        new MAM::Object(name => 'Fund'),
                    ],
                    selections => [
                        new MAM::Selection(
                            object => 'Allocation',
                            name   => 'Id'
                        ),
                        new MAM::Selection(
                            object => 'Allocation',
                            name   => 'Fund'
                        ),
                        new MAM::Selection(
                            object => 'Allocation',
                            name   => 'Amount'
                        ),
                        new MAM::Selection(
                            object => 'Allocation',
                            name   => 'CreditLimit'
                        ),
                        new MAM::Selection(
                            object => 'Allocation',
                            name   => 'StartTime'
                        ),
                        new MAM::Selection(
                            object => 'Allocation',
                            name   => 'EndTime'
                        ),
                        new MAM::Selection(
                            object => 'Fund',
                            name   => 'Priority'
                        ),
                    ],
                    conditions => [
                        new MAM::Condition(
                            object  => 'Allocation',
                            name    => 'Fund',
                            subject => 'Fund',
                            value   => 'Id',
                        ),
                        @allocationConditions,
                    ],
                    options =>
                      [new MAM::Option(name => 'Unique', value => 'True')]
                );
                foreach my $row (@{$results->{data}})
                {
                    my $id          = $row->[0];
                    my $fund        = $row->[1];
                    my $amount      = $row->[2] || 0;
                    my $creditLimit = $row->[3] || 0;
                    my $startTime   = $row->[4];
                    my $endTime     = $row->[5];
                    my $priority    = $row->[6] || 0;

                    push @{$fundAllocations{$fund}}, {
                        id        => $id,
                        fund      => $fund,
                        balance   => $amount + $creditLimit,
                        startTime => $startTime,
                        endTime   => $endTime,
                  # If you want it to take precedence over endTime, use millions
                        weight => 100 * int(($infinity - $endTime) / 86400) +
                          10 * $priority,
                    };
                }

      #use Data::Dumper;
      #print Data::Dumper->Dump([\%fundAllocations], ['fundAllocations']), "\n";

       # Next, we prune off the funds that violate the constraints
       # In order to limit the list of constraints we must loop on, we do a join
       # on eligible allocations instead of constraining it to a huge fund list
                my %fundConstraints = ();
                my %limbo           = ();

                # Build up an association table for quick checks
                my %associations = ();
                foreach
                  my $association (MAM::Cache->associationLookup(undef, undef))
                {
                    my $parent =
                      MAM::Cache->getObjectProperty($association, 'Parent');
                    next unless defined $usageRecordProperties{$parent};
                    my $child =
                      MAM::Cache->getObjectProperty($association, 'Child');
                    $associations{$parent}{$child} = $association;
                }

# SELECT DISTINCT Constraint.Id,Constraint.Fund,Constraint.Name,Constraint.Value FROM Constraint,Allocation WHERE Constraint.Fund=Allocation.Fund AND Allocation.StartTime<=$quoteStartTime AND Allocation.EndTime>$quoteStartTime
                $results = $database->select(
                    objects => [
                        new MAM::Object(name => 'Constraint'),
                        new MAM::Object(name => 'Allocation'),
                    ],
                    selections => [
                        new MAM::Selection(
                            object => 'Constraint',
                            name   => 'Id'
                        ),
                        new MAM::Selection(
                            object => 'Constraint',
                            name   => 'Fund'
                        ),
                        new MAM::Selection(
                            object => 'Constraint',
                            name   => 'Name'
                        ),
                        new MAM::Selection(
                            object => 'Constraint',
                            name   => 'Value'
                        ),
                    ],
                    conditions => [
                        new MAM::Condition(
                            object  => 'Constraint',
                            name    => 'Fund',
                            subject => 'Allocation',
                            value   => 'Fund',
                        ),
                        @allocationConditions,
                    ],
                    options =>
                      [new MAM::Option(name => 'Unique', value => 'True')]
                );
                foreach my $row (@{$results->{data}})
                {
                    my $id     = $row->[0];
                    my $fundId = $row->[1];
                    my $name   = $row->[2];
                    my $value  = $row->[3];

                 # Short circuit if this constraint has no associated allocation
                    next unless defined $fundAllocations{$fundId};
                    push @{$fundConstraints{$fundId}{$name}}, $value;
                }

                my %accepted    = ();
                my %rejected    = ();
                my %memberships = ();
              FUNDID:
                foreach my $fundId (sort { $a <=> $b } keys %fundConstraints)
                {
                    foreach my $constraintName (
                        sort keys %{$fundConstraints{$fundId}})
                    {
                        foreach my $constraintVal (
                            @{$fundConstraints{$fundId}{$constraintName}})
                        {
                            my $constraintValue = $constraintVal;  # Make a copy
                            my $accessAllowed = $constraintValue !~ s/^!//;

                   # If there is a usage record property matching the constraint
                            if (defined $usageRecordProperties{$constraintName}
                                && $usageRecordProperties{$constraintName} eq
                                $constraintValue)
                            {
                                if ($accessAllowed)
                                {
                                    # Check if the usage record property is a
                                    # parent in an association
                                    if (defined $associations{$constraintName})
                                    {
                                        foreach my $childName (
                                            keys
                                            %{$associations{$constraintName}})
                                        {
                                            my $association =
                                              $associations{$constraintName}
                                              {$childName};

                                          # If so, the child should be specified
                                            if (defined $usageRecordProperties{
                                                    $childName})
                                            {
                             # If there is a matching member found, add the fund

                                                # Check for a match unless the
                                                # fund has a matching constraint
                                                unless (
                                                    exists
                                                    $fundConstraints{$fundId}
                                                    {$childName})
                                                {
                                              # Use cache for membership results
                                                    unless (
                                                        exists $memberships{
                                                            $association})
                                                    {
                           # e.g. Account==chemistry && (Name==amy || Name==ANY)
                                                        my @membershipConditions
                                                          = (
                                                            new MAM::Condition(
                                                                name =>
                                                                  $constraintName,
                                                                value =>
                                                                  $usageRecordProperties{
                                                                    $constraintName
                                                                  },
                                                            ),
                                                            new MAM::Condition(
                                                                name => 'Name'
                                                                , # Assume simple association
                                                                value =>
                                                                  $usageRecordProperties{
                                                                    $childName},
                                                                conj  => 'And',
                                                                group => '+1',
                                                            ),
                                                            new MAM::Condition(
                                                                name => 'Name'
                                                                , # Assume simple association
                                                                value => 'ANY',
                                                                conj  => 'Or',
                                                                group => '-1',
                                                            ),
                                                          );
                                                        push
                                                          @membershipConditions,
                                                          new MAM::Condition(
                                                            name  => 'Active',
                                                            value => 'True'
                                                          )
                                                          if MAM::Cache
                                                          ->attributeExists(
                                                            $association,
                                                            'Active');
                                                        my $subResults =
                                                          $database->select(
                                                            object =>
                                                              $association,
                                                            conditions =>
                                                              \@membershipConditions,
                                                          );
                                                        $memberships{
                                                            $association} =
                                                          scalar
                                                          @{$subResults->{data}
                                                          };
                                                    }
                                                    unless (
                                                        $memberships{
                                                            $association})
                                                    {
                                                        return
                                                          new MAM::Response()
                                                          ->failure(
                                                            '740',
                                                            "$childName $usageRecordProperties{$childName} is not a valid member of $constraintName $usageRecordProperties{$constraintName}"
                                                          );
                                                    }
                                                }
                                            }
                                            else
                                            {
                                                return new MAM::Response()
                                                  ->failure(
                                                    '314',
                                                    "A $childName must be specified when specifying a $constraintName"
                                                  );
                                            }
                                        }
                                    }
                                    $accepted{$fundId}{$constraintName} = 1;
                              # Add to the allocation weight for each constraint
                                    foreach my $allocation (
                                        @{$fundAllocations{$fundId}})
                                    {
                                        $allocation->{weight} += 1;
                                    }
                                }
                                else
                                {
                                    if (defined $fundAllocations{$fundId})
                                    {
                                        # Move them to limbo
                                        @{$limbo{$fundId}} =
                                          @{$fundAllocations{$fundId}};
                                        delete $fundAllocations{$fundId};
                                        next FUNDID;
                                    }
                                }
                            }
         # There is not a usage record property matching the positive constraint
         # or the filter does not exist or the constraint is negative
                            else
                            {
                                if ($accessAllowed
                                    || !
                                    defined
                                    $usageRecordProperties{$constraintName})
                                {
                                    if (defined $fundAllocations{$fundId})
                                    {
                                        # Provisionally Reject the constraint
                                        $rejected{$fundId}{$constraintName} = 1;
                                    }
                                }
                            }
                        }
                    }
                }

               # Move all funds to limbo that have rejected constraints and that
               # do not have accepted constraints
                foreach my $fundId (sort { $a <=> $b } keys %rejected)
                {
                    foreach my $constraintName (sort keys %{$rejected{$fundId}})
                    {
                        unless (exists $accepted{$fundId}{$constraintName}
                            || ! defined $fundAllocations{$fundId})
                        {
                            # Move them to limbo
                            @{$limbo{$fundId}} = @{$fundAllocations{$fundId}};
                            delete $fundAllocations{$fundId};
                        }
                    }
                }

      #print Data::Dumper->Dump([\%fundAllocations], ['fundAllocations']), "\n";

                # Then, we build up fund linkages for ancestor funds
                my %ancestors = ();
                # SELECT Fund,Id FROM FundFund WHERE Overflow=True
                $results = $database->select(
                    object     => 'FundFund',
                    selections => [
                        new MAM::Selection(name => 'Fund'),
                        new MAM::Selection(name => 'Id'),
                    ],
                    conditions =>
                      [new MAM::Condition(name => 'Overflow', value => 'True')],
                );
                foreach my $row (@{$results->{data}})
                {
                    my $parent = $row->[0];
                    my $child  = $row->[1];
                    $ancestors{$child}{$parent} = 1;
                }

            # Add ancestors a generation at a time until there is no more growth
                my $distance = 0;
                while (1)
                {
                    $distance++;
                    my $count = scalar keys %fundAllocations;
                    foreach
                      my $fundId (sort { $a <=> $b } keys %fundAllocations)
                    {
                        foreach my $parent (keys %{$ancestors{$fundId}})
                        {
                            if ($limbo{$parent})
                            {
                      # Move it back from limbo with negative distance as weight
                                foreach my $allocation (@{$limbo{$parent}})
                                {
                                    $allocation->{weight} = $distance * 100000 -
                                      ($allocation->{endTime});
                                }
                                @{$fundAllocations{$parent}} =
                                  @{$limbo{$parent}};
                                delete $limbo{$parent};
                            }
                        }
                    }
                    last if $count == scalar keys %fundAllocations;
                }

      #print Data::Dumper->Dump([\%fundAllocations], ['fundAllocations']), "\n";

                if ($log->is_debug())
                {
                    $log->debug('Feasible fund list is: ' . join ',',
                        sort keys %fundAllocations);
                }

                # Begin building allocations list
                foreach my $fundId (sort { $a <=> $b } keys %fundAllocations)
                {
                    foreach my $allocation (@{$fundAllocations{$fundId}})
                    {
                        $allocations{$allocation->{id}} = $allocation;
                    }
                }
            }
        }

        # Instance
        my $instance = $usageRecordProperties{Instance};

        # Look to see if a quote template is specified so we can use it for
        # override rates

        # Quote Id
        # Quote templates may be passed in as options or as in the usage record
        # data properties or may it be extracted from the already existing usage
        # record. When passed in as a usage record data property, it will
        # replace the current value in the usage record.
        # Any of the sources will be used in the quote amount calculation if
        # available with a passed in usage record data property trumping an
        # already stored value and the option trumping both. A guaranteed quote
        # will always store itself back into the usage record.
        # Thus it is possible to request a quaranteed quote against a usage
        # record that already has a stored quote value while also passing in a
        # usage record quote property and a quote option. The passed in usage
        # record quote property will temporarily replace the preexisting value
        # while the option will be used in the calculated quote amount and the
        # resulting guaranteed quote id will then be stored in the usage record
        # clobbering the one passed in via the data property.
        # Of course, this one case makes no sense but the options offer the
        # full spectrum of flexibility. The normal case will probably be to
        # pass in the quote as an option so it does not set the value in the
        # usage record.
        my $quoteId =
          defined $request->getOptionValue('Quote')
          ? $request->getOptionValue('Quote')
          : $usageRecordProperties{Quote};
        if (defined $quoteId)
        {
# Check to see if the quote exists and is not expired or claimed by a different instance
# SELECT Pinned,Instance,UsageRecord FROM Quote WHERE Id=$quoteId AND StartTime<=$quoteStartTime AND EndTime>$quoteStartTime
            my $subResults = $database->select(
                object     => 'Quote',
                selections => [
                    new MAM::Selection(name => 'Pinned'),
                    new MAM::Selection(name => 'Instance'),
                    new MAM::Selection(name => 'UsageRecord')
                ],
                conditions => [
                    new MAM::Condition(name => 'Id', value => $quoteId),
                    new MAM::Condition(
                        name  => 'StartTime',
                        value => $quoteStartTime,
                        op    => 'LE'
                    ),
                    new MAM::Condition(
                        name  => 'EndTime',
                        value => $quoteStartTime,
                        op    => 'GT'
                    )
                ]
            );
            if (@{$subResults->{data}})
            {
                my $row              = ${$subResults->{data}}[0];
                my $quotePinned      = $row->[0];
                my $quoteInstance    = $row->[1];
                my $quoteUsageRecord = $row->[2];

            # Reject all pinned quotes when either the usage instance is not
            # specified or it is specified but does not match the quote instance
                if (   defined $quotePinned
                    && $quotePinned eq 'True'
                    && ((! defined $instance)
                        || (defined $quoteInstance
                            && $quoteInstance ne $instance)
                    )
                  )
                {
                    # Nullify the quote if absent, expired or used
                    $quoteMessage .= "\nWarning -- Quote ($quoteId) is pinned";
                    $quoteMessage .= " by instance ($quoteInstance)"
                      if defined $quoteInstance;
                    $quoteMessage .= ' and cannot be used';
                    $quoteMessage .= ' anonymously'
                      if ! defined $quoteInstance;
                    $warnings++;

                    undef $quoteId;
                    @usageRecordAssignments =
                      grep { $_->getName() ne 'Quote' } @usageRecordAssignments;
                }
# Accept all unpinned quotes or
# ( pinned quotes where a usage instance has been specified
# in the charge and ( where the quote instance is either
# uninitialized or matches the usage instance ))
# if ((! defined $quotePinned) || $quotePinned ne 'True' || (defined $instance &&
# (!(defined $quoteInstance) || $quoteInstance eq $instance)))
                else
                {
                    # Set the quote instance if pinned and uninitialized
                    if (   defined $quotePinned
                        && $quotePinned eq 'True'
                        && defined $instance
                        && ! defined $quoteInstance)
                    {
                        my $subRequest = new MAM::Request(
                            database    => $database,
                            object      => 'Quote',
                            action      => 'Modify',
                            actor       => $SUPER_USER,
                            assignments => [
                                new MAM::Assignment(
                                    name  => 'Instance',
                                    value => $instance
                                )
                            ],
                            conditions => [
                                new MAM::Condition(
                                    name  => 'Id',
                                    value => $quoteId
                                )
                            ]
                        );
                        my $subResponse =
                          MAM::Base->modify($subRequest, $requestId);
                        if ($subResponse->getStatus() eq 'Failure')
                        {
                            return new MAM::Response()->failure(
                                $subResponse->getCode(),
                                'Unable to modify quote: '
                                  . $subResponse->getMessage()
                            );
                        }
                    }
                }
            }
            else
            {
                # Nullify the quote if absent or expired
                $quoteMessage .=
                  "\nWarning -- Quote ($quoteId) is invalid and cannot be used";
                $warnings++;

                undef $quoteId;
                @usageRecordAssignments =
                  grep { $_->getName() ne 'Quote' } @usageRecordAssignments;
            }
        }

# Build a list of conflicting liens. Also get Id, Instance,
# and/or UsageRecord from lien if we have enough info.
# SELECT Lien.Id,Lien.UsageRecord,Lien.Instance,LienAllocation.Id,LienAllocation.Amount FROM Lien,LienAllocation WHERE Lien.Id=LienAllocation.Lien AND Lien.StartTime<=$now AND Lien.EndTime>$now AND [Id=$lienId|UsageRecord=$id|Instance=$instance]
        my %liens           = ();
        my %lienAllocations = ();
        if (   $accounting_mode ne 'usage-tracking'
            && $accounting_mode ne 'notional-charging'
            && $accounting_mode ne 'fast-allocation')
        {
            if (defined $instance || defined $id)
            {
                my %instances = ();
                my %ids       = ();
                my @objects   = (
                    new MAM::Object(name => 'Lien'),
                    new MAM::Object(name => 'LienAllocation'),
                );

                my @selections = (
                    new MAM::Selection(object => 'Lien', name => 'Id'),
                    new MAM::Selection(
                        object => 'Lien',
                        name   => 'UsageRecord'
                    ),
                    new MAM::Selection(
                        object => 'Lien',
                        name   => 'Instance'
                    ),
                    new MAM::Selection(
                        object => 'LienAllocation',
                        name   => 'Id'
                    ),
                    new MAM::Selection(
                        object => 'LienAllocation',
                        name   => 'Amount'
                    ),
                );
                my @conditions = (
                    new MAM::Condition(
                        object  => 'Lien',
                        name    => 'Id',
                        subject => 'LienAllocation',
                        value   => 'Lien',
                    ),
                );
                if (defined $id)
                {
                    push @conditions,
                      new MAM::Condition(
                        object => 'Lien',
                        name   => 'UsageRecord',
                        value  => $id
                      );
                }
                elsif (defined $instance)
                {
                    push @conditions,
                      new MAM::Condition(
                        object => 'Lien',
                        name   => 'Instance',
                        value  => $instance
                      );
                }
                $results = $database->select(
                    objects    => \@objects,
                    selections => \@selections,
                    conditions => \@conditions
                );
                foreach my $row (@{$results->{data}})
                {
                    my $lienId = $row->[0];
                    $liens{$lienId}++;
                    $ids{$row->[1]}++       if defined $row->[1];
                    $instances{$row->[2]}++ if defined $row->[2];
                    my $allocationId = $row->[3];
                    my $lienAmount = $row->[4] || 0;

                    $lienAllocations{$lienId, $allocationId} = $lienAmount;
                    if ($allocations{$allocationId})
                    {
                        $allocations{$allocationId}{balance} += $lienAmount;
                    }
                }
                $id = ${[keys %ids]}[0]
                  if (! defined $id && scalar keys %ids == 1);
                $instance = ${[keys %instances]}[0]
                  if (! defined $instance && scalar keys %instances == 1);
            }

            if ($log->is_debug())
            {
                $log->debug(
                    'Lien allocation list is: ' . join ',',
                    map { '(' . (join ',', split $;, $_) . ')' }
                      keys %lienAllocations
                );
            }
        }

        # Determine whether to create quote and usage records
        if ($guarantee || $accounting_mode eq 'usage-tracking')
        {
            unless (defined $id)
            {
                # Create usage record
                my $subRequest = new MAM::Request(
                    database => $database,
                    object   => 'UsageRecord',
                    action   => 'Create',
                    actor    => $SUPER_USER,
                );
            # Invoke the proxy so it prepares the subRequest by setting defaults
                MAM::Proxy->new(
                    database  => $database,
                    request   => $subRequest,
                    requestId => $requestId
                );
                my $subResponse =
                  MAM::Base->create($subRequest, $requestId, 'NoLog');

                if ($subResponse->getStatus() eq 'Success')
                {
                    # Figure out the id of the usage record we just created
                    $id = $subResponse->getDatumValue('Id');
                    $usageRecordProperties{Id} = $id;
                    $chargeProperties{Id}      = $id;

                    # Log the transaction
                    MAM::Bank->logTransaction(
                        database    => $database,
                        requestId   => $requestId,
                        txnId       => $database->nextId('Transaction'),
                        object      => 'UsageRecord',
                        action      => 'Create',
                        actor       => $actor,
                        assignments => [
                            $subRequest->getAssignments(),
                            new MAM::Assignment(
                                name  => 'Id',
                                value => $id
                            ),
                            new MAM::Assignment(
                                name  => 'UsageRecord',
                                value => $id
                            ),
                        ],
                        options => \@options,
                        count   => 1,
                    );
                    $createdUsageRecord = 1;
                }
                else
                {
                    return new MAM::Response()->failure($subResponse->getCode(),
                        'Unable to create usage record: '
                          . $subResponse->getMessage());
                }
            }

            # Create the quote
            if ($accounting_mode ne 'usage-tracking')
            {
                my @quoteAssignments = (
                    new MAM::Assignment(
                        name  => 'StartTime',
                        value => $quoteStartTime
                    ),
                    new MAM::Assignment(
                        name  => 'EndTime',
                        value => $quoteEndTime
                    ),
                    new MAM::Assignment(name => 'UsageRecord', value => $id),
                    new MAM::Assignment(
                        name  => 'Duration',
                        value => $quoteDuration
                    ),
                );
                my $description = $request->getOptionValue('Description');
                push @quoteAssignments,
                  new MAM::Assignment(name => 'Instance', value => $instance)
                  if defined $instance;
                push @quoteAssignments,
                  new MAM::Assignment(
                    name  => 'Description',
                    value => $description
                  ) if defined $description;
                my $subRequest = new MAM::Request(
                    database    => $database,
                    object      => 'Quote',
                    action      => 'Create',
                    actor       => $SUPER_USER,
                    assignments => \@quoteAssignments,
                );

            # Invoke the proxy so it prepares the subRequest by setting defaults
                MAM::Proxy->new(
                    database  => $database,
                    request   => $subRequest,
                    requestId => $requestId
                );
                my $subResponse = MAM::Base->create($subRequest, $requestId);
                if ($subResponse->getStatus() eq 'Failure')
                {
                    return new MAM::Response()->failure($subResponse->getCode(),
                        'Unable to create quote: '
                          . $subResponse->getMessage());
                }

                # Figure out the id of the quote we just created
                $newQuoteId = $subResponse->getDatumValue('Id');
                push @usageRecordAssignments,
                  new MAM::Assignment(name => 'Quote', value => $newQuoteId);
                $usageRecordProperties{Quote} = $newQuoteId;
                $chargeProperties{Quote}      = $newQuoteId;
            }
        }

        my $quote           = $request->getOptionValue('Charge');
        my $itemizedCharges = $request->getOptionValue('ItemizedCharges')
          || '';

        # Charge must be a non-negative real number
        if (defined $quote && $quote !~ /^\d*\.?\d+(e\+\d+)?$/)
        {
            return new MAM::Response()->failure('317',
                'The specified quote amount must be a non-negative real number'
            );
        }

        # Calculate charge if not already specified
        if ($accounting_mode ne 'usage-tracking')
        {
            MAM::Bank->calculateCharge(
                request               => $request,
                requestId             => $requestId,
                usageRecordProperties => \%chargeProperties,
                quote                 => $newQuoteId,
                quote_template        => $quoteId,
                duration              => $quoteDuration,
                charge_ref            => \$quote,
                itemizedCharges_ref   => \$itemizedCharges,
                data_ref              => \@data,
                metadata_ref          => \%metadata,
                message_ref           => \$quoteMessage,
            );

            if ($log->is_debug())
            {
                $log->debug("The amount to be quoted is: $quote");
            }
        }
        else { $quote = 0; }

        if ($costOnly)
        {
            $quote =
              sprintf("%.${currency_precision}f", $quote / $time_division);
            my $message = "Successfully quoted $quote credits";
            my $response =
              new MAM::Response()
              ->success($quote, \@data, \%metadata, $message);

            return $response;
        }

        if ($guarantee)
        {
            # Modify the quote amount
            my $subRequest = new MAM::Request(
                database => $database,
                object   => 'Quote',
                action   => 'Modify',
                actor    => $SUPER_USER,
                assignments =>
                  [new MAM::Assignment(name => 'Amount', value => $quote)],
                conditions =>
                  [new MAM::Condition(name => 'Id', value => $newQuoteId)]
            );
            my $subResponse = MAM::Base->modify($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Failure')
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    'Unable to modify quote: ' . $subResponse->getMessage());
            }
        }

        # Log the transaction
        my @quoteOptions = @options;
        if ($accounting_mode ne 'usage-tracking')
        {
            push @quoteOptions,
              ( new MAM::Option(name => 'Amount', value => $quote),
                new MAM::Option(
                    name  => 'ItemizedCharges',
                    value => $itemizedCharges
                )
              );
        }
        if (defined $id)
        {
            push @quoteOptions,
              ( new MAM::Option(name => 'Id',          value => $id),
                new MAM::Option(name => 'UsageRecord', value => $id)
              );
        }
        MAM::Bank->logTransaction(
            database  => $database,
            requestId => $requestId,
            txnId     => $database->nextId('Transaction'),
            object    => 'UsageRecord',
            action    => 'Quote',
            actor     => $actor,
            conditions =>
              [new MAM::Condition(name => 'Amount', value => $quote)],
            assignments => \@usageRecordAssignments,
            options     => \@quoteOptions,
            count       => 1
        );

        # Fail if there are no allocations to quote against
        if (   $accounting_mode ne 'usage-tracking'
            && $accounting_mode ne 'notional-charging'
            && ! scalar keys %allocations)
        {
            my $message =
              'Insufficient funds: There are no valid allocations to satisfy the quote';

            return new MAM::Response()->failure('740', $message);
        }

        # Verify sufficient funds exist to satisfy the quote
        # I want to always check for eligible funds even if quote is 0
        if (   $accounting_mode ne 'usage-tracking'
            && $accounting_mode ne 'notional-charging'
            && $quote >= 0)
        {
            my $remaining = $quote;

            # Factor in liens if performing lien-based allocation enforcement
            if ($accounting_mode ne 'fast-allocation')
            {
# Subtract liens from allocation balances (including my own)
# SELECT GroupBy(LienAllocation.Id),Sum(LienAllocation.Amount) FROM Lien,LienAllocation WHERE Lien.Id=LienAllocation.Lien [AND LienAllocation.Id=<Id> ...]
                my @subSelections = (
                    new MAM::Selection(
                        object => 'LienAllocation',
                        name   => 'Id',
                        op     => 'GroupBy',
                    ),
                    new MAM::Selection(
                        object => 'LienAllocation',
                        name   => 'Amount',
                        op     => 'Sum'
                    ),
                );
                my @subConditions = (
                    new MAM::Condition(
                        object  => 'Lien',
                        name    => 'Id',
                        subject => 'LienAllocation',
                        value   => 'Lien',
                    ),
                );

                # Constrain to list of accepted allocation ids
                my $firstTime = 1;
                my $counter   = 0;
                foreach my $id (keys %allocations)
                {
                    $counter++;
                    my $lastTime =
                      ($counter == scalar keys %allocations) ? 1 : 0;

                    if ($firstTime && $lastTime)
                    {
                        push @subConditions,
                          new MAM::Condition(
                            object => 'LienAllocation',
                            name   => 'Id',
                            value  => $id,
                            conj   => 'And',
                            group  => '0'
                          );
                    }
                    elsif ($firstTime)
                    {
                        $firstTime = 0;
                        push @subConditions,
                          new MAM::Condition(
                            object => 'LienAllocation',
                            name   => 'Id',
                            value  => $id,
                            conj   => 'And',
                            group  => '+1'
                          );
                    }
                    elsif ($lastTime)
                    {
                        push @subConditions,
                          new MAM::Condition(
                            object => 'LienAllocation',
                            name   => 'Id',
                            value  => $id,
                            conj   => 'Or',
                            group  => '-1'
                          );
                    }
                    else
                    {
                        push @subConditions,
                          new MAM::Condition(
                            object => 'LienAllocation',
                            name   => 'Id',
                            value  => $id,
                            conj   => 'Or',
                            group  => '0'
                          );
                    }
                }

                $results = $database->select(
                    objects => [
                        new MAM::Object(name => 'Lien'),
                        new MAM::Object(name => 'LienAllocation'),
                    ],
                    selections => \@subSelections,
                    conditions => \@subConditions
                );
                foreach my $row (@{$results->{data}})
                {
                    my $allocationId = $row->[0];
                    my $lienAmount   = $row->[1];

                    if (exists $allocations{$allocationId})
                    {
                        $allocations{$allocationId}{balance} -= $lienAmount;
                    }
                }
            }

            # Sort the allocations by weight
            my @allocations =
              sort { $allocations{$b}{weight} <=> $allocations{$a}{weight} }
              keys %allocations;

            if ($log->is_debug())
            {
                $log->debug('Sorted allocation list is: ' . join ',',
                    @allocations);
            }

       # Use simple iteration if using active or point-in-time allocation debits
            if ((   ($serverConfig->{'debit.activeallocationonly'} || 'false')
                    =~ /^true$/i
                )
                || (($serverConfig->{'debit.pointintimeonly'} || 'false') =~
                    /^true$/i)
              )
            {
                # Iterate through the allocation list
                foreach my $allocationId (@allocations)
                {
                    # Break out if we are done
                    last if $remaining <= 0;

# The amount I subtract from this allocation should be the minimum of:
# 1. $remaining (the amount left to withdraw)
# 2. $allocations{$allocationId}{balance} (the amount available within this allocation (considering liens and credit))
                    $balance = $allocations{$allocationId}{balance};
                    my $subQuote = min($remaining, $balance);
                    next unless $subQuote > 0;
                    $quoteChunk{$allocationId} += $subQuote;
                    $remaining -= $subQuote;
                    if ($log->is_trace())
                    {
                        $log->trace(
                            "Quoting $subQuote credits from allocation $allocationId"
                        );
                    }
                }
            }

            # Iterate over segment boundaries if using uniform allocation debits
            else
            {
                # Build up the quote segment edges
                my %edge = ();
                $edge{$quoteStartTime}++;
                $edge{$quoteEndTime}++;
                foreach my $allocationId (@allocations)
                {
                    my $allocationStartTime =
                      $allocations{$allocationId}{startTime};
                    my $allocationEndTime =
                      $allocations{$allocationId}{endTime};
                    $edge{$allocationStartTime}++
                      if $allocationStartTime > $quoteStartTime;
                    $edge{$allocationEndTime}++
                      if $allocationEndTime < $quoteEndTime;
                }

                # Iterate over quote segments, building up quote chunks
                my ($allocationFound, $leadingEdge);
                my @continuityMessages = ();
                # First pass going for uniform proportional distribution
                if ($remaining > 0)
                {
                    foreach my $trailingEdge (sort keys %edge)
                    {
                      # Handle case where quote start time equals quote end time
                        if (scalar keys %edge == 1)
                        {
                            $leadingEdge = $trailingEdge;
                        }

                        if (defined $leadingEdge)
                        {
                            my $proportion =
                              ($quoteStartTime == $quoteEndTime)
                              ? sprintf("%.${currency_precision}f",
                                ($quote / $time_division))
                              : sprintf(
                                "%.${currency_precision}f",
                                (   ($quote * ($trailingEdge - $leadingEdge)) /
                                      ( $time_division *
                                          ($quoteEndTime - $quoteStartTime)
                                      )
                                )
                              );
                            my $subRemaining = min($proportion, $remaining);
                         # Final segment should incorporate loss due to rounding
                            if (($trailingEdge == $quoteEndTime)
                                && ($remaining - $subRemaining <=
                                    10**(-$currency_precision))
                              )
                            {
                                $subRemaining = $remaining;
                            }
                            next unless $subRemaining > 0;
                            $allocationFound = 0;
                            foreach my $allocationId (@allocations)
                            {
                                if ((   $allocations{$allocationId}{startTime}
                                        <= $leadingEdge
                                    )
                                    && ($allocations{$allocationId}{endTime} >=
                                        $trailingEdge)
                                  )
                                {
                                    $balance =
                                      $allocations{$allocationId}{balance};
                                    my $subQuote = min($subRemaining, $balance);
                                    next unless $subQuote > 0;
                                    $quoteChunk{$allocationId} += $subQuote;
                                    $subRemaining -= $subQuote;
                                    $remaining    -= $subQuote;
                                    $allocations{$allocationId}{balance} -=
                                      $subQuote;
                                    $allocationFound = 1;

                                    if ($log->is_trace())
                                    {
                                        $log->trace(
                                            "Quoting $subQuote credits from allocation $allocationId for interval ($leadingEdge, $trailingEdge)"
                                        );
                                    }
                                    last if $subRemaining <= 0;
                                }
                            }
                            if (! $allocationFound)
                            {
                                my $continuityMessage = sprintf(
                                    "Allocation discontinuity of %.${currency_precision}f credits detected between %s and %s.",
                                    $subRemaining, toHRT($leadingEdge),
                                    toHRT($trailingEdge));
                                if ((   $serverConfig->{
                                            'debit.requirecontinuousallocations'
                                        }
                                        || 'false'
                                    ) =~ /^true$/i
                                  )
                                {
                                    return new MAM::Response()
                                      ->failure('740', $continuityMessage);
                                }
                                elsif (
                                    (   $serverConfig->{
                                            'debit.suppresscontinuitywarnings'}
                                        || 'false'
                                    ) !~ /^true$/i
                                  )
                                {
                                    push @continuityMessages,
                                      $continuityMessage;
                                }
                                $log->trace($continuityMessage)
                                  if $log->is_trace();
                            }
                            elsif ($subRemaining > 0)
                            {
                                my $continuityMessage = sprintf(
                                    "Allocation deficiency of %.${currency_precision}f credits detected between %s and %s.",
                                    $subRemaining, toHRT($leadingEdge),
                                    toHRT($trailingEdge));
                                if ((   $serverConfig->{
                                            'debit.requirecontinuousallocations'
                                        }
                                        || 'false'
                                    ) =~ /^true$/i
                                  )
                                {
                                    return new MAM::Response()
                                      ->failure('740', $continuityMessage);
                                }
                                elsif (
                                    (   $serverConfig->{
                                            'debit.suppresscontinuitywarnings'}
                                        || 'false'
                                    ) !~ /^true$/i
                                  )
                                {
                                    push @continuityMessages,
                                      $continuityMessage;
                                }
                                $log->trace($continuityMessage)
                                  if $log->is_trace();
                            }
                        }

                        last if $remaining <= 0;
                        $leadingEdge = $trailingEdge;
                    }
                }

            # If there were gaps in the first pass allocation, initiate a second
            # pass going from earliest to latest and allocating the maximum
            # possible in each allocation segment
                if ($remaining > 0)
                {
                    undef $leadingEdge;
                    foreach my $trailingEdge (sort keys %edge)
                    {
                        if (defined $leadingEdge)
                        {
                            foreach my $allocationId (@allocations)
                            {
                                if ((   $allocations{$allocationId}{startTime}
                                        <= $leadingEdge
                                    )
                                    && ($allocations{$allocationId}{endTime} >=
                                        $trailingEdge)
                                  )
                                {
                                    $balance =
                                      $allocations{$allocationId}{balance};
                                    my $subQuote = min($remaining, $balance);
                                    next unless $subQuote > 0;
                                    $quoteChunk{$allocationId} += $subQuote;
                                    $remaining -= $subQuote;
                                    $allocations{$allocationId}{balance} -=
                                      $subQuote;
                                    my $continuityMessage = sprintf(
                                        "Continuity deficit of %.${currency_precision}f credits satisfied from allocation $allocationId.",
                                        $subQuote);
                                    if ((   $serverConfig->{
                                                'debit.suppresscontinuitywarnings'
                                            }
                                            || 'false'
                                        ) !~ /^true$/i
                                      )
                                    {
                                        push @continuityMessages,
                                          $continuityMessage;
                                    }
                                    $log->trace($continuityMessage)
                                      if $log->is_trace();
                                    if ($log->is_trace())
                                    {
                                        $log->trace(
                                            "Quoting $subQuote credits from allocation $allocationId to satisfy continuity deficit"
                                        );
                                    }
                                    last if $remaining <= 0;
                                }
                            }
                        }

                        last if $remaining <= 0;
                        $leadingEdge = $trailingEdge;
                    }
                }

                if (@continuityMessages)
                {
                    $warningMessage .= "\nWarning -- " . join "\n",
                      @continuityMessages;
                    $warnings++;
                }
            }

# use Data::Dumper;
# print Data::Dumper->Dump([\%quoteChunk], ['quoteChunk']), "\n";
# print Data::Dumper->Dump([\@continuityMessages], ['continuityMessages']), "\n";

            # Fail if not all amount could be debited
            if ($remaining > 0)
            {
                my $message = 'Insufficient balance to satisfy quote';
                $message .= " (Instance $instance)" if defined $instance;

                # Append balance info to data if balance feedback requested
                if (   $balanceFeedback
                    && $accounting_mode ne 'usage-tracking'
                    && $accounting_mode ne 'notional-charging'
                    && scalar @data == 1)
                {
                    my $response =
                      new MAM::Response()->failure('782', $message);
                    my $datum = new MAM::Datum($data[0]->{_element});
                    $datum->delProperty('Amount');
                    $datum->setProperty('Balance', $balance);
                    $response->addDatum($datum);
                    delete $metadata{'Amount'};
                    $metadata{'Balance'} = 'Currency';
                    $response->setMetadata(\%metadata);
                    return $response;
                }
                else
                {
                    return new MAM::Response()->failure('782', $message);
                }
            }
        }

        # Modify the usage record with gleaned attributes
        if (defined $id)
        {
            # Modify the existing usage record
            my @usageRecordOptions = ();
            push @usageRecordOptions,
              new MAM::Option(name => 'Instance', value => $instance)
              if defined $instance;
            my $subRequest = new MAM::Request(
                database    => $database,
                object      => 'UsageRecord',
                action      => 'Modify',
                actor       => $SUPER_USER,
                conditions  => [new MAM::Condition(name => 'Id', value => $id)],
                options     => \@usageRecordOptions,
                assignments => \@usageRecordAssignments,
            );
            my $subResponse = MAM::Base->modify($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Failure')
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    'Unable to modify usage record: '
                      . $subResponse->getMessage());
            }
        }

        # Append balance info to data if balance feedback requested
        if (   $balanceFeedback
            && $accounting_mode ne 'usage-tracking'
            && $accounting_mode ne 'notional-charging'
            && scalar @data == 1)
        {
            my $datumElement   = $data[0]->{_element};
            my $lastChild      = $datumElement->lastChild;
            my $balanceElement = new XML::LibXML::Element('Balance');
            $balanceElement->appendText($balance);
            $datumElement->insertAfter($balanceElement, $lastChild);
            $metadata{'Balance'} = 'Currency';
        }

        $quote = sprintf("%.${currency_precision}f", $quote / $time_division);

        if ($accounting_mode eq 'usage-tracking')
        {
            $message = "Successfully recorded usage information";
        }
        else
        {
            $message = "Successfully quoted $quote credits";
        }
        $message .= " for instance $instance" if defined $instance;
        $message .= " with quote id $newQuoteId"
          if ($guarantee && $accounting_mode ne 'usage-tracking');
        $message .= " and usage record id $id" if $createdUsageRecord;
        if ($warnings)
        {
            $response =
              new MAM::Response()
              ->warning($quote, \@data, \%metadata,
                $message . $warningMessage . $quoteMessage);
        }
        else
        {
            $response = new MAM::Response()
              ->success($quote, \@data, \%metadata, $message . $quoteMessage);
        }
        return $response;
    }
    else
    {
        return new MAM::Response()
          ->failure('315', "The Bank class does not implement $object Quote");
    }
}

# ----------------------------------------------------------------------------
# $response = refund($request, $requestId);
# ----------------------------------------------------------------------------

# Refund (UsageRecord)
sub refund
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object          = $request->getObject();
    my $actor           = $request->getActor();
    my $database        = $request->getDatabase();
    my $instance        = $request->getOptionValue('Instance');
    my $id              = $request->getOptionValue('Id');
    my $amount          = $request->getOptionValue('Amount');
    my $allocation      = $request->getOptionValue('Allocation');
    my @options         = $request->getOptions();
    my $now             = $requestTime;
    my $accounting_mode = $serverConfig->{'accounting.mode'}
      || $ACCOUNTING_MODE;
    my $accountingModeOption = $request->getOptionValue('AccountingMode');
    my $currency_precision   = $serverConfig->{'currency.precision'} || 0;
    my $showHours            = $request->getOptionValue('ShowHours');
    my $time_division = (defined $showHours && $showHours eq 'True') ? 3600 : 1;
    my $refundSum     = 0;
    my $warningMessage = '';
    my $warnings       = 0;

    # Scale amount into seconds if showHours was specified
    $amount = $amount * $time_division if $time_division > 1;

    if ($object ne 'UsageRecord')
    {
        return new MAM::Response()
          ->failure('315', "The Bank class does not implement $object Refund");
    }

    # Either Id or Instance must be specified
    if ((! defined $id) && (! defined $instance))
    {
        return new MAM::Response()->failure('314',
            'Either a usage record id (Id) or an instance name (Intance) must be specified'
        );
    }

    # Demote accounting mode if option is lower than configuration
    if (defined $accountingModeOption)
    {
        if ((   $accountingModeOption eq 'usage-tracking'
                && (   $accounting_mode eq 'usage-tracking'
                    || $accounting_mode eq 'notional-charging'
                    || $accounting_mode eq 'fast-allocation'
                    || $accounting_mode eq 'strict-allocation')
            )
            || ($accountingModeOption eq 'notional-charging'
                && (   $accounting_mode eq 'notional-charging'
                    || $accounting_mode eq 'fast-allocation'
                    || $accounting_mode eq 'strict-allocation')
            )
            || ($accountingModeOption eq 'fast-allocation'
                && (   $accounting_mode eq 'fast-allocation'
                    || $accounting_mode eq 'strict-allocation')
            )
            || (   $accountingModeOption eq 'strict-allocation'
                && $accounting_mode eq 'strict-allocation')
          )
        {
            $accounting_mode = $accountingModeOption;
        }
        else
        {
            $warningMessage .=
              "\nWarning -- The specified accounting mode ($accountingModeOption) is being ignored because it is more strict than the accounting mode configured in Moab Accounting Manager ($accounting_mode).";
            $warnings++;
        }
    }

    if ($accounting_mode eq 'usage-tracking')
    {
        return new MAM::Response()
          ->warning(0,
            "No action taken since accounting mode is 'usage-tracking'");
    }

# First look for the usage to be refunded in the UsageRecord Table
# SELECT * FROM UsageRecord WHERE (Instance=<instance_name>|Id=<usage_record_id>)

    my $charge;
    my @subSelections = (
        new MAM::Selection(name => 'Id'),
        new MAM::Selection(name => 'Instance'),
        new MAM::Selection(name => 'Charge'),
    );
    push @subSelections, new MAM::Selection(name => 'User')
      if MAM::Cache->attributeExists('UsageRecord', 'User');
    push @subSelections, new MAM::Selection(name => 'Account')
      if MAM::Cache->attributeExists('UsageRecord', 'Account');
    push @subSelections, new MAM::Selection(name => 'Machine')
      if MAM::Cache->attributeExists('UsageRecord', 'Machine');

    my @subConditions = ();
    push @subConditions,
      new MAM::Condition(
        name  => 'Instance',
        value => $instance
      ) if defined $instance;
    push @subConditions,
      new MAM::Condition(
        name  => 'Id',
        value => $id
      ) if defined $id;

    my @subOptions = (new MAM::Option(name => 'Limit', value => '20'));
    my $results = $database->select(
        object     => 'UsageRecord',
        selections => \@subSelections,
        conditions => \@subConditions,
        options    => \@subOptions
    );

    # No usage records were matched
    if (@{$results->{data}} == 0)
    {
        return new MAM::Response()->failure('740',
            'There are no usage records with charges matching the specified instance name'
        );
    }

    # Multiple usage records were matched
    elsif (@{$results->{data}} > 1)
    {
        my $usageRecords = '';
        foreach my $row (@{$results->{data}})
        {
            my $col = 0;
            my ($user, $account, $machine);
            my $id       = $row->[$col++];
            my $instance = $row->[$col++];
            my $charge   = $row->[$col++];
            $user = $row->[$col++]
              if MAM::Cache->attributeExists('UsageRecord', 'User');
            $account = $row->[$col++]
              if MAM::Cache->attributeExists('UsageRecord', 'Account');
            $machine = $row->[$col++]
              if MAM::Cache->attributeExists('UsageRecord', 'Machine');

            $usageRecords .= "\t$id\t[instance=$instance";
            $usageRecords .= ",user=$user" if defined $user;
            $usageRecords .= ",account=$account" if defined $account;
            $usageRecords .= ",machine=$machine" if defined $machine;
            $usageRecords .= ",charge=$charge]\n";
        }
        return new MAM::Response()->failure('740',
            "There are multiple usage records matching the specified instance name.\nPlease respecify using one of the following usage record ids:\n"
              . $usageRecords);
    }

    # A unique usage record was matched
    else
    {
        my $row = ${$results->{data}}[0];
        $id       = $row->[0];
        $instance = $row->[1];
        $charge   = $row->[2];
    }

    # Amount must be positive and not greater than charge
    if (defined $amount)
    {
        if ($amount > $charge)
        {
            return new MAM::Response()->failure('740',
                "Refund amount ($amount) cannot exceed charged amount ($charge)"
            );
        }
    }
    else
    {
        $amount = $charge ? $charge : 0;
    }

    # Amount must be a non-negative real number
    if ($amount !~ /^\d*\.?\d+(e\+\d+)?$/)
    {
        return new MAM::Response()
          ->failure('317',
            'The refund amount must be a non-negative real number');
    }

    # Fail if there are no charges to refund
    if ($amount == 0)
    {
        return new MAM::Response()
          ->failure('740', 'There are no charges to refund');
    }

    # If the allocation is specified we can just issue the refund
    if (   $accounting_mode ne 'usage-tracking'
        && $accounting_mode ne 'notional-charging'
        && defined $allocation)
    {
        # See if allocation exists and obtain its fund, start time and end time
        # SELECT Fund,EndTime,StartTime FROM Allocation WHERE Id=$allocation

        my $results = $database->select(
            object     => 'Allocation',
            selections => [
                new MAM::Selection(name => 'Fund'),
                new MAM::Selection(name => 'StartTime'),
                new MAM::Selection(name => 'EndTime'),
            ],
            conditions => [
                new MAM::Condition(
                    name  => 'Id',
                    value => $allocation,
                )
            ],
        );

        unless (@{$results->{data}})
        {
            return new MAM::Response()
              ->failure('740',
                "The specified allocation ($allocation) does not exist.");
        }

        my $fund                = ${$results->{data}}[0]->[0];
        my $allocationStartTime = ${$results->{data}}[0]->[1];
        my $allocationEndTime   = ${$results->{data}}[0]->[2];

        # Credit the allocation
        my $subRequest = new MAM::Request(
            database    => $database,
            object      => 'Allocation',
            action      => 'Modify',
            actor       => $SUPER_USER,
            assignments => [
                new MAM::Assignment(
                    name  => 'Amount',
                    value => $amount,
                    op    => 'Inc'
                )
            ],
            conditions => [
                new MAM::Condition(
                    name  => 'Fund',
                    value => $fund
                ),
                new MAM::Condition(
                    name  => 'Id',
                    value => $allocation
                )
            ]
        );
        my $subResponse = MAM::Base->modify($subRequest, $requestId, 'NoLog');
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Unable to credit allocation: ' . $subResponse->getMessage());
        }

        # Log the transaction
        MAM::Bank->logTransaction(
            database    => $database,
            requestId   => $requestId,
            txnId       => $database->nextId('Transaction'),
            object      => 'Allocation',
            action      => 'Modify',
            actor       => $actor,
            assignments => [
                new MAM::Assignment(
                    name  => 'Amount',
                    value => $amount,
                )
            ],
            conditions => [
                new MAM::Condition(
                    name  => 'Id',
                    value => $allocation
                )
            ],
            count      => 1,
            fund       => $fund,
            allocation => $allocation
        );

        # Obtain remaining allocation amount
        $results = $database->select(
            object     => 'Allocation',
            selections => [new MAM::Selection(name => 'Amount')],
            conditions => [
                new MAM::Condition(
                    name  => 'Id',
                    value => $allocation
                )
            ]
        );
        my $remainingAllocation = ${$results->{data}}[0]->[0];

        # Log the transaction
        my ($balance, $delta);
        if (   $allocationStartTime <= $now
            && $allocationEndTime > $now)
        {
            $delta   = $amount;
            $balance = $remainingAllocation;
        }
        else
        {
            $delta   = 0;
            $balance = 0;
        }

        # Log the transaction
        MAM::Bank->logTransaction(
            database  => $database,
            requestId => $requestId,
            txnId     => $database->nextId('Transaction'),
            object    => 'UsageRecord',
            action    => 'Refund',
            actor     => $actor,
            options   => [
                @options,
                new MAM::Option(
                    name  => 'Amount',
                    value => $amount
                ),
                new MAM::Option(
                    name  => 'Id',
                    value => $id
                ),
                new MAM::Option(
                    name  => 'UsageRecord',
                    value => $id
                ),
                new MAM::Option(
                    name  => 'Instance',
                    value => $instance
                ),
            ],
            count      => 1,
            fund       => $fund,
            delta      => $delta,
            balance    => $balance,
            remaining  => $remainingAllocation,
            allocation => $allocation,
        );
        $refundSum += $amount;
    }

# Else if the mode is strict or fast allocation,
# look up the charges from the transaction log to see what allocations to charge
    elsif ($accounting_mode ne 'usage-tracking'
        && $accounting_mode ne 'notional-charging')
    {
        my $remaining = $amount;

# SELECT Allocation,Amount FROM Transaction WHERE Object=UsageRecord AND Action=Charge AND UsageRecord=$id
        $results = $database->select(
            object     => 'Transaction',
            selections => [
                new MAM::Selection(name => 'Id'),
                new MAM::Selection(name => 'Fund'),
                new MAM::Selection(name => 'Allocation'),
                new MAM::Selection(name => 'Amount'),
            ],
            conditions => [
                new MAM::Condition(
                    name  => 'Object',
                    value => 'UsageRecord'
                ),
                new MAM::Condition(
                    name  => 'Action',
                    value => 'Charge'
                ),
                new MAM::Condition(
                    name  => 'UsageRecord',
                    value => $id
                )
            ]
        );
        foreach my $row (@{$results->{data}})
        {
            my $transaction       = $row->[0];
            my $fund              = $row->[1];
            my $allocation        = $row->[2];
            my $transactionAmount = $row->[3];

            my $subRefund = min($transactionAmount, $remaining);
            next unless $subRefund;

            unless (defined $allocation)
            {
                return new MAM::Response()->failure('740',
                    "No allocation is recorded for the charge of amount $transactionAmount in transaction $transaction. Please reissue the refund and specify an allocation to refund into."
                );
            }

            # Credit the allocation
            my $subRequest = new MAM::Request(
                database    => $database,
                object      => 'Allocation',
                action      => 'Modify',
                actor       => $SUPER_USER,
                assignments => [
                    new MAM::Assignment(
                        name  => 'Amount',
                        value => $subRefund,
                        op    => 'Inc'
                    )
                ],
                conditions => [
                    new MAM::Condition(
                        name  => 'Fund',
                        value => $fund
                    ),
                    new MAM::Condition(
                        name  => 'Id',
                        value => $allocation
                    )
                ]
            );
            my $subResponse =
              MAM::Base->modify($subRequest, $requestId, 'NoLog');
            if ($subResponse->getStatus() eq 'Failure')
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    'Unable to credit allocation: '
                      . $subResponse->getMessage());
            }

            # Log the transaction
            MAM::Bank->logTransaction(
                database    => $database,
                requestId   => $requestId,
                txnId       => $database->nextId('Transaction'),
                object      => 'Allocation',
                action      => 'Modify',
                actor       => $actor,
                assignments => [
                    new MAM::Assignment(
                        name  => 'Amount',
                        value => $subRefund,
                    )
                ],
                conditions => [
                    new MAM::Condition(
                        name  => 'Id',
                        value => $allocation
                    )
                ],
                count      => 1,
                fund       => $fund,
                allocation => $allocation
            );

            # Obtain remaining allocation amount, start time and end time
            my $results = $database->select(
                object     => 'Allocation',
                selections => [
                    new MAM::Selection(name => 'Amount'),
                    new MAM::Selection(name => 'StartTime'),
                    new MAM::Selection(name => 'EndTime'),
                ],
                conditions => [
                    new MAM::Condition(
                        name  => 'Id',
                        value => $allocation
                    )
                ]
            );
            my $remainingAllocation = ${$results->{data}}[0]->[0];
            my $allocationStartTime = ${$results->{data}}[0]->[1];
            my $allocationEndTime   = ${$results->{data}}[0]->[2];

            # Log the transaction
            my ($balance, $delta);
            if (   $allocationStartTime <= $now
                && $allocationEndTime > $now)
            {
                $delta   = $subRefund;
                $balance = $remainingAllocation;
            }
            else
            {
                $delta   = 0;
                $balance = 0;
            }
            MAM::Bank->logTransaction(
                database  => $database,
                requestId => $requestId,
                txnId     => $database->nextId('Transaction'),
                object    => 'UsageRecord',
                action    => 'Refund',
                actor     => $actor,
                options   => [
                    @options,
                    new MAM::Option(
                        name  => 'Amount',
                        value => $subRefund
                    ),
                    new MAM::Option(name => 'Id', value => $id),
                    new MAM::Option(
                        name  => 'UsageRecord',
                        value => $id
                    ),
                    new MAM::Option(
                        name  => 'Instance',
                        value => $instance
                    ),
                ],
                count      => 1,
                fund       => $fund,
                delta      => $delta,
                balance    => $balance,
                remaining  => $remainingAllocation,
                allocation => $allocation,
            );
            $refundSum += $subRefund;
            $remaining -= $subRefund;
            last unless $remaining > 0;
        }

        # Bail if transaction log charge data was insufficient to determine
        # the entirety of the funds and amounts to refund
        if ($remaining > 0)
        {
            return new MAM::Response()->failure('740',
                "The usage charge data could not be found in the transaction logs.\nPlease specify the allocation to credit."
            );
        }
    }

    # If this is notional charging, we just log the transaction
    elsif ($accounting_mode eq 'notional-charging')
    {
        # Log the transaction
        MAM::Bank->logTransaction(
            database  => $database,
            requestId => $requestId,
            txnId     => $database->nextId('Transaction'),
            object    => 'UsageRecord',
            action    => 'Refund',
            actor     => $actor,
            options   => [
                @options,
                new MAM::Option(
                    name  => 'Amount',
                    value => $amount
                ),
                new MAM::Option(name => 'Id', value => $id),
                new MAM::Option(
                    name  => 'UsageRecord',
                    value => $id
                ),
                new MAM::Option(
                    name  => 'Instance',
                    value => $instance
                ),
            ],
            count => 1,
        );
        $refundSum += $amount;
    }

    # Update the usage record amount
    my $subRequest = new MAM::Request(
        database    => $database,
        object      => 'UsageRecord',
        action      => 'Modify',
        actor       => $SUPER_USER,
        assignments => [
            new MAM::Assignment(
                name  => 'Charge',
                value => $amount,
                op    => 'Dec'
            )
        ],
        conditions => [
            new MAM::Condition(
                name  => 'Id',
                value => $id
            )
        ]
    );
    my $subResponse = MAM::Base->modify($subRequest, $requestId);
    if ($subResponse->getStatus() eq 'Failure')
    {
        return new MAM::Response()->failure($subResponse->getCode(),
            'Unable to modify usage record: ' . $subResponse->getMessage());
    }

    # Build response
    my $response;
    $refundSum =
      sprintf("%.${currency_precision}f", $refundSum / $time_division);
    my $message = "Successfully refunded $refundSum credits";
    $message .= " to usage record $id"    if defined $id;
    $message .= " for instance $instance" if defined $instance;
    $message .= $warningMessage           if $warnings;
    if ($warnings)
    {
        $response = new MAM::Response()->warning($refundSum, $message);
    }
    else
    {
        $response = new MAM::Response()->success($refundSum, $message);
    }

    # Add data to response
    my $datum = new MAM::Datum('UsageRecord');
    $datum->setProperty('Id',       $id)       if defined $id;
    $datum->setProperty('Instance', $instance) if defined $instance;
    $datum->setProperty('Amount',   $refundSum);
    $response->addDatum($datum);
    $response->setDataType('Id', 'Integer');
    $response->setDataType('Instance', 'String');
    $response->setDataType('Amount', 'Currency');

    return $response;
}

# ----------------------------------------------------------------------------
# $response = refresh($request, $requestId);
# ----------------------------------------------------------------------------

# Refresh (Allocation)
sub refresh
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object          = $request->getObject();
    my $actor           = $request->getActor();
    my $database        = $request->getDatabase();
    my @options         = $request->getOptions();
    my $now             = $requestTime;
    my $activationSum   = 0;
    my $deactivationSum = 0;
    my $results;
    my $currency_precision = $serverConfig->{'currency.precision'} || 0;
    my $time_division =
      defined $request->getOptionValue('ShowHours') ? 3600 : 1;

    if ($object ne 'Allocation')
    {
        return new MAM::Response()
          ->failure('315', "The Bank class does not implement $object Refresh");
    }

    # Activate allocations that are newly activated or no longer expired

# Get a list of allocations to be activated
# SELECT Id,Fund,Amount FROM Allocation WHERE ( StartTime<=$now AND EndTime>$now ) AND Active='False'
    $results = $database->select(
        object     => 'Allocation',
        selections => [
            new MAM::Selection(name => 'Id'),
            new MAM::Selection(name => 'Fund'),
            new MAM::Selection(name => 'Amount')
        ],
        conditions => [
            new MAM::Condition(
                name  => 'StartTime',
                value => $now,
                op    => 'LE',
                conj  => 'And',
                group => '+1'
            ),
            new MAM::Condition(
                name  => 'EndTime',
                value => $now,
                op    => 'GT',
                conj  => 'And',
                group => '-1'
            ),
            new MAM::Condition(name => 'Active', value => 'False')
        ]
    );
    foreach my $row (@{$results->{data}})
    {
        my $id     = $row->[0];
        my $fundId = $row->[1];
        my $amount = $row->[2];

        # Construct an individual modify based on fund id in list
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'Allocation',
            action   => 'Modify',
            actor    => $SUPER_USER,
            assignments =>
              [new MAM::Assignment(name => 'Active', value => 'True')],
            conditions => [
                new MAM::Condition(
                    name  => 'Fund',
                    value => $fundId
                ),
                new MAM::Condition(name => 'Id', value => $id)
            ]
        );
        my $subResponse = MAM::Base->modify($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                "Unable to activate allocation ($id): "
                  . $subResponse->getMessage());
        }

        $activationSum += $amount;

        # Log the transaction
        my @activateOptions = @options;
        push @activateOptions, new MAM::Option(name => 'Id', value => $id);
        MAM::Bank->logTransaction(
            database   => $database,
            requestId  => $requestId,
            txnId      => $database->nextId('Transaction'),
            object     => 'Allocation',
            action     => 'Activate',
            actor      => $actor,
            options    => \@activateOptions,
            count      => 1,
            fund       => $fundId,
            delta      => $amount,
            balance    => $amount,
            remaining  => $amount,
            allocation => $id
        );
    }

    # Deactivate allocations that are newly expired or no longer active

# Get a list of allocations to be deactivated
# SELECT Id,Fund,Amount FROM Allocation WHERE ( StartTime>$now OR EndTime<=$now ) AND Active='True'
    $results = $database->select(
        object     => 'Allocation',
        selections => [
            new MAM::Selection(name => 'Id'),
            new MAM::Selection(name => 'Fund'),
            new MAM::Selection(name => 'Amount')
        ],
        conditions => [
            new MAM::Condition(
                name  => 'StartTime',
                value => $now,
                op    => 'GT',
                conj  => 'And',
                group => '+1'
            ),
            new MAM::Condition(
                name  => 'EndTime',
                value => $now,
                op    => 'LE',
                conj  => 'Or',
                group => '-1'
            ),
            new MAM::Condition(name => 'Active', value => 'True')
        ]
    );
    foreach my $row (@{$results->{data}})
    {
        my $id     = $row->[0];
        my $fundId = $row->[1];
        my $amount = $row->[2];

        # Construct an individual modify based on fund id in list
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'Allocation',
            action   => 'Modify',
            actor    => $SUPER_USER,
            assignments =>
              [new MAM::Assignment(name => 'Active', value => 'False')],
            conditions => [
                new MAM::Condition(
                    name  => 'Fund',
                    value => $fundId
                ),
                new MAM::Condition(name => 'Id', value => $id)
            ]
        );
        my $subResponse = MAM::Base->modify($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                "Unable to deactivate allocation ($id): "
                  . $subResponse->getMessage());
        }

        $deactivationSum += $amount;

        # Log the transaction
        my @deactivateOptions = @options;
        push @deactivateOptions, new MAM::Option(name => 'Id', value => $id);
        MAM::Bank->logTransaction(
            database   => $database,
            requestId  => $requestId,
            txnId      => $database->nextId('Transaction'),
            object     => 'Allocation',
            action     => 'Deactivate',
            actor      => $actor,
            options    => \@deactivateOptions,
            count      => 1,
            fund       => $fundId,
            delta      => 0 - $amount,
            balance    => 0,
            remaining  => $amount,
            allocation => $id
        );
    }

    # Repair allocations that have Active == NULL
    my $repairCount = 0;

    # Get a list of NULL allocations
    # SELECT Id,Fund,StartTime,EndTime FROM Allocation WHERE Active IS NULL
    $results = $database->select(
        object     => 'Allocation',
        selections => [
            new MAM::Selection(name => 'Id'),
            new MAM::Selection(name => 'Fund'),
            new MAM::Selection(name => 'StartTime'),
            new MAM::Selection(name => 'EndTime')
        ],
        conditions => [new MAM::Condition(name => 'Active', value => 'NULL')]
    );
    foreach my $row (@{$results->{data}})
    {
        my $id        = $row->[0];
        my $fundId    = $row->[1];
        my $startTime = $row->[2];
        my $endTime   = $row->[3];

        # Construct an individual modify based on fund id in list
        my $activeness;
        if ($startTime <= $now && $endTime > $now)
        {
            $activeness = 'True';
        }
        else { $activeness = 'False'; }
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'Allocation',
            action   => 'Modify',
            actor    => $SUPER_USER,
            assignments =>
              [new MAM::Assignment(name => 'Active', value => $activeness)],
            conditions => [
                new MAM::Condition(
                    name  => 'Fund',
                    value => $fundId
                ),
                new MAM::Condition(name => 'Id', value => $id)
            ]
        );
        my $subResponse = MAM::Base->modify($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                "Unable to repair allocation ($id) to set Active=$activeness: "
                  . $subResponse->getMessage());
        }
        $repairCount++;

        # Log the transaction
        MAM::Bank->logTransaction(
            database   => $database,
            requestId  => $requestId,
            txnId      => $database->nextId('Transaction'),
            object     => 'Allocation',
            action     => 'Repair',
            actor      => $actor,
            count      => 1,
            fund       => $fundId,
            allocation => $id
        );
    }

    $activationSum =
      sprintf("%.${currency_precision}f", $activationSum / $time_division);
    $deactivationSum =
      sprintf("%.${currency_precision}f", $deactivationSum / $time_division);
    my $message =
      "Successfully activated $activationSum credits and deactivated $deactivationSum credits";
    $message .= " and repaired $repairCount allocations" if $repairCount;
    return new MAM::Response()->success($message);
}

# ----------------------------------------------------------------------------
# $response = reserve($request, $requestId);
# ----------------------------------------------------------------------------

# Reserve (UsageRecord)
sub reserve
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object          = $request->getObject();
    my $actor           = $request->getActor();
    my $database        = $request->getDatabase();
    my @options         = $request->getOptions();
    my $now             = $requestTime;
    my $message         = '';
    my $quoteMessage    = '';
    my $reserveMessage  = '';
    my $warningMessage  = '';
    my $warnings        = 0;
    my $accounting_mode = $serverConfig->{'accounting.mode'}
      || $ACCOUNTING_MODE;
    my $accountingModeOption = $request->getOptionValue('AccountingMode');
    my $currency_precision = $serverConfig->{'currency.precision'} || 0;
    my $time_division =
      defined $request->getOptionValue('ShowHours') ? 3600 : 1;
    my $createdUsageRecord = 0;
    my $balance            = 0;
    my $infinity           = 2147483647;
    my %allocations        = ();
    my @data               = ();
    my %metadata           = ();
    my %fundAllocations    = ();
    my %lienChunk          = ();
    my $results;
    my $response;

    if ($object ne 'UsageRecord')
    {
        return new MAM::Response()
          ->failure('315', "The Bank class does not implement $object Reserve");
    }

    my $replace;
    my $replaceOption = $request->getOptionValue('Replace');
    if (defined $replaceOption
        && $replaceOption =~ /^true$/i)
    {
        $replace = 1;
    }
    else
    {
        $replace = 0;
    }

    my $modify;
    my $modifyOption = $request->getOptionValue('Modify');
    if (defined $modifyOption
        && $modifyOption =~ /^true$/i)
    {
        $modify = 1;
    }
    else
    {
        $modify = 0;
    }

    my $balanceFeedback;
    my $balanceFeedbackOption = $request->getOptionValue('BalanceFeedback');
    if (defined $balanceFeedbackOption
        && $balanceFeedbackOption =~ /^true$/i)
    {
        $balanceFeedback = 1;
    }
    else
    {
        $balanceFeedback = 0;
    }

    my $itemize;
    my $itemizeOption = $request->getOptionValue('Itemize');
    if (defined $itemizeOption && $itemizeOption =~ /^true$/i)
    {
        $itemize = 1;
    }
    else
    {
        $itemize = 0;
    }

    if ($balanceFeedback && $itemize)
    {
        return new MAM::Response()
          ->failure('318',
            'The Itemize and BalanceFeedback options are mutually exclusive');
    }

    # Demote accounting mode if option is lower than configuration
    if (defined $accountingModeOption)
    {
        if ((   $accountingModeOption eq 'usage-tracking'
                && (   $accounting_mode eq 'usage-tracking'
                    || $accounting_mode eq 'notional-charging'
                    || $accounting_mode eq 'fast-allocation'
                    || $accounting_mode eq 'strict-allocation')
            )
            || ($accountingModeOption eq 'notional-charging'
                && (   $accounting_mode eq 'notional-charging'
                    || $accounting_mode eq 'fast-allocation'
                    || $accounting_mode eq 'strict-allocation')
            )
            || ($accountingModeOption eq 'fast-allocation'
                && (   $accounting_mode eq 'fast-allocation'
                    || $accounting_mode eq 'strict-allocation')
            )
            || (   $accountingModeOption eq 'strict-allocation'
                && $accounting_mode eq 'strict-allocation')
          )
        {
            $accounting_mode = $accountingModeOption;
        }
        else
        {
            $warningMessage .=
              "\nWarning -- The specified accounting mode ($accountingModeOption) is being ignored because it is more strict than the accounting mode configured in Moab Accounting Manager ($accounting_mode).";
            $warnings++;
        }
    }

    if ($replace && $modify)
    {
        return new MAM::Response()
          ->failure('318',
            'The Replace and Modify options are mutually exclusive');
    }

    if (   $accounting_mode eq 'usage-tracking'
        || $accounting_mode eq 'notional-charging'
        || $accounting_mode eq 'fast-allocation')
    {
        if ($replace)
        {
            return new MAM::Response()->failure('318',
                "The Replace option may not be used when the accounting mode is $accounting_mode"
            );
        }

        if ($modify)
        {
            return new MAM::Response()->failure('318',
                "The Modify option may not be used when the accounting mode is $accounting_mode"
            );
        }
    }

    # Check if data is present and parse out data properties
    my $data = $request->getDataElement();
    my $doc  = XML::LibXML::Document->new();
    $doc->setDocumentElement($data);
    my @rows = $data->childNodes();

    # We currently only support a solitary usage record
    if (@rows == 1)
    {
        return new MAM::Response()
          ->failure('314', 'The Data must be a Usage Record')
          unless $rows[0]->nodeName eq 'UsageRecord';
    }
    elsif (@rows > 1)
    {
        return new MAM::Response()
          ->failure('314', 'Only one UsageRecord may be specified in the Data');
    }

    # Build up a list of usage record properties from the data
    my %dataProperties       = ();
    my %chargeDataProperties = ();
    if (@rows > 0)
    {
        foreach my $col ($rows[0]->childNodes())
        {
            my $propertyName = $col->nodeName();
            my $increment    = $col->getAttribute('increment');
            if ($col->hasChildNodes())    # Simple properties only
            {
                my $propertyValue   = $col->textContent();
                my $consumptionRate = $col->getAttribute('consumptionRate');
                $propertyValue .= "*$consumptionRate"
                  if defined $consumptionRate;
                $dataProperties{$propertyName} = $propertyValue;
                if (defined $increment)
                {
                    $chargeDataProperties{$propertyName} = $increment;
                }
                else
                {
                    $chargeDataProperties{$propertyName} = $propertyValue;
                }
            }
            elsif (defined $increment)
            {
                $chargeDataProperties{$propertyName} = $increment;
            }
        }
    }

    # Id (usage record id)
    my $id;
    my $idOption = $request->getOptionValue('Id');
    if (! defined $idOption && ! keys %dataProperties)
    {
        return new MAM::Response()
          ->failure('314',
            'Either an Id option or usage record data is required');
    }
    my $idProperty = $dataProperties{Id};
    if (defined $idOption)
    {
        return new MAM::Response()->failure('318',
            "The specified Id option ($idOption) differs from the specified Id usage record property ($idProperty)"
        ) if (defined $idProperty && $idProperty ne $idOption);
        $id = $idOption;
    }
    else
    {
        $id = $idProperty;
    }

    # Look up existing usage record properties if id is known
    my %usageRecordProperties = ();
    my %chargeProperties      = ();
    if (defined $id)
    {
        my $subRequest = new MAM::Request(
            database   => $database,
            object     => 'UsageRecord',
            action     => 'Query',
            actor      => $SUPER_USER,
            conditions => [new MAM::Condition(name => 'Id', value => $id)],
        );
        # Invoke the proxy so it prepares the subRequest by setting defaults
        MAM::Proxy->new(
            database  => $database,
            request   => $subRequest,
            requestId => $requestId
        );
        # But use select so we can parse the data results easier
        # SELECT * FROM UsageRecord WHERE Id=$id
        my $subResults = $database->select(
            object     => 'UsageRecord',
            selections => [$subRequest->getSelections()],
            conditions => [new MAM::Condition(name => 'Id', value => $id)],
        );
        if (@{$subResults->{data}})
        {
            my $row   = ${$subResults->{data}}[0];
            my $cols  = $subResults->{cols};
            my $names = $subResults->{names};
            for (my $i = 0; $i < $cols; $i++)
            {
                my $name  = $names->[$i];
                my $value = $row->[$i];     # Null comes in as undef
                if (defined $value)
                {
                    $usageRecordProperties{$name} = $value;
                    # We don't want to autocharge for stored values like
                    # CPUTime which might be incremental in nature.
                    # Numeric values that we want to charge for will have to be
                    # passed in with the data, even if we are passing a UR id
                    $chargeProperties{$name} = $value if $value =~ /^\D/;
                }
            }
        }
        else
        {
            return new MAM::Response()
              ->failure('740', "No usage record found for specified id ($id)");
        }
    }

    # Merge data properties in with existing usage record and charge properties
    foreach my $name (keys %dataProperties)
    {
        $usageRecordProperties{$name} = $dataProperties{$name};
    }
    foreach my $name (keys %chargeDataProperties)
    {
        $chargeProperties{$name} = $chargeDataProperties{$name};
    }

    # Employ object instance overrides for specified usage record properties
    my @overrideCheckProperties = keys %usageRecordProperties;
    while (my $propertyName = shift @overrideCheckProperties)
    {
        # If the usage record property has an object associated with it
        if (MAM::Cache->objectExists($propertyName))
        {
            # Iterate over its attributes
            foreach my $attribute (
                MAM::Cache->listAttributesWithoutPropertyValue(
                    $propertyName, 'Values', undef
                )
              )
            {
                my $values =
                  MAM::Cache->getAttributeProperty($propertyName, $attribute,
                    'Values');
                # If an attribute is an override attribute
                if ($values =~ /^@([:?!]=)(\w+)/)
                {
                    my ($overrideType, $overrideName) = ($1, $2);
# See if the instance has a value set for this attribute
# Assume the primary key is Name
# SELECT $attribute FROM $propertyName WHERE Name=$usageRecordProperties{$propertyName}
                    my $subResults = $database->select(
                        object     => "$propertyName",
                        selections => [new MAM::Selection(name => $attribute)],
                        conditions => [
                            new MAM::Condition(
                                name  => 'Name',
                                value => $usageRecordProperties{$propertyName}
                            )
                        ]
                    );
                    # If the instance override has a value
                    if (@{$subResults->{data}}
                        && defined ${$subResults->{data}}[0]->[0])
                    {
                        my $overrideValue = ${$subResults->{data}}[0]->[0];
                        # Override it with the value if the usage record
                        # property is undefined or if it has a different
                        # value and this is an assign-always override
                        if ((! defined $usageRecordProperties{$overrideName})
                            || (   $overrideType eq ':='
                                && $usageRecordProperties{$overrideName} ne
                                $overrideValue)
                          )
                        {
                            $usageRecordProperties{$overrideName} =
                              $overrideValue;
                            $chargeProperties{$overrideName} = $overrideValue;
                            push @overrideCheckProperties, $overrideName;
                        }
                        # Or return with a failure if the usage record
                        # property is defined but has a different
                        # value and this is an assign-verify override
                        elsif ($overrideType eq '!='
                            && defined $usageRecordProperties{$overrideName}
                            && $usageRecordProperties{$overrideName} ne
                            $overrideValue)
                        {
                            return new MAM::Response()->failure('740',
                                "Specified $overrideName ($usageRecordProperties{$overrideName}) conflicts with ${propertyName}-designated $overrideName ($overrideValue)"
                            );
                        }
                    }
                }
            }
        }
    }

    # Build up a list of usage record attributes and set defaults
    my %usageRecordAttributes = ();
    foreach my $attributeName (
        MAM::Cache->listAttributesWithPropertyValue(
            'UsageRecord', 'Hidden', 'False'
        )
      )
    {
        $usageRecordAttributes{$attributeName} = 1;

        # Set default values for all undefined usage record properties that are
        # first class objects with a default value
        my $defaultValue =
          MAM::Cache->getObjectProperty($attributeName, 'DefaultValue');
        if (! defined $usageRecordProperties{$attributeName}
            && defined $defaultValue)
        {
            $usageRecordProperties{$attributeName} = $defaultValue;
            $chargeProperties{$attributeName}      = $defaultValue;
        }
    }

    # Build up list of usage record assignments
    # Verify that every specified usage record property exists if it is a value-
    # constrained object and that it is active if it has an active field
    my @usageRecordAssignments = ();
    foreach my $propertyName (sort keys %usageRecordProperties)
    {
        if ($usageRecordAttributes{$propertyName})
        {
            my $propertyValue = $usageRecordProperties{$propertyName};
            # Remove coefficients and consumption rates
            $propertyValue =~ s/\*(\d*\.?\d+)$//;
            push @usageRecordAssignments,
              new MAM::Assignment(
                name  => $propertyName,
                value => $propertyValue
              )
              unless ($propertyName eq 'Id'
                || $propertyName eq 'Charge');

            my $values =
              MAM::Cache->getAttributeProperty('UsageRecord', $propertyName,
                'Values');
            if (   defined $values
                && $values =~ /^@(?:[:?!]=)?(\w+)/
                && MAM::Cache->objectExists($1))
            {
                my $propertyObject = $1;
                my @primaryKeys =
                  MAM::Cache->listAttributesWithPropertyValue($propertyObject,
                    'PrimaryKey', 'True');

                # We can only do this if it has a single primary key
                if (@primaryKeys == 1)
                {
   # SELECT $filterKey[,Active] FROM $filterObject WHERE $filterKey=$filterValue
                    my $propertyKey   = $primaryKeys[0];
                    my $propertyValue = $usageRecordProperties{$propertyName};
                    my @validitySelections =
                      (new MAM::Selection(name => $propertyKey));
                    push @validitySelections,
                      new MAM::Selection(name => 'Active')
                      if MAM::Cache->attributeExists($propertyObject, 'Active');

                    my @validityConditions = (
                        new MAM::Condition(
                            name  => $propertyKey,
                            value => $propertyValue
                        )
                    );
                    my $subResults = $database->select(
                        object     => $propertyObject,
                        selections => \@validitySelections,
                        conditions => \@validityConditions,
                    );
                    unless (@{$subResults->{data}})
                    {
                        # Check to see if the instance should be auto-generated
                        if (MAM::Cache->getObjectProperty($propertyObject,
                                'AutoGen') eq 'True'
                          )
                        {
                            # Create the instance
                            my @autoGenAssignments = (
                                new MAM::Assignment(
                                    name  => $propertyKey,
                                    value => $propertyValue
                                )
                            );
                            if (MAM::Cache->attributeExists(
                                    $propertyObject, 'Description'
                                )
                              )
                            {
                                push @autoGenAssignments,
                                  new MAM::Assignment(
                                    name  => 'Description',
                                    value => 'Auto-Generated'
                                  );
                            }
                            my $subRequest = new MAM::Request(
                                database    => $database,
                                object      => $propertyObject,
                                action      => 'Create',
                                actor       => $SUPER_USER,
                                assignments => \@autoGenAssignments,
                            );
            # Invoke the proxy so it prepares the subRequest by setting defaults
                            MAM::Proxy->new(
                                database  => $database,
                                request   => $subRequest,
                                requestId => $requestId
                            );
                            my $subResponse =
                              MAM::Base->create($subRequest, $requestId);
                            if ($subResponse->getStatus() eq 'Failure')
                            {
                                my $message =
                                  "Unable to auto-gen $propertyObject $propertyValue: "
                                  . $subResponse->getMessage();
                                $log->error($message);
                                throw MAM::Exception($subResponse->getCode(),
                                    $message);
                            }
                        }
                        else
                        {

                            return new MAM::Response()->failure('740',
                                "$propertyObject ($propertyName) $propertyValue does not exist"
                            );
                        }
                    }
                    elsif (exists ${$subResults->{data}}[0]->[1]
                        && ${$subResults->{data}}[0]->[1] ne 'True')
                    {
                        return new MAM::Response()
                          ->failure('740',
                            "$propertyObject $propertyValue is not active");
                    }
                }
            }
        }
    }

    # Start Time, End Time and Duration
    my ($lienStartTime, $lienEndTime, $lienDuration);
    my $startTimeOption           = $request->getOptionValue('StartTime');
    my $startTimeProperty         = $usageRecordProperties{StartTime};
    my $endTimeOption             = $request->getOptionValue('EndTime');
    my $endTimeProperty           = $usageRecordProperties{EndTime};
    my $durationOption            = $request->getOptionValue('Duration');
    my $graceDurationOption       = $request->getOptionValue('GraceDuration');
    my $requestedDurationProperty = $usageRecordProperties{RequestedDuration};

    if (defined $startTimeOption)
    {
        $lienStartTime = $startTimeOption;
    }
    else    # Use default if duration was not obtained via option
    {
        $lienStartTime = $now;
    }

    if (defined $durationOption)
    {
        $lienDuration = $durationOption;
        if (defined $endTimeOption)
        {
            $lienEndTime = $endTimeOption;
        }
        elsif (defined $graceDurationOption)
        {
            if ($graceDurationOption !~ /^\d+$/)
            {
                return new MAM::Response()
                  ->failure('317',
                    'The grace duration must be an integer number of seconds');
            }
            $lienEndTime =
              $lienStartTime + $lienDuration + $graceDurationOption;
        }
        else
        {
            $lienEndTime = $lienStartTime + $lienDuration;
        }
    }
    elsif (defined $endTimeOption)
    {
        $lienEndTime  = $endTimeOption;
        $lienDuration = $lienEndTime - $lienStartTime;
    }
    else
    {
        # Override current start time with start time property if specified
        if ((! defined $startTimeOption) && defined $startTimeProperty)
        {
            $lienStartTime = $startTimeProperty;
        }

        if (defined $requestedDurationProperty)
        {
            $lienDuration = $requestedDurationProperty;
            if (defined $endTimeProperty)
            {
                $lienEndTime = $endTimeProperty;
            }
            else
            {
                $lienEndTime = $lienStartTime + $lienDuration;
            }
        }
        elsif (defined $endTimeProperty)
        {
            $lienEndTime  = $endTimeProperty;
            $lienDuration = $lienEndTime - $lienStartTime;
        }
        else
        {
            $lienEndTime  = $lienStartTime;
            $lienDuration = 0;
        }
    }
    if ($lienDuration !~ /^\d+$/)
    {
        return new MAM::Response()->failure('317',
            "The duration ($lienDuration) must be a non-negative integer number of seconds"
        );
    }

    if (   $accounting_mode ne 'usage-tracking'
        && $accounting_mode ne 'notional-charging')
    {
        # Refresh Allocations
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'Allocation',
            action   => 'Refresh'
        );
        my $subResponse = MAM::Bank->refresh($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Failed while refreshing allocations: '
                  . $subResponse->getMessage());
        }

# First, we build a list of funds along with their eligible allocations
# We need to do this because the trivial fund might have no constraints
# SELECT DISTINCT Allocation.Id,Allocation.Fund,Allocation.Amount,Allocation.CreditLimit,Allocation.StartTime,Allocation.EndTime,Fund.Priority FROM Allocation,Fund WHERE Allocation.Fund=Fund.Id AND Allocation.StartTime<=$lienStartTime AND Allocation.EndTime>$lienStartTime
        my @allocationConditions = ();
        if (($serverConfig->{'debit.activeallocationonly'} || 'false') =~
            /^true$/i)
        {
            push @allocationConditions,
              ( new MAM::Condition(
                    object => 'Allocation',
                    name   => 'StartTime',
                    value  => $now,
                    op     => 'LE',
                ),
                new MAM::Condition(
                    object => 'Allocation',
                    name   => 'EndTime',
                    value  => $now,
                    op     => 'GT',
                )
              );
        }
        elsif (
            ($serverConfig->{'debit.pointintimeonly'} || 'false') =~ /^true$/i)
        {
            push @allocationConditions,
              ( new MAM::Condition(
                    object => 'Allocation',
                    name   => 'StartTime',
                    value  => $lienStartTime,
                    op     => 'LE',
                ),
                new MAM::Condition(
                    object => 'Allocation',
                    name   => 'EndTime',
                    value  => $lienStartTime,
                    op     => 'GT',
                )
              );
        }
        else
        {
            push @allocationConditions,
              ( new MAM::Condition(
                    object => 'Allocation',
                    name   => 'StartTime',
                    value  => $lienEndTime,
                    op     => 'LE',
                ),
                new MAM::Condition(
                    object => 'Allocation',
                    name   => 'EndTime',
                    value  => $lienStartTime,
                    op     => 'GT',
                )
              );
        }
        $results = $database->select(
            objects => [
                new MAM::Object(name => 'Allocation'),
                new MAM::Object(name => 'Fund'),
            ],
            selections => [
                new MAM::Selection(object => 'Allocation', name => 'Id'),
                new MAM::Selection(
                    object => 'Allocation',
                    name   => 'Fund'
                ),
                new MAM::Selection(
                    object => 'Allocation',
                    name   => 'Amount'
                ),
                new MAM::Selection(
                    object => 'Allocation',
                    name   => 'CreditLimit'
                ),
                new MAM::Selection(
                    object => 'Allocation',
                    name   => 'StartTime'
                ),
                new MAM::Selection(
                    object => 'Allocation',
                    name   => 'EndTime'
                ),
                new MAM::Selection(
                    object => 'Fund',
                    name   => 'Priority'
                ),
            ],
            conditions => [
                new MAM::Condition(
                    object  => 'Allocation',
                    name    => 'Fund',
                    subject => 'Fund',
                    value   => 'Id',
                ),
                @allocationConditions,
            ],
            options => [new MAM::Option(name => 'Unique', value => 'True')]
        );
        foreach my $row (@{$results->{data}})
        {
            my $id          = $row->[0];
            my $fund        = $row->[1];
            my $amount      = $row->[2] || 0;
            my $creditLimit = $row->[3] || 0;
            my $startTime   = $row->[4];
            my $endTime     = $row->[5];
            my $priority    = $row->[6] || 0;

            push @{$fundAllocations{$fund}}, {
                id        => $id,
                fund      => $fund,
                balance   => $amount + $creditLimit,
                startTime => $startTime,
                endTime   => $endTime,
                # If you want it to take precedence over endTime, use millions
                weight => 100 * int(($infinity - $endTime) / 86400) +
                  10 * $priority,
            };
        }

       # Next, we prune off the funds that violate the constraints
       # In order to limit the list of constraints we must loop on, we do a join
       # on eligible allocations instead of constraining it to a huge fund list
        my %fundConstraints = ();
        my %limbo           = ();

        # Build up an association table for quick checks
        my %associations = ();
        foreach my $association (MAM::Cache->associationLookup(undef, undef))
        {
            my $parent = MAM::Cache->getObjectProperty($association, 'Parent');
            next unless defined $usageRecordProperties{$parent};
            my $child = MAM::Cache->getObjectProperty($association, 'Child');
            $associations{$parent}{$child} = $association;
        }

# SELECT DISTINCT Constraint.Id,Constraint.Fund,Constraint.Name,Constraint.Value FROM Constraint,Allocation WHERE Constraint.Fund=Allocation.Fund AND Allocation.StartTime<=$lienStartTime AND Allocation.EndTime>$lienStartTime
        $results = $database->select(
            objects => [
                new MAM::Object(name => 'Constraint'),
                new MAM::Object(name => 'Allocation'),
            ],
            selections => [
                new MAM::Selection(object => 'Constraint', name => 'Id'),
                new MAM::Selection(
                    object => 'Constraint',
                    name   => 'Fund'
                ),
                new MAM::Selection(object => 'Constraint', name => 'Name'),
                new MAM::Selection(
                    object => 'Constraint',
                    name   => 'Value'
                ),
            ],
            conditions => [
                new MAM::Condition(
                    object  => 'Constraint',
                    name    => 'Fund',
                    subject => 'Allocation',
                    value   => 'Fund',
                ),
                @allocationConditions,
            ],
            options => [new MAM::Option(name => 'Unique', value => 'True')]
        );
        foreach my $row (@{$results->{data}})
        {
            my $id     = $row->[0];
            my $fundId = $row->[1];
            my $name   = $row->[2];
            my $value  = $row->[3];

            # Short circuit if this constraint has no associated allocation
            next unless defined $fundAllocations{$fundId};
            push @{$fundConstraints{$fundId}{$name}}, $value;
        }

        my %accepted    = ();
        my %rejected    = ();
        my %memberships = ();
      FUNDID:
        foreach my $fundId (sort { $a <=> $b } keys %fundConstraints)
        {
            foreach my $constraintName (sort keys %{$fundConstraints{$fundId}})
            {
                foreach my $constraintVal (
                    @{$fundConstraints{$fundId}{$constraintName}})
                {
                    my $constraintValue = $constraintVal;    # Make a copy
                    my $accessAllowed = $constraintValue !~ s/^!//;

                   # If there is a usage record property matching the constraint
                    if (defined $usageRecordProperties{$constraintName}
                        && $usageRecordProperties{$constraintName} eq
                        $constraintValue)
                    {
                        if ($accessAllowed)
                        {
                            # Check if the usage record property is a
                            # parent in an association
                            if (defined $associations{$constraintName})
                            {
                                foreach my $childName (
                                    keys %{$associations{$constraintName}})
                                {
                                    my $association =
                                      $associations{$constraintName}
                                      {$childName};

                                    # If so, the child should be specified
                                    if (defined
                                        $usageRecordProperties{$childName})
                                    {
                             # If there is a matching member found, add the fund

                                        # Check for a match unless the
                                        # fund has a matching constraint
                                        unless (
                                            exists $fundConstraints{$fundId}
                                            {$childName})
                                        {

                                            # Use cache for membership results
                                            unless (
                                                exists
                                                $memberships{$association})
                                            {
                           # e.g. Account==chemistry && (Name==amy || Name==ANY)
                                                my @membershipConditions = (
                                                    new MAM::Condition(
                                                        name => $constraintName,
                                                        value =>
                                                          $usageRecordProperties{
                                                            $constraintName},
                                                    ),
                                                    new MAM::Condition(
                                                        name => 'Name'
                                                        , # Assume simple association
                                                        value =>
                                                          $usageRecordProperties{
                                                            $childName},
                                                        conj  => 'And',
                                                        group => '+1',
                                                    ),
                                                    new MAM::Condition(
                                                        name => 'Name'
                                                        , # Assume simple association
                                                        value => 'ANY',
                                                        conj  => 'Or',
                                                        group => '-1',
                                                    ),
                                                );
                                                push @membershipConditions,
                                                  new MAM::Condition(
                                                    name  => 'Active',
                                                    value => 'True'
                                                  )
                                                  if
                                                  MAM::Cache->attributeExists(
                                                    $association, 'Active');
                                                my $subResults =
                                                  $database->select(
                                                    object => $association,
                                                    conditions =>
                                                      \@membershipConditions,
                                                  );
                                                $memberships{$association} =
                                                  scalar @{$subResults->{data}};
                                            }
                                            unless ($memberships{$association})
                                            {
                                                return new MAM::Response()
                                                  ->failure(
                                                    '740',
                                                    "$childName $usageRecordProperties{$childName} is not a valid member of $constraintName $usageRecordProperties{$constraintName}"
                                                  );
                                            }
                                        }
                                    }
                                    else
                                    {
                                        return new MAM::Response()->failure(
                                            '314',
                                            "A $childName must be specified when specifying a $constraintName"
                                        );
                                    }
                                }
                            }
                            $accepted{$fundId}{$constraintName} = 1;
                            # Add to the allocation weight for each constraint
                            foreach
                              my $allocationRef (@{$fundAllocations{$fundId}})
                            {
                                $allocationRef->{weight} += 1;
                            }
                        }
                        else
                        {
                            if (defined $fundAllocations{$fundId})
                            {
                                # Move them to limbo
                                @{$limbo{$fundId}} =
                                  @{$fundAllocations{$fundId}};
                                delete $fundAllocations{$fundId};
                                next FUNDID;
                            }
                        }
                    }
         # There is not a usage record property matching the positive constraint
         # or the filter does not exist or the constraint is negative
                    else
                    {
                        if ($accessAllowed
                            || !
                            defined $usageRecordProperties{$constraintName})
                        {
                            if (defined $fundAllocations{$fundId})
                            {
                                # Provisionally Reject the constraint
                                $rejected{$fundId}{$constraintName} = 1;
                            }
                        }
                    }
                }
            }
        }

        # Move all funds to limbo that have rejected constraints and that
        # do not have accepted constraints
        foreach my $fundId (sort { $a <=> $b } keys %rejected)
        {
            foreach my $constraintName (sort keys %{$rejected{$fundId}})
            {
                unless (exists $accepted{$fundId}{$constraintName}
                    || ! defined $fundAllocations{$fundId})
                {
                    # Move them to limbo
                    @{$limbo{$fundId}} = @{$fundAllocations{$fundId}};
                    delete $fundAllocations{$fundId};
                }
            }
        }

        # Then, we build up fund linkages for ancestor funds
        my %ancestors = ();
        # SELECT Fund,Id FROM FundFund WHERE Overflow=True
        $results = $database->select(
            object     => 'FundFund',
            selections => [
                new MAM::Selection(name => 'Fund'),
                new MAM::Selection(name => 'Id'),
            ],
            conditions =>
              [new MAM::Condition(name => 'Overflow', value => 'True')],
        );
        foreach my $row (@{$results->{data}})
        {
            my $parent = $row->[0];
            my $child  = $row->[1];
            $ancestors{$child}{$parent} = 1;
        }

        # Add ancestors a generation at a time until there is no more growth
        my $distance = 0;
        while (1)
        {
            $distance++;
            my $count = scalar keys %fundAllocations;
            foreach my $fundId (sort { $a <=> $b } keys %fundAllocations)
            {
                foreach my $parent (keys %{$ancestors{$fundId}})
                {
                    if ($limbo{$parent})
                    {
                      # Move it back from limbo with negative distance as weight
                        foreach my $allocation (@{$limbo{$parent}})
                        {
                            $allocation->{weight} =
                              $distance * 100000 - $allocation->{endTime};
                        }
                        @{$fundAllocations{$parent}} = @{$limbo{$parent}};
                        delete $limbo{$parent};
                    }
                }
            }
            last if $count == scalar keys %fundAllocations;
        }

        if ($log->is_debug())
        {
            $log->debug('Feasible fund list is: ' . join ',',
                sort keys %fundAllocations);
        }

        # Begin building allocations list
        foreach my $fundId (sort { $a <=> $b } keys %fundAllocations)
        {
            foreach my $allocation (@{$fundAllocations{$fundId}})
            {
                $allocations{$allocation->{id}} = $allocation;
            }
        }
    }

    # Instance
    my $instance = $usageRecordProperties{Instance};

   # Look up the quote from the usage record so we can use it for quoted rates
   # Postpone for now since this could be quite expensive for large usage tables

    # Quote Id
    my $quoteId = $usageRecordProperties{Quote};
   # We may need to determine usage record id from quote or use the quoted rates
    if (defined $quoteId)
    {
# Check to see if the quote exists and is not expired or claimed by another instance
# SELECT Pinned,Instance,UsageRecord FROM Quote WHERE Id=$quoteId AND StartTime<=$lienStartTime AND EndTime>$lienStartTime
        my $subResults = $database->select(
            object     => 'Quote',
            selections => [
                new MAM::Selection(name => 'Pinned'),
                new MAM::Selection(name => 'Instance'),
                new MAM::Selection(name => 'UsageRecord')
            ],
            conditions => [
                new MAM::Condition(name => 'Id', value => $quoteId),
                new MAM::Condition(
                    name  => 'StartTime',
                    value => $lienStartTime,
                    op    => 'LE'
                ),
                new MAM::Condition(
                    name  => 'EndTime',
                    value => $lienStartTime,
                    op    => 'GT'
                )
            ]
        );
        if (@{$subResults->{data}})
        {
            my $row              = ${$subResults->{data}}[0];
            my $quotePinned      = $row->[0];
            my $quoteInstance    = $row->[1];
            my $quoteUsageRecord = $row->[2];

            # Reject all pinned quotes when either the usage instance is not
            # specified or it is specified but does not match the quote instance
            if (   defined $quotePinned
                && $quotePinned eq 'True'
                && ((! defined $instance)
                    || (defined $quoteInstance
                        && $quoteInstance ne $instance)
                )
              )
            {
                # Nullify the quote if absent, expired or used
                $quoteMessage .= "\nWarning -- Quote ($quoteId) is pinned";
                $quoteMessage .= " by instance ($quoteInstance)"
                  if defined $quoteInstance;
                $quoteMessage .= ' and cannot be used';
                $quoteMessage .= ' anonymously' if ! defined $quoteInstance;
                $warnings++;

                undef $quoteId;
                @usageRecordAssignments =
                  grep { $_->getName() ne 'Quote' } @usageRecordAssignments;
            }
# Accept all unpinned quotes or
# ( pinned quotes where a usage instance has been specified
# in the charge and ( where the quote instance is either
# uninitialized or matches the usage instance ))
# if ((! defined $quotePinned) || $quotePinned ne 'True' || (defined $instance &&
# ((! defined $quoteInstance) || $quoteInstance eq $instance)))
            else
            {
                # Set usage record id from quote
                $id = $quoteUsageRecord
                  if defined $quoteUsageRecord && ! defined $id;

                # Set the quote instance if pinned and uninitialized
                if (   defined $quotePinned
                    && $quotePinned eq 'True'
                    && defined $instance
                    && ! defined $quoteInstance)
                {
                    my $subRequest = new MAM::Request(
                        database    => $database,
                        object      => 'Quote',
                        action      => 'Modify',
                        actor       => $SUPER_USER,
                        assignments => [
                            new MAM::Assignment(
                                name  => 'Instance',
                                value => $instance
                            )
                        ],
                        conditions => [
                            new MAM::Condition(
                                name  => 'Id',
                                value => $quoteId
                            )
                        ]
                    );
                    my $subResponse =
                      MAM::Base->modify($subRequest, $requestId);
                    if ($subResponse->getStatus() eq 'Failure')
                    {
                        return new MAM::Response()->failure(
                            $subResponse->getCode(),
                            'Unable to modify quote: '
                              . $subResponse->getMessage()
                        );
                    }
                }
            }
        }
        else
        {
            # Nullify the quote if absent or expired
            $quoteMessage .=
              "\nWarning -- Quote ($quoteId) is invalid and cannot be used";
            $warnings++;
            $warnings++;

            undef $quoteId;
            @usageRecordAssignments =
              grep { $_->getName() ne 'Quote' } @usageRecordAssignments;
        }
    }

    # Lien Id
    my $lienId = $usageRecordProperties{Lien};

# Build a list of conflicting liens. Also get Id, Instance,
# and/or UsageRecord from lien if we have enough info.
# SELECT Lien.Id,Lien.UsageRecord,Lien.Instance,LienAllocation.Id,LienAllocation.Amount FROM Lien,LienAllocation WHERE Lien.Id=LienAllocation.Lien AND Lien.StartTime<=$now AND Lien.EndTime>$now AND [Id=$lienId|UsageRecord=$id|Instance=$instance]
    my %liens           = ();
    my %lienAllocations = ();
    if (   $accounting_mode ne 'usage-tracking'
        && $accounting_mode ne 'notional-charging'
        && $accounting_mode ne 'fast-allocation')
    {
        if (defined $lienId || defined $instance || defined $id)
        {
            my %instances = ();
            my %ids       = ();
            my @objects   = (
                new MAM::Object(name => 'Lien'),
                new MAM::Object(name => 'LienAllocation'),
            );

            my @selections = (
                new MAM::Selection(object => 'Lien', name => 'Id'),
                new MAM::Selection(
                    object => 'Lien',
                    name   => 'UsageRecord'
                ),
                new MAM::Selection(
                    object => 'Lien',
                    name   => 'Instance'
                ),
                new MAM::Selection(
                    object => 'LienAllocation',
                    name   => 'Id'
                ),
                new MAM::Selection(
                    object => 'LienAllocation',
                    name   => 'Amount'
                ),
            );
            my @conditions = (
                new MAM::Condition(
                    object  => 'Lien',
                    name    => 'Id',
                    subject => 'LienAllocation',
                    value   => 'Lien',
                ),
            );
            if (defined $lienId)
            {
                push @conditions,
                  new MAM::Condition(
                    object => 'Lien',
                    name   => 'Id',
                    value  => $lienId
                  );
            }
            elsif (defined $id)
            {
                push @conditions,
                  new MAM::Condition(
                    object => 'Lien',
                    name   => 'UsageRecord',
                    value  => $id
                  );
            }
            elsif (defined $instance)
            {
                push @conditions,
                  new MAM::Condition(
                    object => 'Lien',
                    name   => 'Instance',
                    value  => $instance
                  );
            }
            $results = $database->select(
                objects    => \@objects,
                selections => \@selections,
                conditions => \@conditions
            );
            foreach my $row (@{$results->{data}})
            {
                my $lienId = $row->[0];
                $liens{$lienId}++;
                $ids{$row->[1]}++       if defined $row->[1];
                $instances{$row->[2]}++ if defined $row->[2];
                my $allocationId = $row->[3];
                my $lienAmount = $row->[4] || 0;

                $lienAllocations{$lienId, $allocationId} = $lienAmount;
                if ($allocations{$allocationId})
                {
                    $allocations{$allocationId}{balance} += $lienAmount;
                }
            }
            $lienId = ${[keys %liens]}[0]
              if (! defined $lienId
                && scalar keys %liens == 1);
            $id = ${[keys %ids]}[0]
              if ( ! defined $id
                && scalar keys %ids == 1
                && ($replace || $modify));
            $instance = ${[keys %instances]}[0]
              if (! defined $instance && scalar keys %instances == 1);
        }

        if ($log->is_debug())
        {
            $log->debug(
                'Lien allocation list is: ' . join ',',
                map { '(' . (join ',', split $;, $_) . ')' }
                  keys %lienAllocations
            );
        }
    }

    # If Replace option is specified, remove conflicting Liens
    # If Modify option is specified, remove conflicting LienAllocations
    # DELETE FROM Lien WHERE [Id=<id>...]
    if (scalar keys %liens
        && ($replace || $modify))
    {
        my ($deleteObject, $deleteName);
        if ($replace)
        {
            $deleteObject = 'Lien';
            $deleteName   = 'Id';
        }
        elsif ($modify)
        {
            $deleteObject = 'LienAllocation';
            $deleteName   = 'Lien';
        }
        # Constrain to list of accepted allocation ids
        my $firstTime        = 1;
        my $counter          = 0;
        my @deleteConditions = ();
        foreach my $id (keys %liens)
        {
            $counter++;
            my $lastTime = ($counter == scalar keys %liens) ? 1 : 0;

            if ($firstTime && $lastTime)
            {
                push @deleteConditions,
                  new MAM::Condition(
                    name  => $deleteName,
                    value => $id,
                    conj  => 'And',
                    group => '0'
                  );
            }
            elsif ($firstTime)
            {
                $firstTime = 0;
                push @deleteConditions,
                  new MAM::Condition(
                    name  => $deleteName,
                    value => $id,
                    conj  => 'And',
                    group => '+1'
                  );
            }
            elsif ($lastTime)
            {
                push @deleteConditions,
                  new MAM::Condition(
                    name  => $deleteName,
                    value => $id,
                    conj  => 'Or',
                    group => '-1'
                  );
            }
            else
            {
                push @deleteConditions,
                  new MAM::Condition(
                    name  => $deleteName,
                    value => $id,
                    conj  => 'Or',
                    group => '0'
                  );
            }
        }
        my $subRequest = new MAM::Request(
            database   => $database,
            object     => $deleteObject,
            action     => 'Delete',
            actor      => $SUPER_USER,
            conditions => \@deleteConditions,
        );
        my $subResponse = MAM::Base->delete($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Unable to delete liens: ' . $subResponse->getMessage());
        }
        my $count = $subResponse->getCount();
        if ($count)
        {
            $reserveMessage .=
                "\n$count lien"
              . ($count == 1 ? ' was'      : 's were')
              . ($modify     ? ' modified' : ' removed');
        }
    }

    my $reserve         = $request->getOptionValue('Charge');
    my $itemizedCharges = $request->getOptionValue('ItemizedCharges')
      || '';

    # Charge must be a non-negative real number
    if (defined $reserve && $reserve !~ /^\d*\.?\d+(e\+\d+)?$/)
    {
        return new MAM::Response()
          ->failure('317',
            'The specified lien amount must be a non-negative real number');
    }

    # Calculate charge if not already specified
    # This is only needed when enforcing allocations
    if (   $accounting_mode ne 'usage-tracking'
        && $accounting_mode ne 'notional-charging')
    {
        MAM::Bank->calculateCharge(
            request               => $request,
            requestId             => $requestId,
            usageRecordProperties => \%chargeProperties,
            quote                 => $quoteId,
            duration              => $lienDuration,
            charge_ref            => \$reserve,
            itemizedCharges_ref   => \$itemizedCharges,
            data_ref              => \@data,
            metadata_ref          => \%metadata,
            message_ref           => \$reserveMessage,
        );

        if ($log->is_debug())
        {
            $log->debug("The amount to be reserved is: $reserve");
        }
    }
    else { $reserve = 0; }

    # Fail if there are no allocations to reserve against
    if (   $accounting_mode ne 'usage-tracking'
        && $accounting_mode ne 'notional-charging'
        && ! scalar keys %allocations)
    {
        my $message =
          'Insufficient funds: There are no valid allocations against which to make the lien';
        $message .= " (Instance $instance)" if defined $instance;
        return new MAM::Response()->failure('740', $message);
    }

    # Create new usage record if necessary
    if (! defined $id)
    {
        # Create a new usage record
        my @usageRecordOptions = ();
        push @usageRecordOptions,
          new MAM::Option(name => 'Instance', value => $instance)
          if defined $instance;
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'UsageRecord',
            action   => 'Create',
            actor    => $SUPER_USER,
            options  => \@usageRecordOptions,
        );
        # Invoke the proxy so it prepares the subRequest by setting defaults
        MAM::Proxy->new(
            database  => $database,
            request   => $subRequest,
            requestId => $requestId
        );
        my $subResponse = MAM::Base->create($subRequest, $requestId, 'NoLog');

        if ($subResponse->getStatus() eq 'Success')
        {
            # Figure out the id of the usage record we just created
            $id                        = $subResponse->getDatumValue('Id');
            $createdUsageRecord        = 1;
            $usageRecordProperties{Id} = $id;

            # Log the transaction
            MAM::Bank->logTransaction(
                database    => $database,
                requestId   => $requestId,
                txnId       => $database->nextId('Transaction'),
                object      => 'UsageRecord',
                action      => 'Create',
                actor       => $actor,
                assignments => [
                    $subRequest->getAssignments(),
                    new MAM::Assignment(
                        name  => 'Id',
                        value => $id
                    ),
                    new MAM::Assignment(
                        name  => 'UsageRecord',
                        value => $id
                    ),
                ],
                options => \@usageRecordOptions,
                count   => 1,
            );
        }
        else
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Unable to create usage record: ' . $subResponse->getMessage());
        }
    }

    if (   $accounting_mode ne 'usage-tracking'
        && $accounting_mode ne 'notional-charging'
        && $accounting_mode ne 'fast-allocation')
    {
        # Create the lien if it does not already exist
        if (! $modify || ! defined $lienId || ! defined $liens{$lienId})
        {
            my @lienAssignments = (
                new MAM::Assignment(
                    name  => 'StartTime',
                    value => $lienStartTime
                ),
                new MAM::Assignment(
                    name  => 'EndTime',
                    value => $lienEndTime
                ),
                new MAM::Assignment(
                    name  => 'Duration',
                    value => $lienDuration
                ),
                new MAM::Assignment(
                    name  => 'UsageRecord',
                    value => $id
                ),
            );
            push @lienAssignments,
              new MAM::Assignment(
                name  => 'Instance',
                value => $instance
              ) if defined $instance;
            my $description = $request->getOptionValue('Description');
            push @lienAssignments,
              new MAM::Assignment(
                name  => 'Description',
                value => $description
              ) if defined $description;

            my $subRequest = new MAM::Request(
                database    => $database,
                object      => 'Lien',
                action      => 'Create',
                actor       => $SUPER_USER,
                assignments => \@lienAssignments,
            );
            my $subResponse = MAM::Base->create($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Failure')
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    'Unable to create lien: ' . $subResponse->getMessage());
            }

            # Figure out the id of the lien we just created
            $lienId = $subResponse->getDatumValue('Id');
            $usageRecordProperties{Lien} = $lienId;
        }

        # Or modify its properties if the modify option is True
        elsif ($modify && defined $lienId && defined $liens{$lienId})
        {
            my @lienAssignments = (
                new MAM::Assignment(
                    name  => 'StartTime',
                    value => $lienStartTime
                ),
                new MAM::Assignment(
                    name  => 'EndTime',
                    value => $lienEndTime
                ),
                new MAM::Assignment(
                    name  => 'Duration',
                    value => $lienDuration
                ),
                new MAM::Assignment(
                    name  => 'UsageRecord',
                    value => $id
                ),
            );
            push @lienAssignments,
              new MAM::Assignment(
                name  => 'Instance',
                value => $instance
              ) if defined $instance;

            my $subRequest = new MAM::Request(
                database    => $database,
                object      => 'Lien',
                action      => 'Modify',
                actor       => $SUPER_USER,
                assignments => \@lienAssignments,
                conditions  => [
                    new MAM::Condition(
                        name  => 'Id',
                        value => $lienId
                    )
                ],
            );
            my $subResponse = MAM::Base->modify($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Failure')
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    'Unable to modify lien: ' . $subResponse->getMessage());
            }
        }

        # Place Liens against the Fund(s)
        if ($reserve > 0)
        {
            my $remaining = $reserve;

# Subtract all liens from allocation balances (including my own)
# SELECT GroupBy(LienAllocation.Id),Sum(LienAllocation.Amount) FROM Lien,LienAllocation WHERE Lien.Id=LienAllocation.Lien [AND LienAllocation.Id=<Id> ...]
            my @subSelections = (
                new MAM::Selection(
                    object => 'LienAllocation',
                    name   => 'Id',
                    op     => 'GroupBy',
                ),
                new MAM::Selection(
                    object => 'LienAllocation',
                    name   => 'Amount',
                    op     => 'Sum'
                ),
            );
            my @subConditions = (
                new MAM::Condition(
                    object  => 'Lien',
                    name    => 'Id',
                    subject => 'LienAllocation',
                    value   => 'Lien',
                ),
            );

            # Constrain to list of accepted allocation ids
            my $firstTime = 1;
            my $counter   = 0;
            foreach my $id (keys %allocations)
            {
                $counter++;
                my $lastTime = ($counter == scalar keys %allocations) ? 1 : 0;

                if ($firstTime && $lastTime)
                {
                    push @subConditions,
                      new MAM::Condition(
                        object => 'LienAllocation',
                        name   => 'Id',
                        value  => $id,
                        conj   => 'And',
                        group  => '0'
                      );
                }
                elsif ($firstTime)
                {
                    $firstTime = 0;
                    push @subConditions,
                      new MAM::Condition(
                        object => 'LienAllocation',
                        name   => 'Id',
                        value  => $id,
                        conj   => 'And',
                        group  => '+1'
                      );
                }
                elsif ($lastTime)
                {
                    push @subConditions,
                      new MAM::Condition(
                        object => 'LienAllocation',
                        name   => 'Id',
                        value  => $id,
                        conj   => 'Or',
                        group  => '-1'
                      );
                }
                else
                {
                    push @subConditions,
                      new MAM::Condition(
                        object => 'LienAllocation',
                        name   => 'Id',
                        value  => $id,
                        conj   => 'Or',
                        group  => '0'
                      );
                }
            }

            $results = $database->select(
                objects => [
                    new MAM::Object(name => 'Lien'),
                    new MAM::Object(name => 'LienAllocation'),
                ],
                selections => \@subSelections,
                conditions => \@subConditions
            );
            foreach my $row (@{$results->{data}})
            {
                my $allocationId = $row->[0];
                my $lienAmount   = $row->[1];

                if (exists $allocations{$allocationId})
                {
                    $allocations{$allocationId}{balance} -= $lienAmount;
                }
            }

            # Sort the allocations by weight
            my @allocations =
              sort { $allocations{$b}{weight} <=> $allocations{$a}{weight} }
              keys %allocations;

            if ($log->is_debug())
            {
                $log->debug('Sorted allocation list is: ' . join ',',
                    @allocations);
            }

       # Use simple iteration if using active or point-in-time allocation debits
            if ((   ($serverConfig->{'debit.activeallocationonly'} || 'false')
                    =~ /^true$/i
                )
                || (($serverConfig->{'debit.pointintimeonly'} || 'false') =~
                    /^true$/i)
              )
            {
                # Iterate through the allocation list
                foreach my $allocationId (@allocations)
                {
                    # Break out if we are done
                    last if $remaining <= 0;

# The amount I subtract from this allocation should be the minimum of:
# 1. $remaining (the amount left to withdraw)
# 2. $allocations{$allocationId}{balance} (the amount available within this allocation (considering liens and credit))
                    $balance = $allocations{$allocationId}{balance};
                    my $subLien = min($remaining, $balance);
                    next unless $subLien > 0;
                    $lienChunk{$allocationId} += $subLien;

                    $remaining -= $subLien;
                    if ($log->is_trace())
                    {
                        $log->trace(
                            "Reserving $subLien credits from allocation $allocationId"
                        );
                    }
                }
            }

            # Iterate over segment boundaries if using uniform allocation debits
            else
            {
                # Build up the lien segment edges
                my %edge = ();
                $edge{$lienStartTime}++;
                $edge{$lienEndTime}++;
                foreach my $allocationId (@allocations)
                {
                    my $allocationStartTime =
                      $allocations{$allocationId}{startTime};
                    my $allocationEndTime =
                      $allocations{$allocationId}{endTime};
                    $edge{$allocationStartTime}++
                      if $allocationStartTime > $lienStartTime;
                    $edge{$allocationEndTime}++
                      if $allocationEndTime < $lienEndTime;
                }

                # Iterate over lien segments, building up lien chunks
                my ($allocationFound, $leadingEdge);
                my @continuityMessages = ();
                # First pass going for uniform proportional distribution
                if ($remaining > 0)
                {
                    foreach my $trailingEdge (sort keys %edge)
                    {
                        # Handle case where lien start time equals lien end time
                        if (scalar keys %edge == 1)
                        {
                            $leadingEdge = $trailingEdge;
                        }

                        if (defined $leadingEdge)
                        {
                            my $proportion =
                              ($lienStartTime == $lienEndTime)
                              ? sprintf("%.${currency_precision}f",
                                ($reserve / $time_division))
                              : sprintf(
                                "%.${currency_precision}f",
                                (   ($reserve * ($trailingEdge - $leadingEdge))
                                    / ( $time_division *
                                          ($lienEndTime - $lienStartTime)
                                    )
                                )
                              );
                            my $subRemaining = min($proportion, $remaining);
                         # Final segment should incorporate loss due to rounding
                            if (($trailingEdge == $lienEndTime)
                                && ($remaining - $subRemaining <=
                                    10**(-$currency_precision))
                              )
                            {
                                $subRemaining = $remaining;
                            }
                            next unless $subRemaining > 0;
                            $allocationFound = 0;
                            foreach my $allocationId (@allocations)
                            {
                                if ((   $allocations{$allocationId}{startTime}
                                        <= $leadingEdge
                                    )
                                    && ($allocations{$allocationId}{endTime} >=
                                        $trailingEdge)
                                  )
                                {
                                    $balance =
                                      $allocations{$allocationId}{balance};
                                    my $subLien = min($subRemaining, $balance);
                                    next unless $subLien > 0;
                                    $lienChunk{$allocationId} += $subLien;
                                    $subRemaining -= $subLien;
                                    $remaining    -= $subLien;
                                    $allocations{$allocationId}{balance} -=
                                      $subLien;
                                    $allocationFound = 1;

                                    if ($log->is_trace())
                                    {
                                        $log->trace(
                                            "Reserving $subLien credits from allocation $allocationId for interval ($leadingEdge, $trailingEdge)"
                                        );
                                    }
                                    last if $subRemaining <= 0;
                                }
                            }
                            if (! $allocationFound)
                            {
                                my $continuityMessage = sprintf(
                                    "Allocation discontinuity of %.${currency_precision}f credits detected between %s and %s.",
                                    $subRemaining, toHRT($leadingEdge),
                                    toHRT($trailingEdge));
                                if ((   $serverConfig->{
                                            'debit.requirecontinuousallocations'
                                        }
                                        || 'false'
                                    ) =~ /^true$/i
                                  )
                                {
                                    return new MAM::Response()
                                      ->failure('740', $continuityMessage);
                                }
                                elsif (
                                    (   $serverConfig->{
                                            'debit.suppresscontinuitywarnings'}
                                        || 'false'
                                    ) !~ /^true$/i
                                  )
                                {
                                    push @continuityMessages,
                                      $continuityMessage;
                                }
                                $log->trace($continuityMessage)
                                  if $log->is_trace();
                            }
                            elsif ($subRemaining > 0)
                            {
                                my $continuityMessage = sprintf(
                                    "Allocation deficiency of %.${currency_precision}f credits detected between %s and %s.",
                                    $subRemaining, toHRT($leadingEdge),
                                    toHRT($trailingEdge));
                                if ((   $serverConfig->{
                                            'debit.requirecontinuousallocations'
                                        }
                                        || 'false'
                                    ) =~ /^true$/i
                                  )
                                {
                                    return new MAM::Response()
                                      ->failure('740', $continuityMessage);
                                }
                                elsif (
                                    (   $serverConfig->{
                                            'debit.suppresscontinuitywarnings'}
                                        || 'false'
                                    ) !~ /^true$/i
                                  )
                                {
                                    push @continuityMessages,
                                      $continuityMessage;
                                }
                                $log->trace($continuityMessage)
                                  if $log->is_trace();
                            }
                        }

                        last if $remaining <= 0;
                        $leadingEdge = $trailingEdge;
                    }
                }

            # If there were gaps in the first pass allocation, initiate a second
            # pass going from earliest to latest and allocating the maximum
            # possible in each allocation segment
                if ($remaining > 0)
                {
                    undef $leadingEdge;
                    foreach my $trailingEdge (sort keys %edge)
                    {
                        if (defined $leadingEdge)
                        {
                            foreach my $allocationId (@allocations)
                            {
                                if ((   $allocations{$allocationId}{startTime}
                                        <= $leadingEdge
                                    )
                                    && ($allocations{$allocationId}{endTime} >=
                                        $trailingEdge)
                                  )
                                {
                                    $balance =
                                      $allocations{$allocationId}{balance};
                                    my $subLien = min($remaining, $balance);
                                    next unless $subLien > 0;
                                    $lienChunk{$allocationId} += $subLien;
                                    $remaining -= $subLien;
                                    $allocations{$allocationId}{balance} -=
                                      $subLien;
                                    my $continuityMessage = sprintf(
                                        "Continuity deficit of %.${currency_precision}f credits satisfied from allocation $allocationId.",
                                        $subLien);
                                    if ((   $serverConfig->{
                                                'debit.suppresscontinuitywarnings'
                                            }
                                            || 'false'
                                        ) !~ /^true$/i
                                      )
                                    {
                                        push @continuityMessages,
                                          $continuityMessage;
                                    }
                                    $log->trace($continuityMessage)
                                      if $log->is_trace();
                                    if ($log->is_trace())
                                    {
                                        $log->trace(
                                            "Reserving $subLien credits from allocation $allocationId to satisfy continuity deficit"
                                        );
                                    }
                                    last if $remaining <= 0;
                                }
                            }
                        }

                        last if $remaining <= 0;
                        $leadingEdge = $trailingEdge;
                    }
                }

                if (@continuityMessages)
                {
                    $warningMessage .= "\nWarning -- " . join "\n",
                      @continuityMessages;
                    $warnings++;
                }
            }

# use Data::Dumper;
# print Data::Dumper->Dump([\%lienChunk], ['lienChunk']), "\n";
# print Data::Dumper->Dump([\@continuityMessages], ['continuityMessages']), "\n";

            # Iterate through the lien chunks
            foreach my $allocationId (sort keys %lienChunk)
            {
                my $fundId  = $allocations{$allocationId}{fund};
                my $subLien = $lienChunk{$allocationId};

                # Create the lienAllocation
                if (   $accounting_mode ne 'usage-tracking'
                    && $accounting_mode ne 'notional-charging'
                    && $accounting_mode ne 'fast-allocation')
                {
                    my $subRequest = new MAM::Request(
                        database    => $database,
                        object      => 'LienAllocation',
                        action      => 'Create',
                        actor       => $SUPER_USER,
                        assignments => [
                            new MAM::Assignment(
                                name  => 'Lien',
                                value => $lienId
                            ),
                            new MAM::Assignment(
                                name  => 'Id',
                                value => $allocationId
                            ),
                            new MAM::Assignment(
                                name  => 'Fund',
                                value => $fundId
                            ),
                            new MAM::Assignment(
                                name  => 'Amount',
                                value => $subLien
                            )
                        ]
                    );
                    my $subResponse =
                      MAM::Base->create($subRequest, $requestId);
                    if ($subResponse->getStatus() eq 'Failure')
                    {
                        return new MAM::Response()->failure(
                            $subResponse->getCode(),
                            'Unable to create lienAllocation: '
                              . $subResponse->getMessage()
                        );
                    }
                }
            }

            # Fail if not all amount could be debited
            if ($remaining > 0)
            {
                my $message = 'Insufficient balance to reserve usage';
                $message .= " (Instance $instance)" if defined $instance;

                # Append balance info to data if balance feedback requested
                if (   $balanceFeedback
                    && $accounting_mode ne 'usage-tracking'
                    && $accounting_mode ne 'notional-charging'
                    && scalar @data == 1)
                {
                    my $response =
                      new MAM::Response()->failure('782', $message);
                    my $datum = new MAM::Datum($data[0]->{_element});
                    $datum->delProperty('Amount');
                    $datum->setProperty('Balance', $balance);
                    $response->addDatum($datum);
                    delete $metadata{'Amount'};
                    $metadata{'Balance'} = 'Currency';
                    $response->setMetadata(\%metadata);
                    return $response;
                }
                else
                {
                    return new MAM::Response()->failure('782', $message);
                }
            }
        }
    }

    # Log the transaction
    my @reserveOptions = @options;
    if (   $accounting_mode ne 'usage-tracking'
        && $accounting_mode ne 'notional-charging'
        && $accounting_mode ne 'fast-allocation')
    {
        push @reserveOptions,
          new MAM::Option(
            name  => 'ItemizedCharges',
            value => $itemizedCharges
          );
    }
    if (defined $id)
    {
        push @reserveOptions,
          ( new MAM::Option(name => 'Id',          value => $id),
            new MAM::Option(name => 'UsageRecord', value => $id)
          );
    }
    MAM::Bank->logTransaction(
        database    => $database,
        requestId   => $requestId,
        txnId       => $database->nextId('Transaction'),
        object      => 'UsageRecord',
        action      => 'Reserve',
        actor       => $actor,
        assignments => \@usageRecordAssignments,
        options     => [
            @reserveOptions,
            new MAM::Option(name => 'Amount', value => $reserve)
        ],
        count => 1
    );

    # Modify the usage record with gleaned attributes
    if ($id)
    {
        # Modify the existing usage record
        my @options = ();
        push @options, new MAM::Option(name => 'Instance', value => $instance)
          if defined $instance;
        my $subRequest = new MAM::Request(
            database    => $database,
            object      => 'UsageRecord',
            action      => 'Modify',
            actor       => $SUPER_USER,
            conditions  => [new MAM::Condition(name => 'Id', value => $id)],
            options     => \@options,
            assignments => \@usageRecordAssignments,
        );
        my $subResponse = MAM::Base->modify($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Unable to modify usage record: ' . $subResponse->getMessage());
        }
    }

    $reserve = sprintf("%.${currency_precision}f", $reserve / $time_division);
    if (   $accounting_mode eq 'usage-tracking'
        || $accounting_mode eq 'notional-charging'
        || $accounting_mode eq 'fast-allocation')
    {
        $message = "Successfully recorded usage information";
    }
    else
    {
        $message = "Successfully reserved $reserve credits";
    }
    $message .= " with lien id $lienId"         if defined $lienId;
    $message .= " for instance $instance"       if defined $instance;
    $message .= " and created usage record $id" if $createdUsageRecord;

    # Prepend data with Lien and Usage Record Ids
    foreach my $datum (@data)
    {
        my $datumElement = $datum->{_element};
        my $firstChild   = $datumElement->firstChild;
        if ($createdUsageRecord)
        {
            my $idElement = new XML::LibXML::Element('UsageRecord');
            $idElement->appendText($id);
            $datumElement->insertBefore($idElement, $firstChild);
            $firstChild = $idElement;
            $metadata{'UsageRecord'} = 'Integer';
        }
        if (defined $lienId)
        {
            my $lienIdElement = new XML::LibXML::Element('Lien');
            $lienIdElement->appendText($lienId);
            $datumElement->insertBefore($lienIdElement, $firstChild);
            $metadata{'Lien'} = 'Integer';
        }
    }

    # Append balance info to data if balance feedback requested
    if (   $balanceFeedback
        && $accounting_mode ne 'usage-tracking'
        && $accounting_mode ne 'notional-charging'
        && scalar @data == 1)
    {
        my $datumElement   = $data[0]->{_element};
        my $lastChild      = $datumElement->lastChild;
        my $balanceElement = new XML::LibXML::Element('Balance');
        $balanceElement->appendText($balance);
        $datumElement->insertAfter($balanceElement, $lastChild);
        $metadata{'Balance'} = 'Currency';
    }

    # Return response
    if ($warnings)
    {
        $response =
          new MAM::Response()
          ->warning($reserve, \@data, \%metadata,
            $message . $warningMessage . $quoteMessage . $reserveMessage);
    }
    else
    {
        $response =
          new MAM::Response()
          ->success($reserve, \@data, \%metadata,
            $message . $quoteMessage . $reserveMessage);
    }
    return $response;
}

# ----------------------------------------------------------------------------
# $response = reset($request, $requestId);
# ----------------------------------------------------------------------------

# Reset (Fund)
sub reset
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object   = $request->getObject();
    my $actor    = $request->getActor();
    my $database = $request->getDatabase();
    my $id       = $request->getOptionValue('Id');
    my @options  = $request->getOptions();
    my $now      = $requestTime;

    # Fund Reset issues a resetting deposit using the default deposit
    # and the current time as the start time
    if ($object eq 'Fund')
    {
        my $count    = 0;
        my $warnings = 0;
        my @messages = ();
        my $response;

        my $filterType = $request->getOptionValue('FilterType')
          || 'NonExclusive';
        if (   $filterType ne 'ExactMatch'
            && $filterType ne 'Exclusive'
            && $filterType ne 'NonExclusive')
        {
            return new MAM::Response()->failure('317',
                'The FilterType option must be one of Exclusive, NonExclusive or ExactMatch'
            );
        }

        # Parse out filters
        my %filters = ();
        foreach my $option (@options)
        {
            if ($option->getName() eq 'Filter')
            {
                my $filter = $option->getValue();
                if ($filter =~ /^([^\s=!]+)=([^\s=]+)$/)
                {
                    $filters{$1} = $2;
                }
            }
        }

        # Refresh Allocations
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'Allocation',
            action   => 'Refresh'
        );
        my $subResponse = MAM::Bank->refresh($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Failed while refreshing allocations: '
                  . $subResponse->getMessage());
        }

        # First, we build a list of funds
        # We need to do this separately from the constraint query
        # because the trivial fund might have no constraints
        # SELECT Id FROM Fund [WHERE Id=$id]
        my %funds = ();
        my @fundSelections =
          (new MAM::Selection(name => 'Id'),
            new MAM::Selection(name => 'Name'));
        my @fundConditions = ();
        push @fundConditions,
          new MAM::Condition(
            name  => 'Id',
            value => $id
          ) if defined $id;
        my $results = $database->select(
            object     => 'Fund',
            selections => \@fundSelections,
            conditions => \@fundConditions,
        );
        foreach my $row (@{$results->{data}})
        {
            my $id   = $row->[0];
            my $name = $row->[1];
            $funds{$id} = defined $name ? $name : '';
        }

        # Next, we prune off the funds that violate the constraints
        my %fundConstraints = ();

        # SELECT Id,Fund,Name,Value FROM Constraint
        $results = $database->select(
            object     => 'Constraint',
            selections => [
                new MAM::Selection(name => 'Id'),
                new MAM::Selection(name => 'Fund'),
                new MAM::Selection(name => 'Name'),
                new MAM::Selection(name => 'Value'),
            ],
        );
        foreach my $row (@{$results->{data}})
        {
            my $id     = $row->[0];
            my $fundId = $row->[1];
            my $name   = $row->[2];
            my $value  = $row->[3];

            push @{$fundConstraints{$fundId}{$name}}, $value;
        }

        if ($filterType eq 'ExactMatch')
        {
          FUNDID:
            foreach my $fundId (sort { $a <=> $b } keys %funds)
            {
             # Reject if there are any filters without corresponding constraints
                foreach my $filterName (sort keys %filters)
                {
                    if (   (! defined $fundConstraints{$fundId})
                        || (! defined $fundConstraints{$fundId}{$filterName})
                        || (@{$fundConstraints{$fundId}{$filterName}} != 1)
                        || (${$fundConstraints{$fundId}{$filterName}}[0] ne
                            $filters{$filterName})
                      )
                    {
                        delete $funds{$fundId};
                        next FUNDID;
                    }
                }

             # Reject if there are any constraints without corresponding filters
                next unless exists $fundConstraints{$fundId};
                foreach
                  my $constraintName (sort keys %{$fundConstraints{$fundId}})
                {
                    if (   (! defined $filters{$constraintName})
                        || (@{$fundConstraints{$fundId}{$constraintName}} != 1)
                        || (${$fundConstraints{$fundId}{$constraintName}}[0] ne
                            $filters{$constraintName})
                      )
                    {
                        delete $funds{$fundId};
                        next FUNDID;
                    }
                }
            }
        }

        elsif ($filterType eq 'NonExclusive')
        {
          FUNDID:
            foreach my $fundId (sort { $a <=> $b } keys %fundConstraints)
            {
                my %constraintRejected = ();
                foreach my $filterName (sort keys %filters)
                {
                    my $filterValue = $filters{$filterName};
                    my $accepted    = 0;
                    my $rejected    = 0;

                    foreach my $constraintName (
                        sort keys %{$fundConstraints{$fundId}})
                    {
                        foreach my $constraintVal (
                            @{$fundConstraints{$fundId}{$constraintName}})
                        {
                            my $constraintValue = $constraintVal;  # Make a copy
                            my $accessAllowed = $constraintValue !~ s/^!//;
                            if (exists $constraintRejected{$constraintName}
                                {$constraintValue})
                            {
                                $rejected = 1;   # This constraint was rejected
                                next;            # Don't allow it to be accepted
                            }

                   # Same Name  |-1|-1|-1| 0| 0| 0|+1|+1|+1|
                   # Other Name |-1| 0|+1|-1| 0|+1|-1| 0|+1|
                   # Accpt/Rjct |R |R |R |  |  |R | A| A|RA|
                   # Result     |X |X |X |/ |/ |X |/ |/ |/ |
                   # A filter should be accepted unless rejected unless accepted

                            # We have a matching constraint name
                            if ($filterName eq $constraintName)
                            {
                                # Mark as accepted if it is a positive
                                # constraint with a matching value
                                if (   $accessAllowed
                                    && $filterValue eq $constraintValue)
                                {
                                    $accepted = 1;
                                }
                                # Mark as rejected if it is a non-matching value
                                elsif (

                                    $accessAllowed
                                    && $filterValue ne $constraintValue
                                  )
                                {
                                    $rejected = 1;
                                    $constraintRejected{$constraintName}
                                      {$constraintValue} = 1;
                                }
                               # Immediately delete if it is explicitly rejected
                                elsif (! $accessAllowed
                                    && $filterValue eq $constraintValue)
                                {
                                    delete $funds{$fundId};
                                    next FUNDID;
                                }
                            }

                 # If the constraint is in a parent association with this filter
                 # check it for membership
                 # We could do some caching here but we are not doing it yet
                            elsif (
                                $accessAllowed
                                && MAM::Cache->associationLookup(
                                    $constraintName, $filterName)
                                && (! exists $fundConstraints{$fundId}
                                    {$filterName} || ! grep ! /!/,
                                    @{$fundConstraints{$fundId}{$filterName}}
                                )
                              )
                            {
                                my ($association) =
                                  MAM::Cache->associationLookup($constraintName,
                                    $filterName);
                                # If there is not a matching member found
                                # reject the fund
                                my @membershipConditions = (
                                    new MAM::Condition(
                                        name =>
                                          'Name',    # Assume simple association
                                        value => $filterValue,
                                        conj  => 'And',
                                        group => '+1',
                                    ),
                                    new MAM::Condition(
                                        name =>
                                          'Name',    # Assume simple association
                                        value => 'ANY',
                                        conj  => 'Or',
                                        group => '-1',
                                    )
                                );
                                push @membershipConditions,
                                  new MAM::Condition(
                                    name  => 'Active',
                                    value => 'True'
                                  )
                                  if MAM::Cache->attributeExists($association,
                                    'Active');
                                my $subResults = $database->select(
                                    object     => $association,
                                    selections => [
                                        new MAM::Selection(
                                            name => $constraintName
                                        )
                                    ],
                                    conditions => \@membershipConditions,
                                );
                                my $keep = 0;
                                foreach my $row (@{$subResults->{data}})
                                {
                                    my $parent = $row->[0];
                                    if ($parent eq $constraintValue)
                                    {
                                        $keep = 1;
                                    }
                                }
                                if ($keep) { $accepted = 1; }
                                else
                                {
                                    $rejected = 1;
                                    $constraintRejected{$constraintName}
                                      {$constraintValue} = 1;
                                }
                            }

                  # If the constraint is in a child association with this filter
                  # check it for membership
                  # We could do some caching here but we are not doing it yet
                            elsif (
                                $accessAllowed
                                && MAM::Cache->associationLookup($filterName,
                                    $constraintName)
                                && !
                                exists $fundConstraints{$fundId}{$filterName}
                              )
                            {
                                my ($association) =
                                  MAM::Cache->associationLookup($filterName,
                                    $constraintName);
                                # If there is not a matching member found
                                # reject the fund
                                my @membershipConditions = (
                                    new MAM::Condition(
                                        name  => $filterName,
                                        value => $filterValue,
                                    )
                                );
                                push @membershipConditions,
                                  new MAM::Condition(
                                    name  => 'Active',
                                    value => 'True'
                                  )
                                  if MAM::Cache->attributeExists($association,
                                    'Active');
                                my $subResults = $database->select(
                                    object => $association,
                                    selections =>
                                      [new MAM::Selection(name => 'Name')]
                                    ,    # Simple name
                                    conditions => \@membershipConditions,
                                );
                                my $keep = 0;
                                foreach my $row (@{$subResults->{data}})
                                {
                                    my $child = $row->[0];
                                    if ($child eq $constraintValue)
                                    {
                                        $keep = 1;
                                    }
                                }
                                if ($keep) { $accepted = 1; }
                                else
                                {
                                    $rejected = 1;
                                    $constraintRejected{$constraintName}
                                      {$constraintValue} = 1;
                                }
                            }
                        }
                    }

                    if ($rejected && ! $accepted)
                    {
                        # Delete them
                        delete $funds{$fundId};
                        next FUNDID;
                    }
                }
            }
        }

        elsif ($filterType eq 'Exclusive')
        {
            foreach my $fundId (sort { $a <=> $b } keys %fundConstraints)
            {
                foreach
                  my $constraintName (sort keys %{$fundConstraints{$fundId}})
                {
                    my %accepted = ();
                    my %rejected = ();
                    foreach my $constraintValue (
                        @{$fundConstraints{$fundId}{$constraintName}})
                    {
                        my $accessAllowed = $constraintValue !~ s/^!//;

                        # If there is a filter matching the constraint
                        if (defined $filters{$constraintName}
                            && $filters{$constraintName} eq $constraintValue)
                        {
                            # It's a matching positive constraint
                            if ($accessAllowed)
                            {
                                my @associations =
                                  MAM::Cache->associationLookup($constraintName,
                                    undef);

                                # There are associations
                                if (@associations)
                                {
                                    foreach my $association (@associations)
                                    {
                                        my $child =
                                          MAM::Cache->getObjectProperty(
                                            $association, 'Child');

                                       # Check if the usage record property is a
                                       # parent in an association
                                        if (defined $filters{$child})
                                        {
                                            # Check for a match unless the
                                            # fund has a matching constraint
                                            unless (
                                                exists $fundConstraints{$fundId}
                                                {$child})
                                            {
                             # If there is a matching member found, add the fund
                             # We could improve upon this by caching the results
                                                my @membershipConditions = (
                                                    new MAM::Condition(
                                                        name => $constraintName,
                                                        value => $filters{
                                                            $constraintName},
                                                    ),
                                                    new MAM::Condition(
                                                        name => 'Name'
                                                        , # Assume simple association
                                                        value =>
                                                          $filters{$child},
                                                        conj  => 'And',
                                                        group => '+1',
                                                    ),
                                                    new MAM::Condition(
                                                        name => 'Name'
                                                        , # Assume simple association
                                                        value =>
                                                          $filters{$child},
                                                        conj  => 'And',
                                                        group => '+1',
                                                    ),
                                                    new MAM::Condition(
                                                        name => 'Name'
                                                        , # Assume simple association
                                                        value => 'ANY',
                                                        conj  => 'Or',
                                                        group => '-1',
                                                    ),
                                                );
                                                push @membershipConditions,
                                                  new MAM::Condition(
                                                    name  => 'Active',
                                                    value => 'True'
                                                  )
                                                  if
                                                  MAM::Cache->attributeExists(
                                                    $association, 'Active');
                                                my $subResults =
                                                  $database->select(
                                                    object => $association,
                                                    conditions =>
                                                      \@membershipConditions,
                                                  );
                                                if (@{$subResults->{data}})
                                                {
                                                    $accepted{$constraintName}
                                                      = 1;
                                                }
                                                else
                                                {
                                                    $rejected{$constraintName}
                                                      = 1;
                                                }
                                            }
                                            $accepted{$constraintName} = 1;
                                        }
                                        # Child associations must be specified
                                        # for an Exclusive filter
                                        else
                                        {
                                            return new MAM::Response()
                                              ->failure(
                                                '314',
                                                "A $child must be specified when specifying a $constraintName"
                                              );
                                        }
                                    }
                                }
                                # Accept it if there are no associations
                                else
                                {
                                    $accepted{$constraintName} = 1;
                                }
                            }
                            # Reject the matching negative constraint
                            else
                            {
                                $rejected{$constraintName} = 1;
                            }
                        }
                    # There is not a filter matching the positive constraint
                    # or the filter does not exist or the constraint is negative
                        else
                        {
                            if ($accessAllowed
                                || ! defined $filters{$constraintName})
                            {
                                $rejected{$constraintName} = 1;
                            }
                        }
                    }
                    # Constraints of the same name are Ored while
                    # Constraints of different names are Anded
                    # Delete if there is a constraint name
                    # that is rejected and not accepted
                    foreach my $name (sort keys %rejected)
                    {
                        if (! $accepted{$name} && $funds{$fundId})
                        {
                            # Delete them
                            delete $funds{$fundId};
                        }
                    }
                }
            }
        }

        if ($log->is_debug())
        {
            $log->debug('Reset fund list is: ' . join ',', sort keys %funds);
        }

        # Iterate through matched funds issuing a resetting deposit
        # using the default deposit amount
        foreach my $fund (sort keys %funds)
        {
            # Invoke Fund Deposit
            my $subRequest = new MAM::Request(
                database => $database,
                object   => 'Fund',
                action   => 'Deposit',
                actor    => $actor,
                options  => [
                    new MAM::Option(name => 'Id',    value => $fund),
                    new MAM::Option(name => 'Reset', value => 'True'),
                ],
            );
            my $subResponse = MAM::Bank->deposit($subRequest, $requestId);
            $warnings++ if $subResponse->getStatus() eq 'Failure';
            $count += $subResponse->getCount();
            push @messages, $subResponse->getMessage();
        }

        # Log the transaction
        MAM::Bank->logTransaction(
            database  => $database,
            requestId => $requestId,
            txnId     => $database->nextId('Transaction'),
            object    => 'Fund',
            action    => 'Reset',
            actor     => $actor,
            options   => \@options,
            count     => 1,
        );

        if ($warnings)
        {
            $response =
              new MAM::Response()->warning($count, join("\n", @messages));
        }
        else
        {
            $response =
              new MAM::Response()->success($count, join("\n", @messages));
        }
        return $response;
    }

    else
    {
        return new MAM::Response()
          ->failure('315', "The Bank class does not implement $object Reset");
    }
}

# ----------------------------------------------------------------------------
# $response = start($request, $requestId);
# ----------------------------------------------------------------------------

# Start (Payment)
sub start
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object   = $request->getObject();
    my $database = $request->getDatabase();
    my $actor    = $request->getActor();
    my @options  = $request->getOptions();
    my $now      = $requestTime;
    my $warnings = '';
    my $response;

    if ($object ne 'Payment')
    {
        return new MAM::Response()
          ->failure('315', "The Bank class does not implement $object Start");
    }

    # Id
    my $id = $request->getOptionValue('Id');
    if (! defined $id)
    {
        return new MAM::Response()
          ->failure('314', 'An Id option must be specified');
    }

# SELECT ChargePeriod,EndTime,ReservePattern,UpdateEvent FROM Payment WHERE Id=$id
    my $results = $database->select(
        object     => 'Payment',
        selections => [
            new MAM::Selection(name => 'ChargePeriod'),
            new MAM::Selection(name => 'EndTime'),
            new MAM::Selection(name => 'ReservePattern'),
            new MAM::Selection(name => 'UpdateEvent'),
        ],
        conditions => [
            new MAM::Condition(
                name  => 'Id',
                value => $id,
            ),
        ],
    );

    my ($chargePeriod, $endTime, $reservePattern, $updateEvent);
    if (@{$results->{data}})
    {
        my $row = ${$results->{data}}[0];
        $chargePeriod   = $row->[0];
        $endTime        = $row->[1];
        $reservePattern = $row->[2];
        $updateEvent    = $row->[3];
    }
    else
    {
        return new MAM::Response()
          ->failure('317', "Payment Start: Invalid payment id ($id)");
    }
    unless (defined $reservePattern)
    {
        return new MAM::Response()
          ->failure('740', 'Unable to determine payment reserve mode');
    }

    # UPDATE Payment SET StartTime=$now WHERE Id=$id
    my $subMessage = '';
    {
        my $subRequest = new MAM::Request(
            database    => $database,
            object      => 'Payment',
            action      => 'Modify',
            actor       => $SUPER_USER,
            assignments => [
                new MAM::Assignment(
                    name  => 'StartTime',
                    value => $now,
                ),
            ],
            conditions => [
                new MAM::Condition(
                    name  => 'Id',
                    value => $id
                )
            ],
        );
        my $subResponse = MAM::Base->modify($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            $subMessage = "\nWarning -- Unable to modify payment start time";
        }
    }

    # Determine starting fire time if a charge period is specified
    my $fireTime;
    my @eventAssignments = ();
    if (defined $chargePeriod)
    {
        # Set the rearm period
        push @eventAssignments,
          new MAM::Assignment(
            name  => 'RearmPeriod',
            value => $chargePeriod
          );

        # Adjust the base time and the delta as necessary
        my $baseTime  = $now;
        my $startTime = $now;
        my $delta     = 0;

        if ($log->is_trace())
        {
            $log->trace(
                "Calling toDMF($chargePeriod, \\\$baseTime[=$baseTime], \\\$delta"
            );
        }
        my $recur = toDMF($chargePeriod, \$baseTime, \$delta);
        if ($log->is_debug())
        {
            $log->debug("Calculated frequency description string ($recur)");
        }
        if ($recur =~ /^[\d\*:]+$/)
        {
            if ($log->is_trace())
            {
                $log->trace(
                    "Calling ParseRecur($recur, \"epoch $baseTime\", \"epoch "
                      . ($startTime + 1)
                      . "\", \"epoch "
                      . ($startTime + $delta + 1)
                      . "\")");
            }
            my @fireDates = ParseRecur(
                $recur,
                "epoch $baseTime",
                'epoch ' . ($startTime + 1),
                'epoch ' . ($startTime + $delta + 1)
            );
            if (@fireDates)
            {
                my $fireDate = shift @fireDates;
                if ($fireDate =~
                    /^(\d{4})(\d{2})(\d{2})(\d{2}):(\d{2}):(\d{2})$/)
                {
                    my ($year, $mon, $day, $hour, $min, $sec) =
                      ($1, $2, $3, $4, $5, $6);
                    $fireTime = timelocal($sec, $min, $hour, $day, $mon - 1,
                        $year - 1900);
                    if ($log->is_trace())
                    {
                        $log->trace(
                            "Calculated starting fire time ($fireTime)");
                    }

                    push @eventAssignments,
                      new MAM::Assignment(
                        name  => 'FireTime',
                        value => $fireTime
                      );
                }
                else
                {
                    return new MAM::Response()->failure('740',
                        "Unable to calculate update event fire time: The calculated fire date is not in the expected format ($fireDate)"
                    );
                }
            }
            else
            {
                return new MAM::Response()->failure('740',
                    'Unable to calculate update event fire time: No fire time was obtained'
                );
            }
        }
        else
        {
            return new MAM::Response()
              ->failure('740',
                "Unable to calculate update event fire time: $recur");
        }
    }
    else
    {
        push @eventAssignments,
          new MAM::Assignment(name => 'ArmTime', value => $now);
        $warnings .=
          "\nWarning -- No payment charge period specified. Manual payment firing will be required";
    }

    # Update the update event fire time
    if (defined $updateEvent)
    {
        # Go ahead and set it even if it is past the end time in
        # case the end time is extended. The end event will clean
        # up the update event.
        # UPDATE Event SET FireTime=$fireTime WHERE Id=$updateEvent
        my $subRequest = new MAM::Request(
            database    => $database,
            object      => 'Event',
            action      => 'Modify',
            actor       => $SUPER_USER,
            assignments => \@eventAssignments,
            conditions  => [
                new MAM::Condition(
                    name  => 'Id',
                    value => $updateEvent,
                )
            ],
        );
        my $subResponse = MAM::Base->modify($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Unable to arm payment update event: '
                  . $subResponse->getMessage());
        }
    }

    # Obtain a list of update components
    # SELECT Id FROM PaymentComponent WHERE Payment=$id AND Type='Update'
    $results = $database->select(
        object     => 'PaymentComponent',
        selections => [new MAM::Selection(name => 'Id'),],
        conditions => [
            new MAM::Condition(
                name  => 'Payment',
                value => $id,
            ),
            new MAM::Condition(
                name  => 'Type',
                value => 'Update',
            ),
        ],
    );

    # Make a lien
    my $lienCount = 0;
    if ($reservePattern eq 'Periodic' || $reservePattern eq 'Combined')
    {
        my $lienEndTime;
        if ($reservePattern eq 'Periodic')
        {
            if (defined $fireTime && defined $endTime)
            {
                $lienEndTime = min($fireTime, $endTime);
            }
            elsif (defined $fireTime) { $lienEndTime = $fireTime; }
            else
            {
                return new MAM::Response()->failure('314',
                    'The payment charge period must be specified when using the Periodic reservation model'
                );
            }
        }
        else    # ($reservePattern eq 'Combined')
        {
            if (defined $endTime)
            {
                $lienEndTime = $endTime;
            }
            else
            {
                return new MAM::Response()->failure('314',
                    'The payment end time must be specified when using the Combined reservation model'
                );
            }
        }
        my $duration = $lienEndTime - $now;

        # Iterate over update components
        foreach my $row (@{$results->{data}})
        {
            my $usageRecord = $row->[0];

            # UsageRecord Reserve
            my @reserveOptions = (
                new MAM::Option(name => 'Id',       value => $usageRecord),
                new MAM::Option(name => 'Duration', value => $duration),
            );
            push @reserveOptions,
              new MAM::Option(name => 'Modify', value => 'True')
              if $reservePattern eq 'Combined';
            my $subRequest = new MAM::Request(
                database => $database,
                object   => 'UsageRecord',
                action   => 'Reserve',
                actor    => $SUPER_USER,
                options  => \@reserveOptions,
            );
            my $subResponse = MAM::Bank->reserve($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Failure')
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    'Failed creating payment starting lien: '
                      . $subResponse->getMessage());
            }
            $lienCount++;
        }
    }

    # Log the transaction
    MAM::Bank->logTransaction(
        database  => $database,
        requestId => $requestId,
        txnId     => $database->nextId('Transaction'),
        object    => 'Payment',
        action    => 'Start',
        actor     => $actor,
        options   => \@options,
        count     => 1,
    );

    # Return response
    my $message = "Successfully started payment ($id)";
    $message .= " and created $lienCount lien" . ($lienCount == 1 ? '' : 's')
      if $lienCount;
    $message .= $subMessage if $subMessage;
    if ($warnings || $subMessage)
    {
        $response = new MAM::Response()->warning(1, $message . $warnings);
    }
    else
    {
        $response = new MAM::Response()->success(1, $message);
    }
    return $response;
}

# ----------------------------------------------------------------------------
# $response = stop($request, $requestId);
# ----------------------------------------------------------------------------

# Stop (Fund, Payment)
sub stop
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $actor    = $request->getActor();
    my $database = $request->getDatabase();
    my $object   = $request->getObject();
    my @options  = $request->getOptions();
    my $now      = $requestTime;

    # Fund Stop will end all active allocations
    if ($object eq 'Fund')
    {
        my $id         = $request->getOptionValue('Id');
        my $filterType = $request->getOptionValue('FilterType')
          || 'NonExclusive';
        if (   $filterType ne 'ExactMatch'
            && $filterType ne 'Exclusive'
            && $filterType ne 'NonExclusive')
        {
            return new MAM::Response()->failure('317',
                'The FilterType option must be one of Exclusive, NonExclusive or ExactMatch'
            );
        }

        # Parse out filters
        my %filters = ();
        foreach my $option (@options)
        {
            if ($option->getName() eq 'Filter')
            {
                my $filter = $option->getValue();
                if ($filter =~ /^([^\s=!]+)=([^\s=]+)$/)
                {
                    $filters{$1} = $2;
                }
            }
        }

        # Refresh Allocations
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'Allocation',
            action   => 'Refresh'
        );
        my $subResponse = MAM::Bank->refresh($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Failed while refreshing allocations: '
                  . $subResponse->getMessage());
        }

        # First, we build a list of funds
        # We need to do this separately from the constraint query
        # because the trivial fund might have no constraints
        # SELECT Id FROM Fund [WHERE Id=$id]
        my %funds = ();
        my @fundSelections =
          (new MAM::Selection(name => 'Id'),
            new MAM::Selection(name => 'Name'));
        my @fundConditions = ();
        push @fundConditions,
          new MAM::Condition(
            name  => 'Id',
            value => $id
          ) if defined $id;
        my $results = $database->select(
            object     => 'Fund',
            selections => \@fundSelections,
            conditions => \@fundConditions,
        );
        foreach my $row (@{$results->{data}})
        {
            my $id   = $row->[0];
            my $name = $row->[1];
            $funds{$id} = defined $name ? $name : '';
        }

        # Next, we prune off the funds that violate the constraints
        my %fundConstraints = ();

        # SELECT Id,Fund,Name,Value FROM Constraint
        $results = $database->select(
            object     => 'Constraint',
            selections => [
                new MAM::Selection(name => 'Id'),
                new MAM::Selection(name => 'Fund'),
                new MAM::Selection(name => 'Name'),
                new MAM::Selection(name => 'Value'),
            ],
        );
        foreach my $row (@{$results->{data}})
        {
            my $id     = $row->[0];
            my $fundId = $row->[1];
            my $name   = $row->[2];
            my $value  = $row->[3];

            push @{$fundConstraints{$fundId}{$name}}, $value;
        }

        if ($filterType eq 'ExactMatch')
        {
          FUNDID:
            foreach my $fundId (sort { $a <=> $b } keys %funds)
            {
             # Reject if there are any filters without corresponding constraints
                foreach my $filterName (sort keys %filters)
                {
                    if (   (! defined $fundConstraints{$fundId})
                        || (! defined $fundConstraints{$fundId}{$filterName})
                        || (@{$fundConstraints{$fundId}{$filterName}} != 1)
                        || (${$fundConstraints{$fundId}{$filterName}}[0] ne
                            $filters{$filterName})
                      )
                    {
                        delete $funds{$fundId};
                        next FUNDID;
                    }
                }

             # Reject if there are any constraints without corresponding filters
                next unless exists $fundConstraints{$fundId};
                foreach
                  my $constraintName (sort keys %{$fundConstraints{$fundId}})
                {
                    if (   (! defined $filters{$constraintName})
                        || (@{$fundConstraints{$fundId}{$constraintName}} != 1)
                        || (${$fundConstraints{$fundId}{$constraintName}}[0] ne
                            $filters{$constraintName})
                      )
                    {
                        delete $funds{$fundId};
                        next FUNDID;
                    }
                }
            }
        }

        elsif ($filterType eq 'NonExclusive')
        {
          FUNDID:
            foreach my $fundId (sort { $a <=> $b } keys %fundConstraints)
            {
                my %constraintRejected = ();
                foreach my $filterName (sort keys %filters)
                {
                    my $filterValue = $filters{$filterName};
                    my $accepted    = 0;
                    my $rejected    = 0;

                    foreach my $constraintName (
                        sort keys %{$fundConstraints{$fundId}})
                    {
                        foreach my $constraintVal (
                            @{$fundConstraints{$fundId}{$constraintName}})
                        {
                            my $constraintValue = $constraintVal;  # Make a copy
                            my $accessAllowed = $constraintValue !~ s/^!//;
                            if (exists $constraintRejected{$constraintName}
                                {$constraintValue})
                            {
                                $rejected = 1;   # This constraint was rejected
                                next;            # Don't allow it to be accepted
                            }

                   # Same Name  |-1|-1|-1| 0| 0| 0|+1|+1|+1|
                   # Other Name |-1| 0|+1|-1| 0|+1|-1| 0|+1|
                   # Accpt/Rjct |R |R |R |  |  |R | A| A|RA|
                   # Result     |X |X |X |/ |/ |X |/ |/ |/ |
                   # A filter should be accepted unless rejected unless accepted

                            # We have a matching constraint name
                            if ($filterName eq $constraintName)
                            {
                                # Mark as accepted if it is a positive
                                # constraint with a matching value
                                if (   $accessAllowed
                                    && $filterValue eq $constraintValue)
                                {
                                    $accepted = 1;
                                }
                                # Mark as rejected if it is a non-matching value
                                elsif (

                                    $accessAllowed
                                    && $filterValue ne $constraintValue
                                  )
                                {
                                    $rejected = 1;
                                    $constraintRejected{$constraintName}
                                      {$constraintValue} = 1;
                                }
                               # Immediately delete if it is explicitly rejected
                                elsif (! $accessAllowed
                                    && $filterValue eq $constraintValue)
                                {
                                    delete $funds{$fundId};
                                    next FUNDID;
                                }
                            }

                 # If the constraint is in a parent association with this filter
                 # check it for membership
                 # We could do some caching here but we are not doing it yet
                            elsif (
                                $accessAllowed
                                && MAM::Cache->associationLookup(
                                    $constraintName, $filterName)
                                && (! exists $fundConstraints{$fundId}
                                    {$filterName} || ! grep ! /!/,
                                    @{$fundConstraints{$fundId}{$filterName}}
                                )
                              )
                            {
                                my ($association) =
                                  MAM::Cache->associationLookup($constraintName,
                                    $filterName);
                                # If there is not a matching member found
                                # reject the fund
                                my @membershipConditions = (
                                    new MAM::Condition(
                                        name =>
                                          'Name',    # Assume simple association
                                        value => $filterValue,
                                        conj  => 'And',
                                        group => '+1',
                                    ),
                                    new MAM::Condition(
                                        name =>
                                          'Name',    # Assume simple association
                                        value => 'ANY',
                                        conj  => 'Or',
                                        group => '-1',
                                    )
                                );
                                push @membershipConditions,
                                  new MAM::Condition(
                                    name  => 'Active',
                                    value => 'True'
                                  )
                                  if MAM::Cache->attributeExists($association,
                                    'Active');
                                my $subResults = $database->select(
                                    object     => $association,
                                    selections => [
                                        new MAM::Selection(
                                            name => $constraintName
                                        )
                                    ],
                                    conditions => \@membershipConditions,
                                );
                                my $keep = 0;
                                foreach my $row (@{$subResults->{data}})
                                {
                                    my $parent = $row->[0];
                                    if ($parent eq $constraintValue)
                                    {
                                        $keep = 1;
                                    }
                                }
                                if ($keep) { $accepted = 1; }
                                else
                                {
                                    $rejected = 1;
                                    $constraintRejected{$constraintName}
                                      {$constraintValue} = 1;
                                }
                            }

                  # If the constraint is in a child association with this filter
                  # check it for membership
                  # We could do some caching here but we are not doing it yet
                            elsif (
                                $accessAllowed
                                && MAM::Cache->associationLookup($filterName,
                                    $constraintName)
                                && !
                                exists $fundConstraints{$fundId}{$filterName}
                              )
                            {
                                my ($association) =
                                  MAM::Cache->associationLookup($filterName,
                                    $constraintName);
                                # If there is not a matching member found
                                # reject the fund
                                my @membershipConditions = (
                                    new MAM::Condition(
                                        name  => $filterName,
                                        value => $filterValue,
                                    )
                                );
                                push @membershipConditions,
                                  new MAM::Condition(
                                    name  => 'Active',
                                    value => 'True'
                                  )
                                  if MAM::Cache->attributeExists($association,
                                    'Active');
                                my $subResults = $database->select(
                                    object => $association,
                                    selections =>
                                      [new MAM::Selection(name => 'Name')]
                                    ,    # Simple name
                                    conditions => \@membershipConditions,
                                );
                                my $keep = 0;
                                foreach my $row (@{$subResults->{data}})
                                {
                                    my $child = $row->[0];
                                    if ($child eq $constraintValue)
                                    {
                                        $keep = 1;
                                    }
                                }
                                if ($keep) { $accepted = 1; }
                                else
                                {
                                    $rejected = 1;
                                    $constraintRejected{$constraintName}
                                      {$constraintValue} = 1;
                                }
                            }
                        }
                    }

                    if ($rejected && ! $accepted)
                    {
                        # Delete them
                        delete $funds{$fundId};
                        next FUNDID;
                    }
                }
            }
        }

        elsif ($filterType eq 'Exclusive')
        {
            foreach my $fundId (sort { $a <=> $b } keys %fundConstraints)
            {
                foreach
                  my $constraintName (sort keys %{$fundConstraints{$fundId}})
                {
                    my %accepted = ();
                    my %rejected = ();
                    foreach my $constraintValue (
                        @{$fundConstraints{$fundId}{$constraintName}})
                    {
                        my $accessAllowed = $constraintValue !~ s/^!//;

                        # If there is a filter matching the constraint
                        if (defined $filters{$constraintName}
                            && $filters{$constraintName} eq $constraintValue)
                        {
                            # It's a matching positive constraint
                            if ($accessAllowed)
                            {
                                my @associations =
                                  MAM::Cache->associationLookup($constraintName,
                                    undef);

                                # There are associations
                                if (@associations)
                                {
                                    foreach my $association (@associations)
                                    {
                                        my $child =
                                          MAM::Cache->getObjectProperty(
                                            $association, 'Child');

                                       # Check if the usage record property is a
                                       # parent in an association
                                        if (defined $filters{$child})
                                        {
                                            # Check for a match unless the
                                            # fund has a matching constraint
                                            unless (
                                                exists $fundConstraints{$fundId}
                                                {$child})
                                            {
                             # If there is a matching member found, add the fund
                             # We could improve upon this by caching the results
                                                my @membershipConditions = (
                                                    new MAM::Condition(
                                                        name => $constraintName,
                                                        value => $filters{
                                                            $constraintName},
                                                    ),
                                                    new MAM::Condition(
                                                        name => 'Name'
                                                        , # Assume simple association
                                                        value =>
                                                          $filters{$child},
                                                        conj  => 'And',
                                                        group => '+1',
                                                    ),
                                                    new MAM::Condition(
                                                        name => 'Name'
                                                        , # Assume simple association
                                                        value =>
                                                          $filters{$child},
                                                        conj  => 'And',
                                                        group => '+1',
                                                    ),
                                                    new MAM::Condition(
                                                        name => 'Name'
                                                        , # Assume simple association
                                                        value => 'ANY',
                                                        conj  => 'Or',
                                                        group => '-1',
                                                    ),
                                                );
                                                push @membershipConditions,
                                                  new MAM::Condition(
                                                    name  => 'Active',
                                                    value => 'True'
                                                  )
                                                  if
                                                  MAM::Cache->attributeExists(
                                                    $association, 'Active');
                                                my $subResults =
                                                  $database->select(
                                                    object => $association,
                                                    conditions =>
                                                      \@membershipConditions,
                                                  );
                                                if (@{$subResults->{data}})
                                                {
                                                    $accepted{$constraintName}
                                                      = 1;
                                                }
                                                else
                                                {
                                                    $rejected{$constraintName}
                                                      = 1;
                                                }
                                            }
                                            $accepted{$constraintName} = 1;
                                        }
                                        # Child associations must be specified
                                        # for an Exclusive filter
                                        else
                                        {
                                            return new MAM::Response()
                                              ->failure(
                                                '314',
                                                "A $child must be specified when specifying a $constraintName"
                                              );
                                        }
                                    }
                                }
                                # Accept it if there are no associations
                                else
                                {
                                    $accepted{$constraintName} = 1;
                                }
                            }
                            # Reject the matching negative constraint
                            else
                            {
                                $rejected{$constraintName} = 1;
                            }
                        }
                    # There is not a filter matching the positive constraint
                    # or the filter does not exist or the constraint is negative
                        else
                        {
                            if ($accessAllowed
                                || ! defined $filters{$constraintName})
                            {
                                $rejected{$constraintName} = 1;
                            }
                        }
                    }

                    # Constraints of the same name are Ored while
                    # Constraints of different names are Anded
                    # Delete if there is a constraint name
                    # that is rejected and not accepted
                    foreach my $name (sort keys %rejected)
                    {
                        if (! $accepted{$name} && $funds{$fundId})
                        {
                            # Delete them
                            delete $funds{$fundId};
                        }
                    }
                }
            }
        }

        if ($log->is_debug())
        {
            $log->debug('Feasible fund list is: ' . join ',', sort keys %funds);
        }

        # Look for active allocations within the matched funds
        # SELECT Id FROM Allocation WHERE Active=True AND Fund IN @funds

        my @conditions =
          (new MAM::Condition(name => 'Active', value => 'True'));

        # Constrain to list of accepted allocation ids
        my $firstTime = 1;
        my $counter   = 0;
        foreach my $fund (keys %funds)
        {
            $counter++;
            my $lastTime = ($counter == scalar keys %funds) ? 1 : 0;

            if ($firstTime && $lastTime)
            {
                push @conditions,
                  new MAM::Condition(
                    name  => 'Fund',
                    value => $fund,
                    conj  => 'And',
                    group => '0'
                  );
            }
            elsif ($firstTime)
            {
                $firstTime = 0;
                push @conditions,
                  new MAM::Condition(
                    name  => 'Fund',
                    value => $fund,
                    conj  => 'And',
                    group => '+1'
                  );
            }
            elsif ($lastTime)
            {
                push @conditions,
                  new MAM::Condition(
                    name  => 'Fund',
                    value => $fund,
                    conj  => 'Or',
                    group => '-1'
                  );
            }
            else
            {
                push @conditions,
                  new MAM::Condition(
                    name  => 'Fund',
                    value => $fund,
                    conj  => 'Or',
                    group => '0'
                  );
            }
        }

        $results = $database->select(
            object     => 'Allocation',
            selections => [new MAM::Selection(name => 'Id')],
            conditions => \@conditions
        );
        my $count = 0;
        foreach my $row (@{$results->{data}})
        {
            my $allocation = $row->[0];

            # Set the end time to now for each matched allocation
            my $subRequest = new MAM::Request(
                database => $database,
                object   => 'Allocation',
                action   => 'Modify',
                actor    => $SUPER_USER,
                assignments =>
                  [new MAM::Assignment(name => 'EndTime', value => $now)],
                conditions =>
                  [new MAM::Condition(name => 'Id', value => $allocation)],
            );
            my $subResponse = MAM::Bank->modify($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Failure')
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    'Failed modifying allocation: '
                      . $subResponse->getMessage());
            }
            $count += $subResponse->getCount();
        }

        return new MAM::Response()->success($count,
            "Successfully stopped $count allocation"
              . ($count == 1 ? '' : 's'));
    }

    # Stop payment plan
    elsif ($object eq 'Payment')
    {
        my $chargeMessage = '';
        my $warnings      = '';
        my $response;

        # Id
        my $id = $request->getOptionValue('Id');
        if (! defined $id)
        {
            return new MAM::Response()
              ->failure('314', 'An Id option must be specified');
        }

# SELECT ChargePattern,ReservePattern,StartTime,UpdateEvent FROM Payment WHERE Id=$id
        my $results = $database->select(
            object     => 'Payment',
            selections => [
                new MAM::Selection(name => 'ChargePattern'),
                new MAM::Selection(name => 'ReservePattern'),
                new MAM::Selection(name => 'StartTime'),
                new MAM::Selection(name => 'UpdateEvent'),
            ],
            conditions => [
                new MAM::Condition(
                    name  => 'Id',
                    value => $id,
                ),
            ],
        );

        my ($chargePattern, $reservePattern, $startTime, $updateEvent);
        if (@{$results->{data}})
        {
            my $row = ${$results->{data}}[0];
            $chargePattern  = $row->[0];
            $reservePattern = $row->[1];
            $startTime      = $row->[2];
            $updateEvent    = $row->[3];
        }
        else
        {
            return new MAM::Response()
              ->failure('317', "Payment Stop: Invalid payment id ($id)");
        }
        unless (defined $chargePattern)
        {
            return new MAM::Response()
              ->failure('740', 'Unable to determine payment charge mode');
        }
        unless (defined $reservePattern)
        {
            return new MAM::Response()
              ->failure('740', 'Unable to determine payment reserve mode');
        }

        # UPDATE Payment SET StopTime=$now WHERE Id=$id
        my $subMessage = '';
        {
            my $subRequest = new MAM::Request(
                database    => $database,
                object      => 'Payment',
                action      => 'Modify',
                actor       => $SUPER_USER,
                assignments => [
                    new MAM::Assignment(
                        name  => 'StopTime',
                        value => $now,
                    ),
                ],
                conditions => [
                    new MAM::Condition(
                        name  => 'Id',
                        value => $id
                    )
                ],
            );
            my $subResponse = MAM::Base->modify($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Failure')
            {
                $subMessage = "\nWarning -- Unable to modify payment stop time";
            }
        }

        # Obtain event details
        my ($armTime, $fireTime, $endTime);
        if (defined $updateEvent)
        {
            # SELECT ArmTime,FireTime,EndTime FROM Event WHERE Id=$id
            my $results = $database->select(
                object     => 'Event',
                selections => [
                    new MAM::Selection(name => 'ArmTime'),
                    new MAM::Selection(name => 'FireTime'),
                    new MAM::Selection(name => 'EndTime'),
                ],
                conditions => [
                    new MAM::Condition(
                        name  => 'Id',
                        value => $updateEvent,
                    ),
                ],
            );
            if (@{$results->{data}})
            {
                my $row = ${$results->{data}}[0];
                $armTime  = $row->[0];
                $fireTime = $row->[1];
                $endTime  = $row->[2];
            }

            # Clear the update event fire time
            # UPDATE Event SET FireTime=NULL WHERE Id=$updateEvent
            my $subRequest = new MAM::Request(
                database    => $database,
                object      => 'Event',
                action      => 'Modify',
                actor       => $SUPER_USER,
                assignments => [
                    new MAM::Assignment(
                        name  => 'FireTime',
                        value => 'NULL',
                    )
                ],
                conditions => [
                    new MAM::Condition(
                        name  => 'Id',
                        value => $updateEvent,
                    )
                ],
            );
            my $subResponse = MAM::Base->modify($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Failure')
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    'Unable to disarm payment update event: '
                      . $subResponse->getMessage());
            }
        }

        # Obtain a list of update components
        # SELECT Id FROM PaymentComponent WHERE Payment=$id AND Type='Update'
        $results = $database->select(
            object     => 'PaymentComponent',
            selections => [new MAM::Selection(name => 'Id'),],
            conditions => [
                new MAM::Condition(
                    name  => 'Payment',
                    value => $id,
                ),
                new MAM::Condition(
                    name  => 'Type',
                    value => 'Update',
                ),
            ],
        );

        # Handle the update components
        my $chargeAmount = 0;
        if (@{$results->{data}})
        {
          # Make a final charge (prorated in the case of a Periodic charge mode)
            if (   $chargePattern eq 'Combined'
                || $chargePattern eq 'Periodic')
            {
                unless (defined $armTime)
                {
                    return new MAM::Response()->failure('740',
                        'Unable to calculate payment stop charge without an update event arm time'
                    );
                }

                my $chargeStartTime = $now;
                if ($chargePattern eq 'Periodic' && defined $armTime)
                {
                    $chargeStartTime = $armTime;
                }
                elsif ($chargePattern eq 'Combined' && defined $startTime)
                {
                    $chargeStartTime = $startTime;
                }

                my $chargeEndTime = $now;
                my $earliestEventTime;
                if (defined $fireTime && defined $endTime)
                {
                    $earliestEventTime = min($fireTime, $endTime);
                }
                elsif (defined $fireTime)
                {
                    $earliestEventTime = $fireTime;
                }
                elsif (defined $endTime) { $earliestEventTime = $endTime; }
                if (defined $earliestEventTime)
                {
                    if (   ($now - $earliestEventTime) >= 0
                        && ($now - $earliestEventTime) < 30)
                    {
                        # Event occurred less than 30 seconds ago
                        # Assume an event triggered this stop
                        # So we will use the event fire time
                        $chargeEndTime = $earliestEventTime;
                    }
                }

                my $duration = $chargeEndTime - $chargeStartTime;

                # Iterate over update components
                foreach my $row (@{$results->{data}})
                {
                    my $usageRecord = $row->[0];

                    my @chargeOptions = (
                        new MAM::Option(
                            name  => 'Id',
                            value => $usageRecord
                        ),
                        new MAM::Option(
                            name  => 'Duration',
                            value => $duration
                        ),
                    );
                    push @chargeOptions,
                      new MAM::Option(
                        name  => 'Incremental',
                        value => 'True'
                      ) if $reservePattern eq 'Combined';

                    # UsageRecord Charge
                    my $subRequest = new MAM::Request(
                        database => $database,
                        object   => 'UsageRecord',
                        action   => 'Charge',
                        actor    => $SUPER_USER,
                        options  => \@chargeOptions,
                    );
                    my $subResponse =
                      MAM::Bank->charge($subRequest, $requestId);
                    if ($subResponse->getStatus() eq 'Failure')
                    {
                        return new MAM::Response()->failure(
                            $subResponse->getCode(),
                            'Failed issuing payment stopping update charge: '
                              . $subResponse->getMessage()
                        );
                    }
                    $chargeAmount += $subResponse->getCount();
                    $chargeMessage .= "\n" . $subResponse->getMessage();
                }
            }

            # Delete the lien
            if ($reservePattern eq 'Periodic' && $chargePattern eq 'None')
            {
                # Iterate over update components
                foreach my $row (@{$results->{data}})
                {
                    my $usageRecord = $row->[0];

                    # Lien Delete
                    my $subRequest = new MAM::Request(
                        database   => $database,
                        object     => 'Lien',
                        action     => 'Delete',
                        actor      => $SUPER_USER,
                        conditions => [
                            new MAM::Condition(
                                name  => 'UsageRecord',
                                value => $usageRecord
                            )
                        ],
                    );
                    my $subResponse =
                      MAM::Base->delete($subRequest, $requestId);
                    if ($subResponse->getStatus() eq 'Failure')
                    {
                        $warnings .= "\n"
                          . 'Unable to remove update component lien: '
                          . $subResponse->getMessage();
                    }
                    $chargeMessage .= $subResponse->getMessage();
                }
            }
        }

        # Log the transaction
        MAM::Bank->logTransaction(
            database  => $database,
            requestId => $requestId,
            txnId     => $database->nextId('Transaction'),
            object    => 'Payment',
            action    => 'Stop',
            actor     => $actor,
            options   => \@options,
            count     => 1,
        );

        # Return response
        my $message = "Successfully stopped payment ($id)";
        $message .= $subMessage if $subMessage;
        if ($warnings || $subMessage)
        {
            $response = new MAM::Response()->warning(1, $message . $warnings);
        }
        else
        {
            $response = new MAM::Response()->success(1, $message);
        }
        return $response;
    }

    else
    {
        return new MAM::Response()
          ->failure('315', "The Bank class does not implement $object Stop");
    }
}

# ----------------------------------------------------------------------------
# $response = transfer($request, $requestId);
# ----------------------------------------------------------------------------

# Transfer (Fund)
sub transfer
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object             = $request->getObject();
    my $actor              = $request->getActor();
    my $database           = $request->getDatabase();
    my $amount             = $request->getOptionValue('Amount');
    my $fromId             = $request->getOptionValue('FromId');
    my $toId               = $request->getOptionValue('ToId');
    my $fromAllocation     = $request->getOptionValue('FromAllocation');
    my $toAllocation       = $request->getOptionValue('ToAllocation');
    my @options            = $request->getOptions();
    my $currency_precision = $serverConfig->{'currency.precision'} || 0;
    my $showHours          = $request->getOptionValue('ShowHours');
    my $useHours           = $request->getOptionValue('UseHours');
    my $now                = $requestTime;
    my %allocations        = ();
    my $active             = 0;
    my $time_division = (defined $showHours && $showHours eq 'True') ? 3600 : 1;
    my ($sourceFund, $destinationFund);

    # Scale amount into seconds if useHours was specified
    $amount = $amount * 3600
      if (defined $useHours && $useHours =~ /^true$/i);
    my $remaining = $amount;

    if ($object ne 'Fund')
    {
        return new MAM::Response()
          ->failure('315',
            "The Bank class does not implement $object Transfer");
    }

    # Amount must be non-negative and real
    if (! defined $amount)
    {
        return new MAM::Response()
          ->failure('314', 'The transfer amount must be specified');
    }
    if ($amount !~ /^\d*\.?\d+(e\+\d+)?$/)
    {
        return new MAM::Response()
          ->failure('317',
            'The transfer amount must be a non-negative real number');
    }

    # Either FromId or FromAllocation must be specified
    if ((! defined $fromId) && (! defined $fromAllocation))
    {
        return new MAM::Response()->failure('314',
            'Either a source fund id (FromId) or a source allocation id (FromAllocation) must be specified'
        );
    }

    # Either ToId or ToAllocation must be specified
    if ((! defined $toId) && (! defined $toAllocation))
    {
        return new MAM::Response()->failure('314',
            'Either a destination fund id (ToId) or a destination allocation id (ToAllocation) must be specified'
        );
    }

    # Refresh Allocations
    my $subRequest = new MAM::Request(
        database => $database,
        object   => 'Allocation',
        action   => 'Refresh'
    );
    my $subResponse = MAM::Bank->refresh($subRequest, $requestId);
    if ($subResponse->getStatus() eq 'Failure')
    {
        return new MAM::Response()->failure($subResponse->getCode(),
            'Failed while refreshing allocations: '
              . $subResponse->getMessage());
    }

# Look for candidate source allocations
# SELECT Id,Amount,CreditLimit,StartTime,EndTime FROM Allocation WHERE AND [Fund=$fromId | Id=$fromAllocation]

    my @selections = (
        new MAM::Selection(name => 'Id'),
        new MAM::Selection(name => 'Fund'),
        new MAM::Selection(name => 'Amount'),
        new MAM::Selection(name => 'CreditLimit'),
        new MAM::Selection(name => 'StartTime'),
        new MAM::Selection(name => 'EndTime'),
    );
    my @sourceConditions = ();

    if (defined $fromId)
    {
        push @sourceConditions,
          new MAM::Condition(name => 'Fund', value => $fromId);
    }

    if (defined $fromAllocation)
    {
        push @sourceConditions,
          new MAM::Condition(name => 'Id', value => $fromAllocation);
    }
    else
    {
        push @sourceConditions,
          new MAM::Condition(
            name  => 'StartTime',
            value => $now,
            op    => 'LE'
          );
        push @sourceConditions,
          new MAM::Condition(
            name  => 'EndTime',
            value => $now,
            op    => 'GT'
          );
    }

    my $results = $database->select(
        object     => 'Allocation',
        selections => \@selections,
        conditions => \@sourceConditions
    );

    # Fail if there are no valid allocations to withdraw from
    unless (@{$results->{data}})
    {
        return new MAM::Response()->failure('740',
            'There are no valid allocations from which to perform the transfer'
        );
    }

    foreach my $row (@{$results->{data}})
    {
        my $allocationId = $row->[0];
        $sourceFund = $row->[1];
        my $allocationAmount      = $row->[2];
        my $allocationCreditLimit = $row->[3];
        my $allocationStartTime   = $row->[4];
        my $allocationEndTime     = $row->[5];

        $allocations{$allocationId}{startTime} = $allocationStartTime;
        $allocations{$allocationId}{endTime}   = $allocationEndTime;
        $allocations{$allocationId}{balance} =
          $allocationAmount + $allocationCreditLimit;
    }

# Subtract active liens from allocation balances
# Lookup lien amounts against these funds
# SELECT LienAllocation.Id,LienAllocation.Amount FROM Lien,LienAllocation WHERE Lien.Id=LienAllocation.Lien AND Lien.StartTime<=$now AND Lien.EndTime>$now AND Fund=$sourceFund

    my @objects = (
        new MAM::Object(name => 'Lien'),
        new MAM::Object(name => 'LienAllocation'),
    );

    @selections = (
        new MAM::Selection(
            object => 'LienAllocation',
            name   => 'Id'
        ),
        new MAM::Selection(
            object => 'LienAllocation',
            name   => 'Amount'
        ),
    );

    my @lienConditions = (
        new MAM::Condition(
            object  => 'Lien',
            name    => 'Id',
            subject => 'LienAllocation',
            value   => 'Lien'
        ),
    );
    if (defined $fromId)
    {
        push @lienConditions,
          new MAM::Condition(
            object => 'LienAllocation',
            name   => 'Fund',
            value  => $fromId
          );
    }
    if (defined $fromAllocation)
    {
        push @lienConditions,
          new MAM::Condition(
            object => 'LienAllocation',
            name   => 'Id',
            value  => $fromAllocation
          );
    }
    else
    {
        push @lienConditions,
          new MAM::Condition(
            object => 'Lien',
            name   => 'StartTime',
            value  => $now,
            op     => 'LE'
          );
        push @lienConditions,
          new MAM::Condition(
            object => 'Lien',
            name   => 'EndTime',
            value  => $now,
            op     => 'GT'
          );
    }

    $results = $database->select(
        objects    => \@objects,
        selections => \@selections,
        conditions => \@lienConditions
    );
    foreach my $row (@{$results->{data}})
    {
        my $allocationId = $row->[0];
        my $lienAmount   = $row->[1];

        if (exists $allocations{$allocationId})
        {
            $allocations{$allocationId}{balance} -= $lienAmount;
        }
    }

    # Sort the allocations by weight (earliest endtime first)
    my @allocations =
      sort { $allocations{$a}{endTime} <=> $allocations{$b}{endTime} }
      keys %allocations;

    # Iterate through the allocation list
    foreach my $allocationId (@allocations)
    {
        # The amount I subtract from this allocation should be the minimum of:
        # 1. $remaining (the amount left to withdraw)
        # 2. $allocations{$allocationId}{balance}
        my $transferAmount =
          min($remaining, $allocations{$allocationId}{balance});
        my $sourceAllocationStartTime = $allocations{$allocationId}{startTime};
        my $sourceAllocationEndTime   = $allocations{$allocationId}{endTime};
        next unless $transferAmount;

        # Debit the allocation
        my $subRequest = new MAM::Request(
            database    => $database,
            object      => 'Allocation',
            action      => 'Modify',
            actor       => $SUPER_USER,
            assignments => [
                new MAM::Assignment(
                    name  => 'Amount',
                    value => $transferAmount,
                    op    => 'Dec'
                ),
                new MAM::Assignment(
                    name  => 'Allocated',
                    value => $transferAmount,
                    op    => 'Dec'
                )
            ],
            conditions => [
                new MAM::Condition(
                    name  => 'Fund',
                    value => $sourceFund
                ),
                new MAM::Condition(
                    name  => 'Id',
                    value => $allocationId
                )
            ]
        );
        my $subResponse = MAM::Base->modify($subRequest, $requestId, 'NoLog');
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Unable to debit allocation: ' . $subResponse->getMessage());
        }

        # Log the transaction
        MAM::Bank->logTransaction(
            database    => $database,
            requestId   => $requestId,
            txnId       => $database->nextId('Transaction'),
            object      => 'Allocation',
            action      => 'Modify',
            actor       => $actor,
            assignments => [
                new MAM::Assignment(
                    name  => 'Amount',
                    value => $transferAmount,
                )
            ],
            conditions => [
                new MAM::Condition(
                    name  => 'Id',
                    value => $allocationId
                )
            ],
            count      => 1,
            fund       => $sourceFund,
            allocation => $allocationId
        );

        # Obtain remaining allocation amount
        my $results = $database->select(
            object     => 'Allocation',
            selections => [new MAM::Selection(name => 'Amount')],
            conditions => [
                new MAM::Condition(
                    name  => 'Id',
                    value => $allocationId
                )
            ]
        );
        my $remainingAllocation = ${$results->{data}}[0]->[0];

        my @debitOptions = @options;
        push @debitOptions, new MAM::Option(name => 'Id', value => $sourceFund);
        push @debitOptions,
          new MAM::Option(name => 'Child', value => $allocationId);
        if (   $sourceAllocationStartTime <= $now
            && $sourceAllocationEndTime > $now)
        {
            MAM::Bank->logTransaction(
                database   => $database,
                requestId  => $requestId,
                txnId      => $database->nextId('Transaction'),
                object     => 'Fund',
                action     => 'Transfer',
                actor      => $actor,
                options    => \@debitOptions,
                count      => 1,
                fund       => $sourceFund,
                delta      => 0 - $transferAmount,
                balance    => $remainingAllocation,
                remaining  => $remainingAllocation,
                allocation => $allocationId
            );
        }
        else
        {
            MAM::Bank->logTransaction(
                database   => $database,
                requestId  => $requestId,
                txnId      => $database->nextId('Transaction'),
                object     => 'Fund',
                action     => 'Transfer',
                actor      => $actor,
                options    => \@debitOptions,
                count      => 1,
                fund       => $sourceFund,
                delta      => 0,
                balance    => 0,
                remaining  => $remainingAllocation,
                allocation => $allocationId
            );
        }
        $remaining -= $transferAmount;

        # We transfer to destination in chunks because we need to find
        # allocations with start and end times that are compatible with
        # the source allocation.

        # Obtain all allocations for the destination fund and store in a hash
        # SELECT Id,StartTime,EndTime FROM Allocation [WHERE Fund=$toId]

        my %preexistingAllocation = ();
        my @destinationConditions = ();
        push @destinationConditions,
          new MAM::Condition(name => 'Fund', value => $toId)
          if defined $toId;
        $results = $database->select(
            object     => 'Allocation',
            selections => [
                new MAM::Selection(name => 'Id'),
                new MAM::Selection(name => 'Fund'),
                new MAM::Selection(name => 'StartTime'),
                new MAM::Selection(name => 'EndTime'),
            ],
            conditions => \@destinationConditions,
        );

        # Populate the preexisting allocation hash with the results
        foreach my $row (@{$results->{data}})
        {
            my $id        = $row->[0];
            my $fund      = $row->[1];
            my $startTime = $row->[2];
            my $endTime   = $row->[3];

            $preexistingAllocation{$id}{fund}      = $fund;
            $preexistingAllocation{$id}{startTime} = $startTime;
            $preexistingAllocation{$id}{endTime}   = $endTime;
        }

        $allocationId = 0;    # Reuse allocation id for destination allocation
        $remainingAllocation = 0;    # Reuse remaining allocation

        my ($destinationAllocationStartTime, $destinationAllocationEndTime);

        # If allocation specified
        if (defined $toAllocation)
        {
            if (exists $preexistingAllocation{$toAllocation})
            {
                $allocationId = $toAllocation;
            }
            else
            {
                return new MAM::Response()->failure('740',
                    "The specified destination allocation ($toAllocation) is not valid"
                );
            }
        }

        # No allocation id specified. Try to pick an appropriate existing one
        else
        {
            # Look for an allocation in the destination fund matching the start
            # and end time of the source allocation.
            foreach my $id (reverse sort keys %preexistingAllocation)
            {
                if ($preexistingAllocation{$id}{startTime} ==
                       $sourceAllocationStartTime
                    && $preexistingAllocation{$id}{endTime} ==
                    $sourceAllocationEndTime)
                {
                    $allocationId = $id;
                    last;
                }
            }

            # If no allocation was matched, look for an active allocation
            unless ($allocationId)
            {
                foreach my $id (reverse sort keys %preexistingAllocation)
                {
                    if (   $preexistingAllocation{$id}{startTime} <= $now
                        && $preexistingAllocation{$id}{endTime} > $now)
                    {
                        $allocationId = $id;
                        last;
                    }
                }
            }
        }

        # An Allocation already exists
        if ($allocationId)
        {
            $destinationFund = $preexistingAllocation{$allocationId}{fund};
            $destinationAllocationStartTime =
              $preexistingAllocation{$allocationId}{startTime};
            $destinationAllocationEndTime =
              $preexistingAllocation{$allocationId}{endTime};

            # Credit the allocation
            my $subRequest = new MAM::Request(
                database    => $database,
                object      => 'Allocation',
                action      => 'Modify',
                actor       => $SUPER_USER,
                assignments => [
                    new MAM::Assignment(
                        name  => 'Amount',
                        value => $transferAmount,
                        op    => 'Inc'
                    ),
                    new MAM::Assignment(
                        name  => 'Allocated',
                        value => $transferAmount,
                        op    => 'Inc'
                    )
                ],
                conditions => [
                    new MAM::Condition(
                        name  => 'Fund',
                        value => $destinationFund
                    ),
                    new MAM::Condition(
                        name  => 'Id',
                        value => $allocationId
                    )
                ]
            );
            my $subResponse =
              MAM::Base->modify($subRequest, $requestId, 'NoLog');
            if ($subResponse->getStatus() eq 'Failure')
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    'Unable to credit allocation: '
                      . $subResponse->getMessage());
            }

            # Log the transaction
            MAM::Bank->logTransaction(
                database    => $database,
                requestId   => $requestId,
                txnId       => $database->nextId('Transaction'),
                object      => 'Allocation',
                action      => 'Modify',
                actor       => $actor,
                assignments => [
                    new MAM::Assignment(
                        name  => 'Amount',
                        value => $transferAmount,
                    )
                ],
                conditions => [
                    new MAM::Condition(
                        name  => 'Id',
                        value => $allocationId
                    )
                ],
                count      => 1,
                fund       => $destinationFund,
                allocation => $allocationId
            );

            # Obtain remaining allocation amount
            my $results = $database->select(
                object     => 'Allocation',
                selections => [new MAM::Selection(name => 'Amount')],
                conditions => [
                    new MAM::Condition(
                        name  => 'Id',
                        value => $allocationId
                    )
                ]
            );
            $remainingAllocation = ${$results->{data}}[0]->[0];
        }

        # An Allocation does not exist (but a target fund was specified)
        else
        {
            $destinationFund                = $toId;
            $destinationAllocationStartTime = $sourceAllocationStartTime;
            $destinationAllocationEndTime   = $sourceAllocationEndTime;

            my $enforceDiscreteOption =
              $request->getOptionValue('EnforceDiscrete');
            my $enforceDiscreteValue =
              defined $enforceDiscreteOption
              ? $enforceDiscreteOption
              : ($serverConfig->{'allocation.enforcediscrete'} || 'true');
            my $enforceDiscreteAllocations =
              ($enforceDiscreteValue =~ /^false$/i) ? 0 : 1;

            # If we are enforcing discrete allocations, make sure this new
            # allocation does not conflict with an existing one
            if ($enforceDiscreteAllocations)
            {
                foreach my $id (sort keys %preexistingAllocation)
                {
                    my $startTime = $preexistingAllocation{$id}{startTime};
                    my $endTime   = $preexistingAllocation{$id}{endTime};
                    # If an existing allocation does not (end before or
                    # start after) our proposed one, it conflicts
                    if (! (    ($endTime <= $destinationAllocationStartTime)
                            || ($startTime >= $destinationAllocationEndTime)
                        )
                      )
                    {
                        my $message = sprintf(
                            "The resulting new allocation with start time (%s) and end time (%s) would overlap with an existing allocation (%d) in the same fund (%d) with start time (%s) and end time (%s). You can create a new active allocation in the fund via an empty deposit and reissue the refund or rerun the command with the EnforceDiscrete option set to False (and risk that the created allocation may conflict with future allocations)",
                            toHRT($destinationAllocationStartTime),
                            toHRT($destinationAllocationEndTime),
                            $id,
                            $destinationFund,
                            toHRT($startTime),
                            toHRT($endTime)
                        );
                        return new MAM::Response()->failure('740', $message);
                    }
                }
            }

            my $destinationActive;
            if (   $destinationAllocationStartTime <= $now
                && $destinationAllocationEndTime > $now)
            {
                $destinationActive = 'True';
            }
            else
            {
                $destinationActive = 'False';
            }

            # Create a new one
            my $subRequest = new MAM::Request(
                database    => $database,
                object      => 'Allocation',
                action      => 'Create',
                actor       => $SUPER_USER,
                assignments => [
                    new MAM::Assignment(
                        name  => 'Fund',
                        value => $destinationFund
                    ),
                    new MAM::Assignment(
                        name  => 'Active',
                        value => $destinationActive
                    ),
                    new MAM::Assignment(
                        name  => 'StartTime',
                        value => $destinationAllocationStartTime
                    ),
                    new MAM::Assignment(
                        name  => 'EndTime',
                        value => $destinationAllocationEndTime
                    ),
                    new MAM::Assignment(
                        name  => 'Amount',
                        value => $transferAmount
                    ),
                    new MAM::Assignment(
                        name  => 'CreditLimit',
                        value => 0
                    ),
                    new MAM::Assignment(
                        name  => 'InitialDeposit',
                        value => 0
                    ),
                    new MAM::Assignment(
                        name  => 'Allocated',
                        value => $transferAmount
                    ),
                ]
            );
            my $subResponse = MAM::Base->create($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Failure')
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    'Unable to create allocation: '
                      . $subResponse->getMessage());
            }

            # Figure out the id of the allocation we just created
            $allocationId        = $subResponse->getDatumValue('Id');
            $remainingAllocation = $transferAmount;
        }

        # Log the transaction
        my @creditOptions = @options;
        push @creditOptions,
          new MAM::Option(name => 'Id', value => $destinationFund);
        push @creditOptions,
          new MAM::Option(name => 'Child', value => $allocationId);
        if (   $destinationAllocationStartTime <= $now
            && $destinationAllocationEndTime > $now)
        {
            MAM::Bank->logTransaction(
                database    => $database,
                requestId   => $requestId,
                txnId       => $database->nextId('Transaction'),
                object      => 'Fund',
                action      => 'Transfer',
                actor       => $actor,
                options     => \@creditOptions,
                count       => 1,
                fund        => $destinationFund,
                delta       => $transferAmount,
                balance     => $remainingAllocation,
                remaining   => $remainingAllocation,
                allocation  => $allocationId,
                assignments => [
                    new MAM::Assignment(
                        name  => 'Amount',
                        value => $transferAmount,
                    )
                ],
            );
        }
        else
        {
            MAM::Bank->logTransaction(
                database    => $database,
                requestId   => $requestId,
                txnId       => $database->nextId('Transaction'),
                object      => 'Fund',
                action      => 'Transfer',
                actor       => $actor,
                options     => \@creditOptions,
                count       => 1,
                fund        => $destinationFund,
                delta       => 0,
                balance     => 0,
                remaining   => $remainingAllocation,
                allocation  => $allocationId,
                assignments => [
                    new MAM::Assignment(
                        name  => 'Amount',
                        value => $transferAmount,
                    )
                ],
            );
        }

        # Break out if we are done
        last if $remaining <= 0;
    }

    # Fail if not all amount could be transferred
    if ($remaining > 0)
    {
        return new MAM::Response()
          ->failure('782',
            'Insufficient balance to perform requested transfer');
    }

    $amount = sprintf("%.${currency_precision}f", $amount / $time_division);
    return new MAM::Response()->success($amount,
        "Successfully transferred $amount credits from fund $sourceFund to fund $destinationFund"
    );
}

# ----------------------------------------------------------------------------
# $response = withdraw($request, $requestId);
# ----------------------------------------------------------------------------

# Withdraw (Fund)
sub withdraw
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object      = $request->getObject();
    my $actor       = $request->getActor();
    my $database    = $request->getDatabase();
    my $amount      = $request->getOptionValue('Amount');
    my $id          = $request->getOptionValue('Id');
    my $allocation  = $request->getOptionValue('Allocation');
    my $ignoreLiens = $request->getOptionValue('IgnoreLiens')
      || 'False';
    my @options            = $request->getOptions();
    my %allocations        = ();
    my $currency_precision = $serverConfig->{'currency.precision'} || 0;
    my $showHours          = $request->getOptionValue('ShowHours');
    my $useHours           = $request->getOptionValue('UseHours');
    my $now                = $requestTime;
    my $time_division = (defined $showHours && $showHours eq 'True') ? 3600 : 1;

    # Scale amount into seconds if useHours was specified
    $amount = $amount * 3600
      if (defined $useHours && $useHours =~ /^true$/i);
    my $remaining = $amount;

    my $filterType = $request->getOptionValue('FilterType')
      || 'NonExclusive';
    if (   $filterType ne 'ExactMatch'
        && $filterType ne 'Exclusive'
        && $filterType ne 'NonExclusive')
    {
        return new MAM::Response()->failure('317',
            'The FilterType option must be one of Exclusive, NonExclusive or ExactMatch'
        );
    }

    if ($object ne 'Fund')
    {
        return new MAM::Response()
          ->failure('315',
            "The Bank class does not implement $object Withdraw");
    }

    # Amount must be non-negative and real
    if (! defined $amount)
    {
        return new MAM::Response()->failure('314', 'Amount must be specified');
    }
    if ($amount !~ /^\d*\.?\d+(e\+\d+)?$/)
    {
        return new MAM::Response()
          ->failure('317',
            'The withdrawal amount must be a non-negative real number');
    }

    # Invoke base Allocation Modify
    if ($amount <= 0)
    {
        return new MAM::Response()->failure('317', 'Amount must be positive');
    }

    # Parse out filters
    my %filters = ();
    foreach my $option (@options)
    {
        if ($option->getName() eq 'Filter')
        {
            my $filter = $option->getValue();
            if ($filter =~ /^([^\s=!]+)=([^\s=]+)$/)
            {
                $filters{$1} = $2;
            }
        }
    }

    # Refresh Allocations
    my $subRequest = new MAM::Request(
        database => $database,
        object   => 'Allocation',
        action   => 'Refresh'
    );
    my $subResponse = MAM::Bank->refresh($subRequest, $requestId);
    if ($subResponse->getStatus() eq 'Failure')
    {
        return new MAM::Response()->failure($subResponse->getCode(),
            'Failed while refreshing allocations: '
              . $subResponse->getMessage());
    }

# First, we build a list of funds
# We need to do this separately from the constraint query
# because the trivial fund might have no constraints
# SELECT [DISTINCT] [Fund.]Id FROM Fund[,Allocation] [WHERE Allocation.Fund=Fund.Id AND Allocation.Id=$allocation] [AND [Fund.]Id=$id]
    my %funds = ();
    my @fundObjects = (new MAM::Object(name => 'Fund'));
    my @fundSelections =
      (new MAM::Selection(name => 'Id'), new MAM::Selection(name => 'Name'));
    my @fundConditions = ();
    push @fundConditions,
      new MAM::Condition(
        name  => 'Id',
        value => $id
      ) if defined $id;
    my @fundOptions = ();
    if (defined $allocation)
    {
        push @fundObjects, new MAM::Object(name => 'Allocation');
        foreach my $selection (@fundSelections)
        {
            $selection->setObject('Fund');
        }
        foreach my $condition (@fundConditions)
        {
            $condition->setObject('Fund');
        }
        push @fundConditions,
          ( new MAM::Condition(
                object  => 'Fund',
                name    => 'Id',
                subject => 'Allocation',
                value   => 'Fund'
            ),
            new MAM::Condition(
                object => 'Allocation',
                name   => 'Id',
                value  => $allocation,
            )
          );
        push @fundOptions, new MAM::Option(name => 'Unique', value => 'True');
    }
    my $results = $database->select(
        objects    => \@fundObjects,
        selections => \@fundSelections,
        conditions => \@fundConditions,
        options    => \@fundOptions,
    );
    foreach my $row (@{$results->{data}})
    {
        my $id   = $row->[0];
        my $name = $row->[1];
        $funds{$id} = defined $name ? $name : '';
    }

    # Next, we prune off the funds that violate the constraints
    my %fundConstraints = ();

    # SELECT Id,Fund,Name,Value FROM Constraint
    $results = $database->select(
        object     => 'Constraint',
        selections => [
            new MAM::Selection(name => 'Id'),
            new MAM::Selection(name => 'Fund'),
            new MAM::Selection(name => 'Name'),
            new MAM::Selection(name => 'Value'),
        ],
    );
    foreach my $row (@{$results->{data}})
    {
        my $id     = $row->[0];
        my $fundId = $row->[1];
        my $name   = $row->[2];
        my $value  = $row->[3];

        push @{$fundConstraints{$fundId}{$name}}, $value;
    }

    if ($filterType eq 'ExactMatch')
    {
      FUNDID:
        foreach my $fundId (sort { $a <=> $b } keys %funds)
        {
            # Reject if there are any filters without corresponding constraints
            foreach my $filterName (sort keys %filters)
            {
                if (   (! defined $fundConstraints{$fundId})
                    || (! defined $fundConstraints{$fundId}{$filterName})
                    || (@{$fundConstraints{$fundId}{$filterName}} != 1)
                    || (${$fundConstraints{$fundId}{$filterName}}[0] ne
                        $filters{$filterName})
                  )
                {
                    delete $funds{$fundId};
                    next FUNDID;
                }
            }

            # Reject if there are any constraints without corresponding filters
            next unless exists $fundConstraints{$fundId};
            foreach my $constraintName (sort keys %{$fundConstraints{$fundId}})
            {
                if (   (! defined $filters{$constraintName})
                    || (@{$fundConstraints{$fundId}{$constraintName}} != 1)
                    || (${$fundConstraints{$fundId}{$constraintName}}[0] ne
                        $filters{$constraintName})
                  )
                {
                    delete $funds{$fundId};
                    next FUNDID;
                }
            }
        }
    }

    elsif ($filterType eq 'NonExclusive')
    {
      FUNDID:
        foreach my $fundId (sort { $a <=> $b } keys %fundConstraints)
        {
            my %constraintRejected = ();
            foreach my $filterName (sort keys %filters)
            {
                my $filterValue = $filters{$filterName};
                my $accepted    = 0;
                my $rejected    = 0;

                foreach
                  my $constraintName (sort keys %{$fundConstraints{$fundId}})
                {
                    foreach my $constraintVal (
                        @{$fundConstraints{$fundId}{$constraintName}})
                    {
                        my $constraintValue = $constraintVal;    # Make a copy
                        my $accessAllowed = $constraintValue !~ s/^!//;
                        if (exists $constraintRejected{$constraintName}
                            {$constraintValue})
                        {
                            $rejected = 1;    # This constraint was rejected
                            next;             # Don't allow it to be accepted
                        }

                   # Same Name  |-1|-1|-1| 0| 0| 0|+1|+1|+1|
                   # Other Name |-1| 0|+1|-1| 0|+1|-1| 0|+1|
                   # Accpt/Rjct |R |R |R |  |  |R | A| A|RA|
                   # Result     |X |X |X |/ |/ |X |/ |/ |/ |
                   # A filter should be accepted unless rejected unless accepted

                        # We have a matching constraint name
                        if ($filterName eq $constraintName)
                        {
                            # Mark as accepted if it is a positive
                            # constraint with a matching value
                            if (   $accessAllowed
                                && $filterValue eq $constraintValue)
                            {
                                $accepted = 1;
                            }
                            # Mark as rejected if it is a non-matching value
                            elsif (

                                $accessAllowed
                                && $filterValue ne $constraintValue
                              )
                            {
                                $rejected = 1;
                                $constraintRejected{$constraintName}
                                  {$constraintValue} = 1;
                            }
                            # Immediately delete if it is explicitly rejected
                            elsif (! $accessAllowed
                                && $filterValue eq $constraintValue)
                            {
                                delete $funds{$fundId};
                                next FUNDID;
                            }
                        }

                 # If the constraint is in a parent association with this filter
                 # check it for membership
                 # We could do some caching here but we are not doing it yet
                        elsif (
                            $accessAllowed
                            && MAM::Cache->associationLookup($constraintName,
                                $filterName)
                            && (! exists $fundConstraints{$fundId}{$filterName}
                                || ! grep ! /!/,
                                @{$fundConstraints{$fundId}{$filterName}}
                            )
                          )
                        {
                            my ($association) =
                              MAM::Cache->associationLookup($constraintName,
                                $filterName);
                            # If there is not a matching member found
                            # reject the fund
                            my @membershipConditions = (
                                new MAM::Condition(
                                    name => 'Name',  # Assume simple association
                                    value => $filterValue,
                                    conj  => 'And',
                                    group => '+1',
                                ),
                                new MAM::Condition(
                                    name  => 'Name', # Assume simple association
                                    value => 'ANY',
                                    conj  => 'Or',
                                    group => '-1',
                                )
                            );
                            push @membershipConditions,
                              new MAM::Condition(
                                name  => 'Active',
                                value => 'True'
                              )
                              if MAM::Cache->attributeExists($association,
                                'Active');
                            my $subResults = $database->select(
                                object => $association,
                                selections =>
                                  [new MAM::Selection(name => $constraintName)],
                                conditions => \@membershipConditions,
                            );
                            my $keep = 0;
                            foreach my $row (@{$subResults->{data}})
                            {
                                my $parent = $row->[0];
                                if ($parent eq $constraintValue)
                                {
                                    $keep = 1;
                                }
                            }
                            if ($keep) { $accepted = 1; }
                            else
                            {
                                $rejected = 1;
                                $constraintRejected{$constraintName}
                                  {$constraintValue} = 1;
                            }
                        }

                  # If the constraint is in a child association with this filter
                  # check it for membership
                  # We could do some caching here but we are not doing it yet
                        elsif (
                            $accessAllowed
                            && MAM::Cache->associationLookup($filterName,
                                $constraintName)
                            && ! exists $fundConstraints{$fundId}{$filterName}
                          )
                        {
                            my ($association) =
                              MAM::Cache->associationLookup($filterName,
                                $constraintName);
                            # If there is not a matching member found
                            # reject the fund
                            my @membershipConditions = (
                                new MAM::Condition(
                                    name  => $filterName,
                                    value => $filterValue,
                                )
                            );
                            push @membershipConditions,
                              new MAM::Condition(
                                name  => 'Active',
                                value => 'True'
                              )
                              if MAM::Cache->attributeExists($association,
                                'Active');
                            my $subResults = $database->select(
                                object => $association,
                                selections =>
                                  [new MAM::Selection(name => 'Name')]
                                ,    # Simple name
                                conditions => \@membershipConditions,
                            );
                            my $keep = 0;
                            foreach my $row (@{$subResults->{data}})
                            {
                                my $child = $row->[0];
                                if ($child eq $constraintValue)
                                {
                                    $keep = 1;
                                }
                            }
                            if ($keep) { $accepted = 1; }
                            else
                            {
                                $rejected = 1;
                                $constraintRejected{$constraintName}
                                  {$constraintValue} = 1;
                            }
                        }
                    }
                }

                if ($rejected && ! $accepted)
                {
                    # Delete them
                    delete $funds{$fundId};
                    next FUNDID;
                }
            }
        }
    }

    elsif ($filterType eq 'Exclusive')
    {
        foreach my $fundId (sort { $a <=> $b } keys %fundConstraints)
        {
            foreach my $constraintName (sort keys %{$fundConstraints{$fundId}})
            {
                my %accepted = ();
                my %rejected = ();
                foreach my $constraintValue (
                    @{$fundConstraints{$fundId}{$constraintName}})
                {
                    my $accessAllowed = $constraintValue !~ s/^!//;

                    # If there is a filter matching the constraint
                    if (defined $filters{$constraintName}
                        && $filters{$constraintName} eq $constraintValue)
                    {
                        # It's a matching positive constraint
                        if ($accessAllowed)
                        {
                            my @associations =
                              MAM::Cache->associationLookup($constraintName,
                                undef);

                            # There are associations
                            if (@associations)
                            {
                                foreach my $association (@associations)
                                {
                                    my $child = MAM::Cache->getObjectProperty(
                                        $association, 'Child');
                                    # Check if the usage record property is a
                                    # parent in an association
                                    if (defined $filters{$child})
                                    {
                                        # Check for a match unless the
                                        # fund has a matching constraint
                                        unless (
                                            exists $fundConstraints{$fundId}
                                            {$child})
                                        {
                             # If there is a matching member found, add the fund
                             # We could improve upon this by caching the results
                                            my @membershipConditions = (
                                                new MAM::Condition(
                                                    name => $constraintName,
                                                    value =>
                                                      $filters{$constraintName},
                                                ),
                                                new MAM::Condition(
                                                    name => 'Name'
                                                    , # Assume simple association
                                                    value => $filters{$child},
                                                    conj  => 'And',
                                                    group => '+1',
                                                ),
                                                new MAM::Condition(
                                                    name => 'Name'
                                                    , # Assume simple association
                                                    value => 'ANY',
                                                    conj  => 'Or',
                                                    group => '-1',
                                                ),
                                            );
                                            push @membershipConditions,
                                              new MAM::Condition(
                                                name  => 'Active',
                                                value => 'True'
                                              )
                                              if MAM::Cache->attributeExists(
                                                $association, 'Active');
                                            my $subResults = $database->select(
                                                object => $association,
                                                conditions =>
                                                  \@membershipConditions,
                                            );
                                            if (@{$subResults->{data}})
                                            {
                                                $accepted{$constraintName} = 1;
                                            }
                                            else
                                            {
                                                $rejected{$constraintName} = 1;
                                            }
                                        }
                                        $accepted{$constraintName} = 1;
                                    }
                                    # Child associations must be specified
                                    # for an Exclusive filter
                                    else
                                    {
                                        return new MAM::Response()->failure(
                                            '314',
                                            "A $child must be specified when specifying a $constraintName"
                                        );
                                    }
                                }
                            }
                            # Accept it if there are no associations
                            else
                            {
                                $accepted{$constraintName} = 1;
                            }
                        }
                        # Reject the matching negative constraint
                        else
                        {
                            $rejected{$constraintName} = 1;
                        }
                    }
                    # There is not a filter matching the positive constraint
                    # or the filter does not exist or the constraint is negative
                    else
                    {
                        if ($accessAllowed
                            || ! defined $filters{$constraintName})
                        {
                            $rejected{$constraintName} = 1;
                        }
                    }
                }

                # Constraints of the same name are Ored while
                # Constraints of different names are Anded
                # Delete if there is a constraint name
                # that is rejected and not accepted
                foreach my $name (sort keys %rejected)
                {
                    if (! $accepted{$name} && $funds{$fundId})
                    {
                        # Delete them
                        delete $funds{$fundId};
                    }
                }
            }
        }
    }

    if ($log->is_debug())
    {
        $log->debug('Feasible fund list is: ' . join ',', sort keys %funds);
    }

    # Return immediately if there are no matching funds
    if (scalar keys %funds == 0)
    {
        return new MAM::Response()
          ->failure('740', 'No funds could be matched for the withdrawal');
    }

    # List options if multiple funds were matched
    elsif (scalar keys %funds > 1)
    {
        my $message =
          "Multiple funds were matched for the withdrawal.\nPlease respecify using one of the following fund ids:\n"
          . join("\n", map {"$_\t[$funds{$_}]"} sort keys %funds);
        return new MAM::Response()->failure('740', $message);
    }

    # We should have narrowed the fund down to a single id
    $id = ${[keys %funds]}[0] unless $id;

 # Look for candidate source allocations
 # SELECT Id,Amount,CreditLimit,StartTime,EndTime FROM Allocation WHERE Fund=$id
    my @selections = (
        new MAM::Selection(name => 'Id'),
        new MAM::Selection(name => 'Amount'),
        new MAM::Selection(name => 'CreditLimit'),
        new MAM::Selection(name => 'StartTime'),
        new MAM::Selection(name => 'EndTime'),
    );

    my @conditions = (new MAM::Condition(name => 'Fund', value => $id));

    # Restrict to allocation if we have it narrowed down to one
    if ($allocation)
    {
        push @conditions,
          new MAM::Condition(
            name  => 'Id',
            value => $allocation,
          );
    }

    # Otherwise use active allocations
    else
    {
        push @conditions,
          new MAM::Condition(
            name  => 'StartTime',
            value => $now,
            op    => 'LE'
          );
        push @conditions,
          new MAM::Condition(
            name  => 'EndTime',
            value => $now,
            op    => 'GT'
          );
    }

    $results = $database->select(
        object     => 'Allocation',
        selections => \@selections,
        conditions => \@conditions
    );

    # Fail if there are no valid allocations to withdraw from
    unless (@{$results->{data}})
    {
        return new MAM::Response()->failure('740',
            'Insufficient funds: There are no valid allocations from which to withdraw'
        );
    }

    foreach my $row (@{$results->{data}})
    {
        my $allocationId          = $row->[0];
        my $allocationAmount      = $row->[1];
        my $allocationCreditLimit = $row->[2];
        my $allocationStartTime   = $row->[3];
        my $allocationEndTime     = $row->[4];

        $allocations{$allocationId}{startTime} = $allocationStartTime;
        $allocations{$allocationId}{endTime}   = $allocationEndTime;
        $allocations{$allocationId}{weight}    = $allocationEndTime;
        $allocations{$allocationId}{balance} =
          $allocationAmount + $allocationCreditLimit;
    }

# Subtract active liens from allocation balances
# Lookup lien amounts against these funds
# SELECT LienAllocation.Id,LienAllocation.Amount FROM Lien,LienAllocation WHERE Lien.Id=LienAllocation.Lien AND Lien.StartTime<=$now AND Lien.EndTime>$now AND Fund=$fromId
    if ($ignoreLiens ne 'True')
    {
        my @objects = (
            new MAM::Object(name => 'Lien'),
            new MAM::Object(name => 'LienAllocation'),
        );

        @selections = (
            new MAM::Selection(
                object => 'LienAllocation',
                name   => 'Id'
            ),
            new MAM::Selection(
                object => 'LienAllocation',
                name   => 'Amount'
            ),
        );

        @conditions = (
            new MAM::Condition(
                object  => 'Lien',
                name    => 'Id',
                subject => 'LienAllocation',
                value   => 'Lien'
            ),
            new MAM::Condition(
                object => 'LienAllocation',
                name   => 'Fund',
                value  => $id
            ),
        );

        if (defined $allocation)
        {
            push @conditions,
              new MAM::Condition(
                object => 'LienAllocation',
                name   => 'Id',
                value  => $allocation
              );
        }
        else
        {
            push @conditions,
              new MAM::Condition(
                object => 'Lien',
                name   => 'StartTime',
                value  => $now,
                op     => 'LE'
              );
            push @conditions,
              new MAM::Condition(
                object => 'Lien',
                name   => 'EndTime',
                value  => $now,
                op     => 'GT'
              );
        }

        $results = $database->select(
            objects    => \@objects,
            selections => \@selections,
            conditions => \@conditions
        );
        foreach my $row (@{$results->{data}})
        {
            my $allocationId = $row->[0];
            my $lienAmount   = $row->[1];

            if (exists $allocations{$allocationId})
            {
                $allocations{$allocationId}{balance} -= $lienAmount;
            }
        }
    }

    # Sort the allocations by weight (earliest endtime first)
    my @allocations =
      sort { $allocations{$a}{endTime} <=> $allocations{$b}{endTime} }
      keys %allocations;

    # Iterate through the allocation list
    foreach my $allocationId (@allocations)
    {
# The amount I subtract from this allocation should be the minimum of:
# 1. $remaining (the amount left to withdraw)
# 2. $allocations{$allocationId}{balance} (the amount supportable by this allocation itself (and constrained by credit limit))
        my $debitAmount = min($remaining, $allocations{$allocationId}{balance});
        my $allocationStartTime = $allocations{$allocationId}{startTime};
        my $allocationEndTime   = $allocations{$allocationId}{endTime};
        next unless $debitAmount > 0;

        # Debit the allocation
        my $subRequest = new MAM::Request(
            database    => $database,
            object      => 'Allocation',
            action      => 'Modify',
            actor       => $SUPER_USER,
            assignments => [
                new MAM::Assignment(
                    name  => 'Amount',
                    value => $debitAmount,
                    op    => 'Dec'
                ),
                new MAM::Assignment(
                    name  => 'Allocated',
                    value => $debitAmount,
                    op    => 'Dec'
                )
            ],
            conditions => [
                new MAM::Condition(name => 'Fund', value => $id),
                new MAM::Condition(
                    name  => 'Id',
                    value => $allocationId
                )
            ]
        );
        my $subResponse = MAM::Base->modify($subRequest, $requestId, 'NoLog');
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Unable to debit allocation: ' . $subResponse->getMessage());
        }

        # Log the transaction
        MAM::Bank->logTransaction(
            database    => $database,
            requestId   => $requestId,
            txnId       => $database->nextId('Transaction'),
            object      => 'Allocation',
            action      => 'Modify',
            actor       => $actor,
            assignments => [$subRequest->getAssignments()],
            conditions  => [$subRequest->getConditions()],
            count       => 1,
            allocation  => $allocationId
        );

        # Obtain remaining allocation amount
        my $results = $database->select(
            object     => 'Allocation',
            selections => [new MAM::Selection(name => 'Amount')],
            conditions => [
                new MAM::Condition(
                    name  => 'Id',
                    value => $allocationId
                )
            ]
        );
        my $remainingAllocation = ${$results->{data}}[0]->[0];

        my @creditOptions = @options;
        push @creditOptions,
          new MAM::Option(name => 'Child', value => $allocationId);
        if ($allocationStartTime <= $now && $allocationEndTime > $now)
        {
            MAM::Bank->logTransaction(
                database   => $database,
                requestId  => $requestId,
                txnId      => $database->nextId('Transaction'),
                object     => 'Fund',
                action     => 'Withdraw',
                actor      => $actor,
                options    => \@creditOptions,
                count      => 1,
                fund       => $id,
                delta      => 0 - $debitAmount,
                balance    => $remainingAllocation,
                remaining  => $remainingAllocation,
                allocation => $allocationId
            );
        }
        else
        {
            MAM::Bank->logTransaction(
                database   => $database,
                requestId  => $requestId,
                txnId      => $database->nextId('Transaction'),
                object     => 'Fund',
                action     => 'Withdraw',
                actor      => $actor,
                options    => \@creditOptions,
                count      => 1,
                fund       => $id,
                delta      => 0,
                balance    => 0,
                remaining  => $remainingAllocation,
                allocation => $allocationId
            );
        }
        $remaining -= $debitAmount;

        # Break out if we are done
        last if $remaining <= 0;
    }

    # Fail if not all amount could be debited
    if ($remaining > 0)
    {
        return new MAM::Response()
          ->failure('782',
            'Insufficient balance to perform requested withdrawal');
    }

    $amount = sprintf("%.${currency_precision}f", $amount / $time_division);
    return new MAM::Response()
      ->success($amount, "Successfully withdrew $amount credits from fund $id");
}

# ----------------------------------------------------------------------------
# $response = undelete($request, $requestId);
# ----------------------------------------------------------------------------

# Undelete (Allocation, Fund)
sub undelete
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object     = $request->getObject();
    my $actor      = $request->getActor();
    my $database   = $request->getDatabase();
    my @conditions = $request->getConditions();
    my @options    = $request->getOptions();
    my $now        = $requestTime;

    if ($object eq 'Allocation')
    {
        # Refresh Allocations
        my $subRequest = new MAM::Request(
            database => $database,
            object   => 'Allocation',
            action   => 'Refresh'
        );
        my $subResponse = MAM::Bank->refresh($subRequest, $requestId);
        if ($subResponse->getStatus() eq 'Failure')
        {
            return new MAM::Response()->failure($subResponse->getCode(),
                'Failed while refreshing allocations: '
                  . $subResponse->getMessage());
        }

# Get a list of allocations to be undeleted
# SELECT Id,Fund,Amount,StartTime,EndTime FROM Allocation WHERE Deleted=True AND <Conditions>
        push @conditions,
          new MAM::Condition(name => 'Deleted', value => 'True');
        my $results = $database->select(
            object     => 'Allocation',
            selections => [
                new MAM::Selection(name => 'Id'),
                new MAM::Selection(name => 'Fund'),
                new MAM::Selection(name => 'Amount'),
                new MAM::Selection(name => 'StartTime'),
                new MAM::Selection(name => 'EndTime')
            ],
            conditions => \@conditions
        );
        foreach my $row (@{$results->{data}})
        {
            my $id        = $row->[0];
            my $fund      = $row->[1];
            my $amount    = $row->[2];
            my $startTime = $row->[3];
            my $endTime   = $row->[4];

            # Log the transaction (includes the new fund id and delta)
            my ($delta, $balance);
            if ($startTime <= $now && $endTime > $now)
            {
                $delta   = $amount;
                $balance = $amount;
            }
            else
            {
                $delta   = 0;
                $balance = 0;
            }

            MAM::Bank->logTransaction(
                database   => $database,
                requestId  => $requestId,
                txnId      => $database->nextId('Transaction'),
                object     => 'Allocation',
                action     => 'Undelete',
                actor      => $actor,
                conditions => [
                    new MAM::Condition(name => 'Id', value => $id),
                    new MAM::Condition(
                        name  => 'Description',
                        value => 'Redundant entry includes fund and delta'
                    )
                ],
                options    => \@options,
                count      => 1,
                fund       => $fund,
                delta      => $delta,
                balance    => $balance,
                remaining  => $amount,
                allocation => $id
            );
        }

        # Undelete allocations by calling base Allocation Undelete
        my $baseResponse = MAM::Base->undelete($request, $requestId);

        return $baseResponse;
    }

    elsif ($object eq 'Fund')
    {
        # Get a list of funds to be undeleted
        # SELECT Id FROM Fund WHERE Deleted=True && <Conditions>
        push @conditions,
          new MAM::Condition(name => 'Deleted', value => 'True');
        my $results = $database->select(
            object     => 'Fund',
            selections => [new MAM::Selection(name => 'Id')],
            conditions => \@conditions
        );
        foreach my $row (@{$results->{data}})
        {
            my $id = $row->[0];

            # Undelete all allocations associated with this fund
            my $subRequest = new MAM::Request(
                database   => $database,
                object     => 'Allocation',
                action     => 'Undelete',
                actor      => $SUPER_USER,
                conditions => [new MAM::Condition(name => 'Fund', value => $id)]
            );
            my $subResponse = MAM::Bank->undelete($subRequest, $requestId);
            if ($subResponse->getStatus() eq 'Failure')
            {
                return new MAM::Response()->failure($subResponse->getCode(),
                    'Unable to undelete allocation: '
                      . $subResponse->getMessage());
            }
        }

        # Undelete funds by calling base Fund Undelete
        my $baseResponse = MAM::Base->undelete($request, $requestId);

        return $baseResponse;
    }

    else
    {
        return new MAM::Response()
          ->failure('315',
            "The Bank class does not implement $object Undelete");
    }
}

# ----------------------------------------------------------------------------
# $response = update($request, $requestId);
# ----------------------------------------------------------------------------

# Update (Payment)
sub update
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $object   = $request->getObject();
    my $database = $request->getDatabase();
    my $actor    = $request->getActor();
    my @options  = $request->getOptions();
    my $now      = $requestTime;
    my $warnings = 0;
    my $response;

    if ($object ne 'Payment')
    {
        return new MAM::Response()
          ->failure('315', "The Bank class does not implement $object Update");
    }

    # Id
    my $id = $request->getOptionValue('Id');
    if (! defined $id)
    {
        return new MAM::Response()
          ->failure('314', 'An Id option must be specified');
    }

# SELECT ChargePattern,ChargePeriod,EndTime,ReservePattern,UpdateEvent FROM Payment WHERE Id=$id
    my $results = $database->select(
        object     => 'Payment',
        selections => [
            new MAM::Selection(name => 'ChargePattern'),
            new MAM::Selection(name => 'ChargePeriod'),
            new MAM::Selection(name => 'EndTime'),
            new MAM::Selection(name => 'ReservePattern'),
            new MAM::Selection(name => 'UpdateEvent'),
        ],
        conditions => [
            new MAM::Condition(
                name  => 'Id',
                value => $id,
            ),
        ],
    );

    unless (@{$results->{data}})
    {
        return new MAM::Response()
          ->failure('740', "Payment Update: Invalid payment id ($id)");
    }

    my $chargePattern  = ${$results->{data}}[0]->[0];
    my $chargePeriod   = ${$results->{data}}[0]->[1];
    my $endTime        = ${$results->{data}}[0]->[2];
    my $reservePattern = ${$results->{data}}[0]->[3];
    my $updateEvent    = ${$results->{data}}[0]->[4];
    unless (defined $chargePattern)
    {
        return new MAM::Response()
          ->failure('740',
            'Payment Update: Unable to determine payment charge mode');
    }
    unless (defined $reservePattern)
    {
        return new MAM::Response()
          ->failure('740',
            'Payment Update: Unable to determine payment reserve mode');
    }

    # Obtain event details
    my ($armTime, $fireTime);
    if (defined $updateEvent)
    {
        # SELECT ArmTime,FireTime FROM Event WHERE Id=$id
        my $results = $database->select(
            object     => 'Event',
            selections => [
                new MAM::Selection(name => 'ArmTime'),
                new MAM::Selection(name => 'FireTime'),
            ],
            conditions => [
                new MAM::Condition(
                    name  => 'Id',
                    value => $updateEvent,
                ),
            ],
        );
        if (@{$results->{data}})
        {
            my $row = ${$results->{data}}[0];
            $armTime  = $row->[0];
            $fireTime = $row->[1];
        }
    }

    # Invoke update charges and liens
    my $message      = '';
    my $messageCount = 0;
    my $chargeAmount = 0;
    if ($chargePattern eq 'Periodic' || $reservePattern eq 'Periodic')
    {
        # SELECT Id FROM PaymentComponent WHERE Payment=$id AND Type='Update'
        $results = $database->select(
            object     => 'PaymentComponent',
            selections => [new MAM::Selection(name => 'Id'),],
            conditions => [
                new MAM::Condition(
                    name  => 'Payment',
                    value => $id,
                ),
                new MAM::Condition(
                    name  => 'Type',
                    value => 'Update',
                ),
            ],
        );

        # Iterate over Update components

        # Invoke charges if mode is Periodic
        if ($chargePattern eq 'Periodic')
        {
            my $prevUpdateStartTime = defined $armTime ? $armTime : $now;
            # If fire time is not defined, we are calling this manually so use
            # now. If fire time is defined, if we call update manually before
            # the fire time now will be less than fire time so we will use now,
            # if we call payment update from within the event and the event
            # takes a few seconds or is equal, we want to use fire time, or if
            # we are catching up on old fire time that is past due then
            # firetime will be less than now and we will want to use fire time.
            my $prevUpdateEndTime =
              defined $fireTime
              ? ($now > $fireTime ? $fireTime : $now)
              : $now;
            my $chargeDuration = $prevUpdateEndTime - $prevUpdateStartTime;

            # Only call charge if there are non-zero charges to make
            if (@{$results->{data}} && $chargeDuration)
            {
                foreach my $row (@{$results->{data}})
                {
                    my $usageRecord = $row->[0];

                    my @chargeOptions = (
                        new MAM::Option(
                            name  => 'Id',
                            value => $usageRecord
                        ),
                        new MAM::Option(
                            name  => 'Duration',
                            value => $chargeDuration
                        ),
                    );
                    push @chargeOptions,
                      new MAM::Option(
                        name  => 'Incremental',
                        value => 'True'
                      ) if $reservePattern eq 'Combined';

                    # UsageRecord Charge
                    my $subRequest = new MAM::Request(
                        database => $database,
                        object   => 'UsageRecord',
                        action   => 'Charge',
                        actor    => $SUPER_USER,
                        options  => \@chargeOptions,
                    );
                    my $subResponse =
                      MAM::Bank->charge($subRequest, $requestId);
                    if ($subResponse->getStatus() eq 'Failure')
                    {
                        return new MAM::Response()->failure(
                            $subResponse->getCode(),
                            'Failed in payment update charge: '
                              . $subResponse->getMessage()
                        );
                    }
                    $chargeAmount += $subResponse->getCount();
                    $message .=
                      $messageCount++
                      ? ("\n" . $subResponse->getMessage())
                      : $subResponse->getMessage();
                }

                # At this point we should isolate the charges from potential
                # failures in the lien. If we get a failure in the
                # subsequent liens, we would like to return a failure
                # without rolling back the charges.
                $database->getHandle()->commit();
            }
        }

        # Invoke liens if mode is Periodic
        if ($reservePattern eq 'Periodic')
        {
            # If fire time is not defined, we are calling this manually so use
            # now. If fire time is defined, if we call update manually before
            # the fire time now will be less than fire time so we will use now,
            # if we call payment update from within the event and the event
            # takes a few seconds or is equal, we want to use fire time, or if
            # we are catching up on old fire time that is past due then
            # firetime will be less than now and we will want to use fire time.
            my $nextUpdateStartTime =
              defined $fireTime
              ? ($now > $fireTime ? $fireTime : $now)
              : $now;

            # Use a fast estimate for the rearm period duration
            # This is not perfect and will sometimes make a lien larger
            # than necessary, but it avoids the expense of a call to ParseRecur
            my $periodDuration = 0;
            if (defined $chargePeriod)
            {
                (my $rearmString = $chargePeriod) =~ s/\s//g;
                if ($rearmString =~ /^(\w+)(?:(?:(?:@\w+)?(?:~|\^)?)|!)$/)
                {
                    my $period = $1;
                    while ($period =~ s/^(\d+)([a-zA-Z]+)//)
                    {
                        my ($number, $designator) = ($1, $2);
                        if ($designator =~ /^ye?a?r?s?$/i)
                        {
                            $periodDuration += $number * 366 * 86400;
                        }
                        elsif ($designator =~ /^mon?t?h?s?$/i)
                        {
                            $periodDuration += $number * 31 * 86400;
                        }
                        elsif ($designator =~ /^we?e?k?s?$/i)
                        {
                            $periodDuration += $number * 7 * 86400;
                        }
                        elsif ($designator =~ /^da?y?s?$/i)
                        {
                            $periodDuration += $number * 86400;
                        }
                        elsif ($designator =~ /^ho?u?r?s?$/i)
                        {
                            $periodDuration += $number * 3600;
                        }
                        elsif ($designator =~ /^min?u?t?e?s?$/i)
                        {
                            $periodDuration += $number * 60;
                        }
                        else
                        {
                            return new MAM::Response()->failure('317',
                                "Payment Update: Invalid payment charge period designator ($designator)"
                            );
                        }
                    }
                }
                else
                {
                    return new MAM::Response()->failure('317',
                        "Payment Update: Charge period ($chargePeriod) must be of the form: <period>[[@<instant>][~|^]|!]"
                    );
                }
            }
            else
            {
                return new MAM::Response()->failure('314',
                    'Payment Update: The payment charge period must be specified when using the Periodic reservation model'
                );
            }

            # Lien duration should be the duration to the end time or
            # the duration to the next firing, whichever is shorter.
            my $lienDuration = $periodDuration;
            if ($endTime)
            {
                my $endDuration = $endTime - $nextUpdateStartTime;
                $lienDuration = $endDuration
                  if $endDuration < $periodDuration;
            }

            foreach my $row (@{$results->{data}})
            {
                my $usageRecord = $row->[0];

                my @reserveOptions = (
                    new MAM::Option(name => 'Id', value => $usageRecord),
                    new MAM::Option(
                        name  => 'Duration',
                        value => $lienDuration
                    ),
                    new MAM::Option(name => 'Modify', value => 'True'),
                );

                # UsageRecord Reserve
                my $subRequest = new MAM::Request(
                    database => $database,
                    object   => 'UsageRecord',
                    action   => 'Reserve',
                    actor    => $SUPER_USER,
                    options  => \@reserveOptions,
                );
                my $subResponse = MAM::Bank->reserve($subRequest, $requestId);
                if ($subResponse->getStatus() eq 'Failure')
                {
                    return new MAM::Response()->failure($subResponse->getCode(),
                        'Failed in payment update lien: '
                          . $subResponse->getMessage());
                }
                $message .=
                  $messageCount++
                  ? ("\n" . $subResponse->getMessage())
                  : $subResponse->getMessage();
            }
        }
    }

    # Update the arm time of the update event to the present so duration can
    # be correctly calculated for later charges
    # UPDATE Event SET ArmTime=$now WHERE Id=$updateEvent
    my $subRequest = new MAM::Request(
        database    => $database,
        object      => 'Event',
        action      => 'Modify',
        actor       => $SUPER_USER,
        assignments => [
            new MAM::Assignment(
                name  => 'ArmTime',
                value => $now,
            )
        ],
        conditions => [
            new MAM::Condition(
                name  => 'Id',
                value => $updateEvent,
            )
        ],
    );
    my $subResponse = MAM::Base->modify($subRequest, $requestId);
    if ($subResponse->getStatus() eq 'Failure')
    {
        $message .= "\nWarning -- Unable to arm payment update event: "
          . $subResponse->getMessage();
        $warnings++;
    }

    # Log the transaction
    MAM::Bank->logTransaction(
        database  => $database,
        requestId => $requestId,
        txnId     => $database->nextId('Transaction'),
        object    => 'Payment',
        action    => 'Update',
        actor     => $actor,
        options   => \@options,
        count     => 1,
    );

    # Return response
    if ($warnings)
    {
        $response = new MAM::Response()
          ->warning($chargeAmount, 'Payment Update: ' . $message);
    }
    else
    {
        $response = new MAM::Response()
          ->success($chargeAmount, 'Payment Update: ' . $message);
    }
    return $response;
}

# ----------------------------------------------------------------------------
# $response = usage($request, $requestId);
# ----------------------------------------------------------------------------

# Custom Bank Usage
sub usage
{
    my ($class, $request, $requestId) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (', join(', ', @_[1 .. $#_]), ')');
    }

    my $action  = $request->getAction();
    my $object  = $request->getObject();
    my @objects = $request->getObjects();

    # Begin the usage
    my $usage = "<Request action=\"$action\">\n";

    # Bank actions do not support multiple objects
    if (@objects > 1)
    {
        return new MAM::Response()
          ->failure('313', 'Bank actions do not suppport multiple objects');
    }

    # Handle disabled commands
    if ($action eq 'Create' && $object eq 'Allocation')
    {
        return new MAM::Response(
            status => 'Warning',
            code   => '144',
            message =>
              'This Bank action is disallowed because it bypasses funding'
        );
    }

    # Add the object to the usage
    $usage .= "    <Object>$object</Object>\n";

    # Begin (Payment)
    if ($action eq 'Begin' && $object eq 'Payment')
    {
        $usage .= "    <Option name=\"Id\">{Payment Id}</Option>\n";
        $usage .= "    [<Option name=\"Start\">False</Option>]\n";
    }

    # Charge (UsageRecord)
    elsif ($action eq 'Charge' && $object eq 'UsageRecord')
    {
        $usage .= "    <Data>\n";
        $usage .= "      <UsageRecord>\n";
        $usage .= "        [<Id>{Usage Record Id}</Id>]\n";
        $usage .= "        [<Type>{Usage Record Type}</Type>]\n";
        $usage .= "        [<Instance>{Instance Name}</Instance>]\n";
        $usage .= "        [<Duration>{Integer Number}</Duration>]\n";
        $usage .= "        [<StartTime>YYYY-MM-DD[ hh:mm:ss]</StartTime>]\n";
        $usage .= "        [<EndTime>YYYY-MM-DD[ hh:mm:ss]</EndTime>]\n";
        $usage .= "        [<Quote>{Quote Id}</Quote>]\n";
        $usage .= "        [<Lien>{Lien Id}</Lien>]\n";
        $usage .=
          "        [<{Charge Factor Name}>{Charge Factor Value}</{Charge Factor Name}>]*\n";
        $usage .=
          "        [<{Fund Constraint Name}>{Fund Constraint Value}</{Fund_Constraint Name}>]*\n";
        $usage .=
          "        [<{Usage Property Name}>{Usage Property Value}</{Usage Property Name}>]*\n";
        $usage .= "      </UsageRecord>\n";
        $usage .= "    </Data>\n";
        $usage .= "    |<Option name=\"Id\">{Usage Record Id}</Option>\n";
        $usage .= "    [<Option name=\"Duration\">{Integer Number}</Option>]\n";
        $usage .=
          "    [<Option name=\"StartTime\">YYYY-MM-DD[ hh:mm:ss]</Option>]\n";
        $usage .=
          "    [<Option name=\"EndTime\">YYYY-MM-DD[ hh:mm:ss]</Option>]\n";
        $usage .= "    [<Option name=\"Fund\">{Fund Id}</Option>]\n";
        $usage .=
          "    [<Option name=\"Charge\">{Non-Negative Real Number}</Option>]\n";
        $usage .=
          "    [<Option name=\"ItemizedCharges\">{Itemized Charges}</Option>]\n";
        $usage .=
          "    [<Option name=\"Incremental\">True|False (False)</Option>]\n";
        $usage .=
          "    [<Option name=\"Rate\">{Charge Rate Type}:{Charge Rate Name}[{{Charge Rate Value}}]={Charge Rate Amount}</Option>]*\n";
        $usage .=
          "    [<Option name=\"Itemize\">True|False (False)</Option>|<Option name=\"BalanceFeedback\">True|False (False)</Option>]\n";
        $usage .=
          "    [<Option name=\"AccountingMode\">usage-tracking|notional-charging|fast-allocation|strict-allocation</Option>]\n";
        $usage .=
          "    [<Option name=\"ShowHours\">True|False (False)</Option>]\n";
    }

    # Create (Fund)
    elsif ($action eq 'Create' && $object eq 'Fund')
    {
        $usage .=
          "    [<Set name=\"Name\" [op=\"Assign (Assign)\"]>{Fund Name}</Set>]\n";
        $usage .=
          "    [<Set name=\"Description\" [op=\"Assign (Assign)\"]>{Description}</Set>]\n";
        $usage .=
          "    [<Option name=\"Constraint\">{Constraint Name}=[!]{Constraint Value}</Option>]*\n";
    }

    # Create (Payment)
    elsif ($action eq 'Create' && $object eq 'Payment')
    {
        $usage .= "    [<Data>\n";
        $usage .= "      <Begin*|End*|Update*>\n";
        $usage .= "        [<Id>{Usage Record Id}</Id>]\n";
        $usage .= "        [<Type>{Usage Record Type}</Type>]\n";
        $usage .= "        [<Instance>{Instance Name}</Instance>]\n";
        $usage .= "        [<Duration>{Integer Number}</Duration>]\n";
        $usage .= "        [<Quote>{Quote Id}</Quote>]\n";
        $usage .=
          "        [<{Charge Factor Name}>{Charge Factor Value}</{Charge Factor Name}>]*\n";
        $usage .=
          "        [<{Fund Constraint Name}>{Fund Constraint Value}</{Fund Constraint Name}>]*\n";
        $usage .=
          "        [<{Usage Property Name}>{Usage Property Value}</{Usage Property Name}>]*\n";
        $usage .= "      </Begin|End|Update>\n";
        $usage .= "    </Data>]\n";
        $usage .=
          "    [<Set name=\"BeginEvent\" [op=\"Assign (Assign)\"]>{Integer Number}</Set>]*\n";
        $usage .=
          "    [<Set name=\"BeginTime\" [op=\"Assign (Assign)\"]>YYYY-MM-DD[ hh:mm:ss]</Set>]\n";
        $usage .=
          "    [<Set name=\"ChargePattern\" [op=\"Assign (Assign)\"]>Combined|None|Periodic (Periodic)</Set>]*\n";
        $usage .=
          "    [<Set name=\"ChargePeriod\" [op=\"Assign (Assign)\"]>{Charge Interval}</Set>]\n";
        $usage .=
          "    [<Set name=\"Description\" [op=\"Assign (Assign)\"]>{Description}</Set>]\n";
        $usage .=
          "    [<Set name=\"Duration\" [op=\"Assign (Assign)\"]>{Integer Number}</Set>]\n";
        $usage .=
          "    [<Set name=\"EndEvent\" [op=\"Assign (Assign)\"]>{Integer Number}</Set>]*\n";
        $usage .=
          "    [<Set name=\"EndTime\" [op=\"Assign (Assign)\"]>YYYY-MM-DD[ hh:mm:ss]</Set>]\n";
        $usage .=
          "    [<Set name=\"Notify\" [op=\"Assign (Assign)\"]>{Notification URL}</Set>]\n";
        $usage .=
          "    [<Set name=\"ReservePattern\" [op=\"Assign (Assign)\"]>Combined|None|Periodic (Periodic)</Set>]*\n";
        $usage .=
          "    [<Set name=\"UpdateEvent\" [op=\"Assign (Assign)\"]>{Integer Number}</Set>]*\n";
        $usage .= "    [<Option name=\"Begin\">True|False (False)</Option>]\n";
        $usage .= "    [<Option name=\"Delete\">True|False (False)</Option>]\n";
        $usage .= "    [<Option name=\"Quote\">True|False (False)</Option>]\n";
        $usage .= "    [<Option name=\"QuoteDuration\">{Integer Number}</Option>]\n";
    }

    # Create (Account)
    elsif ($action eq 'Create' && $object eq 'Account')
    {
        $usage .=
          "    <Set name=\"Name\" [op=\"Assign (Assign)\"]>{Account Name}</Set>\n";
        $usage .=
          "    [<Set name=\"Active\" [op=\"Assign (Assign)\"]>True|False (True)</Set>]\n";
        $usage .=
          "    [<Set name=\"Organization\" [op=\"Assign (Assign)\"]>{Organization}</Set>]\n";
        $usage .=
          "    [<Set name=\"Description\" [op=\"Assign (Assign)\"]>{Description}</Set>]\n";
        $usage .= "    [<Option name=\"CreateFund\">True|False</Option>]\n";
    }

    # Deposit (Fund)
    elsif ($action eq 'Deposit' && $object eq 'Fund')
    {
        $usage .= "    [<Option name=\"Id\">{Fund Id}</Option>]\n";
        $usage .=
          "    [<Option name=\"Allocation\">{Allocation Id}</Option>]\n";
        $usage .=
          "    [<Option name=\"Amount\">{Non-Negative Number}</Option>]\n";
        $usage .=
          "    [<Option name=\"CreditLimit\">{Non-Negative Number}</Option>]\n";
        $usage .=
          "    [<Option name=\"StartTime\">YYYY-MM-DD[ hh:mm:ss]</Option>]\n";
        $usage .=
          "    [<Option name=\"EndTime\">YYYY-MM-DD[ hh:mm:ss]</Option>]\n";
        $usage .=
          "    [<Option name=\"Ancestors\">{List of Ancestor Ids}</Option>]\n";
        $usage .=
          "    [<Option name=\"Description\">{Transaction Description}</Option>]\n";
        $usage .=
          "    [<Option name=\"Filter\">{Filter Name}={Filter Value}</Option>]*\n";
        $usage .=
          "    [<Option name=\"FilterType\">ExactMatch|Exclusive|NonExclusive (NonExclusive)</Option>]\n";
        $usage .= "    [<Option name=\"Reset\">True|False (False)</Option>]\n";
        $usage .=
          "    [<Option name=\"EnforceDiscrete\">True|False (True)</Option>]\n";
        $usage .= "    [<Option name=\"CreateFund\">True|False</Option>]\n";
        $usage .=
          "    [<Option name=\"ShowHours\">True|False (False)</Option>]\n";
        $usage .=
          "    [<Option name=\"UseHours\">True|False (False)</Option>]\n";
    }

    # End (Payment)
    elsif ($action eq 'End' && $object eq 'Payment')
    {
        $usage .= "    <Option name=\"Id\">{Payment Id}</Option>\n";
        $usage .= "    [<Option name=\"Stop\">False</Option>]\n";
        $usage .= "    [<Option name=\"Delete\">True</Option>]\n";
    }

    # Query (Account)
    elsif ($action eq 'Query' && $object eq 'Account')
    {
        $usage .=
          "    [<Get name=\"Name\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"Active\" [op=\"Sort|Tros|Count|GroupBy\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"Organization\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"Description\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Where name=\"Name\" [op=\"EQ|NE|GT|GE|LT|LE|Match (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Account Name}</Where>]*\n";
        $usage .=
          "    [<Where name=\"Active\" [op=\"EQ|NE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>True|False</Where>]\n";
        $usage .=
          "    [<Where name=\"Organization\" [op=\"EQ|NE|GT|GE|LT|LE|Match (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Organization Name}</Where>]\n";
        $usage .=
          "    [<Where name=\"Description\" [op=\"EQ|NE|GT|GE|LT|LE|Match (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Description}</Where>]\n";
        $usage .= "    [<Option name=\"Filter\">User={User Name}</Option>]*\n";
        $usage .=
          "    [<Option name=\"Time\">YYYY-MM-DD[ hh:mm:ss]</Option>]\n";
        $usage .= "    [<Option name=\"Unique\">True|False (False)</Option>]\n";
        $usage .= "    [<Option name=\"ChunkSize\">{Integer Number}</Option>\n";
        $usage .= "    [<Option name=\"Limit\">{Integer Number}</Option>\n";
        $usage .= "    [<Option name=\"Offset\">{Integer Number}</Option>]]\n";
    }

    # Query (Allocation)
    elsif ($action eq 'Query' && $object eq 'Allocation')
    {
        $usage .=
          "    [<Get name=\"Id\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"Fund\" [op=\"Sort|Tros|Count|GroupBy|Max|Min|Sum|Average\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"StartTime\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"EndTime\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"Amount\" [op=\"Sort|Tros|Count|GroupBy|Max|Min|Sum|Average\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"CreditLimit\" [op=\"Sort|Tros|Count|GroupBy|Max|Min|Sum|Average\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"InitialDeposit\" [op=\"Sort|Tros|Count|GroupBy|Max|Min|Sum|Average\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"Allocated\" [op=\"Sort|Tros|Count|GroupBy|Max|Min|Sum|Average\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"Active\" [op=\"Sort|Tros|Count|GroupBy\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"Description\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Where name=\"Id\" [op=\"EQ|NE|GT|GE|LT|LE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Integer Number}</Where>]*\n";
        $usage .=
          "    [<Where name=\"Fund\" [op=\"EQ|NE|GT|GE|LT|LE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Integer Number}</Where>]\n";
        $usage .=
          "    [<Where name=\"StartTime\" [op=\"EQ|NE|GT|GE|LT|LE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>YYYY-MM-DD[ hh:mm:ss]|-Infinity|Infinity|Now</Where>]\n";
        $usage .=
          "    [<Where name=\"EndTime\" [op=\"EQ|NE|GT|GE|LT|LE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>YYYY-MM-DD[ hh:mm:ss]|-Infinity|Infinity|Now</Where>]\n";
        $usage .=
          "    [<Where name=\"Amount\" [op=\"EQ|NE|GT|GE|LT|LE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Non-Negative Number}</Where>]\n";
        $usage .=
          "    [<Where name=\"CreditLimit\" [op=\"EQ|NE|GT|GE|LT|LE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Non-Negative Number}</Where>]\n";
        $usage .=
          "    [<Where name=\"InitialDeposit\" [op=\"EQ|NE|GT|GE|LT|LE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Non-Negative Number}</Where>]\n";
        $usage .=
          "    [<Where name=\"Allocated\" [op=\"EQ|NE|GT|GE|LT|LE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Non-Negative Number}</Where>]\n";
        $usage .=
          "    [<Where name=\"Active\" [op=\"EQ|NE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>True|False</Where>]\n";
        $usage .=
          "    [<Where name=\"Description\" [op=\"EQ|NE|GT|GE|LT|LE|Match (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Description}</Where>]\n";
        $usage .=
          "    [<Option name=\"Filter\">{Filter Name}={Filter Value}</Option>]*\n";
        $usage .=
          "    [<Option name=\"FilterType\">ExactMatch|Exclusive|NonExclusive (NonExclusive)</Option>]\n";
        $usage .=
          "    [<Option name=\"IncludeAncestors\">True|False (False)</Option>]\n";
        $usage .=
          "    [<Option name=\"Time\">YYYY-MM-DD[ hh:mm:ss]</Option>]\n";
        $usage .= "    [<Option name=\"Unique\">True|False (False)</Option>]\n";
        $usage .= "    [<Option name=\"ChunkSize\">{Integer Number}</Option>\n";
        $usage .= "    [<Option name=\"Limit\">{Integer Number}</Option>\n";
        $usage .= "    [<Option name=\"Offset\">{Integer Number}</Option>]]\n";
    }

    # Query (Fund)
    elsif ($action eq 'Query' && $object eq 'Fund')
    {
        $usage .=
          "    [<Get name=\"Id\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"Name\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"Description\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Where name=\"Id\" [op=\"EQ|NE|GT|GE|LT|LE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Integer Number}</Where>]*\n";
        $usage .=
          "    [<Where name=\"Name\" [op=\"EQ|NE|GT|GE|LT|LE|Match (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Fund Name}</Where>]\n";
        $usage .=
          "    [<Where name=\"Description\" [op=\"EQ|NE|GT|GE|LT|LE|Match (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Description}</Where>]\n";
        $usage .= "    [<Option name=\"Active\">True|False</Option>]\n";
        $usage .=
          "    [<Option name=\"Filter\">{Filter Name}={Filter Value}</Option>]*\n";
        $usage .=
          "    [<Option name=\"FilterType\">ExactMatch|Exclusive|NonExclusive (NonExclusive)</Option>]\n";
        $usage .=
          "    [<Option name=\"IncludeAncestors\">True|False (False)</Option>]\n";
        $usage .=
          "    [<Option name=\"IncludeDescendants\">True|False (False)</Option>]\n";
        $usage .=
          "    [<Option name=\"Time\">YYYY-MM-DD[ hh:mm:ss]</Option>]\n";
        $usage .= "    [<Option name=\"Unique\">True|False (False)</Option>]\n";
        $usage .= "    [<Option name=\"ChunkSize\">{Integer Number}</Option>\n";
        $usage .= "    [<Option name=\"Limit\">{Integer Number}</Option>\n";
        $usage .= "    [<Option name=\"Offset\">{Integer Number}</Option>]]\n";
    }

    # Query (Lien)
    elsif ($action eq 'Query' && $object eq 'Lien')
    {
        $usage .=
          "    [<Get name=\"Id\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"Instance\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"UsageRecord\" [op=\"Sort|Tros|Count|GroupBy|Max|Min|Sum|Average\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"StartTime\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"EndTime\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"Duration\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"Description\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Where name=\"Id\" [op=\"EQ|NE|GT|GE|LT|LE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Integer Number}</Where>]*\n";
        $usage .=
          "    [<Where name=\"Instance\" [op=\"EQ|NE|GT|GE|LT|LE|Match (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Instance Name}</Where>]\n";
        $usage .=
          "    [<Where name=\"UsageRecord\" [op=\"EQ|NE|GT|GE|LT|LE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Integer Number}</Where>]\n";
        $usage .=
          "    [<Where name=\"StartTime\" [op=\"EQ|NE|GT|GE|LT|LE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>YYYY-MM-DD[ hh:mm:ss]|-Infinity|Infinity|Now</Where>]\n";
        $usage .=
          "    [<Where name=\"EndTime\" [op=\"EQ|NE|GT|GE|LT|LE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>YYYY-MM-DD[ hh:mm:ss]|-Infinity|Infinity|Now</Where>]\n";
        $usage .=
          "    [<Where name=\"Duration\" [op=\"EQ|NE|GT|GE|LT|LE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Integer Number}</Where>]\n";
        $usage .=
          "    [<Where name=\"Description\" [op=\"EQ|NE|GT|GE|LT|LE|Match (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Description}</Where>]\n";
        $usage .=
          "    [<Option name=\"FilterType\">AttributedTo|ImpingesUpon (AttributedTo)</Option>]\n";
        $usage .=
          "    [<Option name=\"Filter\">{Filter Name}={Filter Value}</Option>]*\n";
        $usage .= "    [<Option name=\"Active\">True|False</Option>]\n";
        $usage .=
          "    [<Option name=\"IgnoreAncestors\">True|False (False)</Option>]\n";
        $usage .=
          "    [<Option name=\"Time\">YYYY-MM-DD[ hh:mm:ss]</Option>]\n";
        $usage .= "    [<Option name=\"Unique\">True|False (False)</Option>]\n";
        $usage .= "    [<Option name=\"ChunkSize\">{Integer Number}</Option>\n";
        $usage .= "    [<Option name=\"Limit\">{Integer Number}</Option>\n";
        $usage .= "    [<Option name=\"Offset\">{Integer Number}</Option>]]\n";
    }

    # Query (Quote)
    elsif ($action eq 'Query' && $object eq 'Quote')
    {
        $usage .=
          "    [<Get name=\"Id\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"Amount\" [op=\"Sort|Tros|Count|GroupBy|Max|Min|Sum|Average\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"Pinned\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"Instance\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"UsageRecord\" [op=\"Sort|Tros|Count|GroupBy|Max|Min|Sum|Average\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"StartTime\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"EndTime\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"Duration\" [op=\"Sort|Tros|Count|GroupBy|Max|Min|Sum|Average\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"Description\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Where name=\"Id\" [op=\"EQ|NE|GT|GE|LT|LE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Integer Number}</Where>]*\n";
        $usage .=
          "    [<Where name=\"Amount\" [op=\"EQ|NE|GT|GE|LT|LE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Non-Negative Real Number}</Where>]\n";
        $usage .=
          "    [<Where name=\"Pinned\" [op=\"EQ|NE|GT|GE|LT|LE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>True|False</Where>]\n";
        $usage .=
          "    [<Where name=\"Instance\" [op=\"EQ|NE|GT|GE|LT|LE|Match (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Instance Name}</Where>]\n";
        $usage .=
          "    [<Where name=\"UsageRecord\" [op=\"EQ|NE|GT|GE|LT|LE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Integer Number}</Where>]\n";
        $usage .=
          "    [<Where name=\"StartTime\" [op=\"EQ|NE|GT|GE|LT|LE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>YYYY-MM-DD[ hh:mm:ss]|-Infinity|Infinity|Now</Where>]\n";
        $usage .=
          "    [<Where name=\"EndTime\" [op=\"EQ|NE|GT|GE|LT|LE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>YYYY-MM-DD[ hh:mm:ss]|-Infinity|Infinity|Now</Where>]\n";
        $usage .=
          "    [<Where name=\"Duration\" [op=\"EQ|NE|GT|GE|LT|LE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Integer Number}</Where>]\n";
        $usage .=
          "    [<Where name=\"Description\" [op=\"EQ|NE|GT|GE|LT|LE|Match (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Description}</Where>]\n";
        $usage .=
          "    [<Option name=\"Filter\">{Filter Name}={Filter Value}</Option>]*\n";
        $usage .= "    [<Option name=\"Active\">True|False</Option>]\n";
        $usage .=
          "    [<Option name=\"Time\">YYYY-MM-DD[ hh:mm:ss]</Option>]\n";
        $usage .= "    [<Option name=\"Unique\">True|False (False)</Option>]\n";
        $usage .= "    [<Option name=\"ChunkSize\">{Integer Number}</Option>\n";
        $usage .= "    [<Option name=\"Limit\">{Integer Number}</Option>\n";
        $usage .= "    [<Option name=\"Offset\">{Integer Number}</Option>]]\n";
    }

    # Query (User)
    elsif ($action eq 'Query' && $object eq 'User')
    {
        $usage .=
          "    [<Get name=\"Name\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"Active\" [op=\"Sort|Tros|Count|GroupBy\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"CommonName\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"PhoneNumber\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"EmailAddress\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"DefaultAccount\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Get name=\"Description\" [op=\"Sort|Tros|Count|GroupBy|Max|Min\"]></Get>]\n";
        $usage .=
          "    [<Where name=\"Name\" [op=\"EQ|NE|GT|GE|LT|LE|Match (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Account Name}</Where>]*\n";
        $usage .=
          "    [<Where name=\"Active\" [op=\"EQ|NE (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>True|False</Where>]\n";
        $usage .=
          "    [<Where name=\"CommonName\" [op=\"EQ|NE|GT|GE|LT|LE|Match (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Full Name}</Where>]\n";
        $usage .=
          "    [<Where name=\"PhoneNumber\" [op=\"EQ|NE|GT|GE|LT|LE|Match (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Phone Number}</Where>]\n";
        $usage .=
          "    [<Where name=\"EmailAddress\" [op=\"EQ|NE|GT|GE|LT|LE|Match (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Email Address}</Where>]\n";
        $usage .=
          "    [<Where name=\"DefaultAccount\" [op=\"EQ|NE|GT|GE|LT|LE|Match (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Account Name}</Where>]\n";
        $usage .=
          "    [<Where name=\"Description\" [op=\"EQ|NE|GT|GE|LT|LE|Match (EQ)\"] [conj=\"And|Or (And)\"] [group=\"{Integer Number}\"]>{Description}</Where>]\n";
        $usage .=
          "    [<Option name=\"Filter\">Account={Account Name}</Option>]*\n";
        $usage .=
          "    [<Option name=\"Time\">YYYY-MM-DD[ hh:mm:ss]</Option>]\n";
        $usage .= "    [<Option name=\"Unique\">True|False (False)</Option>]\n";
        $usage .= "    [<Option name=\"ChunkSize\">{Integer Number}</Option>\n";
        $usage .= "    [<Option name=\"Limit\">{Integer Number}</Option>\n";
        $usage .= "    [<Option name=\"Offset\">{Integer Number}</Option>]]\n";
    }

    # Quote (Payment)
    elsif ($action eq 'Quote' && $object eq 'Payment')
    {
        $usage .= "    <Data>\n";
        $usage .= "      <Begin*|End*|Update*>\n";
        $usage .= "        [<Type>{Usage Record Type}</Type>]\n";
        $usage .= "        [<Instance>{Instance Name}</Instance>]\n";
        $usage .= "        [<Duration>{Integer Number}</Duration>]\n";
        $usage .= "        [<Quote>{Quote Id}</Quote>]\n";
        $usage .=
          "        [<{Charge Factor Name}>{Charge Factor Value}</{Charge Factor Name}>]*\n";
        $usage .=
          "        [<{Fund Constraint Name}>{Fund Constraint Value}</{Fund Constraint Name}>]*\n";
        $usage .=
          "        [<{Usage Property Name}>{Usage Property Value}</{Usage Property Name}>]*\n";
        $usage .= "      </Begin|End|Update>\n";
        $usage .= "    </Data>\n";
        $usage .= "    |<Option name=\"Id\">{Payment Id}</Option>\n";
        $usage .= "    [<Option name=\"Duration\">{Integer Number}</Option>]\n";
        $usage .=
          "    [<Option name=\"CostOnly\">True|False (False)</Option>]\n";
        $usage .=
          "    [<Option name=\"Itemize\">True|False (False)</Option>]\n";
        $usage .=
          "    [<Option name=\"ShowHours\">True|False (False)</Option>]\n";
    }

    # Quote (UsageRecord)
    elsif ($action eq 'Quote' && $object eq 'UsageRecord')
    {
        $usage .= "    <Data>\n";
        $usage .= "      <UsageRecord>\n";
        $usage .= "        [<Id>{Usage Record Id}</Id>]\n";
        $usage .= "        [<Type>{Usage Record Type}</Type>]\n";
        $usage .= "        [<Instance>{Instance Name}</Instance>]\n";
        $usage .= "        [<Duration>{Integer Number}</Duration>]\n";
        $usage .= "        [<StartTime>YYYY-MM-DD[ hh:mm:ss]</StartTime>]\n";
        $usage .= "        [<EndTime>YYYY-MM-DD[ hh:mm:ss]</EndTime>]\n";
        $usage .=
          "        [<{Charge Factor Name}>{Charge Factor Value}</{Charge Factor Name}>]*\n";
        $usage .=
          "        [<{Fund Constraint Name}>{Fund Constraint Value}</{Fund Constraint Name}>]*\n";
        $usage .=
          "        [<{Usage Property Name}>{Usage Property Value}</{Usage Property Name}>]*\n";
        $usage .= "      </UsageRecord>\n";
        $usage .= "    </Data>\n";
        $usage .= "    [<Option name=\"Id\">{Usage Record Id}</Option>]\n";
        $usage .=
          "    [<Option name=\"StartTime\">YYYY-MM-DD[ hh:mm:ss]</Option>]\n";
        $usage .=
          "    [<Option name=\"EndTime\">YYYY-MM-DD[ hh:mm:ss]</Option>]\n";
        $usage .= "    [<Option name=\"Duration\">{Integer Number}</Option>]\n";
        $usage .=
          "    [<Option name=\"Description\">{Quote Description}</Option>]\n";
        $usage .=
          "    [<Option name=\"GraceDuration\">{Integer Number}</Option>]\n";
        $usage .= "    [<Option name=\"Charge\">{Decimal Number}</Option>]\n";
        $usage .=
          "    [<Option name=\"ItemizedCharges\">{Itemized Charges}</Option>]\n";
        $usage .=
          "    [<Option name=\"CostOnly\">True|False (False)</Option>|<Option name=\"Guarantee\">True|False (False)</Option>]\n";
        $usage .= "    [<Option name=\"Quote\">{Quote Template Id}</Option>]\n";
        $usage .=
          "    [<Option name=\"Rate\">{Charge Rate Name}[{{Charge Rate Value}}]={Charge Rate Amount}</Option>]*\n";
        $usage .=
          "    [<Option name=\"Itemize\">True|False (False)</Option>|<Option name=\"BalanceFeedback\">True|False (False)</Option>]\n";
        $usage .=
          "    [<Option name=\"AccountingMode\">usage-tracking|notional-charging|fast-allocation|strict-allocation</Option>]\n";
        $usage .=
          "    [<Option name=\"ShowHours\">True|False (False)</Option>]\n";
    }

    # Refund (UsageRecord)
    elsif ($action eq 'Refund' && $object eq 'UsageRecord')
    {
        $usage .=
          "    <Option name=\"Id\">{Usage Record Id}</Option>|<Option name=\"Instance\">{Instance Name}</Option>\n";
        $usage .=
          "    [<Option name=\"Allocation\">{Allocation Id}</Option>]\n";
        $usage .= "    [<Option name=\"Amount\">{Refund Amount}</Option>]\n";
        $usage .=
          "    [<Option name=\"Description\">{Transaction Description}</Option>]\n";
        $usage .=
          "    [<Option name=\"AccountingMode\">usage-tracking|notional-charging|fast-allocation|strict-allocation</Option>]\n";
        $usage .=
          "    [<Option name=\"ShowHours\">True|False (False)</Option>]\n";
    }

    # Reserve (UsageRecord)
    elsif ($action eq 'Reserve' && $object eq 'UsageRecord')
    {
        $usage .= "    <Data>\n";
        $usage .= "      <UsageRecord>\n";
        $usage .= "        [<Id>{Usage Record Id}</Id>]\n";
        $usage .= "        [<Type>{Usage Record Type}</Type>]\n";
        $usage .= "        [<Instance>{Instance Name}</Instance>]\n";
        $usage .= "        [<Duration>{Integer Number}</Duration>]\n";
        $usage .= "        [<Quote>{Quote Id}</Quote>]\n";
        $usage .= "        [<StartTime>YYYY-MM-DD[ hh:mm:ss]</StartTime>]\n";
        $usage .= "        [<EndTime>YYYY-MM-DD[ hh:mm:ss]</EndTime>]\n";
        $usage .=
          "        [<{Charge Factor Name}>{Charge Factor Value}</{Charge Factor Name}>]*\n";
        $usage .=
          "        [<{Fund Constraint Name}>{Fund Constraint Value}</{Fund Constraint Name}>]*\n";
        $usage .=
          "        [<{Usage Property Name}>{Usage Property Value}</{Usage Property Name}>]*\n";
        $usage .= "      </UsageRecord>\n";
        $usage .= "    </Data>\n";
        $usage .= "    |<Option name=\"Id\">{Usage Record Id}</Option>\n";
        $usage .=
          "    [<Option name=\"StartTime\">YYYY-MM-DD[ hh:mm:ss]</Option>]\n";
        $usage .=
          "    [<Option name=\"EndTime\">YYYY-MM-DD[ hh:mm:ss]</Option>]\n";
        $usage .= "    [<Option name=\"Duration\">{Integer Number}</Option>]\n";
        $usage .=
          "    [<Option name=\"Description\">{Lien Description}</Option>]\n";
        $usage .=
          "    [<Option name=\"GraceDuration\">{Integer Number}</Option>]\n";
        $usage .=
          "    [<Option name=\"Charge\">{Non-Negative Number}</Option>]\n";
        $usage .=
          "    [<Option name=\"ItemizedCharges\">{Itemized Charges}</Option>]\n";
        $usage .=
          "    [<Option name=\"Replace\">True|False (False)</Option>|<Option name=\"Modify\">True|False (False)</Option>]\n";
        $usage .=
          "    [<Option name=\"Rate\">{Charge Rate Name}[{{Charge Rate Value}}]={Charge Rate Amount}</Option>]*\n";
        $usage .=
          "    [<Option name=\"Itemize\">True|False (False)</Option>|<Option name=\"BalanceFeedback\">True|False (False)</Option>]\n";
        $usage .=
          "    [<Option name=\"AccountingMode\">usage-tracking|notional-charging|fast-allocation|strict-allocation</Option>]\n";
        $usage .=
          "    [<Option name=\"ShowHours\">True|False (False)</Option>]\n";
    }

    # Reset (Fund)
    elsif ($action eq 'Reset' && $object eq 'Fund')
    {
        $usage .= "    [<Option name=\"Id\">{Fund Id}</Option>]\n";
        $usage .=
          "    [<Option name=\"Filter\">{Filter Name}={Filter Value}</Option>]*\n";
        $usage .=
          "    [<Option name=\"FilterType\">ExactMatch|Exclusive|NonExclusive (NonExclusive)</Option>]\n";
        $usage .=
          "    [<Option name=\"ShowHours\">True|False (False)</Option>]\n";
    }

    # Reset (Allocation)
    elsif ($action eq 'Reset' && $object eq 'Allocation')
    {
        $usage .= "    [<Option name=\"Id\">{Allocation Id}</Option>]\n";
        $usage .=
          "    [<Option name=\"ShowHours\">True|False (False)</Option>]\n";
    }

    # Start (Payment)
    elsif ($action eq 'Start' && $object eq 'Payment')
    {
        $usage .= "    <Option name=\"Id\">{Payment Id}</Option>\n";
    }

    # Stop (Fund)
    elsif ($action eq 'Stop' && $object eq 'Fund')
    {
        $usage .= "    [<Option name=\"Id\">{Fund Id}</Option>]\n";
        $usage .=
          "    [<Option name=\"Filter\">{Filter Name}={Filter Value}</Option>]*\n";
        $usage .=
          "    [<Option name=\"FilterType\">ExactMatch|Exclusive|NonExclusive (NonExclusive)</Option>]\n";
    }

    # Stop (Payment)
    elsif ($action eq 'Stop' && $object eq 'Payment')
    {
        $usage .= "    <Option name=\"Id\">{Payment Id}</Option>\n";
        $usage .=
          "    [<Option name=\"FilterType\">ExactMatch|Exclusive|NonExclusive (NonExclusive)</Option>]\n";
    }

    # Transfer (Fund)
    elsif ($action eq 'Transfer' && $object eq 'Fund')
    {
        $usage .= "    { <Option name=\"FromId\">{Source Fund Id}</Option>\n";
        $usage .=
          "    | <Option name=\"FromAllocation\">{Source Allocation Id}</Option> }\n";
        $usage .=
          "    { <Option name=\"ToId\">{Destination Fund Id}</Option>\n";
        $usage .=
          "    | <Option name=\"ToAllocation\">{Destination Allocation Id}</Option> }\n";
        $usage .=
          "    <Option name=\"Amount\">{Non-Negative Real Number}</Option>\n";
        $usage .=
          "    [<Option name=\"Description\">{Transaction Description}</Option>]\n";
        $usage .=
          "    [<Option name=\"EnforceDiscrete\">True|False (True)</Option>]\n";
        $usage .=
          "    [<Option name=\"ShowHours\">True|False (False)</Option>]\n";
        $usage .=
          "    [<Option name=\"UseHours\">True|False (False)</Option>]\n";
    }

    # Update (Payment)
    elsif ($action eq 'Update' && $object eq 'Payment')
    {
        $usage .= "    <Option name=\"Id\">{Payment Id}</Option>\n";
    }

    # Withdraw (Fund)
    elsif ($action eq 'Withdraw' && $object eq 'Fund')
    {
        $usage .=
          "    <Option name=\"Amount\">{Non-Negative Real Number}</Option>\n";
        $usage .= "    [<Option name=\"Id\">{Fund Id}</Option>]\n";
        $usage .= "    [<Option name=\"Allocation\">{Allocation Id}</Option>]\n";
        $usage .=
          "    [<Option name=\"Description\">{Transaction Description}</Option>]\n";
        $usage .=
          "    [<Option name=\"Filter\">{Filter Name}={Filter Value}</Option>]*\n";
        $usage .=
          "    [<Option name=\"FilterType\">ExactMatch|Exclusive|NonExclusive (NonExclusive)</Option>]\n";
        $usage .=
          "    [<Option name=\"IgnoreLiens\">True|False (False)</Option>]\n";
        $usage .=
          "    [<Option name=\"ShowHours\">True|False (False)</Option>]\n";
        $usage .=
          "    [<Option name=\"UseHours\">True|False (False)</Option>]\n";
    }

    # Fall through to Base actions
    else
    {
        return new MAM::Response()
          ->failure('313', "Unsupported bank usage: ($action)");
    }

    # Terminate the usage
    $usage .= "    [<Option name=\"ShowHidden\">True|False (False)</Option>]\n";
    $usage .= "    [<Option name=\"ShowUsage\">True|False (False)</Option>]\n";
    $usage .= '</Request>';

    return new MAM::Response(
        status  => 'Success',
        code    => '010',
        message => $usage
    );
}

# ----------------------------------------------------------------------------
# logTransaction(database => $database,
#                   object => $object,
#                   action => $action,
#                   actor => $actor,
#                   assignments => \@assignments,
#                   conditions => \@conditions,
#                   options => \@options,
#                   data => \@data,
#                   count => $count,
#                   fund => $fund,
#                   allocation => $allocation,
#                   delta => $delta,
#                   balance => $balance,
#                   remaining => $remaining,
#                   requestId => $requestId,
#                   txnId => $txnId
#                   );
# ----------------------------------------------------------------------------

# Log Transaction
sub logTransaction
{
    my ($class, %arg) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (',
            join(', ', map {"$_ => $arg{$_}"} keys %arg), ')');
    }

    # Declare and Initialize variables
    my $database    = $arg{database};
    my $dbh         = $database->getHandle();
    my $object      = $arg{object};
    my $action      = $arg{action} || 'Create';
    my $actor       = $arg{actor};
    my @assignments = $arg{assignments} ? @{$arg{assignments}} : ();
    my @conditions  = $arg{conditions} ? @{$arg{conditions}} : ();
    my @options     = $arg{options} ? @{$arg{options}} : ();
    my @data        = $arg{data} ? @{$arg{data}} : ();
    my $count       = $arg{count};
    my $amount      = $arg{amount};
    my $delta       = $arg{delta};
    my $balance     = $arg{balance};
    my $remaining   = $arg{remaining};
    my $fund        = $arg{fund};
    my $allocation  = $arg{allocation};
    my $requestId   = $arg{requestId};
    my $txnId       = $arg{txnId};
    my %child       = ();
    my %instance    = ();
    my %key         = ();
    my %details     = ();
    my $description = '';
    my $now         = $requestTime;
    my $firstTime;
    my $duration;
    my $usageRecord;

    # Identify discriminators by finding the simple difference
    # between the regular Transaction attributes and the actual list
    my (%discriminators, %nonDiscriminators);
    @nonDiscriminators{
        'Id',           'Object',           'Action',
        'Actor',        'Name',             'Child',
        'Count',        'Details',          'Description',
        'CreationTime', 'ModificationTime', 'Deleted',
        'RequestId',    'TransactionId',    'Instance',
        'Amount',       'Delta',            'Fund',
        'Allocation',   'UsageRecord',      'Duration',
      }
      = ();
    foreach my $item (MAM::Cache->listAttributes('Transaction'))
    {
        %{$discriminators{$item}} = ()
          unless exists $nonDiscriminators{$item};
    }

    # Iterate over assignments
    my $association = MAM::Cache->getObjectProperty($object, 'Association');
    my $parent      = MAM::Cache->getObjectProperty($object, 'Parent');
    foreach my $assignment (@assignments)
    {
        my $name  = $assignment->getName();
        my $value = $assignment->getValue();
        $value =~ s/'/''/g;    # Escape quotes in values
        my $op = $assignment->getOperator();

        # Append key or child
        if ($name eq 'Name' || $name eq 'Id')
        {
            if ($association eq 'True')
            {
                $child{$value}++;
            }
            else
            {
                $key{$value}++;
            }
        }

        # Append parent to key if association
        elsif ($association eq 'True' && $name eq $parent)
        {
            $key{$value}++;
        }

        # Append instance
        elsif ($name eq 'Instance')
        {
            $instance{$value}++;
        }

        # Append discriminators
        elsif (exists $discriminators{$name})
        {
            ${$discriminators{$name}}{$value}++;
        }

        # Set amount
        elsif ($name eq 'Amount' && ! defined $amount)
        {
            $amount = $value;
        }

        # Set usage record
        elsif ($name eq 'UsageRecord')
        {
            $usageRecord ||= $value;
        }

        # Set duration
        elsif ($name eq 'Duration')
        {
            $duration = $value;
        }

        # Append as detail
        else
        {
            my $detail = $name;
            if (! $op)
            {
                $detail .= '=';    # Use = for assignments
            }
            else
            {
                if    ($op eq 'Assign') { $detail .= '='; }
                elsif ($op eq 'Inc')    { $detail .= '+='; }
                elsif ($op eq 'Dec')    { $detail .= '-='; }
                else                    { $detail .= '{' + $op + '}'; }
            }
            $detail .= $value;
            $details{$detail}++;
        }
    }

    # Iterate over conditions
    foreach my $condition (@conditions)
    {
        my $name  = $condition->getName();
        my $value = $condition->getValue();
        my $op    = $condition->getOperator();

        # Append key or child
        if ($name eq 'Name' || $name eq 'Id')
        {
            if ($association eq 'True')
            {
                $child{$value}++;
            }
            else
            {
                $key{$value}++;
            }
        }

        # Append parent to key if association
        elsif ($association eq 'True' && $name eq $parent)
        {
            $key{$value}++;
        }

        # Append instance
        elsif ($name eq 'Instance')
        {
            $instance{$value}++;
        }

        # Set usage record
        elsif ($name eq 'UsageRecord')
        {
            $usageRecord ||= $value;
        }

        # Append discriminators
        elsif (exists $discriminators{$name})
        {
            ${$discriminators{$name}}{$value}++;
        }

        # Append as detail
        else
        {
            my $detail = $name;
            if (! $op)
            {
                $detail .= '==';
            }
            else
            {
                if    ($op eq 'EQ')    { $detail .= '=='; }
                elsif ($op eq 'GT')    { $detail .= '>'; }
                elsif ($op eq 'GE')    { $detail .= '>='; }
                elsif ($op eq 'LT')    { $detail .= '<'; }
                elsif ($op eq 'LE')    { $detail .= '<='; }
                elsif ($op eq 'NE')    { $detail .= '!='; }
                elsif ($op eq 'Match') { $detail .= '~='; }
                else                   { $detail .= '{' + $op + '}'; }
            }
            $detail .= $value;
            $details{$detail}++;
        }
    }

    # Iterate over options
    foreach my $option (@options)
    {
        my $name  = $option->getName();
        my $value = $option->getValue();
        my $op    = $option->getOperator();

        # Append key or child
        if ($name eq 'Name' || $name eq 'Id')
        {
            if ($association eq 'True')
            {
                $child{$value}++;
            }
            else
            {
                $key{$value}++;
            }
        }

        # Append parent to key if association
        elsif ($association eq 'True' && $name eq $parent)
        {
            $key{$value}++;
        }

        # Append instance
        elsif ($name eq 'Instance')
        {
            $instance{$value}++;
        }

        # Append discriminators
        elsif (exists $discriminators{$name})
        {
            ${$discriminators{$name}}{$value}++;
        }

        # Set amount
        elsif ($name eq 'Amount' && ! defined $amount)
        {
            $amount = $value;
        }

        # Set usage record
        elsif ($name eq 'UsageRecord')
        {
            $usageRecord ||= $value;
        }

        # Set duration
        elsif ($name eq 'Duration')
        {
            $duration = $value;
        }

        # Append description
        elsif ($name eq 'Description')
        {
            $description = $value;
        }

        # Explicity set child if specified
        elsif ($name eq 'Child')
        {
            $child{$value}++;
        }

        # Append as detail
        else
        {
            my $detail = $name;
            if (! $op)
            {
                $detail .= ':=';
            }
            else
            {
                if   ($op eq 'Not') { $detail .= ':!'; }
                else                { $detail .= '{' + $op + '}'; }
            }
            $detail .= $value;
            $details{$detail}++;
        }
    }

    # Iterate over data
    foreach my $datum (@data)
    {
        my $name = $datum->getName();
        if ($name eq 'UsageRecord')
        {
            # Append instance
            if ($datum->getPropertyValue('Instance'))
            {
                $instance{$datum->getPropertyValue('Instance')}++;
            }

            # Append discriminators
            foreach my $key (sort keys %discriminators)
            {
                if ($datum->getPropertyValue($key))
                {
                    ${$discriminators{$key}}{$datum->getPropertyValue($key)}++;
                }
            }

            # Set charge
            if ($datum->getPropertyValue('Charge') && ! defined $amount)
            {
                $amount = $datum->getPropertyValue('Charge');
            }

            # Do not clobber duration option
            elsif ($datum->getPropertyValue('Duration')
                && ! defined $duration)
            {
                $duration = $datum->getPropertyValue('Duration');
            }

            # Append usage record details
            $details{$datum->toString()}++;
        }
    }

    # Add fields to sql names and values
    my $names  = 'g_object,g_action,g_actor';
    my $values = "'$object','$action','$actor'";

    # Log key
    if (keys %key)
    {
        $names .= ',g_key';
        $values .= ",'" . join(',', sort keys %key) . "'";
    }

    # Log child
    if (keys %child)
    {
        $names .= ',g_child';
        $values .= ",'" . join(',', sort keys %child) . "'";
    }

    # Log instance
    if (keys %instance)
    {
        $names .= ',g_instance';
        $values .= ",'" . join(',', sort keys %instance) . "'";
    }

    # Log discriminators
    foreach my $key (sort keys %discriminators)
    {
        if (keys %{$discriminators{$key}})
        {
            my $name = toDBC($key);
            my $value = join ',', sort keys %{$discriminators{$key}};
            $names  .= ",$name";
            $values .= ",'$value'";
        }
    }

    # Log count
    if (defined $count)
    {
        $names  .= ',g_count';
        $values .= ",$count";
    }

    # Log amount
    if (defined $amount)
    {
        $names .= ',g_amount';
        if ($amount =~ /inf/i)
        {
            $values .= ",'$amount'";
        }
        else
        {
            $values .= ",$amount";
        }
    }

    # Log fund
    if (defined $fund)
    {
        $names  .= ',g_fund';
        $values .= ",$fund";
    }

    # Log allocation
    if (defined $allocation)
    {
        $names  .= ',g_allocation';
        $values .= ",$allocation";
    }

    # Log delta
    if (defined $delta)
    {
        $names .= ',g_delta';
        if ($delta =~ /inf/i)
        {
            $values .= ",'$delta'";
        }
        else
        {
            $values .= ",$delta";
        }
    }

    # Log Balance
    if (defined $balance)
    {
        $names .= ',g_balance';
        if ($balance =~ /inf/i)
        {
            $values .= ",'$balance'";
        }
        else
        {
            $values .= ",$balance";
        }
    }

    # Log Remaining
    if (defined $remaining)
    {
        $names .= ',g_remaining';
        if ($remaining =~ /inf/i)
        {
            $values .= ",'$remaining'";
        }
        else
        {
            $values .= ",$remaining";
        }
    }

    # Log usage record
    if (defined $usageRecord)
    {
        $names  .= ',g_usage_record';
        $values .= ",$usageRecord";
    }

    # Log duration
    if (defined $duration)
    {
        $names  .= ',g_duration';
        $values .= ",$duration";
    }

    # Log description
    if ($description ne '')
    {
        $names  .= ',g_description';
        $values .= ",'$description'";
    }

    # Log details
    if (keys %details)
    {
        my $details = join ',', sort keys %details;
        # Truncate details value if necessary
        if (length $details >= 1024)
        {
            $details = substr($details, 0, 1024);
        }
        $names  .= ',g_details';
        $values .= ",'$details'";
    }

    # Log creationdate, modificationdate, requestId, txnId and id
    $names .=
      ',g_creation_time,g_modification_time,g_request_id,g_transaction_id,g_id';
    $values .= ",$now,$now,$requestId,$txnId,$txnId";

    # Add column list
    my $sql = "INSERT INTO g_transaction ($names) VALUES ($values)";

    # Perform SQL Update
    if ($log->is_debug())
    {
        $log->debug("SQL Update: $sql");
    }
    $dbh->do($sql);
}

# ----------------------------------------------------------------------------
# ($charge, $itemizedCharges, $message, \@data) = calculateCharge(
#                   request => $request,
#                   requestId => $requestId,
#                   usageRecordProperties => \%usageRecordProperties,
#                   quote => $quoteId,
#                   quote_template => $quoteTemplateId,
#                   duration => $duration,
#                   charge_ref => \$charge,
#                   itemizedCharges_ref => \$itemizedCharges,
#                   data_ref => \@data,
#                   metadata_ref => \%metadata,
#                   message_ref => \$chargeMessage,
#                   );
# ----------------------------------------------------------------------------

# Calculate Charge
sub calculateCharge
{
    my ($class, %arg) = @_;

    if ($log->is_debug())
    {
        $log->debug('Invoked with arguments: (',
            join(', ', map { "$_ => " . ($arg{$_} || 'undef') } keys %arg),
            ')');
    }

    # Declare and Initialize variables
    my $request                   = $arg{request};
    my $requestId                 = $arg{requestId};
    my $usageRecordProperties_ref = $arg{usageRecordProperties};
    my $quoteId                   = $arg{quote} || '';
    my $quoteTemplateId           = $arg{quote_template} || '';
    my $duration                  = $arg{duration} || 0;
    my $charge_ref                = $arg{charge_ref};
    my $itemizedCharges_ref       = $arg{itemizedCharges_ref};
    my $data_ref                  = $arg{data_ref};
    my $metadata_ref              = $arg{metadata_ref};
    my $message_ref               = $arg{message_ref};
    my $attributes_ref            = $arg{attributes_ref};

    my $action   = $request->getAction();
    my $database = $request->getDatabase();
    my %usageRecordProperties =
      defined $usageRecordProperties_ref
      ? %$usageRecordProperties_ref
      : ();
    my $charge =
      defined $charge_ref
      ? $$charge_ref
      : $request->getOptionValue('Charge');
    my $itemizedCharges =
      defined $itemizedCharges_ref
      ? $$itemizedCharges_ref
      : $request->getOptionValue('ItemizedCharges') || '';
    my $message =
      defined $message_ref
      ? $$message_ref
      : '';
    my @data =
      defined $data_ref
      ? @$data_ref
      : ();
    my %metadata =
      defined $metadata_ref
      ? %$metadata_ref
      : ();
    my @options = $request->getOptions();

    my $guarantee;
    my $guaranteeOption = $request->getOptionValue('Guarantee');
    if (defined $guaranteeOption && $guaranteeOption =~ /^True$/i)
    {
        $guarantee = 1;
    }
    else
    {
        $guarantee = 0;
    }

    my $itemize;
    my $itemizeOption = $request->getOptionValue('Itemize');
    if (defined $itemizeOption && $itemizeOption =~ /^true$/i)
    {
        $itemize = 1;
    }
    else
    {
        $itemize = 0;
    }
    my $chargeItemization;
    if (($serverConfig->{'charge.itemization'} || 'false') =~ /^true$/i)
    {
        $chargeItemization = 1;
    }
    else
    {
        $chargeItemization = 0;
    }
    my $currency_precision = $serverConfig->{'currency.precision'} || 0;

    # Calculate charge if not already specified
    unless (defined $charge)
    {
        $charge = 0;
        my %chargeRate          = ();
        my %specifiedChargeRate = ();

        # Look up the current charge rates
        # SELECT Name,Value,Amount FROM ChargeRate
        my $results = $database->select(
            object     => 'ChargeRate',
            selections => [
                new MAM::Selection(name => 'Name'),
                new MAM::Selection(name => 'Value'),
                new MAM::Selection(name => 'Amount')
            ]
        );
        foreach my $row (@{$results->{data}})
        {
            my $name   = $row->[0];
            my $value  = $row->[1] || '';
            my $amount = $row->[2];

            $chargeRate{$name}{$value} = $amount;
        }

        # Note: For Reserve and Charge actions, the quote argument may be
        # either a quote or a quote template, whereas for Quote actions, the
        # quote argument will be the new quote and the quote_template
        # argument will be the quote template.

        # Override with quoted charge rates
        if (   (($action eq 'Reserve' || $action eq 'Charge') && $quoteId ne '')
            || ($action eq 'Quote' && $quoteTemplateId ne ''))
        {
            my $operativeQuoteId =
              $action eq 'Quote' ? $quoteTemplateId : $quoteId;
   # Look up the quoted charge rates
   # SELECT Name,Value,Amount FROM QuoteChargeRate WHERE Quote=$operativeQuoteId
            my $results = $database->select(
                object     => 'QuoteChargeRate',
                selections => [
                    new MAM::Selection(name => 'Name'),
                    new MAM::Selection(name => 'Value'),
                    new MAM::Selection(name => 'Amount')
                ],
                conditions => [
                    new MAM::Condition(
                        name  => 'Quote',
                        value => $operativeQuoteId
                    )
                ]
            );
            foreach my $row (@{$results->{data}})
            {
                my $name   = $row->[0];
                my $value  = $row->[1] || '';
                my $amount = $row->[2];

                if ($action eq 'Quote')
                {
                    $chargeRate{$name}{$value}          = $amount;
                    $specifiedChargeRate{$name}{$value} = $amount;
                }
                else { $chargeRate{$name}{$value} = $amount; }
            }
        }

        # Override with specified charge rates
        foreach my $option (@options)
        {
            if ($option->getName() eq 'Rate')
            {
                my $rateOptionValue = $option->getValue();

                unless ($rateOptionValue =~
                    /^(\w+)(?:{(\w+|\d+(?:-\d+)?(?:,\d+(?:-\d+)?)*)})?=(\S+)$/)
                {
                    $message .=
                      "\nWarning -- Invalid charge rate string ($rateOptionValue)";
                    next;
                }
                my ($name, $value, $amount) = ($1, $2, $3);
                $value = '' unless defined $value;

                $chargeRate{$name}{$value}          = $amount;
                $specifiedChargeRate{$name}{$value} = $amount;
            }
        }

        # Initialize charge tallies
        my $scalingFactor      = 1;
        my @scalingExpressions = ();
        my @chargeExpressions  = ();

        # First iterate over the charge multipliers with multiplier rates so
        # the scaling factors are populated for the additive rates
      NAME: foreach my $name (keys %chargeRate)
        {
            my %values          = %{$chargeRate{$name}};
            my %quoteChargeRate = ();

            # Look for this rate in the usage record properties
            if (defined $usageRecordProperties{$name})
            {
                my $propertyValue = $usageRecordProperties{$name};
                $propertyValue =~ s/^{(.*)}$/$1/;
                my $isMultiplier      = 0;
                my $chargeRateApplied = 0;

              PART: foreach my $propertyValuePart (split /,/, $propertyValue)
                {
                    my $effectivePropertyValuePart = $propertyValuePart;

                    # Iterate through charge rate values
                    # Default (null) values should sort last
                  VALUE: foreach my $value (
                        sort { ($a || '~') cmp($b || '~') }
                        keys %values
                      )
                    {
                        my $rateExpression  = $values{$value};
                        my $valueExpression = $value;

                        next NAME unless $rateExpression =~ /^\*([\d\.]+)$/;
                        my $amount = $1;
                        $isMultiplier = 1;

                        # Parse out and handle conditions
                        if ($valueExpression =~ s/^(.*)\?//)
                        {
                            my $valueConditions = $1;
                            my ($conditionsMatched, $conditionsMissed) = (0, 0);
                            foreach my $valueCondition (split /[\|\&]/,
                                $valueConditions)
                            {
                                if ($valueCondition =~ /^(\w+)=(\S+)$/)
                                {
                                    my ($conditionName, $conditionValue) =
                                      ($1, $2);
                                    if (defined
                                        $usageRecordProperties{$conditionName}
                                        && $usageRecordProperties{
                                            $conditionName} eq $conditionValue)
                                    {
                                        $conditionsMatched++;
                                    }
                                    else
                                    {
                                        $conditionsMissed++;
                                    }
                                }
                                else
                                {
                                    $message .=
                                      "\nWarning -- Invalid value condition ($valueCondition) for charge rate $name";
                                    next VALUE;
                                }
                            }
                            if ($valueConditions =~ /\|/)
                            {
                                next unless $conditionsMatched;
                            }
                            else
                            {
                                next if $conditionsMissed;
                            }
                        }

                        # Continue unless value expression matches

                        # Numeric-valued charge rates
                        if (   $valueExpression !~ /^[[:alpha:]]/
                            && $propertyValuePart =~ /^\d*\.?\d+$/)
                        {
                            my $matches = 0;
                            foreach my $valueRange (
                                length $valueExpression
                                ? (split /,/, $valueExpression, -1)
                                : ('')
                              )
                            {
                                if ($valueRange =~
                                    /^([\d\.]+)?(?:(-|>=?|=?[<>]=?)([\d\.]+))?$/
                                  )
                                {
                                    my ($leftBound, $comparison, $rightBound) =
                                      ($1, $2, $3);
                                    my $expression;
                                    # a-b or a[=]<[=]b
                                    if (   defined $leftBound
                                        && defined $rightBound
                                        && defined $comparison
                                        && $comparison !~ />/)
                                    {
                                        my $leftOperator =
                                          (      $comparison eq '-'
                                              || $comparison =~ /^=</)
                                          ? '>='
                                          : '>';
                                        my $rightOperator =
                                          (      $comparison eq '-'
                                              || $comparison =~ /<=$/)
                                          ? '<='
                                          : '<';
                                        $expression =
                                          "$propertyValuePart $leftOperator $leftBound && $propertyValuePart $rightOperator $rightBound";
                                    }
                                    # (>|<)[=]b
                                    elsif (defined $rightBound
                                        && defined $comparison
                                        && $comparison !~ /^=/
                                        && ! defined $leftBound)
                                    {
                                        $expression =
                                          "$propertyValuePart $comparison $rightBound";
                                    }
                                    # a
                                    elsif (defined $leftBound
                                        && ! defined $rightBound
                                        && ! defined $comparison)
                                    {
                                        $expression =
                                          "$propertyValuePart == $leftBound";
                                    }
                                    # null
                                    elsif (! defined $leftBound
                                        && ! defined $rightBound
                                        && ! defined $comparison)
                                    {
                                        $matches = 1;
                                        last;
                                    }
                                    # Invalid
                                    else
                                    {
                                        $message .=
                                          "\nWarning -- Invalid value expression ($valueRange) for standard charge rate $name";
                                        next;
                                    }
                                    # Test the expression
                                    if (eval $expression)
                                    {
                                        $matches = 1;
                                        last;
                                    }
                                }
                                else
                                {
                                    $message .=
                                      "\nWarning -- Invalid value expression ($valueRange) for standard charge rate $name";
                                    next;
                                }
                            }
                            next unless $matches;
                        }

                        # Complex charge rates
                        elsif ($valueExpression =~ /^[[:alpha:]]/
                            && $propertyValuePart =~ /:/)
                        {
                            my $matches = 0;
                            $propertyValuePart =~ s/\"//g;

                            # name-valued part charge rate
                            if ($propertyValuePart eq $valueExpression)
                            {
                                $matches = 1;
                            }
                            # numeric-valued part charge rate
                            elsif ($propertyValuePart =~ /([^:]+):(.*)/)
                            {
                                my ($partName, $partValue) = ($1, $2);
                                if ($partName eq $valueExpression)
                                {
                                    if ($partValue =~ /^\d*\.?\d+$/)
                                    {
                                        $matches = 1;
                                        $effectivePropertyValuePart =
                                          $partValue;
                                    }
                                    else
                                    {
                                        $message .=
                                          "\nWarning -- Non-numeric part value ($partValue) for $partName in complex property $name";
                                        next;
                                    }
                                }
                            }
                            next unless $matches;
                        }

                        # Name-valued charge rates
                        elsif ($valueExpression =~ /^[[:alpha:]]|^$/
                            && $propertyValuePart =~ /^[[:alpha:]][\w\.\-]*$/)
                        {
                            my $matches = 0;
                            foreach my $altValue (
                                length $valueExpression
                                ? (split /,/, $valueExpression, -1)
                                : ('')
                              )
                            {
                                if (   $altValue eq ''
                                    || $altValue eq $propertyValuePart)
                                {
                                    $matches = 1;
                                    last;
                                }
                            }
                            next unless $matches;
                        }

                        # Unable to derive if name based or value based
                        else
                        {
                            $message .=
                              "\nWarning -- Unable to categorize the testing of $name $propertyValuePart against $valueExpression as name-valued or numeric-valued";
                            next;
                        }

                        # Apply and record charges
                        $quoteChargeRate{$value} = $rateExpression;
                        $chargeRateApplied++;

                        my $scalingExpression = '';
                        if ($effectivePropertyValuePart =~
                            /^\d*\.?\d+$/)    # Numeric property
                        {
                            $scalingFactor *= $effectivePropertyValuePart;
                            $scalingExpression .=
                              "$effectivePropertyValuePart [$name] * ";
                        }
                        $scalingFactor *= $amount;
                        $scalingExpression .= "$amount [ChargeRate{$name}";
                        $scalingExpression .= "{$value}"
                          if $value ne '';
                        $scalingExpression .= ']';

                        push @scalingExpressions, $scalingExpression;

                        last;
                    }
                    if ($effectivePropertyValuePart =~ /^[[:alpha:]]/
                        && ! $chargeRateApplied)
                    {
                        $message .=
                          "\nWarning -- Charge rate for $name $propertyValuePart is not defined";
                    }
                    delete $chargeRate{$name} if $isMultiplier;
                }
            }

            # Specified charge rates override matched standard rates
            foreach my $value (keys %{$specifiedChargeRate{$name}})
            {
                $quoteChargeRate{$value} = $specifiedChargeRate{$name}{$value}
                  if $specifiedChargeRate{$name}{$value} =~ /^\*([\d\.]+)$/;
            }

            # Create associated quote charge rates
            if (   $action eq 'Quote'
                && $quoteId ne ''
                && $guarantee)
            {
                foreach my $value (keys %quoteChargeRate)
                {
                    # Create a quote charge rate
                    my $subRequest = new MAM::Request(
                        database    => $database,
                        object      => 'QuoteChargeRate',
                        action      => 'Create',
                        actor       => $SUPER_USER,
                        assignments => [
                            new MAM::Assignment(
                                name  => 'Quote',
                                value => $quoteId
                            ),
                            new MAM::Assignment(
                                name  => 'Name',
                                value => $name
                            ),
                            new MAM::Assignment(
                                name  => 'Value',
                                value => $value
                            ),
                            new MAM::Assignment(
                                name  => 'Amount',
                                value => $quoteChargeRate{$value}
                            )
                        ]
                    );
                    my $subResponse =
                      MAM::Base->create($subRequest, $requestId);
                    if ($subResponse->getStatus() eq 'Failure')
                    {
                        return new MAM::Response()->failure(
                            $subResponse->getCode(),
                            'Unable to create quote charge rate: '
                              . $subResponse->getMessage()
                        );
                    }
                }
            }
        }

        # Next iterate over the additive rates (preadd and postadd)
        foreach my $name (keys %chargeRate)
        {
            my %values          = %{$chargeRate{$name}};
            my %quoteChargeRate = ();

            # Look for this rate in the usage record properties
            if (defined $usageRecordProperties{$name})
            {
                my $propertyValue = $usageRecordProperties{$name};
                $propertyValue =~ s/^{(.*)}$/$1/;
                my $chargeRateApplied = 0;

              PART: foreach my $propertyValuePart (split /,/, $propertyValue)
                {
                    my ($propertyCoefficient, $consumptionRate);
                    if ($propertyValuePart =~
                        /^([[:alpha:]][\w\.\-]*)\*(\d*\.?\d+)$/)
                    {
                        ($propertyValuePart, $propertyCoefficient) = ($1, $2);
                    }
                    elsif ($propertyValuePart =~ /^(\d*\.?\d+)\*(\d*\.?\d+)$/)
                    {
                        ($propertyValuePart, $consumptionRate) = ($1, $2);
                    }
                    my $effectivePropertyValuePart = $propertyValuePart;

                    # Iterate through charge rate values
                    # Default (null) values should sort last
                  VALUE: foreach my $value (
                        sort { ($a || '~') cmp($b || '~') }
                        keys %values
                      )
                    {
                        my $rateExpression  = $values{$value};
                        my $valueExpression = $value;

                        unless ($rateExpression =~
                            /^(?:([\d\.]+)\+)$|^(?:\+?([\d\.]+))(?:\/([1-9]\d*))?(?:\/([smhdWMY]))?$/
                          )
                        {
                            $message .=
                              "\nWarning -- Invalid rate expression ($rateExpression) for charge rate $name";
                            next VALUE;
                        }
                        my ($postAdd, $preAdd, $divisor, $timeUnit) =
                          ($1, $2, $3, $4, $5);
                        my $amount = defined $preAdd ? $preAdd : $postAdd;
                        $amount /= $divisor if defined $divisor;
                        if (defined $timeUnit)
                        {
                            if    ($timeUnit eq 's') { $amount /= 1; }
                            elsif ($timeUnit eq 'm') { $amount /= 60; }
                            elsif ($timeUnit eq 'h') { $amount /= 3600; }
                            elsif ($timeUnit eq 'd') { $amount /= 86400; }
                            elsif ($timeUnit eq 'w') { $amount /= 604800; }
                            elsif ($timeUnit eq 'M') { $amount /= 2629744; }
                            elsif ($timeUnit eq 'Y') { $amount /= 31556926; }
                        }

                        # Parse out and handle conditions
                        if ($valueExpression =~ s/^(.*)\?//)
                        {
                            my $valueConditions = $1;
                            my ($conditionsMatched, $conditionsMissed) = (0, 0);
                            foreach my $valueCondition (split /[\|\&]/,
                                $valueConditions)
                            {
                                if ($valueCondition =~ /^(\w+)=(\S+)$/)
                                {
                                    my ($conditionName, $conditionValue) =
                                      ($1, $2);
                                    if (defined
                                        $usageRecordProperties{$conditionName}
                                        && $usageRecordProperties{
                                            $conditionName} eq $conditionValue)
                                    {
                                        $conditionsMatched++;
                                    }
                                    else
                                    {
                                        $conditionsMissed++;
                                    }
                                }
                                else
                                {
                                    $message .=
                                      "\nWarning -- Invalid value condition ($valueCondition) for charge rate $name";
                                    next VALUE;
                                }
                            }
                            if ($valueConditions =~ /\|/)
                            {
                                next unless $conditionsMatched;
                            }
                            else
                            {
                                next if $conditionsMissed;
                            }
                        }

                        # Continue unless value expression matches

                        # Numeric-valued charge rates
                        if (   $valueExpression !~ /^[[:alpha:]]/
                            && $propertyValuePart =~ /^\d*\.?\d+$/)
                        {
                            my $matches = 0;
                            foreach my $valueRange (
                                length $valueExpression
                                ? (split /,/, $valueExpression, -1)
                                : ('')
                              )
                            {
                                if ($valueRange =~
                                    /^([\d\.]+)?(?:(-|>=?|=?[<>]=?)([\d\.]+))?$/
                                  )
                                {
                                    my ($leftBound, $comparison, $rightBound) =
                                      ($1, $2, $3);
                                    my $expression;
                                    # a-b or a[=]<[=]b
                                    if (   defined $leftBound
                                        && defined $rightBound
                                        && defined $comparison
                                        && $comparison !~ />/)
                                    {
                                        my $leftOperator =
                                          (      $comparison eq '-'
                                              || $comparison =~ /^=</)
                                          ? '>='
                                          : '>';
                                        my $rightOperator =
                                          (      $comparison eq '-'
                                              || $comparison =~ /<=$/)
                                          ? '<='
                                          : '<';
                                        $expression =
                                          "$propertyValuePart $leftOperator $leftBound && $propertyValuePart $rightOperator $rightBound";
                                    }
                                    # (>|<)[=]b
                                    elsif (defined $rightBound
                                        && defined $comparison
                                        && $comparison !~ /^=/
                                        && ! defined $leftBound)
                                    {
                                        $expression =
                                          "$propertyValuePart $comparison $rightBound";
                                    }
                                    # a
                                    elsif (defined $leftBound
                                        && ! defined $rightBound
                                        && ! defined $comparison)
                                    {
                                        $expression =
                                          "$propertyValuePart == $leftBound";
                                    }
                                    # null
                                    elsif (! defined $leftBound
                                        && ! defined $rightBound
                                        && ! defined $comparison)
                                    {
                                        $matches = 1;
                                        last;
                                    }
                                    # Invalid
                                    else
                                    {
                                        $message .=
                                          "\nWarning -- Invalid value expression ($valueRange) for standard charge rate $name";
                                        next;
                                    }
                                    # Test the expression
                                    if (eval $expression)
                                    {
                                        $matches = 1;
                                        last;
                                    }
                                }
                                else
                                {
                                    $message .=
                                      "\nWarning -- Invalid value expression ($valueRange) for standard charge rate $name";
                                    next;
                                }
                            }
                            next unless $matches;
                        }

                        # Complex charge rates
                        elsif ($valueExpression =~ /^[[:alpha:]]/
                            && $propertyValuePart =~ /:/)
                        {
                            my $matches = 0;
                            $propertyValuePart =~ s/\"//g;

                            # name-valued complex charge rate
                            if ($propertyValuePart eq $valueExpression)
                            {
                                $matches = 1;
                            }
                            # numeric-valued complex charge rate
                            elsif ($propertyValuePart =~ /([^:]+):(.*)/)
                            {
                                my ($partName, $partValue) = ($1, $2);
                                if ($partName eq $valueExpression)
                                {
                                    if ($partValue =~ /^\d*\.?\d+$/)
                                    {
                                        $matches = 1;
                                        $effectivePropertyValuePart =
                                          $partValue;
                                    }
                                    else
                                    {
                                        $message .=
                                          "\nWarning -- Non-numeric part value ($partValue) for $partName in complex property $name";
                                        next;
                                    }
                                }
                            }
                            next unless $matches;
                        }

                        # Name-valued charge rates
                        elsif ($valueExpression =~ /^[[:alpha:]]|^$/
                            && $propertyValuePart =~ /^[[:alpha:]][\w\.\-]*$/)
                        {
                            my $matches = 0;
                            foreach my $altValue (
                                length $valueExpression
                                ? (split /,/, $valueExpression, -1)
                                : ('')
                              )
                            {
                                if (   $altValue eq ''
                                    || $altValue eq $propertyValuePart)
                                {
                                    $matches = 1;
                                    last;
                                }
                            }
                            next unless $matches;
                        }

                        # Unable to derive if name based or value based
                        else
                        {
                            $message .=
                              "\nWarning -- Unable to categorize the testing of $name $propertyValuePart against $valueExpression as name-valued or numeric-valued";
                            next;
                        }

                        # Apply and record charges
                        $quoteChargeRate{$value} = $rateExpression;
                        $chargeRateApplied++;

                        my @chargeAssignments = ();
                        my $datum;
                        if (   $chargeItemization
                            && $action eq 'Charge')
                        {
                            push @chargeAssignments,
                              ( new MAM::Assignment(
                                    name  => 'Name',
                                    value => $name
                                ),
                                new MAM::Assignment(
                                    name  => 'Value',
                                    value => $propertyValuePart
                                ),
                                new MAM::Assignment(
                                    name  => 'Rate',
                                    value => $rateExpression
                                )
                              );
                            push @chargeAssignments,
                              new MAM::Assignment(
                                name  => 'UsageRecord',
                                value => $usageRecordProperties{'Id'}
                              ) if exists $usageRecordProperties{'Id'};
                            push @chargeAssignments,
                              new MAM::Assignment(
                                name  => 'Instance',
                                value => $usageRecordProperties{'Instance'}
                              ) if exists $usageRecordProperties{'Instance'};
                        }
                        if ($itemize)
                        {
                            $datum = new MAM::Datum('Charge');
                            if (exists $usageRecordProperties{'Quote'}
                                && $action eq 'Quote')
                            {
                                $datum->setProperty('Quote',
                                    $usageRecordProperties{'Quote'});
                                $metadata{'Quote'} = 'Integer';
                            }
                            if (exists $usageRecordProperties{'Lien'}
                                && $action eq 'Reserve')
                            {
                                $datum->setProperty('Lien',
                                    $usageRecordProperties{'Lien'});
                                $metadata{'Lien'} = 'Integer';
                            }
                            if (exists $usageRecordProperties{'Id'})
                            {
                                $datum->setProperty('UsageRecord',
                                    $usageRecordProperties{'Id'});
                                $metadata{'UsageRecord'} = 'Integer';
                            }
                            if (exists $usageRecordProperties{'Instance'})
                            {
                                $datum->setProperty('Instance',
                                    $usageRecordProperties{'Instance'});
                                $metadata{'Instance'} = 'String';
                            }
                            $datum->setProperty('Name', $name);
                            $metadata{'Name'} = 'String';
                            $datum->setProperty('Value', $propertyValuePart);
                            $metadata{'Value'} = 'String';
                        }
                        my $chargeExpression = '';
                        my $chargeAmount     = $amount;
                        if ($effectivePropertyValuePart =~
                            /^\d*\.?\d+$/)    # Numeric property
                        {
                            $chargeAmount *= $effectivePropertyValuePart;
                            $chargeExpression .=
                              "$effectivePropertyValuePart [$name] * ";
                        }
                        $chargeExpression .= "$amount [ChargeRate{$name}";
                        $chargeExpression .= "{$value}"
                          if $value ne '';
                        $chargeExpression .= ']';
                        if (defined $propertyCoefficient
                          )    # Name-valued charge with coefficient
                        {
                            $chargeExpression .=
                              " * $propertyCoefficient [coefficient]";
                        }
                        if (defined $timeUnit)    # Resource charge rate
                        {
                            if (defined $consumptionRate)
                            {
                                $chargeExpression .=
                                  " * $consumptionRate [ConsumptionRate]";
                            }
                            $chargeAmount *= $duration;
                            $chargeExpression .= " * $duration [Duration]";
                            push @chargeAssignments,
                              new MAM::Assignment(
                                name  => 'Duration',
                                value => $duration
                              )
                              if ( $chargeItemization
                                && $action eq 'Charge');
                            if ($itemize)
                            {
                                $datum->setProperty('Duration', $duration);
                                $metadata{'Duration'} = 'Integer';
                            }
                        }
                        else
                        {
                            if ($itemize)
                            {
                                $datum->setProperty('Duration', '');
                                $metadata{'Duration'} = 'Integer';
                            }
                        }
                        if ($itemize)
                        {
                            $datum->setProperty('Rate', $amount);
                            $metadata{'Rate'} = 'String';
                        }
                        my $scalingRate;
                        if (defined $preAdd)    # Factor in multipliers
                        {
                            $scalingRate = $scalingFactor;
                            $scalingRate *= $consumptionRate
                              if defined $consumptionRate;
                            $scalingRate *= $propertyCoefficient
                              if defined $propertyCoefficient;
                        }
                        else
                        {
                            $scalingRate = $propertyCoefficient
                              if defined $propertyCoefficient;
                        }
                        if (defined $scalingRate)
                        {
                            $chargeAmount *= $scalingRate
                              if $scalingRate != 1;
                            $chargeExpression .=
                              ' * ' . join(' * ', @scalingExpressions)
                              if @scalingExpressions;
                            push @chargeAssignments,
                              new MAM::Assignment(
                                name  => 'ScalingFactor',
                                value => $scalingRate
                              )
                              if ( $chargeItemization
                                && $action eq 'Charge');
                            if ($itemize)
                            {
                                $datum->setProperty('ScalingFactor',
                                    $scalingRate);
                                $metadata{'ScalingFactor'} = 'Float';
                            }
                        }
                        else
                        {
                            if ($itemize)
                            {
                                $datum->setProperty('ScalingFactor', '');
                                $metadata{'ScalingFactor'} = 'Float';
                            }
                        }

                        # Round the charge amount to the currency precision
                        $chargeAmount =
                          sprintf("%.${currency_precision}f", $chargeAmount);
                        push @chargeAssignments,
                          new MAM::Assignment(
                            name  => 'Amount',
                            value => $chargeAmount
                          )
                          if ( $chargeItemization
                            && $action eq 'Charge');
                        $charge += $chargeAmount;

                        push @chargeExpressions, $chargeExpression;
                        push @chargeAssignments,
                          new MAM::Assignment(
                            name  => 'Details',
                            value => $chargeExpression
                          )
                          if ( $chargeItemization
                            && $action eq 'Charge');

                        if ($itemize)
                        {
                            $datum->setProperty('Amount',  $chargeAmount);
                            $datum->setProperty('Details', $chargeExpression);
                            $metadata{'Amount'}  = 'Currency';
                            $metadata{'Details'} = 'String';
                            push @data, $datum;
                        }

                        # Record itemized charges if needed
                        if (   $chargeItemization
                            && $action eq 'Charge')
                        {
                            # Create a charge
                            my $subRequest = new MAM::Request(
                                database    => $database,
                                object      => 'Charge',
                                action      => 'Create',
                                actor       => $SUPER_USER,
                                assignments => \@chargeAssignments,
                            );
                            my $subResponse =
                              MAM::Base->create($subRequest, $requestId);
                            if ($subResponse->getStatus() eq 'Failure')
                            {
                                return new MAM::Response()->failure(
                                    $subResponse->getCode(),
                                    'Unable to create charge record: '
                                      . $subResponse->getMessage()
                                );
                            }
                        }

                        last;
                    }
                    if ($effectivePropertyValuePart =~ /^[[:alpha:]]/
                        && ! $chargeRateApplied)
                    {
                        $message .=
                          "\nWarning -- Charge rate for $name $propertyValuePart is not defined";
                    }
                }
            }

            # Specified charge rates override matched standard rates
            foreach my $value (keys %{$specifiedChargeRate{$name}})
            {
                $quoteChargeRate{$value} = $specifiedChargeRate{$name}{$value}
                  unless $specifiedChargeRate{$name}{$value} =~ /^\*([\d\.]+)$/;
            }

            # Create associated quote charge rates
            if (   $action eq 'Quote'
                && $quoteId ne ''
                && $guarantee)
            {
                foreach my $value (keys %quoteChargeRate)
                {
                    # Create a quote charge rate
                    my $subRequest = new MAM::Request(
                        database    => $database,
                        object      => 'QuoteChargeRate',
                        action      => 'Create',
                        actor       => $SUPER_USER,
                        assignments => [
                            new MAM::Assignment(
                                name  => 'Quote',
                                value => $quoteId
                            ),
                            new MAM::Assignment(
                                name  => 'Name',
                                value => $name
                            ),
                            new MAM::Assignment(
                                name  => 'Value',
                                value => $value
                            ),
                            new MAM::Assignment(
                                name  => 'Amount',
                                value => $quoteChargeRate{$value}
                            )
                        ]
                    );
                    my $subResponse =
                      MAM::Base->create($subRequest, $requestId);
                    if ($subResponse->getStatus() eq 'Failure')
                    {
                        return new MAM::Response()->failure(
                            $subResponse->getCode(),
                            'Unable to create quote charge rate: '
                              . $subResponse->getMessage()
                        );
                    }
                }
            }
        }

        $itemizedCharges = '';
        if (@chargeExpressions)
        {
            $itemizedCharges .= join(' + ', @chargeExpressions);
        }
        else { $itemizedCharges .= '0 [ConsumedResources]'; }
        $itemizedCharges .= " = $charge";
    }

    # If Charge is not included in usage record properties,
    # look for itemized charges in an option
    else
    {
        if (! defined $itemizedCharges)
        {
            $itemizedCharges = "$charge [externally calculated] = $charge";
        }
    }

    # Build up charge datum if not itemized or no chargeable resources
    unless (@data)
    {
        my $datum = new MAM::Datum('Charge');
        if (exists $usageRecordProperties{'Quote'}
            && $action eq 'Quote')
        {
            $datum->setProperty('Quote', $usageRecordProperties{'Quote'});
            $metadata{'Quote'} = 'Integer';
        }
        if (exists $usageRecordProperties{'Lien'}
            && $action eq 'Reserve')
        {
            $datum->setProperty('Lien', $usageRecordProperties{'Lien'});
            $metadata{'Lien'} = 'Integer';
        }
        if (exists $usageRecordProperties{'Id'})
        {
            $datum->setProperty('UsageRecord', $usageRecordProperties{'Id'});
            $metadata{'UsageRecord'} = 'Integer';
        }
        if (exists $usageRecordProperties{'Instance'})
        {
            $datum->setProperty('Instance', $usageRecordProperties{'Instance'});
            $metadata{'Instance'} = 'String';
        }
        $datum->setProperty('Amount', $charge);
        $metadata{'Amount'} = 'Currency';
        push @data, $datum;
    }

    # Rereference the read-write parameters for visibility outside of subroutine
    $$charge_ref          = $charge;
    $$itemizedCharges_ref = $itemizedCharges;
    $$message_ref         = $message;
    @$data_ref            = @data;
    %$metadata_ref        = %metadata;

    return ($charge, $itemizedCharges, $message, \@data);
}

1;
